/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/**********************************************************************************
***	dbmalloc.c	a simple to use malloc debugger				***
***********************************************************************************
***										***
***	by		Markus Eichinger and Sebastian Albrecht			***
***			Institut of Medical Optics, Oettingenstrasse 67		***
***			80538 Muenchen, Germany, August, 1998			***
***										***
***	1. Checks of free(..) and realloc(...) with invalid pointers.		***
***	2. Checks boundary integrity of each allocated array.			***
***	3. Statistic about number of allocated buffers and space.		***
***	4. In case of an error the line-number and file-name is printed out	***
***	   where the error did occur.						***
***										***
***	How to use it:								***
***	1. a.) Include dbmalloc.h with #include "dbmalloc.h" in each file of 	***
***	       your source code in which a malloc(...), calloc(...),		***
***	       realloc(...), free(...) or strdup(...) statement is called.	***
***	       You must include it after any include statements of		***
***	       stdio.h, malloc.h and string.h.					***	
***	   b.) Include ini_mallocdebug(...),testmalloc(...) and 		***
***	       statisticmalloc(...) in your program.				***
***	   c.) Recompile your source code including dbmalloc.c and link it.	***
***										***
***	2. ini_mallocdebug(int ininow, int dblevel,int patmode,			***
***			   int presize,int postsize,int fillbyte,int nodeid)	***
***		Call this initialisation function before any allocs are called	***
***		by your program.                                                ***
***		int ininow:	0 -->	do not initialize buffer list. Should 	***
***					only be used if you like to change the	***
***					debug level dblevel.			***
***				1 -->	do initialize the buffer list. Use this ***
***					always first time you call this function***
***		int dblevel:	0 -->	Create buffer list. Test buffer integr.	***
***					only with function testmalloc(...).	***
***				1 -->	Test buffer integrity also each time 	***
***					before any alloc/free function is called***
***				2 -->	Test also for alloc size == 0		***
***				3 -->	Print also info about each alloc/free	***
***					call (note that this will slow down your***
***					program if you use malloc/free often.	***
***										*&*
***				4 -->	Print even more info			***
***										***
***		int patmode:		before and after each allocated buffer	***
***					an extra space before (PRE) and after	***
***					(POST) each array is allocated. This	***
***					PRE and POST space is filled with a	***
***					certain bit pattern. If the bit pattern	***
***					is destroyed during programm execution	***
***					it is very likely that your program	***
***					has an malloc-problem with that buffer.	***
***				0 -->	bit pattern is 0x00 for each byte	***
***				1 -->	bit pattern is 0xff			***
***				2 -->	bit pattern is random			***
***				3 -->	bit pattern is 0x00ff			***
***				4 -->	bit pattern is 0xff00			***
***			       -1 -->	do not change value of patmode.		***
***		int presize:	8,16,24,32,40					***
***		int postsize:	8,16,24,32,40					***
***					length of the extra space in bytes.	***
***					recomended values are 8,16,24,32,40.	***
***					all other values will lead to unaligned	***
***					buffers which will slow down your progr.***
***				-1 -->	do not change value of per- or postsize.***
***		int fillbyte:	-1 -->	Do not initialize malloc/strdup buffers	***
***			    0..255 -->	Initialize with given byte.		***
***		int nodeid:	   -->  Any integer number which you like to	***
***				   	specify the node-id for parallel progr.	***
***										***
***		Do not change patmode,presize,postsize after initialisation.	***
***		You may change however the debug-level and the fillbyte		***
***		any time after initialization with				***
***		with ini_mallocdebug(0,newlevel,-1,-1,-1,0,0)			***
***										***
***	3. testmalloc(char *infostr,int mode)					***
***		You can call this function anywhere in your program. This 	***
***		function will test the integrity of the extra space for each	***
***		allocated buffer. If the integrity is destroyed an error message***
***		and a lot of helpful information including the infostring will  ***
***		be printed. Thus, the infostring helps to trace in which part 	***
***		of the program testmalloc failed.				***
***		int mode:	0 -->	do not exit if testmalloc(...) failed.	***
***					testmalloc(...) will return with -1 if	***
***					it failed, otherwise it returns with	***
***					the number of allocated bytes.		***
***					Now you have the possibility to trace	***
***					the contents of variables in your prog.	***
***					with the construction:			***
***					if (testmalloc(...)==-1) {		***
***						printf("Test: %d\n",i);		***
***						exit(-1);			***
***					}					***
***				1 -->	do exit if testmalloc(...) failed.	***
***										***
***	4. statisticmalloc(char *infostr, int mode)				***
***		This function will print information about the current		***
***		allocation status. It also test the buffer integrity.		***
***		int mode:	0 --> Print only summary statistic.		***
***				1 --> Print verbose buffer statistics.		***
***										***
**********************************************************************************/

#define INC_IN_DBMALLOC	/* flag used in dbmalloc.h */
#define MAXENTRY	1000
#define	MAXFREELIST	10

#include <stdio.h>
#include <string.h>

#if defined(NEXT) || defined(RS6K) || defined(SUN4) || defined(LINUX64) || defined(MAC)
#include <stdlib.h>
#elif defined(PARIX) || defined(PARIXPPC)
#include <sys/memory.h>
#elif defined(PARIXHAL)
#include </export/epx/epx/include/epx/memory.h>
#else
#include <malloc.h>
#endif

#include "dbmalloc.h"

#if DEBUGMALLOC
/**** zero bit muster *******/
static	unsigned char muster0[41]={
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

/**** 0xff bit muster *******/
static	unsigned char muster1[41]={
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff};

/**** arbitrary bit muster *******/
static	unsigned char muster2[41]={
	0x3f,0x12,0x7f,0xe3,0x6a,0x61,0xef,0x1f,
	0x2f,0x9a,0x02,0x34,0x88,0x33,0xee,0xf4,
	0xe7,0x41,0x93,0x5f,0xaa,0xb4,0x4b,0x65,
	0xa6,0x3f,0x02,0x34,0x88,0x9a,0x02,0x34,
	0x88,0x2f,0x9a,0x7f,0xe3,0x6a,0x88,0x9a,0x9a};

static	unsigned char muster3[41]={
	0x00,0xff,0x00,0xff,0x00,0x00,0xff,0x00,
	0x00,0xff,0x00,0xff,0x00,0x00,0xff,0x00,
	0x00,0xff,0x00,0xff,0x00,0x00,0xff,0x00,
	0x00,0xff,0x00,0xff,0x00,0x00,0xff,0x00,
	0x00,0xff,0x00,0xff,0x00,0x00,0xff,0x00,0x00};

static	unsigned char muster4[41]={
	0xff,0x00,0xff,0x00,0x00,0xff,0x00,0xff,
	0xff,0x00,0xff,0x00,0x00,0xff,0x00,0xff,
	0xff,0x00,0xff,0x00,0x00,0xff,0x00,0xff,
	0xff,0x00,0xff,0x00,0x00,0xff,0x00,0xff,
	0xff,0x00,0xff,0x00,0x00,0xff,0x00,0xff,0x00};

static	char	tstr[2][10] = {"SHIT"," OK "};
static	char	idstr[5][10]= {"calloc ","malloc ","realloc","strdup ","free   "};
static	int	mustermode=0;

#endif

static	unsigned char	curbitmuster[41];

static	int	debuglevel=0;
static	int	presize=16;
static	int	postsize=16;

static	unsigned char* preptr[MAXENTRY];
static	unsigned char* postptr[MAXENTRY];
static	unsigned char* bufptr[MAXENTRY];
static	int	buflen[MAXENTRY];
static	int	linenum[MAXENTRY];
static	char	filename[MAXENTRY][20];
static	int	idtag[MAXENTRY];

static	int	curentries=0;
static	int	lastentry=-1;

static	int	freelist[MAXFREELIST];
static	int	freelistind=-1;
static	int	fillbyte=-1;
static	int	nodeid=0;

static	char	tempstr[500];


int
ini_mallocdebug(int ininow,int dblevel,int bitpattern,
		int pres,int posts,int fillb,int a)
{
#if DEBUGMALLOC
	debuglevel=dblevel;
	fillbyte  =fillb;
	nodeid    = a;
	
	if (ininow) {
		int i;
		curentries = 0;
		freelistind=-1;
		for (i=0;i<MAXENTRY;i++) {
			preptr[i] = (unsigned char*) NULL;
			postptr[i]= (unsigned char*) NULL;
			bufptr[i] = (unsigned char*) NULL;
			buflen[i] = 0;
			linenum[i]= 0;
			idtag[i]  =-1;
			filename[i][0] = 0;
		}
	}
	
	if (pres!=-1)	presize=pres;
	if (posts!=-1)	postsize=posts;

	if (bitpattern!=-1) {
		mustermode = bitpattern;
	
		if (mustermode==0)	memcpy(curbitmuster,muster0,40);
		else
		if (mustermode==1)	memcpy(curbitmuster,muster1,40);
		else	
		if (mustermode==2)	memcpy(curbitmuster,muster2,40);
		else	
		if (mustermode==3)	memcpy(curbitmuster,muster3,40);
		else	
		if (mustermode==4)	memcpy(curbitmuster,muster4,40);
		else			memcpy(curbitmuster,muster2,40);
	}

#endif
	return 0;	
}

static	int	strdupflag=0;

void*
debugcalloc(int myline,char *myfile,size_t numobj,size_t sobj)
{
	int i,flen=0,len=0,offset=0;
	size_t internlen=0;

	if (sobj==0)	len=numobj;
	else 		len=numobj*sobj;

	flen=strlen(myfile);
	if (flen<18)	offset=0;
	else		offset=flen-18;
	
	if (debuglevel>=1) {
		sprintf(tempstr,"DBM-test before alloc in L=%d <%s>",
			myline,myfile+offset);
		testmalloc(tempstr,1);
	}

	if ((debuglevel>=2)&&(len<=0)) {
		printf("DBM(%2d): WARNING: alloc-request with len=%d in line=%d in file <%s>\n",
			nodeid,len,myline,myfile+offset);
	}
			
	/*** find empty space in list ***/
	if (freelistind>-1) {	/*** use the free list index stack ***/
		lastentry=freelist[freelistind];
		freelistind--;
		if (preptr[lastentry]!=(unsigned char*)NULL) { /*** really free? ***/
			printf("DBM(%2d): Fatal error in free list management in dbmalloc.c\n",
				nodeid);
			fflush(stdout);
			exit(-1);
		}
	} else {		/*** search the total list ***/
		for (i=0,lastentry=-1;i<MAXENTRY;i++) {
			if (preptr[i]==(unsigned char*)NULL) {
				lastentry=i;
				break;	
			}
		}
	}

	if ((lastentry<0)||(lastentry>=MAXENTRY)) {
		printf("DBM(%2d): Increase MAXENTRY(=%d) for dbmalloc.c!\n",
			MAXENTRY,nodeid);
		fflush(stdout);
		exit(-1);	
	}

	/*** now we have an empty entry ***/	
	internlen = len+presize+postsize;
	if (debuglevel>=3) {
		if (sobj!=0) 
			printf("DBM(%2d) %4d: calloc  : S=%7d L=%5d <%s> ",
				nodeid,lastentry,len,myline,myfile+offset);
		else
		if (strdupflag==1)
			printf("DBM(%2d) %4d: strdup  : S=%7d L=%5d <%s> ",
				nodeid,lastentry,len,myline,myfile+offset);
		else			
			printf("DBM(%2d) %4d: malloc  : S=%7d L=%5d <%s> ",
				nodeid,lastentry,len,myline,myfile+offset);
		fflush(stdout);
	}
		
	/*** do the malloc ***/
	preptr[lastentry] = (unsigned char*) malloc(internlen);
	if (preptr[lastentry]==(unsigned char*)NULL) {
		printf("DBM(%2d): Not enough space available detected in dbmalloc.c\n",
			nodeid);
		printf("DBM(%2d): Called from line %d in file <%s>\n",
			nodeid,myline,myfile+offset);
		printf("DBM(%2d): You tried to allocate %d (%d) bytes!\n",
			nodeid,len,(int) internlen);
		fflush(stdout);
		exit(-1);	
	}
	if (sobj!=0) {
		idtag[lastentry]=0;
		memset(preptr[lastentry],0,internlen);
	} else {
		if (strdupflag==1)	idtag[lastentry]=3;
		else			idtag[lastentry]=1;

		if (fillbyte!=-1) memset(preptr[lastentry],fillbyte,internlen);
	}

	/*** do the entry bookkeeping ***/
	bufptr[lastentry] = preptr[lastentry]+presize;
	postptr[lastentry]= bufptr[lastentry]+len;
	buflen[lastentry] = len;
	linenum[lastentry]= myline;
	strcpy(filename[lastentry],myfile+offset);

	/*** fill with bits ***/
	memcpy(preptr[lastentry] ,curbitmuster,presize );
	memcpy(postptr[lastentry],curbitmuster,postsize);
	
	curentries++;

	if (debuglevel>=3) {
		printf("ptr=%p\n",bufptr[lastentry]);
		fflush(stdout);
	}
	return (void*) bufptr[lastentry];
}

void*
debugstrdup(int myline,char *myfile,const char *str)
{
	int	len=strlen(str);
	char *ptr;
	
	strdupflag=1;
	ptr=(char*)debugcalloc(myline,myfile,len+1,0);
	strdupflag=0;

	strcpy(ptr,str);

	return (void*) ptr;
}

void*
debugrealloc(int myline,char *myfile,void *oldp,size_t newlen)
{
	int i,flen,offset=0;
	unsigned char *newptr;

	flen=strlen(myfile);
	if (flen<18)	offset=0;
	else		offset=flen-18;

	if (debuglevel>=1) {
		sprintf(tempstr,"DBM-test before realloc in L=%d <%s>",
			myline,myfile+offset);
		testmalloc(tempstr,1);
	}	

	if ((debuglevel>=2)&&(newlen<=0)) {
		printf("DBM(%2d): WARNING: realloc-request with len=%d in line=%d in file <%s>\n",
			nodeid,(int) newlen,myline,myfile+offset);
		fflush(stdout);
	}

	for (i=0;i<MAXENTRY;i++) {
		if (preptr[i]==(unsigned char*)NULL) continue;
		
		if (bufptr[i]==(unsigned char*)oldp) {
			if (debuglevel>=3) {
				printf("DBM(%2d) %4d: realloc : S=%7d L=%5d <%s> OldPtr=%p OldS=%7d ",
					nodeid,lastentry,(int) newlen,myline,myfile+offset,oldp,buflen[i]);
				fflush(stdout);	
			}
			newptr=(unsigned char*)realloc(preptr[i],newlen+presize+postsize);
			if (newptr==(unsigned char*)NULL) {
				printf("DBM(%2d): Not enough space detected in realloc.\n",
					nodeid);
				printf("DBM(%2d): Called from line %d in file <%s>.",
					nodeid,myline,myfile+offset);
				printf("DBM(%2d): You tried to allocate %d (%d) bytes!\n",
					nodeid,(int) newlen,(int) (newlen+presize+postsize));
				fflush(stdout);
				exit(-1);	
			}
			preptr[i] = (unsigned char*) newptr;
			bufptr[i] = preptr[i]+presize;
			postptr[i]= bufptr[i]+newlen;
			buflen[i] = newlen;
			linenum[i]= myline;
			strcpy(filename[i],myfile+offset);
			
			idtag[i]=2;
	
			/*** fill with bits ***/
			memcpy(preptr[i] ,curbitmuster,presize );
			memcpy(postptr[i],curbitmuster,postsize);
			
			if (debuglevel>=3) {
				printf("ptr=%p\n",bufptr[i]);
				fflush(stdout);
			}
			return (void*) bufptr[i];	
		}		
	}
	
	printf("DBM(%2d): Fatal error in realloc(%p,%d)\n",nodeid,oldp,(int) newlen);
	printf("DBM(%2d): You did try to realloc a buffer which has not been allocated.\n",nodeid);
	printf("DBM(%2d): Realloc was called from line %d in file <%s>\n",
		nodeid,myline,myfile);
	fflush(stdout);

	statisticmalloc("Called from realloc",1);
	
	exit(-1);
	return (void*)NULL;
}

int
debugfree(int myline,char *myfile, void* ptr)
{
	int i,flen,offset=0;

	flen=strlen(myfile);
	if (flen<18)	offset=0;
	else		offset=flen-18;
	
	if (debuglevel>=1) {
		sprintf(tempstr,"DBM-test before free in L=%d <%s>",
			myline,myfile+offset);
		testmalloc(tempstr,1);
	}

	for (i=0;i<MAXENTRY;i++) {
		if (preptr[i]==(unsigned char*)NULL) continue;
		
		if (bufptr[i]==ptr) {	/*** delete entry ***/
			if (debuglevel>=3) {
				printf("DBM(%2d) %4d: free    : S=%7d L=%5d <%s> ptr=%p ",
					nodeid,i,buflen[i],myline,myfile+offset,bufptr[i]);
				fflush(stdout);
			}

			free(preptr[i]);
			preptr[i] = (unsigned char*) NULL;
			postptr[i]= (unsigned char*) NULL;
			bufptr[i] = (unsigned char*) NULL;
			buflen[i] = 0;
			linenum[i]= 0;
			filename[i][0]=0;
			if (debuglevel>=3) {
				printf("ok.\n");
				fflush(stdout);
			}

			/*** manage the free list ***/
			freelistind++;
			if (freelistind<MAXFREELIST) {
				freelist[freelistind]=i;
			} else freelistind=MAXFREELIST-1;

			curentries--;
			return 0;				
		}		
	}
	
	/*** no entry found,... thats an error ***/
	printf("DBM(%2d): Fatal error in debugfree(%p)\n",nodeid,ptr);
	printf("DBM(%2d): You did try to free a buffer which has not been allocated.\n",nodeid);
	printf("DBM(%2d): Free was called from line %d in file <%s>\n",nodeid,myline,myfile+offset);
	fflush(stdout);

	statisticmalloc("Called from free",1);

	exit(-1);
	return 0;
}

int
statisticmalloc(char *infostr,int mode)
{
#if DEBUGMALLOC
	int i,okflagpre=0,okflagpost=0;
	int allocatedlen=0;
	int reallen=0,oneshit=0;
	int internmode=0;

	internmode=mode;

	printf("DBM(%2d): Current malloc statistics: <%s>\n",nodeid,infostr);

do_statistic_now:
	allocatedlen=0;
	reallen=0;
	oneshit=0;	
	okflagpre=0;
	okflagpost=0;

	for (i=0;i<MAXENTRY;i++) {
		
		if (preptr[i]==(unsigned char*)NULL) continue;

		if (memcmp(preptr[i] ,curbitmuster,presize )==0) okflagpre =1;
		else						 okflagpre =0;
		
		if (memcmp(postptr[i],curbitmuster,postsize)==0) okflagpost=1;
		else						 okflagpost=0;

		if ((okflagpre==0)||(okflagpost==0)) oneshit=1;
	
		if (internmode==1) {
			printf("DBM(%2d) %4d: %s : len=%8d  ptr=%p  PRE=<%s> POST=<%s> L=%6d <%s>\n",
				nodeid,i,idstr[idtag[i]],
				buflen[i],bufptr[i],
				tstr[okflagpre],tstr[okflagpost],
				linenum[i],filename[i]);	
			fflush(stdout);
		}
		
		allocatedlen+=buflen[i];
		reallen     +=buflen[i]+presize+postsize;
	}
	
	printf("DBM(%2d): open buffers         : %9d\n",nodeid,curentries);
	printf("DBM(%2d): allocated len        : %9d bytes\n",nodeid,allocatedlen);
	printf("DBM(%2d): really allocated len : %9d bytes\n",nodeid,reallen);	
	fflush(stdout);
		
	if (oneshit) {
		printf("DBM: ******************************************\n");
		printf("DBM: *** At least one buffer is corrupted ! ***\n");
		printf("DBM: ******************************************\n");	
		if (internmode==0) {
			internmode=1;
			printf("DBM(%2d): Automatic verbose printout follows:\n",
				nodeid);	
			goto do_statistic_now;
		}
		return -1;
	}	
	fflush(stdout);
#endif
	return 0;
}

int
testmalloc(char *infostr, int mode)
{
#if DEBUGMALLOC
	int i;
	
	if (debuglevel>=4) {
		printf("DBM(%2d): testmalloc : <%s> ",nodeid,infostr);
		fflush(stdout);
	}
	
	for (i=0;i<MAXENTRY;i++) {
		if (preptr[i]==(unsigned char*)NULL) continue;

		if (memcmp(preptr[i] ,curbitmuster,presize )!=0) {
			printf("DBM(%2d): Fatal error in PRE-area in testmalloc <%s>\n",
				nodeid,infostr);
			printf("DBM(%2d): for buffer entry %d at ptr=%p with len=%d bytes\n",
				nodeid,i,bufptr[i],buflen[i]);
			printf("DBM(%2d): allocated with %s in line %d in file <%s>\n",
				nodeid,idstr[idtag[i]],linenum[i],filename[i]);
						
			fflush(stdout);
			
			statisticmalloc("Called from testmalloc",1);
			
			if (mode==0)	return -1;
			else		exit(-1);		
		}
		
		if (memcmp(postptr[i],curbitmuster,postsize)!=0) {
			printf("DBM(%2d): Fatal error in POST-area in testmalloc <%s>\n",
				nodeid,infostr);
			printf("DBM(%2d): for buffer entry %d at ptr=%p with len=%d bytes\n",
				nodeid,i,bufptr[i],buflen[i]);
			printf("DBM(%2d): allocated with %s in line %d in file <%s>\n",
				nodeid,idstr[idtag[i]],linenum[i],filename[i]);
							
			fflush(stdout);
			
			statisticmalloc("Called from testmalloc",1);
			
			if (mode==0)	return -1;
			else		exit(-1);
		}
	}

	if (debuglevel>=4) {
		printf(" ok\n");
		fflush(stdout);
	}
#endif
	return 0;	
}
