/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: eganaly.c	by M.Eichinger, Munich			**
**	Belongs2: EGO							**
**	Contents: Energy output to screen and file,			**	
**		  coordinate output to file,				**
**		  saving of restart file.				**
**		  doing the system-call.				**		
**		  Writing out forces.					**
*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>

#include "ego.h"
#include "eganaly.h"
#include "egloader.h"
#include "util_lib.h"
#include "version.h"
#include "egbound.h"
#include "egpressure.h"
#include "egqminter.h"

#define DEBUG		0

#if defined(NEXT)
#define FPRINTF_ERROR	0 /* no error returns 0 */
#else
#define FPRINTF_ERROR	1 /* no error returns number of characters printed */
#endif

static void	writeout_energy_header(LOAD_MSG_ALL_NODE *global_data, FILE *to_file);

static int	ene2file_newstyle(ENERGY_INFO *energy,
				LOAD_MSG_ALL_NODE *global_data,
				double temperature,
				double tot,double time);

static int	maybe_perform_systemcall(LOAD_MSG_ALL_NODE *global_data,
					 int curfilenum);
					
static int	writeout_force_type(	int force_type,
					char *file_name,
					LOAD_MSG_ALL_NODE *global_data,
					int *force_id,
					STRUCT_FORCE_INFO *finfo);

static void	writeout_pickbond(	char *file_name,
					LOAD_MSG_ALL_NODE *global_data,
					STRUCT_PICK_BOND *pinfo);

static void	writeout_pickangle(	char *file_name,
					LOAD_MSG_ALL_NODE *global_data,
					STRUCT_PICK_BOND *pinfo);

static void	writeout_pickdihe(	char *file_name,
					LOAD_MSG_ALL_NODE *global_data,
					STRUCT_PICK_BOND *pinfo);
	
static	FILE	*to_file;

static	int	cur_atoms_insortbuf=0;
static  int 	print_to_file=FALSE;
static	char	tmprestartfile[MAX_STRING_SIZE];

static	int 	time_accu_count=0 ,time_accu_count_tot=0;
static	double	time_accu_user=0.0,time_accu_tot_user=0.0;
static	double	time_accu_real=0.0,time_accu_tot_real=0.0;
static	FILE 	*r_file;
static	int	cycle_finished;
static	int	file_open=0;

/***************************************************************************/
void   initialize_analyze(void)
{
	time_accu_count     = 0;
	time_accu_count_tot = 0;
	time_accu_user      = 0.0;
	time_accu_tot_user  = 0.0;
	time_accu_real      = 0.0;
	time_accu_tot_real  = 0.0;
	
	cycle_finished = TRUE;
	file_open      = FALSE;		/* important the first time only */
	cur_atoms_insortbuf=0;
	print_to_file  = FALSE;
}

void   initialize_restart(void)
{
	r_file = NULL;
}


/****************************************************************************/
void close_output_file(	LOAD_MSG_ALL_NODE *global_data,
			int file_count)
{
	if (file_open) {
		fclose(to_file);
		file_open = FALSE;
		if (global_data!=NULL) 
			maybe_perform_systemcall(global_data,file_count);
#if DEBUG
txtout("in close_output_file: One EGO-outputfile completed.");
#endif		
	}
}

void trick_cur_atoms_insortbuf(int nr_atoms)
{
  cur_atoms_insortbuf = nr_atoms;
}

void sort_to_analybuffer(int new_clustering,
			TRAVELLING_COORDINATES *coords,
                        LOAD_MSG_ALL_NODE *global_data)
{
	int i;

	if (cur_atoms_insortbuf==global_data->total_no_atoms) 
		cur_atoms_insortbuf = 0;

	/*** do it in any case to atomdata[j].ko ***/
	for (i=0;i<coords->no_atoms;i++) {
		int j=global_data->new2old_map[i+global_data->first_atom[coords->node]];
		global_data->atomdata[j].ko[0] = coords->x[i];
		global_data->atomdata[j].ko[1] = coords->y[i];
		global_data->atomdata[j].ko[2] = coords->z[i];
	}

	/*** and only in case before new clustering  ***/
	/*** also to atomdata[j].vel	   	     ***/
	if (new_clustering==1) {
		for (i=0;i<coords->no_atoms;i++) {
			int j=global_data->new2old_map[i+global_data->first_atom[coords->node]];
			global_data->atomdata[j].vel[0] = coords->x[i];
			global_data->atomdata[j].vel[1] = coords->y[i];
			global_data->atomdata[j].vel[2] = coords->z[i];
		}
	}

	cur_atoms_insortbuf += coords->no_atoms;
	if (cur_atoms_insortbuf>global_data->total_no_atoms) {
		errout("Fatal error: in sort_to_analybuffer (eganalyze.c)\n");
		errout("Got illegal no of atoms %d!=%d!",
			cur_atoms_insortbuf,global_data->total_no_atoms);
		pex(-1);
	}

	if (global_data->rotation_mode) {
	  for (i=0;i<global_data->nr_of_rot_atoms;i++) {
	    if (global_data->rot_t1_x[i] != 0.0) {
	      global_data->rot1_x[i] = global_data->rot_t1_x[i];
	      global_data->rot1_y[i] = global_data->rot_t1_y[i];
	      global_data->rot1_z[i] = global_data->rot_t1_z[i];
	      global_data->rotation_torque[i]  = global_data->rotation_t_torque[i];
	    }
	    if (global_data->rot_t2_x[i] != 0.0) {
	      global_data->rot2_x[i] = global_data->rot_t2_x[i];
	      global_data->rot2_y[i] = global_data->rot_t2_y[i];
	      global_data->rot2_z[i] = global_data->rot_t2_z[i];
	    }
	  }
	}
}

/**************************************************************************/
/*** A EGO output file is created starting with the ctl.lis file	***/
/*** and the coordinate data.						***/
/**************************************************************************/	

static	char	file_name[MAX_STRING_SIZE];

void writeout_coordinates(	LOAD_MSG_ALL_NODE *global_data,
				double time_step,	
				int file_count,
				char *version_string)
{
	int	i,res;
#if DEBUG
txtout("in writeout_coordinates: filecount=%d  cycle_finished=%d\n",
file_count,cycle_finished);
#endif
 	if (cycle_finished) {
		/***************************************************/
		/*** put out some header-information		 ***/
		/***************************************************/

		/*** close old file and maybe perform systemcall ***/
		close_output_file(global_data,file_count);

 		generate_egofname(file_name,global_data->outpath,file_count);
#if DEBUG
txtout("in writeout_coordinates: now opening file <%s>\n",file_name);
#endif
      		if ((to_file=fopen(file_name,"w"))==NULL) {
			errout("Fatal error (%d) %s: Could not open EGO output file <%s>\n",
				errno,strerror(errno),file_name);
			pex(-1);
 		}
 		cycle_finished = FALSE;
 		file_open      = TRUE;

      		if (fprintf(to_file,"REMARK Outputfile of %s, %s \n",
			VERSION,EGO_MACHINE) < (5*FPRINTF_ERROR)) {
			errout("Fatal error: Could not write EGO header to file <%s>\n",
				file_name);
			pex(-1);
		}
      		fprintf(to_file,"REMARK %s\n",version_string);
		fprintf(to_file,"REMARK %s ===========================================\n",
				global_data->ctlfile);
		writeout_ctlfile(to_file,global_data);

		/*** write out Position of springs if pulling mode 3 ***/
		if (global_data->flag_pull==3) {
			res=fprintf(to_file,"[BEGINSPRINGCOORD]\n%f  %f  %f\n%f %f %f %f\n\n",
				global_data->pullspringpos[0],
				global_data->pullspringpos[1],
				global_data->pullspringpos[2],
				global_data->pullspringpos[3],
				global_data->pullspringpos[4],
				global_data->pullspringpos[5],
				global_data->pullspringpos[6]);
			if (res < (6 *FPRINTF_ERROR)) {
				errout("Fatal error: Could not write spring coordinates to file <%s>\n",
					file_name);
				pex(-1);
			}
		}

		res=fprintf(to_file,"[BEGINCOORD]\n%d  %d  %d  %f\n",
			global_data->writeout_nr_atoms,
			global_data->cycles_passed,
			global_data->Analyse_nth,
			time_step*1.0e15);
		if (res < (5 *FPRINTF_ERROR)) {
			errout("Fatal error: Could not write EGO header to file <%s>\n",
				file_name);
			pex(-1);
		}
   	}

	if (cur_atoms_insortbuf==global_data->total_no_atoms) {
		/***************************************************/
		/*** put out the coordinates and energy-header   ***/
		/***************************************************/
		for (i=0;i<global_data->total_no_atoms;i++) {
			if (global_data->writeout_atoms[i]!=0) {
				if (fprintf(to_file,"%# 10.7f %# 10.7f %# 10.7f\n",
					  global_data->atomdata[i].ko[0],
					  global_data->atomdata[i].ko[1],
					  global_data->atomdata[i].ko[2]) < (5*FPRINTF_ERROR)) {
					  
					errout("Fatal error: Could not write coordinates to file <%s>\n",
						file_name);
					pex(-1);
				}	
			}
		}
		
		writeout_energy_header(global_data, to_file);
	}
	
	return;
}

static void
writeout_energy_header(LOAD_MSG_ALL_NODE *global_data, FILE *to_file)
{
	int err=0;

	err|=fprintf(to_file,"[BEGINENERGY]\n");
	if (any_sbound_set(global_data->sbound_droff)) 
	  err|=fprintf(to_file,"%d  %d\n",NR_OF_COMMENT_LINES,NR_OF_ENERGY_ROWS+QMENERGIES);
	else
	  err|=fprintf(to_file,"%d  %d\n",2,NR_OF_ENERGY_ROWS+QMENERGIES);
	err|=fprintf(to_file,"[Clock]=seconds, [Temperature]=K, [Energies]=kcal/mol, [Pressures]=atm\n");
	err|=fprintf(to_file,"Step Clock Temperature Total Kinetic ");
	err|=fprintf(to_file,"Electrostatic VDW Bonded Angle Dihedral Improper ");
	err|=fprintf(to_file,"Restraint SBound HBond Flooding FloodStrength FloodDist ");
	err|=fprintf(to_file,"Immobilisation PullInDir PullOutDir PullInDir2 PullOutDir2 ");
	if (global_data->rotation_mode) {
	  err|=fprintf(to_file,"Rotforce1 Rotforce2 Rotforce3 Rotenergy ");
	}
	err|=fprintf(to_file,"QM-total QM-extern");
	err|=fprintf(to_file,"\n");
		
	if (any_sbound_set(global_data->sbound_droff)) 
	{ /* we write out these header lines ONLY if there are real values to follow them afterwards */
	  err|=fprintf(to_file,"PSphere PCyl ");
	  err|=fprintf(to_file,"PPlane-X PPlane+X PPlane-Y PPlane+Y PPlane-Z PPlane+Z ");
	  err|=fprintf(to_file,"PEdge-Y-Z PEdge+Y-Z PEdge-Y+Z PEdge+Y+Z PEdge-X-Z PEdge+X-Z ");
	  err|=fprintf(to_file,"PEdge-X+Z PEdge+X+Z PEdge-X-Y PEdge+X-Y PEdge-X+Y PEdge+X+Y ");
	  err|=fprintf(to_file,"PCorner--- PCorner+-- PCorner-+- PCorner++- PCorner--+ PCorner+-+ PCorner-++ PCorner+++ ");
	  err|=fprintf(to_file,"DSphere DCyl ");
	  err|=fprintf(to_file,"DPlane-X DPlane+X DPlane-Y DPlane+Y DPlane-Z DPlane+Z ");
	  err|=fprintf(to_file,"DEdge-Y-Z DEdge+Y-Z DEdge-Y+Z DEdge+Y+Z DEdge-X-Z DEdge+X-Z ");
	  err|=fprintf(to_file,"DEdge-X+Z DEdge+X+Z DEdge-X-Y DEdge+X-Y DEdge-X+Y DEdge+X+Y ");
	  err|=fprintf(to_file,"DCorner--- DCorner+-- DCorner-+- DCorner++- DCorner--+ DCorner+-+ DCorner-++ DCorner+++");
	  err|=fprintf(to_file,"\n");
  
	  err|=fprintf(to_file,"PVir11 PVir12 PVir13 PVir21 PVir22 PVir23 PVir31 PVir32 PVir33 PVir ");
	  err|=fprintf(to_file,"Pex11 Pex12 Pex13 Pex21 Pex22 Pex23 Pex31 Pex32 Pex33 Pex "); 
	  err|=fprintf(to_file,"Vir11 Vir12 Vir13 Vir21 Vir22 Vir23 Vir31 Vir32 Vir33 Vir ");
	  err|=fprintf(to_file,"Vex11 Vex12 Vex13 Vex21 Vex22 Vex23 Vex31 Vex32 Vex33 Vex ");
	  err|=fprintf(to_file,"Vke11 Vke12 Vke13 Vke21 Vke22 Vke23 Vke31 Vke32 Vke33 Vke ");
	  err|=fprintf(to_file,"Ekin11 Ekin12 Ekin13 Ekin21 Ekin22 Ekin23 Ekin31 Ekin32 Ekin33 Ekin");
	  err|=fprintf(to_file,"\n");                                               
  
	  if (err < (5*FPRINTF_ERROR)) {
		  errout("Could not write energy header to file <%s>\n",file_name);
		  pex(-1);
	  }
	}
	fflush(to_file);
	return;
}


void writeout_energy(	ENERGY_INFO *energy,
			double temperature, double temperature_h2o, double temperature_protein,
			double time_user,double time_real,
			LOAD_MSG_ALL_NODE *global_data,
			double time_step)
{
	double tot;
        char mminfo[50];
#if DEBUG
txtout("In writeout_energy: temperature %f fileopen=%d\n",
temperature,to_file);
#endif						
	/******************************************************************/
	/*** Now we have a very flexible output selection for energies: ***/
	/*** step_of_energy > 0  then print EVERY n-th step		***/
	/*** step_of_energy = 0  print no energies at all		***/
	/******************************************************************/
	print_to_file = (	
		(global_data->step_of_energy > 0) && 
		((global_data->cycles_passed%global_data->step_of_energy)==0)
		);

	pressure_update(global_data, energy); /* update all kind of pressure things:
			    * constant pressure/density mode, virial computation, etc. */
#if DEBUG
	txtout("In writeout_energy: after pressure_update\n");
#endif	
	tot  = energy->elec;		/*** sum up all energies ***/
	tot += energy->vdw;
	tot += energy->bond;
	tot += energy->angle;
	tot += energy->dihe;
	tot += energy->impr;
	tot += energy->kinetic;
	tot += energy->restraint;
	tot += energy->hbond;
	tot += energy->sbound;
	tot += energy->flooding;
	tot += energy->flooding;
	tot += energy->qm_energy[0] + energy->qm_energy[1] + global_data->qmenergyoffset;
	
 	time_accu_count++;
	time_accu_count_tot++;
 	time_accu_user      	+= time_user;
	time_accu_tot_user	+= time_user;
 	time_accu_real      	+= time_real;
	time_accu_tot_real	+= time_real;	
  
	/**********************************************************/
	/*** out to screen and set cycle.finished 		***/
	/**********************************************************/
	if (	(!cycle_finished)||
		((global_data->total_no_atoms>global_data->no_nodes*LIMIT_NO_ATOMS)&&
			print_to_file)) {

 		double femtosec=(double)global_data->cycles_passed*(time_step*1.0e15);
		double curtime;
		char   metric[5];

		if (femtosec<1000.0) {
			curtime = femtosec;
			strcpy(metric,"fs");
		} else
		if (femtosec<1000000.0) {
			curtime = femtosec/1000.0;
			strcpy(metric,"ps");
		} else {
			curtime = femtosec/1000000.0;
			strcpy(metric,"ns");
		}
	
		txtout("Center of mass (COM): %# 10.7g %# 10.7g %# 10.7g A.\n",
			energy->C[0],energy->C[1],energy->C[2]);

		txtout("Immob. deviation(1).: %# 10.7g %# 10.7g %# 10.7g A.\n",
			global_data->deltaimob[0],
			global_data->deltaimob[1],
			global_data->deltaimob[2]);

		if (global_data->flag_selimob>=2) {
		  txtout("Immob. deviation(2).: %# 10.7g %# 10.7g %# 10.7g A.\n",
			 global_data->deltaimob[3],
			 global_data->deltaimob[4],
			 global_data->deltaimob[5]);
		  if (global_data->flag_selimob==3) {
		    txtout("Immob. deviation(3).: %# 10.7g %# 10.7g %# 10.7g A.\n",
			   global_data->deltaimob[6],
			   global_data->deltaimob[7],
			   global_data->deltaimob[8]);
		  }
		}
		txtout("Ekin COM............: %# 10.7g %# 10.7g %# 10.7g kcal/mol.\n",
			0.5*energy->P[0]*energy->P[0]/global_data->total_mass,
			0.5*energy->P[1]*energy->P[1]/global_data->total_mass,
			0.5*energy->P[2]*energy->P[2]/global_data->total_mass);

		txtout("Angular momentum....: %# 10.7g %# 10.7g %# 10.7g AMU*A^2/fs.\n",
			energy->L[0]*time_step*1.0e15,
			energy->L[1]*time_step*1.0e15,
			energy->L[2]*time_step*1.0e15);

		print_pressure(&(energy->pressure),global_data->sbound_droff,
				PRESSURE_MODE);
		print_pressure(&(energy->density), global_data->sbound_droff,
				DENSITY_MODE);
		print_vir_pressure(&(energy->pressure));

                if (global_data->nr_qmatoms==0) {
                        strcpy(mminfo,"\n");
                } else {
                        strcpy(mminfo,"(only MM fragment)\n");
                }

                txtout("ELECTROSTATIC energy: %# 10.7g kcal/mol. %s",
                        energy->elec,mminfo);
                txtout("VAN DER WAALS energy: %# 10.7g kcal/mol.\n",
                        energy->vdw);
                txtout("BONDED energy.......: %# 10.7g kcal/mol. %s",
                        energy->bond,mminfo);
                txtout("ANGLE energy........: %# 10.7g kcal/mol. %s",
                        energy->angle,mminfo);
                txtout("DIHEDRAL energy.....: %# 10.7g kcal/mol. %s",
                        energy->dihe,mminfo);
                txtout("IMPROPER energy.....: %# 10.7g kcal/mol. %s",
                        energy->impr,mminfo);

		txtout("RESTRAINT energy....: %# 10.7g kcal/mol.\n",
			energy->restraint);
		txtout("SBOUND energy.......: %# 10.7g kcal/mol.\n",
			energy->sbound);
/**we dont have any H-bonds in EGO
		txtout("HBOND energy........: %# 10.7g kcal/mol.\n",
			energy->hbond);
**/
		txtout("FLOODING energy.....: %# 10.7g kcal/mol.\n",
			energy->flooding);
		txtout("FLOODING strength...: %# 10.7g kT.\n",
			energy->flood_energy_strength);
		txtout("FLOODING distance...: %# 10.7g A^2.\n",
			energy->flooding_distance);
		txtout("IMMOBIL. energy.....: %# 10.7g kcal/mol.\n",
			energy->imobil);
		txtout("PULLING forces......: %# 10.7g %# 10.7g %# 10.7g %# 10.7g pN.\n",
			energy->pull_indir[0],energy->pull_outdir[0],
			energy->pull_indir[1],energy->pull_outdir[1]);

		if (global_data->rotation_mode) {
		  txtout("ROTATION forces......: %# 10.7g %# 10.7g %# 10.7g pN.\n",
			 energy->rot_force[0],energy->rot_force[1],energy->rot_force[2]);
		  txtout("Rotation energy.....: %# 10.7g kcal/mol.\n",
			 global_data->rotation_energy);
		}
		txtout("KINETIC energy: T,H2O,P.....: %# 10.7g %# 10.7g %# 10.7g kcal/mol.\n",
			energy->kinetic,energy->kinetic_h2o,energy->kinetic_protein);
                txtout("QM energy intern....: %# 10.7g kcal/mol.\n",
                        energy->qm_energy[0]+global_data->qmenergyoffset);
                txtout("QM energy elec.ext..: %# 10.7g kcal/mol.\n",
                        energy->qm_energy[1]);	
		txtout("TOTAL energy........: %# 10.7g kcal/mol.\n",
			tot);	
		txtout("TEMPERATURE.........: %# 6.5g K %# 6.5g K %# 6.5g K.\n",
			temperature,temperature_h2o,temperature_protein);

		if (global_data->minimisation) {
		txtout("Gradient (max,ave)..: %# 10.7g %# 10.7g kcal/mol A\n",
			energy->gradinfo[0],energy->gradinfo[1]/global_data->total_no_atoms);	
		}
		
		txtout("Integration step....: % 6d (%.5g %s)\n",
			global_data->cycles_passed,
			curtime,metric);
		txtout("Time per step.......: %# .4g    %# .4g s.\n",
			time_user,time_real);

		if ( (!cycle_finished)&&(time_accu_count>1)) {
			txtout("**********************************************\n");
			txtout("Average time last %3d steps: % .4g  % .4g s.\n",
				time_accu_count,
				time_accu_user/(double)time_accu_count,
				time_accu_real/(double)time_accu_count);
			txtout("Average time since startup : % .4g  % .4g s.\n",
				time_accu_tot_user/(double)time_accu_count_tot,
				time_accu_tot_real/(double)time_accu_count_tot);
			txtout("*********************************************\n");
			time_accu_user  = 0.0;
			time_accu_real  = 0.0;			
   			time_accu_count = 0;		
		}
  
		txtout("\n");
   		cycle_finished = TRUE;
	}
#if DEBUG
	txtout("In writeout_energy: after print to screen print_to_file=%d.\n",print_to_file);
#endif
	if (print_to_file) { int i;
		if ((i=ene2file_newstyle(energy,global_data,
					temperature,tot,time_user)) != 0) {
			errout("Fatal error: Could not write energy data to file <%s>. Error = %d\n",
				file_name,i);
			pex(-1);
		}			
		fflush(to_file);
		print_to_file=FALSE;
	}
#if DEBUG
	txtout("In writeout_energy: after print to file.\n");
#endif	
	return;
} /* end of writeout_energy */


static int
ene2file_newstyle(	ENERGY_INFO *energy,
			LOAD_MSG_ALL_NODE *global_data,
			double temperature,
			double tot,
			double time)
{
	int stepnum;
        int i,j;
	stepnum = global_data->cycles_passed;
#if DEBUG
txtout("in ene2file_newstyle: fileopen=%d\n",file_open);
#endif

	if (fprintf(to_file,
	    "E %7d %# 6.3g %#7.4g %#.12g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g %#.7g ",
			stepnum,
			time,
			temperature,
			tot,
			energy->kinetic,
			energy->elec,
			energy->vdw,
			energy->bond,
			energy->angle,
			energy->dihe,
			energy->impr,
			energy->restraint,
			energy->sbound,
			energy->hbond,
			energy->flooding,
			energy->flood_energy_strength,
			energy->flooding_distance,
			energy->imobil,
			energy->pull_indir[0],
			energy->pull_outdir[0],
			energy->pull_indir[1],
			energy->pull_outdir[1]) < (5*FPRINTF_ERROR)) return -1;

	if (global_data->rotation_mode) {
	  if (fprintf(to_file,
		      "%#.7g %#.7g %#.7g %#.7g ",
		      energy->rot_force[0],
		      energy->rot_force[1],
		      energy->rot_force[2],
		      global_data->rotation_energy) < (4*FPRINTF_ERROR)) return -1;
	}
	if (fprintf(to_file,
		    "%#.12g %#.12g\n",
		    energy->qm_energy[0],
		    energy->qm_energy[1]) < (2*FPRINTF_ERROR)) return -1;

				
	if (any_sbound_set(global_data->sbound_droff)) {
		if (fprintf(to_file,"P ") < (2*FPRINTF_ERROR)) return -2;
		if (fprint_pressure(to_file,&(energy->pressure)) < (5*FPRINTF_ERROR)) return -3;
		if (fprint_pressure(to_file,&(energy->density)) < (5*FPRINTF_ERROR)) return -4;
		if (fprintf(to_file,"\n") < (1*FPRINTF_ERROR)) return -5;
		if (fprintf(to_file,"V ") < (2*FPRINTF_ERROR)) return -6;        
		if (fprintf_vir_pressure(to_file,&(energy->pressure)) < (5*FPRINTF_ERROR)) return -7;
		
		/*rok191196 last output in Virial line, Energy-tensor*/
		for(i=0;i<3;i++)
			for(j=0;j<3;j++) {
	       			if (fprintf(to_file,"%.7g ",
	       					energy->ekin_tensor[i][j]
	       			    ) < (1*FPRINTF_ERROR)) return -8;
			}
		if (fprintf(to_file,"%.7g ",energy->ekin_trace) < (1*FPRINTF_ERROR)) return -9;

		if (fprintf(to_file,"\n") < (1*FPRINTF_ERROR)) return -10;
	}
	return 0;
}

/*************************************************************************/
/*** The idea is to write a second restart file just in case something ***/
/*** goes wrong while writing it, so that one can ALWAYS resort to the ***/
/*** first restart file. If the writing was successfull, then the first***/
/*** one is removed and the second obtains the name of the first one.  ***/
/*** This name is the only valid name for a restart file.              ***/
/*************************************************************************/

void   restart(	int restartlevel,	
		int next_step,
		int next_file,
		double time_step,
		LOAD_MSG_ALL_NODE *global_data,
		ENERGY_INFO *energy)
{
	int    i;

	strcpy(tmprestartfile,global_data->restartfile);
	strcat(tmprestartfile,".tmp");
#if DEBUG
	txtout("In restart: restartlevel %d next_step %d next_file %d <%s>.\n",
		restartlevel,next_step,next_file,tmprestartfile);
#endif
	if (restartlevel==1) {

		if (r_file!=NULL) {	/*** close and erase old one ***/
			fclose(r_file);	
			r_file = NULL;
		}
		remove(tmprestartfile);	
		
		r_file = fopen(tmprestartfile,"w"); /*** open new one ***/
		if (r_file==NULL) {
			errout("An error occured while opening the tmp restart file <%s> !\n",
				tmprestartfile);
			pex(-1);
 		} else {
			/* write some header lines */
			fprintf(r_file,"Old time step was %e seconds.\n",time_step);
			fprintf(r_file,"Next integration step number: %d.\n",next_step+1);
			fprintf(r_file,"Next file number: %d.\n",next_file);
			fprintf(r_file,"The data are in the following fixed format:\n");
			fprintf(r_file,"one complete set of coordinates x2,y2,z2\n");
			if (fprintf(r_file,"a second complete set of coordinates x1,y1,z1\n") < (5*FPRINTF_ERROR)) {
				errout("Could not write header to tmp restart file !\n");
				pex(-1);	
			}	
		}
#if DEBUG
		txtout("In restart: after opening new tmp-restart-file <%s>.\n",
			tmprestartfile);
#endif
	} /* end of open new file */

#if DEBUG
	txtout("In restart: cur_atoms_insortbuf %d  total_no_atoms %d.\n",
		cur_atoms_insortbuf,global_data->total_no_atoms);
#endif
 	/*** Test weather we have a complete set of coordinates ***/
	if (cur_atoms_insortbuf==global_data->total_no_atoms) {
		/*** append the coordinates to file ***/
		for (i=0;i<global_data->total_no_atoms;i++) {		
			if (fprintf(r_file,"%.12e %.12e %.12e\n",
				global_data->atomdata[i].ko[0],
				global_data->atomdata[i].ko[1],
				global_data->atomdata[i].ko[2]) < (5*FPRINTF_ERROR)) {
				errout("Fatal error: Could not write coordinates to tmp restart file.\n");
				pex(-1);	
			}	
		}
		if (fflush(r_file)!=0) {
			errout("Fatal error: Could not write (fflush) coordinates to tmp restart file.\n");
			pex(-1);	
		}
		
		if (restartlevel==1) {
#if DEBUG
			txtout("In restart: First part of restart file written.\n");
#endif
		} else
		if (restartlevel==2) {
			/*** add COM for immobilisation ***/
			if (fprintf(r_file,"%d %.12e %.12e %.12e\n", 
				global_data->start_refC,
				global_data->refC[0],
				global_data->refC[1],
				global_data->refC[2])<(5*FPRINTF_ERROR)) {
				errout("Fatal error: Could not write data (refC(1)) to tmp restart file.\n");
				pex(-1);
			}
			if (global_data->flag_selimob>=2) {
			  if (fprintf(r_file,"%.12e %.12e %.12e\n", 
				      global_data->refC[3],
				      global_data->refC[4],
				      global_data->refC[5])<(4*FPRINTF_ERROR)) {
			              errout("Fatal error: Could not write data (refC(2)) to tmp restart file.\n");
				      pex(-1);
			  }
			  if (global_data->flag_selimob==3) {
			    if (fprintf(r_file,"%.12e %.12e %.12e\n", 
					global_data->refC[6],
					global_data->refC[7],
					global_data->refC[8])<(4*FPRINTF_ERROR)) {
			                errout("Fatal error: Could not write data (refC(3)) to tmp restart file.\n");
					pex(-1);
			    }
			  }
			}
			/*** add pulling start position ***/
			if (fprintf(r_file,"%.12e %.12e %.12e %.12e %.12e %.12e %.12e\n",	
				global_data->pullspringpos[0],
				global_data->pullspringpos[1],
				global_data->pullspringpos[2],
				global_data->pullspringpos[3],
				global_data->pullspringpos[4],
				global_data->pullspringpos[5],
				global_data->pullspringpos[6])<(6*FPRINTF_ERROR)) {
				errout("Fatal error: Could not write data to tmp restart file.\n");
				pex(-1);
			}
#if DEBUG
			txtout("In restart: after pullspringpos writing.\n");
#endif				
			/*** add SBOUND positions ***/
			write_sbound_droff(r_file,global_data->sbound_droff);
#if DEBUG
			txtout("In restart: after write_sbound_droff writing.\n");
#endif
			/*** add QM bounding box ***/
			if (fprintf(r_file,"QMBOX_X: %.12e %.12e\n",global_data->qmboxext[0],
								global_data->qmboxext[1])<(5*FPRINTF_ERROR)) {
				errout("Fatal error: Could not write data to tmp restart file.\n");
				pex(-1);
			}
			if (fprintf(r_file,"QMBOX_Y: %.12e %.12e\n",global_data->qmboxext[2],
								global_data->qmboxext[3])<(5*FPRINTF_ERROR)) {
				errout("Fatal error: Could not write data to tmp restart file.\n");
				pex(-1);
			}
			if (fprintf(r_file,"QMBOX_Z: %.12e %.12e\n",global_data->qmboxext[4],
								global_data->qmboxext[5])<(5*FPRINTF_ERROR)) {
				errout("Fatal error: Could not write data to tmp restart file.\n");
				pex(-1);
			}
#if DEBUG
			txtout("In restart: after QMBOX writing.\n");
#endif			
			/*** add moving average for flooding energy ***/
			if (fprintf(r_file,"FLOOD_ENERGY_AVERAGE:  %.12e\n",
				energy->flood_energy_average)<(5*FPRINTF_ERROR)) {
				errout("Fatal error: Could not write data to tmp restart file.\n");
				pex(-1);
			}
			if (fprintf(r_file,"FLOOD_ENERGY_STRENGTH: %.12e\n",
				energy->flood_energy_strength)<(5*FPRINTF_ERROR)) {
				errout("Fatal error: Could not write data to tmp restart file.\n");
				pex(-1);
			}
			if (fprintf(r_file,"FLOOD_DISTANCE: %.12e\n",
				energy->flooding_distance)<(5*FPRINTF_ERROR)) {
				errout("Fatal error: Could not write data to tmp restart file.\n");
				pex(-1);
			}
#if DEBUG
			txtout("In restart: after FLOODING writing.\n");
#endif
			if (global_data->rotation_mode) {
			  if (fprintf(r_file,"Rotation_No._of_Atoms: %d\n",
				      global_data->nr_of_rot_atoms)<(1*FPRINTF_ERROR)){
			    errout("Fatal error: Could not write nr_of_rot_atoms to tmp restart file.\n");
			    pex(-1);
			  }
			  for (i=0;i<global_data->nr_of_rot_atoms;i++) {		
			    if (fprintf(r_file,"%d %.12e %.12e %.12e %.12e %.12e %.12e %.12e %.12e %.12e %.12e\n",
					global_data->rotation_map[i],
					global_data->rot_x[i],
					global_data->rot_y[i],
					global_data->rot_z[i],
					global_data->rot1_x[i],
					global_data->rot1_y[i],
					global_data->rot1_z[i],
					global_data->rot2_x[i],
					global_data->rot2_y[i],
					global_data->rot2_z[i],
					global_data->rotation_torque[i]) < (12*FPRINTF_ERROR)) {
					
			      errout("Fatal error: Could not write rotation coordinates to tmp restart file.\n");
			      pex(-1);	
			    }	
			  }
			  if (fprintf(r_file,"%.12e\n",
				      global_data->rotation_energy)<(1*FPRINTF_ERROR)){
			    errout("Fatal error: Could not write rotation energy to tmp restart file.\n");
			    pex(-1);
			  }
			}		

			/*** restart file now completed --> close ***/
			if (fclose(r_file)!=0) {
				errout("Fatal error: Could not write (fclose) coordinates to tmp restart file.\n");
				pex(-1);	
			}	
			r_file = NULL;
			
			/*** and rename the temporary restart file. ***/
			remove(global_data->restartfile);
			rename(tmprestartfile,global_data->restartfile);
			txtout("Restart file <%s> updated !\n",global_data->restartfile);
		}
	}
} /* end of procedure restart */


/**************************************************************************/
/*** Here the system call from EGO is periodically envoked		***/
/*** This system call usually calls a shell script, which does some     ***/
/*** manipulations to your EGO output files (e.g. compressing,converting)**/
/*** A sample shell-script using ego2crd is given in ./utils/ego2crd.sh ***/
/**************************************************************************/
static	int	filecount=0;	/*** number of files completed ***/
static	char	callstr[4*MAX_STRING_SIZE];
static	char	curcrdname[MAX_STRING_SIZE];

static int
maybe_perform_systemcall(LOAD_MSG_ALL_NODE *global_data,int curfilenum)
{
	int 	res,len;
	int	firstfilenum=-1;
	int	lastfilenum=-1;

	/*** do nothing in that cases ***/
	if (global_data->syscallstride<1)		return 0;	
	if (global_data->syscallstring[0]==0)		return 0;
	if (curfilenum==0)				return 0;	
	if ((curfilenum%global_data->syscallstride)!=0)	return 0;

	/*** now we have global_data->syscallstride files ***/	
	lastfilenum = curfilenum-1;
	firstfilenum= curfilenum-global_data->syscallstride;

	generate_egofname(curcrdname,"",firstfilenum);
	len = strlen(curcrdname);
	curcrdname[len-4] = 0;	/*** truncate .ego ***/
	strcat(curcrdname,"-");
	len = strlen(curcrdname);
	generate_egofname(curcrdname+len,"",lastfilenum);
	len = strlen(curcrdname);
	curcrdname[len-4] = 0;	/*** truncate .ego ***/

	sprintf(callstr,"%s %d %d %s %s %s %s",	global_data->syscallstring,
					firstfilenum,
					lastfilenum,
					curcrdname,
					global_data->outpath,
					global_data->lispath,
					global_data->workpath);
#if DEBUG
	txtout("System call started: <%s> ...",callstr);
#endif
	res = (int)system(callstr)/256;
	if (res==0) {
#if DEBUG
		txtout("Exit: OK\n");
#endif
	} else {
		if (res==99) {
		  global_data->final_cycle_no = global_data->Analyse_nth + global_data->cycles_passed;
		  errout("-------------------------------------------------------\n");
		  errout("Exit with 99 from syscall. Will exit ego at step %d.\n",global_data->final_cycle_no);
		  errout("-------------------------------------------------------\n\n");
		  }
		else {
		  errout("Exit with %d from syscall ! Check any problems.\n",res);
		  }
	}

	firstfilenum = curfilenum;
	filecount = 1;

	return 0;
}

int
writeout_myforces(	LOAD_MSG_ALL_NODE *global_data,
			STRUCT_FORCES *my_force)
{
	int difstep=global_data->cycles_passed-global_data->forceout_start_step;
#if DEBUG
	txtout("In writeout_myforces: Step %d atoms %d Step %d Stride %d test %d\n",
		global_data->cycles_passed,
		global_data->forceout_nr_atoms,
		global_data->forceout_start_step,
		global_data->forceout_stride_step,
		difstep%global_data->forceout_stride_step);
#endif
	if (global_data->forceout_nr_atoms<=0)			return 0;
	if (difstep<0) 						return 0;
	if (global_data->forceout_stride_step<=0)		return 0;
	if ((difstep%global_data->forceout_stride_step)!=0) 	return 0;	

	if (my_force->total.len)
		writeout_force_type(	FORCE_TOTAL,FFN_TOTAL,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->total));

	if (my_force->elecvdw.len) {
		writeout_force_type(	FORCE_ELECVDW,FFN_ELECVDW,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->elecvdw));
	}
	if (my_force->bond.len) {
		writeout_force_type(	FORCE_BOND,FFN_BOND,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->bond));
	}
	if (my_force->elec.len) {
		writeout_force_type(	FORCE_ELEC,FFN_ELEC,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->elec));
	}
	if (my_force->vdw.len) {
		writeout_force_type(	FORCE_VDW,FFN_VDW,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->vdw));
	}
	if (my_force->angle.len) {
		writeout_force_type(	FORCE_ANGLE,FFN_ANGLE,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->angle));
	}
	if (my_force->dihe.len) {
		writeout_force_type(	FORCE_DIHE,FFN_DIHE,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->dihe));
	}
	if (my_force->impr.len) {
		writeout_force_type(	FORCE_IMPR,FFN_IMPR,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->impr));
	}
	if (my_force->restr.len) {
		writeout_force_type(	FORCE_RESTR,FFN_RESTR,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->restr));
	}
	if (my_force->sbound.len) {
		writeout_force_type(	FORCE_SBOUND,FFN_SBOUND,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->sbound));
	}
	if (my_force->intern.len) {
		writeout_force_type(	FORCE_INTERN,FFN_INTERN,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->intern));
	}
	if (my_force->flood.len) {
		writeout_force_type(	FORCE_FLOOD,FFN_FLOOD,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->flood));
	}
	if (my_force->kin.len) {
		writeout_force_type(	FORCE_KIN,FFN_KIN,
					global_data,
					global_data->atom_sele_flag1,
					&(my_force->kin));
	}

	if (my_force->pickbond.len) {
		writeout_pickbond(	FFN_PICKBOND,
					global_data,
					&(my_force->pickbond));
	}

	if (my_force->pickelec.len) {
		writeout_pickbond(	FFN_PICKELEC,
					global_data,
					&(my_force->pickelec));
	}

	if (my_force->pickvdw.len) {
		writeout_pickbond(	FFN_PICKVDW,
					global_data,
					&(my_force->pickvdw));
	}

	if (my_force->pickangle.len) {
		writeout_pickangle(	FFN_PICKANGLE,
					global_data,
					&(my_force->pickangle));
	}

	if (my_force->pickdihe.len) {
		writeout_pickdihe(	FFN_PICKDIHE,
					global_data,
					&(my_force->pickdihe));
	}

	if (my_force->pickimpr.len) {
		writeout_pickdihe(	FFN_PICKIMPR,
					global_data,
					&(my_force->pickimpr));
	}
#if DEBUG
	txtout("After writeout_myforces\n");
#endif
	return 0;
}

/******************************************************************/
/*** Lets append the forces to the specific files: 		***/
/*** As there is a (hidden) minus sign in the integrator the	***/
/*** EGO-forces have the opposite direction as they "physically"***/
/*** have. Therefore we write the "negative" EGO-forces to file.***/
/******************************************************************/
static int
writeout_force_type(	int force_type,
			char *file_name,
			LOAD_MSG_ALL_NODE *global_data,
			int *force_id,
			STRUCT_FORCE_INFO *finfo)
{
	FILE	*forceh;
	int	newfile=0,i;
	char	fname[MAX_STRING_SIZE];

	strcpy(fname,global_data->outpath);
	strcat(fname,global_data->idstr);
	strcat(fname,file_name);

#if DEBUG
	txtout("Writeout forcetype=%d <%s>\n",force_type,fname);
#endif
	if ((forceh=fopen(fname,"r"))==NULL) { /*** test existance of file ***/
		newfile = 1;
	} else {
		fclose(forceh);
	}

	if ((forceh=fopen(fname,"aw"))==NULL) { /*** open for appending ***/
		errout("Fatal error: Could not open force file <%s>\n",
			file_name);
		pex(-1);
 	}

	if (newfile) {		/*** add header information if new file ***/
		char headstr[MAX_STRING_SIZE];
		sprintf(headstr,"REMARK Force file %s\n",file_name);
			
		write_force_header(forceh,global_data,
				force_type,finfo->len,
				headstr,"BEGINFORCE");
	}
	
	/*** write out atomid, energy, position and forces ***/
	fprintf(forceh,"Step %-8d\n",global_data->cycles_passed);
	for (i=0;i<finfo->len;i++) {
		int offs=FORCEINFOLEN*i;
		fprintf(forceh,"%-6d %# 10.5f   %# 10.5f %# 10.5f %# 10.5f   %# 10.5f %# 10.5f %# 10.5f\n",
			 finfo->atom_id[i]+1,
			 finfo->force[offs + 0], /*** Energy ***/
			 finfo->force[offs + 4], /*** Kox ***/
			 finfo->force[offs + 5], /*** Koy ***/
			 finfo->force[offs + 6], /*** Koz ***/
			-finfo->force[offs + 1], /*** Fx ***/
			-finfo->force[offs + 2], /*** Fy ***/
			-finfo->force[offs + 3]);/*** Fz ***/
	}
	fclose(forceh);
	
	finfo->cur_len = 0;	/*** very important ***/
	return 0;
}

static void
writeout_pickbond(	char *file_name,
			LOAD_MSG_ALL_NODE *global_data,
			STRUCT_PICK_BOND *pinfo)
{
	FILE	*forceh;
	int	newfile=0,i,k;
	char	fname[MAX_STRING_SIZE];

	strcpy(fname,global_data->outpath);
	strcat(fname,global_data->idstr);
	strcat(fname,file_name);

#if DEBUG
	txtout("Writeout PICKBOND <%s>\n",fname);
#endif
	if ((forceh=fopen(fname,"r"))==NULL) { /*** test existance of file ***/
		newfile = 1;
	} else {
		fclose(forceh);
	}

	if ((forceh=fopen(fname,"aw"))==NULL) { /*** open for appending ***/
		errout("Fatal error: Could not open force file <%s>\n",
			file_name);
		pex(-1);
 	}

	if (newfile) {		/*** add header information if new file ***/		
     		fprintf(forceh,"REMARK Force file %s\n",file_name);
		fprintf(forceh,"REMARK %s ===========================================\n",global_data->ctlfile);
		writeout_ctlfile(forceh,global_data);
		fprintf(forceh,"[BEGINFORCE]   Distance: A     E: kcal/mol     Fx, Fy Fz: kcal/(molA)\n");
		fprintf(forceh,"Picked %d\n",pinfo->len);
		fprintf(forceh,"Atom1  Atom2     Atom-Str1  Atom-Str2\n");
		for (i=0;i<2*pinfo->len;i+=2) {
			int bp1=pinfo->atom_ids[i  ];
			int bp2=pinfo->atom_ids[i+1];
			fprintf(forceh,"%-6d %-6d    %-4s       %-4s\n",
				bp1+1,
				bp2+1,
				global_data->atomdata[bp1].atomstr,
				global_data->atomdata[bp2].atomstr);
		}
	}

	fprintf(forceh,"Step %-8d\n",global_data->cycles_passed);
	for (i=0,k=0;i<2*pinfo->len;i+=2,k+=5) {
		int bp1=pinfo->atom_ids[i  ];
		int bp2=pinfo->atom_ids[i+1];
		fprintf(forceh,"%-6d %-6d   % 10.5f % 10.5f   % 10.5f % 10.5f % 10.5f   % 10.5f % 10.5f % 10.5f\n",
			bp1+1,bp2+1,	
			pinfo->force[k+4],  /*** Distance ***/
			pinfo->force[k  ],  /*** Energy   ***/
			-pinfo->force[k+1], /*** Fx1 ***/
			-pinfo->force[k+2], /*** Fy1 ***/
			-pinfo->force[k+3], /*** Fz1 ***/
			pinfo->force[k+1],  /*** Fx2 ***/
			pinfo->force[k+2],  /*** Fy2 ***/
			pinfo->force[k+3]   /*** Fz2 ***/
		);
	}
	fclose(forceh);
	return;

}

static void
writeout_pickangle(	char *file_name,
			LOAD_MSG_ALL_NODE *global_data,
			STRUCT_PICK_BOND *pinfo)
{
	FILE	*forceh;
	int	newfile=0,i,k;
	char	fname[MAX_STRING_SIZE];

	strcpy(fname,global_data->outpath);
	strcat(fname,global_data->idstr);
	strcat(fname,file_name);

#if DEBUG
	txtout("Writeout PICKANGLE <%s>\n",fname);
#endif
	if ((forceh=fopen(fname,"r"))==NULL) { /*** test existance of file ***/
		newfile = 1;
	} else {
		fclose(forceh);
	}

	if ((forceh=fopen(fname,"aw"))==NULL) { /*** open for appending ***/
		errout("Fatal error: Could not open force file <%s>\n",
			file_name);
		pex(-1);
 	}

	if (newfile) {		/*** add header information if new file ***/		
     		fprintf(forceh,"REMARK Force file %s\n",file_name);
		fprintf(forceh,"REMARK %s ==================\n",global_data->ctlfile);
		writeout_ctlfile(forceh,global_data);
		fprintf(forceh,"[BEGINFORCE]   Angle: Degree   E: kcal/mol   Fx, Fy Fz: kcal/(molA)\n");
		fprintf(forceh,"Picked %d\n",pinfo->len);
		fprintf(forceh,"Atom1  Atom2  Atom3     Atom-Str1  Atom-Str2  Atom-Str3\n");
		for (i=0;i<3*pinfo->len;i+=3) {
			int bp1=pinfo->atom_ids[i  ];
			int bp2=pinfo->atom_ids[i+1];
			int bp3=pinfo->atom_ids[i+2];
			fprintf(forceh,"%-6d %-6d %-6d    %-4s       %-4s       %-4s\n",
				bp1+1,
				bp2+1,
				bp3+1,
				global_data->atomdata[bp1].atomstr,
				global_data->atomdata[bp2].atomstr,
				global_data->atomdata[bp3].atomstr);
		}
	}

	fprintf(forceh,"Step %-8d\n",global_data->cycles_passed);
	for (i=0,k=0;i<3*pinfo->len;i+=3,k+=8) {
		int bp1=pinfo->atom_ids[i  ];
		int bp2=pinfo->atom_ids[i+1];
		int bp3=pinfo->atom_ids[i+2];
		fprintf(forceh,"%-6d %-6d %-6d   % 10.5f % 10.5f   % 10.5f % 10.5f % 10.5f   % 10.5f % 10.5f % 10.5f   % 10.5f % 10.5f % 10.5f\n",
			bp1+1,bp2+1,bp3+1,	
			pinfo->force[k+7],  			/*** Angle  ***/
			pinfo->force[k  ],  			/*** Energy ***/
			-pinfo->force[k+1], 			/*** Fx1 ***/
			-pinfo->force[k+2], 			/*** Fy1 ***/
			-pinfo->force[k+3], 			/*** Fz1 ***/
			pinfo->force[k+1]+pinfo->force[k+4],	/*** Fx2 ***/
			pinfo->force[k+2]+pinfo->force[k+5],	/*** Fy2 ***/
			pinfo->force[k+3]+pinfo->force[k+6],	/*** Fz2 ***/
			-pinfo->force[k+4], 			/*** Fx3 ***/
			-pinfo->force[k+5], 			/*** Fy3 ***/
			-pinfo->force[k+6]  			/*** Fz3 ***/
		); 
	}
	fclose(forceh);
	return;

}

static void
writeout_pickdihe(	char *file_name,
			LOAD_MSG_ALL_NODE *global_data,
			STRUCT_PICK_BOND *pinfo)
{
	FILE	*forceh;
	int	newfile=0,i,k;
	char	fname[MAX_STRING_SIZE];

	strcpy(fname,global_data->outpath);
	strcat(fname,global_data->idstr);
	strcat(fname,file_name);

#if DEBUG
	txtout("Writeout PICKANGLE <%s>\n",fname);
#endif
	if ((forceh=fopen(fname,"r"))==NULL) { /*** test existance of file ***/
		newfile = 1;
	} else {
		fclose(forceh);
	}

	if ((forceh=fopen(fname,"aw"))==NULL) { /*** open for appending ***/
		errout("Fatal error: Could not open force file <%s>.\n",
			fname);
		pex(-1);
 	}

	if (newfile) {		/*** add header information if new file ***/		
     		fprintf(forceh,"REMARK Force file %s\n",file_name);
		fprintf(forceh,"REMARK %s =================\n",global_data->ctlfile);
		writeout_ctlfile(forceh,global_data);
		fprintf(forceh,"[BEGINFORCE]   Angle: Degree   E: kcal/mol      Fx, Fy Fz: kcal/(molA)\n");
		fprintf(forceh,"Picked %d\n",pinfo->len);
		fprintf(forceh,"Atom1  Atom2  Atom3  Atom4     Atom-Str1  Atom-Str2  Atom-Str3  Atom-Str4\n");
		for (i=0;i<4*pinfo->len;i+=4) {
			int bp1=pinfo->atom_ids[i  ];
			int bp2=pinfo->atom_ids[i+1];
			int bp3=pinfo->atom_ids[i+2];
			int bp4=pinfo->atom_ids[i+3];
			fprintf(forceh,"%-6d %-6d %-6d %-6d    %-4s       %-4s       %-4s       %-4s\n",
				bp1+1,
				bp2+1,
				bp3+1,
				bp4+1,
				global_data->atomdata[bp1].atomstr,
				global_data->atomdata[bp2].atomstr,
				global_data->atomdata[bp3].atomstr,
				global_data->atomdata[bp4].atomstr);
		}
	}

	fprintf(forceh,"Step %-8d\n",global_data->cycles_passed);
	for (i=0,k=0;i<4*pinfo->len;i+=4,k+=14) {
		int bp1=pinfo->atom_ids[i  ];
		int bp2=pinfo->atom_ids[i+1];
		int bp3=pinfo->atom_ids[i+2];
		int bp4=pinfo->atom_ids[i+3];
		fprintf(forceh,"%-6d %-6d %-6d %-6d   % 10.5f % 10.5f   % 10.5f % 10.5f % 10.5f   % 10.5f % 10.5f % 10.5f   % 10.5f % 10.5f % 10.5f   % 10.5f % 10.5f % 10.5f\n",
			bp1+1,bp2+1,bp3+1,bp4+1,	
			pinfo->force[k+13], /*** Angle ***/
			pinfo->force[k   ], /*** Energy ***/
			-pinfo->force[k+1], /*** Fx1 ***/
			-pinfo->force[k+2], /*** Fy1 ***/
			-pinfo->force[k+3], /*** Fz1 ***/
			-pinfo->force[k+4], /*** Fx2 ***/
			-pinfo->force[k+5], /*** Fy2 ***/
			-pinfo->force[k+6], /*** Fz2 ***/
			-pinfo->force[k+7], /*** Fx3 ***/
			-pinfo->force[k+8], /*** Fy3 ***/
			-pinfo->force[k+9], /*** Fz3 ***/
			-pinfo->force[k+10],/*** Fx4 ***/
			-pinfo->force[k+11],/*** Fy4 ***/
			-pinfo->force[k+12] /*** Fz4 ***/
		);
	}
	fclose(forceh);
	return;

}
