/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*********************************************************************************
**										**
**	Module	: boundary.c	by M.Eichinger, H.Heller, H.Grubmueller, Munich	**
**	Belongs2: EGO		The parallel MD-Simulation programm		**
**	Contents: HARMONIC, STOCHASTIC and SBOUND-Potential.			**
**		  All functions that describe the boundary of the		**
**		  simulation volume.						**																	
**		  The boundary can be specified in several ways:		**
**		  1) Col. Q and B of the pdb file contain friction and		**
**		     harmonic force constants on an atom-by-atom basis.		**
**		     To activate this, you have to set HARMONIC and/or 		**
**		     STOCHASTIC TRUE (in ctl.lis), but the radius and location	**
**		     of the SBOUND sphere/planes MUST be set to zero.		**
**		     Curvature must be set to zero.				**
**		  2) If you set the radius of the SBOUND sphere to a positive	**
**		     value you define a spherical boundary of given radius.	**
**		     You can add SBOUND planes to obtain, e.g., rounded caps.	**
**		     The curvature parameter must be set to zero.		**
**		     The Q column (friction) is ignored.			**
**		     If you set STOCHASTIC TRUE (in ctl.lis), you add the	**
**		     appropriate stochastic forces. You can specify their	**
**		     range separately. 						**
**		     If you set HARMONIC TRUE (in ctl.lis), you can add an	**
**		     additional harmonic potential according to the B col.	**
**                3) If you set the radius of the SBOUND sphere to a negative   **
**                   value you define an irregularly shaped boundary from       **
**                   the data in bounday.lis, which can be created by the       **
**                   utility program SOLVATE 1.0.                               **
**                   The Q column (friction) is ignored.                        **
**                   If you set STOCHASTIC TRUE (in ctl.lis), you add the       **
**                   appropriate stochastic forces. You can specify their       **
**                   range separately.                                          **
**                   If you set HARMONIC TRUE (in ctl.lis), you can add an      **
**                   additional harmonic potential according to the B col.      **
**		  4) If you set the location of any SBOUND plane to a non-zero	**
**		     value, the curvature parameter defines the			**
**		     curvature of the corners and edges of the rectangular	**
**		     prism. A radius of 0.0 is NOT recommended, but possible.	**
**		     If you use a radius of 0.0, you can select single planes,	**
**		     e.g., only infinite parallel XZ planes. Planes through the	**
**		     origin will be ignored. If you have a positive radius, all	**
**		     planes will be active, even if they contain the origin.	**
**		     The Q column (friction) is ignored.			**
**		     If you set STOCHASTIC TRUE (in ctl.lis), you add the	**
**		     appropriate stochastic forces. You can specify their	**
**		     range separately. 						**
**		     If you set HARMONIC TRUE (in ctl.lis), you can add an	**
**		     additional harmonic potential according to the B col.	**
**										**
**  	Modif's: 22.6.1995, H. Heller: added round corners to rectangular	**
**		 prism and reordered all procedures.				**
**		 introduced two sets of SBOUND2 params: water and other		**		
**	         9.1.1997, H. Grubmueller: added irregularly shaped boundary    **		
**		31.7.1997, H. Heller: added torques				**		
**										**
*********************************************************************************/

#include <math.h>
#include <stdio.h>

#include "ego.h"
#include "egbound.h"
#include "util_lib.h"
#include "egalloc.h"
#include "egpressure.h"

#define DEBUG 0
/* DEBUG -1:	general debug printouts */
/* DEBUG  0:	default: NO debugging */
/* DEBUG  1:	debug planar SBOUND pressures */
/* DEBUG  2:	debug moving SBOUND */
/* DEBUG  3:	debug special mode */
/* DEBUG  4:	debug SBOUND potential in edges */
/* DEBUG  5:    debug ranged SBOUND */
/* DEBUG  6:	debug SBOUND potential in corners */
/* DEBUG  7:	debug torques */
/* DEBUG  8:	debug torques: compute average force from torques */

/******** START: PATCH to modify the harmonic potential to only act in selected directions ****/
/* Possible values are defined bitwise:
	00000001	1: harmonic in X
	00000010	2: harmonic in Y
	00000100	4: harmonic in Z
	00000111	7: harmonic in X, Y, Z (default)	
	00001000	8: asymmetric harmonic restraints */

/* #define PATCH_HARMONIC	10  *//* only in Y and only inward */
#define PATCH_HARMONIC	7 /* default */

/* if the compiler is gcc, print a warning while compiling: */
#if defined(NEXT) || defined(SUN4)  /* only gcc compiler knows about #warning directive */
#if (PATCH_HARMONIC & 7) != 7
#warning "ATTENTION: the Patch for HARMONIC is ON"
#endif
#if (PATCH_HARMONIC & 8) == 8
#warning "ATTENTION: Patch for HARMONIC is ASYMMETRIC"
#endif
#endif
/******** END: PATCH to modify the harmonic potential to only act in selected directions ****/



/******** START: PATCH FOR MOVING SBOUND ***********/
#define PATCH_MS FALSE

#if (PATCH_MS==TRUE)
static int sboundcounter=0; /* used in printing out SBOUND position to screen */
static double sboundmove = 0.0004; /* speed in AA per integration step */
/* static double sboundmove = 0.0; */
static int sboundindex1 = 2; /* add to this one (2=-Y) */
static int sboundindex2 = 3; /* subtract from this one (3=+Y) */
#endif
/******** END: PATCH FOR MOVING SBOUND ***********/

/******** START: PATCH FOR ADDL CONSTANT FORCE ***********/
#define PATCH_CF FALSE

#if (PATCH_CF==TRUE)
static double const_force = 0.001;	/*** enter a positive force here ***/
static double middle_plane_dist = 2.5;	/*** Distance from the middle of the
					 *** membrane at Y=0 in which NO constant force
					 *** will be active. ***/

#endif
/******** END: PATCH FOR ADDL CONSTANT FORCE ***********/

/******** PATCH to switch TORQUES on or off: ***********/
#define DO_TORQUES FALSE	/* set it to TRUE if you want torques */

#if defined(NEXT) || defined(SUN4)  /* only gcc compiler knows about #warning directive */
#if (DO_TORQUES == TRUE) 
#warning "ATTENTION: TORQUES are ACTIVE, energy may not be conserved!"
#else
#warning "ATTENTION: TORQUES are OFF"
#endif
#endif




#if DEBUG == 8
static double average_torque_force=0.0;
static int    average_torque_count=0;
#endif

static PRESSURE old_positions; /* also used in constant_density mode for sliding average */
static int first_call_to_density;
static RANGE range;	/* holds range definitions for ranged SBOUND */
static RANGE special;	/* holds range definitions for special pressure mode */
static int my_number = -999; /* for debugging purposes */

/******** Local functions:	**********/
static void calc_spheric_sbound ( LOAD_MSG_ONE_NODE *my_data,
				  LOAD_MSG_ALL_NODE *global_data,
				  int start, int stop,  /* only atoms start <= i <= stop will be considered */
				  double *sboundpos,	/** x,y,z,R **/
				  double sbound_p1_h2o,
				  double sbound_force_h2o,
				  double sbound_p1,
				  double sbound_force,
				  double *pressure,
				  double *density,
				  double *forces);

static void calc_cylindric_sbound ( LOAD_MSG_ONE_NODE *my_data,
				    LOAD_MSG_ALL_NODE *global_data,
				    double *sboundpos,	/** x,y,R **/
				    double sbound_p1_h2o,	/* width of potential */
				    double sbound_force_h2o, /* depth of potential */
				    double sbound_p1,	/* width of potential */
				    double sbound_force, /* depth of potential */
				    PRESSURE *pressure, /* pressure elements */
				    PRESSURE *density   /* density elements */
				    );

static void calc_in_edge (LOAD_MSG_ONE_NODE *my_data,
			  LOAD_MSG_ALL_NODE *global_data,
			  int atom_index,
			  int *octant, int edge,
			  double *sbound_droff,
			  double *pressure,
			  double *density);

static void calc_in_corner (LOAD_MSG_ONE_NODE *my_data,
			    LOAD_MSG_ALL_NODE *global_data,
			    int    atom_index,
			    int    *octant,
			    double *sbound_droff,
			    double *pressure,
			    double *density);

static int in_edge (double x, double y, double z, int *octant,
	            double *sbound_droff);

static int in_corner (double x, double y, double z, 
		      int *octant,
		      double *sbound_droff);

static void set_octant (double x, double y, double z,
			double *sbound_droff,
			int *octant);

static void calc_planar_sbound (LOAD_MSG_ONE_NODE *my_data,
				LOAD_MSG_ALL_NODE *global_data,
				double *sbound_droff,
				PRESSURE *pressure,
				PRESSURE *density);

static double calc_greaterplane_sbound (double sbound_p1,
					double sbound_force,
					double atompos,
					double sboundpos,
					double *forceatom,
					double *SB_num);

static double calc_lesserplane_sbound ( double sbound_p1,
					double sbound_force,
					double atompos,
					double sboundpos,
					double *forceatom,
					double *SB_num);

static double calc_arbshaped_sbound ( LOAD_MSG_ONE_NODE *my_data,
				      LOAD_MSG_ALL_NODE *global_data,
				      double sbound_p1_h2o,
				      double sbound_force_h2o,
				      double sbound_p1,
				      double sbound_force,
				      double *pressure,
				      double *density);

static double calc_numerical_sbound (LOAD_MSG_ONE_NODE *my_data,
				     LOAD_MSG_ALL_NODE *global_data,
				     double *Gamma);

static double density_grad_curve(LOAD_MSG_ALL_NODE *global_data,
				 double x, double y, double z, 
				 double *gx, double *gy, double *gz, double *g, 
				 double *cr, double *curvature);

static double scale_edge_potential (double Px, double Py,
				    double Ox, double Oy,
				    double *Dx, double *Dy);

static double scale_corner_potential (double x, double y, double z,
				      double *origin,
				      int *octant,
				      double *sbound_droff,
				      double D[3]);

static int in_stochastic_region ( double *sbound_droff,
				  double *xyzko,
				  double layerthick,
				  double *offset);

static void calc_random_force (double rfd, double *xyzforce);

static void add_stochastic_forces ( LOAD_MSG_ONE_NODE *my_data,
				    LOAD_MSG_ALL_NODE *global_data,
				    double *rnd_force_std_deviation,
				    double *Gamma);

static void add_harmonic_boundary ( LOAD_MSG_ONE_NODE *my_data);

#if 0
/* this is not needed any more, but I like to keep it here, just in case... (HH) */
static int max3 (double Px, double Py, double Pz);
#endif 

static void   calc_radial_distance (double *xyzko,
			            double Ox, double Oy, double Oz,
			            double radius, double maxoffset,
			            int *active,
			            double *offset);

static int     use_water_params(int i, LOAD_MSG_ONE_NODE *my_data);

static void compute_area (PRESSURE *area, double *sbound_droff);
static void compute_pressure (PRESSURE *pressure, PRESSURE *area, int mode);
static void constant_pressure( PRESSURE *pressure,     /* instantaneous pressure, already converted to atm */
			PRESSURE *area,
			LOAD_MSG_ALL_NODE *global_data);
static void constant_density ( PRESSURE *pressure,     /* instantaneous density */
			PRESSURE *area,
			LOAD_MSG_ALL_NODE *global_data);
static void constant_special ( 	LOAD_MSG_ALL_NODE *global_data, 
				double *special_YP_pos, double *special_YM_pos,
				int     special_YP_count,  int  special_YM_count);
static void set_special (int no_atoms, double *y, int *atom_id);
static int is_water(char *res_string);
static int assign_water (LOAD_MSG_ONE_NODE *my_data, int a_water_atom_index, WATER *selected_water);
static int select_water (LOAD_MSG_ONE_NODE *my_data, int a_water_atom_index, WATER *selected_water);
static void compute_dipol_moment (LOAD_MSG_ONE_NODE *my_data, WATER *selected_water, VECTOR *dipol);
static double vector_length (VECTOR *vector);
static double scalar_product (VECTOR *v1, VECTOR *v2);
static double compute_pol_angle (VECTOR *dipol, VECTOR *radius);
static void compute_torque (LOAD_MSG_ONE_NODE *my_data, WATER *selected_water, 
			    VECTOR *normal, /* vector normal to the boundary, 
			                       pointing outward (into the vacuum) */
			    VECTOR *dipole,
			    double dist, /* distance (of C.M.) to surface */
			    EVECTOR *torque);
static double M (double dist, double theta);
static double EM (double dist, double theta);
static void cross_product (VECTOR *a, VECTOR *b, VECTOR *c);
static void compute_torque_forces (LOAD_MSG_ONE_NODE *my_data, int a_water_atom_index, VECTOR *normal,
				   double dist);


#define sign(a) (a>=0 ? 1 : -1)


/**************************************************************************/

/** sbound_droff[]:  -x, +x, -y, +y, -z, +z , ori_x, ori_y, ori_z, radius, curve, x_cyl  y_cyl  R_cyl
 **                   0   1   2   3   4   5       6      7      8       9     10     11     12     13
 **/

/**************************************************************************/
/***** is called in every integration step once to compute all boundary ***/
/***** contributions						   	***/

/******************************************************************/
/************************** SBOUND potential **********************/
/******************************************************************/

void   boundary ( LOAD_MSG_ONE_NODE *my_data,
		  LOAD_MSG_ALL_NODE *global_data,
		  double *rnd_force_std_deviation,
		  double *Gamma,   /* changes every step, if any sbound
				    * set! */
		  PRESSURE *pressure,
		  PRESSURE *density)

{
#if DEBUG == 8
  average_torque_force = 0.0;
  average_torque_count = 0;
#endif

  /** check for stochastic forces **/
  if (global_data->stochastic && global_data->restarting) 
  {	/* global_data->restarting is set TRUE in every
	 * but the very first step */
        /* if radius is negativ we have a numerical sbound
	 * in that case, we calc the stochastic forces in 
	 * direct in calc_numerical_sbound */
      if ((global_data->sbound_droff[R_SPHERE] >= (-SMALL)) || (global_data->sbound_droff[R_SPHERE] < (-1.0))) {
      add_stochastic_forces(my_data, global_data, 
			    rnd_force_std_deviation,
			    Gamma);	/* changes every step,
					 * if any sbound set!  */
    }
  }

  /*** check for harmonic boundary from B column of pdb file ***/
  if (global_data->harmonic) 
  {
    add_harmonic_boundary(my_data);
  }
  
  set_special(my_data->no_atoms, my_data->y_new, my_data->atom_id); 
	
  /*************************************************/
  /******** Spheric version ************************/
  /*************************************************/
  if (global_data->sbound_droff[R_SPHERE]>SMALL) 
  {
    calc_spheric_sbound(my_data,
			global_data,
			0, (my_data->no_atoms)-1,
			global_data->sbound_droff + 6,
			SBOUND_P1_H2O,	  /* width of potential for water */
			SBOUND_FORCE_H2O, /* depth of potential for water */
			SBOUND_P1,	/* width of potential for lipids */
			SBOUND_FORCE,   /* depth of potential for lipids */
			&(pressure->sphere),
			&(density->sphere),
			NULL
			);
  }

  /*************************************************/
  /******** Cylindric version *********************/
  /*************************************************/
  if (global_data->sbound_droff[R_CYL]>SMALL) 
  {
    calc_cylindric_sbound(my_data,
			  global_data,
			  global_data->sbound_droff + 11,
			  SBOUND_P1_H2O,	/* width of potential */
			  SBOUND_FORCE_H2O, 	/* depth of potential */
			  SBOUND_P1,	  /* width of potential */
			  SBOUND_FORCE,   /* depth of potential */
			  pressure,
			  density
			  );

  }

  /*************************************************/
  /******** Arbitrarily shaped version *************/
  /*************************************************/
  if (global_data->sbound_droff[R_SPHERE]> (-1) && global_data->sbound_droff[R_SPHERE] < -SMALL)
    {

    /*** commented out untill correct control parameter 
                 parameters:
     			SBOUND_P1_H2O       width of potential 
                        SBOUND_FORCE_H2O    depth of potential 
                        SBOUND_P1           width of potential 
                        SBOUND_FORCE        depth of potential 
	
    calc_arbshaped_sbound(my_data,
			global_data,
			SBOUND_P1_H2O,	
			SBOUND_FORCE_H2O, 
			SBOUND_P1,	
			SBOUND_FORCE, 
			&(pressure->sphere),
			&(density->sphere)
			);
    ***/

    calc_numerical_sbound (my_data,
			   global_data,
			   Gamma
			  );
#if debug == -1
	errout("boundary: After calc_numerical_sbound.\n");
#endif

  }
 
 if (global_data->sbound_droff[R_SPHERE]< (-1))
  {
    calc_arbshaped_sbound(my_data,
                        global_data,
                        SBOUND_P1_H2O,
                        SBOUND_FORCE_H2O,
                        SBOUND_P1,
                        SBOUND_FORCE,
                        &(pressure->sphere),
                        &(density->sphere)
                        );
  }



  /*************************************************/
  /********* Version with planes *******************/
  /*************************************************/
  if (any_planar_sbound_set(global_data->sbound_droff)) 
  {
    calc_planar_sbound (my_data,
			global_data,
			global_data->sbound_droff,
			pressure,
			density
			);

  }

/******** START: PATCH FOR ADDL CONSTANT FORCE ***********/
#if (PATCH_CF == TRUE)
  {
    /* Due to the inverted sign in the integration routines, the sign
     * is just opposite of what you would think!!! */
    int    i;

    for (i = 0; i < my_data->no_atoms; i++)
    {
      if (my_data->y_new[i] > middle_plane_dist)
      {
	my_data->force_restr[4*i+2] += const_force;
      }
      if (my_data->y_new[i] < -middle_plane_dist)
      {
	my_data->force_restr[4*i+2] -= const_force;
      }
    }
  }
#endif
/******** END: PATCH FOR ADDL CONSTANT FORCE ***********/

#if DEBUG == 2
  txtout("-X: %f\n",global_data->sbound_droff[XM]);
#endif

  ini_range(global_data->sbound_range); /* reset the range for the next integration step */

#if DEBUG == 8
  if(average_torque_count > 0)
  errout("%d: Average force exerted by torques: %f\n",
         my_number,average_torque_force / (double) average_torque_count);
#endif

} /* end of boundary() */


/************************************************************************/
void   zero_pressure (PRESSURE *pressure)
{
  int i,j;

#if DEBUG == 1
  errout("zero_pressure\n");
#endif

  pressure->sphere = 0.0;
  
  pressure->cyl_rad = 0.0;

  for (i=0;i<6;i++)
    pressure->plane_planar[i] = 0.0; 
  for (i=0;i<12;i++)
    pressure->plane_edge[i] = 0.0;
  for (i=0;i<8;i++)
    pressure->plane_corner[i] = 0.0;
  /*rok140696,rok251096,rok051196*/
  for(i=0;i<3;i++)
    for(j=0;j<3;j++)
      {
      pressure->virial[i][j] = 0.0;
      pressure->virial_ext[i][j] = 0.0;
      pressure->virial_ekin[i][j] = 0.0;
      }
  /*rok241096,rok251096,rok051196*/
      pressure->virial_tot = 0.0;
      pressure->virial_ext_tot = 0.0;
      pressure->virial_ekin_tot = 0.0;
  return;
}


/**************************************************************************/
/*** computes the areas for pressure calculation ***/
static void compute_area (PRESSURE *area, double *sbound_droff)
{ /* is called on the MASTER */
/** sbound_droff[]:  -x, +x, -y, +y, -z, +z , ori_x, ori_y, ori_z, radius, curve, x_cyl  y_cyl  R_cyl
 **                   0   1   2   3   4   5       6      7      8       9     10     11     12     13
 **/

  /* If multiple SBOUNDs are set (e.g., sphere intersected with planes)
   * we have a serious problem. This case is not yet implemented!
   */
  
  zero_pressure(area);
   
  /* sphere */
  if (sbound_droff[R_SPHERE]>SMALL) 
  {
    area->sphere = 4.0*PI*sbound_droff[R_SPHERE]*sbound_droff[R_SPHERE];
    return;
  }

  /* cylinder */
  if (sbound_droff[R_CYL]>SMALL) 
  {
    double h;
    h = sbound_droff[ZP]-sbound_droff[ZM]; /* length of the cylinder */
    area->cyl_rad = 2.0*PI*sbound_droff[R_CYL]*h;
    area->plane_planar[ZM] = area->plane_planar[ZP] = PI*sbound_droff[R_CYL]*sbound_droff[R_CYL];
    return;
  }

  /* planes, round and sharp */
  if (any_planar_sbound_set(sbound_droff)) 
  {
    if (sbound_droff[R_CURVE] < SMALL)
    {/* we have sharp corners, easy to compute,
      * but we get a problem if not all faces are defined 
      */
      area->plane_planar[XM]=(sbound_droff[YP]-sbound_droff[YM]) * (sbound_droff[ZP]-sbound_droff[ZM]);
      area->plane_planar[YM]=(sbound_droff[XP]-sbound_droff[XM]) * (sbound_droff[ZP]-sbound_droff[ZM]);
      area->plane_planar[ZM]=(sbound_droff[XP]-sbound_droff[XM]) * (sbound_droff[YP]-sbound_droff[YM]);
      area->plane_planar[XP]=area->plane_planar[XM]; /* same area for both sides */
      area->plane_planar[XP]=area->plane_planar[YM];
      area->plane_planar[ZP]=area->plane_planar[ZM];
    }
    else /* round corners */
    {
      /* first we set the planar faces */
      area->plane_planar[XM]=(sbound_droff[YP]-sbound_droff[YM]-2.0*sbound_droff[R_CURVE]) * 
                             (sbound_droff[ZP]-sbound_droff[ZM]-2.0*sbound_droff[R_CURVE]);
      area->plane_planar[YM]=(sbound_droff[XP]-sbound_droff[XM]-2.0*sbound_droff[R_CURVE]) * 
      			     (sbound_droff[ZP]-sbound_droff[ZM]-2.0*sbound_droff[R_CURVE]);
      area->plane_planar[ZM]=(sbound_droff[XP]-sbound_droff[XM]-2.0*sbound_droff[R_CURVE]) * 
      			     (sbound_droff[YP]-sbound_droff[YM]-2.0*sbound_droff[R_CURVE]);
      area->plane_planar[XP]=area->plane_planar[XM];
      area->plane_planar[YP]=area->plane_planar[YM];
      area->plane_planar[ZP]=area->plane_planar[ZM];
      
      /* now we set the edge areas */
      area->plane_edge[0] = PI*sbound_droff[R_CURVE]*(sbound_droff[XP]-sbound_droff[XM]);
      area->plane_edge[1] = area->plane_edge[0];
      area->plane_edge[2] = area->plane_edge[0];
      area->plane_edge[3] = area->plane_edge[0];
      area->plane_edge[4] = PI*sbound_droff[R_CURVE]*(sbound_droff[YP]-sbound_droff[YM]);
      area->plane_edge[5] = area->plane_edge[4];
      area->plane_edge[6] = area->plane_edge[4];
      area->plane_edge[7] = area->plane_edge[4];
      area->plane_edge[8] = PI*sbound_droff[R_CURVE]*(sbound_droff[ZP]-sbound_droff[ZM]);
      area->plane_edge[9] = area->plane_edge[8];
      area->plane_edge[10]= area->plane_edge[8];
      area->plane_edge[11]= area->plane_edge[8];
      
      /* now we set the corner areas */
      area->plane_corner[0] = PI*sbound_droff[R_CURVE]*sbound_droff[R_CURVE]/2.0;
      area->plane_corner[1] = area->plane_corner[0] ;
      area->plane_corner[2] = area->plane_corner[0] ;
      area->plane_corner[3] = area->plane_corner[0] ;
      area->plane_corner[4] = area->plane_corner[0] ;
      area->plane_corner[5] = area->plane_corner[0] ;
      area->plane_corner[6] = area->plane_corner[0] ;
      area->plane_corner[7] = area->plane_corner[0] ;
    }
    return;
  }
  return;
}

/**************************************************************************/
/*** computes the pressure from force and area, does unit conversion    ***/
/*** in case of keep_press_const=2 we compute the density               ***/
static void compute_pressure (PRESSURE *pressure, PRESSURE *area, int mode)
{ /* is called on the MASTER !! */
  int i;
  /* converts from kcal/mol/AA/AA^2 to atm*/
  double conv; 
  
  if (mode == PRESSURE_MODE)
    conv = KCALPN*1.0e8/1.013e5; /* real pressure in atm */
  else if (mode == DENSITY_MODE)
    conv = 1.0; /* heavy particle density in SBOUND area */
  else
    conv = 0.0; /* no pressure at all */

  if(area->sphere > SMALL)
  { /* we have a sphere */
    pressure->sphere = pressure->sphere / area->sphere * conv;
    return;
  }

  if(area->cyl_rad > SMALL)
  { /* we have a cylinder */
    pressure->cyl_rad = pressure->cyl_rad / area->cyl_rad * conv;
    if(area->plane_planar[4] > SMALL)
    {
      pressure->plane_planar[4] = pressure->plane_planar[4] / area->plane_planar[4] * conv;
      pressure->plane_planar[5] = pressure->plane_planar[5] / area->plane_planar[5] * conv;
    }
    return;
  }
  
  /* here more sanity checking is necessary. may be own routine, used in read_ctl_file? */
  if(area->plane_corner[0] > SMALL)
  { /* we have round corners */
    for(i=0;i<6;i++)
      pressure->plane_planar[i] = pressure->plane_planar[i] / area->plane_planar[i] * conv;
    for(i=0;i<12;i++)
      pressure->plane_edge[i]   = pressure->plane_edge[i]   / area->plane_edge[i]   * conv;
    for(i=0;i<8;i++)
      pressure->plane_corner[i] = pressure->plane_corner[i] / area->plane_corner[i] * conv;
    return;
  }
  
  /* we have either sharp corners or no SBOUND at all */
  for(i=0;i<6;i++)
    if(area->plane_planar[i] > SMALL)
      pressure->plane_planar[i] = pressure->plane_planar[i] / area->plane_planar[i] * conv;
    else
      pressure->plane_planar[i] = 0.0;
  
  return;
}

/**************************************************************************/
void print_pressure(PRESSURE *pressure, double *sbound_droff, int mode)
{
  char Mode[255], unit[255];
  if (mode == PRESSURE_MODE)
  {
    strcpy(Mode,"Pressure");
    strcpy(unit,"[atm]");
  }
  else if (mode == DENSITY_MODE)
  {
    strcpy(Mode,"Density");
    strcpy(unit,"");
  }
  else
  {
    errout("Fatal error: Wrong mode %d in print_pressure!\n");
    pex(-1);
  }
  
  /* sphere */
  if ((sbound_droff[R_SPHERE]>SMALL)&&(mode==PRESSURE_MODE))
  {
    txtout("%s sphere.....: % 11.5g %s\n",Mode,pressure->sphere,unit);
    return;
  }
  
  /* cylinder */
  if (sbound_droff[R_CYL]>SMALL) 
  {
    txtout("%s cylinder...: % 11.5g   % 11.5g   % 11.5g %s\n",Mode,
            pressure->cyl_rad,pressure->plane_planar[4],pressure->plane_planar[5],
	    unit);
    return;
  }

  /* planes, round and sharp */
  if (any_planar_sbound_set(sbound_droff)) 
  {
    if (sbound_droff[R_CURVE] < SMALL)
    {/* we have sharp corners, easy to compute */
      txtout("%s on planes..: %11.5g  %11.5g  %11.5g  %11.5g  %11.5g  %11.5g %s\n",Mode,
	      pressure->plane_planar[XM],pressure->plane_planar[XP],
	      pressure->plane_planar[YM],pressure->plane_planar[YP],
	      pressure->plane_planar[ZM],pressure->plane_planar[ZP],
	      unit);
      return;
    }
    else
    {
      txtout("%s w. corners1: %11.5g  %11.5g  %11.5g  %11.5g  %11.5g  %11.5g %s\n",Mode,
	      pressure->plane_planar[XM],pressure->plane_planar[XP],
	      pressure->plane_planar[YM],pressure->plane_planar[YP],
	      pressure->plane_planar[ZM],pressure->plane_planar[ZP],
	      unit);
      txtout("%s w. corners2: %11.5g  %11.5g  %11.5g  %11.5g  %11.5g  %11.5g %s\n",Mode,
	      pressure->plane_edge[0],pressure->plane_edge[1],
	      pressure->plane_edge[2],pressure->plane_edge[3],
	      pressure->plane_edge[4],pressure->plane_edge[5],
	      unit);
      txtout("%s w. corners3: %11.5g  %11.5g  %11.5g  %11.5g  %11.5g  %11.5g %s\n",Mode,
	      pressure->plane_edge[6],pressure->plane_edge[7],
	      pressure->plane_edge[8],pressure->plane_edge[9],
	      pressure->plane_edge[10],pressure->plane_edge[11],
	      unit);
      txtout("%s w. corners4: %11.5g  %11.5g  %11.5g  %11.5g  %11.5g  %11.5g  %11.5g  %11.5g %s\n",Mode,
	      pressure->plane_corner[0],pressure->plane_corner[1],
	      pressure->plane_corner[2],pressure->plane_corner[3],
	      pressure->plane_corner[4],pressure->plane_corner[5],
	      pressure->plane_corner[6],pressure->plane_corner[7],
	      unit);
      return;
    }
  }
  return; /* no SBOUND at all, nothing to print. */   
}


/**************************************************************************/
int fprint_pressure(FILE *to_file, PRESSURE *pressure)
{
  return (
   fprintf(to_file," %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g %.7g",
   pressure->sphere,
   pressure->cyl_rad,
   pressure->plane_planar[XM],
   pressure->plane_planar[XP],
   pressure->plane_planar[YM],
   pressure->plane_planar[YP],
   pressure->plane_planar[ZM],
   pressure->plane_planar[ZP],
   pressure->plane_edge[0],
   pressure->plane_edge[1],
   pressure->plane_edge[2],
   pressure->plane_edge[3],
   pressure->plane_edge[4],
   pressure->plane_edge[5],
   pressure->plane_edge[6],
   pressure->plane_edge[7],
   pressure->plane_edge[8],
   pressure->plane_edge[9],
   pressure->plane_edge[10],
   pressure->plane_edge[11],
   pressure->plane_corner[0],
   pressure->plane_corner[1],
   pressure->plane_corner[2],
   pressure->plane_corner[3],
   pressure->plane_corner[4],
   pressure->plane_corner[5],
   pressure->plane_corner[6],
   pressure->plane_corner[7]));
}

/**************************************************************************/
void ini_constant_pressure(double *sbound_droff, int mode)
{ /* here we initialize the local static data structure */
  
  if (mode == PRESSURE_MODE) /* only for real constant PRESSURE mode! */
  {
    /* sphere */
    if (sbound_droff[R_SPHERE]>SMALL) 
    {
      errout("Constant pressure for SPHERE not yet implemented!");
      pex(-2);
    }
    
    /* cylinder */
    if (sbound_droff[R_CYL]>SMALL) 
    {
      errout("Constant pressure for CYLINDER not yet implemented!");
      pex(-2);
    }
  
    /* planes, round and sharp */
    if (any_planar_sbound_set(sbound_droff)) 
    { int i;
      for (i=0; i<6; i++)
      {
	old_positions.plane_planar[i] = sbound_droff[i];
      }
    }
  }
  else if ((mode == DENSITY_MODE)||(mode == SPECIAL_MODE))
  { 
    int i;
    old_positions.sphere = 0.0;
    old_positions.cyl_rad = 0.0;    
    for (i=0; i<6; i++)
    {
      old_positions.plane_planar[i] = 0.0;
    }
    for (i=0; i<12; i++)
    {
      old_positions.plane_edge[i] = 0.0;
    }
    for (i=0; i<8; i++)
    {
      old_positions.plane_corner[i] = 0.0;
    }
    
    first_call_to_density = TRUE;
  }

  return;
}

#define INTEGRATEPISTON(ko_old,ko_new,F_h_new)\
ko_brand_new = ko_new + ((((1.0-g)*(ko_new-ko_old))-(F_h_new[i]*c))*alpha); \
v = half_rez_time_step * (ko_brand_new-ko_old); \
kin_energy+=((v*v)*m); \
ko_old=ko_new; \
ko_new=ko_brand_new;

/**************************************************************************/
static void constant_pressure( PRESSURE *pressure,     /* instantaneous pressure, already converted to atm */
			PRESSURE *area,
			LOAD_MSG_ALL_NODE *global_data)
{ 
  PRESSURE local_pressure;
  int      i;
  double conv = KCALPN*1.0e8/1.013e5; 

  /* sphere */
  if (global_data->sbound_droff[R_SPHERE]>SMALL) 
  {
    /* This is a tricky part:
      * If we have the MOVING SBOUND PATCH active AND the constant
      * pressure mode is switched on in the ctl.lis file, then we misuse
      * the constant pressure mode, ignore the outside pressure value
      * and instead move the sbound position every integration step.
      * Please keep in mind that this function is called on the MASTER! 
      * In case of MOVING SBOUND PATCH active we keep the unused but
      * declared variables to warn the user at compile time that the
      * patch is active! */
      
/******** START: PATCH FOR MOVING SBOUND ***********/
#if (PATCH_MS==TRUE)
    global_data->sbound_droff[R_SPHERE] += sboundmove; /* use + or - as you need */
    if (++sboundcounter == 100) 
    {
      sboundcounter = 0;
      txtout("Current Sbound[%d] = %f\n",
	      R_SPHERE, global_data->sbound_droff[R_SPHERE]);
    }
#else
/******** END: PATCH FOR MOVING SBOUND ***********/
    errout("Constant pressure for SPHERE not yet implemented!");
    pex(-2);
#endif
  }
  
  /* cylinder */
  if (global_data->sbound_droff[R_CYL]>SMALL) 
  {
    errout("Constant pressure for CYLINDER not yet implemented!");
    pex(-2);
  }

  /* planes, round and sharp */
  if (any_planar_sbound_set(global_data->sbound_droff)) 
  {
    /* This is a tricky part:
      * If we have the MOVING SBOUND PATCH active AND the constant
      * pressure mode is switched on in the ctl.lis file, then we misuse
      * the constant pressure mode, ignore the outside pressure value
      * and instead move the sbound position every integration step.
      * Please keep in mind that this function is called on the MASTER! 
      * In case of MOVING SBOUND PATCH active we keep the unused but
      * declared variables to warn the user at compile time that the
      * patch is active! */
      
/******** START: PATCH FOR MOVING SBOUND ***********/
#if (PATCH_MS==TRUE)
    global_data->sbound_droff[sboundindex1] += sboundmove;
    global_data->sbound_droff[sboundindex2] -= sboundmove;
    if (++sboundcounter == 100) 
    {
      sboundcounter = 0;
      txtout("Current Sbound[%d] = %f  Sbound[%d] = %f\n",
	      sboundindex1, global_data->sbound_droff[sboundindex1],
	      sboundindex2, global_data->sbound_droff[sboundindex2]);
    }
#else
/******** END: PATCH FOR MOVING SBOUND ***********/

    /* \cite{FELL95} strongly suggests to use Langevin dynamics on the piston
     * to avoid strong fluctuations of the pressure */
         
    /* the following variables are used in INTEGRATEPISTON */
    double m = PISTON_MASS;
    double c = global_data->time_step * global_data->time_step / m;
    double fbeta = PISTON_FRICTION * TIMEFAC;
    double g = fbeta * global_data->time_step / 2.0; /* the friction constant, 
    						      * 0.0=no friction */
    double rnd_force_std_deviation = sqrt(TWO * m * fbeta * KBOLTZ * 
			global_data->target_temp/global_data->time_step);
    double alpha = ONE / (ONE + g); 
    double ko_brand_new,v; 
    double half_rez_time_step = 0.5/global_data->time_step;
    double kin_energy=0.0;
    
    double xyzforce[3];
      
    for (i=0; i<6; i++)
    {
      /* We move the SBOUND *ONLY* if the outside pressure is not exactly 0.0.
         This way we can specify s.th. like constant normal pressure and constant surface area! */
      if (global_data->outside_pressure.plane_planar[i] != 0.0)
      {
	local_pressure.plane_planar[i] = pressure->plane_planar[i]-global_data->outside_pressure.plane_planar[i];
	/* now we convert back to forces */
	local_pressure.plane_planar[i] = local_pressure.plane_planar[i] * area->plane_planar[i] / conv;
	/* now we have to get the sign right: invert for the + planes */
	local_pressure.plane_planar[i] *= 1 - (i%2)*2;
	
	/** sbound_droff[]:  -x, +x, -y, +y, -z, +z , ori_x, ori_y, ori_z, radius, curve, x_cyl  y_cyl  R_cyl
	  **                   0   1   2   3   4   5       6      7      8       9     10     11     12     13
	  **/
	kin_energy=0.0;
	
	/* now we add the random force */
	calc_random_force (rnd_force_std_deviation, xyzforce);
	local_pressure.plane_planar[i] -= xyzforce[0];
  
	INTEGRATEPISTON ( old_positions.plane_planar[i],
			  global_data->sbound_droff[i],
			  local_pressure.plane_planar);
#if DEBUG == 2
	errout("Plane %d, energy: %f; ", i, kin_energy);
#endif
	} /* if... */
    } /* for loop */
#if DEBUG == 2
      errout("\n");
#endif

#endif /* SBOUND MOVE */
  }

  return;
}

/**************************************************************************/
static void constant_density ( PRESSURE *density,     /* instantaneous density */
			PRESSURE *area,
			LOAD_MSG_ALL_NODE *global_data)
{ /* this routine is called on the MASTER */
  int      i;
  double delta_press = 0.1; /* density only adjusted if more than 10% deviation */
  double delta_r = 0.0004;  /* move per step in AA. Should be scaled with integration step size */

  /** sbound_droff[]:  -x, +x, -y, +y, -z, +z , ori_x, ori_y, ori_z, radius, curve, x_cyl  y_cyl  R_cyl
   **                   0   1   2   3   4   5       6      7      8       9     10     11     12     13
   **/

  if (first_call_to_density)
  { /* in this case we initialize the average pressure values */
    first_call_to_density = FALSE;
    old_positions.sphere = density->sphere;
    old_positions.cyl_rad = density->cyl_rad;    
    for (i=0; i<6; i++)
    {
      old_positions.plane_planar[i] = density->plane_planar[i];
    }
    for (i=0; i<12; i++)
    {
      old_positions.plane_edge[i] = density->plane_edge[i];
    }
    for (i=0; i<8; i++)
    {
      old_positions.plane_corner[i] = density->plane_corner[i];
    }
  }
  else
  { /* now we make a sliding average */
    old_positions.sphere = (density->sphere + old_positions.sphere) / 2.0;
    old_positions.cyl_rad = (density->cyl_rad + old_positions.cyl_rad) / 2.0;    
    for (i=0; i<6; i++)
    {
      old_positions.plane_planar[i] = (density->plane_planar[i] + old_positions.plane_planar[i]) / 2.0;
    }
    for (i=0; i<12; i++)
    {
      old_positions.plane_edge[i] = (density->plane_edge[i] + old_positions.plane_edge[i]) / 2.0;
    }
    for (i=0; i<8; i++)
    {
      old_positions.plane_corner[i] = (density->plane_corner[i] + old_positions.plane_corner[i]) / 2.0;
    }
  }

  /* sphere */
  if (global_data->sbound_droff[R_SPHERE]>SMALL) 
  {
    if (old_positions.sphere > ((global_data->outside_pressure.sphere) * (1.0+delta_press)))
      global_data->sbound_droff[R_SPHERE] += delta_r;
    else
    if (old_positions.sphere < ((global_data->outside_pressure.sphere) * (1.0-delta_press)))
      global_data->sbound_droff[R_SPHERE] -= delta_r;
  }
  
  /* cylinder */
  if (global_data->sbound_droff[R_CYL]>SMALL) 
  {
    if (old_positions.cyl_rad > ((global_data->outside_pressure.cyl_rad) * (1.0+delta_press)))
      global_data->sbound_droff[R_CYL] += delta_r;
    else
    if (old_positions.cyl_rad < ((global_data->outside_pressure.cyl_rad) * (1.0-delta_press)))
      global_data->sbound_droff[R_CYL] -= delta_r;
  }

  /* planes, round and sharp (we look only at the density in the planar parts) */
  for (i=0; i < 6; i++)
    /* We move the SBOUND *ONLY* if the outside pressure (density) is not exactly 0.0.
	This way we can specify s.th. like constant normal density and constant surface area! */
    if((area->plane_planar[i] > SMALL) && (global_data->outside_pressure.plane_planar[i] != 0.0))
    {
      if (old_positions.plane_planar[i] > ((global_data->outside_pressure.plane_planar[i]) * (1.0+delta_press)))
      {
	if ((i%2) == 0) /* negative planes */
	{
	  global_data->sbound_droff[i] -= delta_r;
	}
	else /* positive plane */
	{
	  global_data->sbound_droff[i] += delta_r;
	}
      }

      if (old_positions.plane_planar[i] < ((global_data->outside_pressure.plane_planar[i]) * (1.0-delta_press)))
      {
	if ((i%2) == 0) /* negative planes */
	{
	  global_data->sbound_droff[i] += delta_r;
	}
	else  /* positive plane */
	{
	  global_data->sbound_droff[i] -= delta_r;
	}
      }
    }

  return;
}

/**************************************************************************/
static void constant_special (	LOAD_MSG_ALL_NODE *global_data, 
				double *special_YP_pos, double *special_YM_pos,
				int     special_YP_count,  int  special_YM_count)
{ /* Here we use the membrane itself as a pressure sensor:
   * We try to keep the P43 distance at the value in global_data->outside_pressure.plane_planar[2]
   * This function is called on the MASTER 
   */
  double delta_r = 0.000003;  /* move per step in AA. Should be scaled with integration step size */
  double P43_dist_dev;

  /** sbound_droff[]:  -x, +x, -y, +y, -z, +z , ori_x, ori_y, ori_z, radius, curve, x_cyl  y_cyl  R_cyl
   **                   0   1   2   3   4   5       6      7      8       9     10     11     12     13
   **/


  /* sphere */
  if (global_data->sbound_droff[R_SPHERE]>SMALL) 
  {
    errout("Constant special mode for SPHERE not yet implemented!");
    pex(-2);
  }
  
  /* cylinder */
  if (global_data->sbound_droff[R_CYL]>SMALL) 
  {
    errout("Constant special mode for CYLINDER not yet implemented!");
    pex(-2);
  }

  if(any_planar_sbound_set(global_data->sbound_droff))
  {
    /* planes, round and sharp (we look only at the P43 distance) */
    /* We move the Y SBOUND *ONLY*  if more than 0.3 off the ideal distance */

#if DEBUG == 3
errout("%d: Special: YP_count = %d, YM_count=%d, YP_pos = %f, YM_pos = %f\n",my_number, 
	special_YP_count, special_YM_count, *special_YP_pos, *special_YM_pos);
#endif
    if ((special_YP_count==0) || (special_YM_count==0))
    { /* we did not have any entries, probably we are switched off */
      *special_YP_pos =  global_data->outside_pressure.plane_planar[2]/2.0; /* we set them so that */
      *special_YM_pos = -global_data->outside_pressure.plane_planar[2]/2.0; /* they will match */
    }
    else
    {    
    *special_YP_pos = *special_YP_pos / (double)(special_YP_count); /* we have to divide because we */
    *special_YM_pos = *special_YM_pos / (double)(special_YM_count); /* added up the contributions from all nodes */
    }
    
#if DEBUG == 3
errout("%d: P43_dist should be %f, but is %f\n",my_number, global_data->outside_pressure.plane_planar[2],
					    *special_YP_pos - *special_YM_pos);
#endif

    if ( fabs((P43_dist_dev=((*special_YP_pos - *special_YM_pos) - global_data->outside_pressure.plane_planar[2]))) > 0.3)
    {
      if (P43_dist_dev > 0.0) /* dist too big, shrink planes! */
      {
	  global_data->sbound_droff[YP] -= delta_r;
	  global_data->sbound_droff[YM] += delta_r;
      }
     else		  /* dist too small, increase plane distance! */
      {
	  global_data->sbound_droff[YP] += delta_r;
	  global_data->sbound_droff[YM] -= delta_r;
      }
    }
  }
  return;
}


/**************************************************************************/
static void calc_spheric_sbound ( LOAD_MSG_ONE_NODE *my_data,
				  LOAD_MSG_ALL_NODE *global_data,
				  int start, int stop,  /* only atoms start <= i < stop will be considered */
				  double *sboundpos,	/** x,y,z,R **/
				  double sbound_p1_h2o,	/* width of potential */
				  double sbound_force_h2o,	/* depth of potential */
				  double sbound_p1_other,	/* width of potential */
				  double sbound_force_other,	/* depth of potential */
				  double *pressure,	/* pressure, radial element */
				  double *SB_num,	/* atom count in SB region */
				  double *forces	/* if NULL then we add the forces directly.
				  			   if !NULL then we don't add the forces and 
							   return them in forces[4] instead. */
)
{
  int    i;
  double delta,delta2;
  double dist,
         sboundx,sboundy,sboundz,
         sboundr,sboundrad,
         fac;
  double minoffset;
  double sbound_p1, sbound_force;
  VECTOR radius;


  sboundx = sboundpos[0]; /* x, y, z coordinates of origin of sphere */
  sboundy = sboundpos[1];
  sboundz = sboundpos[2];

  for (i = start; i <= stop; i++)
  {
    int	   ty= (int)my_data->typ[i];
    double m = global_data->AB[ty][ty];
    
    if(m > 4.0) /* we ignore hydrogen atoms */
    {
      radius.x = my_data->x_new[i] - sboundx; /* distance to origin of sphere */
      radius.y = my_data->y_new[i] - sboundy;
      radius.z = my_data->z_new[i] - sboundz;
      dist = vector_length (&radius); /* distance to origin of sphere */

      if(use_water_params(i,my_data))
      {
	sbound_p1 = sbound_p1_h2o;
	sbound_force = sbound_force_h2o;
        sboundr = sboundpos[3] - SBOUND_FILLING_H2O; /* sboundpos[3] is given by the user
							and is the value how far the user thinks
							that atoms should fill up the SBOUND region.
							SBOUND_FILLING_H2O is the offset from minimum to
							how far water actually fills the SBOUND under 
							normal conditions.
							sboundr therefore gives the position where
							the user sets the SBOUND minimum to be */
      }
      else
      {
	sbound_p1 = sbound_p1_other;
	sbound_force = sbound_force_other;
        sboundr = sboundpos[3] - SBOUND_FILLING; /* Here we take into account that the SBOUND2
						    potential is not completely filled to the rim. 
						    sboundr is how far the SBOUND2 potential is 
						    actually filled with lipids. */
      }
      minoffset = sqrt (sbound_p1 / 2.0); /* position of the minimum relative
      					     to where the SBOUND starts to be nonzero */
      sboundrad = sboundr - minoffset ; /* this is  where the SBOUND actually starts */

#if DEBUG == 4
printf ("CS: forces=%p, start=%d, stop=%d, m=%f, dist=%f, sbr=%f\n",forces,start,stop,m,dist, sboundrad);
#endif
      if (dist >= sboundrad)
      { /* only now are we inside the SBOUND region */
	delta = dist - sboundrad; /* how far into the SBOUND we are */
	delta2 = delta * delta;
	fac = sbound_force * 2.0 * delta * (2.0 * delta2 - sbound_p1) / dist;
	if (forces == NULL)
	{
	  my_data->force_sbound[4*i]   += sbound_force * delta2 * (delta2 - sbound_p1);
	  my_data->force_sbound[4*i+1] += radius.x * fac;
	  my_data->force_sbound[4*i+2] += radius.y * fac;
	  my_data->force_sbound[4*i+3] += radius.z * fac;
	}
	else
	{
	  forces[0] = sbound_force * delta2 * (delta2 - sbound_p1);
	  forces[1] = radius.x * fac;
	  forces[2] = radius.y * fac;
	  forces[3] = radius.z * fac;
#if DEBUG == 4
printf ("CS: f0=%f, f1=%f(%f), f2=%f, f3=%f\n",forces[0],forces[1],radius.x * fac,forces[2],forces[3]);
#endif
	}

       	/* now we add the force for pressure calculations
	  * the sign is also positive here, since the force sign
	  * will be inverted in the integrator.
	  * fac is either positive or negative.  */
	*pressure += (dist * fac);
        (*SB_num)++; /* count the heavy atoms in the SBOUND */
	
	/* Now we add the torque forces: */
	compute_torque_forces (my_data, i, &radius, sboundpos[3]-dist);
	/* the last parameter gives the distance to the wall */
      }
    }
  }
  return;
}

/**************************************************************************/
static void calc_cylindric_sbound ( LOAD_MSG_ONE_NODE *my_data,
				    LOAD_MSG_ALL_NODE *global_data,
				    double *sboundpos, /** x,y,R **/
				    double sbound_p1_h2o,	  /* width of potential */
				    double sbound_force_h2o,/* depth of potential */
				    double sbound_p1_other,	  /* width of potential */
				    double sbound_force_other,/* depth of potential */
				    PRESSURE *pressure,  /* pressure elements */
				    PRESSURE *density    /* density elements */
)
{
  /* the cylinder is parallel to the Z axis! */
  int    i;
  double delta, delta2;
  double sbound_p1, sbound_force;
  double dist,sboundx,sboundy,sboundrad,fac;
  double minoffset, sboundr ;
  VECTOR radius;

  sboundx = sboundpos[0];
  sboundy = sboundpos[1];

  for (i = 0; i < my_data->no_atoms; i++) 
  {
    int	ty    = (int)my_data->typ[i];
    double	m     = global_data->AB[ty][ty];
    if(m > 4.0) /* we ignore hydrogen atoms */
    {
      radius.x = my_data->x_new[i] - sboundx;
      radius.y = my_data->y_new[i] - sboundy;
      radius.z = 0.0;
      dist = vector_length(&radius); /* distance to axis of cylinder */

      if(use_water_params(i,my_data))
      {
	sbound_p1 = sbound_p1_h2o;
	sbound_force = sbound_force_h2o;
        sboundr = sboundpos[2] - SBOUND_FILLING_H2O; /* sboundpos[2] is given by the user
							and is the value how far the user thinks
							that atoms should fill up the SBOUND region.
							SBOUND_FILLING_H2O is the offset from minimum to
							how far water actually fills the SBOUND under 
							normal conditions.
							sboundr therefore gives the position where
							the user sets the SBOUND minimum to be */
      }
      else
      {
	sbound_p1 = sbound_p1_other;
	sbound_force = sbound_force_other;
        sboundr = sboundpos[2] - SBOUND_FILLING;
      }
      minoffset = sqrt(sbound_p1 / 2.0);
      sboundrad = sboundr - minoffset ;
      if (dist >= sboundrad)
      { 
	delta = dist - sboundrad; /* position of the minimum relative
      			             to where the SBOUND starts to be nonzero */
	delta2 = delta * delta; /* this is  where the SBOUND actually starts */
	my_data->force_sbound[4*i] += sbound_force * delta2 * (delta2 - sbound_p1);
	fac = sbound_force * 2.0 * delta * (2.0 * delta2 - sbound_p1) / dist;
	my_data->force_sbound[4*i+1] += radius.x * fac;
	my_data->force_sbound[4*i+2] += radius.y * fac;
	
	/* now we add the force for pressure calculations
	  * the sign is also positive here, since the force sign
	  * will be inverted in the integrator.
	  * fac is either positive or negative.  */
	pressure->cyl_rad += (dist * fac);
        (density->cyl_rad)++; /* count atoms in SBOUND */

	/* Now we add the torque forces: */
	compute_torque_forces (my_data, i, &radius, sboundpos[2]-dist);
	/* the last parameter gives the distance to the wall */
      }
    }
  }
  return;
}



/**************************************************************************/
/***** returns  sbound_energy ***/
#define DISTANCE_TIMES_WATER_SHELL_THICKNESS 2.0
static double calc_arbshaped_sbound (LOAD_MSG_ONE_NODE *my_data,
				     LOAD_MSG_ALL_NODE *global_data,
			             double sbound_p1_h2o,	/* width of potential */
			             double sbound_force_h2o,	/* depth of potential */
			             double sbound_p1_other,	/* width of potential */
			             double sbound_force_other,	/* depth of potential */
				     double *pressure,	/* pressure radial element */
				     double *SB_num	/* atom number in SB region */
)
{
  int    i;
  double SCALE_SBOUND = 0.6666666667;
  double sbound_energy = 0.0;
  double fac;
  double minoffset, delta, delta2, sboundrad2;
  double sbound_p1, sbound_force;
  double gx,gy,gz,dd,x0,y0,z0,x1,y1,z1;
  double estimated_distance;
  double gr,cur,s2,kk,xx0,yy0,yy1,yy2;
  double curvature;
  double dist_0,dist_1,dd_0;

  for (i = 0; i < my_data->no_atoms; i++)
  {
    int	   ty= (int)my_data->typ[i];
    double m = global_data->AB[ty][ty];
    double e_help;
    if(m > 4.0) /* we ignore hydrogen atoms */
    {
          x0 = my_data->x_new[i];
          y0 = my_data->y_new[i];
          z0 = my_data->z_new[i];
	  dd=density_grad_curve(global_data,x0,y0,z0,&gx,&gy,&gz,&gr,&cur,&curvature);
	  yy0 = dd;
	  dd_0 = dd;
	  yy1 = sqrt(gx*gx+gy*gy+gz*gz);
	  yy2 = cur;
	  s2  = yy0*yy0/(yy1*yy1-yy0*yy2);
	  kk  = yy0*exp(0.5*yy1*yy1/(yy1*yy1-yy0*yy2));
	  xx0 = yy0*yy1/(yy1*yy1-yy0*yy2);
	  /* dist_0 = fabs(xx0 - sqrt(2.0*s2*log(kk/global_data->arbs_surface_level))); xxx */
	  dist_0 = - xx0 + sqrt(2.0*s2*log(kk/global_data->arbs_surface_level));
	  if (dist_0<DISTANCE_TIMES_WATER_SHELL_THICKNESS* global_data->arbs_water_shell_thickness) {
            x1 = x0 - dist_0*gx/yy1;
            y1 = y0 - dist_0*gy/yy1;
            z1 = z0 - dist_0*gz/yy1;
	    dd=density_grad_curve(global_data,x1,y1,z1,&gx,&gy,&gz,&gr,&cur,&curvature);
	    yy0 = dd;
	    yy1 = sqrt(gx*gx+gy*gy+gz*gz);
	    yy2 = cur;
	    s2  = yy0*yy0/(yy1*yy1-yy0*yy2);
	    kk  = yy0*exp(0.5*yy1*yy1/(yy1*yy1-yy0*yy2));
	    xx0 = yy0*yy1/(yy1*yy1-yy0*yy2);
	    /* dist_1 = fabs(xx0 - sqrt(2.0*s2*log(kk/global_data->arbs_surface_level))); xxx */
	    dist_1 = - xx0 + sqrt(2.0*s2*log(kk/global_data->arbs_surface_level));
	    estimated_distance = dist_0 + dist_1;
            gx =  gx / yy1 ;
            gy =  gy / yy1 ;
            gz =  gz / yy1 ;
	    }
	  else {
	    curvature=0.0;
	    gx=0.0;
	    gy=0.0;
	    gz=0.0;
            estimated_distance=10.0;
	    }
      /* txtout("curve %f\n",curvature); */
      /* in estimated_distance we now have the distance of atom i
	 to the boundary. (gx,gy,gz) is a normal to the boundary.
	 curvature is the avergage boundary curvature to be 
	 used for atom i.   */

      if(use_water_params(i,my_data))
      {
	sbound_p1 = sbound_p1_h2o;
	sbound_force = sbound_force_h2o;
      }
      else
      {
	sbound_p1 = sbound_p1_other;
	sbound_force = sbound_force_other;
      }
      minoffset = sqrt (sbound_p1 / 2.0);

      /* minoffset is the distance from the beginning to the minimum of
       * the sbound potential  */

      if (global_data->sbound_droff[R_SPHERE] >= -1.9 && global_data->sbound_droff[R_SPHERE] < -1.0)
        {


      estimated_distance = estimated_distance - 2.0;
 
      /* now the sbound minimum is 2 A moved towards the origin */
 
      sboundrad2 = estimated_distance*estimated_distance;
      if (estimated_distance <= minoffset)
      {  
	delta = minoffset - estimated_distance;
	delta2 = delta * delta;
	e_help = SCALE_SBOUND * sbound_force * delta2 * (delta2 - sbound_p1);
	sbound_energy += e_help;
	my_data->force_sbound[4*i] += e_help;
	fac = SCALE_SBOUND * sbound_force * 2.0 * delta * (2.0 * delta2 - sbound_p1);
	my_data->force_sbound[4*i+1] -= gx * fac;
	my_data->force_sbound[4*i+2] -= gy * fac;
	my_data->force_sbound[4*i+3] -= gz * fac;

       	/* now we add the force for pressure calculations
	  * the sign is also positive here, since the force sign
	  * will be inverted in the integrator.
	  * fac is either positive or negative.  */
	*pressure += (fac);
        (*SB_num)++; /* count the heavy atoms in the SBOUND */
      }
    }
   }
  }
  return sbound_energy;
}


/**************************************************************************/
/***** returns  sbound_energy ***/
#define RHO_FOR_SBOUND_INTERPOLATE (0.65*sqrt(3.0))
#define GAUSS_CUTOFF 4.0
#define HEIGHT_OF_SBOUND_CHUTE -0.3
#define MIN_SBOUND_CHUTE -2.0
static double calc_numerical_sbound (LOAD_MSG_ONE_NODE *my_data,
				     LOAD_MSG_ALL_NODE *global_data,
				     double *Gamma)
{
  double sbound_energy,eny_one_atom;
  double x,y,z,xmin,xmax,ymin,ymax,zmin,zmax;
  double dx,dy,dz,dx2,dy2,dd;
  double fx1,fy1,fz1,fx2,fy2,fz2;
  double r2,sig2,sig22,step,norm,ee,eecut,eesig2;
  double eesig2dx,eesig2dy,eesig2dz,tail;
  static double expfkt[20000];
  static int ini=0;
  long int    i,j,k,l;
  long int    jmin,kmin,lmin;
  long int    jmax,kmax,lmax;
  long int    atype,idx;
  long int    end_of_string,diff_string,equal_string;
  char   *my_resname,*wish_resname;
  STRUCT_POTGRID_ELEMENT *pot_el;

  step = global_data->potgrid.stepsize;
  xmin = global_data->potgrid.xmin;
  xmax = global_data->potgrid.xmax;
  ymin = global_data->potgrid.ymin;
  ymax = global_data->potgrid.ymax;
  zmin = global_data->potgrid.zmin;
  zmax = global_data->potgrid.zmax;

  sig2 = 1.0/(2.0*RHO_FOR_SBOUND_INTERPOLATE*RHO_FOR_SBOUND_INTERPOLATE);
  /* cutoff radius is always 4 grid points */
  r2 = (GAUSS_CUTOFF*step) * (GAUSS_CUTOFF*step);
  eecut = exp(-sig2*r2);
  sbound_energy = 0.0;

  if (ini<1) {
    for (i=0; i<20000;i++) {
      ee = (((float) i) / 1000.0) * sig2;  /* i = distance^2 * 1000 */
      expfkt[i] = exp(-ee);
    }
    ini=1;
  }

  for (i = 0; i < my_data->no_atoms; i++) {
    /* we ignore hydrogens (mass lower than 4.0) */
    atype = (int) my_data->typ[i];
    if(global_data->AB[atype][atype] > 4.0) {
      my_resname = my_data->res_string[i];
      j = 0;
      equal_string=FALSE;
      do {
	wish_resname = global_data->sbound_res_string[j]; 
	end_of_string = FALSE;
	diff_string   = FALSE;
        for (k=0;!(end_of_string || diff_string);k++) {
	  end_of_string = (wish_resname[k] == '\0');
	  diff_string  = (wish_resname[k] != my_resname[k]);
	}
	equal_string = (end_of_string && !(diff_string));
	j++;
      } while (!(equal_string || (wish_resname[0]=='\0') || (wish_resname[0]=='*')));
      if (equal_string || (wish_resname[0]=='*')) {
        x = my_data->x_new[i];
        y = my_data->y_new[i];
        z = my_data->z_new[i];
        if ( (x<xmin) || (x>xmax) || (y<ymin) || (y>ymax) || (z<zmin) || (z>zmax) ) {
          errout("Fatal error: atom leaved SBound region.\n");
          errout("atom position: (%f,%f,%f)\n",x,y,z);
          errout("range x (%f,%f), ",xmin,xmax);
          errout("range y (%f,%f), ",ymin,ymax);
          errout("range z (%f,%f)\n\n",zmin,zmax);
          pex(-1);
        }
        jmin = (int) floor((x-xmin)/step) - GAUSS_CUTOFF;
        jmax = jmin + 2*GAUSS_CUTOFF + 1;
        jmin = (jmin > 0) ? jmin : 0; 
        jmax = (jmax < global_data->potgrid.nx) ? jmax : (global_data->potgrid.nx-1); 
        kmin = (int) floor((y-ymin)/step) - GAUSS_CUTOFF;
        kmax = kmin + 2*GAUSS_CUTOFF + 1;
        kmin = (kmin > 0) ? kmin : 0; 
        kmax = (kmax < global_data->potgrid.ny) ? kmax : (global_data->potgrid.ny-1); 
        lmin = (int) floor((z-zmin)/step) - GAUSS_CUTOFF;
        lmax = lmin + 2*GAUSS_CUTOFF + 1;
        lmin = (lmin > 0) ? lmin : 0; 
        lmax = (lmax < global_data->potgrid.nz) ? lmax : (global_data->potgrid.nz-1); 
    
	sig22 =  -2.0 * sig2;
        eny_one_atom = 0.0;
        norm = 0.0;
        fx1 = fy1 = fz1 = 0.0;
        fx2 = fy2 = fz2 = 0.0;
    
        for (j=jmin;j<=jmax;j++) {
          dx = x - (j*step + xmin);
	  dx2 = dx * dx;
          for (k=kmin;k<=kmax;k++) {
	    dy = y - (k*step + ymin);
	    dy2 = dy * dy;
            for (l=lmin;l<=lmax;l++) {
	      dz = z - (l*step + zmin);
	      dd = dx2 + dy2 + dz*dz;
	      if (dd <= r2) {
		dd *= 1000.0;
		idx = (long int)  dd;
		tail = dd - idx;
		ee = (1.0-tail)*expfkt[idx] + tail*expfkt[idx+1];
	        norm += ee - eecut;
	        eesig2 = sig22 * ee;
	        eesig2dx = dx * eesig2;
	        eesig2dy = dy * eesig2;
	        eesig2dz = dz * eesig2;
	        pot_el = global_data->potgrid.pt[j][k][l];
	        if (pot_el != NULL) {
	          eny_one_atom += pot_el->sbound * (ee-eecut);
	          fx1 += pot_el->sbound * eesig2dx;
	          fy1 += pot_el->sbound * eesig2dy;
	          fz1 += pot_el->sbound * eesig2dz;
	        }
	        fx2 += eesig2dx;
	        fy2 += eesig2dy;
	        fz2 += eesig2dz;
	      }
            }
          }
        }
    
        if (norm < SMALL) {
          errout("Fatal error: Can not calculate SBound.\n");
          errout("atom nr %d, norm=%f\n, eny_one_atom=%f",i,norm, eny_one_atom);
          errout("This should really not happen!\n\n");
          pex(-1);
        }
    
        eny_one_atom /= norm;
        sbound_energy += eny_one_atom;
        my_data->force_sbound[4*i  ] += eny_one_atom;
        my_data->force_sbound[4*i+1] += ( fx1 - fx2*eny_one_atom ) / norm;
        my_data->force_sbound[4*i+2] += ( fy1 - fy2*eny_one_atom ) / norm;
        my_data->force_sbound[4*i+3] += ( fz1 - fz2*eny_one_atom ) / norm;

        if (global_data->stochastic && global_data->restarting) {
	  Gamma[i] = 0.0;
	  if (eny_one_atom < HEIGHT_OF_SBOUND_CHUTE) {
            double fbetafac = TWO * KBOLTZ * global_data->target_temp / global_data->time_step;
	    double sclfac = 0.0;
            double fbeta, rfd;
            double xyzforce[3];

	    sclfac = (eny_one_atom - HEIGHT_OF_SBOUND_CHUTE) / MIN_SBOUND_CHUTE;
	    fbeta = sclfac * TIMEFAC * global_data->sbound_fbetagradient;
	    Gamma[i] = fbeta * global_data->time_step / TWO;
	    rfd = sqrt (fbetafac * global_data->AB[atype][atype] * fbeta);

	    calc_random_force (rfd, xyzforce);

	    my_data->force_sbound[4*i+1] -= xyzforce[0];
	    my_data->force_sbound[4*i+2] -= xyzforce[1];
	    my_data->force_sbound[4*i+3] -= xyzforce[2];
	  }
	}
      }
    }
  }
        
  /*** do we have to add the force for pressure calculations here??? ***/

  return sbound_energy;
}

/**************************************************************************/
static double density_grad_curve(LOAD_MSG_ALL_NODE *global_data,
				 double x, double y, double z, 
				 double *gx, double *gy, double *gz, double *g, 
				 double *cr, double *curvature)
{
	int r;
        double sum,prod;
	double w11,w12,w13;
	double w21,w22,w23;
	double w31,w32,w33;
	double t11,t12,t13;
	double t21,t22,t23;
	double t31,t32,t33;
	double a11,a12,a13;
	double a21,a22,a23;
	double a31,a32,a33;
	double m11,m12,m13;
	double m21,m22,m23;
	double m31,m32,m33;
	double cxx,cxy,cxz,cyx,cyy,cyz,czx,czy,czz;
	double dx,dy,dz,ex;
	double g0x,g0y,g0z;
	double v1,v2,v3;
	double temp;

	sum=0.0;
	*gx = 0.0;
	*gy = 0.0;
	*gz = 0.0;
	cxx = 0.0;
	cxy = 0.0;
	cxz = 0.0;
	cyx = 0.0;
	cyy = 0.0;
	cyz = 0.0;
	czx = 0.0;
	czy = 0.0;
	czz = 0.0;
	for (r=0; r<(global_data->arbs_ncodebook); r++) {
          dx=x-global_data->arbs_gauss_y[0][r];
          dy=y-global_data->arbs_gauss_y[1][r];
          dz=z-global_data->arbs_gauss_y[2][r];
	  w11=global_data->arbs_gauss_w[0][0][r];
	  w12=global_data->arbs_gauss_w[0][1][r];
	  w13=global_data->arbs_gauss_w[0][2][r];
	  w21=global_data->arbs_gauss_w[1][0][r];
	  w22=global_data->arbs_gauss_w[1][1][r];
	  w23=global_data->arbs_gauss_w[1][2][r];
	  w31=global_data->arbs_gauss_w[2][0][r];
	  w32=global_data->arbs_gauss_w[2][1][r];
	  w33=global_data->arbs_gauss_w[2][2][r];
	  t11 = w11/global_data->arbs_gauss_rho2[0][r];        /* T := rho2^(-1) * W^T */
	  t12 = w21/global_data->arbs_gauss_rho2[0][r];
	  t13 = w31/global_data->arbs_gauss_rho2[0][r];
	  t21 = w12/global_data->arbs_gauss_rho2[1][r];
	  t22 = w22/global_data->arbs_gauss_rho2[1][r];
	  t23 = w32/global_data->arbs_gauss_rho2[1][r];
	  t31 = w13/global_data->arbs_gauss_rho2[2][r];
	  t32 = w23/global_data->arbs_gauss_rho2[2][r];
	  t33 = w33/global_data->arbs_gauss_rho2[2][r];
          a11 = w11*t11 + w12*t21 + w13*t31;  /* A := W * T    ( == W * rho2^(-1) * W^T)  */
          a12 = w11*t12 + w12*t22 + w13*t32;
          a13 = w11*t13 + w12*t23 + w13*t33;
          a21 = w21*t11 + w22*t21 + w23*t31;
          a22 = w21*t12 + w22*t22 + w23*t32;
          a23 = w21*t13 + w22*t23 + w23*t33;
          a31 = w31*t11 + w32*t21 + w33*t31;
          a32 = w31*t12 + w32*t22 + w33*t32;
          a33 = w31*t13 + w32*t23 + w33*t33;
          v1 = a11*dx + a12*dy + a13*dz;      /* v := A*x */
          v2 = a21*dx + a22*dy + a23*dz;
          v3 = a31*dx + a32*dy + a33*dz;
	  prod = dx*v1 + dy*v2 + dz*v3;       /* prod := x^T * v   ( == x^T * A * x )  */
	  if (prod<100.0) {
	    m11 = -a11 + v1*v1;               /* M := -A + (Ax)(Ax)^T */
	    m12 = -a12 + v1*v2;
	    m13 = -a13 + v1*v3;
	    m21 = -a21 + v2*v1;
	    m22 = -a22 + v2*v2;
	    m23 = -a23 + v2*v3;
	    m31 = -a31 + v3*v1;
	    m32 = -a32 + v3*v2;
	    m33 = -a33 + v3*v3;
	    ex = global_data->arbs_gauss_height[r]*exp(-prod/2.0);
	    sum += ex;
	    *gx -= (v1*ex);
	    *gy -= (v2*ex);
	    *gz -= (v3*ex);
	    cxx += (m11*ex);
	    cxy += (m12*ex);
	    cxz += (m13*ex);
	    cyx += (m21*ex);
	    cyy += (m22*ex);
	    cyz += (m23*ex);
	    czx += (m31*ex);
	    czy += (m32*ex);
	    czz += (m33*ex);
	    }
	  }
	*g =sqrt((*gx)*(*gx) + (*gy)*(*gy) + (*gz)*(*gz));
	if ((*g)!=0.0) {
	  g0x = (*gx)/(*g);
	  g0y = (*gy)/(*g);
	  g0z = (*gz)/(*g);
	  }
	else {
	  g0x = 0.0;
	  g0y = 0.0;
	  g0z = 0.0;
	  *gx = 1.0e-15;
	  *gy = 0.0;
	  *gz = 0.0;
	  *g  = 1.0e-15;
	  }
        v1 = cxx*g0x + cxy*g0y + cxz*g0z;
        v2 = cyx*g0x + cyy*g0y + cyz*g0z;
        v3 = czx*g0x + czy*g0y + czz*g0z;
	*cr = v1*g0x + v2*g0y + v3*g0z;
	temp=sqrt((*gx)*(*gx)+(*gy)*(*gy)+(*gz)*(*gz));
	if (temp!=0.0) {
	  *curvature=-0.5*((*gx)*(*gx)*(cyy+czz) + (*gy)*(*gy)*(cxx+czz) + (*gz)*(*gz)*(cxx+cyy)
	  	        -2.0*((*gx)*(*gy)*cxy + (*gx)*(*gz)*cxz + (*gy)*(*gz)*cyz))/(temp*temp*temp);
	  }
	else {
	  *curvature = 0.0;
	  }
	return (sum);
}



/******************************************************************/
static void calc_planar_sbound (LOAD_MSG_ONE_NODE *my_data,
				LOAD_MSG_ALL_NODE *global_data,
				double *sbound_droff,
				PRESSURE *pressure,
				PRESSURE *density)
{
  double force;
  double sbound_p1, sbound_force;
  double sboundfilling=0.0; /* this is the distance of the user specified sbound position and the
  		               potential minimum */
  VECTOR XPnormal, YPnormal, ZPnormal, XMnormal, YMnormal, ZMnormal;

  /** sbound_droff[]:  -x, +x, -y, +y, -z, +z , ori_x, ori_y, ori_z, radius, curve, x_cyl  y_cyl  R_cyl
   **                   0   1   2   3   4   5       6      7      8       9     10     11     12     13
   **/
   
  XPnormal.x = 1.0;
  XPnormal.y = 0.0;
  XPnormal.z = 0.0;
  
  YPnormal.x = 0.0;
  YPnormal.y = 1.0;
  YPnormal.z = 0.0;

  ZPnormal.x = 0.0;
  ZPnormal.y = 0.0;
  ZPnormal.z = 1.0;  
   
  XMnormal.x = -1.0;
  XMnormal.y = 0.0;
  XMnormal.z = 0.0;
  
  YMnormal.x = 0.0;
  YMnormal.y = -1.0;
  YMnormal.z = 0.0;

  ZMnormal.x = 0.0;
  ZMnormal.y = 0.0;
  ZMnormal.z = -1.0;  
   
  if (sbound_droff[R_CURVE] < SMALL)
  {
    /* we have sharp corners, easy to compute */
    int    i;

#if DEBUG == 1
    errout("Sharp corners\n");
#endif
    for (i = 0; i < my_data->no_atoms; i++)
    {
      int	ty;
      double	m ;

      ty    = (int)my_data->typ[i];
      m     = global_data->AB[ty][ty];      

      if(m > 4.0) /* we ignore hydrogen atoms */
      {

#if DEBUG == 1
    errout("Found a heavy atom\n");
#endif
	if(use_water_params(i,my_data))
	{
	  sbound_p1 = SBOUND_P1_H2O;
	  sbound_force = SBOUND_FORCE_H2O;
	  sboundfilling = SBOUND_FILLING_H2O;
	}
	else
	{
	  sbound_p1 = SBOUND_P1;
	  sbound_force = SBOUND_FORCE;
	  sboundfilling = SBOUND_FILLING;
	}

        if (fabs (sbound_droff[XM]) > SMALL)
	{				/*** x minus ***/
#if DEBUG == 1
    errout("Before calc_greaterplane_sbound()\n");
#endif
	  my_data->force_sbound[4*i] += 
		    calc_greaterplane_sbound (sbound_p1, sbound_force,
					      my_data->x_new[i],
					      sbound_droff[XM] + sboundfilling,
					      &(force),&(density->plane_planar[XM]));
#if DEBUG == 1
    errout("After calc_greaterplane_sbound(), Xyz: %f %f %f -energy=%f\n", my_data->x_new[i],my_data->y_new[i],my_data->z_new[i],my_data->force_sbound[4*i]);
#endif
	  my_data->force_sbound[4*i+1] += force;
	  pressure->plane_planar[XM] -= force;
	  /* Now we add the torque forces: */
	  compute_torque_forces (my_data, i, &XMnormal, my_data->x_new[i]-sbound_droff[XM]);
	  /* the last parameter gives the distance to the wall */
	}
  
	if (fabs (sbound_droff[YM]) > SMALL)
	{				/*** y minus ***/
	  my_data->force_sbound[4*i] += calc_greaterplane_sbound (sbound_p1, sbound_force,
	         my_data->y_new[i],
		 sbound_droff[YM] + sboundfilling,
		 &(force),&(density->plane_planar[YM]));
#if DEBUG == 1
    errout("After calc_greaterplane_sbound(), xYz: %f %f %f -energy=%f\n", my_data->x_new[i],my_data->y_new[i],my_data->z_new[i],my_data->force_sbound[4*i]);
#endif
	  my_data->force_sbound[4*i+2] += force;
	  pressure->plane_planar[YM] -= force;
	  compute_torque_forces (my_data, i, &YMnormal, my_data->y_new[i]-sbound_droff[YM]);
	}
  
	if (fabs (sbound_droff[ZM]) > SMALL)
	{				/*** z minus ***/
	  my_data->force_sbound[4*i] += calc_greaterplane_sbound (sbound_p1, sbound_force,
	         my_data->z_new[i],
		 sbound_droff[ZM] + sboundfilling,
		 &(force),&(density->plane_planar[ZM]));
#if DEBUG == 1
    errout("After calc_greaterplane_sbound(), xyZ: %f %f %f -energy=%f\n", my_data->x_new[i],my_data->y_new[i],my_data->z_new[i],my_data->force_sbound[4*i]);
#endif
	  my_data->force_sbound[4*i+3] += force;
	  pressure->plane_planar[ZM] -= force;
	  compute_torque_forces (my_data, i, &ZMnormal, my_data->z_new[i]-sbound_droff[ZM]);
	}
  
	if (fabs (sbound_droff[XP]) > SMALL)
	{				/*** x plus ***/
	  my_data->force_sbound[4*i] += calc_lesserplane_sbound (sbound_p1, sbound_force,
	         my_data->x_new[i],
		 sbound_droff[XP] - sboundfilling,
		 &(force),&(density->plane_planar[XP]));
#if DEBUG == 1
    errout("After calc_greaterplane_sbound(), Xyz: %f %f %f +energy=%f\n", my_data->x_new[i],my_data->y_new[i],my_data->z_new[i],my_data->force_sbound[4*i]);
#endif
	  my_data->force_sbound[4*i+1] += force;
	  pressure->plane_planar[XP] += force;
	  compute_torque_forces (my_data, i, &XPnormal, sbound_droff[XP]-my_data->x_new[i]);
	}
  
	if (fabs (sbound_droff[YP]) > SMALL)
	{				/*** y plus ***/
	  my_data->force_sbound[4*i] += calc_lesserplane_sbound (sbound_p1, sbound_force,
	         my_data->y_new[i],
		 sbound_droff[YP] - sboundfilling,
		 &(force),&(density->plane_planar[YP]));
#if DEBUG == 1
    errout("After calc_greaterplane_sbound(), xYz: %f %f %f +energy=%f\n", my_data->x_new[i],my_data->y_new[i],my_data->z_new[i],my_data->force_sbound[4*i]);
#endif
	  my_data->force_sbound[4*i+2] += force;
	  pressure->plane_planar[YP] += force;
	  compute_torque_forces (my_data, i, &YPnormal, sbound_droff[YP]-my_data->y_new[i]);
	}
  
	if (fabs (sbound_droff[ZP]) > SMALL)
	{				/*** z plus ***/
	  my_data->force_sbound[4*i] += calc_lesserplane_sbound (sbound_p1, sbound_force,
	         my_data->z_new[i],
		 sbound_droff[ZP] - sboundfilling,
		 &(force),&(density->plane_planar[ZP]));
#if DEBUG == 1
    errout("After calc_greaterplane_sbound(), xyZ: %f %f %f +energy=%f\n", my_data->x_new[i],my_data->y_new[i],my_data->z_new[i],my_data->force_sbound[4*i]);
#endif
	  my_data->force_sbound[4*i+3] += force;
	  pressure->plane_planar[ZP] += force;
	  compute_torque_forces (my_data, i, &ZPnormal, sbound_droff[ZP]-my_data->z_new[i]);
	}
      } /* if hydrogen */
    } /* for */
  } /* if sharp corners */
  else
  {
    /* we have round corners */
    /* In this case we have all 6 sides active, even if some of them are
     * placed through the origin. */

    int    i;
    int    octant[3],
           edge=-1, corner;

#if DEBUG == 1
  errout("Round corners\n");
  errout("typ[0]=%d\n",my_data->typ[0]);
#endif
    for (i = 0; i < my_data->no_atoms; i++)
    {
      int	ty    = (int)my_data->typ[i];
      double	m     = global_data->AB[ty][ty];
      
      if (m > 4.0) /* we ignore hydrogen atoms */
      {
	int select_plane=-5;
  
	if(use_water_params(i,my_data))
	{
	  sbound_p1 = SBOUND_P1_H2O;
	  sbound_force = SBOUND_FORCE_H2O;
	  sboundfilling = SBOUND_FILLING_H2O;
	}
	else
	{
	  sbound_p1 = SBOUND_P1;
	  sbound_force = SBOUND_FORCE;
	  sboundfilling = SBOUND_FILLING;
	}

	set_octant (my_data->x_new[i], my_data->y_new[i], my_data->z_new[i], sbound_droff, octant);
  
	/*** '=' is correct here ***/
	if ((corner = in_corner (my_data->x_new[i], my_data->y_new[i], my_data->z_new[i],
			        octant, sbound_droff)))
	{
	  select_plane = (octant[0]+1)/2+(octant[1]+1)+(octant[2]+1)*2;
	  if ((select_plane<0)||(select_plane>=8)) {
		errout("Fatal error: in calc_planar_sbound #1 %d\n",select_plane);
		pex(-1);	
	  } 	
	  calc_in_corner (my_data,
	  		  global_data, 
			  i,
			  octant,
			  sbound_droff,
			  &(pressure->plane_corner[select_plane]),
			  &(density-> plane_corner[select_plane])
			  );
	}
	else /*** '=' is right here ***/
	if ((edge = in_edge (my_data->x_new[i], my_data->y_new[i], my_data->z_new[i], octant, sbound_droff)))
	{
	  
	  /* edge=0: no edge; edge=1: along X; edge=2: along Y; edge=3: along Z */
	  select_plane = (edge-1)*4; /* select along which side we go */
	  if (edge == 1) /* along X */
	  {
	    select_plane += (octant[1]+1)/2 + (octant[2]+1);
	  }
	  else if (edge == 2) /* along Y */
	  {
	    select_plane += (octant[0]+1)/2 + (octant[2]+1);
	  }
	  else if (edge == 3) /* along Z */
	  {
	    select_plane += (octant[0]+1)/2 + (octant[1]+1);
	  }
	  else
	  {
	    errout("Fatal error: Wrong edge %d detected! Should only be 1, 2, or 3.\n",edge);
	    pex(-1);
	  }
 	  if ((select_plane<0)||(select_plane>=12)) {
		errout("Fatal error: in calc_planar_sbound #2 %d\n",select_plane);
		pex(-1);	
	  } 
	  calc_in_edge (my_data,
			global_data, 
			i,
			octant, edge, sbound_droff,
			&(pressure->plane_edge[select_plane]),
			&(density ->plane_edge[select_plane]));
	}
	else
	{
	  /*** x minus ***/
	  my_data->force_sbound[4*i] += calc_greaterplane_sbound (sbound_p1, sbound_force,
	         my_data->x_new[i],
		 sbound_droff[XM] + sboundfilling,
		 &(force),&(density->plane_planar[XM]));
	  my_data->force_sbound[4*i+1] += force;
	  pressure->plane_planar[XM] -= force;
	  compute_torque_forces (my_data, i, &XMnormal, my_data->x_new[i]-sbound_droff[XM]);
	  
	  /*** y minus ***/
	  my_data->force_sbound[4*i] += calc_greaterplane_sbound (sbound_p1, sbound_force,
	         my_data->y_new[i],
		 sbound_droff[YM] + sboundfilling,
		 &(force),&(density->plane_planar[YM]));
	  my_data->force_sbound[4*i+2] += force;
	  pressure->plane_planar[YM] -= force;
	  compute_torque_forces (my_data, i, &YMnormal, my_data->y_new[i]-sbound_droff[YM]);
	  
	  /*** z minus ***/
	  my_data->force_sbound[4*i] += calc_greaterplane_sbound (sbound_p1, sbound_force,
	         my_data->z_new[i],
		 sbound_droff[ZM] + sboundfilling,
		 &(force),&(density->plane_planar[ZM]));
	  my_data->force_sbound[4*i+3] += force;
	  pressure->plane_planar[ZM] -= force;
	  compute_torque_forces (my_data, i, &ZMnormal, my_data->z_new[i]-sbound_droff[ZM]);
	  
	  /*** x plus ***/
	  my_data->force_sbound[4*i] += calc_lesserplane_sbound (sbound_p1, sbound_force,
	         my_data->x_new[i],
		 sbound_droff[XP] - sboundfilling,
		 &(force),&(density->plane_planar[XP]));
	  my_data->force_sbound[4*i+1] += force;
	  pressure->plane_planar[XP] += force;
	  compute_torque_forces (my_data, i, &XPnormal, sbound_droff[XP]-my_data->x_new[i]);
	  
	  /*** y plus ***/
	  my_data->force_sbound[4*i] += calc_lesserplane_sbound (sbound_p1, sbound_force,
	         my_data->y_new[i],
		 sbound_droff[YP] - sboundfilling,
		 &(force),&(density->plane_planar[YP]));
	  my_data->force_sbound[4*i+2] += force;
	  pressure->plane_planar[YP] += force;
	  compute_torque_forces (my_data, i, &YPnormal, sbound_droff[YP]-my_data->y_new[i]);
	  
	  /*** z plus ***/
	  my_data->force_sbound[4*i] += calc_lesserplane_sbound (sbound_p1, sbound_force,
	         my_data->z_new[i],
		 sbound_droff[ZP] - sboundfilling,
		 &(force),&(density->plane_planar[ZP]));
	  my_data->force_sbound[4*i+3] += force;
	  pressure->plane_planar[ZP] += force;
	  compute_torque_forces (my_data, i, &ZPnormal, sbound_droff[ZP]-my_data->z_new[i]);
	}
      } /* if not hydrogen */
    } /* for */
  } /* round corners */

  return ;
}

/**************************************************************************/
static void calc_in_corner (LOAD_MSG_ONE_NODE *my_data,
			    LOAD_MSG_ALL_NODE *global_data,
			    int    atom_index,
			    int    *octant,
			    double *sbound_droff,
			    double *pressure,
			    double *density)
{
  double sphere[4];
  double scale_factor, D[3], forces[4] = {0,0,0,0};

  sphere[3] = sbound_droff[R_CURVE];	/* radius of curvature */
  sphere[0] = sbound_droff[(octant[0] + 1) / 2 + 0] - sphere[3] * octant[0];	/* origin X */
  sphere[1] = sbound_droff[(octant[1] + 1) / 2 + 2] - sphere[3] * octant[1];	/* origin Y */
  sphere[2] = sbound_droff[(octant[2] + 1) / 2 + 4] - sphere[3] * octant[2];	/* origin Z */

  scale_factor = scale_corner_potential (my_data->x_new[atom_index], 
  					 my_data->y_new[atom_index], 
					 my_data->z_new[atom_index],
					 sphere,
					 octant,
					 sbound_droff,
					 D       );	/* SBOUND_FORCE has to be
							 * multiplied to deepen the
							 * potential */
  /* We used to rescale the potential by scale_factor here, 
   * but now this is taken care of when we apply the product rule 
   * further down.
   * d2 is 1.0 at the boundary to the
   * planar boundaries and 8.0 at the
   * maximum under 45 degrees 
   * We rescale both, water and lipids */
  
  calc_spheric_sbound ( my_data,
			global_data,
			atom_index, atom_index,
			sphere,
			SBOUND_P1_H2O, SBOUND_FORCE_H2O,
			SBOUND_P1, SBOUND_FORCE,
			pressure,
			density,
			forces
			);
  /* Now we apply the product rule: D(ab) = a*Db + Da*b */
  my_data->force_sbound[4*atom_index+0] += scale_factor*forces[0]; /* The energy */
  my_data->force_sbound[4*atom_index+1] += scale_factor*forces[1] + D[0]*forces[0];
  my_data->force_sbound[4*atom_index+2] += scale_factor*forces[2] + D[1]*forces[0];
  my_data->force_sbound[4*atom_index+3] += scale_factor*forces[3] + D[2]*forces[0];


  return;
}


/**************************************************************************/
static double scale_corner_potential (double Px, double Py, double Pz,
			              double *origin,
			              int *octant,
			              double *sbound_droff,
				      double D[3])
{
  double sx, sy, sz, scale;
  double dx, dy, dz; /* (x,y,z) of the radial distance from the origin point (sphere center) */
  double SB2, pxy, pyz, pxz;

  /* To set the potential depth we project the origin-point vector
     on the three planes that form the corner and compute the angle (in those planes)
     with one of the two axes that span the plane (doesn't matter with which one). 
     The sin(2*\alpha) reaches unity at 45deg and zero at 0 and 90deg.
     We use then the same scaling as for the edges for all three axes:
     1 + sin^n(2*\alpha) = 1 + (s*sin(\alpha)*cos(\alpha))^n = 1 + (2*dx*dy/(dx^2 + dy^2))^n
     Then we multiply all three together, which results in an 8-fold increase of the
     potential at the tip of the corner, a twofold increase where the corner meets the edges, 
     and no increase where the corner meets the planes.
     */

  dx = (Px - origin[0]);
  dy = (Py - origin[1]);
  dz = (Pz - origin[2]);
  
  sx = fabs(2*dz*dy/(dz*dz + dy*dy));
  sx = pow (sx, (double)SBOUND_EXPONENT);
  sx = 1.0 + sx; /* 1...2 */

  sy = fabs(2*dx*dz/(dx*dx + dz*dz));
  sy = pow (sy, (double)SBOUND_EXPONENT);
  sy = 1.0 + sy; /* 1...2 */
  
  sz = fabs(2*dx*dy/(dx*dx + dy*dy));
  sz = pow (sz, (double)SBOUND_EXPONENT);
  sz = 1.0 + sz; /* 1...2 */
  
  scale = sx * sy * sz; /* 1...8 */
  
  /* Now we need the derivatives to apply the product rule later on */
  SB2 = pow(2,SBOUND_EXPONENT);
  pxy = pow(dx*dy/(dx*dx + dy*dy),SBOUND_EXPONENT);
  pyz = pow(dy*dz/(dy*dy + dz*dz),SBOUND_EXPONENT);
  pxz = pow(dx*dz/(dx*dx + dz*dz),SBOUND_EXPONENT);
  D[0] = SB2*(1 + SB2*pxy)*(-dx*dx + dz*dz)*pxz*
	  (1 + SB2*pyz)*
	  SBOUND_EXPONENT/(dx*dx*dx + dx*dz*dz) + 
	  SB2*(-dx*dx + dy*dy)*pxy*
	  (1 + SB2*pxz)*(1 + SB2*pyz)*
	  SBOUND_EXPONENT/(dx*dx*dx + dx*dy*dy);
  D[1] = SB2*(1 + SB2*pxy)*(-dy*dy + dz*dz)*pyz*
	  (1 + SB2*pxz)*
	  SBOUND_EXPONENT/(dy*dy*dy + dy*dz*dz) + 
	  SB2*(dx*dx - dy*dy)*pxy*
	  (1 + SB2*pxz)*(1 + SB2*pyz)*
	  SBOUND_EXPONENT/(dx*dx*dy + dy*dy*dy);
  D[2] = SB2*(1 + SB2*pxy)*(dy*dy - dz*dz)*pyz*
	  (1 + SB2*pxz)*SBOUND_EXPONENT
	  /(dy*dy*dz + dz*dz*dz) + 
	  SB2*(1 + SB2*pow(dx*dy/(dx*dx + dy*dy),
	  SBOUND_EXPONENT))*(dx*dx - dz*dz)*
	  pxz*(1 + SB2*
	  pyz)*SBOUND_EXPONENT
	  /(dx*dx*dz + dz*dz*dz);
     
#if DEBUG == 6
printf ("SCP: Px=%f, Py=%f, Pz=%f, Ox=%f, Oy=%f, Oz=%f, scale=%f\n",
         Px,Py,Pz, origin[0],origin[1],origin[2], scale);
#endif

  return scale;
}

/**************************************************************************/
#if 0
/* this is not needed any more, but I like to keep it here, just in case... (HH) */
static int max3 (double a, double b, double c)
{
  int    i;

  if (a < b)
  {
    if (b < c)
    {
      i = 2;
    }
    else
    {
      i = 1;
    }
  }
  else
  {
    if (a < c)
    {
      i = 2;
    }
    else
    {
      i = 0;
    }
  }
  return i;
}
#endif

/**************************************************************************/
static void calc_in_edge (LOAD_MSG_ONE_NODE *my_data,
			  LOAD_MSG_ALL_NODE *global_data,
			  int atom_index,
			  int *octant, int edge,
			  double *sbound_droff,
			  double *pressure,
			  double *density)

{
  double sphere[4];		/* origin, radius */
  double d2=0.0, D[3]={0,0,0}, forces[4]={0,0,0,0}; /* The arrays have to be cleared,
  						       as it CAN be that we go into
						       calc_spheric_sbound() but nothing
						       is done in there because we are in fact
						       outside of the SBOUND range. This can
						       happen if the radius of curvature is greater than 
						       the SBOUND reaches inward (about 5A). */

  /** sbound_droff[]:  -x, +x, -y, +y, -z, +z , ori_x, ori_y, ori_z, radius, curve, x_cyl  y_cyl  R_cyl
   **                   0   1   2   3   4   5       6      7      8       9     10     11     12     13
   **/

  /* to set the potential depth we use the squared radial distance from the
   * surface of the sphere to the outer rectangle as the scaling parameter:
   * d2 */
  sphere[3] = sbound_droff[R_CURVE];
  if (edge == 1)		/* along X */
  {
    sphere[0] = my_data->x_new[atom_index];
    sphere[1] = sbound_droff[(octant[1] + 1) / 2 + 2] - octant[1] * sbound_droff[R_CURVE];
    sphere[2] = sbound_droff[(octant[2] + 1) / 2 + 4] - octant[2] * sbound_droff[R_CURVE];
    d2 = scale_edge_potential (my_data->y_new[atom_index], my_data->z_new[atom_index], sphere[1], sphere[2], &(D[1]), &(D[2]));
  }
  else
  if (edge == 2)		/* along Y */
  {
    sphere[0] = sbound_droff[(octant[0] + 1) / 2 + 0] - octant[0] * sbound_droff[R_CURVE];
    sphere[1] = my_data->y_new[atom_index];
    sphere[2] = sbound_droff[(octant[2] + 1) / 2 + 4] - octant[2] * sbound_droff[R_CURVE];
    d2 = scale_edge_potential (my_data->x_new[atom_index], my_data->z_new[atom_index], sphere[0], sphere[2], &(D[0]), &(D[2]));
  }
  else
  if (edge == 3)		/* along Z */
  {
    sphere[0] = sbound_droff[(octant[0] + 1) / 2 + 0] - octant[0] * sbound_droff[R_CURVE];
    sphere[1] = sbound_droff[(octant[1] + 1) / 2 + 2] - octant[1] * sbound_droff[R_CURVE];
    sphere[2] = my_data->z_new[atom_index];
    d2 = scale_edge_potential (my_data->x_new[atom_index], my_data->y_new[atom_index], sphere[0], sphere[1], &(D[0]), &(D[1]));
  }
  else
  {
    fprintf (stderr, "Illegal edge = %d detected in calc_in_edge.\n", edge);
    pex (-1);
  }

  /* We used to rescale the potential by *d2 here, 
   * but now this is taken care of when we apply the product rule 
   * further down.
   * d2 is 1.0 at the boundary to the
   * planar boundaries and 2.0 at the
   * maximum under 45 degrees 
   * We rescale both, water and lipids */
  
  calc_spheric_sbound ( my_data,
			global_data,
			atom_index, atom_index,
			sphere,
			SBOUND_P1_H2O, SBOUND_FORCE_H2O,
			SBOUND_P1, SBOUND_FORCE,
			pressure,
			density,
			forces); 

#if DEBUG == 4
printf ("CIE: Dx=%f, Dy=%f, Dz=%f, f0=%f, f2=%f\n",D[0],D[1],D[2],forces[0],forces[2]);
#endif
  /* Now we apply the product rule: D(ab) = a*Db + Da*b */
  my_data->force_sbound[4*atom_index+0] += d2*forces[0]; /* The energy */
  my_data->force_sbound[4*atom_index+1] += d2*forces[1] + D[0]*forces[0];
  my_data->force_sbound[4*atom_index+2] += d2*forces[2] + D[1]*forces[0];
  my_data->force_sbound[4*atom_index+3] += d2*forces[3] + D[2]*forces[0];

  
  return;
}


/**************************************************************************/
static double scale_edge_potential (double Px, double Py,	/* (x,y) of point */
			            double Ox, double Oy,	/* (x,y) of origin */
				    double *Dx, double *Dy )	/* the derivative, needed to scale the forces */
{ /* This routine is used to scale the potential in curved edges DEEPER to
     compensate the additional surface tension caused by the curvature. 
     It should be between 1.0 and 2.0, depending on how far away from the
     flat planes the point (Px,Py) lies. */
     
  double dx,dy; /* (radial) distance of point to origin (origin is the center of the
  		   curvature of the edge) */
  double scale; /* The scaling coefficient */

  dx = (Px - Ox);
  dy = (Py - Oy);

  /* Say, for edges along the X and Z axes \alpha is the angle with 
     the Y axis (only between 0...45deg, -45...0deg gets folded over due to
     the fabs() calls above), then we scale like
     1 + sin^n(2*\alpha) = 1 + (2*sin(\alpha)*cos(\alpha))^n = 1 + (2*dx*dy/(dx^2 + dy^2))^n
     which produces values between 1...2  */

  scale = fabs(2*dx*dy/(dx*dx + dy*dy));	/* This is between 0...1 for an angle of 0...45deg. */
  scale = pow (scale, (double)SBOUND_EXPONENT);	/* n-th power, to make it as steep as needed. */

  /* Now we have to normalize to get 1 at the outsides and 2 in the middle */
  scale = 1.0 + scale;
  
  /* The next two expressions were derived by Mma, so they look horrible, but I
     am confident that they are correct. */
  *Dx = (pow(2,SBOUND_EXPONENT)*(-dx + dy)*(dx + dy)*pow((dx*dy)/
        (dx*dx + dy*dy),SBOUND_EXPONENT)*(SBOUND_EXPONENT))/
        (dx*dx*dx + dx*dy*dy);
  *Dy = (pow(2,SBOUND_EXPONENT)*( dx - dy)*(dx + dy)*pow((dx*dy)/
        (dx*dx + dy*dy),SBOUND_EXPONENT)*(SBOUND_EXPONENT))/
        (dx*dx*dy + dy*dy*dy);

#if DEBUG == 4
printf ("SEP: Px=%f, Py=%f, Ox=%f, Oy=%f, scale=%f, Dx=%f, Dy=%f\n",Px,Py,Ox,Oy,scale,*Dx,*Dy);
#endif
  return scale;
}

/**************************************************************************/
static int in_edge (double x, double y, double z, int *octant,
		    double *sbound_droff)
{;
  /** sbound_droff[]:  -x, +x, -y, +y, -z, +z , ori_x, ori_y, ori_z, radius, curve, x_cyl  y_cyl  R_cyl
   **                   0   1   2   3   4   5       6      7      8       9     10     11     12     13
   **/

  /** We return which kind of edge we have:
   ** 0	= no edge
   ** 1 = along X axis
   ** 2 = along Y axis
   ** 3 = along Z axis  **/


  if ((octant[0] * x > (octant[0] * sbound_droff[(octant[0] + 1) / 2    ] - sbound_droff[R_CURVE])) &&
      (octant[1] * y > (octant[1] * sbound_droff[(octant[1] + 1) / 2 + 2] - sbound_droff[R_CURVE])))
    return 3;

  if ((octant[0] * x > (octant[0] * sbound_droff[(octant[0] + 1) / 2    ] - sbound_droff[R_CURVE])) &&
      (octant[2] * z > (octant[2] * sbound_droff[(octant[2] + 1) / 2 + 4] - sbound_droff[R_CURVE])))
    return 2;

  if ((octant[1] * y > (octant[1] * sbound_droff[(octant[1] + 1) / 2 + 2] - sbound_droff[R_CURVE])) &&
      (octant[2] * z > (octant[2] * sbound_droff[(octant[2] + 1) / 2 + 4] - sbound_droff[R_CURVE])))
    return 1;

  return 0;
}


/**************************************************************************/
static int in_corner (double x, double y, double z, int *octant,
	              double *sbound_droff)
{;
  /** sbound_droff[]:  -x, +x, -y, +y, -z, +z , ori_x, ori_y, ori_z, radius, curve, x_cyl  y_cyl  R_cyl
   **                   0   1   2   3   4   5       6      7      8       9     10     11     12     13
   **/
  return (((octant[0] * x) > (octant[0] * sbound_droff[(octant[0] + 1) / 2 + 0] - sbound_droff[R_CURVE])) &&
	  ((octant[1] * y) > (octant[1] * sbound_droff[(octant[1] + 1) / 2 + 2] - sbound_droff[R_CURVE])) &&
	  ((octant[2] * z) > (octant[2] * sbound_droff[(octant[2] + 1) / 2 + 4] - sbound_droff[R_CURVE])));
}


/**************************************************************************/
static void set_octant (double x, double y, double z,
			double *sbound_droff,
			int *octant)
{;
  /* octant is defined by the SBOUND planes, not the sign of the coordinates */
  /** sbound_droff[]:  -x, +x, -y, +y, -z, +z , ori_x, ori_y, ori_z, radius, curve, x_cyl  y_cyl  R_cyl
   **                   0   1   2   3   4   5       6      7      8       9     10     11     12     13
   **/
  octant[0] = sign (x - (sbound_droff[XM] + sbound_droff[XP]) / 2.0); /* +1 for +X, -1 for -X */
  octant[1] = sign (y - (sbound_droff[YM] + sbound_droff[YP]) / 2.0); /* +1 for +Y, -1 for -Y */
  octant[2] = sign (z - (sbound_droff[ZM] + sbound_droff[ZP]) / 2.0); /* +1 for +Z, -1 for -Z */
  return;
}


/**************************************************************************/
static double calc_greaterplane_sbound (double sbound_p1,
					double sbound_force,
					double atompos,
					double sboundr,	/** input **/
					double *forceatom,     /** output **/
					double *SB_num 
					)	
{ /* for NEGATIVE SBOUND */
  double delta,
         delta2,
         sbound_energy = 0.0;
  double sboundrad = sboundr + sqrt (sbound_p1 / 2.0);
 
  if (atompos <= sboundrad)
  {
    (*SB_num)++; /* here we count the heavy atoms in the boundary region */
    delta = atompos - sboundrad;
    delta2 = delta * delta;
    sbound_energy += sbound_force * delta2 * (delta2 - sbound_p1);
    *forceatom = (sbound_force * 2.0 * delta * (2.0 * delta2 - sbound_p1));
  }
  else
    *forceatom = 0.0;
    
  return sbound_energy;
}

/**************************************************************************/
static double calc_lesserplane_sbound ( double sbound_p1,
					double sbound_force,
					double atompos,
					double sboundr,	/** input **/
					double *forceatom,	/** output **/
					double *SB_num
					)
{ /* for POSITIVE SBOUND */
  double delta,
         delta2,
         sbound_energy = 0.0;
  double sboundrad = sboundr - sqrt (sbound_p1 / 2.0); /* where the SBOUND begins */
  
  if (atompos >= sboundrad) /* only in this case do we feel an SBOUND */
  {
    (*SB_num)++; /* here we count the heavy atoms in the boundary region */
    delta = atompos - sboundrad;
    delta2 = delta * delta;
    sbound_energy += sbound_force * delta2 * (delta2 - sbound_p1);
    *forceatom = (sbound_force * 2.0 * delta * (2.0 * delta2 - sbound_p1));
  }
  else
    *forceatom = 0.0;
    
  return sbound_energy;
}

/**********************************************************************************/
/*** returns TRUE if atom is in sboundregion, otherwise FALSE  			***/
/*** *offset is the distance into the sbound-region in angstroem + maxoffset	***/
/**********************************************************************************/
static int in_stochastic_region ( double *sbound_droff,
				  double *xyzko,
				  double maxoffset, /* how far inward, measuring from
						       where the user specifies the position of the potential, 
						       we will have the stochastic forces active */
				  double *offset)
{
  int    active = FALSE;

  *offset = 9999999.0;		/* distance into the sbound-region in
				 * angstroem plus additional layer thickness
				 * plus half width of potential */

  /*** test the planes ***/
  /*** test X- plane ***/
  if ((fabs (sbound_droff[XM]) > SMALL) &&
      (xyzko[0] < (sbound_droff[XM] + maxoffset)))
  {
    *offset = min (*offset, sbound_droff[XM] + maxoffset - xyzko[0]);
    active = TRUE;
  }

  /*** test X+ plane ***/
  if ((fabs (sbound_droff[XP]) > SMALL) &&
      (xyzko[0] > (sbound_droff[XP] - maxoffset)))
  {
    *offset = min (*offset, xyzko[0] - (sbound_droff[XP] - maxoffset));
    active = TRUE;
  }

  /*** test Y- plane ***/
  if ((fabs (sbound_droff[YM]) > SMALL) &&
      (xyzko[1] < (sbound_droff[YM] + maxoffset)))
  {
    *offset = min (*offset, sbound_droff[YM] + maxoffset - xyzko[1]);
    active = TRUE;
  }

  /*** test Y+ plane ***/
  if ((fabs (sbound_droff[YP]) > SMALL) &&
      (xyzko[1] > (sbound_droff[YP] - maxoffset)))
  {
    *offset = min (*offset, xyzko[1] - sbound_droff[YP] + maxoffset);
    active = TRUE;
  }

  /*** test Z- plane ***/
  if ((fabs (sbound_droff[ZM]) > SMALL) &&
      (xyzko[2] < (sbound_droff[ZM] + maxoffset)))
  {
    *offset = min (*offset, sbound_droff[ZM] + maxoffset - xyzko[2]);
    active = TRUE;
  }

  /*** test Z+ plane ***/
  if ((fabs (sbound_droff[ZP]) > SMALL) &&
      (xyzko[2] > (sbound_droff[ZP] - maxoffset)))
  {
    *offset = min (*offset, xyzko[2] - sbound_droff[ZP] + maxoffset);
    active = TRUE;
  }

  /*** test the sphere ***/
  if (fabs (sbound_droff[R_SPHERE]) > SMALL) 
  {
    calc_radial_distance(xyzko, 
		    sbound_droff[SP_ORI_X], sbound_droff[SP_ORI_Y], sbound_droff[SP_ORI_Z],
		    sbound_droff[R_SPHERE], maxoffset, &active, offset);
  }

  /*** test the cylinder ***/
  if (fabs(sbound_droff[R_CYL]) > SMALL) 
  {
    calc_radial_distance(xyzko, 
		         sbound_droff[ORI_X_CYL], sbound_droff[ORI_Y_CYL], xyzko[2],
		         sbound_droff[R_CYL], maxoffset, &active, offset);
  }

  /*** test the round corners ***/
  if (sbound_droff[R_CURVE] > SMALL)
  {
    int    octant[3];
    int    edge;

    set_octant (xyzko[0], xyzko[1], xyzko[2], sbound_droff, octant);

    if (in_corner (xyzko[0], xyzko[1], xyzko[2],
		   octant, sbound_droff))
    {
      calc_radial_distance (xyzko,
			    sbound_droff[(octant[0] + 1) / 2    ] - sbound_droff[R_CURVE] * octant[0],
			    sbound_droff[(octant[1] + 1) / 2 + 2] - sbound_droff[R_CURVE] * octant[1],
			    sbound_droff[(octant[2] + 1) / 2 + 4] - sbound_droff[R_CURVE] * octant[2],
			    sbound_droff[R_CURVE], maxoffset, &active, offset);
    }
    else if ((edge=in_edge (xyzko[0], xyzko[1], xyzko[2],octant, sbound_droff)))
    {
      double sphere[3];

      if (edge == 1)		/* along X */
      {
	sphere[0] = xyzko[0];
	sphere[1] = sbound_droff[(octant[1] + 1) / 2 + 2] - octant[1] * sbound_droff[R_CURVE];
	sphere[2] = sbound_droff[(octant[2] + 1) / 2 + 4] - octant[2] * sbound_droff[R_CURVE];
      }
      else
      if (edge == 2)		/* along Y */
      {
	sphere[0] = sbound_droff[(octant[0] + 1) / 2 + 0] - octant[0] * sbound_droff[R_CURVE];
	sphere[1] = xyzko[1];
	sphere[2] = sbound_droff[(octant[2] + 1) / 2 + 4] - octant[2] * sbound_droff[R_CURVE];
      }
      else
      if (edge == 3)		/* along Z */
      {
	sphere[0] = sbound_droff[(octant[0] + 1) / 2 + 0] - octant[0] * sbound_droff[R_CURVE];
	sphere[1] = sbound_droff[(octant[1] + 1) / 2 + 2] - octant[1] * sbound_droff[R_CURVE];
	sphere[2] = xyzko[2];
      }
      calc_radial_distance (xyzko,
			    sphere[0], sphere[1], sphere[2],
			    sbound_droff[R_CURVE], maxoffset, &active, offset);
    }
    else
    {
      if (xyzko[0] < sbound_droff[0] + maxoffset)
      {
	*offset = min (*offset, sbound_droff[0] + maxoffset - xyzko[0]);
	active = TRUE;
      }

      if (xyzko[0] > sbound_droff[1] - maxoffset)
      {
	*offset = min (*offset, xyzko[0] - sbound_droff[1] + maxoffset);
	active = TRUE;
      }

      if (xyzko[1] < sbound_droff[2] + maxoffset)
      {
	*offset = min (*offset, sbound_droff[2] + maxoffset - xyzko[1]);
	active = TRUE;
      }

      if (xyzko[1] > sbound_droff[3] - maxoffset)
      {
	*offset = min (*offset, xyzko[1] - sbound_droff[3] + maxoffset);
	active = TRUE;
      }

      if (xyzko[2] < sbound_droff[4] + maxoffset)
      {
	*offset = min (*offset, sbound_droff[4] + maxoffset - xyzko[2]);
	active = TRUE;
      }

      if (xyzko[2] > sbound_droff[5] - maxoffset)
      {
	*offset = min (*offset, xyzko[2] - sbound_droff[5] + maxoffset);
	active = TRUE;
      }
    }
  }

  return active;
}

/**************************************************************************/
static void   calc_radial_distance (double *xyzko,
			            double Ox, double Oy, double Oz,
			            double radius, double maxoffset,
			            int *active,
			            double *offset)
{
  double dx = xyzko[0] - Ox;
  double dy = xyzko[1] - Oy;
  double dz = xyzko[2] - Oz;
  double dist = sqrt(dx * dx + dy * dy + dz * dz);

  if (dist > radius - maxoffset) 
  {
    *offset = min(*offset, dist - radius + maxoffset);
    *active = TRUE;
  }
  return;
}


/**************************************************************************/
/*** returns TRUE, if any planar sbound set ***/
int any_planar_sbound_set (double *sbound_droff) 
{
	int    i;
	for (i = 0; i < 6; i++) if (fabs (sbound_droff[i]) > SMALL) return TRUE;
	return FALSE;
}

/**************************************************************************/
/*** returns TRUE, if any sbound set ***/
int    any_sbound_set (double *sbound_droff) 
{
	int    i;

	for (i = 0; i < 6; i++) 
	if (fabs (sbound_droff[i])>SMALL) return TRUE;

	if (sbound_droff[R_SPHERE]>SMALL)		return TRUE;
        if (sbound_droff[R_SPHERE]< (-SMALL)) 		return TRUE;
	if (sbound_droff[R_CYL]>SMALL)			return TRUE;

	return FALSE;
}

/**************************************************************************/
/*** generates a gaussian random distribution 				***/
/*** with standard deviation of rfd for  	  			***/
/*** a random direction in three dimensions   				***/
/**************************************************************************/
static void calc_random_force (double rfd, double *xyzforce)
{
  double rg = 6.0;
  double ax = 0.0,
         ay = 0.0,
         az = 0.0;
  int    j;

  for (j = 0; j < 12; j++)
  {
    ax += ego_rand();
    ay += ego_rand();
    az += ego_rand();
  }
  
  xyzforce[0] = - (ax - rg) * rfd;
  xyzforce[1] = - (ay - rg) * rfd;
  xyzforce[2] = - (az - rg) * rfd;
}

/******************************************************************/
/************** Stochastic Forces *********************************/
/******************************************************************/
static void add_stochastic_forces (	LOAD_MSG_ONE_NODE *my_data,
					LOAD_MSG_ALL_NODE *global_data,
					double *rnd_force_std_deviation,
					double *Gamma	/* changes every step,
							 * if any sbound set! */
)
{
  double xyzforce[3],
         xyzko[3];

  if (any_sbound_set (global_data->sbound_droff))
  {
    /*************************************************************************/
    /*** the new way: each atom in sbound region feels a rnd force 	   ***/
    /*************************************************************************/
    double fbetafac = TWO * KBOLTZ * global_data->target_temp / global_data->time_step;
    double offset;		/*** distance from sbound-region in angstroem ***/
    double fbeta,
           rfd,
           minwidth ;	/* distance from inner start to minimum of potential */
    double maxoffset;	 /* maxoffset is max. 	*/
	   		 /* extension of	*/
			 /* stochastic forces,	*/
			 /* measured from	*/
			 /* potential minimum 	*/
    int    typ;
    int    i;

    for (i = 0; i < my_data->no_atoms; i++)
    {
      xyzko[0] = my_data->x_new[i];
      xyzko[1] = my_data->y_new[i];
      xyzko[2] = my_data->z_new[i];
      offset = Gamma[i] = 0.0;

      if(use_water_params(i,my_data))
      {
        minwidth = sqrt (SBOUND_P1_H2O / 2.0) +	/* distance from inner start to what the user */
      	     SBOUND_FILLING_H2O; 		/* specifies as the position of the potential */
     
      }
      else
      {
        minwidth = sqrt (SBOUND_P1 / 2.0) +	/* distance from inner start to what the user */
		     SBOUND_FILLING;		/* specifies as the position of the potential */
      }
      maxoffset = global_data->sbound_layerthick + minwidth; /* how far inward, measuring from
      								what the user specifies as 
								the position of the potential, we will
								have the stochastic forces active */
  
      if (in_stochastic_region (global_data->sbound_droff, xyzko,
				maxoffset, &offset))
      {				/* inside the SBOUND region offset is always
				 * POSITIVE. It gives the distance from where the stochastic forces
				 * region starts to where the current atom is. Shortest distance that is. */
				/* offset > 0 if INSIDE potential, negative
				 * if outside */
	typ = (int)my_data->typ[i];
	if (offset > maxoffset)
	{			/* CONSTANT random force region */
	  fbeta = TIMEFAC * global_data->sbound_fbetagradient;
	}
	else
	{			/* random force linear decrease */
	  fbeta =  offset / maxoffset *
	           TIMEFAC * global_data->sbound_fbetagradient ;
	}
	Gamma[i] = fbeta * global_data->time_step / TWO;
	rfd = sqrt (fbetafac * global_data->AB[typ][typ] * fbeta);

	calc_random_force (rfd, xyzforce);

	my_data->force_sbound[4*i+1] -= xyzforce[0];
	my_data->force_sbound[4*i+2] -= xyzforce[1];
	my_data->force_sbound[4*i+3] -= xyzforce[2];
      }
    }
  } /*** any sbound set ***/
  else
  {
    /***********************************************************************/
    /*** thats the old way: each atom with fbeta!=0 feels a random force ***/
    /***********************************************************************/
    int    i;
    for (i = 0; i < my_data->no_atoms; i++) 
    {
      if (rnd_force_std_deviation[i] != 0.0)
      {
	calc_random_force (rnd_force_std_deviation[i], xyzforce);

	my_data->force_sbound[4*i+1] -= xyzforce[0];
	my_data->force_sbound[4*i+2] -= xyzforce[1];
	my_data->force_sbound[4*i+3] -= xyzforce[2];
      }
    }
  }
}

/******************************************************************/
/************************* Harmonic Restraints ********************/
/******************************************************************/
/** []double harmon_x, harmon_y, harmon_z  equilibrium positions***/
/*** of the atoms; []double harmon_k  all the force parameters; ***/
/*** each atom in the boundary (surface) is fixed at its 	***/
/*** equilibrium position by a harmonic potential;		***/
/*** energy: K*(r - r0)^2 = K*((x-x0)^2 + (y-y0)^2 + (z-z0)^2	***/
/*** force: 2*K*(x - x0) in "wrong" direction because		***/
/*** "-" is hidden in the VERLET-integrator.           		***/
/******************************************************************/
static void add_harmonic_boundary (LOAD_MSG_ONE_NODE *my_data)
{
  double K,	 /* K is the force constant */
  	  bx=0.0,by=0.0,bz=0.0; /* the differences from the reference position */
  int    i,ii;

  for (i=0,ii=0;i<my_data->no_atoms;i++,ii+=4) 
  {
    K  = my_data->harmon_k[i];
    if (K != 0.0)
    {
#if  (PATCH_HARMONIC & 1) == 1 /* harmonic in X direction */
      bx = my_data->x_new[i] - my_data->harmon_x[i];
#if  (PATCH_HARMONIC & 8) == 8 /* asymmetric */
      if((my_data->harmon_x[i] >= 0) && (bx < 0)) bx = 0.0;
      if((my_data->harmon_x[i] <= 0) && (bx > 0)) bx = 0.0;
#endif
#endif
#if  (PATCH_HARMONIC & 2) == 2 /* harmonic in Y direction */
      by = my_data->y_new[i] - my_data->harmon_y[i];
#if  (PATCH_HARMONIC & 8) == 8 /* asymmetric */
      if((my_data->harmon_y[i] >= 0) && (by < 0)) by = 0.0;
      if((my_data->harmon_y[i] <= 0) && (by > 0)) by = 0.0;
#endif
#endif
#if  (PATCH_HARMONIC & 4) == 4 /* harmonic in Z direction */
      bz = my_data->z_new[i] - my_data->harmon_z[i];
#if  (PATCH_HARMONIC & 8) == 8 /* asymmetric */
      if((my_data->harmon_z[i] >= 0) && (bz < 0)) bz = 0.0;
      if((my_data->harmon_z[i] <= 0) && (bz > 0)) bz = 0.0;
#endif
#endif

      my_data->force_restr[ii  ] += K * (bx*bx + by*by + bz*bz); /* this is the ENERGY!! */
      K = K * 2.0;
      my_data->force_restr[ii+1] += K * bx;
      my_data->force_restr[ii+2] += K * by;
      my_data->force_restr[ii+3] += K * bz;
    }
  }
}

/************************************************************************/
/* Gamma conforms with the famous fluctuation/dissipation theorem and is
 * the friction coefficient to take energy out of the system that has
 * been brought into it by the random forces. changing the value of Gamma
 * can be used  to simulate a temperature coupling to a heat bath to heat
 * or cool the protein. */

void   compute_Gamma(	LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			double *Gamma,
			double *rnd_force_std_deviation,
			int node_number)
{
	int    i,typ;
	double fbeta,mass;
	
	my_number = node_number; /* now it is available globally in this file */
	if ((global_data->stochastic)&&(!any_sbound_set(global_data->sbound_droff))) {
		for (i=0;i<my_data->no_atoms;i++) {
			typ   = (int)my_data->typ[i];
			fbeta = my_data->beta[i] * TIMEFAC;
			mass  = global_data->AB[typ][typ];
			if (	(my_data->beta[i]<0.0) ||
				(mass<0.0) ) {
				errout("Fatal error: Negative friction coeffcient=%f or mass=%f not allowed!\n",
					my_data->beta[i],mass);
				pex(-1);	
			}

			rnd_force_std_deviation[i] = 
				sqrt(TWO * mass * fbeta * KBOLTZ * 
					global_data->target_temp/global_data->time_step);

			Gamma[i] = fbeta * global_data->time_step / 2.0;
		}
	} else {
		for (i = 0; i < my_data->no_atoms; i++) {
			Gamma[i]                   = 0.0;
			rnd_force_std_deviation[i] = 0.0;
		}
	}
}



/************************************************************************/
void ini_range (int range_used)
{ /* Reset the current range definition.
   * This function should be called from init_node() in egnolib.c 
   * and the end of boundary() in this module.
   *
   * The ranged SBOUND, which is activated via the ctl.lis-switch SBOUNDMODE 
   * works as follows:
   * SBOUNDMODE = 0: switched off, normal SBOUND: waters use water params, lipids use lipid params
   * SBOUNDMODE = 1: whether we use water or lipid params depends on the position of the heavy atom:
   *		     if it is inside the range (set in define_range()), we use lipid params, even
   *		     if it is a water, outside the range we use water params for all atoms.
   *		     The range is now defined by the position of the glycerol backbone, which is held
   *		     in the global static variable "special" and updated in each step.
   * SBOUNDMODE = 2: We use water params for all molecules!
   * SBOUNDMODE = 3: Something special to the numerical_sbound() routines of Volker Groll.
   */
#if DEBUG == 5
printf("Range: set=%d, used(is)=%d, used(should)=%d YP=%f, YM=%f\n",
        range.range_set,range.range_used, range_used, range.YP_pos,range.YM_pos);
#endif
  range.range_set = FALSE;
  range.range_used = range_used;
  range.YP_pos = 0.0;
  range.YM_pos = 0.0;
  range.YP_count = 0;
  range.YM_count = 0;
}

/************************************************************************/
void ini_special (int range_used)
{ /* Reset the current "special" definition, i.e., the P43 distance.
   * This function should be called from init_for_each_step() in egnolib.c 
   */
#if DEBUG == 3
printf("%d: Special: set=%d, used(is)=%d, used(should)=%d YP_sum_pos=%f, YP_count=%d, YM_sum_pos=%f, YM_count=%d\n",my_number,
        special.range_set, special.range_used, range_used, 
	special.YP_pos, special.YP_count, special.YM_pos, special.YM_count);
#endif
  special.range_set = FALSE;
  special.range_used = range_used;
  special.YP_pos = 0.0;
  special.YM_pos = 0.0;
  special.YP_count = 0;
  special.YM_count = 0;
}

/************************************************************************/
void transfer_special (ENERGY_INFO *energy)
{/* called from egnode.c at the end of the main loop 
  * and from do_all_calculations() in egcontro.c */
    energy->special_YP_pos = special.YP_pos;
    energy->special_YM_pos = special.YM_pos;
    energy->special_YP_count = special.YP_count;
    energy->special_YM_count = special.YM_count;
}

/************************************************************************/
void define_range (int no_atoms, double *y, int *atom_id)
{ /* In this function we set "range".
   * Currently we determine the average position of the glycerol
   * backbone to serve as a divider between water and lipid sbound params.
   * We identify the glycerol atoms as C19, C20, C21 */
  /* This function is called from gather_non_pair_atoms() in egnode.c and
   * from do_all_calculations() in egcontro.c */

  if (range.range_used == 1)
  {
    int i;
    for (i=0; i<no_atoms; i++)
    {
      if ( atom_id[i] < (200*52) )
      { /* we have a lipid */
	if ( atom_id[i] < (100*52))
	{ /* we have a lipid in the upper layer */
	  if ( (((atom_id[i]-18) % 52) == 0) ||
	      (((atom_id[i]-19) % 52) == 0) ||
	      (((atom_id[i]-20) % 52) == 0)   )
	  { /* we have one of the glycerol atoms! */
	    range.YP_pos =(range.YP_pos*range.YP_count + y[i])/(double)(++range.YP_count);
	    range.range_set = TRUE;
	  }
	}
	else
	{ /* we have a lipid in the lower layer */
	  if ( (((atom_id[i]-18) % 52) == 0) ||
	      (((atom_id[i]-19) % 52) == 0) ||
	      (((atom_id[i]-20) % 52) == 0)   )
	  { /* we have one of the glycerol atoms! */
	    range.YM_pos =(range.YM_pos*range.YM_count + y[i])/(double)(++range.YM_count);
	    range.range_set = TRUE;
	  }
	}
      }
    }
  }
}

/************************************************************************/
static void set_special (int no_atoms, double *y, int *atom_id)
{ /* In this function we set data for "special" pressure mode (3) 
   * currently we determine the average distance of the P43
   * to serve as a pressure gauge.
   * We identify the P43 atoms as P43 */
  /* This function is called from boundary() in this module */
#if DEBUG == 3
errout("%d: Special: setting the P43 positions. special.range_used=%d, no_atoms=%d\n", my_number,
	special.range_used, no_atoms);
#endif

  if (special.range_used == SPECIAL_MODE)
  {
    int i;
    for (i=0; i<no_atoms; i++)
    {
      if ( atom_id[i] < (200*52) )
      { /* we have a lipid */
	if ( atom_id[i] < (100*52))
	{ /* we have a lipid in the upper layer */
	  if ( (((atom_id[i]-42) % 52) == 0) )
	  { /* we have one of the P43 atoms! */
	    special.YP_pos += y[i];
	    ++special.YP_count;
	    special.range_set = TRUE;
#if DEBUG == 3
errout("%d: Special: adding P43 in upper layer, YP_count=%d, YP_sum_pos=%f\n", my_number,
	special.YP_count,special.YP_pos);
#endif
	  }
	}
	else
	{ /* we have a lipid in the lower layer */
	  if ( (((atom_id[i]-42) % 52) == 0) )
	  { /* we have one of the glycerol atoms! */
	    special.YM_pos += y[i];
	    ++special.YM_count;
	    special.range_set = TRUE;
#if DEBUG == 3
errout("%d: Special: adding P43 in lower layer, YM_count=%d, YM_sum_pos=%f\n", my_number, 
	special.YM_count,special.YM_pos);
#endif
	  }
	}
      }
    }
  }
}


/************************************************************************/
static int use_water_params(int i, LOAD_MSG_ONE_NODE *my_data)
{
  if ((range.range_used == 1) && range.range_set)
  { /* in this case we select the SBOUND parameters according to 
     * a range criterion. In the present case it is: between the glycerol
     * backbones or not.
     */
    if ( (my_data->y_new[i] >= range.YM_pos) && (my_data->y_new[i] <= range.YP_pos) )
      return FALSE;
    else
      return TRUE;
  }
  else if (range.range_used == 2)
  { /* use water parameters ONLY for all atoms */
    return TRUE;
  }
  else
    return( is_water(my_data->res_string[i]) );
}

/************************************************************************/
static int is_water(char *res_string)
{
  return ((strcmp(res_string,"TIP3")==0) ||
  	  (strcmp(res_string,"HOH" )==0) );
}

/************************************************************************/
void pressure_update(LOAD_MSG_ALL_NODE *global_data, ENERGY_INFO *energy)
{ /* This routine is called on the MASTER from eganaly.c!! */
	PRESSURE area; /* contains the areas by which to divide the forces */

	/*** now we compute the areas. This has to be done every step, since the ***/
	/*** boundary might be moving, to achieve constant pressure 		 ***/
	compute_area(&area,global_data->sbound_droff);
	compute_pressure(&(energy->pressure),&area, PRESSURE_MODE);
	compute_pressure(&(energy->density), &area, DENSITY_MODE);
	
	if( global_data->keep_press_const == PRESSURE_MODE )
		constant_pressure(&(energy->pressure),&area,global_data);
	else if( global_data->keep_press_const == DENSITY_MODE )
		constant_density(&(energy->density),&area,global_data);
	else if( global_data->keep_press_const == SPECIAL_MODE )
		constant_special(global_data, 
		&(energy->special_YP_pos), &(energy->special_YM_pos),
		  energy->special_YP_count,  energy->special_YM_count);
	else if( global_data->keep_press_const == VOLUME_MODE )
		/* don't do anything */;
	else
		{
		  errout("Fatal error: Wrong MODE (%d) found for constant pressure!\n",
		 	 global_data->keep_press_const);
		  pex(-1);
		}
			
        calc_vir_pressure(	&(energy->pressure),
        			energy->kinetic,
        			global_data->sbound_droff,
                                energy->ekin_tensor);
}


/************************************************************************/
/* Here we have some routines usefull in the computation of the torque
   forces on water molecules in the SBOUND2 region. 
   The derivation is from Robert Kossmann's thesis, 
   implementation by Helmut Heller.
   
   Basic functions that need to know about torques:
    calc_spheric_sbound():	DONE.
    calc_cylindric_sbound():	DONE.
    calc_planar_sbound():	DONE.
   Other functions may need to be modified to pass the hydrogen coordinates down:
    calc_in_corner():		DONE.
    calc_in_edge():		DONE.
    round_corners():		DONE.
    adjust_sbound():		DONE.
*/


static int select_water(LOAD_MSG_ONE_NODE *my_data, int a_water_atom_index, WATER *selected_water)
{
  /* selects the three water atoms belonging to this selected water */
  int water_atom_count=0;
  int residue_id;
  
  /* Ini: */
  selected_water->O = -1;
  selected_water->H1 = -1;
  selected_water->H2 = -1;
  
  /* Sometimes we build fake my_data structures with just one atom.
     In such a case we can't select water molecules: */
  if(my_data->no_atoms < 3) return -1;

#if DEBUG == 7
  errout("Possible water molecules for torque computation\n");
#endif
  
  /* if not a water, we have an error: */
  if(!is_water(my_data->res_string[a_water_atom_index])) return (-1);
  
  /* now we look for the other two atoms. The idea is to find all atoms with the same
     res_id which are continous. This is not fool-proof: we should also check the residue name */
  residue_id = my_data->res_id[a_water_atom_index];
  
  /* first we go down to the beginning: */
  while ((residue_id == my_data->res_id[a_water_atom_index]) && (a_water_atom_index>=0)) 
    a_water_atom_index--; 
  a_water_atom_index++; /* go back into the range */
  
  /* assign the atoms: */
  while ((residue_id == my_data->res_id[a_water_atom_index]) && (a_water_atom_index <  my_data->no_atoms))
  {
    water_atom_count +=  assign_water(my_data, a_water_atom_index, selected_water);
    a_water_atom_index++;
  }
  
  if (water_atom_count == 3) 
    return 0; /* all three atoms were assigned */
  else
    return -1;
}


static int assign_water (LOAD_MSG_ONE_NODE *my_data, int a_water_atom_index, WATER *selected_water)
{
  /* The return value is 1 if the atom could be assigned, 0 otherwise */
  if(strcmp(my_data->atom_string[a_water_atom_index],"OH2")==0) 
    {selected_water->O=a_water_atom_index;return 1;}
  if(strcmp(my_data->atom_string[a_water_atom_index],"H1")==0) 
    {selected_water->H1=a_water_atom_index;return 1;}
  if(strcmp(my_data->atom_string[a_water_atom_index],"H2")==0) 
    {selected_water->H2=a_water_atom_index;return 1;}
  return 0;
}


static void compute_dipol_moment (LOAD_MSG_ONE_NODE *my_data, WATER *selected_water, VECTOR *dipol)
{  
  dipol->x = my_data->q[selected_water->H1] * 
  				(my_data->x_new[selected_water->H1] +  
				 my_data->x_new[selected_water->H2] - 
			     2 * my_data->x_new[selected_water->O]);
  dipol->y = my_data->q[selected_water->H1] * 
  				(my_data->y_new[selected_water->H1] +  
				 my_data->y_new[selected_water->H2] - 
			     2 * my_data->y_new[selected_water->O]);
  dipol->z = my_data->q[selected_water->H1] * 
  				(my_data->z_new[selected_water->H1] +  
				 my_data->z_new[selected_water->H2] - 
			     2 * my_data->z_new[selected_water->O]);
  return;
}


static double vector_length(VECTOR *vector)
{
  return (sqrt( scalar_product(vector,vector) ));
}


static double compute_pol_angle(VECTOR *dipol, VECTOR *normal)
{
  double theta;
  theta = scalar_product(dipol, normal);
  theta = theta / (vector_length(dipol) * vector_length(normal));
  theta = acos(theta) ;
  return(theta);
}


static double scalar_product(VECTOR *v1, VECTOR *v2)
{
  return (v1->x*v2->x + v1->y*v2->y + v1->z*v2->z);
}


static void cross_product(VECTOR *a, VECTOR *b, VECTOR *c)
{
  c->x = a->y * b->z - a->z * b->y;
  c->y = a->z * b->x - a->x * b->z;
  c->z = a->x * b->y - a->y * b->x;

  return;
}


static double M (double distance, double theta)
{  
  double d;
  
  /* Since we can have atoms outside the wall, we have to somehow extrapolate
     the torque amplitude for these distances (distance < 0). It is not obvious how
     to do that, so for now we just keep the amplitude at its maximum value. */
  if (distance < 0.0) d=0.0; else d=distance;
  
  /* Distance "distance" in Angstrom, theta in radian.
     This function comes from Kossi's thesis, page 69, Eq.(4.35). 
     Attention: the phase in Kossi's thesis is given in RADIAN. */
  return (-3.391534 * exp(-d) * sin(2.0*(theta + 3.06484)));
}


static double EM (double distance, double theta)
{ /* Compute the energy (potential) of the torque */
  double d;
  
  /* Since we can have atoms outside the wall, we have to somehow extrapolate
     the torque amplitude for these distances (distance < 0). It is not obvious how
     to do that, so for now we just keep the amplitude at its maximum value. */
  if (distance < 0.0) d=0.0; else d=distance;
  
  /* Distance "distance" in Angstrom, theta in radian.
     This function stems from Kossi's thesis, page 69, Eq.(4.35), but was integrated
     over theta. 
     Attention: the phase in Kossi's thesis is given in RADIAN. */
  return (-1.695757 * exp(-d) * cos(2.0*(theta + 3.06484)));
}


static void compute_torque (LOAD_MSG_ONE_NODE *my_data, WATER *selected_water, 
			    VECTOR *normal, /* vector normal to the boundary, 
			                       pointing outward (into the vacuum).
					       It is not necessary to have length one, but can 
					       be of any length. */
			    VECTOR *dipole,
			    double distance, /* distance (of C.M.) to surface */
			    EVECTOR *torque)
{
  double length, m;
  
  cross_product(normal, dipole, &(torque->pos));
  length = vector_length(&(torque->pos));
  if (length < SMALL) 
  { /* The torque is too small and we set it to zero: */
    torque->pos.x = 0.0;
    torque->pos.y = 0.0;
    torque->pos.z = 0.0;
    torque->energy = 0.0;
  }
  else
  { double theta; /* angle of dipole moment with normale */
    theta = compute_pol_angle(dipole,normal);
    m = M(distance, theta);
    torque->pos.x = m * torque->pos.x / length ;
    torque->pos.y = m * torque->pos.y / length ;
    torque->pos.z = m * torque->pos.z / length ; /* now we have length m */
    torque->energy = EM(distance, theta);
  }
  return;
}

static void compute_torque_forces (LOAD_MSG_ONE_NODE *my_data, int a_water_atom_index, VECTOR *normal,
				   double distance /* distance (of C.M.) to surface */ )
{
#if DO_TORQUE == TRUE
  WATER selected_water;
  VECTOR force;
  VECTOR dipole;
  EVECTOR torque;
  double length, m, dipole_length;
  int i;

#if DEBUG == 7
  errout("In torque computation\n");
#endif

  /* Select the whole water molecule or return if it is not a water atom. */
  if (select_water(my_data, a_water_atom_index, &selected_water) != 0) return;

#if DEBUG == 7
  errout("Got a water molecule for torque computation\n");
#endif
  
  /*  This function comes from Kossi's thesis, page 50, Eq.(4.23). */
  compute_dipol_moment(my_data, &selected_water, &dipole);
  compute_torque(my_data, &selected_water, normal, &dipole, distance, &torque);
  cross_product(&dipole, &(torque.pos), &force);
  
  /* Kossi has an error in his thesis, page 50, eq(4.23) that Berthold Heymann noticed:
     Kossi forgot to divide through the length (of the dipole, multiplied with the charge),
     furthermore, he forgot a "-" sign in the torque to force conversion! */
  length = vector_length(&force);
  dipole_length = vector_length(&dipole);
  m = vector_length(&(torque.pos));
  force.x = -(m * force.x / length) / dipole_length * my_data->q[selected_water.H1];
  force.y = -(m * force.y / length) / dipole_length * my_data->q[selected_water.H1];
  force.z = -(m * force.z / length) / dipole_length * my_data->q[selected_water.H1];
  
  /* Surprisingly, the torque also gives a RADIAL (in the normal direction) 
     force contribution! We get it analytically by building the derivative of the
     torque (amplitude) corresponding to the distance. Since the dependence is
     only that of a simple exponential ( exp(-d) ), all we get is a sign
     (which is swallowed, since for the force we need the NEGATIVE gradient). 
     Therefore, we can use the EM() function, which is in torque.energy.
     This force should only act on the oxygen, as all the other radial forces, too,
     only act on the heavy partner. 
     Because of the inverted force sign in EGO, we have to use -=. 
     BUT: because our distance is measured inward from the SBOUND,
          we get one more "-" sign, so we use here and only here += instead.
     ... and again BUT: experimentally shown that it should be -= . */

  length = vector_length(normal);
  i = selected_water.O;
  my_data->force_sbound[4*i+1] -= torque.energy * normal->x / length;
  my_data->force_sbound[4*i+2] -= torque.energy * normal->y / length;
  my_data->force_sbound[4*i+3] -= torque.energy * normal->z / length;

   /* I think it should read -= in the following force additions,
      because of the inverted forces in EGO */
   
  /* now we have the force on hydrogen 1: */
  i = selected_water.H1;
  my_data->force_sbound[4*i  ] += torque.energy/4.0;
  my_data->force_sbound[4*i+1] -= force.x;
  my_data->force_sbound[4*i+2] -= force.y;
  my_data->force_sbound[4*i+3] -= force.z;

  /* now we have the force on hydrogen 2: */
  i = selected_water.H2;
  my_data->force_sbound[4*i  ] += torque.energy/4.0;
  my_data->force_sbound[4*i+1] -= force.x;
  my_data->force_sbound[4*i+2] -= force.y;
  my_data->force_sbound[4*i+3] -= force.z;

  /* now we have the force on the oxygen: */
  i = selected_water.O;
  my_data->force_sbound[4*i  ] += torque.energy/2.0;
  my_data->force_sbound[4*i+1] -= (-2.0)*force.x;
  my_data->force_sbound[4*i+2] -= (-2.0)*force.y;
  my_data->force_sbound[4*i+3] -= (-2.0)*force.z;

#if DEBUG == 8
  average_torque_force += vector_length(&force)*4.0;
  average_torque_count += 3; /* we add three atoms */
#endif

#endif /* DO_TORQUE */
  return;
}

