/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: egclust.c	by M.Eichinger, Munich			**
**	Belongs2: EGO							**
**	Contents: clustering algorithm,	load-balancing			**	
**									**		
**  	Modif's:							**
**									**
*************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ego.h"
#include "timing.h"
#include "util_lib.h"
#include "egclust.h"
#include "egbolib.h"
#include "egalloc.h"
#include "egmarti.h"

#define DEBUG_INFO	0	/*** 0, 1 or 2 ***/
#define DEBUG		0
#define IFCHECKS	1
#define DEBUG_LOAD	2

#define BALANCE_COMPUTE 0	/*** 0 = optimize memory balance, ***/
				/*** 1 = optimize computation balance ***/
#define	CHARGEWAGEADD	0.2


/************* defined functions in cluster.c *************************/
static	int print_levelinfo(	int level_nr,
				struct_levelinfo *levelinfo,
				double used_time);

static	int units_to_cluster0(	int nr_of_atoms,
				char *unit_filen,
				struct_levelinfo *levelinfo,
				struct_cluster_reduced *cluster[],
				int *child_nr[],
				LOAD_MSG_ALL_NODE *global_data);

static	int make_cluster_stat(	char *cntfn,int level,
				struct_levelinfo *levelinfo,
				struct_cluster_reduced *cluster[],
				int *atom_nrs,
				LOAD_MSG_ALL_NODE *global_data);

static	int get_all_childs( 	int level,
				int parent,
				struct_cluster_reduced *cluster[],
				int *atom_nrs);

static	int resort_tree(	struct_treeinfo *treeinfo,
				struct_levelinfo *levelinfo,
				struct_cluster_reduced *cluster[],
				int *child_nr[]);

static int do_load_balancing(	int lev,int no_nodes,
				struct_treeinfo *treeinfo,
				struct_cluster_reduced *parentdat,
				struct_cluster_reduced *cluster[],
				int *atom_nrs);

static int	make_simvol_bbox(int nr_of_atoms);

static double	get_aufwand(int lev, int parent,struct_cluster_reduced *cluster[]);

static
int	buildup_other_levels(	int level,
				struct_levelinfo *levelinfo,
				struct_cluster_reduced *cluster[],
				int *child_nr[],
				int num_subchilds);

static
int	conect_units_cluster(	struct_levelinfo *levelinfo,
				struct_cluster_reduced *cluster[],
				int *child_nr[],
				int num_levels);

int
do_neuro_clustini(int numvec,double *xvec,double *yvec,double *zvec,
		int numcode,double *xcode,double *ycode,double *zcode);

static int
do_clustering(	int level,int levels,
		int inpn,double *xinpvec,double *yinpvec,double *zinpvec,
		int outn,double *xoutvec,double *youtvec,double *zoutvec,
		struct_cluster_reduced *cluster[],struct_levelinfo *levelinfo,
		int startoutn);

static
int make_parent_list(	int nr_of_vecs, double *vecx, double *vecy, double *vecz,
			int nr_of_code, double *codex, double *codey, double *codez,
			int *parentlist,
			int mincount,int maxcount,int optimum);

static int
do_fixpoint_iteration(	int level,int levels,
			int inpn,double *xinpvec,double *yinpvec,double *zinpvec,
			int outn,double *xoutvec,double *youtvec,double *zoutvec,
			struct_cluster_reduced *cluster[],int startoutn);
static int analyze_neutrals(struct_cluster_reduced *cluster[]);

/**********************************************************************/
static	int	ego_start_up=1;

static	int	nr_of_units=-1;
static	int	nr_of_neutrals=-1;
static	int	nr_of_polar=-1;
static	int	nr_of_charged=-1;
static	int	max_nr_fields=-1;

static	char		sdum[MAX_STRING_SIZE];	/* space for dummy-strings */
static	STRUCT_ATOMDATA *atom_data;
static	struct_cluster_reduced	*tempcluster;
static	int	*tempchild,tempbuffers_here=0;
static	double	simvol_bbox[6]; /*** bound box of the total simvolumen	***/
				/*** needed for load-balancing only !	***/

static	int	*pararr;	/*** holds for every unit the parentnumber of each level ***/
static	double	*xinpvec,*yinpvec,*zinpvec; /*** input  vectors x,y,z ***/
static	double	*xoutvec,*youtvec,*zoutvec; /*** output vectors x,y,z ***/
static	int	*inpparents;	/*** parent of the inputvector	***/
static	int	*parentscount;	/*** parent of the inputvector	***/

int	CLU_free_clusters(	struct_treeinfo *treeinfo,
				struct_levelinfo *levelinfo,
				struct_cluster_reduced *cluster[],
				int *child_nr[])
{
	int	i=levelinfo[0].nr_of_clusters; /*** no warning ***/

	for (i=0;i<treeinfo->levels;i++) {

		if (cluster[i] !=NULL)	myfree(cluster[i]);
		if (child_nr[i]!=NULL)	myfree(child_nr[i]);

		cluster[i] = NULL;
		child_nr[i]= NULL;
	}

	if (tempbuffers_here) {
		myfree(tempchild);
		myfree(tempcluster);
		tempbuffers_here=0;
	}

	if (!ego_start_up) {
		myfree(pararr);
		myfree(xinpvec);
		myfree(yinpvec);
		myfree(zinpvec);
		myfree(xoutvec);
		myfree(youtvec);
		myfree(zoutvec);
		myfree(inpparents);
		myfree(parentscount);
		ego_start_up=1;
	}
	free_neuroclust_ms();
	return 0;
}

/******************************************************************************/

static int algo=0;
static int neutral_branches=0;

int	CLU_buildup_clusters(	int nr_of_atoms, int no_of_nodes,
				STRUCT_ATOMDATA *atomdata,
				char *unitfn,
				struct_treeinfo *treeinfo,
				struct_levelinfo *levelinfo,
				struct_cluster_reduced *cluster[],
				int *child_nr[],
				LOAD_MSG_ALL_NODE *global_data)
{	int	i,j,htime;
	int	num_parents;	/*** number of parents on level ***/
	int	level,cur_parent;
	int	inpn,outn;	/*** number of input vectors ***/
	int	minlimit=0,maxlimit=0,absminlimit=0,optimum=0;
	double	used_sec[MAX_NO_LEVELS+1];

	set_allocinfo("buildup_clusters");
	
	/*** atom_data buffer for holding atom-data is a static pointer in this modul ***/
	atom_data=atomdata;
	algo=treeinfo->algo;

	/*** get the bounding-box of sim-volume ***/
	make_simvol_bbox(nr_of_atoms);
	txtout("Bounding-Box:  X: %# 7.4g \t %# 7.4g \t dx= %# 7.4g\n",
		simvol_bbox[0],simvol_bbox[1],simvol_bbox[1]-simvol_bbox[0]);
 	txtout("               Y: %# 7.4g \t %# 7.4g \t dy= %# 7.4g\n",
		simvol_bbox[2],simvol_bbox[3],simvol_bbox[3]-simvol_bbox[2]);
	txtout("               Z: %# 7.4g \t %# 7.4g \t dz= %# 7.4g\n",
		simvol_bbox[4],simvol_bbox[5],simvol_bbox[5]-simvol_bbox[4]);

	/*** readin the structural units ***/
	htime=init_timer();
	start_timer(htime);
	if (units_to_cluster0(	nr_of_atoms,
				unitfn,
				levelinfo,
				cluster,
				child_nr,
				global_data)==-1) {
                errout("Fatal error in Unit-File !\n");
		pex(-1);
	}

	stop_timer(htime);
	used_sec[0]=(double)get_rt_timer(htime);
	free_timer(htime);
	print_levelinfo(0,levelinfo,used_sec[0]);

	/**********************************************************************/
	if (treeinfo->branches<0) {	/*** signal for do it yourselv ***/
		double	dummy[MAX_NO_LEVELS+2],fac;
		int	i,ideal,dummy1[MAX_NO_LEVELS+2];

		treeinfo->levels = set_cluster_params(nr_of_atoms,
						dummy,dummy1);

		ideal = levelinfo[0].nr_of_clusters;
		for (i=1;i<treeinfo->levels;i++) {
			ideal /= treeinfo->clusterseq[i];
		}
		if (ideal<no_of_nodes) {
                        errout("Too many nodes for this molecule!\n");
                        errout("Use not more than %d nodes in your control file.\n",ideal);
                        errout("Best choice may be %d.\n",ideal/2);
			pex(-1);
		}
		fac = (double) ideal / no_of_nodes;
		treeinfo->branches = no_of_nodes * (int)(fac+0.25);
	}

	if (ego_start_up) {
		if (	 (algo==2) ||
			((algo==0)&&(analyze_neutrals(cluster)==1)) ) {
			neutral_branches = (treeinfo->branches*nr_of_neutrals)/nr_of_units;
			if (neutral_branches==0) neutral_branches=1;
			txtout("Separate clustering of neutral units with %d branches.\n",
				neutral_branches);
		} else {
			txtout("No separate clustering of neutral units!\n");			
			neutral_branches = 0;
		}
	}
#if (DEBUG_INFO==2)
        txtout("Number of NEUTRAL Branches %d of total %d\n",
		neutral_branches,treeinfo->branches);
#endif
	outn = treeinfo->branches;
	if (ego_start_up) {
		set_allocinfo("egclust ego-startup");
		pararr		= alloc_ints(levelinfo[0].nr_of_clusters*treeinfo->levels+10);
		xinpvec 	= alloc_doubles(levelinfo[0].nr_of_clusters+10);
		yinpvec 	= alloc_doubles(levelinfo[0].nr_of_clusters+10);
		zinpvec 	= alloc_doubles(levelinfo[0].nr_of_clusters+10);

		xoutvec 	= alloc_doubles(outn+10);
		youtvec 	= alloc_doubles(outn+10);
		zoutvec 	= alloc_doubles(outn+10);
		inpparents	= alloc_ints(levelinfo[0].nr_of_clusters+10);
		parentscount	= alloc_ints(levelinfo[0].nr_of_clusters+10);
		set_allocinfo("XXX");
	}

	nr_of_units=levelinfo[0].nr_of_clusters;

	/*** ini for first level ***/
	if (neutral_branches==0) {
               num_parents = 1;    /*** there is one "whole" parent ***/
	       for (i=0,j=treeinfo->levels-1;i<levelinfo[0].nr_of_clusters;j+=treeinfo->levels,i++) {
		     pararr[j] = 0;
	       }
	} else {
	       num_parents = 2;    /*** one for neutrals, one for charged/polar units ***/
	       for (i=0,j=treeinfo->levels-1;i<levelinfo[0].nr_of_clusters;j+=treeinfo->levels,i++) {
		     if (cluster[0][i].typ==NEUTRAL) pararr[j] = 0; /*** neutrals units ***/
		     else			     pararr[j] = 1; /*** charged/polar  ***/
	       }
	}

	minlimit=absminlimit=1;
	for (i=2;i<treeinfo->levels;i++) {
		absminlimit *= treeinfo->clusterseq[i];
	}
	for (level=treeinfo->levels-1;level>=1;level--) {
		htime=init_timer();
		start_timer(htime);
#if (DEBUG_INFO==2)
txtout("basic loop: level %d num_parents %d outn %d\n",level,num_parents,outn);
txtout("Minlimit %d\n",absminlimit);
#endif
		for (cur_parent=0;cur_parent<num_parents;cur_parent++) {
			int k,l,offset,myoutn=outn;
			int startoutn=0;
			if ((neutral_branches!=0)&&(level==treeinfo->levels-1)) {
				if (cur_parent==0) {
					myoutn    = neutral_branches;
					startoutn = 0;
				}			
				if (cur_parent==1) { 
					myoutn    = outn-neutral_branches;		
					startoutn = neutral_branches;
				}		
			}
						
			/*** extract input vectors of subbranch ***/
			inpn = 0;
			for (k=0,l=level;k<levelinfo[0].nr_of_clusters;k++,l+=treeinfo->levels) {
				if (pararr[l]==cur_parent) {
					xinpvec[inpn]	= cluster[0][k].cen_geom[0];
					yinpvec[inpn]	= cluster[0][k].cen_geom[1];
					zinpvec[inpn++] = cluster[0][k].cen_geom[2];
				}
			}

			optimum   = max(absminlimit+1,(int)((inpn+myoutn+1)/myoutn));
			minlimit  = max(absminlimit,optimum - (int)((optimum+10)/10));
			maxlimit  = max(3,          optimum + (int)((optimum+10)/10));
			
#if (DEBUG_INFO==2)
txtout("start do_neuro_clustering cur_parent %d with %d vectors, %d codebooks.\n",
	cur_parent,inpn,myoutn);
txtout("Max-limit %d\n",maxlimit);
#endif
			do_clustering(	level,treeinfo->levels,
					inpn  ,xinpvec,yinpvec,zinpvec,
					myoutn,xoutvec,youtvec,zoutvec,
					cluster,levelinfo,startoutn);

#if (DEBUG_INFO==2)
txtout("before make parent list.\n");
#endif
			make_parent_list(	inpn  ,xinpvec,yinpvec,zinpvec,
						myoutn,xoutvec,youtvec,zoutvec,
						inpparents,minlimit,maxlimit,optimum);
#if (DEBUG_INFO==2)
txtout("after make parent list.\n");
#endif
			/*** sort inpparents to pararr ***/
			inpn = 0;
			if (	(neutral_branches!=0) &&
				(level==treeinfo->levels-1) &&
				(cur_parent==1)	) {
				offset = cur_parent*neutral_branches;
			} else {
				offset = cur_parent*myoutn;
			}	
			for (k=0,l=level;k<levelinfo[0].nr_of_clusters;k++,l+=treeinfo->levels) {
				if (pararr[l]==cur_parent) {
					pararr[l-1] = offset + inpparents[inpn];
					inpn++;
				}
			}
		}
		if (	(neutral_branches!=0) &&
		    
			(level==treeinfo->levels-1)) {
			num_parents = outn;
		} else {	
			num_parents *= outn;
		}
		levelinfo[level].nr_of_clusters   = num_parents;
		levelinfo[level+1].nr_of_childs   = num_parents;
#if (DEBUG_INFO==2)
txtout("levelinfo[%d].nr_of_clusters = %d\n",level,levelinfo[level].nr_of_clusters);
#endif
		outn      = treeinfo->clusterseq[level];
		absminlimit /= treeinfo->clusterseq[i];
		absminlimit  = max(1,absminlimit);
		
		stop_timer(htime);
		used_sec[level]=(double)get_rt_timer(htime);
		free_timer(htime);
	}

	if (treeinfo->levels>1)
		conect_units_cluster(	levelinfo,
					cluster,
					child_nr,
					treeinfo->levels);

	for (level=2;level<treeinfo->levels;level++) {
		buildup_other_levels(level,levelinfo,cluster,child_nr,
			treeinfo->clusterseq[level]);
	}
#if (DEBUG_INFO==2)
txtout("Before resort tree first time\n.");
#endif
	resort_tree(	treeinfo,
			levelinfo,
			cluster,
			child_nr);
#if (DEBUG_INFO==2)
txtout("After resort tree first time.\n");
#endif
	for (j=1;j<treeinfo->levels;j++) {
		print_levelinfo(j,levelinfo,used_sec[j]);
	}
	for (j=0;j<treeinfo->levels;j++) {
		char	str[100];
                sprintf(str,"cluster%1d.out",j);
		make_cluster_stat(str,j,levelinfo,cluster,child_nr[0],global_data);
	}
#if (DEBUG_INFO==2)
txtout("Before load balancing.\n");
#endif
	do_load_balancing(	treeinfo->levels-1,
				no_of_nodes,
				treeinfo,
				cluster[treeinfo->levels-1],
				cluster,
				child_nr[0]);
#if (DEBUG_INFO==2)
txtout("Before resort tree second time.\n");
#endif
	resort_tree(	treeinfo,
			levelinfo,
			cluster,
			child_nr);

	ego_start_up=0; /*** from now we do do_fixpoint_iteration ***/
	return 0;
}

/******************************************************************************/

static int
do_clustering(	int level,int levels,
		int inpn,double *xinpvec,double *yinpvec,double *zinpvec,
		int outn,double *xoutvec,double *youtvec,double *zoutvec,
		struct_cluster_reduced *cluster[],
		struct_levelinfo *levelinfo,
		int startoutn)
{
	int a=levelinfo[0].nr_of_clusters;a++;

	if ((ego_start_up)||(level<levels-1)) {
		do_neuroclust_ms(inpn,xinpvec,yinpvec,zinpvec,
					outn,xoutvec,youtvec,zoutvec,1);
		if (level==levels-1) {
			do_neuroclust_ms(inpn,xinpvec,yinpvec,zinpvec,
					outn,xoutvec,youtvec,zoutvec,0);

		}
	} else {
#if (DEBUG_INFO==2)
txtout("do fixpoint iteration %d %d inp %d out %d\n",level,levels,inpn,outn);
#endif
		do_fixpoint_iteration(	level,levels,
					inpn,xinpvec,yinpvec,zinpvec,
					outn,xoutvec,youtvec,zoutvec,
					cluster,startoutn);
	}
	return 0;
}

static int
do_fixpoint_iteration(	int level,int levels,
			int inpn,double *xinpvec,double *yinpvec,double *zinpvec,
			int outn,double *xoutvec,double *youtvec,double *zoutvec,
			struct_cluster_reduced *cluster[],
			int startoutn)
{
	int k,l,i;
	double t1=xinpvec[0],t2=yinpvec[0],t3=zinpvec[0];
	t1++; t2++; t3++; /*** no warnings ***/

	for (i=0;i<outn;i++) {
		xoutvec[i] = 0.0;
		youtvec[i] = 0.0;
		zoutvec[i] = 0.0;
	}

	for (k=0;k<nr_of_units;k++)		inpparents[k] = 0;

	inpn = 0;
	for (k=0,l=level-1;k<nr_of_units;k++,l+=levels) {
		int parent=pararr[l]-startoutn;
		if ((parent<0)||(parent>=outn)) continue;
	
		xoutvec[parent] += cluster[0][k].cen_geom[0];
		youtvec[parent] += cluster[0][k].cen_geom[1];
		zoutvec[parent] += cluster[0][k].cen_geom[2];
		inpparents[parent]++;
	}

	for (i=0;i<outn;i++) {
		xoutvec[i] /= inpparents[i];
		youtvec[i] /= inpparents[i];
		zoutvec[i] /= inpparents[i];
	}
#if (DEBUG_INFO==2)
txtout("Exit fixpoint iteration\n");
#endif
	return 0;
}

static
int make_parent_list(	int nr_of_vecs, double *vecx,  double *vecy,  double *vecz,
			int nr_of_code, double *codex, double *codey, double *codez,
			int *parentlist,int minlimit,int maxlimit,int optimum)
{
	int	veci,codei;
	int	repeat_correction=2;
#if (DEBUG_INFO==2)
txtout("Enter make_parent_list with minlimit %d maxlimit %d optimum %d\n",
minlimit,maxlimit,optimum);
#endif
	for (veci=0;veci<nr_of_vecs;veci++) { /*** loop over input vectors ***/
		double	mindist  = -1.0;
		parentlist[veci] = -1;
		for (codei=0;codei<nr_of_code;codei++) { /*** loop over code-books ***/
			double dx=vecx[veci]-codex[codei];
			double dy=vecy[veci]-codey[codei];
			double dz=vecz[veci]-codez[codei];
			double dist = dx*dx + dy*dy + dz*dz;
			if ((mindist<0.0)||(dist<mindist)) {
				parentlist[veci] = codei;
				mindist = dist;
			}
		}
	}
	
	/**** check weather every parent got enough childs ! ***/
	for (codei=0;codei<nr_of_code;codei++) parentscount[codei]=0;
	for (veci=0;veci<nr_of_vecs;veci++)    parentscount[parentlist[veci]]++;

do_correction:
	for (codei=0;codei<nr_of_code;codei++) {
		double dx,dy,dz,dist;
		int minopt=(int)(optimum+minlimit)/2;
		int maxopt=(int)(optimum+maxlimit)/2;
#if (DEBUG_INFO==2)
txtout("Correct-level %d: Cluster %3d has %5d units.\n",
repeat_correction,codei,parentscount[codei]);
#endif
		if (	(parentscount[codei]>=minlimit) &&
			(parentscount[codei]<=maxlimit)) continue;
		
		if (parentscount[codei]<minlimit) {
#if (DEBUG_INFO>=1)			
			txtout("Cluster %4d with too few  children (%3d<%3d),...",
				codei,parentscount[codei],minlimit);
#endif
			while (parentscount[codei]<minopt) {		
				double	mindist=-1.0;
				int     newveci=-1;	
				for (veci=0;veci<nr_of_vecs;veci++) {
					if (parentlist[veci]==codei) continue;
					dx=vecx[veci]-codex[codei];
					dy=vecy[veci]-codey[codei];
					dz=vecz[veci]-codez[codei];
					dist = dx*dx + dy*dy + dz*dz;
					if ( ((mindist<0.0)||(dist<mindist)) &&
				      	      (parentscount[parentlist[veci]]>minlimit) ) {
						newveci = veci;
						mindist = dist;
					}
				}
				if (newveci==-1) {
#if (DEBUG_INFO>=1)					
					errout("\nCould not correct parents list %d %d\n",
						codei,minopt);					
#endif
					break;	
				} else {
					parentscount[parentlist[newveci]]--;
					parentlist[newveci]=codei;
					parentscount[codei]++;
				}				
			}
#if (DEBUG_INFO>=1)			
			txtout("corrected to %d!\n",parentscount[codei]);
#endif		
		}	
		
		/*** Here we reduce the number of childs of parent codei ***/
		/*** loop over all childs belonging to parent codei.	 ***/
		/*** find the child which has the smallest distance 	 ***/
		/*** to all other parent except parent codei	    	 ***/
		if (parentscount[codei]>maxlimit) {
#if (DEBUG_INFO>=1)			
			txtout("Cluster %4d with too many children (%3d>%3d),...",
				codei,parentscount[codei],maxlimit);
#endif
			while (parentscount[codei]>maxopt) {
				double	mindist =-1.0;
				int     newcodei=-1,newveci=-1;	
				int	i;
			
				for (veci=0;veci<nr_of_vecs;veci++) {
					if (parentlist[veci]!=codei) continue;
					for (i=0;i<nr_of_code;i++) {
						if (i==codei) continue;
						if (parentscount[i]>parentscount[codei]) continue;
						
						dx=vecx[veci]-codex[i];
						dy=vecy[veci]-codey[i];
						dz=vecz[veci]-codez[i];
						dist = dx*dx + dy*dy + dz*dz;
						if ((mindist<0.0)||(dist<mindist)) {
							newcodei = i;
							newveci  = veci;
							mindist  = dist;
						}
					}
				}
				
				if (newveci==-1) {
#if (DEBUG_INFO>=1)					
					errout("\nCould not correct parents list %d %d\n",
						codei,maxopt);					
#endif
					break;
				} else {
					parentscount[codei]--;
					parentlist[newveci]=newcodei;
					parentscount[newcodei]++;
				}
			}
#if (DEBUG_INFO>=1)			
			txtout("corrected to %d!\n",parentscount[codei]);
#endif		
		}		
	}
	
	repeat_correction--;
	if (repeat_correction>0)	goto do_correction;
	else				return 0;
}

static
int	buildup_other_levels(	int level,
				struct_levelinfo *levelinfo,
				struct_cluster_reduced *cluster[],
				int *child_nr[],
				int num_subchilds)
{
	int parent,child_pos,child;
#if (DEBUG_INFO==2)
txtout("enter buildup_other_levels %d: %d %d\n",level,
	levelinfo[level].nr_of_clusters,
	levelinfo[level-1].nr_of_clusters);
#endif
	if (cluster[level]==NULL) {
#if (DEBUG_INFO==2)
txtout("Allocated clusters on level %d: %d\n",
level,levelinfo[level].nr_of_clusters);
#endif
		cluster[level] = alloc_cluster_reduced(levelinfo[level].nr_of_clusters+5);
	}
	if (child_nr[level]==NULL) {
		child_nr[level] = alloc_ints(levelinfo[level-1].nr_of_clusters+5);
	}

	child_pos=0;
	for (parent=0;parent<levelinfo[level].nr_of_clusters;parent++) {
		cluster[level][parent].nr_childs = num_subchilds;
		cluster[level][parent].child_pos = child_pos;
		child_pos+=cluster[level][parent].nr_childs;
	}

	for (child=0;child<levelinfo[level-1].nr_of_clusters;child++) {
		child_nr[level][child] = child;
	}
#if (DEBUG_INFO==2)
txtout("exit buildup_other_levels\n");
#endif
	return 0;
}

static
int	conect_units_cluster(	struct_levelinfo *levelinfo,
				struct_cluster_reduced *cluster[],
				int *child_nr[],
				int num_levels)
{
	int	i;
	int	child_pos,parent,unit;
#if (DEBUG_INFO==2)
txtout("enter conect_units_cluster %d %d ptr %p %p\n",
	levelinfo[1].nr_of_clusters,
	levelinfo[0].nr_of_clusters,
	cluster[1],child_nr[1]);
#endif
	if (cluster[1]==NULL) {
		cluster[1]  = alloc_cluster_reduced(levelinfo[1].nr_of_clusters+5);
	}
	if (child_nr[1]==NULL) {
		child_nr[1] = alloc_ints(levelinfo[0].nr_of_clusters+5);
	}

	/*** ini number of childs ***/
	for (parent=0;parent<levelinfo[1].nr_of_clusters;parent++) {
		cluster[1][parent].nr_childs = 0;
	}
	for (unit=0,i=0;unit<levelinfo[0].nr_of_clusters;unit++,i+=num_levels) {
		parent = pararr[i];
#if IFCHECKS
		if ((parent<0)||(parent>=levelinfo[1].nr_of_clusters)) {
                        errout("Fatal error: should never happen in egclust.c");
			pex(-1);
		}
#endif
		cluster[1][parent].nr_childs++;
	}
	/*** if any parent has no child this is it shit ! --> we correct it by hand ***/
	for (parent=0,child_pos=0;parent<levelinfo[1].nr_of_clusters;parent++) {
		int badnum=0;
		if (cluster[1][parent].nr_childs==0) {
			badnum++;
		}
		if (badnum>0) {
                        errout("Found %d parents with no children !\n",badnum);
			pex(-1);
		}
	}
#if (DEBUG_INFO==2)
txtout("enter conect_units_cluster Phase-2 %d %d ptr %p %p\n",
	levelinfo[1].nr_of_clusters,
	levelinfo[0].nr_of_clusters,
	cluster[1],child_nr[1]);
#endif
	/*** set up positions ***/
	for (parent=0,child_pos=0;parent<levelinfo[1].nr_of_clusters;parent++) {
		cluster[1][parent].child_pos = child_pos;
		child_pos+=cluster[1][parent].nr_childs;
	}
#if (DEBUG_INFO==2)
txtout("enter conect_units_cluster Phase-3 %d %d ptr %p %p\n",
	levelinfo[1].nr_of_clusters,
	levelinfo[0].nr_of_clusters,
	cluster[1],child_nr[1]);
#endif
	for (unit=0,i=0;unit<levelinfo[0].nr_of_clusters;unit++,i+=num_levels) {
		parent=pararr[i];
		child_nr[1][cluster[1][parent].child_pos++] = unit;
	}
	for (parent=0;parent<levelinfo[1].nr_of_clusters;parent++) {
		cluster[1][parent].child_pos -= cluster[1][parent].nr_childs;
/**
txtout("parent %d  childs %d\n",parent,cluster[1][parent].nr_childs);
**/
	}

	levelinfo[1].nr_of_childs   = levelinfo[0].nr_of_clusters;
#if (DEBUG_INFO==2)
txtout("conect_units_cluster\n");
#endif
	return 0;
}

/**********************************************************************/

/***********************************************************************************/
/*** now we want to resort the cluster-structure in a way so that the		****/
/*** subchilds are given by the index-range :					*****
 cluster[l][p].child_pos, ...., cluster[l][p].child_pos + cluster[l][p].nr_childs - 1
**** so we do later not need *child_nr[] any more !! Only the child_nr[0]-array  ***/
/*** (which contains the atom-number of the units) is needed in loader.c, for	 ***/
/*** resorting the atoms. after that even child_nr[0] is not needed any more *******/
/***********************************************************************************/
static int
resort_tree(	struct_treeinfo *treeinfo,
		struct_levelinfo *levelinfo,
		struct_cluster_reduced *cluster[],
		int *child_nr[])
{
	int	level;
	int	parent,child;
	int	newchild,i;

	/********** here the resorting takes place ************************/
	if (!tempbuffers_here) {
#if (DEBUG_INFO==2)
txtout("Alloc in resort tree %d %d\n",levelinfo[0].nr_of_clusters,treeinfo->levels);
#endif
		tempchild  = alloc_ints(levelinfo[0].nr_of_clusters+5);
		tempcluster= alloc_cluster_reduced(levelinfo[0].nr_of_clusters+5);
		tempbuffers_here=1;
	}

	for (level=treeinfo->levels-2;level>=0;level--) {  /***** for each level ****/
		newchild=0;
#if (DEBUG_INFO==2)
txtout("Level %d  parents %d\n",level+1,levelinfo[level+1].nr_of_clusters);
#endif
		memcpy(tempcluster,cluster[level],
			sizeof(struct_cluster_reduced)*levelinfo[level].nr_of_clusters);

						/** for each parent on this level ****/
		for (parent=0;parent<levelinfo[level+1].nr_of_clusters;parent++) {

			int cur_childpos = cluster[level+1][parent].child_pos;


#if DEBUG
txtout("    parent %d    child_pos %d    newchild %d  subchilds %d\n",
parent,cur_childpos,newchild,cluster[level+1][parent].nr_childs);
#endif

						/*** copy child-data to right position ***/
			cluster[level+1][parent].child_pos=newchild; /* correct position !! */
			for (i=0;i<cluster[level+1][parent].nr_childs;i++) {
				child=child_nr[level+1][cur_childpos+i];

#if DEBUG
txtout("        copy child %d to %d\n",child,newchild);
#endif
				tempchild[newchild] = newchild;

				memcpy( &(cluster[level][newchild++]),
					&(tempcluster[child]),
					sizeof(struct_cluster_reduced) );
			}
		}
#if DEBUG
txtout("        recopy it %d \n",levelinfo[level+1].nr_of_childs);
#endif

		memcpy(child_nr[level+1],tempchild,
				sizeof(int)*levelinfo[level+1].nr_of_childs);
	}

	/*** now we want to determin the ranges of cluster-numbers on each level***/
	/*** for each branch *************************************************/
	/*** we get this by going down each branch always on the very left ***/
	/*** right  and side (only works if sorted by the way abouve !!) *****/
	{
	int	toplevel=treeinfo->levels-1;
	int	nr_branches=levelinfo[toplevel].nr_of_clusters;
	int	level,branch;

	for (level=toplevel;level>=0;level--) {
		if (levelinfo[level].first_cluster!=NULL) myfree(levelinfo[level].first_cluster);
		if (levelinfo[level].last_cluster !=NULL) myfree(levelinfo[level].last_cluster);

		levelinfo[level].first_cluster = alloc_ints(nr_branches);
		levelinfo[level].last_cluster  = alloc_ints(nr_branches);
	}

	for (branch=0;branch<nr_branches;branch++) {

		int left =	  cluster[toplevel][branch].child_pos;
		int right= left + cluster[toplevel][branch].nr_childs - 1;

		levelinfo[toplevel].first_cluster[branch] = branch;
		levelinfo[toplevel].last_cluster[branch]  = branch;
#if DEBUG
txtout("\nRange for Branch %d:\n",branch);
#endif
		for (level=toplevel-1;level>=0;level--) {
			levelinfo[level].first_cluster[branch] = left;
			levelinfo[level].last_cluster[branch]  = right;

#if DEBUG
txtout("l %2d  (% 3d  % 3d)  ",level,left,right);
#endif
			left = cluster[level][left ].child_pos;
			right= cluster[level][right].child_pos+
					cluster[level][right].nr_childs-1;
		}
	}

	}
	return 0;
}

/**************************************************************************/

static
int	print_levelinfo(	int level_nr,
				struct_levelinfo *levelinfo,
				double used_time)
{
        txtout("Level % 2d: Nr_of_clusters % 4d    Nr_of_Children % 4d   (Time: % .4f) \n",
			level_nr,
			levelinfo[level_nr].nr_of_clusters,
			levelinfo[level_nr].nr_of_childs,
			used_time);
	return 0;
}

static
int	make_cluster_stat(	char *cntfn,
				int level,
				struct_levelinfo *levelinfo,
				struct_cluster_reduced *cluster[],
				int *atom_nrs,
				LOAD_MSG_ALL_NODE *global_data)
{
	int	i,j;
	FILE	*h_cnt;
	int	*temp_buf,maxatoms=1000;
	double	avegyr=0.0,sigmagyr=0.0;
	double	sum_charge=0.0;

	temp_buf = alloc_ints(maxatoms);

	/************** open statistic-file *********************************/
        if ((h_cnt=fopen(cntfn,"w"))==NULL) {
                errout("Can not open statistic file for unit-counts.");
		return -1;
	}
        fprintf(h_cnt,"Distribution of atoms to units.\n");
        fprintf(h_cnt,"%d\n",0);
        fprintf(h_cnt,"Atoms  Count units: total  charges neutral  polar  charged\n");
        fprintf(h_cnt,"\n");
        fprintf(h_cnt,"%d\n",levelinfo[level].nr_of_clusters);
        fprintf(h_cnt,"Atoms   Charge              Dipol                     RefP                       BBox    R-Gyr\n");

	for (i=0;i<levelinfo[level].nr_of_clusters;i++) {
		int	k,sum_atoms;
		double	dip[3];
		double	rgyr=0.0;

				/****************  get me the charges ****/
		dip[0]=0.0;
		dip[1]=0.0;
		dip[2]=0.0;
				/******* radius of Gyration **************/
		sum_atoms=CLU_get_all_atoms(level,i,&maxatoms,&temp_buf,cluster,atom_nrs);

		k=temp_buf[0];
		cluster[level][i].bbox[1]=cluster[level][i].bbox[0]=atom_data[k].ko[0];
		cluster[level][i].bbox[3]=cluster[level][i].bbox[2]=atom_data[k].ko[1];
		cluster[level][i].bbox[5]=cluster[level][i].bbox[4]=atom_data[k].ko[2];

		cluster[level][i].cen_geom[0]=0.0;
		cluster[level][i].cen_geom[1]=0.0;
		cluster[level][i].cen_geom[2]=0.0;
		cluster[level][i].cen_geom[3]=0.0;
		cluster[level][i].status     =0;
		
		sum_charge = 0.0;
		cluster[level][i].typ=NEUTRAL;

		for (j=0;j<sum_atoms;j++) {
			k=temp_buf[j];
			sum_charge		      += atom_data[k].charge;
			if (fabs(sum_charge)>SMALL) cluster[level][i].typ=POLAR;
			cluster[level][i].cen_geom[0] += atom_data[k].ko[0];
			cluster[level][i].cen_geom[1] += atom_data[k].ko[1];
			cluster[level][i].cen_geom[2] += atom_data[k].ko[2];
		}
		cluster[level][i].cen_geom[0] /= (double) sum_atoms;
		cluster[level][i].cen_geom[1] /= (double) sum_atoms;
		cluster[level][i].cen_geom[2] /= (double) sum_atoms;

		if (fabs(sum_charge)>SMALL) cluster[level][i].typ=CHARGED;
		cluster[level][i].moment[0] = sum_charge*global_data->f_elstat;

                fprintf(h_cnt,"%5d % 5.2f   ",sum_atoms,cluster[level][i].moment[0]/global_data->f_elstat);
                fprintf(h_cnt,"% 6.3f % 6.3f % 6.3f",dip[0],dip[1],dip[2]);
                fprintf(h_cnt,"  % 6.3f % 6.3f % 6.3f      ",
				cluster[level][i].cen_geom[0],
				cluster[level][i].cen_geom[1],
				cluster[level][i].cen_geom[2]);

		for (j=0;j<sum_atoms;j++) {
			double d1,d2,d3;
			k=temp_buf[j];
			d1 = cluster[level][i].cen_geom[0] - atom_data[k].ko[0];
			d2 = cluster[level][i].cen_geom[1] - atom_data[k].ko[1];
			d3 = cluster[level][i].cen_geom[2] - atom_data[k].ko[2];
			rgyr += d1*d1 + d2*d2 + d3*d3;

			if (cluster[level][i].bbox[0]>atom_data[k].ko[0])
					cluster[level][i].bbox[0]=atom_data[k].ko[0];

			if (cluster[level][i].bbox[1]<atom_data[k].ko[0])
					cluster[level][i].bbox[1]=atom_data[k].ko[0];

			if (cluster[level][i].bbox[2]>atom_data[k].ko[1])
					cluster[level][i].bbox[2]=atom_data[k].ko[1];

			if (cluster[level][i].bbox[3]<atom_data[k].ko[1])
					cluster[level][i].bbox[3]=atom_data[k].ko[1];

			if (cluster[level][i].bbox[4]>atom_data[k].ko[2])
					cluster[level][i].bbox[4]=atom_data[k].ko[2];

			if (cluster[level][i].bbox[5]<atom_data[k].ko[2])
					cluster[level][i].bbox[5]=atom_data[k].ko[2];
		}

		rgyr=sqrt(rgyr/sum_atoms);

				/************* boundry box *****************/
                fprintf(h_cnt,"% 6.3f % 6.3f % 6.3f",
			cluster[level][i].bbox[1]-cluster[level][i].bbox[0],
			cluster[level][i].bbox[3]-cluster[level][i].bbox[2],
			cluster[level][i].bbox[5]-cluster[level][i].bbox[4]);
                fprintf(h_cnt,"   % 6.3f\n",rgyr);
		avegyr	 += rgyr;
		sigmagyr += rgyr*rgyr;
	}

        fprintf(h_cnt,"\n");
	myfree(temp_buf);

	avegyr /= levelinfo[level].nr_of_clusters;
	sigmagyr = sqrt(fabs(sigmagyr/levelinfo[level].nr_of_clusters - avegyr*avegyr));
	fclose(h_cnt);
        txtout("Cl.-Statistic-file <%s> written: <RGyr>= %.4g Sig = %.4g\n",
		cntfn,avegyr,sigmagyr);
	return 0;
}

/*****************************************************************************/


static	int	*atom_buf;
static	int	cur_size=0;
static	int	cur_atoms=0;
static	double	normalized_cur_atoms=0.0;

/***** this functions returns all atom-numbers, which belong to the cluster**/
/***** with number "parent" at the level "level"  ***************************/

int	CLU_get_all_atoms(int level,	/** level-index: 0 -> unitlevel,...**/
			int parent,	/** parent-index ***/
			int *nr,	/** nr of elements for which buffer is allocated */
			int **buf,	/** the buffer for the atom-indices (OLD-EGO) */
			struct_cluster_reduced *cluster[],
			int *old_atom_nrs)
{
	atom_buf=*buf;
	cur_size=*nr;
	cur_atoms=0;
	normalized_cur_atoms=0.0;

	get_all_childs(level,parent,cluster,old_atom_nrs);

	*nr=cur_size;
	*buf=atom_buf;
	return cur_atoms;
}

/**************************************************************************/

static
int	get_all_childs( int level,
			int parent,
			struct_cluster_reduced *cluster[],
			int *old_atom_nrs)
{
	if (level==0) { 	/* now we reached the atom-level */
		int i;		/* fill atom_buf with the atom-numbers */

		int cur_pos=cluster[0][parent].child_pos;
		for (i=0;i<cluster[0][parent].nr_childs;i++) {
			if (cur_size<=cur_atoms) {
				cur_size+=50;
				atom_buf=(int*)merealloc(atom_buf,
					sizeof(int)*cur_size,
					sizeof(int)*(cur_size-50));
			}
			atom_buf[cur_atoms++]=old_atom_nrs[cur_pos++];

			normalized_cur_atoms++; 		
			
			/*** here wage the atoms in charged or polar slightly ***/
			/*** more than atoms in neutral clusters              ***/
			if (cluster[0][parent].typ!=NEUTRAL) {
				normalized_cur_atoms+=CHARGEWAGEADD;
			}
		}
	} else {		/* we are at a higher level, so we step down. */
		int i;		/* recursive call of get_all_childs ***********/
		int nr_childs=cluster[level][parent].nr_childs;
		int child_pos=cluster[level][parent].child_pos;
		for (i=0;i<nr_childs;i++) {
			int subchild=child_pos++;
			get_all_childs(level-1,subchild,cluster,old_atom_nrs);
		}
	}
	return 0;
}

/****************************************************************************/

static int
make_simvol_bbox(int nr_of_atoms)
{
	int i;

	simvol_bbox[0]=simvol_bbox[1]=atom_data[0].ko[0];
	simvol_bbox[2]=simvol_bbox[3]=atom_data[0].ko[1];
	simvol_bbox[4]=simvol_bbox[5]=atom_data[0].ko[2];

	for (i=0;i<nr_of_atoms;i++) {
		if (simvol_bbox[0]>atom_data[i].ko[0]) simvol_bbox[0]=atom_data[i].ko[0];
		if (simvol_bbox[1]<atom_data[i].ko[0]) simvol_bbox[1]=atom_data[i].ko[0];
		if (simvol_bbox[2]>atom_data[i].ko[1]) simvol_bbox[2]=atom_data[i].ko[1];
		if (simvol_bbox[3]<atom_data[i].ko[1]) simvol_bbox[3]=atom_data[i].ko[1];
		if (simvol_bbox[4]>atom_data[i].ko[2]) simvol_bbox[4]=atom_data[i].ko[2];
		if (simvol_bbox[5]<atom_data[i].ko[2]) simvol_bbox[5]=atom_data[i].ko[2];
	}
	return 0;
}

static
int	units_to_cluster0(	int nr_of_atoms,
				char *unit_filen,
				struct_levelinfo *levelinfo,
				struct_cluster_reduced *cluster[],
				int *child_nr[],
				LOAD_MSG_ALL_NODE *global_data)
{
	FILE	*fopen();
	FILE	*h_unitfile;
	int	i,j,testi;
	int	child_pos;

        if ((h_unitfile=fopen(unit_filen,"r"))==NULL) {
                errout("Can not open unit-file <%s>.\n",unit_filen);
		return -1;
	}
						/* skip header */
	skip_line(h_unitfile,1);

	/*************** get number of units ********************************/
        fscanf(h_unitfile,"%s%s%s%d",sdum,sdum,sdum,&nr_of_units);
	if (nr_of_units<1) {
                errout("Illegal number (%d) of units.\n",nr_of_units);
		fclose(h_unitfile);
		return -1;
	}
        fscanf(h_unitfile,"%s%s%s%d",sdum,sdum,sdum,&nr_of_neutrals);
	if (nr_of_neutrals<0) {
                errout("Illegal number (%d) of neutral units.\n",nr_of_neutrals);
		fclose(h_unitfile);
		return -1;
	}
        fscanf(h_unitfile,"%s%s%s%d",sdum,sdum,sdum,&nr_of_polar);
	if (nr_of_polar<0) {
                errout("Illegal number (%d) of polar units.\n",nr_of_polar);
		fclose(h_unitfile);
		return -1;
	}
        fscanf(h_unitfile,"%s%s%s%d",sdum,sdum,sdum,&nr_of_charged);
	if (nr_of_charged<0) {
                errout("Illegal number (%d) of charged units.\n",nr_of_charged);
		fclose(h_unitfile);
		return -1;
	}
        fscanf(h_unitfile,"%s%s%s%s%d",sdum,sdum,sdum,sdum,&max_nr_fields);
	if (max_nr_fields<1) {
                errout("Illegal number (%d) of fields.",max_nr_fields);
		fclose(h_unitfile);
		return -1;
	}

	skip_line(h_unitfile,2);

	/************ now we scanned the header ***************************/
	/************ making now top-level: level = 0 *********************/
        /*** here don't have to care about the changin size of the buffers */
        /****because they don't change their size 			****/
	if (cluster[0]==NULL)	cluster[0]  = alloc_cluster_reduced(nr_of_units+5);
	if (child_nr[0]==NULL)	child_nr[0] = alloc_ints(nr_of_atoms+5);

	/************ lets get data of units *****************************/
	child_pos=0;
	for (i=0;i<nr_of_units;i++) {
		int	cur_child;
		int	idum,nr_childs,utype;
		double	sum_charge=0.0;

		/***** basic data of unit *******/
                if (fscanf(h_unitfile,"%d %d %d %d %d",
			&testi,
			&nr_childs,
			&idum,
			&idum,
			&utype)!=5) {
                        errout("Can not readin line of units-file.\n");
			myfree(cluster[0]);
			myfree(child_nr[0]);
			fclose(h_unitfile);
			return -1;
		}
		cluster[0][i].nr_childs =  nr_childs;
		cluster[0][i].typ	=  utype;

		if (i!=testi) {
                        errout("Illegal numbering of units.");
                        errout("Occured in line %d: %d\n",i,testi);
			myfree(cluster[0]);
			myfree(child_nr[0]);
			fclose(h_unitfile);
			return -1;
		}

		cluster[0][i].child_pos=child_pos;

		/**************************************************************************/
		/*** scanin the child-nr`s and make boundry-box 			***/
                /*** on this level the child-nr's are the atom-numbers, which belong    ***/
		/*** to this unit							***/
		/**************************************************************************/
		cur_child		= read_int(h_unitfile);
		child_nr[0][child_pos++]= cur_child;
		sum_charge	       += atom_data[cur_child].charge;

		cluster[0][i].bbox[0]=cluster[0][i].bbox[1]=atom_data[cur_child].ko[0];
		cluster[0][i].bbox[2]=cluster[0][i].bbox[3]=atom_data[cur_child].ko[1];
		cluster[0][i].bbox[4]=cluster[0][i].bbox[5]=atom_data[cur_child].ko[2];

		for (j=0;j<cluster[0][i].nr_childs-1;j++) {
			cur_child		= read_int(h_unitfile);
			child_nr[0][child_pos++]= cur_child;
			sum_charge	       += atom_data[cur_child].charge;

			if (cluster[0][i].bbox[0]>atom_data[cur_child].ko[0])
					cluster[0][i].bbox[0]=atom_data[cur_child].ko[0];

			if (cluster[0][i].bbox[1]<atom_data[cur_child].ko[0])
					cluster[0][i].bbox[1]=atom_data[cur_child].ko[0];

			if (cluster[0][i].bbox[2]>atom_data[cur_child].ko[1])
					cluster[0][i].bbox[2]=atom_data[cur_child].ko[1];

			if (cluster[0][i].bbox[3]<atom_data[cur_child].ko[1])
					cluster[0][i].bbox[3]=atom_data[cur_child].ko[1];

			if (cluster[0][i].bbox[4]>atom_data[cur_child].ko[2])
					cluster[0][i].bbox[4]=atom_data[cur_child].ko[2];

			if (cluster[0][i].bbox[5]<atom_data[cur_child].ko[2])
					cluster[0][i].bbox[5]=atom_data[cur_child].ko[2];
		}

		cluster[0][i].moment[0]=sum_charge*global_data->f_elstat;

		/********** center of geometry ****************/
		cluster[0][i].cen_geom[0] = 0.0;
		cluster[0][i].cen_geom[1] = 0.0;
		cluster[0][i].cen_geom[2] = 0.0;
		cluster[0][i].cen_geom[3] = 0.0;
		cluster[0][i].status      = 0;

		child_pos=cluster[0][i].child_pos;
		for (j=0;j<cluster[0][i].nr_childs;j++) {
			int childnr=child_nr[0][child_pos++];
			cluster[0][i].cen_geom[0] += atom_data[childnr].ko[0];
			cluster[0][i].cen_geom[1] += atom_data[childnr].ko[1];
			cluster[0][i].cen_geom[2] += atom_data[childnr].ko[2];
		}
		cluster[0][i].cen_geom[0] /= (double)cluster[0][i].nr_childs;
		cluster[0][i].cen_geom[1] /= (double)cluster[0][i].nr_childs;
		cluster[0][i].cen_geom[2] /= (double)cluster[0][i].nr_childs;
	}

	levelinfo[0].nr_of_clusters	=	nr_of_units;
	levelinfo[0].nr_of_childs	=	nr_of_atoms;

	fclose(h_unitfile);
	return 0;
}

static int
get_mininsec(int *nr_atoms,int start1,int stop1,int *mini)
{
	int minv,i;

	minv = 99999999;
	for (i=start1;i<stop1;i++){
		if (nr_atoms[i]<minv) {
			minv = nr_atoms[i];
			*mini = i;
		}
	}
	return minv;
}

static int
get_maxinsec(int *nr_atoms,int start1,int stop1,int *maxi)
{
	int maxv,i;

	maxv = -99;
	for (i=start1;i<stop1;i++){
		if (nr_atoms[i]>maxv) {
			maxv = nr_atoms[i];
			*maxi = i;
		}
	}
	return maxv;
}

static int
get_anyinsec(int *nr_atoms,int start1,int stop1,int *maxi)
{
	int maxv;

	*maxi = start1 + (int)((stop1-start1-1)*ego_rand());
	maxv  = nr_atoms[*maxi];
	return maxv;
}

static int
find_best_exchange(	int *nr_atoms,int cur_nr_cluster,int start1, int stop1,
			int rangideal,int rangmax, int rangmin,
			int sum_atoms,int minv,
			int *maxi)
{
	int new_sum1,new_sum2,i=rangmax,j=rangmin; /*** no warning ***/
	int start2,stop2,ds=stop1-start1;
	int good_new_sum1,good_new_sum2;
	int maxv,abs(int y);

	maxv   = -9;
	*maxi  = 0;
	start2 = 0;
	stop2  = ds;

	good_new_sum1 = sum_atoms;

	for (i=0;i<cur_nr_cluster;i++) {
		if ((i>=start1)&&(i<stop1)) {
			if (i+1==ds) {
				start2 += ds;
				stop2  += ds;
			}
			continue;
		}
		new_sum1 = sum_atoms - minv + nr_atoms[i];

		new_sum2 = 0;
		for (j=start2;j<stop2;j++) new_sum2 += nr_atoms[j];
		good_new_sum2 = new_sum2;

		new_sum2 = new_sum2 - nr_atoms[i] + minv;

		/**** both sections impruve ****/
		if (	(abs(good_new_sum1-rangideal) > abs(new_sum1-rangideal) ) &&
			(abs(good_new_sum2-rangideal) > abs(new_sum2-rangideal) ) ) {

			good_new_sum1 = new_sum1;
			good_new_sum2 = new_sum2;
			maxv  = nr_atoms[i];
			*maxi = i;
		}
		if (i+1==stop2) {
			start2 += ds;
			stop2  += ds;
		}
	}
	if (maxv>=0)	return maxv;
	
	maxv   = -9;
	*maxi  = 0;
	start2 = 0;
	stop2  = ds;

	good_new_sum1 = sum_atoms;

	for (i=0;i<cur_nr_cluster;i++) {
		if ((i>=start1)&&(i<stop1)) {
			if (i+1==ds) {
				start2 += ds;
				stop2  += ds;
			}
			continue;
		}
		new_sum1 = sum_atoms - minv + nr_atoms[i];

		new_sum2 = 0;
		for (j=start2;j<stop2;j++) new_sum2 += nr_atoms[j];
		good_new_sum2 = new_sum2;

		new_sum2 = new_sum2 - nr_atoms[i] + minv;

		/**** both sections impruve ****/
		if (	(new_sum1>rangmin) && (new_sum1<rangmax) &&
			(new_sum2>rangmin) && (new_sum2<rangmax) &&
			abs(new_sum2-new_sum1) < abs(good_new_sum1-good_new_sum2) ) {

			good_new_sum1 = new_sum1;
			good_new_sum2 = new_sum2;
			maxv  = nr_atoms[i];
			*maxi = i;
		}
		if (i+1==stop2) {
			start2 += ds;
			stop2  += ds;
		}
	}
	return maxv;
}


static double
get_aufwand(int lev, int parent,struct_cluster_reduced *cluster[])
{
	double	clcen[3];	/*** center of cluster ***/
	double	clextent[3];
	double	xfac=1.0,yfac=1.0,zfac=1.0;
	double	xlap,ylap,zlap;
	double	totfac = 1.0;

	clcen[0] = (cluster[lev][parent].bbox[0]+cluster[lev][parent].bbox[1])/2.0;
	clcen[1] = (cluster[lev][parent].bbox[2]+cluster[lev][parent].bbox[3])/2.0;
	clcen[2] = (cluster[lev][parent].bbox[4]+cluster[lev][parent].bbox[5])/2.0;

	clextent[0] = 1.0*(cluster[lev][parent].bbox[1]-cluster[lev][parent].bbox[0]);
	clextent[1] = 1.0*(cluster[lev][parent].bbox[3]-cluster[lev][parent].bbox[2]);
	clextent[2] = 1.0*(cluster[lev][parent].bbox[5]-cluster[lev][parent].bbox[4]);

	xlap = simvol_bbox[0] - (clcen[0]-clextent[0]);
	if (xlap>0.0) xfac = 1.0 - xlap/(2.0*clextent[0]);

	xlap = (clcen[0]+clextent[0]) - simvol_bbox[1];
	if (xlap>0.0) xfac = 1.0 - xlap/(2.0*clextent[0]);

	ylap = simvol_bbox[2] - (clcen[1]-clextent[1]);
	if (ylap>0.0) yfac = 1.0 - ylap/(2.0*clextent[1]);

	ylap = (clcen[1]+clextent[1]) - simvol_bbox[3];
	if (ylap>0.0) yfac = 1.0 - ylap/(2.0*clextent[1]);

	zlap = simvol_bbox[4] - (clcen[2]-clextent[2]);
	if (zlap>0.0) zfac = 1.0 - zlap/(2.0*clextent[2]);

	zlap = (clcen[2]+clextent[2]) - simvol_bbox[5];
	if (zlap>0.0) zfac = 1.0 - zlap/(2.0*clextent[2]);

	totfac = xfac*yfac*zfac;
	if ((totfac<0.0) || (totfac>1.00001)){
                errout("Fatal error: totally confusion in get_aufwand %f %f %f\n",
			xfac,yfac,zfac);
		pex(-1);
	}

	if (totfac<0.2) totfac=0.2;
#if DEBUG
txtout("Aufwand %f\n",totfac);
#endif
	return totfac;
}

static int
do_load_balancing(	int lev,int no_nodes,
			struct_treeinfo *treeinfo,
			struct_cluster_reduced *parentdat,
			struct_cluster_reduced *cluster[],
			int *atom_nrs)

{
	int	*nr_atoms;
	int	*temp_buf,maxatoms=1000,node;
	int	start1,stop1,rangmin,rangmax,rangideal,sum_atoms,i,c;
	int	cur_nr_cluster=treeinfo->branches,parent;
	int	mini,maxi,maxv=0,minv=0,kk,fac=10;
	int	branches_per_node=(int)(treeinfo->branches/no_nodes);
	struct_cluster_reduced temp;

	temp_buf  = alloc_ints(maxatoms);
	nr_atoms  = alloc_ints(cur_nr_cluster+10);
	fac = 50;

     for (kk=0;kk<6;kk++) {
	sum_atoms = 0;
	for (parent=0;parent<cur_nr_cluster;parent++) {
		double aufwand;

		aufwand	         = get_aufwand(lev,parent,cluster);
		nr_atoms[parent] = CLU_get_all_atoms(	lev,parent,
							&maxatoms,&temp_buf,
							cluster,atom_nrs);
#if (BALANCE_COMPUTE==1)
		nr_atoms[parent] = (int)(normalized_cur_atoms*aufwand);
#endif
		sum_atoms += nr_atoms[parent];
	}

	rangideal= sum_atoms/no_nodes;
	rangmax  = rangideal+rangideal/fac;
	rangmin  = rangideal-rangideal/fac;
	start1 = 0;
	stop1  = branches_per_node;
	for (node=0;node<no_nodes;node++) {
		sum_atoms = 0;
		for (i=start1;i<stop1;i++) sum_atoms += nr_atoms[i];
#if (DEBUG_LOAD)
if (kk==0)
txtout("Node %2d: before load-balance %5d\n",node,sum_atoms);
#endif
		c = 0;
		while ( ((sum_atoms<rangmin)||(sum_atoms>rangmax))&&
			(c++<branches_per_node)) {
			if (sum_atoms<rangmin) {
				if (kk<3) minv = get_mininsec(nr_atoms,start1,stop1,&mini);
				else	  minv = get_anyinsec(nr_atoms,start1,stop1,&mini);

				maxv = find_best_exchange(nr_atoms,cur_nr_cluster,start1,stop1,
						rangideal,rangmax,rangmin,
						sum_atoms,minv,
						&maxi);
				if (maxv<0) continue;
				sum_atoms = sum_atoms - minv + maxv;
			} else
			if (sum_atoms>rangmax) {
				if (kk<3) maxv = get_maxinsec(nr_atoms,start1,stop1,&maxi);
				else	  maxv = get_anyinsec(nr_atoms,start1,stop1,&maxi);

				minv = find_best_exchange(nr_atoms,cur_nr_cluster,start1,stop1,
						rangideal,rangmax,rangmin,
						sum_atoms,maxv,
						&mini);
				if (minv<0) continue;

				sum_atoms = sum_atoms - maxv + minv;
			}

			nr_atoms[maxi] = minv;
			nr_atoms[mini] = maxv;
			memcpy(&temp,&parentdat[mini],sizeof(struct_cluster_reduced));
			memcpy(&parentdat[mini],&parentdat[maxi],sizeof(struct_cluster_reduced));
			memcpy(&parentdat[maxi],&temp,sizeof(struct_cluster_reduced));
		}
#if (DEBUG_LOAD)
if (kk==5)
txtout("Node %2d: after load-balance  %5d\n",node,sum_atoms);
#endif
		start1 += branches_per_node;
		stop1  += branches_per_node;
	}
    } /**** end for k ****/
	myfree(temp_buf);
	myfree(nr_atoms);
	return cur_nr_cluster;
}

/*** this function has to decide, whether neutral clusters must be clusterd  ***/
/*** seperated from charged/polar clusters. This is not needed if neutral    ***/
/*** clusters are widely spread between charged/polar clusters. This function***/
/*** returns 1 if separate clustering needed ***/
static int
analyze_neutrals(struct_cluster_reduced *cluster[])
{
	double cen_neutral[3],cen_all[3];
	double rgyr_all=0.0,rgyr_neutral=0.0;
	double gyrvoldens_all=0.0,gyrvoldens_neutral=0.0;
	int i;

	if (nr_of_neutrals<6)	return 0;
	cen_neutral[0] = cen_neutral[1] = cen_neutral[2] = 0.0;
	cen_all[0] = cen_all[1] = cen_all[2] = 0.0;

	for (i=0;i<nr_of_units;i++) {
		cen_all[0] += cluster[0][i].cen_geom[0];
		cen_all[1] += cluster[0][i].cen_geom[1];
		cen_all[2] += cluster[0][i].cen_geom[2];
		if (cluster[0][i].typ==NEUTRAL) {
			cen_neutral[0] += cluster[0][i].cen_geom[0];
			cen_neutral[1] += cluster[0][i].cen_geom[1];
			cen_neutral[2] += cluster[0][i].cen_geom[2];
		}
	}
	cen_all[0]/=(double)nr_of_units;
	cen_all[1]/=(double)nr_of_units;
	cen_all[2]/=(double)nr_of_units;

	cen_neutral[0]/=(double)nr_of_neutrals;
	cen_neutral[1]/=(double)nr_of_neutrals;
	cen_neutral[2]/=(double)nr_of_neutrals;

	for (i=0;i<nr_of_units;i++) {
		double d1=cluster[0][i].cen_geom[0] - cen_all[0];
		double d2=cluster[0][i].cen_geom[1] - cen_all[1];
		double d3=cluster[0][i].cen_geom[2] - cen_all[2];
		rgyr_all += d1*d1 + d2*d2 + d3*d3;

		if (cluster[0][i].typ==NEUTRAL) {
			double d1=cluster[0][i].cen_geom[0] - cen_neutral[0];
			double d2=cluster[0][i].cen_geom[1] - cen_neutral[1];
			double d3=cluster[0][i].cen_geom[2] - cen_neutral[2];
			rgyr_neutral += d1*d1 + d2*d2 + d3*d3;
		}
	}
	rgyr_all=sqrt(rgyr_all/(double)nr_of_units);
	rgyr_neutral=sqrt(rgyr_neutral/(double)nr_of_neutrals);

	gyrvoldens_all = (double)nr_of_units/(rgyr_all*rgyr_all*rgyr_all);
	gyrvoldens_neutral= (double)nr_of_neutrals/(rgyr_neutral*rgyr_neutral*rgyr_neutral);
#if (DEBUG_INFO==2)
        txtout("GYR-DENS-ALL %f    NEUTRALS %f   frac %f\n",
		gyrvoldens_all,gyrvoldens_neutral,gyrvoldens_neutral/gyrvoldens_all);
#endif
	if (gyrvoldens_neutral/gyrvoldens_all<0.25)	return 0;

	/*** now we have to do separate clustering ***/
	return 1;
}




