/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/**************************************************************************/
/***	Module: egintegr.c	by M. Eichinger, Munich			***/
/***	Contents:							***/
/***		addup_all_forces(...)	adds up all forces and energies ***/
/***					to my_new_forces which are used	***/
/***					for the integration step.	***/
/***									***/
/***		do_integrate(...)	performs the integration step   ***/
/***					together with SHAKE and 	***/
/***                                    minimisation.			***/
/***									***/
/**************************************************************************/

#define DEBUG		0
#define DEBUGQM 	0
#define DEBUGSCHWING	0
#define DEBUGMAXGRAD 	0

#include <math.h>
#include <stdio.h>
#include <fcntl.h>
#if defined(LINUX)||defined(LINUX64)
#include <sys/types.h>
#include <unistd.h>
#endif
#if defined(NEXT)
#include <sys/file.h>
#include <libc.h>
#endif

#include "ego.h"
#include "egintegr.h"
#include "egalloc.h"
#include "util_lib.h"

static int
calc_hessian(	LOAD_MSG_ONE_NODE *my_data,
		LOAD_MSG_ALL_NODE *global_data,
		TRAVELLING_FORCES *my_new_forces);
	
static int
calc_qmmmanglemom(	LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data,
			double *center,
			double *qmang,double *mmang);


static int
correct_mmforces(	LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data,
			TRAVELLING_FORCES *my_new_forces);

static int
set_qmrotmats(double *veca,double *vecb);

static	double **ttens;
static	int *indx;
static	double *bvec;
static	int	ttensallocated=0;

static int
perform_rot(double *x,double *res);


int do_newrottranscor_part1(int my_number,
			ENERGY_INFO *local_energy,	
			LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data,
			TRAVELLING_PACKETS *packet)
{
	int node,ii,i=my_number;
	double totm=0.0;
	
	/*** loop over all packets from the nodes 	***/
	/*** for calculating center of mass		***/
	local_energy->C[0] = local_energy->C[1] = local_energy->C[2] = 
	  local_energy->imobC[0] = local_energy->imobC[3] = local_energy->imobC[6] = 0.0;

	for (node=0;node<global_data->no_nodes;node++) {
		for (i=0;i<packet[node].coordinates.no_atoms;i++) {
			int	ty    = packet[node].coordinates.typ[i];
			double	m     = global_data->AB[ty][ty];
			/** double	r[3],r2; --unused-- **/
			
			totm+=m;
		
			local_energy->C[0] += packet[node].coordinates.x[i]*m;
			local_energy->C[1] += packet[node].coordinates.y[i]*m;
			local_energy->C[2] += packet[node].coordinates.z[i]*m;
		}
	}
		
	local_energy->C[0]/=totm;
	local_energy->C[1]/=totm;
	local_energy->C[2]/=totm;
#if DEBUG
	txtout("Center of mass: %f %f %f\n",
	local_energy->C[0],local_energy->C[1],local_energy->C[2]);
#endif

	/*** loop over all packets from the nodes	***/
	/*** to calculate the traegheitstensor ttens	***/
	if (ttensallocated==0) {	/*** alloc space for it ***/
		ttens  = dmatrix(1,3,1,3);
		bvec   = dvector(1,3);
		indx   = ivector(1,3);
		ttensallocated=1;
	}
	
	ttens[1][1]=0.0;  ttens[1][2]=0.0;  ttens[1][3]=0.0;
	ttens[2][1]=0.0;  ttens[2][2]=0.0;  ttens[2][3]=0.0;
	ttens[3][1]=0.0;  ttens[3][2]=0.0;  ttens[3][3]=0.0;	
	
	for (node=0;node<global_data->no_nodes;node++) {
		for (i=0;i<packet[node].coordinates.no_atoms;i++) {
			int	ty    = packet[node].coordinates.typ[i];
			double	m     = global_data->AB[ty][ty];
			double	r[3],r2;

			r[0] = packet[node].coordinates.x[i] - local_energy->C[0];
			r[1] = packet[node].coordinates.y[i] - local_energy->C[1];
			r[2] = packet[node].coordinates.z[i] - local_energy->C[2];
			r2=r[0]*r[0] + r[1]*r[1] + r[2]*r[2];
			ttens[1][1]+=m*(r2-r[0]*r[0]);
			ttens[2][1]+=m*(  -r[0]*r[1]); ttens[2][2]+=m*(r2-r[1]*r[1]);
			ttens[3][1]+=m*(  -r[0]*r[2]); ttens[3][2]+=m*(  -r[1]*r[2]); 
			ttens[3][3]+=m*(r2-r[2]*r[2]);
		}
	}

	ttens[1][2]=ttens[2][1];
	ttens[1][3]=ttens[3][1];	
	ttens[2][3]=ttens[3][2];
#if DEBUG	
	txtout("TTensor: %f %f %f\n",ttens[1][1],ttens[1][2],ttens[1][3]);
	txtout("TTensor: %f %f %f\n",ttens[2][1],ttens[2][2],ttens[2][3]);
	txtout("TTensor: %f %f %f\n",ttens[3][1],ttens[3][2],ttens[3][3]);
#endif	

	/******************************************************************/
	/*** Here we calculate some values which we need later         	***/
	/*** on the master in order to calculate some stuff		***/
	/******************************************************************/
	local_energy->kinetic = 0.0;
	local_energy->kinetic_h2o = local_energy->kinetic_protein = 0.0;
	local_energy->P[0] = local_energy->P[1] = local_energy->P[2] = 
	  local_energy->imobC[1] = local_energy->imobC[4] = local_energy->imobC[7] = 0.0;
	local_energy->L[0] = local_energy->L[1] = local_energy->L[2] = 
	  local_energy->imobC[2] = local_energy->imobC[5] = local_energy->imobC[8] = 0.0;
	local_energy->qmftot[0] = local_energy->qmftot[1] = local_energy->qmftot[2] = 0.0;
	local_energy->mmftot[0] = local_energy->mmftot[1] = local_energy->mmftot[2] = 0.0;
	local_energy->qmang[0]  = local_energy->qmang[1]  = local_energy->qmang[2]  = 0.0;
	local_energy->mmang[0]  = local_energy->mmang[1]  = local_energy->mmang[2]  = 0.0;

	for (i=0,ii=0;i<my_data->no_atoms;i++,ii+=4) {
		int	i1=ii+1,i2=ii+2,i3=ii+3;
		int	ty  = my_data->typ[i];
		double	m   = global_data->AB[ty][ty];
		double  vx,vy,vz,r[3],ang[3];

		if (global_data->restarting) {
			vx  = (my_data->x_new[i] - my_data->x_old[i])/global_data->time_step;
			vy  = (my_data->y_new[i] - my_data->y_old[i])/global_data->time_step;
			vz  = (my_data->z_new[i] - my_data->z_old[i])/global_data->time_step;
		} else {
			vx  = my_data->x_old[i];
			vy  = my_data->y_old[i];
			vz  = my_data->z_old[i];
		}

		/*** kinetic energy ***/
		local_energy->kinetic += 0.5*(vx*vx + vy*vy + vz*vz)*m;


		if (strcmp(my_data->res_string[i],"TIP3")==0) {
		  local_energy->kinetic_h2o     += 0.5*(vx*vx + vy*vy + vz*vz)*m;
		} else {
		  local_energy->kinetic_protein += 0.5*(vx*vx + vy*vy + vz*vz)*m;
		}

		/*** Momentum of all atoms ***/
		local_energy->P[0] += m*vx;
		local_energy->P[1] += m*vy;
		local_energy->P[2] += m*vz;

		/*** Angular momentum of all atoms ***/
		local_energy->L[0] += m*( my_data->y_new[i]*vz - my_data->z_new[i]*vy);
		local_energy->L[1] += m*(-my_data->x_new[i]*vz + my_data->z_new[i]*vx);
		local_energy->L[2] += m*( my_data->x_new[i]*vy - my_data->y_new[i]*vx);

		/*** Center of Mass of all immobilised atoms ***/
		if (my_data->atom_sele_flag2[i]&FLAG_IMOBIL1) {
		  local_energy->imobC[0] += m*my_data->x_new[i];
		  local_energy->imobC[1] += m*my_data->y_new[i];
		  local_energy->imobC[2] += m*my_data->z_new[i];
		}
		if (my_data->atom_sele_flag2[i]&FLAG_IMOBIL2) {
		  local_energy->imobC[3] += m*my_data->x_new[i];
		  local_energy->imobC[4] += m*my_data->y_new[i];
		  local_energy->imobC[5] += m*my_data->z_new[i];
		}
		if (my_data->atom_sele_flag2[i]&FLAG_IMOBIL3) {
		  local_energy->imobC[6] += m*my_data->x_new[i];
		  local_energy->imobC[7] += m*my_data->y_new[i];
		  local_energy->imobC[8] += m*my_data->z_new[i];
		}

		/*** calculate total force action on node atoms ***/
		if (	(my_data->atom_sele_flag1[i]&FLAG_QMATOM)&&
		      (!(my_data->atom_sele_flag1[i]&FLAG_QMLINK))
		    ) {
			local_energy->qmftot[0]+=my_data->force_total[i1];/**-my_data->force_vdw[i1];**/
			local_energy->qmftot[1]+=my_data->force_total[i2];/**-my_data->force_vdw[i2];**/
			local_energy->qmftot[2]+=my_data->force_total[i3];/**-my_data->force_vdw[i3];**/
		} else {
			local_energy->mmftot[0]+=my_data->force_total[i1];/**-my_data->force_vdw[i1];**/
			local_energy->mmftot[1]+=my_data->force_total[i2];/**-my_data->force_vdw[i2];**/
			local_energy->mmftot[2]+=my_data->force_total[i3];/**-my_data->force_vdw[i3];**/
		}

		/*** calculate drehmoment on node atoms ***/
		r[0] = my_data->x_new[i] - local_energy->C[0];
		r[1] = my_data->y_new[i] - local_energy->C[1];
		r[2] = my_data->z_new[i] - local_energy->C[2];

		crossprod(r,my_data->force_total+i1,ang);
		if (	(my_data->atom_sele_flag1[i]&FLAG_QMATOM)&&
		      (!(my_data->atom_sele_flag1[i]&FLAG_QMLINK))
		    ) {
			local_energy->qmang[0]+=ang[0];
			local_energy->qmang[1]+=ang[1];
			local_energy->qmang[2]+=ang[2];
		} else {
			local_energy->mmang[0]+=ang[0];
			local_energy->mmang[1]+=ang[1];
			local_energy->mmang[2]+=ang[2];
		}
	}
	return 0;
}

int
do_newrottranscor_part2(int my_number,
			ENERGY_INFO *local_energy,	
			LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data,
			TRAVELLING_PACKETS *packet,
			TRAVELLING_FORCES *my_new_forces)
{
	int i,ii,node;
	double ftot[3],atot[3],ang[3];
	double lenf,lena;
	double d,axis[3],drehm,traeg,maxforce=0;
		
	if (global_data->newtransrotcor==0)	return 0;	
	if (global_data->schwing_mode!=0)	return 0;
	
	ftot[0]=local_energy->qmftot[0]+local_energy->mmftot[0];
	ftot[1]=local_energy->qmftot[1]+local_energy->mmftot[1];
	ftot[2]=local_energy->qmftot[2]+local_energy->mmftot[2];

	atot[0]=local_energy->qmang[0]+local_energy->mmang[0];
	atot[1]=local_energy->qmang[1]+local_energy->mmang[1];
	atot[2]=local_energy->qmang[2]+local_energy->mmang[2];	
	
	lenf = veclen(ftot);
	lena = veclen(atot);

#if DEBUG
	txtout("Total Force before : %f %f %f     %f\n",ftot[0],ftot[1],ftot[2],lenf);
	txtout("Total Drehm before : %f %f %f     %f\n",atot[0],atot[1],atot[2],lena);
#endif

	if (global_data->debug_flags[0]&0x0001)
		txtout("Rot/Trans Correction: F=% f  M=% f\n",lenf,lena);

	if (!(global_data->newtransrotcor&(int)0x01)) goto skip_transcor;

	/*** if full QM calculation or full MM calculation then	***/
	/*** make simple translation correction (f(i) -= F/N)	***/
	/*** and normal rotation correction			***/
	if (	(global_data->nr_qmatoms==global_data->total_no_atoms) ||
		(global_data->nr_qmatoms==0) ) {
		int node;
		double fcor[3];

		fcor[0] = -ftot[0]/global_data->total_no_atoms;
		fcor[1] = -ftot[1]/global_data->total_no_atoms;
		fcor[2] = -ftot[2]/global_data->total_no_atoms;
		
       		for (i=0;i<my_data->no_atoms;i++) {
                	my_new_forces->Fx[i]+=fcor[0];
                	my_new_forces->Fy[i]+=fcor[1];
                	my_new_forces->Fz[i]+=fcor[2];
        	}

        	/*** after correctiong the total force the drehmoment has to be updated ***/
        	for (node=0;node<global_data->no_nodes;node++) {
			for (i=0;i<packet[node].coordinates.no_atoms;i++) {
				/** int	ty    = packet[node].coordinates.typ[i]; --unused-- **/ 
				/** double	m     = global_data->AB[ty][ty]; --unused-- **/
				double	r[3]; /** r2 --unused--;**/
	
				r[0] = packet[node].coordinates.x[i] - local_energy->C[0];
				r[1] = packet[node].coordinates.y[i] - local_energy->C[1];
				r[2] = packet[node].coordinates.z[i] - local_energy->C[2];
				crossprod(r,fcor,ang);
				if (	(my_data->atom_sele_flag1[i]&FLAG_QMATOM)&&
		      			(!(my_data->atom_sele_flag1[i]&FLAG_QMLINK))
		    		) {
					local_energy->qmang[0]+=ang[0];
					local_energy->qmang[1]+=ang[1];
					local_energy->qmang[2]+=ang[2];
				} else {
					local_energy->mmang[0]+=ang[0];
					local_energy->mmang[1]+=ang[1];
					local_energy->mmang[2]+=ang[2];
				}
			}
		}
 		
		atot[0]=local_energy->qmang[0]+local_energy->mmang[0];
		atot[1]=local_energy->qmang[1]+local_energy->mmang[1];
		atot[2]=local_energy->qmang[2]+local_energy->mmang[2];			
		lena = veclen(atot);
	} 
	/*** in hybrid calculations make a special translation correction ***/
	else {
		local_energy->mmftot[0]=-local_energy->mmftot[0];
		local_energy->mmftot[1]=-local_energy->mmftot[1];
		local_energy->mmftot[2]=-local_energy->mmftot[2];

		if (set_qmrotmats(local_energy->qmftot,local_energy->mmftot)!=-1)
			correct_mmforces(my_data,global_data,my_new_forces);

        	/*** after correcting the total force the drehmoment	***/
		/*** has to be updated					***/		
		calc_qmmmanglemom(	my_data,
					global_data,
					local_energy->C,
					local_energy->qmang,
					local_energy->mmang);
				
		atot[0]=local_energy->qmang[0]+local_energy->mmang[0];
		atot[1]=local_energy->qmang[1]+local_energy->mmang[1];
		atot[2]=local_energy->qmang[2]+local_energy->mmang[2];			
		lena = veclen(atot);
	}
	
skip_transcor:
	if (!(global_data->newtransrotcor&(int)0x02)) goto skip_rotcor;
				
#if DEBUG
	txtout("Atot now %f %f %f\n",atot[0],atot[1],atot[2]);
#endif
	/*** calculate rotation axis by inverting the TreagTens ***/
	bvec[1] = atot[0];
	bvec[2] = atot[1];
	bvec[3] = atot[2];
	ludcmp(ttens,3,indx,&d);
	lubksb(ttens,3,indx,bvec);
	if (normvec(bvec+1,axis)<0.0) return 0;
#if DEBUG
	txtout("Drehaxi: %f %f %f\n",axis[0],axis[1],axis[2]);
#endif		

	drehm=dotprod(axis,atot);	/*** drehmoment along axis ***/	
					/*** treagheitsmoment um achse ***/	
	traeg=0.0;
	for (node=0;node<global_data->no_nodes;node++) {
		for (i=0;i<packet[node].coordinates.no_atoms;i++) {
			int	ty    = packet[node].coordinates.typ[i];
			double	m     = global_data->AB[ty][ty];
			double	r[3],l,k;

			r[0] = packet[node].coordinates.x[i] - local_energy->C[0];
			r[1] = packet[node].coordinates.y[i] - local_energy->C[1];
			r[2] = packet[node].coordinates.z[i] - local_energy->C[2];
			
			l=veclen(r);
			k=dotprod(axis,r);
			traeg+=m*(l*l-k*k);
		}
	}
		
	maxforce=0.0;	
	for (i=0,ii=0;i<my_data->no_atoms;i++,ii+=4) {
	  /** int i1=ii+1,i2=ii+2,i3=ii+3;  -- unused --**/
		int	ty    = my_data->typ[i];
		double	m     = global_data->AB[ty][ty];
		double  vec[3],r[3],fac;
	
		r[0]=(my_data->x_new[i] - local_energy->C[0]);
		r[1]=(my_data->y_new[i] - local_energy->C[1]);
		r[2]=(my_data->z_new[i] - local_energy->C[2]);

		crossprod(axis,r,vec);
		fac=drehm*m/traeg;
		
		my_new_forces->Fx[i] -= vec[0]*fac;
		my_new_forces->Fy[i] -= vec[1]*fac;
		my_new_forces->Fz[i] -= vec[2]*fac;
		
		m=veclen(vec)*fac;
		if (m>maxforce) maxforce=m;		
	}
	
	if (global_data->debug_flags[0]&0x0001)
		txtout("Maximum force by angle correction=%f\n",maxforce);
			
skip_rotcor:
	return 0;
}

static int
calc_qmmmanglemom(	LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data,
			double *center,
			double *qmang,
			double *mmang)
{
	double ang[3],r[3];
	int i,ii;

	qmang[0] = qmang[1] = qmang[2] = 0.0;
	mmang[0] = mmang[1] = mmang[2] = 0.0;

	for (i=0,ii=0;i<my_data->no_atoms;i++,ii+=4) {
		int i1=ii+1;

		r[0] = my_data->x_new[i] - center[0];
		r[1] = my_data->y_new[i] - center[1];
		r[2] = my_data->z_new[i] - center[2];

		crossprod(r,my_data->force_total+i1,ang);
		if (	(my_data->atom_sele_flag1[i]&FLAG_QMATOM)&&
		      (!(my_data->atom_sele_flag1[i]&FLAG_QMLINK))
		    ) {
			qmang[0]+=ang[0];
			qmang[1]+=ang[1];
			qmang[2]+=ang[2];
		} else {
			mmang[0]+=ang[0];
			mmang[1]+=ang[1];
			mmang[2]+=ang[2];
		}
	}
	return 0;
}

static double e1[3],e2[3],e3[3],scalfac=1.0;
static double zrot[3][3];
static	int transok=0;

static int
set_qmrotmats(double *veca,double *vecb)
{
	double t,lena,lenb,veclen(double *a),dotprod(double *a,double *b);
	double sz,cz,axb[3],ta[3],tb[3];
	transok=0;
	
	lena=veclen(veca);
	lenb=veclen(vecb);
	if ((lena<0.000001)||(lenb<0.000001)) return -1;
	e1[0]=veca[0]/lena;e1[1]=veca[1]/lena;e1[2]=veca[2]/lena;

	scalfac=lena/lenb;

	crossprod(e1,vecb,e3);
	t=veclen(e3);
	if (t<0.000001) return -1;
	e3[0]/=t;e3[1]/=t;e3[2]/=t;
	
	crossprod(e3,e1,e2);
	t=veclen(e2);
	if (t<0.000001) return -1;
	e2[0]/=t;e2[1]/=t;e2[2]/=t;
	transok=1;
	
	ta[0]=dotprod(veca,e1);
	ta[1]=dotprod(veca,e2);
	ta[2]=dotprod(veca,e3);

	tb[0]=dotprod(vecb,e1);
	tb[1]=dotprod(vecb,e2);
	tb[2]=dotprod(vecb,e3);

	crossprod(ta,tb,axb);
	sz = veclen(axb)/(veclen(ta)*veclen(tb));
	cz = sqrt(1.0-sz*sz);

	if (dotprod(ta,tb)<0.0)	{cz = -cz;}
/****
printf("e1: %f %f %f\n",e1[0],e1[1],e1[2]);
printf("e2: %f %f %f\n",e2[0],e2[1],e2[2]);
printf("e3: %f %f %f\n",e3[0],e3[1],e3[2]);
printf("cz %f  sz %f fac %f\n",cz,sz,scalfac);
****/
	zrot[0][0]= cz;  zrot[0][1]= sz;   zrot[0][2]=0.0;
	zrot[1][0]=-sz;  zrot[1][1]= cz;   zrot[1][2]=0.0;
	zrot[2][0]=0.0;  zrot[2][1]=0.0;   zrot[2][2]=1.0;
		
	return 0;
}

static int
perform_rot(double *x,double *res)
{
	double t[3],tt[3],veclen(double *a),dotprod(double *a,double *b);

	if (!transok) {
		res[0] = x[0];
		res[1] = x[1];
		res[2] = x[2];	
		return 0;
	}
	t[0]=dotprod(x,e1);
	t[1]=dotprod(x,e2);
	t[2]=dotprod(x,e3);		

	tt[0] = zrot[0][0]*t[0] + zrot[0][1]*t[1]                ;
	tt[1] = zrot[1][0]*t[0] + zrot[1][1]*t[1]                ;
	tt[2] = 			          zrot[2][2]*t[2];

	res[0] = (tt[0]*e1[0] + tt[1]*e2[0] + tt[2]*e3[0])*scalfac;
	res[1] = (tt[0]*e1[1] + tt[1]*e2[1] + tt[2]*e3[1])*scalfac;
	res[2] = (tt[0]*e1[2] + tt[1]*e2[2] + tt[2]*e3[2])*scalfac;
	
	return 0;
}

static int
correct_mmforces(	LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data,
			TRAVELLING_FORCES *my_new_forces)
{
	int i,ii;
	for (i=0,ii=0;i<my_data->no_atoms;i++,ii+=4) {
		int i1=ii+1,i2=ii+2,i3=ii+3;

		if (	(my_data->atom_sele_flag1[i]&FLAG_QMATOM)&&
		      (!(my_data->atom_sele_flag1[i]&FLAG_QMLINK))
		    ) continue;
				
		/*** subtract old incorrect forces ***/
		my_data->force_total[i1] -= my_data->force_elecvdw[i1];
		my_data->force_total[i2] -= my_data->force_elecvdw[i2];
		my_data->force_total[i3] -= my_data->force_elecvdw[i3];

		my_new_forces->Fx[i] -= my_data->force_elecvdw[i1];
		my_new_forces->Fy[i] -= my_data->force_elecvdw[i2];
		my_new_forces->Fz[i] -= my_data->force_elecvdw[i3];
		
		/*** correct elec forces ***/
		perform_rot(	my_data->force_elecvdw+i1,
				my_data->force_elecvdw+i1);	
		
		/*** add correct forces ***/			
		my_data->force_total[i1] += my_data->force_elecvdw[i1];
		my_data->force_total[i2] += my_data->force_elecvdw[i2];
		my_data->force_total[i3] += my_data->force_elecvdw[i3];

		my_new_forces->Fx[i] += my_data->force_elecvdw[i1];
		my_new_forces->Fy[i] += my_data->force_elecvdw[i2];
		my_new_forces->Fz[i] += my_data->force_elecvdw[i3];
	}

	return 0;
}

void
addup_all_forces(ENERGY_INFO *energy,
		 LOAD_MSG_ONE_NODE *my_data,
		 LOAD_MSG_ALL_NODE *global_data,
		 TRAVELLING_FORCES *my_new_forces)
{
	int i,ii;

	/******************************************************************/
	/*** first we determine VDW-forces, sum up all internal forces  ***/
	/*** (BOND, ANGLE, DIEHDRAL, IMPROPER) and sum up long range    ***/
	/*** forces on each atoms 					***/
	/******************************************************************/
	for (i=0,ii=0;i<my_data->no_atoms;i++,ii+=4) {
		int i1=ii+1,i2=ii+2,i3=ii+3;

		my_data->force_vdw[ii] = my_data->force_elecvdw[ii]-
					 my_data->force_elec[ii];
					
		my_data->force_vdw[i1] = my_data->force_elecvdw[i1]-
					 my_data->force_elec[i1];
					
		my_data->force_vdw[i2] = my_data->force_elecvdw[i2]-
					 my_data->force_elec[i2];
					
		my_data->force_vdw[i3] = my_data->force_elecvdw[i3]-
					 my_data->force_elec[i3];
/***
txtout("VDW %d %f    %f %f %f\n",
i,my_data->force_vdw[ii],my_data->force_vdw[i1],my_data->force_vdw[i2],my_data->force_vdw[i3]);
txtout("ELE %d %f    %f %f %f\n",
i,my_data->force_elec[ii],my_data->force_elec[i1],my_data->force_elec[i2],my_data->force_elec[i3]);
***/					
		my_data->force_intern[ii] = my_data->force_bond [ii] +
					    my_data->force_angle[ii] +
					    my_data->force_dihe [ii] +
					    my_data->force_impr [ii];

		my_data->force_intern[i1] = my_data->force_bond [i1] +
					    my_data->force_angle[i1] +
					    my_data->force_dihe [i1] +
					    my_data->force_impr [i1];

		my_data->force_intern[i2] = my_data->force_bond [i2] +
					    my_data->force_angle[i2] +
					    my_data->force_dihe [i2] +
					    my_data->force_impr [i2];

		my_data->force_intern[i3] = my_data->force_bond [i3] +
					    my_data->force_angle[i3] +
					    my_data->force_dihe [i3] +
					    my_data->force_impr [i3];
					
					
		my_data->force_total[ii]  = my_data->force_intern[ii]+ 
					    my_data->force_elec[ii]  +
					    my_data->force_vdw[ii];
					    
		my_data->force_total[i1]  = my_data->force_intern[i1]+ 
					    my_data->force_elec[i1]  +
					    my_data->force_vdw[i1];

		my_data->force_total[i2]  = my_data->force_intern[i2]+ 
					    my_data->force_elec[i2]  +
			    		    my_data->force_vdw[i2];
			    		    
		my_data->force_total[i3]  = my_data->force_intern[i3]+ 
					    my_data->force_elec[i3]  + 
					    my_data->force_vdw[i3];
	}

	/**********************************************************/
	/*** In case of QM simulations we overwrite here the 	***/
	/*** classical forces from EGO with the QM forces. Note	***/
	/*** that the QM-forces have the right sign, so that they**/
	/*** must be subtracted as the internal EGO-forces have ***/
	/*** the opposite sign.					***/
	/**********************************************************/
	if (global_data->nr_qmatoms>0) {
		if (global_data->qmatomforce==QMATOMFORCE_QM) {
			for (i=0;i<my_data->qmlocal.cur_len;i++) {
				int j=my_data->qmlocal.atomi[i];
				int ii,i1,i2,i3;
				ii=4*j; i1=ii+1; i2=ii+2; i3=ii+3;

				my_data->force_total[ii] -= my_data->force_elec[ii];

				my_data->force_total[i1]  = 
						my_data->force_vdw[i1] -
						my_data->qmlocal.qm_force[3*i+0];

				my_data->force_total[i2]  = 
						my_data->force_vdw[i2] -
						my_data->qmlocal.qm_force[3*i+1];
			    		    
				my_data->force_total[i3]  = 
						my_data->force_vdw[i3] -
						my_data->qmlocal.qm_force[3*i+2];

				/*** now we update the charges of the atoms in the ***/
				/*** QM fragment. The charge of the link atom was  ***/
				/*** set to -999.9 and has been transfered to the  ***/
				/*** atom to which it is bonded			   ***/				
				if (global_data->qmchargemode==QMCHARGEMODE_VARIABLE) {	
					if (my_data->qmlocal.qm_q[i]>-100.0)
						my_data->q[j] = my_data->qmlocal.qm_q[i];
				}
			}			

		} else
		if (global_data->qmatomforce==QMATOMFORCE_CLASSIC) {
			
			/*** nothing to do in this case		***/
			/*** the classical forces are used	***/
			
		}
	}

	/**********************************************************/
	/*** Add up all force contributions (also the external  ***/
	/*** SBOUND forces) to the buffer my_new_forces which is***/
	/*** used for the integration.				***/
	/**********************************************************/
	for (i=0,ii=0;i<my_data->no_atoms;i++,ii+=4) {
		int i1=ii+1,i2=ii+2,i3=ii+3;	
			
		my_new_forces->Fx[i] += my_data->force_total[i1] + 
					my_data->force_restr[i1] +
					my_data->force_sbound[i1]+
					my_data->force_flood[i1];

		my_new_forces->Fy[i] += my_data->force_total[i2]+ 
					my_data->force_restr[i2]+
					my_data->force_sbound[i2]+
					my_data->force_flood[i2];

		my_new_forces->Fz[i] += my_data->force_total[i3]+ 
					my_data->force_restr[i3]+
					my_data->force_sbound[i3]+
					my_data->force_flood[i3];

		/*** force_elec [ii] contains 1/2 of the elec energy of a charge ***/
		/*** in the field of all other charges, so that the sum is the   ***/
		/*** total electrostatic energy. For a QM/MM calculation the elec***/
		/*** energy of the QM fragment in the presence of the MM elec    ***/
		/*** field is added in eganaly.c. In order to aviod a double     ***/
		/*** counting of the elec energy of the QM fragment we have to   ***/
		/*** subtract for a QM atom my_data->force_elec [ii]		 ***/
		if (my_data->atom_sele_flag1[i]&FLAG_QMATOM) {
			energy->elec         -= my_data->force_elec [ii];
		} else {
			energy->elec         += my_data->force_elec [ii];
		}

		energy->vdw          += my_data->force_vdw  [ii];
/***
txtout("ENE %d  %d  %f   %f\n",i,(my_data->atom_sele_flag1[i]&FLAG_QMATOM),energy->elec,energy->vdw);
***/
		energy->bond         += my_data->force_bond [ii];
		energy->angle        += my_data->force_angle[ii];
		energy->dihe         += my_data->force_dihe [ii];
		energy->impr         += my_data->force_impr [ii];
		energy->restraint    += my_data->force_restr[ii];
		energy->sbound       += my_data->force_sbound[ii];

		/*************************************************
		Note:	flooding-energies currently are 
			added directly to my_new_forces in the
			module-file egrotflood.c!
		**************************************************/
	
		/*** this affects only the forces, not the energies ***/
		if (my_data->atom_sele_flag1[i]&FLAG_FORCEOFF) {
			my_new_forces->Fx[i]=0.0;
			my_new_forces->Fy[i]=0.0;
			my_new_forces->Fz[i]=0.0;
		}
		
		if (my_data->atom_sele_flag1[i]&FLAG_SPEEDOFF) {		
			if (global_data->restarting) {
				my_data->x_old[i] = my_data->x_new[i];
				my_data->y_old[i] = my_data->y_new[i];
				my_data->z_old[i] = my_data->z_new[i];
			} else {
				my_data->x_old[i] = 0.0;
				my_data->y_old[i] = 0.0;
				my_data->z_old[i] = 0.0;				
			}
		}
	}

	apply_specforce(energy,my_data,global_data,my_new_forces);
	
	/**********************************************************/
	/*** In case of QMATOMFORCE_OFF we set			***/
	/*** forces on QMATOMS to zero.				***/
	/**********************************************************/
 	if (global_data->qmatomforce==QMATOMFORCE_OFF) {
		for (i=0;i<my_data->qmlocal.cur_len;i++) {
			int j=my_data->qmlocal.atomi[i];
			my_new_forces->Fx[j]=0.0;
			my_new_forces->Fy[j]=0.0;
			my_new_forces->Fz[j]=0.0;
		}
	}

	/**********************************************************/
	/*** this is a very special case for the calculation for **/
	/*** of the hesse matrix: Only vdw and  		***/
	/*** elec forces will enter into the calculation of the ***/
	/*** hesse matrix. Never tested!			***/	
	/**********************************************************/
	if (global_data->schwing_mode<0) {
		for (i=0,ii=0;i<my_data->no_atoms;i++,ii+=4) {
			int i1=ii+1,i2=ii+2,i3=ii+3;		
			my_new_forces->Fx[i] = my_data->force_elecvdw[i1];
			my_new_forces->Fy[i] = my_data->force_elecvdw[i2];
			my_new_forces->Fz[i] = my_data->force_elecvdw[i3];
		}
	}
	return;
}

static	int	extdyn_between=0;	

static	int	atomsschwing=-1;
static	int	schwing_counter_atom=-1,firsttime=1,filled=0;
static	int	schwing_counter_axis=0;
static	char	mybuf[MAX_STRING_SIZE];
static	int	first_enter=1;
static	int	allocatedsf=0;
static	double	*sfx,*sfy,*sfz;
static	double	inidip[3];

void   do_integrate(	ENERGY_INFO *energy,
			LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data,
			TRAVELLING_FORCES *my_new_forces,
			double *Gamma,int my_number)
{
	int i,maxind=-1;
	double square_time_step=global_data->time_step*global_data->time_step;
	double a=energy->elec;a++; /*** avoid warnings ***/

        /*** calculate maximum and average gradient     ***/
        /*** of forces on this node                     ***/
	energy->gradinfo[0]=0.0;
	energy->gradinfo[1]=0.0;
        for (i=0;i<my_data->no_atoms;i++) {
                double curf;
                curf = sqrt(    my_new_forces->Fx[i]*my_new_forces->Fx[i] +
                                my_new_forces->Fy[i]*my_new_forces->Fy[i] +
                                my_new_forces->Fz[i]*my_new_forces->Fz[i] );
                if (curf>energy->gradinfo[0]) {
                        energy->gradinfo[0] = curf;
			maxind=i;		
                }
                energy->gradinfo[1]  += curf;
	}
	if (global_data->debug_flags[0]&0x0001)
		txtout("Atom with MaxGrad is %d  (PDB %d)\n",
			maxind,my_data->atom_id[maxind]+1);

	if ((global_data->minimisation==1)||(global_data->minimisation==4)) {
		if (first_enter) { /*** initialization ***/
			sfx = alloc_doubles(my_data->no_atoms);
			sfy = alloc_doubles(my_data->no_atoms);
			sfz = alloc_doubles(my_data->no_atoms);			
			first_enter=0;	
			memcpy(sfx,my_new_forces->Fx,sizeof(double)*my_data->no_atoms);
			memcpy(sfy,my_new_forces->Fy,sizeof(double)*my_data->no_atoms);
			memcpy(sfz,my_new_forces->Fz,sizeof(double)*my_data->no_atoms);
			allocatedsf=my_data->no_atoms;
		} else
		if (allocatedsf!=my_data->no_atoms) {
			myfree(sfx);
			myfree(sfy);
			myfree(sfz);
			sfx = alloc_doubles(my_data->no_atoms);
			sfy = alloc_doubles(my_data->no_atoms);
			sfz = alloc_doubles(my_data->no_atoms);			
			first_enter=0;	
			memcpy(sfx,my_new_forces->Fx,sizeof(double)*my_data->no_atoms);
			memcpy(sfy,my_new_forces->Fy,sizeof(double)*my_data->no_atoms);
			memcpy(sfz,my_new_forces->Fz,sizeof(double)*my_data->no_atoms);
			allocatedsf=my_data->no_atoms;	
		}
	}

	if (global_data->minimisation==3) { /*** Newton ***/
		int i;
		
		if (first_enter) { /*** initialization ***/
			sfx = alloc_doubles(my_data->no_atoms);
			sfy = alloc_doubles(my_data->no_atoms);
			sfz = alloc_doubles(my_data->no_atoms);			
			first_enter=0;	
			memcpy(sfx,my_new_forces->Fx,sizeof(double)*my_data->no_atoms);
			memcpy(sfy,my_new_forces->Fy,sizeof(double)*my_data->no_atoms);
			memcpy(sfz,my_new_forces->Fz,sizeof(double)*my_data->no_atoms);
			
			memcpy(my_data->x_old,my_data->x_new,sizeof(double)*my_data->no_atoms);
			memcpy(my_data->y_old,my_data->y_new,sizeof(double)*my_data->no_atoms);
			memcpy(my_data->z_old,my_data->z_new,sizeof(double)*my_data->no_atoms);
			for (i=0;i<my_data->no_atoms;i++) {
				my_data->x_new[i]=my_data->x_old[i] - sfx[i]/1200.0;
				my_data->y_new[i]=my_data->y_old[i] - sfy[i]/1200.0;
				my_data->z_new[i]=my_data->z_old[i] - sfz[i]/1200.0;				
			}
			return;
		}
				
		for (i=0;i<my_data->no_atoms;i++) {
			double fdiffx,fdiffy,fdiffz,diff=0,newp,fac=1.0;
			/* int	ty    = my_data->typ[i];
			double	m     = global_data->AB[ty][ty]; Is this still needed?? HH */
			fdiffx = my_new_forces->Fx[i] - sfx[i];
			fdiffy = my_new_forces->Fy[i] - sfy[i];
			fdiffz = my_new_forces->Fz[i] - sfz[i];
			
			if (fabs(fdiffx)>0.0000001) {
				newp=(my_new_forces->Fx[i]*my_data->x_old[i]-sfx[i]*my_data->x_new[i])/fdiffx;
				diff = fac*global_data->fric_coeff*(newp - my_data->x_new[i]);
				if (diff>global_data->max_move)		diff = global_data->max_move;
				else
				if (diff<(-global_data->max_move))	diff = -global_data->max_move;
			}
			my_data->x_old[i]=my_data->x_new[i];	
			my_data->x_new[i]=my_data->x_new[i] + diff;
			
			if (fabs(fdiffy)>0.0000001) {
				newp=(my_new_forces->Fy[i]*my_data->y_old[i]-sfy[i]*my_data->y_new[i])/fdiffy;
				diff = fac*global_data->fric_coeff*(newp - my_data->y_new[i]);
				if (diff>global_data->max_move)		diff = global_data->max_move;
				else
				if (diff<(-global_data->max_move))	diff = -global_data->max_move;
			}
			my_data->y_old[i]=my_data->y_new[i];	
			my_data->y_new[i]=my_data->y_new[i] + diff;
			if (fabs(fdiffz)>0.0000001) {
				newp=(my_new_forces->Fz[i]*my_data->z_old[i]-sfz[i]*my_data->z_new[i])/fdiffz;
				diff = fac*global_data->fric_coeff*(newp - my_data->z_new[i]);
				if (diff>global_data->max_move)		diff = global_data->max_move;
				else
				if (diff<(-global_data->max_move))	diff = -global_data->max_move;
			}
			my_data->z_old[i]=my_data->z_new[i];	
			my_data->z_new[i]=my_data->z_new[i] + diff;
		}
				
		memcpy(sfx,my_new_forces->Fx,sizeof(double)*my_data->no_atoms);
		memcpy(sfy,my_new_forces->Fy,sizeof(double)*my_data->no_atoms);
		memcpy(sfz,my_new_forces->Fz,sizeof(double)*my_data->no_atoms);	
		return; /*** return is important ***/
	}

	if (global_data->minimisation==2) { /*** Gradient ***/
		int i;
		
		if (first_enter) { /*** initialization ***/
			sfx = alloc_doubles(my_data->no_atoms);
			sfy = alloc_doubles(my_data->no_atoms);
			sfz = alloc_doubles(my_data->no_atoms);			
			first_enter=0;	
			memcpy(sfx,my_new_forces->Fx,sizeof(double)*my_data->no_atoms);
			memcpy(sfy,my_new_forces->Fy,sizeof(double)*my_data->no_atoms);
			memcpy(sfz,my_new_forces->Fz,sizeof(double)*my_data->no_atoms);
			
			memcpy(my_data->x_old,my_data->x_new,sizeof(double)*my_data->no_atoms);
			memcpy(my_data->y_old,my_data->y_new,sizeof(double)*my_data->no_atoms);
			memcpy(my_data->z_old,my_data->z_new,sizeof(double)*my_data->no_atoms);
			for (i=0;i<my_data->no_atoms;i++) {
				int	ty    = my_data->typ[i];
				double	m     = global_data->AB[ty][ty],fac;
				if (m>2.0) fac=2600.0;
				else	   fac=1300.0;
				my_data->x_new[i]=my_data->x_old[i] - sfx[i]/fac;
				my_data->y_new[i]=my_data->y_old[i] - sfy[i]/fac;
				my_data->z_new[i]=my_data->z_old[i] - sfz[i]/fac;				
			}
			return;
		}
				
		for (i=0;i<my_data->no_atoms;i++) {
			double fdiffx,fdiffy,fdiffz,diff=0.0;
			double kdiffx,kdiffy,kdiffz;
			int	ty    = my_data->typ[i];
			double	m     = global_data->AB[ty][ty],facx,facy,facz;

			if (m>2.0) facx=facy=facz=1.0/2600.0;
			else	   facx=facy=facz=1.0/1300.0;
				
			fdiffx = my_new_forces->Fx[i] - sfx[i];
			fdiffy = my_new_forces->Fy[i] - sfy[i];
			fdiffz = my_new_forces->Fz[i] - sfz[i];
			kdiffx = my_data->x_new[i] - my_data->x_old[i];
			kdiffy = my_data->y_new[i] - my_data->y_old[i];
			kdiffz = my_data->z_new[i] - my_data->z_old[i];
			if (fabs(fdiffx)>0.0001) facx = min(0.010,fabs(kdiffx/fdiffx));
			if (fabs(fdiffy)>0.0001) facy = min(0.010,fabs(kdiffy/fdiffy));
			if (fabs(fdiffz)>0.0001) facz = min(0.010,fabs(kdiffz/fdiffz));						
			
			diff = -global_data->fric_coeff*my_new_forces->Fx[i]*facx;
			if (diff>global_data->max_move)		diff = global_data->max_move;
			else
			if (diff<(-global_data->max_move))	diff = -global_data->max_move;
			
			my_data->x_old[i]=my_data->x_new[i];	
			my_data->x_new[i]=my_data->x_new[i] + diff;
			
			diff = -global_data->fric_coeff*my_new_forces->Fy[i]*facy;
			if (diff>global_data->max_move)		diff = global_data->max_move;
			else
			if (diff<(-global_data->max_move))	diff = -global_data->max_move;

			my_data->y_old[i]=my_data->y_new[i];	
			my_data->y_new[i]=my_data->y_new[i] + diff;
			
			diff = -global_data->fric_coeff*my_new_forces->Fz[i]*facz;
			if (diff>global_data->max_move)		diff = global_data->max_move;
			else
			if (diff<(-global_data->max_move))	diff = -global_data->max_move;

			my_data->z_old[i]=my_data->z_new[i];	
			my_data->z_new[i]=my_data->z_new[i] + diff;
		}
				
		memcpy(sfx,my_new_forces->Fx,sizeof(double)*my_data->no_atoms);
		memcpy(sfy,my_new_forces->Fy,sizeof(double)*my_data->no_atoms);
		memcpy(sfz,my_new_forces->Fz,sizeof(double)*my_data->no_atoms);	
		return; /*** return is important ***/
	}


	/**********************************************************/
	/*** for normal mode analysis: In that case in each 	***/
	/*** step one atom is selected and by repositioning it	***/
	/*** in x,y,z direction the massweighted forceconstant 	***/
	/*** matrix is calculated				***/
	/**********************************************************/
	if (global_data->schwing_mode!=0) {
		calc_hessian(my_data,global_data,my_new_forces);
		return; /*** return is important ***/
	}

	/******************************************************************/
	/*** the stepping-mode: note that the velocity of all		***/
	/*** "stepping" atoms is set to zero! In addup_all_forces the	***/
	/*** forces for these atoms are set to zero. 			***/
	/******************************************************************/
	if ( (global_data->flag_pull==2) &&
             (global_data->cycles_passed>=global_data->selimob_start) ) {
		double dstep=global_data->pull_vel*global_data->time_step*TIMEFAC;
		int i;

		if (	(global_data->extdyn_step[2]!=-1) &&
			(++extdyn_between<=global_data->extdyn_step[2]) ) {
			txtout("Free dynamic at step %d\n",global_data->cycles_passed);
		} else {
			txtout("Constrained dynamic at step %d\n",global_data->cycles_passed);
			extdyn_between=0;
			for (i=0;i<my_data->no_atoms;i++) {
				if (!(my_data->atom_sele_flag1[i]&FLAG_PULL)) continue;			
				my_data->x_new[i] += global_data->pull_dir[0]*dstep;
				my_data->y_new[i] += global_data->pull_dir[1]*dstep;
				my_data->z_new[i] += global_data->pull_dir[2]*dstep;
	
				if (global_data->restarting) {
					my_data->x_old[i] = my_data->x_new[i];
					my_data->y_old[i] = my_data->y_new[i];
					my_data->z_old[i] = my_data->z_new[i];
				} else {
					my_data->x_old[i] = 0.0;
					my_data->y_old[i] = 0.0;
					my_data->z_old[i] = 0.0;				
				}				
			}
		}
	}

	/******************************************************************/
	/*** that is the integration for normal MD simulations ************/	
	/******************************************************************/	
	if (global_data->restarting) {	/*** if we have old and new coordinate set   ***/
		int    i;		/*** this after restarts and after the first ***/
					/*** integration step has been done !        ***/
		for (i=0; i<my_data->no_atoms; i++) {
			int	ty    = my_data->typ[i];
			double	m     = global_data->AB[ty][ty];
			double	g     = Gamma[i];
			double	alpha = 1.0 / (1.0 + g);
			double	ko_brand_new,c;

			if (m<=0.0) { /*** set velocity to 0 ***/
				my_data->x_old[i] = my_data->x_new[i];
				my_data->y_old[i] = my_data->y_new[i];
				my_data->z_old[i] = my_data->z_new[i];				
				m = 1.0;
			} 

			if ((global_data->minimisation==4)&&(m>1.5)) m=2.0;

			c = square_time_step / m;
		
			INTEGRATE3 (	my_data->x_old[i],my_data->x_new[i],
					my_new_forces->Fx);

			INTEGRATE3 (	my_data->y_old[i],my_data->y_new[i],
					my_new_forces->Fy);

			INTEGRATE3 (	my_data->z_old[i],my_data->z_new[i],
					my_new_forces->Fz);
		}	
	} else {		/*** in the first integration step we have usually ***/
		int    i;	/*** velocities, and not a second coordinate set   ***/
				/*** x_old are in that case velocities !!! 	   ***/
		for (i=0;i<my_data->no_atoms;i++) {
			int	ty = my_data->typ[i];
			double	m  = global_data->AB[ty][ty];
			double	ko_brand_new,c;

			if (m<=0.0) {
				my_data->x_old[i] = 0.0;
				my_data->y_old[i] = 0.0;
				my_data->z_old[i] = 0.0;				
				m = 1.0;
			}
			c = 0.5 * square_time_step / m;

			INTEGRATE4 (	my_data->x_old[i], my_data->x_new[i],
				 	my_new_forces->Fx);

			INTEGRATE4 (	my_data->y_old[i], my_data->y_new[i],
				 	my_new_forces->Fy);

			INTEGRATE4 (	my_data->z_old[i], my_data->z_new[i],
				 	my_new_forces->Fz);
 		} 

		global_data->restarting = TRUE;	/*** from now on we have positions ***/
						/*** and no longer velocities  	   ***/
	}

	/********************************************************************/
 	/******************** SHAKE *****************************************/
	/********************************************************************/
	if (global_data->shake) {
		int    i;
		int    iteration=0;
		int    within_tolerance=FALSE;
		double max_deviation;
		int    max_deviation_number;

		while (	(iteration < MXNO_SHAKE_ITERATIONS) && 
			(!within_tolerance) ) {

			for (i = 0; i < my_data->shake_number/2; i++) {
				int    hd_ptr = i * 2;
				int    hd = my_data->shake_list[hd_ptr  ] - global_data->first_atom[my_number];
				int    az = my_data->shake_list[hd_ptr+1] - global_data->first_atom[my_number];
				int    hd_type = my_data->typ[hd];
				int    az_type = my_data->typ[az];
				double hd_m    = global_data->AB[hd_type][hd_type];
				double az_m    = global_data->AB[az_type][az_type];

				/* this is the OLD bond	 vector */
				double rx = my_data->x_old[az] - my_data->x_old[hd];
				double ry = my_data->y_old[az] - my_data->y_old[hd];
				double rz = my_data->z_old[az] - my_data->z_old[hd];

				/* sprot = <r, rn>, the scalar product 		 */
				double rnx,rny,rnz,R0,Rn,sprot,c,g;

				/* this is a vector from the antezedent to the hydrogen in the
				 * direction of the new bond */
				rnx = my_data->x_new[az] - my_data->x_new[hd];
				rny = my_data->y_new[az] - my_data->y_new[hd];
				rnz = my_data->z_new[az] - my_data->z_new[hd];
				Rn = rnx * rnx + rny * rny + rnz * rnz;

				/* load R0 (length to be) */

				if (hd_type<az_type)
					R0 = global_data->k_bond[az_type][hd_type];
				else if (hd_type == az_type)
					R0 = global_data->k_bond[global_data->nr_of_atom_typs][hd_type];
				else 
					R0 = global_data->k_bond[hd_type][az_type];
	  
				R0 = R0*R0;		/* we need it only squared */
				sprot = rnx*rx + rny*ry + rnz*rz;
				g = (R0-Rn)/(2.0*sprot*(az_m + hd_m));
				
				/*** having r, rn in sprot gets the normation right: ***/
				/*** first we move the hydrogen                      ***/
				c = -g*az_m;
				my_data->x_new[hd] = my_data->x_new[hd] + c*rx;
				my_data->y_new[hd] = my_data->y_new[hd] + c*ry;
				my_data->z_new[hd] = my_data->z_new[hd] + c*rz;

				/*** then we move the antezedent ***/
				c = g*hd_m;
				my_data->x_new[az] = my_data->x_new[az] + c*rx;
				my_data->y_new[az] = my_data->y_new[az] + c*ry;
				my_data->z_new[az] = my_data->z_new[az] + c*rz;
			} /*** end for every shake-atom: do shake ***/
			
      			/*** test SHAKE constraints 		  ***/
			max_deviation        = 0.0;
			max_deviation_number = -1;
			within_tolerance     = TRUE;

			for (i = 0; i < my_data->shake_number / 2; i++) {
				int    hd_ptr = i * 2;
				int    hd = my_data->shake_list[hd_ptr  ] - global_data->first_atom[my_number];
				int    az = my_data->shake_list[hd_ptr+1] - global_data->first_atom[my_number];
				int    hd_type = my_data->typ[hd];
				int    az_type = my_data->typ[az];
				double x,y,z,len,diff,R0;

				/* a vector from the antezedent to the hydrogen in the
				 * direction of the bond */
				x = my_data->x_new[az] - my_data->x_new[hd];
				y = my_data->y_new[az] - my_data->y_new[hd];
				z = my_data->z_new[az] - my_data->z_new[hd];
				len = sqrt(x*x + y*y + z*z);

				if (hd_type<az_type) 	/*** load R0 (length to be) ***/
					R0 = global_data->k_bond[az_type][hd_type];
				else if (hd_type == az_type)
					R0 = global_data->k_bond[global_data->nr_of_atom_typs][hd_type];
				else
					R0 = global_data->k_bond[hd_type][az_type];
	  
				diff = R0 - len; 	/*** deviation from the ideal length */
				diff = fabs(diff/R0);	/*** now in percent of the ideal length */
				within_tolerance = within_tolerance && (diff<SHAKE_TOLERANCE);

				/* max_deviation, may be commented   */
				if (max_deviation < diff) {
					max_deviation_number = i;
					max_deviation = diff;
				}
      			} /*** end for every shake-atom: test */
			iteration = iteration + 1;
		} /*** end while not shaken ***/

		if (!within_tolerance) {
			errout("\nFatal error: SHAKE has not converged with tolerance %g.\n",
				SHAKE_TOLERANCE);
			errout("max_deviation=%g, atom_nr(max_deviation)=%d iteration %d\n",
				max_deviation,max_deviation_number,iteration);
			pex(-1);
		}
	} /*********************** end of if (shake) ***************************/


	/***********************************************************************/
	/****************** max_move and minimization **************************/
	/***********************************************************************/
	/*** it is the POSITIONS which are scaled! ***/
	if ((global_data->minimisation==1)||(global_data->minimisation==4)) {
		double diff;
		int    i;

		for (i=0;i<my_data->no_atoms;i++) {
			if (global_data->minimisation==4) {
				if ((sfx[i]<0.0)&&(my_new_forces->Fx[i]>0.0)) 
					my_data->x_old[i]=my_data->x_new[i];
				if ((sfx[i]>0.0)&&(my_new_forces->Fx[i]<0.0)) 
					my_data->x_old[i]=my_data->x_new[i];			
	
				if ((sfy[i]<0.0)&&(my_new_forces->Fy[i]>0.0)) 
					my_data->y_old[i]=my_data->y_new[i];
				if ((sfy[i]>0.0)&&(my_new_forces->Fy[i]<0.0)) 
					my_data->y_old[i]=my_data->y_new[i];
	
				if ((sfz[i]<0.0)&&(my_new_forces->Fz[i]>0.0)) 
					my_data->z_old[i]=my_data->z_new[i];
				if ((sfz[i]>0.0)&&(my_new_forces->Fz[i]<0.0)) 
					my_data->z_old[i]=my_data->z_new[i];
			}
			
			INTEGRATE7 (my_data->x_old[i], my_data->x_new[i]);
			INTEGRATE7 (my_data->y_old[i], my_data->y_new[i]);
			INTEGRATE7 (my_data->z_old[i], my_data->z_new[i]);
		}
		
		if (global_data->minimisation==4) {
			memcpy(sfx,my_new_forces->Fx,sizeof(double)*my_data->no_atoms);
			memcpy(sfy,my_new_forces->Fy,sizeof(double)*my_data->no_atoms);
			memcpy(sfz,my_new_forces->Fz,sizeof(double)*my_data->no_atoms);
		}
	}
	return;
}

/***********************************************************************/
/*** note: if global_data->schwing_mode=1 or 2 the real hesse matrix ***/
/*** is calculated. if however global_data->schwing_mode=-1 or -2    ***/
/*** only the hesse matrix of elec and vdw forces is calculated      ***/
/*** see also addup_all_forces(...) at end of the function           ***/
/***********************************************************************/
static int
calc_hessian(	LOAD_MSG_ONE_NODE *my_data,
		LOAD_MSG_ALL_NODE *global_data,
		TRAVELLING_FORCES *my_new_forces)
{		
	int i,j,goon,restart=0;
	char fname[MAX_STRING_SIZE],dipfname[MAX_STRING_SIZE];
	double scalfac=0.6951907/(abs(global_data->schwing_mode)*global_data->schwing_offs);
	double dipscalfac=1.0/(abs(global_data->schwing_mode)*global_data->schwing_offs);
	int fh,fhdip;

	strcpy(fname,global_data->outpath);
	strcat(fname,global_data->idstr);
	strcat(fname,"schwing.out");

	strcpy(dipfname,global_data->outpath);
	strcat(dipfname,global_data->idstr);
	strcat(dipfname,"dipgrad.out");
	
	if (schwing_counter_atom==-1) { /*** initialisation ****/
		int at=-1,an=-1,sa=-1,res=0;

		atomsschwing=0;
		for (i=0;i<my_data->no_atoms;i++) {
			if (my_data->atom_sele_flag1[i]&FLAG_SCHWING) atomsschwing++;
		}			

		sfx = alloc_doubles(my_data->no_atoms);
		sfy = alloc_doubles(my_data->no_atoms);
		sfz = alloc_doubles(my_data->no_atoms);	

		/*** save the forces in the initial state ***/
		memcpy(sfx,my_new_forces->Fx,my_data->no_atoms*sizeof(double));
		memcpy(sfy,my_new_forces->Fy,my_data->no_atoms*sizeof(double));
		memcpy(sfz,my_new_forces->Fz,my_data->no_atoms*sizeof(double));
		
		inidip[0] = global_data->dipol[0];
		inidip[1] = global_data->dipol[1];
		inidip[2] = global_data->dipol[2];
				
#if DEBUGSCHWING
		for (i=0;i<my_data->no_atoms;i++) {
			txtout("SCHWING: Iniforce %d: %f %f %f\n",
				i,sfx[i],sfy[i],sfz[i]);	
		}
#endif
		schwing_counter_atom++;
			
		/*** writeout the header: Number of atoms, masses, and start-coordinates ***/

      		fh=open(fname,O_RDWR,0400|0200);
      		if (fh!=-1) {   /*** check wether there is already a not yet	***/
      				/*** finished hess file.			***/
	      		read(fh,mybuf,24);
	      		res=sscanf(mybuf,"%d %d %d",&at,&an,&sa);
			close(fh);
			if (res!=3) {
				errout("INFO: Hess-calculation already finished (or corrupted hess-file)!\n");
				pex(-1);
			}
		}
		if ((res==3)&&(at==atomsschwing)) {
			txtout("************************************************************\n");
			txtout("Restart-situation will be handled for normal mode analysis.\n");
			txtout("Atom=%d, Axis=%d\n",an+1,sa);
			txtout("************************************************************\n");
			schwing_counter_axis=sa;
			schwing_counter_atom=an;
			global_data->cycles_passed = 3*an*abs(global_data->schwing_mode) + sa;

      			fh=open(fname,O_CREAT|O_RDWR,0400|0200);
			lseek(fh,0L,2);
			
			fhdip=open(dipfname,O_CREAT|O_RDWR,0400|0200);
			lseek(fhdip,0L,2);
			restart=1;

			for (i=0;i<my_data->no_atoms;i++) {
				my_data->x_old[i] = my_data->x_new[i];
				my_data->y_old[i] = my_data->y_new[i];
				my_data->z_old[i] = my_data->z_new[i];
			}
		} else {		
#if DEBUGSCHWING
			txtout("SCHWING: New hess file will be created\n");
#endif
 			/******** for the hesse file ******/
			fh=open(fname,O_CREAT|O_RDWR|O_TRUNC,0400|0200);
			sprintf(mybuf,"%-5d   %-5d   %-5d      \n",atomsschwing,-1,0);
			write(fh,mybuf,strlen(mybuf));
			for (i=0;i<my_data->no_atoms;i++) {
				int ty1 = my_data->typ[i];
				my_data->x_old[i] = my_data->x_new[i];
				my_data->y_old[i] = my_data->y_new[i];
				my_data->z_old[i] = my_data->z_new[i];	
					
				if (!(my_data->atom_sele_flag1[i]&FLAG_SCHWING)) continue;
					
				sprintf(mybuf,"%4d  %-4s    % 10.5f    % 10.5f % 10.5f % 10.5f\n",
					my_data->atom_id[i]+1,
					my_data->atom_string[i],
					global_data->AB[ty1][ty1],
					my_data->x_new[i],
					my_data->y_new[i],
					my_data->z_new[i]);
				write(fh,mybuf,strlen(mybuf));				
			}
			sprintf(mybuf,"Matrix:");			
			write(fh,mybuf,strlen(mybuf));
			
			/******** for the dipgrad file ******/
			fhdip=open(dipfname,O_CREAT|O_RDWR|O_TRUNC,0400|0200);
			sprintf(mybuf,"%-5d\n",atomsschwing);
			write(fhdip,mybuf,strlen(mybuf));
		}
		global_data->restarting=TRUE;
	} else {
#if DEBUGSCHWING
		txtout("SCHWING: Atomschwing %d %d %d\n",
			atomsschwing,
			schwing_counter_atom,
			schwing_counter_axis);
#endif
      		fh=open(fname,O_CREAT|O_RDWR,0400|0200);
		lseek(fh,0L,2);
		
      		fhdip=open(dipfname,O_CREAT|O_RDWR,0400|0200);
		lseek(fhdip,0L,2);
	}
		
	goon=0;
	if (	(schwing_counter_axis==2) ||
		(schwing_counter_axis==4) ||
		(schwing_counter_axis==0) ) {
		filled=1;
	} else	filled=0;
		
	for (i=0,j=0;i<my_data->no_atoms;i++) {
			
		/*** reset to old positions: should be equilibrium positions ***/
		my_data->x_new[i] = my_data->x_old[i];
		my_data->y_new[i] = my_data->y_old[i];
		my_data->z_new[i] = my_data->z_old[i];
			
		/*** skip not selected atoms ***/
		if (!(my_data->atom_sele_flag1[i]&FLAG_SCHWING)) continue;

		/*** write out 3 more values for HESSE-Matrix ***/
		if (!firsttime) {				
			if (	 (abs(global_data->schwing_mode)==1) ||
				((abs(global_data->schwing_mode)==2)&&(filled==1)) ) {
				
				sprintf(mybuf,"% 10.5f % 10.5f % 10.5f   ",
					scalfac*(my_new_forces->Fx[i]-sfx[i]),
					scalfac*(my_new_forces->Fy[i]-sfy[i]),
					scalfac*(my_new_forces->Fz[i]-sfz[i]));
				write(fh,mybuf,strlen(mybuf));
#if DEBUGSCHWING
				txtout("SCHWING: HESSE %d %s\n",i,mybuf);
#endif	
				if (j==0) {
					sprintf(mybuf,"% 10.5f % 10.5f % 10.5f  \n",
						dipscalfac*(global_data->dipol[0]-inidip[0]),
						dipscalfac*(global_data->dipol[1]-inidip[1]),
						dipscalfac*(global_data->dipol[2]-inidip[2]));
					write(fhdip,mybuf,strlen(mybuf));
				}
			} 
		}
			
		/*** set the next atom-positions ***/
		if (j==schwing_counter_atom) {
#if DEBUGSCHWING				
			txtout("SCHWING: Newsetting Atom %d   Axis %d Offs %f\n",
					schwing_counter_atom,schwing_counter_axis,
					global_data->schwing_offs);
#endif
			if (abs(global_data->schwing_mode)==1) {
				switch (schwing_counter_axis) {
				case 0:
					my_data->x_new[i] += global_data->schwing_offs;
					schwing_counter_axis = 1;
					break;
				case 1: 
					my_data->y_new[i] += global_data->schwing_offs;
					schwing_counter_axis = 2;
					break;
				case 2:
					my_data->z_new[i] += global_data->schwing_offs;
					schwing_counter_axis = 0;
					goon=1;
					break;
				default:
					errout("error in schwing\n");
					pex(-1);
					break;			
				} /*** end switch ***/
			} else
			if (abs(global_data->schwing_mode)==2) {
				switch (schwing_counter_axis) {
				case 0:
					my_data->x_new[i] -= global_data->schwing_offs;
					schwing_counter_axis = 1;
					break;
				case 1: 
					memcpy(sfx,my_new_forces->Fx,my_data->no_atoms*sizeof(double));
					memcpy(sfy,my_new_forces->Fy,my_data->no_atoms*sizeof(double));
					memcpy(sfz,my_new_forces->Fz,my_data->no_atoms*sizeof(double));	
					inidip[0] = global_data->dipol[0];
					inidip[1] = global_data->dipol[1];
					inidip[2] = global_data->dipol[2];
					my_data->x_new[i] += global_data->schwing_offs;
					schwing_counter_axis = 2;
					break;
				case 2:
					my_data->y_new[i] -= global_data->schwing_offs;
					schwing_counter_axis = 3;
					break;
				case 3: 
					memcpy(sfx,my_new_forces->Fx,my_data->no_atoms*sizeof(double));
					memcpy(sfy,my_new_forces->Fy,my_data->no_atoms*sizeof(double));
					memcpy(sfz,my_new_forces->Fz,my_data->no_atoms*sizeof(double));	
					inidip[0] = global_data->dipol[0];
					inidip[1] = global_data->dipol[1];
					inidip[2] = global_data->dipol[2];					
					my_data->y_new[i] += global_data->schwing_offs;
					schwing_counter_axis = 4;
					break;
				case 4:
					my_data->z_new[i] -= global_data->schwing_offs;
					schwing_counter_axis = 5;
					break;
				case 5: 
					memcpy(sfx,my_new_forces->Fx,my_data->no_atoms*sizeof(double));
					memcpy(sfy,my_new_forces->Fy,my_data->no_atoms*sizeof(double));
					memcpy(sfz,my_new_forces->Fz,my_data->no_atoms*sizeof(double));	
					inidip[0] = global_data->dipol[0];
					inidip[1] = global_data->dipol[1];
					inidip[2] = global_data->dipol[2];					
					my_data->z_new[i] += global_data->schwing_offs;
					schwing_counter_axis = 0;
					goon=1;
					break;				
				default:
					errout("error in schwing\n");
					pex(-1);
					break;												
				}  /*** end switch ***/	
			} /*** end mode 2 ***/
		} /*** end of if (j==schwing_counter_atom) ***/
		j++;
	}
	
	if (	 (abs(global_data->schwing_mode)==1) ||
		((abs(global_data->schwing_mode)==2)&&(filled==1)) ) { 
		sprintf(mybuf,"\n");
		if (!restart) write(fh,mybuf,strlen(mybuf));
	}
		
	/*** patching of the header ***/
	if (!firsttime) {
		int sa = schwing_counter_axis-1;
		int an = schwing_counter_atom;		
		lseek(fh,0L,0);
		
		if (sa<0) sa = abs(global_data->schwing_mode)*3-1;
		if (abs(global_data->schwing_mode)==2) {
			if (sa==1) sa=0;
			if (sa==3) sa=2;					
			if (sa==5) sa=4;					
		}	
				
		/*** we are ready, put no restart information to first line ***/
		if (an==atomsschwing) {
			sprintf(mybuf,"%-5d                    ",atomsschwing);
			write(fh,mybuf,strlen(mybuf));	
		} else {
			/*** write restart information to the first line ***/
			sprintf(mybuf,"%-5d   %-5d   %-5d   ",
				atomsschwing,
				an,
				sa);
			write(fh,mybuf,strlen(mybuf));	
		}		
	}
						
	close(fh);
	close(fhdip);
	
	if (goon) {
		schwing_counter_atom++;
	}
	firsttime=0;		
#if DEBUGSCHWING
	txtout("SCHWING: New positions:\n");
	for (i=0,j=0;i<my_data->no_atoms;i++) {
			txtout("SCHWING: Step %d  Atom %d: %f %f %f\n",	
			global_data->cycles_passed,
			i,
			my_data->x_new[i],
			my_data->y_new[i],
			my_data->z_new[i]);
	}
	txtout("\n");
#endif	
	return 0;		
} /******************* end of schwing_mode!=0 *********************/

