/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/**************************************************************************
**									***
**	Module	: egintern.c	by M.Eichinger, Munich			***
**	Belongs2: EGO		The parallel MD-Simulation programm	***
**	Contents: CalcIntern_ShortRange(...)				***
**		  	BOND, ANGLE, DIHEDRAL, IMPROPPER, HARMONIC	***
**			STOCHASTIC and SBOUND-Potential.		***
**									***																	
**	CalcIntern_LongRange(...)					***
**		Exact Calculation of VDW and Electrostatic between 	***
**		all intern atoms (only used for exact calculation)	***
**									***											
**************************************************************************/
#define DEBUG 0

#include <math.h>
#include <stdio.h>

#ifndef NEXT
#include <stdlib.h>
#endif

#include "ego.h"
#include "egalloc.h"
#include "egintern.h"
#include "egnomath.h"
#include "egnolib.h"
#include "egbolib.h"
#include "util_lib.h"
#include "egbound.h"
#include "egqminter.h"

/*** You may use this template for any additional potentials 	***/
/*** for a detailed description see below			***/ 
static	void add_userdefined_forces(	ENERGY_INFO *energy,
					LOAD_MSG_ONE_NODE *my_data,
					LOAD_MSG_ALL_NODE *global_data,
					TRAVELLING_FORCES *my_new_forces);

static	void set_A_B(	double *A,double *B,
			double *elecstat_loc_energy,
			int ex14,int typ1,int typ2,
			LOAD_MSG_ALL_NODE *global_data);

/************************************************************************/
/*** compute internal ShortRange interactions: bond, angle, dihe, impr **/
/*** also harmonic restraint forces and Random-Forces are computed     **/
/************************************************************************/

	
void   CalcIntern_ShortRange(	ENERGY_INFO *energy,
				LOAD_MSG_ONE_NODE *my_data,
				LOAD_MSG_ALL_NODE *global_data,
				TRAVELLING_FORCES *my_new_forces,
				double *rnd_force_std_deviation,
				double *Gamma /*** changes every step, if 
						   any sbound set! ***/			
				)
{
	int    i,nr_atom_typs=global_data->nr_of_atom_typs,ii;
	int	bond_ptr = 0;	/* Pointer in int_bond-list */
    				/* this atom is bound first */
	int	bond_partner;
	int	onthisnode[2],forceout_flag[2],myatom_id[2];

	/**********************************************************/
	/*** calculate intern BONDED interaction:		***/
	/*** energy:   Kb*(R-R0)^2				***/
	/*** force:  2*Kb*(R-R0)*(x/R)				***/
	/**********************************************************/
	
	onthisnode[0] = 1;
	onthisnode[1] = 1;

  	for (i=0,ii=0;i<my_data->no_atoms;i++,ii+=4) {
		int	typ1 = my_data->typ[i];
		int	ii1=ii+1,ii2=ii+2,ii3=ii+3;

		double	x = my_data->x_new[i];
		double	y = my_data->y_new[i];
		double	z = my_data->z_new[i];

		while ((bond_partner=my_data->int_bond_list[bond_ptr])!=END_OF_LIST) {
			double 	Kb,R0,dr,kr,bforce[4];
			double	rx = x - my_data->x_new[bond_partner];
			double	ry = y - my_data->y_new[bond_partner];
			double	rz = z - my_data->z_new[bond_partner];
			double 	R  = sqrt(rx*rx + ry*ry + rz*rz);
			int 	typ2 = my_data->typ[bond_partner];
			int	jj=4*bond_partner;

			/*** Bonding parameters: Force-constant and distance ***/
			if (typ1 < typ2) {
				Kb = global_data->k_bond[typ1][typ2];
				R0 = global_data->k_bond[typ2][typ1];
			} else
			if (typ1 > typ2) {
				Kb = global_data->k_bond[typ2][typ1];
				R0 = global_data->k_bond[typ1][typ2];
			} else {
				Kb = global_data->k_bond[typ1][nr_atom_typs];
				R0 = global_data->k_bond[nr_atom_typs][typ1];
			}

			dr  = R - R0;
			kr  = Kb * dr / 2;
			bforce[0] = kr*dr;

			/*** energy ***/
			my_data->force_bond[ii  ] += bforce[0];	
			my_data->force_bond[jj++] += bforce[0];

			kr = 4 * kr / R;

			/*** force x ***/
			bforce[1] = kr * rx;
			my_data->force_bond[ii1]  += bforce[1];
			my_data->force_bond[jj++] -= bforce[1];

			/*** force y ***/
			bforce[2] = kr * ry;
			my_data->force_bond[ii2]  += bforce[2];
			my_data->force_bond[jj++] -= bforce[2];

			/*** force z ***/
			bforce[3] = kr * rz;
			my_data->force_bond[ii3]  += bforce[3];
			my_data->force_bond[jj ]  -= bforce[3];
			
			/*** here we look whether one of the atoms***/
			/*** carries a special mark		  ***/ 
			forceout_flag[0] = my_data->atom_sele_flag1[i           ];
			forceout_flag[1] = my_data->atom_sele_flag1[bond_partner];
			
			if (	forceout_flag[0] || 
				forceout_flag[1]  ) {
					
				myatom_id[0] = my_data->atom_id[i           ];
				myatom_id[1] = my_data->atom_id[bond_partner];
				
				handle_qmbond(	onthisnode,forceout_flag,myatom_id,
						my_data,
						bforce,i,bond_partner);
		
				/*** Here we eventually sort the calculated forces ***/
				/*** to the PICK_BOND buffer which is sent to the  ***/
				/*** master and can be saved to disk		   ***/	
				if (	(forceout_flag[0]&PICK_BOND) &&
					(forceout_flag[1]&PICK_BOND)  ) {
					
					sort_to_pickbond(myatom_id,bforce[0],bforce,R,
							&(my_data->pickbond));				
				}
			}

			bond_ptr++;
		} /*** end while{} over bonded atoms ***/

     		bond_ptr++;		/* skip END_OF_LIST for bond */
  	} /*** end for{} intern BOND interaction ***/
#if DEBUG
	errout("CalcIntern_ShortRange: After bonded.\n");
#endif

	angle_energy(	my_data->int_angle_ptr,
			my_data->int_angle_Nr1,
			my_data->int_angle_Nr2,
			my_data->int_angle_Nr3,
			my_data->int_angle_index,
			my_data->atom_sele_flag1,
			my_data->atom_id,
			my_data->x_new,
			my_data->y_new,
			my_data->z_new,
			global_data->theta0,
			global_data->k_angle,
			global_data->ubk,
			global_data->ubd,
			TRUE,
			my_data->force_angle,
			0,			
			-1,
			NULL,
			&(my_data->pickangle),
			my_data);
#if DEBUG
	errout("CalcIntern_ShortRange: After angle.\n");
#endif

	improper_energy(my_data->int_impr_ptr,
			my_data->int_impr_Nr1,
			my_data->int_impr_Nr2,
			my_data->int_impr_Nr3,
			my_data->int_impr_Nr4,
			my_data->int_impr_index,
			my_data->atom_sele_flag1,
			my_data->atom_id,
			my_data->x_new,
			my_data->y_new,
			my_data->z_new,
			global_data->omega0,
			global_data->k_impr,
			TRUE,
			my_data->force_impr,
			0,
			-1,
			NULL,
			&(my_data->pickimpr),
			my_data);
#if DEBUG
	errout("CalcIntern_ShortRange: After impr.\n");
#endif

	dihedral_energy(my_data->int_dihe_ptr,
			my_data->int_dihe_Nr1,
			my_data->int_dihe_Nr2,
			my_data->int_dihe_Nr3,
			my_data->int_dihe_Nr4,
			my_data->int_dihe_index,
			my_data->atom_sele_flag1,
			my_data->atom_id,
			my_data->x_new,
			my_data->y_new,
			my_data->z_new,
			global_data->phi0,
			global_data->k_dihe,
			global_data->n_dihe,
			TRUE,
			my_data->force_dihe,
			0,
			-1,
			NULL,
			&(my_data->pickdihe),
			my_data);
#if DEBUG
	errout("CalcIntern_ShortRange: After dihe.\n");
#endif

        boundary(my_data, global_data, 
		 rnd_force_std_deviation,
		 Gamma,
		 &(energy->pressure),
		 &(energy->density));
#if DEBUG
	errout("CalcIntern_ShortRange: After boundary.\n");
#endif

	add_userdefined_forces(energy,my_data,global_data,my_new_forces);
#if DEBUG
	errout("CalcIntern_ShortRange: After user.\n");
#endif

	return;

} /*** end of Calculate_ShortRange ***/


void   CalcExtern_Bonds(	TRAVELLING_PACKETS *packet,
				LOAD_MSG_ALL_NODE *global_data,
				LOAD_MSG_ONE_NODE *my_data,
				int my_number)
{
	int	int_first 	= global_data->first_atom[my_number];
	int	ext_bond_len 	= my_data->ext_bond_ptr;
	int	ext_first 	= global_data->first_atom[packet->coordinates.node];
	int	ext_last 	= ext_first + (packet->coordinates.no_atoms-1);
	int	nr_atom_typs	= global_data->nr_of_atom_typs;
	int	list_end 	= END_OF_LIST - int_first;
	int	bond_ptr	= 0;
	int	forceout_flag[2],onthisnode[2],myatom_id[2];

	onthisnode[0] = 0;
	onthisnode[1] = 1;
	
	/* search for <ext_atom> in the external region */
	while (	(bond_ptr<ext_bond_len) && 
		(my_data->ext_bond_list[bond_ptr]<ext_first) ) {
		while ((my_data->ext_bond_list[bond_ptr]) != END_OF_LIST) bond_ptr++;
		bond_ptr++;			/*** skip END_OF_LIST ***/
	}

	while (	(bond_ptr<ext_bond_len) && 
		(my_data->ext_bond_list[bond_ptr]<=ext_last) ) {

      		int	bond_partner;
		int	ext_bond_elem     = my_data->ext_bond_list[bond_ptr] - ext_first;
		int	typ1              = packet->coordinates.typ[ext_bond_elem];

		double	x = packet->coordinates.x[ext_bond_elem];
		double	y = packet->coordinates.y[ext_bond_elem];
		double	z = packet->coordinates.z[ext_bond_elem];

		myatom_id[0]     = packet->coordinates.atom_id[ext_bond_elem];		
		forceout_flag[0] = packet->coordinates.atom_sele_flag1[ext_bond_elem];
		
		bond_ptr++; /*** points now to the first internal element ***/

		bond_partner = my_data->ext_bond_list[bond_ptr] - int_first;
	
		while (bond_partner != list_end) {
			double	Kb,R0,dr,kr,bforce[4];
			double	rx = x - my_data->x_new[bond_partner];
			double	ry = y - my_data->y_new[bond_partner];
			double	rz = z - my_data->z_new[bond_partner];
			double	R  = sqrt(rx*rx + ry*ry + rz*rz);
			int	typ2 = my_data->typ[bond_partner];
			int	jj=4*bond_partner;

			/*** Bonding parameters: Force-constant and distance ***/
			if (typ1 < typ2) {
				Kb = global_data->k_bond[typ1][typ2];
				R0 = global_data->k_bond[typ2][typ1];
			} else 
			if (typ1 > typ2) {
				Kb = global_data->k_bond[typ2][typ1];
				R0 = global_data->k_bond[typ1][typ2];
			} else {
				Kb = global_data->k_bond[typ1][nr_atom_typs];
				R0 = global_data->k_bond[nr_atom_typs][typ1];
			}

			dr = R - R0;
			kr = Kb * dr / 2;
			bforce[0]  = kr*dr;	/*** half of the energy ***/
			/*** add here only half of the energy, because same amount is   ***/
			/*** added on the other node, which exhibits the same situation ***/
			my_data->force_bond[jj++] += bforce[0];

			kr = 4 * kr / R;
			bforce[1] = kr*rx;
			my_data->force_bond[jj++] -= bforce[1];
			
			bforce[2] = kr*ry;
			my_data->force_bond[jj++] -= bforce[2];
			
			bforce[3] = kr*rz;
			my_data->force_bond[jj  ] -= bforce[3];

			/*** Here we eventually sort the calculated forces ***/
			/*** to the PICK_BOND buffer which is sent to the  ***/
			/*** master and can be saved to disk		   ***/
			forceout_flag[1] = my_data->atom_sele_flag1[bond_partner];

			if (	forceout_flag[0] || 
				forceout_flag[1]  ) {
				
				myatom_id[1] = my_data->atom_id[bond_partner];

				handle_qmbond(	onthisnode,forceout_flag,myatom_id,
						my_data,
						bforce,0,bond_partner);
								
				/*** Here we eventually sort the calculated forces ***/
				/*** to the PICK_BOND buffer which is sent to the  ***/
				/*** master and can be saved to disk		   ***/	
				if (	(forceout_flag[0]&PICK_BOND) &&
					(forceout_flag[1]&PICK_BOND)){
					
					sort_to_pickbond(myatom_id,bforce[0],bforce,R,
							&(my_data->pickbond));				
				}
			}

			bond_ptr++;
			bond_partner = my_data->ext_bond_list[bond_ptr] - int_first;
		} /*** end of while{} ***/

		bond_ptr++;	/*** skip the END_OF_LIST ***/
	} /*** end of while{} ***/
}

/************************************************************************/
void   CalcExtern_ShortRange(	LOAD_MSG_ONE_NODE *my_data,
			        LOAD_MSG_ALL_NODE *global_data,
			        EXT_3_ATOM *ext_angle,
			        EXT_4_ATOM *ext_dihe,
			        EXT_4_ATOM *ext_impr,
			        int my_number)
{
	angle_energy(	my_data->ext_angle_ptr,
			my_data->ext_angle_Nr1,
			my_data->ext_angle_Nr2,
			my_data->ext_angle_Nr3,
			my_data->ext_angle_index,
			my_data->atom_sele_flag1,
			my_data->atom_id,
			my_data->x_new,
			my_data->y_new,
			my_data->z_new,
			global_data->theta0,
			global_data->k_angle,
			global_data->ubk,
			global_data->ubd,
			FALSE,
			my_data->force_angle,
			global_data->first_atom[my_number],			
			my_number,
			ext_angle,
			&(my_data->pickangle),
			my_data);

	improper_energy(my_data->ext_impr_ptr,
			my_data->ext_impr_Nr1,
			my_data->ext_impr_Nr2,
			my_data->ext_impr_Nr3,
			my_data->ext_impr_Nr4,
			my_data->ext_impr_index,
			my_data->atom_sele_flag1,
			my_data->atom_id,
			my_data->x_new,
			my_data->y_new,
			my_data->z_new,
			global_data->omega0,
			global_data->k_impr,
			FALSE,
			my_data->force_impr,
			global_data->first_atom[my_number],
			my_number,
			ext_impr,
			&(my_data->pickimpr),
			my_data);

 	dihedral_energy(my_data->ext_dihe_ptr,
			my_data->ext_dihe_Nr1,
			my_data->ext_dihe_Nr2,
			my_data->ext_dihe_Nr3,
			my_data->ext_dihe_Nr4,
			my_data->ext_dihe_index,
			my_data->atom_sele_flag1,
			my_data->atom_id,
			my_data->x_new,
			my_data->y_new,
			my_data->z_new,
			global_data->phi0,
			global_data->k_dihe,
			global_data->n_dihe,
			FALSE,
			my_data->force_dihe,
			global_data->first_atom[my_number],
			my_number,
			ext_dihe,
			&(my_data->pickdihe),
			my_data);
}


/******************************************************************************/
/*** We store only the force on the first noted atom pickbond->atom_ids[i  ],**/
/*** because the force on the second noted atom pickbond->atom_ids[i+1] is   **/
/*** just the negative from the other.					     **/
/*** This function is also used for PICK_ELEC and PICK_VDW in egtree.c	     **/
/******************************************************************************/	
int
sort_to_pickbond(	int *myatom_id,
			double E, double *bforce,double R,
			STRUCT_PICK_BOND *pickbond)
{
	int	i,j;
	double aforce[3];
	
	for (i=0,j=0;i<2*pickbond->len;i+=2,j+=5) {
		if (	(pickbond->atom_ids[i  ]==myatom_id[0]) &&
			(pickbond->atom_ids[i+1]==myatom_id[1]) ) {
			memcpy(aforce,bforce,3*sizeof(double));
		} else
		if (	(pickbond->atom_ids[i  ]==myatom_id[1]) &&
			(pickbond->atom_ids[i+1]==myatom_id[0]) ) {
			memcpy(aforce,bforce,3*sizeof(double));
			aforce[0] = -aforce[0];
			aforce[1] = -aforce[1];
			aforce[2] = -aforce[2];			
		} else	
			continue; /*** not the right tuple ***/

		if (pickbond->cur_len>=pickbond->len)
			goto inconsistent_pickbond;

		pickbond->force[j  ] = E;
		pickbond->force[j+1] = aforce[0];
		pickbond->force[j+2] = aforce[1];
		pickbond->force[j+3] = aforce[2];
		pickbond->force[j+4] = R;
		pickbond->cur_len++;
#if DEBUG_PICK
txtout("PICKBOND %d %d + %d: %f   %f %f %f    %f\n",
pickbond->atom_ids[i  ],
pickbond->atom_ids[i+1],
pickbond->cur_len,E,aforce[0],aforce[1],aforce[2],R);
#endif
		return 0;	/*** data sorted in -> go home ***/
	}
	return 0;

inconsistent_pickbond:
	errout("Fatal error: Inconsistent data in sort_to_pickbond %d %d!\n",
		myatom_id[0],myatom_id[1]);
	errout("This is surely a programming bug!\n");
	pex(-1);
	return -1;
}

				
/************************************************************************/
/****** here Electrostatic and VDW-Forces are calculated ****************/
/****** exact calculation, no cut-off !!!		 ****************/
/************************************************************************/

static
void   set_A_B(	double *A,
		double *B,
		double *elecstat_loc_energy,
		int ex14,
		int typ1,
		int typ2,
		LOAD_MSG_ALL_NODE *global_data)

{
	if (ex14) {
		if (typ1 < typ2) {
			*A = global_data->AB14[typ1][typ2];
			*B = global_data->AB14[typ2][typ1];
		} else
		if (typ1 == typ2) {
			*A = global_data->AB14[typ1][global_data->nr_of_atom_typs];
			*B = global_data->AB14[global_data->nr_of_atom_typs][typ1];
		} else {
			*A = global_data->AB14[typ2][typ1];
			*B = global_data->AB14[typ1][typ2];
		}
		*elecstat_loc_energy = *elecstat_loc_energy * global_data->elec14;
	} else {
		if (typ1 < typ2) {
			*A = global_data->AB[typ1][typ2];
			*B = global_data->AB[typ2][typ1];
		} else
		if (typ1 == typ2) {
			*A = global_data->AB[typ1][global_data->nr_of_atom_typs];
			*B = global_data->AB[global_data->nr_of_atom_typs][typ1];
		} else {
			*A = global_data->AB[typ2][typ1];
			*B = global_data->AB[typ1][typ2];
		}
	}
}

/* set initial values set ex_elem */
    /* Format */
    /* the exclusion list has the following format: <ex_1> <ex_2> ...
     * <END_OF_LIST> <ex_1> <ex_2> ... <END_OF_LIST> ...; for each (local)
     * atom there is one entry; eventually <END_OF_LIST> only; the
     * exclusion atoms <ex_1>, ... are in ascending order and contain all
     * exclusions, especially the reverse ones, too; the atom numbers are
     * "global", i.e. <int_first [...]> is included??? and must be
     * subtracted in order to get the local index for the type arrays; the
     * atom itself, to which the excl_list belongs, is not included in the
     * list! */


void   CalcIntern_LongRange(	  ENERGY_INFO *energy,
			          LOAD_MSG_ALL_NODE *global_data,
			          LOAD_MSG_ONE_NODE *my_data,
			          TRAVELLING_FORCES *my_new_forces,
			          int int_first)
{
	int 	i,j;
	int 	list_end = (END_OF_LIST - int_first);
	int    	ex14_ptr,ex_ptr,ex14_flag;

     	ex_ptr   = 0;		/* pointer in exclusion-list */
    	ex14_ptr = 0;		/* pointer in exclusion-list */

    	for (i = 0; i < my_data->no_atoms; i++) {
		double x,y,z;	
		double Q;
		int typ1 = my_data->typ[i];
		int ex14_elem,ex_elem;

		/* skip my_data->ex_list until in local range set atom according */
		/* to pointer */
		/* this atom is skipped first (local index) */
		ex_elem = my_data->ex_list[ex_ptr] - int_first;
		while ((ex_elem<=i) && (ex_elem!=list_end)) {
			ex_ptr++;
			ex_elem = my_data->ex_list[ex_ptr] - int_first;
		}

		ex14_elem = my_data->ex14_list[ex14_ptr] - int_first;
		while (	(global_data->special_14_excl) &&
	       		(ex14_elem<=i) &&
			(ex14_elem != list_end)) {
			ex14_ptr++;
			ex14_elem = my_data->ex14_list[ex14_ptr] - int_first;
		}

		x = my_data->x_new[i];
		y = my_data->y_new[i];
		z = my_data->z_new[i];
		Q = my_data->q[i] * global_data->f_elstat;
	
		/*------------------------------------- */
		/* electrostatics:                   -- */
		/* ---------------                   -- */
		/* energy: q1*q2*faktor*r.1          -- */
		/* force:  q1*q2*faktor*r.3 * rx     -- */
		/*------------------------------------- */
		/*------------------------------------- */
		/* van der Waals:                    -- */
		/* --------------                    -- */
		/* energy: r.6*(r.6*A[i,j]-B[i,j])   -- */
	        /*------------------------------------- */
		for (j=(i+1);j<my_data->no_atoms;j++) {
			ex14_flag = (global_data->special_14_excl && (j == ex14_elem));
			if (j != ex_elem) {
	    			int 	typ2 = my_data->typ[j];
				double 	q2   = my_data->q[j];
				double 	rx,ry,rz,r_2,r_6,
					elecstat_loc_energy,
					vdw_loc_energy,A,B;
				double	fx_h,fy_h,fz_h;

				rx = x - my_data->x_new[j];
				ry = y - my_data->y_new[j];
				rz = z - my_data->z_new[j];
	    			r_2 = 1.0 / (rx * rx + ry * ry + rz * rz);
	    			r_6 = (r_2 * r_2) * r_2;
	    			/* use the inverse here: later multiply instead of divide! */

	    			/* electrostatics */
	    			if ((Q == ZERO) || (q2 == ZERO)) 
					elecstat_loc_energy = ZERO;
	    			else	elecstat_loc_energy = Q * q2 * sqrt (r_2);
				set_A_B(&A, &B, &elecstat_loc_energy, ex14_flag, typ1,
					typ2, global_data);
				if (ex14_flag) {	/* next ex14 elem */
					ex14_ptr++;
					ex14_elem = my_data->ex14_list[ex14_ptr]-int_first;
				}
				/* vdW */
				vdw_loc_energy = r_6 * (r_6 * A - B);
				energy->elec += elecstat_loc_energy;
				energy->vdw  += vdw_loc_energy;

				elecstat_loc_energy = r_2 * (elecstat_loc_energy +
					 	     (12.0 * vdw_loc_energy + 6.0 * r_6 * B)
						    );
	    			/* forces */
	    			fx_h = elecstat_loc_energy * rx;
	    			fy_h = elecstat_loc_energy * ry;
	    			fz_h = elecstat_loc_energy * rz;

				my_new_forces->Fx[i] -= fx_h;
				my_new_forces->Fy[i] -= fy_h;
				my_new_forces->Fz[i] -= fz_h;

				my_new_forces->Fx[j] += fx_h;
				my_new_forces->Fy[j] += fy_h;
				my_new_forces->Fz[j] += fz_h;
			} else {			/* j is excluded */
				ex_ptr  = ex_ptr + 1;
				ex_elem = my_data->ex_list[ex_ptr] - int_first;
				/* this atom is skipped next */
				/* if it is outside the local region, it does not matter; */
				/* it only means that all the local exclusions are already used. */

				if (ex14_flag) {
					ex14_ptr = ex14_ptr + 1;
					ex14_elem = my_data->ex14_list[ex14_ptr] - int_first;
					/* this atom is skipped next */
				}
			}
      		}	/* end of the inner for loop j */
		/* set next ex_elem; search for end of list */
     		 while (ex_elem != list_end) {	/* search for end of list */
			ex_ptr++;
			ex_elem = my_data->ex_list[ex_ptr] - int_first;	
						/* this atom is skipped next */
		}
     		ex_ptr++;  /* skip end of list */
		/* here: set only pointer and do not access list because this might */
		/* cause a boundary violation (think of the last element!) */
		/* set next ex14_elem; search for end of list */
		/* search for end of list */
		while (global_data->special_14_excl && (ex14_elem != list_end)) {
			/* skip exclusions outside the region! */
			ex14_ptr++;
			ex14_elem = my_data->ex14_list[ex14_ptr] - int_first;
			/* this atom is skipped next */
		}

		/* skip end of list */
		ex14_ptr++;
		/* here: set only pointer and do not access list because this might */
		/* cause a boundary violation (think of the last element!) */
	} /* end of long for loop */
}

/************************************************************************/

void   CalcExtern_LongRange(	TRAVELLING_PACKETS *packet,
				ENERGY_INFO *energy,
				LOAD_MSG_ALL_NODE *global_data,
				LOAD_MSG_ONE_NODE *my_data,
				TRAVELLING_FORCES *my_new_forces,
				int my_number)
{
	int	ext_first = global_data->first_atom[packet->coordinates.node];
	int	list_end = END_OF_LIST - ext_first;
	int	ex_ptr,ex14_ptr,ex14_flag;
	int	i=my_number; /*** no warning ***/

	ex_ptr = 0;		/* pointer in exclusion-list */
	ex14_ptr = 0;		/* pointer in exclusion-list */

	for (i=0;i<my_data->no_atoms;i++) {
		int	ex_elem,ex14_elem;
		double	x,y,z;		
		double	Q;
      				/* setup ex_elem in appropriate range */
		ex_elem = my_data->ex_list[ex_ptr] - ext_first;	/* this atom is skipped first */
		while ((ex_elem < 0) && (ex_elem != list_end)) {
			/* skip exclusions outside the range! */
			/* range starts with j=0 */
			ex_ptr++;
			/* this atom is skipped first */
			ex_elem = my_data->ex_list[ex_ptr] - ext_first;
		}

		/* setup ex14_elem in appropriate range */
		/* this atom is skipped first */
		ex14_elem = my_data->ex14_list[ex14_ptr] - ext_first;
		while (	(ex14_elem < 0) &&
			(ex14_elem != list_end) &&
			global_data->special_14_excl) {
			/* skip exclusions outside the range! */
			/* range starts with j=0 */
			ex14_ptr++;
			/* this atom is skipped first */
			ex14_elem = my_data->ex14_list[ex14_ptr] - ext_first;
		}

		x = my_data->x_new[i];
		y = my_data->y_new[i];
		z = my_data->z_new[i];
		Q = (my_data->q[i]) * global_data->f_elstat;
 
		/* account for the extern interaction; */
		/* is treated here, in order to avoid IF statements for the other
		 * atoms (optimization); */
		/* because of ACTIO = REACTIO, only half the work is needed! */
		 /*------------------------------------- */
	      /* electrostatics:                   -- */
	      /* ---------------                   -- */
	      /* energy: q1*q2*faktor*r.1          -- */
	      /* force:  q1*q2*faktor*r.3 * rx     -- */
	      /*------------------------------------- */
	      /*------------------------------------- */
	      /* van der Waals:                    -- */
	      /* --------------                    -- */
	      /* energy: r.6*(r.6*A[i,j]-B[i,j])   -- */
	      /* force:  r.2*(12*E + 6*r.6*B[i,j]) -- */
	      /*------------------------------------- */

		{
		int	typ1 = my_data->typ[i];
		int	j;

		for (j = 0; j < packet->coordinates.no_atoms; j++) {
			ex14_flag = (global_data->special_14_excl && (j == ex14_elem));

			if (j != ex_elem) {
	     			int 	typ2 = packet->coordinates.typ[j];
				double	q2   = packet->coordinates.q[j];

				/* distance vector */
				/* inverse distance squared */
				/* inverse distance to the 6th power (for vdW) */
				/* additional variables */
		 		double	rx,ry,rz,r_2,r_6,
					elecstat_loc_energy,
					vdw_loc_energy,A,B;
				double	fx_h,fy_h,fz_h;

				/* radius */
				rx = x - packet->coordinates.x[j];
				ry = y - packet->coordinates.y[j];
				rz = z - packet->coordinates.z[j];
				r_2 = 1.0 / (rx*rx + ry*ry + rz*rz);
				r_6 = r_2*r_2*r_2;

				/********* electrostatics *********/
				if ((Q == 0.0) || (q2 == 0.0))
					elecstat_loc_energy = 0.0;
				else	elecstat_loc_energy = Q*q2*sqrt (r_2);

				set_A_B (&A, &B, &elecstat_loc_energy, ex14_flag, typ1,
					typ2, global_data);

				if (ex14_flag) {	/* set next ex14 element */
					ex14_ptr++;
					ex14_elem = my_data->ex14_list[ex14_ptr] - ext_first;
				}

				/********* vdW *******************/
				vdw_loc_energy = r_6 * (r_6 * A - B);

				/*** here we have to add only half of the energy ***/
				/*** because we do calculate it on the other node again ***/	
				energy->elec += elecstat_loc_energy/2.0;
				energy->vdw  += vdw_loc_energy/2.0;
				elecstat_loc_energy = r_2 * (elecstat_loc_energy +
						(12.0*vdw_loc_energy + 6.0*r_6*B));

				fx_h = elecstat_loc_energy * rx;
				fy_h = elecstat_loc_energy * ry;
				fz_h = elecstat_loc_energy * rz;

				my_new_forces->Fx[i] -= fx_h;
				my_new_forces->Fy[i] -= fy_h;
				my_new_forces->Fz[i] -= fz_h;

			} else {	/* skip this atom */
				ex_ptr++;
				ex_elem = my_data->ex_list[ex_ptr] - ext_first;
	
				if (ex14_flag) {
					ex14_ptr++;
					/* this atom is next */
					ex14_elem = my_data->ex14_list[ex14_ptr] - ext_first;
				}
			} /* end else: j was excluded */
		} /* end for (j....) loop */
      		}	/* end of range of inner loop */

		/* set ex_elem for next i-loop */
		/* search end of sub-list for this atom */
		 while (ex_elem != list_end) {	/* search for end of list */
			/* skip exclusions outside the range! */
			ex_ptr++;
			/* this atom skipped next */
			ex_elem = my_data->ex_list[ex_ptr] - ext_first;
		}

		/* skip end of list */
		ex_ptr++;

		/* set ex14_elem for next i-loop search end of sub-list for this */
		/* search end of list */
		while (global_data->special_14_excl && (ex14_elem != list_end)) {
			/* skip exclusions outside the range! */
			ex14_ptr++;
			/* this atom is skipped next */
			ex14_elem = my_data->ex14_list[ex14_ptr] - ext_first;
		}

		/* skip end of list */
		ex14_ptr++;
	} /* end of for (i;...) loop */
}
/**************************************************************************/

/**************************************************************************/
/*** here you can program simple potentials of your own, e.g. special 	***/
/*** forces acting on special atoms:					***/
/*** All structures are defined in ego.h				***/
/*** the moste like needed structure elements are listed below:		***/
/*** my_data->no_atoms		Number of atoms loaded on this node.	***/
/*** my_data->x_new[i]		atom position x of atom i.		***/
/*** my_data->atom_string[i]	atom string as in the PDB/PSF file.	***/
/*** my_data->res_string[i]	residue string as in the PDB/PSF file.	***/
/*** my_data->atom_id[i]	atom_number-1 in the PDB/PSF file.	***/
/*** my_data->res_id[i]		residue_number as in the PDB/PSF file.	***/
/*** my_new_forces->Fx[i]	the x-force on atom i. All forces are   ***/
/***				added to this force buffer !		***/
/*** energy->...		add your energy to one of the elements. ***/
/***				of the energy structure.		***/
/*** global_data->...		These struncture contains all data which***/
/***				is identical to all nodes, e.g. force   ***/
/***				constants, etc.				***/
/**************************************************************************/							

static	void add_userdefined_forces(	ENERGY_INFO *energy,
					LOAD_MSG_ONE_NODE *my_data,
					LOAD_MSG_ALL_NODE *global_data,
					TRAVELLING_FORCES *my_new_forces)
{
	int i=my_data->atom_id[0];	/*** no warnings ***/
	int b=global_data->total_no_atoms;
	double c=my_new_forces->Fx[0];
	double d=energy->elec;
	i++;b++;c++;d++;
#if DONT_INCLUDE_THIS
	/*** this example shows, how to exert a force on a special atom.	***/
	/*** Lets choose the 13th atom in the PDB/PSF which has in   		***/
	/*** the atom_id==12					     		***/  

	for (i=0;i<my_data->no_nodes;i++) { 	/*** scan every atom on node 	***/
		if (my_data->atom_id[i]==12) { 	/*** we found the one		***/
			my_new_forces->Fx[i] += 1.0;/*** add force in x-direction	***/ 
		}
	}
#endif
}
