/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: egmaster.c	by M.Eichinger, Munich			**
**	Belongs2: EGO							**
**	Contents: init of master node,					**	
**		  main-loop of master node				**
**  	Modif's:							**
**									**
*************************************************************************/
#if defined(SEQ)

#else

#include <math.h>
#include <stdio.h>

#define DEBUG		0
			/* DEBUG = 0: no debug at all
			 * DEBUG = 1: standard debugging
			 * DEBUG = 2: debug special mode
			 * DEBUG = 3: get both: special mode debug and standard debug
			 * DEBUG = 8: enable malloc debug
			 */
#define SHOWLOAD	0
/******************************************/
/*** for the sequential version this	***/
/*** file is empty. All things of the	***/
/*** sequential "controller" are located***/
/*** in the file egcontro.c!		***/
/******************************************/


/******************************************/
/*** for the parallel version this file	***/
/*** represents the master.		***/
/******************************************/

#include "ego.h"
#include "timing.h"
#include "egalloc.h"
#include "egbolib.h"
#include "egmach.h"
#include "util_lib.h"
#include "eganaly.h"
#include "egloader.h"
#include "egbound.h"
#include "egcolib.h"
#include "egpull.h"
#include "egcontro.h"
#include "dbmalloc.h"	/*** include always as last include file ***/

static void	zero_energy(	ENERGY_INFO *energy,int no_nodes);
static void     zero_rotation(  LOAD_MSG_ALL_NODE *global_data);
static void 	addup_energy(	ENERGY_INFO *energy,
				ENERGY_INFO *local_energy,
				int other_number);

void	master_sigterm_proc(int signo);

#if defined(PARIX) || defined(PARIXHAL)
#include "egmepvm.h"
#endif

/*** this routine is called once. It handles all the communications ***/
/*** between the master and the slaves while the program is running ***/

void   master(	LOAD_MSG_ALL_NODE *global_data,
		double time_step,
		int file_count,
		char *version_string,
		STRUCT_FORCES *my_force)
{
	int    do_analyze,miss_first,restartlevel,coords_are_coming;
	int    new_clustering;
	TRAVELLING_COORDINATES coordinates;
	ENERGY_INFO energy,local_energy;
	double temperature = 1234567890.0;
	double temperature_h2o     = 1234567890.0;
	double temperature_protein = 1234567890.0;

	double temp_factor,temp_factor_h2o,temp_factor_protein;
	double nodetime[100][NUM_TIMERS],aveloadmiss=0.0;
		
	/* holds the time used for the current integration step (in seconds) */
 	double	cycletime_user = 0.0,cycletime_real = 0.0;	 	
	int	cycletimer,i,k,loadcounter=0;
#if (DEBUG & 1)
txtout("Master: enter master.\n");
#endif
	memset(&energy,0,sizeof(ENERGY_INFO));
	memset(&local_energy,0,sizeof(ENERGY_INFO));

	set_terminate_flag(DONT_STOP);

	/*********************************************************************/
  	/*** 2/(3*N*Kb) in AKMA 3(N-2) is the number of degrees of freedom ***/
   	/*** (velocities) we have to disregard velocity of the center of   ***/
   	/*** mass and the total angular momentum, therefore we take 6 DOF  ***/
   	/*** away. if shaking hydrogens, we further subtract num_shake/2 as***/
	/*** we remove one (velocity) DOF per hydrogen atom. Temperature in***/
	/*** Xplor is calculated with 3N DOF (this is wrong !!)		   ***/
	/*********************************************************************/
	if (global_data->total_no_atoms<3) {
		temp_factor = 2.0 / (
			KBOLTZ*(3.0*global_data->total_no_atoms-global_data->num_shake/2.0));
	} else {
		temp_factor = 2.0 / (
			KBOLTZ*(3.0*(global_data->total_no_atoms-2.0)-global_data->num_shake/2.0));
		temp_factor_h2o = 2.0 / (
			KBOLTZ*(3.0*(global_data->total_h2o_atoms-2.0)-global_data->num_shake_h2o/2.0));
		temp_factor_protein = 2.0 / (
			KBOLTZ*(3.0*(global_data->total_protein_atoms-2.0)-global_data->num_shake_protein/2.0));

	}

	/**************************************************************/
	set_allocinfo("travelling coordinates");
	coordinates.no_atoms 	= (int) (1.2*global_data->maxatoms_per_node);
	coordinates.node 	= -1;
	alloc_TRAVELLING_COORDINATES(&coordinates);

	set_allocinfo("inirestart");
	initialize_restart();
	set_allocinfo("inianalyze");
	initialize_analyze();
	ini_constant_pressure(global_data->sbound_droff, 
			      global_data->keep_press_const);
#if (DEBUG & 8) == 8
	testmalloc("after ini_constant_pressure in egmaster(...)",1);
#endif
	set_allocinfo("XXX");

	miss_first        = 0;
	do_analyze	  = 1;	 
	restartlevel      = 0;
	coords_are_coming = do_analyze;
	new_clustering    = global_data->treeinfo.steps;

#if (DEBUG & 1)
txtout("Master: before send start_nodes.\n");
#endif
	send_start_node(global_data->no_nodes);

	/*** Now we are really computing and now we manipulate the signals ***/
	set_signal_procs(master_sigterm_proc);

	/*** start the timer here ***/
	loadcounter     = 0;
	aveloadmiss     = 0.0;
	cycletime_user  = 0.0;
	cycletime_real  = 0.0;
	cycletimer      = init_timer();
	clear_timer(cycletimer);
	start_timer(cycletimer);

	print_meminfo();
	/**********************************************************/
	/*** here starts the big loop on the master over all 	***/
	/*** integration steps. A corresponding loop exists  	***/
	/*** also on the nodes (see egnode.c)			***/
	/**********************************************************/
	while (global_data->cycles_passed<=global_data->final_cycle_no) {
		double scaling_factor;
		
#if (DEBUG & 8) == 8
testmalloc("begining of cycle loop in egmaster(...)",1);
#endif
#if (DEBUG & 1)
txtout("Master: cycles_passed=%d,final_cycle_no=%d\n",
global_data->cycles_passed,global_data->final_cycle_no);
#endif
		PCI_synchronise(global_data->no_nodes);
 
		/******** adding up all energies and some other stuff ***/
		/******** from all nodes                              ***/
		zero_energy(&energy,global_data->no_nodes);

		if ((global_data->rotation_mode) && (global_data->cycles_passed==global_data->selimob_start)) {
		  energy.rotation_energy       = 0.0;
		  global_data->rotation_energy = 0.0;
		}

		for (i=0;i<global_data->no_nodes;i++) { /*** loop over all nodes ***/
			int other_number;
#if (DEBUG & 1)
txtout("Master: before energy recv count %d of %d \n",i+1,
global_data->no_nodes);
#endif
			/*** get all energies and some other stuff from the nodes ***/
			PCI_energy_recv(i,
					&local_energy,
					global_data->no_nodes,
					&other_number,
					my_force);
#if (DEBUG & 1) || (DEBUG & 2)
txtout("Master: after  energy recv count %d of %d\n",i+1,global_data->no_nodes);
txtout("Master: recieved pulling positions (%f,%f,%f) and (%f,%f,%f,%f)\n",
local_energy.pullspringpos[0],
local_energy.pullspringpos[1],
local_energy.pullspringpos[2],
local_energy.pullspringpos[3],
local_energy.pullspringpos[4],
local_energy.pullspringpos[5],
local_energy.pullspringpos[6]);
#endif
#if (DEBUG & 2)
txtout("Master: YP_pos=%f, YM_pos=%f\n",
local_energy.special_YP_pos,
local_energy.special_YM_pos);
#endif
			/*** add it up ***/
			addup_energy(&energy,&local_energy,other_number);
			memcpy(	&(nodetime[i][0]),
				&(local_energy.time[0]),
				sizeof(double)*NUM_TIMERS);
		}

		energy.pullspringpos[0]=local_energy.pullspringpos[0];
		energy.pullspringpos[1]=local_energy.pullspringpos[1];
		energy.pullspringpos[2]=local_energy.pullspringpos[2];
		
		global_data->pullspringpos[0]=energy.pullspringpos[0];
		global_data->pullspringpos[1]=energy.pullspringpos[1];
		global_data->pullspringpos[2]=energy.pullspringpos[2];
#if (DEBUG & 1)
txtout("Master: before PCI_etcdata_send");
#endif
		/*** send summed up energy-struct data back to nodes ***/
	        PCI_etcdata_send(&energy,global_data);
#if (DEBUG & 1)
txtout("Master: after PCI_etcdata_send");
#endif

		reset_force_buffer(my_force,global_data);
		calc_imob_energy(&energy,global_data);
        	adopt_pull_frame(&energy,global_data);

		if (global_data->rotation_mode) {
		  global_data->rotation_energy += energy.rotation_energy;
		}


#if (DEBUG & 8) == 8
testmalloc("after adopt_pull_frame in egmaster(...)",1);
#endif
		if ((coords_are_coming)||(new_clustering<=1)
		    || ((global_data->rotation_mode) &&
		     (global_data->cycles_passed==global_data->selimob_start))) {
		        zero_rotation(global_data);

			/*** receive all coordiantes from the nodes.	  ***/
			/*** we need these coordiantes for writing it out ***/
			/*** to file or for reclustering		  ***/
			for (i=0;i<global_data->no_nodes;i++) {
#if (DEBUG & 1)
txtout("Master: before recv coords %d of %d.\n",i+1,global_data->no_nodes);
#endif
				PCI_coords_recv(i,&coordinates,global_data);
#if (DEBUG & 1)
txtout("Master: after  recv coords %d of %d.\n",i+1,global_data->no_nodes);
#endif			
				sort_to_analybuffer(new_clustering,&coordinates,global_data);

				if ( 	(restartlevel==1) || /*** write first set  ***/
					(restartlevel==2) ){ /*** write second set ***/
#if (DEBUG & 1)
txtout("Master: write out restart-file level %d  filecount %d\n",restartlevel,file_count);
#endif
					restart(restartlevel,
						global_data->cycles_passed+1,
						file_count,
						time_step,
						global_data,
						&energy);
					
									
					if ((restartlevel==2)&&(get_terminate_flag()==AFTER_FILE)) {
						set_terminate_flag(AFTER_STEP);
					}
				}
	
				if (do_analyze) {
#if (DEBUG & 1)
txtout("Master: write out ego-file %d\n",file_count);
#endif
					writeout_coordinates(	global_data,
								time_step,
								file_count,
								version_string);
				}
			} /**** end of for every node *****/
			if (do_analyze) file_count++;	
#if (DEBUG & 8) == 8
testmalloc("after coords are comming in egmaster(...)",1);
#endif
#if (DEBUG & 1)
txtout("Master: got and anlyzed all coords\n");
#endif
		} /**** end of if coords are comming ****/

		/*** ------------------------------------------ ***/

		if ((global_data->rotation_mode) && 
		    (global_data->cycles_passed==global_data->selimob_start)) {
		  for (i=0,k=0;i<global_data->total_no_atoms;i++) {
		    if (!(global_data->atom_sele_flag2[i] & FLAG_ROT)) continue;
		    global_data->rot_x[k] = global_data->atomdata[i].ko[0];
		    global_data->rot_y[k] = global_data->atomdata[i].ko[1];
		    global_data->rot_z[k] = global_data->atomdata[i].ko[2];
		    global_data->rot1_x[k] = global_data->atomdata[i].ko[0];
		    global_data->rot1_y[k] = global_data->atomdata[i].ko[1];
		    global_data->rot1_z[k] = global_data->atomdata[i].ko[2];
		    global_data->rot2_x[k] = global_data->atomdata[i].ko[0];
		    global_data->rot2_y[k] = global_data->atomdata[i].ko[1];
		    global_data->rot2_z[k] = global_data->atomdata[i].ko[2];

		    global_data->rotation_torque[k]  = 0.0;
		    k++;
		  }
		  PCI_positions_send(global_data);
		}


		/***********************************************************/

		temperature = energy.kinetic * temp_factor;
		temperature_h2o     = energy.kinetic_h2o * temp_factor_h2o;
		temperature_protein = energy.kinetic_protein * temp_factor_protein;

#if (DEBUG & 1)
txtout("Master: after got all energy.\n");
#endif

		/*** get spent time of current integration step ***/
		/*** and reset timer again. 			***/
		stop_timer(cycletimer);
		cycletime_user = (double)get_user_timer(cycletimer);
		cycletime_real = (double)get_rt_timer(cycletimer);		
		clear_timer(cycletimer);
		start_timer(cycletimer);
		
		if (global_data->debug_flags[1]&1) {
			double sumcalc,mincalc=50000.0,maxcalc=0.0;
			txtout("          TREE1  TREE2  OTHER  EXACT  STUFF  =SUMM    COMM1  COMM2  COMENE  COMKO\n");
			for (i=0;i<global_data->no_nodes;i++) {
				sumcalc = nodetime[i][0]+nodetime[i][1]+
				  nodetime[i][2]+nodetime[i][3]+nodetime[i][4];
				if (sumcalc>maxcalc) maxcalc=sumcalc;
				if (sumcalc<mincalc) mincalc=sumcalc;
				txtout("Node %2d: %6.3f %6.3f %6.3f %6.3f %6.3f %6.3f   %6.3f %6.3f %6.3f %6.3f\n",
					i,
					nodetime[i][0],nodetime[i][1],nodetime[i][2],nodetime[i][3],
					nodetime[i][4],sumcalc,
					nodetime[i][5],nodetime[i][6],nodetime[i][7],
					nodetime[i][8]);			
			}
			txtout("Loadmiss: %6.4f\n",maxcalc-mincalc);
			loadcounter++;
			aveloadmiss += maxcalc-mincalc;
		}

		if (!miss_first) {
#if (DEBUG & 1)
txtout("Master: before print analyze\n");
#endif
			writeout_energy(&energy,
		 			temperature,temperature_h2o,temperature_protein,
		 			cycletime_user,cycletime_real,
		 			global_data,
		 			time_step);
#if (DEBUG & 1)
txtout("Master: after print analyze !miss_first\n");
#endif
			writeout_myforces(global_data,my_force);
#if (DEBUG & 1)
txtout("Master: after writeout_myforces\n");
#endif
		}



		if (	(global_data->minimisation) &&
			((global_data->minikrit[0]>0.0)||(global_data->minikrit[1]>0.0)) ) {
			int krit1 = (	(energy.gradinfo[0]<=global_data->minikrit[0]) &&
					(global_data->minikrit[1]<=0.0) );

			int krit2 = (	((energy.gradinfo[1]/global_data->total_no_atoms)<=global_data->minikrit[1]) &&
					(global_data->minikrit[0]<=0.0) );				

			int krit3 = (	((energy.gradinfo[1]/global_data->total_no_atoms)<=global_data->minikrit[1]) &&
					(energy.gradinfo[0]<=global_data->minikrit[0]) );

			if ( krit1 || krit2 || krit3 ) {
				txtout("***************************************************\n");
				txtout("Reached minimisation criterion %g  %g.\n",
					global_data->minikrit[0],global_data->minikrit[1]);
				txtout("Maximum and average gradient was %g and %g at step %d.\n",
					energy.gradinfo[0],(energy.gradinfo[1]/global_data->total_no_atoms),
					global_data->cycles_passed);
				txtout("Simulation will be finished after next restart file!\n");	
				txtout("****************************************************\n");
				set_terminate_flag(AFTER_FILE);
				global_data->minikrit[0]=-1.0;
				global_data->minikrit[1]=-1.0;
			}
		}

	/* notes on equilibration: The equilibration scheme by velocity
	 * rescaling is described in the following paper: H. J. C.
	 * Bedendsen, J. P. M. Postma, W. F. van Gunsteren, A. DiNola, J.
	 * R. Haak; "Molecular dynamics with coupling to an external
	 * bath"; J. Chem. Phys. 81(8), 15. Oct. 1984, pp 3684 - 3690; 
	 *
	 * Proportional scaling gives the least squares local disturbance
	 * satisfying a global constraint. tau_T determines how fast the
	 * rescaling reacts to temperature deviations. */


		if (temperature==0.0) temperature=global_data->target_temp;
		scaling_factor=sqrt(1.0 + 
				(global_data->target_temp/temperature-1.0)*
	       			(time_step/global_data->tau_T));
#if (DEBUG & 1)
txtout("Master: sending rescale-factor %f\n",scaling_factor);
txtout("Master: sending pulling positions (%f,%f,%f) and (%f,%f,%f,%f)\n",
					global_data->pullspringpos[0],
					global_data->pullspringpos[1],
					global_data->pullspringpos[2],
					global_data->pullspringpos[3],
					global_data->pullspringpos[4],
					global_data->pullspringpos[5],
					global_data->pullspringpos[6]);
#endif
	
		if (PCI_send_scaling(scaling_factor,global_data)==-1) break; /*** stop signal ***/
#if (DEBUG & 8) == 8
	testmalloc("after PCI_send_scaling in egmaster(...)",1);
#endif
#if (DEBUG & 1)
txtout("Master: setting new cycle\n");
#endif
		global_data->cycles_passed++;
		set_working_flags(global_data,
				&do_analyze,
				&restartlevel,
				&miss_first,
				&coords_are_coming);

		/*** Here we trigger the reclustering ***/
		/*** we dont want reclustering between writing the restart file ***/
		if (new_clustering==0) {
			if (restartlevel==2) {	/*** do it not yet, but later ***/
				new_clustering=2; 
			} else {	
				stop_timer(cycletimer);
				do_new_clustering(global_data->total_no_atoms,
					global_data,0,my_force);
				PCI_send_update(global_data);
				if (global_data->rotation_mode) {
				  PCI_positions_send(global_data);
				}
				new_clustering = global_data->treeinfo.steps;
				start_timer(cycletimer);
				global_data->cycles_passed--;
				miss_first = 1;
				do_analyze = 0;
#if (DEBUG & 8) == 8
				testmalloc("after clustering in egmaster(...)",1);
#endif
			}
		}
		new_clustering--;

	} /******* end of main while loop ********/

	close_output_file(global_data,file_count);
	
	if (global_data->debug_flags[1]&1) {
		txtout("*************************************************\n");
		txtout("Average loadmiss last %d steps was %6.4f\n",loadcounter,aveloadmiss/loadcounter);
		txtout("*************************************************\n");
	}
	
	txtout("Master: Simulation has finished now with step %d !\n",
		global_data->cycles_passed-1);
#if (DEBUG & 8) == 8
	testmalloc("after simulation on egmaster(...)",1);
#endif
	print_meminfo();

	free_timer(cycletimer);
	free_TRAVELLING_COORDINATES(&coordinates);
	free_loader_buffers(global_data);
	free_my_forces(my_force,global_data);

	txtout("Master: CleanUp ended !\n");
	return;
} /* end of master */

/*** here we initialize the energy struct which we		***/
/*** use to add up energy contributions calulated on the nodes	***/

static void
zero_energy(ENERGY_INFO *energy,int no_nodes)
{
	int i,j=no_nodes;
	energy->elec     = 0.0;
	energy->vdw      = 0.0;
	energy->bond     = 0.0;
	energy->angle    = 0.0;
	energy->dihe     = 0.0;
	energy->impr     = 0.0;
	energy->kinetic  = 0.0;
	energy->kinetic_h2o     = 0.0;
	energy->kinetic_protein = 0.0;
	energy->restraint= 0.0;
	energy->hbond    = 0.0;
	energy->flooding = 0.0;
	energy->sbound   = 0.0;
	energy->imobil   = 0.0;

	energy->gradinfo[0] = 0.0;
	energy->gradinfo[1] = 0.0;	

	for (i=0;i<QMENERGIES;i++) energy->qm_energy[i] = 0.0;
	
	/*** not only energies are in that energy but also stuff	***/
	/*** which we need for the translation/rotation correction	***/
	for (i=0;i<3;i++) {
		energy->C[i]     = 0.0;
		energy->P[i]     = 0.0;
		energy->L[i]     = 0.0;
		energy->qmftot[i]= 0.0;
		energy->mmftot[i]= 0.0;		
		energy->qmang[i] = 0.0;
		energy->mmang[i] = 0.0;
	}
	for (i=0;i<9;i++) {
	  energy->imobC[i] = 0.0;
	}


        /***for (i=0;i<7;i++) {
	  energy->pullspringpos[i]  = 0.0;
	  }***/


	/*** some stuff for pulling ***/	
	energy->pull_indir[0]  = 0.0;
	energy->pull_outdir[0] = 0.0;
	energy->pull_indir[1]  = 0.0;
	energy->pull_outdir[1] = 0.0;

	energy->rot_force[0] = 0.0;
	energy->rot_force[1] = 0.0;
	energy->rot_force[2] = 0.0;
	energy->rotation_energy= 0.0;

	for (i=0;i<NUM_TIMERS;i++) energy->time[i]      = 0.0;

	/*** rok140696   ***/
        energy->dFx = energy->dFy = energy->dFz = 0.0;
	      
	/*** rok170696   ***/
        for(i=0;i<3;i++)
          for(j=0;j<3;j++)
            energy->ekin_tensor[i][j] = 0.0;
        /*** rok191196 trace of ekin tensor ***/
        energy->ekin_trace = 0.0;

	zero_pressure(&(energy->pressure));
	zero_pressure(&(energy->density));
	
	energy->special_YP_pos = 0.0;
	energy->special_YM_pos = 0.0;
	energy->special_YP_count = 0;
	energy->special_YM_count = 0;
	
	return;
}

static void
zero_rotation(LOAD_MSG_ALL_NODE *global_data)
{
  int i;
  for (i=0;i<global_data->nr_of_rot_atoms;i++) {
    global_data->rot_t1_x[i]      = 0.0;
    global_data->rot_t1_y[i]      = 0.0;
    global_data->rot_t1_z[i]      = 0.0;
    global_data->rot_t2_x[i]      = 0.0;
    global_data->rot_t2_y[i]      = 0.0;
    global_data->rot_t2_z[i]      = 0.0;
    global_data->rotation_t_torque[i] = 0.0;
  }
  return;
}

static void 
addup_energy(ENERGY_INFO *energy,ENERGY_INFO *local_energy,int other_number)
{
	int i,j=other_number;

	energy->elec     += local_energy->elec;
	energy->vdw      += local_energy->vdw;
	energy->bond     += local_energy->bond;
	energy->angle    += local_energy->angle;
	energy->dihe     += local_energy->dihe;
	energy->impr     += local_energy->impr;
	energy->kinetic  += local_energy->kinetic;
	energy->kinetic_h2o     += local_energy->kinetic_h2o;
	energy->kinetic_protein += local_energy->kinetic_protein;
	energy->restraint+= local_energy->restraint;
	energy->hbond    += local_energy->hbond;
	energy->flooding += local_energy->flooding;
	energy->flood_energy_strength = local_energy->flood_energy_strength;
	energy->flood_energy_average  = local_energy->flood_energy_average;
	energy->flooding_distance     = local_energy->flooding_distance;
	energy->sbound   += local_energy->sbound;				

	if (local_energy->gradinfo[0]>energy->gradinfo[0]) {
		energy->gradinfo[0] = local_energy->gradinfo[0];
	}
	energy->gradinfo[1] += local_energy->gradinfo[1];

	/*** no += here since we have calculated center of mass ***/
	/*** completly on the nodes				***/
	energy->C[0]   = local_energy->C[0];
	energy->C[1]   = local_energy->C[1];
	energy->C[2]   = local_energy->C[2];

	energy->L[0]   += local_energy->L[0];
	energy->L[1]   += local_energy->L[1];
	energy->L[2]   += local_energy->L[2];

	energy->P[0]   += local_energy->P[0];
	energy->P[1]   += local_energy->P[1];
	energy->P[2]   += local_energy->P[2];

	energy->qmftot[0]  += local_energy->qmftot[0];
	energy->qmftot[1]  += local_energy->qmftot[1];
	energy->qmftot[2]  += local_energy->qmftot[2];

	energy->mmftot[0]  += local_energy->mmftot[0];
	energy->mmftot[1]  += local_energy->mmftot[1];
	energy->mmftot[2]  += local_energy->mmftot[2];
	
	energy->qmang[0]   += local_energy->qmang[0];
	energy->qmang[1]   += local_energy->qmang[1];
	energy->qmang[2]   += local_energy->qmang[2];

	energy->mmang[0]   += local_energy->mmang[0];
	energy->mmang[1]   += local_energy->mmang[1];
	energy->mmang[2]   += local_energy->mmang[2];

	energy->imobC[0]   += local_energy->imobC[0];
	energy->imobC[1]   += local_energy->imobC[1];
	energy->imobC[2]   += local_energy->imobC[2];
	energy->imobC[3]   += local_energy->imobC[3];
	energy->imobC[4]   += local_energy->imobC[4];
	energy->imobC[5]   += local_energy->imobC[5];
	energy->imobC[6]   += local_energy->imobC[6];
	energy->imobC[7]   += local_energy->imobC[7];
	energy->imobC[8]   += local_energy->imobC[8];

        /*** Total kinetic energy tensor***/
        for(i=0;i<3;i++)
          for(j=0;j<3;j++)
            energy->ekin_tensor[i][j] += local_energy->ekin_tensor[i][j];

        /*** rok191196 total trace of ekin tensor ***/
        energy->ekin_trace += local_energy->ekin_trace;
          
        /*** rok140696             ***/
        /*** Total error in force  ***/
        energy->dFx += local_energy->dFx;
        energy->dFy += local_energy->dFy;
        energy->dFz += local_energy->dFz;

	energy->pull_indir[0]    += local_energy->pull_indir[0];
	energy->pull_outdir[0]   += local_energy->pull_outdir[0];
	energy->pull_indir[1]    += local_energy->pull_indir[1];
	energy->pull_outdir[1]   += local_energy->pull_outdir[1];

	energy->rot_force[0]     += local_energy->rot_force[0];
	energy->rot_force[1]     += local_energy->rot_force[1];
	energy->rot_force[2]     += local_energy->rot_force[2];

	energy->rotation_energy  += local_energy->rotation_energy;

	/*** For the SBOUND pressure computation we have to add all forces ***/
	energy->pressure.sphere  += local_energy->pressure.sphere;
	energy->pressure.cyl_rad += local_energy->pressure.cyl_rad;
	for(i=0;i<6;i++)
	  energy->pressure.plane_planar[i] += local_energy->pressure.plane_planar[i];
	for(i=0;i<12;i++)
	  energy->pressure.plane_edge[i]   += local_energy->pressure.plane_edge[i];
	for(i=0;i<8;i++)
	  energy->pressure.plane_corner[i] += local_energy->pressure.plane_corner[i];

	/*** rok140696             ***/
	/*** For the Virial-Tensor ***/
        for(i=0;i<3;i++)
          for(j=0;j<3;j++)
            {
            energy->pressure.virial[i][j]      += local_energy->pressure.virial[i][j];
            energy->pressure.virial_ext[i][j]  += local_energy->pressure.virial_ext[i][j];
            energy->pressure.virial_ekin[i][j] += local_energy->pressure.virial_ekin[i][j];
            }
        /*** rok241096            ***/
        energy->pressure.virial_tot      += local_energy->pressure.virial_tot;
        energy->pressure.virial_ext_tot  += local_energy->pressure.virial_ext_tot;
        energy->pressure.virial_ekin_tot += local_energy->pressure.virial_ekin_tot;


	/*** For the SBOUND density computation we have to add all particle numbers ***/
	energy->density.sphere  += local_energy->density.sphere;
	energy->density.cyl_rad += local_energy->density.cyl_rad;
	for(i=0;i<6;i++)
	  energy->density.plane_planar[i] += local_energy->density.plane_planar[i];
	for(i=0;i<12;i++)
	  energy->density.plane_edge[i]   += local_energy->density.plane_edge[i];
	for(i=0;i<8;i++)
	  energy->density.plane_corner[i] += local_energy->density.plane_corner[i];
	
	energy->special_YP_pos   += local_energy->special_YP_pos;
	energy->special_YM_pos   += local_energy->special_YM_pos;
	energy->special_YP_count += local_energy->special_YP_count;
	energy->special_YM_count += local_energy->special_YM_count;
	
	return;
}

#endif

