/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: egnode.c	by M.Eichinger, Munich			**
**	Belongs2: EGO							**
**	Contents: init of worker nodes,					**	
**		  main-loop for worker nodes				**
**  	Modif's:							**
**									**
*************************************************************************/

/* DEBUG == 1 : standard debugging */
/* DEBUG == 2 : currently not used */
/* DEBUG == 4 : debug QM 	   */
/* DEBUG == 8 : malloc debug 	   */
#define DEBUG			0

#if defined(SEQ)

/**************************************************/
/*** for the sequential version this file	***/
/*** is empty. The main calculation stuff	***/
/*** of the sequential version is located	***/
/*** in egcontro.c, do_all_calculations(...) 	***/
/**************************************************/

#else 

/**************************************************/
/*** for the parallel version this file	contains***/
/*** the main loop of a slave node		***/
/**************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <signal.h>

#include "ego.h"
#include "timing.h"
#include "egalloc.h"
#include "egmach.h"
#include "egnolib.h"
#include "egintern.h"
#include "egintegr.h"
#include "egbolib.h"
#include "egtree.h"
#include "egrotflood.h"
#include "egpull.h"
#include "util_lib.h"
#include "egpressure.h"
#include "egbound.h"
#include "dbmalloc.h"	/*** include always as last include file ***/

#if defined(PARIX) || defined(PARIXHAL) || defined(MPI)
#include "egmepvm.h"
#endif

#define STEP_NEW_WWLIST		64

static	void	node_sigterm_proc(int signo);

static	int	manage_node_qminterface(int nProcs,
					LOAD_MSG_ALL_NODE *global_data,
					LOAD_MSG_ONE_NODE *my_data,
					int my_number);

static	void	gather_non_pair_atoms(	TRAVELLING_COORDINATES *coord,
					LOAD_MSG_ALL_NODE *global_data,
					LOAD_MSG_ONE_NODE *my_data,
					EXT_3_ATOM *ext_angle,
					EXT_4_ATOM *ext_dihe,
					EXT_4_ATOM *ext_impr);

static	int	get_new_data(int *other_number,int no_nodes,int *node_status);
static	int	mystart_timer(int timer);
static	int	mystop_timer(int timer);
static	int	calculate_time(ENERGY_INFO *energy);

				
/**************************************************************************/
/*** This is the main routine to be called on each node 		***/
/**************************************************************************/

static	LOAD_MSG_ONE_NODE 	my_data;
static	LOAD_MSG_ALL_NODE 	global_data;
static	TRAVELLING_PACKETS	*packet;
static	int			my_number;
static	int			*phase_status;
static	int			timerhandle[NUM_TIMERS];
static	int			do_load_output=0;

#if defined(PARIX) || defined(PARIXHAL) || defined(T3D) || defined(MPI)
int    node (void)
#else
int    main (void)
#endif
{
	int	MyProcId,nProcs;
	int	do_analyze,miss_first,restartlevel,coords_are_coming;
	int	tree_mode,steps_after_reclust,new_clustering;
	int	effort_coord,effort_branch;

	ENERGY_INFO 	local_energy,estm_local_energy,dummy_energy;

	EXT_3_ATOM 	*ext_angle;
	EXT_4_ATOM 	*ext_dihe;
	EXT_4_ATOM 	*ext_impr;

	TRAVELLING_FORCES my_new_forces;	/* not really travelling ! */
	TRAVELLING_FORCES my_exact_LRforces;

	double	scaling_factor=1.0;		/*** for equilibration ***/
  	double 	*rnd_force_std_deviation,*Gamma;/* stochastic boundary */

 	int    	my_first_atom;
	int	*node_status,other_number,i;

	ARGUMENTS_THREAD arg_send;
	ARGUMENTS_THREAD arg_recv;

	malloc_init(-2);
#if (DEBUG & 1) == 1
txtout("Node xx: Enter any node\n");
#endif

	/*** shure is shure ***/
	memset(&local_energy,0,sizeof(ENERGY_INFO));
	memset(&estm_local_energy,0,sizeof(ENERGY_INFO));
	memset(&dummy_energy,0,sizeof(ENERGY_INFO));
	memset(&global_data,0,sizeof(LOAD_MSG_ALL_NODE));
	memset(&my_data,0,sizeof(LOAD_MSG_ONE_NODE));

	MyProcId = PCI_enroll();
	PCI_node_setup(MyProcId,&nProcs,&my_number);

	malloc_init(my_number);
#if (DEBUG & 1) == 1
txtout("Node %d: node id now found with nProcs %d\n",my_number,nProcs);
#endif

	/*** receive the data local to this node ***/
	my_data.no_atoms      = 0;
	my_data.no_more_atoms = 0;
	PCI_recv_load_msg_one_node(nProcs,&my_data);

#if (DEBUG & 1) == 1
txtout("Node %d: after PCI_recv_load_msg_one_node\n",my_number);
#endif

	/*** receive the global data common to all nodes ***/
	PCI_recv_load_msg_all_node(nProcs,&global_data);
#if (DEBUG & 1) == 1
txtout("Node %d: after PCI_recv_load_msg_all_node\n",my_number);
#endif
	if (global_data.no_nodes != nProcs) {
		errout("Node %d: Mismatch nr nodes! global_data.no_nodes=%d nProcs=%d\n",
			my_number, global_data.no_nodes, nProcs);
		pex(-1);
	}

	set_signal_procs(node_sigterm_proc);

	/******************************************************************/
	/*** here we ini all our timer-stuff 				***/
	/******************************************************************/
	for (i=0;i<NUM_TIMERS;i++) {
		timerhandle[i]       = init_timer();
		local_energy.time[i] = 0.0;
		dummy_energy.time[i] = 0.0;
		estm_local_energy.time[i] = 0.0;
	}

	set_allocinfo("STUFF");
	node_status 		= alloc_ints(global_data.no_nodes);
	phase_status		= alloc_ints(global_data.no_nodes);
	set_allocinfo("PACKETS");
	packet			= alloc_packets(global_data.no_nodes);
	set_allocinfo("XXX");

	/*****************************************************************/
	/*****************************************************************/
	my_first_atom = global_data.first_atom[my_number];
#if (DEBUG & 1) == 1
txtout("Node %d: before init node my_first_atom %d\n",my_number,my_first_atom);
#endif
	init_node(	&global_data,
			&my_data,
			&my_new_forces,
			&my_exact_LRforces,
	    		my_number,
			&rnd_force_std_deviation,
			&Gamma,
	    		packet,
			&ext_angle,
			&ext_dihe,
			&ext_impr);

	for (i=0;i<global_data.no_nodes;i++) {
		node_status[i] = PACKET_NOT_READY;
	}

#if (DEBUG & 1) == 1
txtout("Node %d: after init node\n",my_number);
#endif

	miss_first        = 0;
	do_analyze	  = 1;	 
	restartlevel      = 0;
	coords_are_coming = do_analyze;
	new_clustering    = global_data.treeinfo.steps;
	steps_after_reclust = 0;
	do_load_output    = global_data.debug_flags[1]&1;

	set_allocinfo("TREE_ini_all");
	TREE_ini_all(my_number,&global_data,&my_data);
	set_allocinfo("XXX");
#if (DEBUG & 1) == 1
txtout("Node %d: after TREE_ini_all\n",my_number);
#endif

	tree_mode = 3;		/*** calc ww-list and calc new local expansion ***/

	set_communication_effort(COMEFFORT_ALL,COMEFFORT_ALL);

	PCI_start_thread_send(	&my_data,
				&global_data,
				packet,
				my_number,
				node_status,
				&arg_send);

	PCI_start_thread_recv(	&my_data,
				&global_data,
				packet,
				my_number,
				node_status,
				&arg_recv);
#if (DEBUG & 1) == 1
txtout("Node %d: after start_threads\n",my_number);
#endif
	print_meminfo();

	recv_start_node();

	local_energy.pullspringpos[0]=global_data.pullspringpos[0];
	local_energy.pullspringpos[1]=global_data.pullspringpos[1];
	local_energy.pullspringpos[2]=global_data.pullspringpos[2];


	/**************************************************************/
	/**** here starts the big cycle over all integration-steps ****/
	/**************************************************************/
	while (global_data.cycles_passed<=global_data.final_cycle_no) {
#if (DEBUG & 1) == 1
txtout("Node %d: cycles_passed=%d,final_cycle_no=%d\n",my_number,
global_data.cycles_passed,global_data.final_cycle_no);
#endif
		mystart_timer(timerhandle[ORGA_STUFF]);
#if (DEBUG & 8) == 8
		testmalloc("before init_for_each_step(...) in egnode",1);
#endif

		init_for_each_step(	&local_energy,
					&estm_local_energy,
					&dummy_energy,		
					&global_data,
					&my_data,
					&my_new_forces,
					&my_exact_LRforces,
					my_number,
					packet,
					node_status);
#if (DEBUG & 8) == 8
		testmalloc("after init_for_each_step(...) in egnode",1);
#endif
		for (i=0;i<global_data.no_nodes;i++) phase_status[i] = 0;
		mystop_timer(timerhandle[ORGA_STUFF]);

#if (DEBUG & 1) == 1
txtout("Node %d: before TREE_calc_multipol\n",my_number);
#endif

		mystart_timer(timerhandle[CALC_TREE1]);
		TREE_start_loop();
		TREE_calc_multipol(	my_number,
					&global_data,
					&(packet[my_number]));
		mystop_timer(timerhandle[CALC_TREE1]);
#if (DEBUG & 8) == 8
		testmalloc("after TREE_calc_multipol(...) in egnode",1);
#endif		
#if (DEBUG & 1) == 1
txtout("Node %d: after TREE_calc_multipol\n",my_number);
#endif
		mystart_timer(timerhandle[ORGA_STUFF]);
		{int level,toplevel;
		toplevel = my_data.levels-1;

		for (level=0;level<=toplevel;level++) {
			int nr_cl=my_data.cl_per_level[level];	
			memcpy(	my_data.cluster[level],
				packet[my_number].branch.cluster[level],
				sizeof(struct_cluster)*nr_cl);
		}
		}
		mystop_timer(timerhandle[ORGA_STUFF]);
		/*****************************************************/
		/********* now all orga-stuff ended ******************/
		/*****************************************************/

#if (DEBUG & 1) == 1
txtout("Node %d: before PCI_start_send\n",my_number);
#endif
		/************ start SEND und RECV-Threads ************/
		mystart_timer(timerhandle[COMMUN1]);
		PCI_start_send(&global_data,
				packet,
				my_number,
				node_status);
		PCI_start_recv();
		mystop_timer(timerhandle[COMMUN1]);

		/********************************************************/
		/***** do all things on your own atoms ******************/
		/********************************************************/
		mystart_timer(timerhandle[CALC_EXACT]);
		if (global_data.debug_flags[0]==1) {
#if (DEBUG & 1) == 1
txtout("Node %d: before CalcIntern_LongRange\n",my_number);
#endif
			CalcIntern_LongRange(	&dummy_energy,
						&global_data,
						&my_data,
						&my_exact_LRforces,
						my_first_atom);
		}
		mystop_timer(timerhandle[CALC_EXACT]);

#if (DEBUG & 1) == 1
txtout("Node %d: before CalcIntern_ShortRange\n",my_number);
#endif
		mystart_timer(timerhandle[CALC_OTHER]);
		CalcIntern_ShortRange(	&local_energy,
	       				&my_data,
	       				&global_data,
	       				&my_new_forces,
	       				rnd_force_std_deviation,
					Gamma);
		mystop_timer(timerhandle[CALC_OTHER]);
#if (DEBUG & 8) == 8
		testmalloc("after CalcIntern_ShortRange(...) in egnode",1);
#endif
#if (DEBUG & 1) == 1
txtout("Node %d: before loop over all pakets\n",my_number);
#endif
		/**********************************************************/
		/*** Now we have to get all atoms from the other nodes.	***/
		/*** We wait for packets (coordinates, multipoles) 	***/
		/*** from any other computation node and work on it!	***/
		/**********************************************************/
		mystart_timer(timerhandle[COMMUN2]);
		while (get_new_data(	&other_number,
					global_data.no_nodes,
					node_status)==WORK_ON_PACKET) {
			mystop_timer(timerhandle[COMMUN2]);

#if (DEBUG & 1) == 1
txtout("Node %d: now start work on packet %d\n",my_number,other_number);
#endif
			mystart_timer(timerhandle[ORGA_STUFF]);
			gather_non_pair_atoms(	&(packet[other_number].coordinates),
				 		&global_data,
						&my_data,
				 		ext_angle,
						ext_dihe,
						ext_impr);
			mystop_timer(timerhandle[ORGA_STUFF]);
#if (DEBUG & 1) == 1
txtout("Node %d: after gather_non_pair_atoms %d\n",my_number,other_number);
#endif
			mystart_timer(timerhandle[CALC_TREE1]);
			if (other_number!=my_number)		
				TREE_calc_multipol(	other_number,
							&global_data,
							&(packet[other_number]));
			mystop_timer(timerhandle[CALC_TREE1]);
 				
			mystart_timer(timerhandle[CALC_TREE2]);
			TREE_calc_LongRange(	my_number,
						&global_data,
						&my_data,
						&(packet[other_number]),
						tree_mode);
			mystop_timer(timerhandle[CALC_TREE2]);

			if (my_number!=other_number) {
				mystart_timer(timerhandle[CALC_OTHER]);
				CalcExtern_Bonds(&(packet[other_number]),
						&global_data,
						&my_data,
						my_number);
				mystop_timer(timerhandle[CALC_OTHER]);

				mystart_timer(timerhandle[CALC_EXACT]);
				if (global_data.debug_flags[0]==1)
					CalcExtern_LongRange(	&(packet[other_number]),
								&dummy_energy,
			   	   				&global_data,
								&my_data,
			      					&my_exact_LRforces,
			      					my_number);
				mystop_timer(timerhandle[CALC_EXACT]);
			}
			mystart_timer(timerhandle[COMMUN2]);
			node_status[other_number] = PACKET_DONE;
			phase_status[other_number]= 1;
		}
#if (DEBUG & 1) == 1
txtout("Node %d: after big loop\n",my_number);
#endif
		PCI_stop_recv();
		PCI_stop_send(global_data.no_nodes);
		mystop_timer(timerhandle[COMMUN2]);
#if (DEBUG & 8) == 8
		testmalloc("after big loop over nodes in egnode",1);
#endif	
		/**************************************************/
		/***** now all work on the other atoms is done ****/
		/**************************************************/

                 mystart_timer(timerhandle[ORGA_STUFF]);
		/**** for flooding and/or rotation-correction ****/
	         for (i=0;i<global_data.no_nodes;i++) { 
			merge_packet_to_buffer(	my_number,
						global_data.total_no_atoms,
						&(packet[i].coordinates));
			/**** for pulling ****/
                        merge_packet_to_buffer_pull( my_number,
                                                global_data.total_no_atoms,
                                                &(packet[i].coordinates),
                                                global_data.atomdata);

		}

		mystop_timer(timerhandle[ORGA_STUFF]);
#if (DEBUG & 8) == 8
		testmalloc("after merge_packet_to_buffer(...) in egnode",1);
#endif		
		/*** calculate ShortRange forces: angle, dihe, impr ***/
		mystart_timer(timerhandle[CALC_OTHER]);
		CalcExtern_ShortRange(	&my_data,
			   		&global_data,
			   		ext_angle,
			   		ext_dihe,
			   		ext_impr,
			   		my_number);
#if (DEBUG & 8) == 8
		testmalloc("after CalcExtern_ShortRange(...) in egnode",1);
#endif
#if (DEBUG & 1) == 1
txtout("Node %d: after external_non_pair\n",my_number);
#endif
		TREE_add_distforces(	my_number,
					&global_data,
					&my_data);

		if (global_data.nr_qmatoms>0) {
			TREE_calc_qmlocal(my_number,
					&global_data,
					&my_data);
		}

#if (DEBUG & 8) == 8
		testmalloc("after TREE_add_distforces(...) in egnode",1);
#endif
#if (DEBUG & 1) == 1
txtout("Node %d: after TREE_add_distforces\n",my_number);
#endif
		do_rot_correction(      &global_data,
					my_data.x_new,
					my_data.y_new,
					my_data.z_new,
					my_data.x_old,
					my_data.y_old,
					my_data.z_old);

#if (DEBUG & 1) == 1
txtout("Node %d: before add_flood_forces.\n",my_number);
#endif
		/***** for grubis flooding ***********************/
		add_flood_forces(	&my_data,
					&global_data,
					my_new_forces.Fx,
					my_new_forces.Fy,
					my_new_forces.Fz,
					&local_energy);
#if (DEBUG & 8) == 8
		testmalloc("after add_flood_forces(...) in egnode",1);
#endif
#if (DEBUG & 1) == 1
txtout("Node %d: before add_imob_forces.\n",my_number);
#endif
		/***** Immobilisation forces *********************/
		add_imob_forces(	&my_data,
					&global_data,
					my_new_forces.Fx,
					my_new_forces.Fy,
					my_new_forces.Fz);
#if (DEBUG & 1) == 1
txtout("Node %d: before add_pull_forces.\n",my_number);
#endif
		/***** Pulling forces ****************************/
		add_pull_forces(	my_number,
                                        &local_energy,	
					&my_data,
					&global_data,
					my_new_forces.Fx,
					my_new_forces.Fy,
					my_new_forces.Fz);

#if (DEBUG & 1) == 1
txtout("Node %d: before add_rot_forces.\n",my_number);
#endif

		add_rotation_forces(    my_number,
					&local_energy,
					&my_data,
					&global_data,
				        my_new_forces.Fx,
					my_new_forces.Fy,
					my_new_forces.Fz);

#if (DEBUG & 1) == 1
		testmalloc("after add_rotation_forces(...) in egnode",1);
#endif
  
		addup_all_forces(&local_energy,
				&my_data,
				&global_data,
				&my_new_forces);
#if (DEBUG & 8) == 8
		testmalloc("after addup_all_forces(...) in egnode",1);
#endif
#if (DEBUG & 1) == 1
txtout("Node %d: before calc_virial.\n",my_number);
#endif

                calc_virial(&local_energy,&my_data,&global_data,&my_new_forces);
#if (DEBUG & 1) == 1
txtout("Node %d: before do_newrottranscor_part1\n",my_number);
#endif
		/*** sum up some data located on nodes and send it to the master ***/
		/*** later with PCI_energy_send(...).				 ***/ 
		do_newrottranscor_part1(my_number,
					&local_energy,	
					&my_data,
					&global_data,packet);
#if (DEBUG & 8) == 8
		testmalloc("after do_newrottranscor_part1(...) in egnode",1);
#endif
		transfer_special(&local_energy); /* transfer the data of the 
						    SPECIAL pressure mode into local_energy */
		mystop_timer(timerhandle[CALC_OTHER]);
#if (DEBUG & 8) == 8
		testmalloc("after transfer_special(...) in egnode",1);
#endif	
#if (DEBUG & 1) == 1
txtout("Node %d: before send energy.\n",my_number);
#endif

		calculate_time(&local_energy);
		mystart_timer(timerhandle[COMMUN_ENE]);
		PCI_energy_send(global_data.no_nodes,
				my_number,
				&local_energy,
				&my_data,
				&global_data);
			
		/*** get summed data from the nodes ***/
		PCI_etcdata_recv(global_data.no_nodes,
				my_number,
				&local_energy,
				&my_data,
				&global_data);
			
		/*** now we have all data to make the trans/rotation correction ***/
		do_newrottranscor_part2(my_number,
					&local_energy,	
					&my_data,
					&global_data,
					packet,
					&my_new_forces);		
#if (DEBUG & 8) == 8
testmalloc("after do_newrottranscor_part2 in egnode",1);
#endif

#if (DEBUG & 1) == 1
txtout("Node %d: after do_newrottranscor_part2.\n",my_number);
#endif

		if (global_data.nr_qmatoms) {
#if (DEBUG & 1) == 1
txtout("Node %d: before recv qmforces.\n",my_number);
#endif
			manage_node_qminterface(nProcs,
					  &global_data,
			 		  &my_data,
					  my_number);
#if (DEBUG & 1) == 1
txtout("Node %d: after  recv qmforces.\n",my_number);
#endif
		}
		mystop_timer(timerhandle[COMMUN_ENE]);
#if (DEBUG & 8) == 8
		testmalloc("before do_integrate(...) in egnode",1);
#endif
		/**********************************************************/
		/*** now we have all forces and do our integration	***/
		/*** my_new_forces.F* holds now the force on each atom	***/
		/**********************************************************/
		mystart_timer(timerhandle[CALC_OTHER]);
		do_integrate(	&local_energy,
		  		&my_data,
		  		&global_data,
		  		&my_new_forces,
				Gamma,
		  		my_number);
		mystop_timer(timerhandle[CALC_OTHER]);
#if (DEBUG & 8) == 8
		testmalloc("after do_integrate(...) in egnode",1);
#endif			
		/********************************************************/
		/**** now we have coordinates for the master	      ***/
		/********************************************************/
		if (	(coords_are_coming) ||
			(new_clustering==0) || (new_clustering==1) 
			|| ((global_data.rotation_mode) && 
			 (global_data.cycles_passed==global_data.selimob_start))) {
#if (DEBUG & 1) == 1
txtout("Node %d: before send coords %d.\n",my_number,coords_are_coming);
#endif
			mystart_timer(timerhandle[COMMUN_KO]);
			PCI_coords_send(global_data.no_nodes,
					my_number,
					&my_data,
					&global_data);

			mystop_timer(timerhandle[COMMUN_KO]);
		}		

		mystart_timer(timerhandle[COMMUN_ENE]);

		if ((global_data.rotation_mode) && 
		    (global_data.cycles_passed==global_data.selimob_start)) {
		  PCI_positions_recv(global_data.no_nodes,
				     my_number,
				     &global_data);
		}

		if (PCI_recv_scaling(	my_number,
					&scaling_factor,
					&global_data)==-1) {
			txtout("Node %d: Received stop signal from master!\n",my_number);
			break;
		}
		mystop_timer(timerhandle[COMMUN_ENE]);
#if (DEBUG & 1) == 1
txtout("Node %d: got rescale-factor %f, delta-imob (1) %f %f %f.\n",
				my_number,scaling_factor,
				global_data.deltaimob[0],
				global_data.deltaimob[1],
				global_data.deltaimob[2]); 
 if (global_data.flag_selimob>=2) {
   txtout("Node %d: delta-imob (2) %f %f %f.\n",
	  my_number,
	  global_data.deltaimob[3],
	  global_data.deltaimob[4],
	  global_data.deltaimob[5]); 
 if (global_data.flag_selimob==3) {
   txtout("Node %d: delta-imob (3) %f %f %f.\n",
	  my_number,
	  global_data.deltaimob[6],
	  global_data.deltaimob[7],
	  global_data.deltaimob[8]); 
 }
 }
txtout("Node %d: got XM=%f\n",my_number, global_data.sbound_droff[0]);
txtout("Node %d: got pulling pos (%f,%f,%f), (%f,%f,%f,%f)\n", 
						my_number,
						global_data.pullspringpos[0],
						global_data.pullspringpos[1],
						global_data.pullspringpos[2],
						global_data.pullspringpos[3],
						global_data.pullspringpos[4],
						global_data.pullspringpos[5],
						global_data.pullspringpos[6]);
#endif

		if (global_data.equilibration) {
			mystart_timer(timerhandle[CALC_OTHER]);
			rescale(fabs(scaling_factor),&global_data,&my_data);
			mystop_timer(timerhandle[CALC_OTHER]);
		}
#if (DEBUG & 8) == 8
		testmalloc("after set next step in egnode",1);
#endif
		/**************************************************/
		/**** new setup for the next integration-step *****/
		/**************************************************/
#if (DEBUG & 1) == 1
txtout("Node %d: setting up new step\n",my_number);
#endif
		global_data.cycles_passed++;
		set_working_flags(&global_data,
				&do_analyze,
				&restartlevel,
				&miss_first,
				&coords_are_coming);

		/*** meaning of bits in tree_mode 			    ***/
		/*** 1.Bit: 0 -> use old ww-list,   1 -> calc new ww-list   ***/
		/*** 2.Bit: 0 -> use old local exp, 1 -> calc new localc exp***/
		get_communication_effort(&effort_coord,&effort_branch);
		effort_branch=COMEFFORT_SKIP;
		steps_after_reclust++;
		if ((steps_after_reclust%STEP_NEW_WWLIST)==0) {
			tree_mode = 3;
			effort_branch = COMEFFORT_REDUCED;
		} else {
			tree_mode = 2;	
		}
		set_communication_effort(COMEFFORT_REDUCED,effort_branch);

		if (new_clustering==0) {
testmalloc("AAAAA in egnode",1);
			if (restartlevel==2) {	/*** do it not yet, but later ***/
				new_clustering=2; 
			} else {		/*** lets do reclustering ***/	
				set_communication_effort(COMEFFORT_ALL,COMEFFORT_ALL);
#if OLDMEM
				free_LOAD_MSG_ONE_NODE(&my_data);
#endif
				PCI_recv_load_msg_one_node(nProcs,&my_data);
				PCI_recv_update(nProcs,&global_data);
				if (global_data.rotation_mode) {
				  PCI_positions_recv(global_data.no_nodes,
						     my_number,
						     &global_data);
				}

#if (DEBUG & 8) == 8
				testmalloc("after recv after clustering",1);
#endif
				reinit_node(	&global_data,
						&my_data,
						&my_new_forces,
						&my_exact_LRforces,
	    					my_number,
						rnd_force_std_deviation,
						Gamma,
						packet);

				TREE_update(my_number,&global_data,&my_data);
#if (DEBUG & 8) == 8
				testmalloc("after ini after clustering",1);
#endif
				my_first_atom  = global_data.first_atom[my_number];
				new_clustering = global_data.treeinfo.steps;
				tree_mode      = 3;
				steps_after_reclust = 0;
		  		reset_merge_buffer();
				do_analyze = 0;
				global_data.cycles_passed--;
			}
		}
		new_clustering--;
   } /*** end of one integration step: while ***/

	/********************************************************************/
	/*** it is good practice to free what we alloc'ed before we exit ****/
  	/********************************************************************/
	txtout("Ordinary stopping now node %d !\n",my_number);
#if (DEBUG & 8) == 8
	testmalloc("after stopping nodes",1);
#endif
	print_meminfo();

	TREE_exit(&global_data);
	myfree(rnd_force_std_deviation);
	myfree(Gamma);

	myfree(node_status);
	myfree(phase_status);

	for (i=0;i<global_data.no_nodes;i++) {
		free_TRAVELLING_PACKETS(&(packet[i]));
	}
	myfree(packet);

	free_TRAVELLING_FORCES(&my_new_forces);
	free_TRAVELLING_FORCES(&my_exact_LRforces);
	myfree(ext_angle);
	myfree(ext_dihe);
	myfree(ext_impr);

	free_LOAD_MSG_ONE_NODE(&my_data);
	myfree(my_data.force_total);
	myfree(my_data.force_elecvdw);
	myfree(my_data.force_bond);
	myfree(my_data.force_elec);
	myfree(my_data.force_vdw);
	myfree(my_data.force_angle);
	myfree(my_data.force_dihe);
	myfree(my_data.force_impr);
	myfree(my_data.force_restr);
	myfree(my_data.force_sbound);
	myfree(my_data.force_flood);
	myfree(my_data.force_intern);
	myfree(my_data.force_kin);

	free_global_data(&global_data);
	free_rot_flood();
        free_pull();
	free_rotation();
	myfree(my_data.inv_rotation_map);

	for (i=0;i<NUM_TIMERS;i++) {
		free_timer(timerhandle[i]);
	}

	PCI_node_cleanup();

	txtout("Node %d: CleanUp ended !\n",my_number);
	return (0);
} /* end of main */

static int
manage_node_qminterface(int nProcs,
			LOAD_MSG_ALL_NODE *global_data,
			LOAD_MSG_ONE_NODE *my_data,
			int my_number)
{
	int difstep=global_data->cycles_passed-global_data->qmstart_step;
	int await_qmresult=0;
#if (DEBUG & 4) == 4
txtout("Enter manage_node_qminterface step %d  start %d.\n",
global_data->cycles_passed,global_data->qmstart_step);
#endif	
	if (difstep<0)	return 0;

	if (	(global_data->qmstart_step>=0) &&
		(difstep==0) ) {
		await_qmresult = 1;
	} else
	if (	(global_data->qmcontin_step>0) &&
		(difstep>0) &&
		((difstep%global_data->qmcontin_step)==0) 
           ) {
		await_qmresult = 1;
	} else {
		await_qmresult = 0;
	}

	if (await_qmresult) {
		PCI_recv_qmforces(nProcs,
				  global_data,
		 		  my_data,
				  my_number);		
	}
	return 0;
}


static	int
get_new_data(int *other_number,int no_nodes,int *node_status)
{
	int	node,number_packets_done;

	do {
		number_packets_done = 0;
		for (node=0;node<no_nodes;node++) {
			if (	(node_status[node]==PACKET_FOR_CALC) &&
				(phase_status[node]==0) ) {
				*other_number = node;
				return WORK_ON_PACKET;
			} else
			if (phase_status[node]==1) {
				number_packets_done++;
			}
		}

		if (number_packets_done==no_nodes) return WORK_DONE;

		PCI_get_packet(&global_data,
				packet,
				my_number,
				node_status);

	} while (number_packets_done<no_nodes);

	return WORK_DONE;
}

static
int mystart_timer(int timer)
{
	if (do_load_output) {
		start_timer(timer);
	}
	return 0;
}

static
int mystop_timer(int timer)
{
	if (do_load_output) {
		stop_timer(timer);
	}
	return 0;
}

static int
calculate_time(ENERGY_INFO *energy)
{
	int i;
	if (do_load_output) {
		for (i=0;i<NUM_TIMERS;i++) {
			energy->time[i] = (double) get_user_timer(timerhandle[i]);
			clear_timer(timerhandle[i]);
		}
	}
	return 0;
}

static
void   gather_non_pair_atoms(	TRAVELLING_COORDINATES *coord,
				LOAD_MSG_ALL_NODE *global_data,
				LOAD_MSG_ONE_NODE *my_data,
				EXT_3_ATOM *ext_angle,
				EXT_4_ATOM *ext_dihe,
				EXT_4_ATOM *ext_impr)
{
	/* note that all atom arrays are global */
	/* and lists,pointers are also global  */
	int    ext_first = global_data->first_atom[coord->node];
	int    ext_last  = ext_first + coord->no_atoms - 1;
	int    i;

#if (DEBUG & 1) == 1
txtout("in gathernonpair: ang %d dihe %d  impr %d  first %d last %d node %d atoms %d\n",
	my_data->ext_angle_ptr,
	my_data->ext_dihe_ptr,
	my_data->ext_impr_ptr,
	ext_first,ext_last,
	coord->node,
	coord->no_atoms);
#endif
	
	/******************* determine the SBOUND range params ******************/
	define_range (coord->len,  coord->y, coord->atom_id);

	
	/******************* work through the angle-list ******************/
	for (i = 0; i < my_data->ext_angle_ptr; i++) { 

		if (inrange(my_data->ext_angle_Nr1[i],ext_first,ext_last)) {
			int    index = my_data->ext_angle_Nr1[i] - ext_first;

			ext_angle[i].a1.x = coord->x[index];
			ext_angle[i].a1.y = coord->y[index];
			ext_angle[i].a1.z = coord->z[index];
			ext_angle[i].a1.node_id = coord->node;
			ext_angle[i].a1.atom_id = coord->atom_id[index];
			ext_angle[i].a1.forceout_atom = coord->atom_sele_flag1[index];
		}

		if (inrange(my_data->ext_angle_Nr2[i],ext_first,ext_last)) {
			int    index = my_data->ext_angle_Nr2[i] - ext_first;

			ext_angle[i].a2.x = coord->x[index];
			ext_angle[i].a2.y = coord->y[index];
			ext_angle[i].a2.z = coord->z[index];
			ext_angle[i].a2.node_id = coord->node;
			ext_angle[i].a2.atom_id = coord->atom_id[index];
			ext_angle[i].a2.forceout_atom = coord->atom_sele_flag1[index];
		}

		if (inrange(my_data->ext_angle_Nr3[i],ext_first,ext_last)) {
      			int    index = my_data->ext_angle_Nr3[i] - ext_first;

			ext_angle[i].a3.x = coord->x[index];
			ext_angle[i].a3.y = coord->y[index];
			ext_angle[i].a3.z = coord->z[index];
			ext_angle[i].a3.node_id = coord->node;
			ext_angle[i].a3.atom_id = coord->atom_id[index];
			ext_angle[i].a3.forceout_atom = coord->atom_sele_flag1[index];
		}
	}

	/************* work on dihedrals *************************************/
	for (i = 0; i < my_data->ext_dihe_ptr; i++) {	/* work through the lists */
		if (inrange(my_data->ext_dihe_Nr1[i],ext_first,ext_last)) {
			int    index = my_data->ext_dihe_Nr1[i] - ext_first;

			ext_dihe[i].a1.x = coord->x[index];
			ext_dihe[i].a1.y = coord->y[index];
			ext_dihe[i].a1.z = coord->z[index];
			ext_dihe[i].a1.node_id = coord->node;
			ext_dihe[i].a1.atom_id = coord->atom_id[index];
			ext_dihe[i].a1.forceout_atom = coord->atom_sele_flag1[index];
		}

		if (inrange(my_data->ext_dihe_Nr2[i],ext_first,ext_last)) {
			int    index = my_data->ext_dihe_Nr2[i] - ext_first;

			ext_dihe[i].a2.x = coord->x[index];
			ext_dihe[i].a2.y = coord->y[index];
			ext_dihe[i].a2.z = coord->z[index];
			ext_dihe[i].a2.node_id = coord->node;
			ext_dihe[i].a2.atom_id = coord->atom_id[index];
			ext_dihe[i].a2.forceout_atom = coord->atom_sele_flag1[index];
		}

 		/* Nr3 */
		if (inrange(my_data->ext_dihe_Nr3[i],ext_first,ext_last)) {
			int    index = my_data->ext_dihe_Nr3[i] - ext_first;

			ext_dihe[i].a3.x = coord->x[index];
			ext_dihe[i].a3.y = coord->y[index];
			ext_dihe[i].a3.z = coord->z[index];
			ext_dihe[i].a3.node_id = coord->node;
			ext_dihe[i].a3.atom_id = coord->atom_id[index];
			ext_dihe[i].a3.forceout_atom = coord->atom_sele_flag1[index];
		}

		/* Nr4 */
		if (inrange(my_data->ext_dihe_Nr4[i],ext_first,ext_last)) {
			int    index = my_data->ext_dihe_Nr4[i] - ext_first;

			ext_dihe[i].a4.x = coord->x[index];
			ext_dihe[i].a4.y = coord->y[index];
			ext_dihe[i].a4.z = coord->z[index];
			ext_dihe[i].a4.node_id = coord->node;
			ext_dihe[i].a4.atom_id = coord->atom_id[index];
			ext_dihe[i].a4.forceout_atom = coord->atom_sele_flag1[index];
		}
	}

	/********** work on extern impropers  ****************************/
	for (i = 0; i < my_data->ext_impr_ptr; i++) {
 
		if (inrange(my_data->ext_impr_Nr1[i],ext_first,ext_last)) {
			int    index = my_data->ext_impr_Nr1[i] - ext_first;

			ext_impr[i].a1.x = coord->x[index];
			ext_impr[i].a1.y = coord->y[index];
			ext_impr[i].a1.z = coord->z[index];
			ext_impr[i].a1.node_id = coord->node;
			ext_impr[i].a1.atom_id = coord->atom_id[index];
			ext_impr[i].a1.forceout_atom = coord->atom_sele_flag1[index];
		}

		if (inrange(my_data->ext_impr_Nr2[i],ext_first,ext_last)) {
			int    index = my_data->ext_impr_Nr2[i] - ext_first;

			ext_impr[i].a2.x = coord->x[index];
			ext_impr[i].a2.y = coord->y[index];
			ext_impr[i].a2.z = coord->z[index];
			ext_impr[i].a2.node_id = coord->node;
			ext_impr[i].a2.atom_id = coord->atom_id[index];
			ext_impr[i].a2.forceout_atom = coord->atom_sele_flag1[index];
		}
 		/* Nr3 */
		if (inrange(my_data->ext_impr_Nr3[i],ext_first,ext_last)) {
			int    index = my_data->ext_impr_Nr3[i] - ext_first;

			ext_impr[i].a3.x = coord->x[index];
			ext_impr[i].a3.y = coord->y[index];
			ext_impr[i].a3.z = coord->z[index];
			ext_impr[i].a3.node_id = coord->node;
			ext_impr[i].a3.atom_id = coord->atom_id[index];
			ext_impr[i].a3.forceout_atom = coord->atom_sele_flag1[index];
		}

        	/* Nr4 */
		if (inrange(my_data->ext_impr_Nr4[i],ext_first,ext_last)) {
			int    index = my_data->ext_impr_Nr4[i] - ext_first;

			ext_impr[i].a4.x = coord->x[index];
			ext_impr[i].a4.y = coord->y[index];
			ext_impr[i].a4.z = coord->z[index];
			ext_impr[i].a4.node_id = coord->node;
			ext_impr[i].a4.atom_id = coord->atom_id[index];
			ext_impr[i].a4.forceout_atom = coord->atom_sele_flag1[index];
		}
	}
}


static
void	node_sigterm_proc(int signo)
{
#if defined(SEQ_NANO) || defined(PARIX) || defined(PARIXHAL) || defined(HPPA)

#else
	char strtime[100];
	get_time_string(strtime);
	
	/* only the following two signals can be used! Using any others may break MPI routines!! */
	switch (signo) {
		case SIGUSR1: 
			errout("Node %d: EGO node process ignores signal SIGUSR1 (%d) at %s!\n",
				my_number,signo,strtime);
			errout("Instead we should forward it to the master.\n");
			break;
		case SIGUSR2: 
			errout("Node %d: EGO node process ignores signal SIGUSR1 (%d) at %s!\n",
				my_number,signo,strtime);
			errout("Instead we should forward it to the master.\n");
			break;
		default:
			errout("Node %d: EGO node process ignores unknown signal in node_sigterm_proc (%d) at %s!\n",
				my_number,signo,strtime);
			break;
	}
	return;
#endif
}


#if !(defined(PARIX) || defined(PARIXHAL) || defined(T3D) || defined(MPI))

/*** for PVM case there are two executables (ego and node), so we need 	***/
/*** a separate pex on the node here.					***/
void   pex(int i)
{
	txtout("Exit by pex(%d) in egnode.c !\n",i);
	PCI_exit(i);
	exit(i);
	return;
}

int    controller(char *ctlname, int host_tid, int nProcs)
{
	return 0;	
}

int	get_terminate_flag()
{
	return DONT_STOP;	
}
#endif	/* special PVM pex() definition */

#endif
