/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#if defined(LINUX) || defined(LINUX64)
#include <sys/types.h>
#include <unistd.h>
#endif

#if defined(NEXT)
#include <sys/file.h>
#include <libc.h>
#endif

#include "ego.h"
#include "version.h"
#include "egalloc.h"
#include "util_lib.h"
#include "timing.h"
#include "egloader.h"
#include "eganaly.h"
#include "egqminter.h"

#if defined(SEQ) 
#include "egcontro.h"
#else
#include "egmach.h"
void   master(	LOAD_MSG_ALL_NODE *global_data,
		double time_step,
		int file_count,
		char *version_string,
		STRUCT_FORCES *my_force);
void   node(void);
#endif

#define DEBUG 0

static	int	controller(char *ctlname,int host_tid,int nProcs);

static	char	restartcmd[MAX_STRING_SIZE];
#if defined(DOS)
static	char	defpath[10]=".";
#endif


int    main(int argc, char *argv[]) /*### the beginning of the EGO program ###*/
{
	char ctl_file[MAX_STRING_SIZE],*tt;
	char strtime[100];
		
#if defined(SEQ) /************* sequential setup ****************/

	get_time_string(strtime);
	txtout("Sequential EGO started at %s.\n",strtime);

	malloc_init(-1);

#else		/************** parallel setup ****************/
	int    MyProcID=0,MasterID=0;
	int    nProcs  =1;

	get_time_string(strtime);
	txtout("Parallel EGO started at %s.\n",strtime);
	
	malloc_init(-1);

#if (defined(T3D) && defined(DEBUG)) && !defined(VPP)
	setvbuf(stdout,NULL,_IONBF,(size_t)0);
#endif
	MasterID = PCI_start(&MyProcID,&nProcs,argc,argv);
	if (MyProcID==MasterID) { /*** parse arguments if you are the master ***/
#endif		
		/************ common block for sequential and parallel ****/
		/************ parsing of the command line *****************/
		if (argc>3) {
			errout("Usage of EGO: %s [<control_file_name> [<restart_command>]]\n",argv[0]);
			errout("Default for <control_file_name> is ctl.lis.\n");
			errout("The <restart_command> (e.g. a UNIX command (e.g. to send an e-mail) or a script)\n");
			errout("is executed if the simulation is not yet finished,\n");
			errout("but EGO has to stop for some reason (time limit or some error).\n");
			exit(-1);
		}

		/*** get the current path ***/
		tt = getenv("PWD");
		if (tt==NULL) {
#if defined(DOS)
			tt=defpath;
#else
			errout("Fatal error: getenv(char *str) does not work.\n");
			errout("You have to define the environment variable PWD. See readme.txt!\n");
			exit(-1);
#endif
		}
		strcpy(ctl_file,tt); strcat(ctl_file,"/");
		
		/*** set the first argument (control-filename) to default ***/
		/*** or use the file name given in argv[1] 		  ***/
		if (argc==1) {
			strcat(ctl_file,"ctl.lis");
		} else {
			if (argv[1][0] == '/') ctl_file[0] = '\0';
			strcat(ctl_file,argv[1]);
		}

		/*** look for the second argument in the command line ***/
		if (argc==3) {
			strcpy(restartcmd,argv[2]);
		} else	restartcmd[0]=0;
	
	/************ now we start the program ***********/
#if defined(SEQ)
	controller(ctl_file,0,1);
#else
	} /*** end of parse arguments if you are the master ***/


#if defined(PARIX) || defined(PARIXHAL) || defined(T3D) || defined(MPI) || defined(VPP)
	/*** for that cases there is only one executable for EGO    ***/
	/*** which includes the loop of the master and the slaves   ***/
	/*** So we call the function controller(...) for the master ***/
	/*** and node() for each slave node depending on MyProcID   ***/
	if (MyProcID==MasterID)
		controller(ctl_file,MyProcID,nProcs);
	else
		node();
#else
	/*** for PVM case there are two executables "ego" and "node" ***/
	/*** now we are in the "ego" executable. Later in egloader.c ***/
	/*** the "node" executable is started for each node (see also***/
	/*** PCI_master_setup(...) )				     ***/
	controller(ctl_file,MyProcID,nProcs);
#endif

	/*** after the simulation has finished or stopped we perform an exit***/
	PCI_exit(0);
#endif

	get_time_string(strtime);
	txtout("EGO stopped at %s!\n",strtime);
	return 0;
}


/****************************************************************************/
static
int controller(char *ctlname,int host_tid,int nProcs) /*### initialisation of EGO on the master ###*/
{
	LOAD_MSG_ALL_NODE	global_data;
	STRUCT_FORCES		my_force;

	double 	time_step;	/* this is the time step in seconds, not
				 * to be mixed up with global_data.time_step
				 * = time_step/TIMEDIV for use by the
				 * nodes in the integration step */
	int    	file_count;
	char	version_string[MAX_STRING_SIZE];

	memset(&global_data,0,sizeof(global_data));
	memset(&my_force,0,sizeof(my_force));
	
	set_version_string(version_string);

	global_data.cycles_passed = 0;
	strcpy(global_data.ctlfile,ctlname);
	
	txtout("\n%s   A Molecular Dynamics Program\n",VERSION);
	txtout("%s\n",EGO_MACHINE);
	txtout("%s\n",version_string);
	txtout("==========================================================\n\n");

	txtout("ProcessID: %u\n",getpid());

	/******************************************************************/
	/*** loader(...) does almost every initialization !		***/
	/*** Reads the ctl.lis file, sets global_data and my_data.	***/
	/*** For parallel EGO data is also distributed to the nodes	***/
	/******************************************************************/
#if DEBUG
errout("Master: starting loader\n");
#endif
	loader(	&global_data,
		&time_step,
		&file_count,
		host_tid,
		nProcs,
		&my_force);
#if DEBUG
errout("Master: finished loader\n");
#endif

		 /***********************************************/		
#if defined(SEQ) /******* Sequential Version ********************/
		 /***********************************************/
	seq_master(&global_data,
		time_step,
		file_count,
		version_string,
		&my_force);

#else		 /***********************************************/
	         /******* Parallel Version: see egmaster.c ******/
		 /***********************************************/
	master(	&global_data,
		time_step,
		file_count,
		version_string,
		&my_force);

#endif

	/**************************************************/
	/*** Program finished: make the clean up 	***/
	/**************************************************/
	free_global_data(&global_data);
	return 0;
}



void	master_sigterm_proc(int signo)
{
#if defined(SEQ_NANO) || defined(PARIX) || defined(PARIXHAL) || defined(HPPA)

#else
	char strtime[100];
	get_time_string(strtime);
	
	/* only the following two signals can be used! Using any others may break MPI routines!! */
	switch (signo) {
		case SIGUSR1: /* This is the only signal which we really react to */
			errout("EGO master detected signal SIGUSR1 (%d) at %s!\n",signo,strtime);
			errout("Program stops after next restart file!\n");	
			set_terminate_flag(AFTER_FILE);
			break;
		case SIGUSR2: 
			errout("Terminating the program after integration step is finished!\n");
			errout("No restart file is written at this stop.\n");
			set_terminate_flag(AFTER_STEP);
			break;
		default:
			errout("EGO master detected unknown signal in master_sigterm_proc (%d) at %s!\n",signo,strtime);
			break;
	}
	
	return;
#endif
}

static	int terminate_flag=DONT_STOP;

void set_terminate_flag(int flag)
{
	terminate_flag=flag;
}

int get_terminate_flag()
{
	return terminate_flag;
}

/*** pex is the abbreviation for normal "Program EXit" ***/
void   pex(int i)
{
	char strtime[100];
	
	txtout("Exit by pex(%d) in ego.c !\n",i);
	close_output_file((LOAD_MSG_ALL_NODE*) NULL,0);
	finish_qmcalculation();

#if defined(SEQ)
	close_extern_dynamic();
#else
	PCI_exit(i);	/* end MPI/PVM communication normally, free buffers, etc. */
#endif
	/*** manage the restart ***/
	if ( (i!=-1) && (restartcmd[0]!=0) ) {
		txtout("Restarting EGO with command <%s>\n",restartcmd);
		system(restartcmd);
	}

	get_time_string(strtime);
	txtout("EGO stopped at %s!\n",strtime);
	exit(i);
	return;
}

