/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/**************************************************************************/
/***    File: egqminter.c       by Markus Eichinger, Munich             ***/
/***    Here most of the QM interface between EGO and CPMD is located   ***/
/***                                                                    ***/
/**************************************************************************/

/*** if fork(..) does not work for starting cpmd then one can use ***/
/*** a system(...) call instead. In that case set NOFORK to 1***/
#define NOFORK          0

#define DEBUG           0
#define DEBUGEXCL       0
#define DEBUG_LINK      0

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <sys/types.h>
#include <unistd.h>
#if defined(NEXT)
#include <sys/file.h>
#include <libc.h>
#endif

#include "ego.h"
#include "util_lib.h"
#include "egqminter.h"
#include "egcontro.h"
#include "egtree.h"
#if defined(SEQ)
#else
#include "egmach.h"
#endif
#include "timing.h"

/*****
#if defined(PARIX)
static int      sleep(int a) { return 0; }
#endif
*****/

static int
parse_qmcontrol(char *keystr,char *selstr,
                STRUCT_ATOMDATA *atom_data,
                LOAD_MSG_ALL_NODE *global_data);

static int
parse_qmlink(   char *keystr,
                char *selstr,
                STRUCT_ATOMDATA *atom_data,
                LOAD_MSG_ALL_NODE *global_data,
                STRUCT_QM_INFO *finfo,
                int *forceout_atoms);

static int      manage_qminterface(     LOAD_MSG_ALL_NODE *global_data,
                                        STRUCT_FORCES *my_force,
                                        int skipcounter);

static int open_qmtraj(LOAD_MSG_ALL_NODE *global_data);
static int close_qmtraj();

static int prepare_qminterface( LOAD_MSG_ALL_NODE *global_data,
                                STRUCT_QM_INFO *finfo);

static int readin_qmresult(     LOAD_MSG_ALL_NODE *global_data,
                                STRUCT_QM_INFO *finfo);

static int patch_qm_template(   LOAD_MSG_ALL_NODE *global_data,
                                STRUCT_QM_INFO *finfo,int add_elec);

static int patch_COLUMBUS(      FILE *htempl,FILE *hout,
                                LOAD_MSG_ALL_NODE *global_data,
                                STRUCT_QM_INFO *finfo,int add_elec);

static int patch_CPMD(          FILE *htempl,FILE *hout,
                                LOAD_MSG_ALL_NODE *global_data,
                                STRUCT_QM_INFO *finfo,int add_elec);

static int gather_qmatoms(      FILE *hout,FILE *htempl,char *sdum,
                                LOAD_MSG_ALL_NODE *global_data,
                                STRUCT_QM_INFO *finfo,int firsttime);

static int writeout_qmelec(     FILE *forceh,
                                LOAD_MSG_ALL_NODE *global_data,
                                STRUCT_QM_INFO *finfo);

static int touch_file(char *fname,char *fpath);
static int remove_file(char *fname,char *fpath);

/*
static int adapt_qmcharges(LOAD_MSG_ALL_NODE *global_data,
                STRUCT_QM_INFO *finfo);
*/

static int      set_QQL_anglefactor(int *atomids, double factor);
static double   get_QQL_anglefactor(int *atomids);

static int      set_bondfactor(int *atomids, double factor);
static double   get_bondfactor(int *atomids);

static int      set_qmrotmats(double wx, double wy, double wz);
static int      rotforw(double *ko,double *res);
static int      rotback(double *ko,double *res);
static int      rothess(double hess[3][3]);

#define FN_QMSTART      "QMSTART"
#define FN_QMCONTINUE   "QMCONTINUE"
#define FN_EGOCONTINUE  "EGOCONTINUE"
#define FN_QMEXIT       "EXIT"
#define FN_EGOEXIT      "EGOEXIT"
#define FN_QMOUT        "qmoutput.out"

#define QMLOCALDIST     4.7
#define CHARGEDOFFSET   3.5

#define MAXQMEXCL       500
#define QMWARNINGDIST   1.99
#define FLAG_ISLINKATOM 0x00000001

#define WITHOUT_ELEC    0
#define WITH_ELEC       1

#define SEC_NO          0
#define SEC_CPMD        1
#define SEC_SYSTEM      2
#define SEC_ATOMS       3
#define SEC_OTHER       10

/*** multiply-factor to convert from Angstroem to atomic units (bohr)   ***/
#define ANGS_TO_BOHR    1.88972688

/*** multiply-factor to convert from kcal/mol to Hartree = 2Ry          ***/
#define KCAL_TO_HART    0.001594153

/*** multiply-factor to convert forces in atomic units (Hartree/bohr)   ***/
/*** to kcal/mol A and to kcal/mol A^2                                  ***/
#define FORCE_AU_TO_KCAL 1185.41
#define HESS_AU_TO_KCAL  2239.5566

static  FILE*   hqmtraj;
static  int     qmtraj_here=0;
static  int     curqmmap=0;
static  int     cpmd_started=0;
static  char    qmoutpath[MAX_STRING_SIZE];
static  char    fncarparrun[MAX_STRING_SIZE];
static  char    fnqmtraj[MAX_STRING_SIZE];
static  char    fndipol[MAX_STRING_SIZE];

static  double  qmenecor=0.0;
static  int     do_qm_restart=0;

#define CPMD_MODE       0
#define COLUMBUS_MODE   1
static  int     interfacemode=CPMD_MODE;

int
parse_qminterface(      char *keywordstr,char* reststr,
                        STRUCT_ATOMDATA *atom_data,
                        int *forceout_atoms,
                        LOAD_MSG_ALL_NODE *global_data,
                        STRUCT_FORCES *my_force,
                        int     *used_atoms)
{
        int i,j,len=strlen(reststr);

        /*** QMLINMODE ********************/
        if (strcmp(keywordstr,"QMROT")==0) {
                double wx,wy,wz;
                if (sscanf(reststr,"%lf %lf %lf",&wx,&wy,&wz)!=3) {
                        errout("Fatal error: Invalid format for QMROT\n");
                        pex(-1);
                }
                set_qmrotmats(wx,wy,wz);
                txtout("Rotating QM-Fragment (wx=%f, wy=%f, wz=%f)\n",wx,wy,wz);
                return 1;
        }

        /*** QMLINMODE ********************/
        if (strcmp(keywordstr,"QMLINKMODE")==0) {
                if ( (strcmp(reststr,"ACTIVEPOS")==0) ||
                     (strcmp(reststr,"SPLAM")==0) ) {
                        global_data->qmlinkmode=QMLINKMODE_ACTIVEPOS;
                        txtout("Using SPLAM link atom model.\n");
                        return 1;
                }
                if (strcmp(reststr,"FIXEDPOS")==0) {
                        global_data->qmlinkmode=QMLINKMODE_FIXEDPOS;
                        txtout("Using FIXEDPOS link atom model.\n");
                        return 1;
                }
                if ( (strcmp(reststr,"MINIPOS")==0) ||
                     (strcmp(reststr,"CLAMP")==0) ) {
                        global_data->qmlinkmode=QMLINKMODE_MINIPOS;
                        txtout("Using CLAMP link atom model.\n");
                        return 1;
                }
                if ( (strcmp(reststr,"BAKOWIES")==0) ||
                     (strcmp(reststr,"LAMP")==0) ){
                        global_data->qmlinkmode=QMLINKMODE_BAKOWIES;
                        txtout("Using LAMP link atom model.\n");
                        return 1;
                }
                errout("Fatal error: Only SPLAM, FIXEDPOS, CLAMP or\nLAMP keyword for QMLINKMODE is allowed.\n");
                pex(-1);
        }

        if (strcmp(keywordstr,"QMCHARGEMODE")==0) {
                if (strcmp(reststr,"EGO")==0) {
                        global_data->qmchargemode  = QMCHARGEMODE_FIXED;
                        txtout("Using QMCHARGEMODE FIXED.\n");
                        return 1;
                }
                if (strcmp(reststr,"QM")==0) {
                        global_data->qmchargemode  = QMCHARGEMODE_VARIABLE;
                        txtout("Using QMCHARGEMODE QM.\n");
                        return 1;
                }
                errout("Fatal error: Only EGO or QM keyword for QMCHARGEMODE is allowed.\n");
                pex(-1);
        }

        if (strcmp(keywordstr,"QMCHARGE")==0) {
                if (sscanf(reststr,"%lf",&global_data->qmcharge)!=1) {
                        errout("Fatal error: No value for QMCHARGE specified.\n");
                        pex(-1);
                }
                txtout("Total charge of the QM-Fragment is %f\n",global_data->qmcharge);
                return 1;
        }
        if (    (strcmp(keywordstr,"QMLINK")==0)||
                (strcmp(keywordstr,"QMNOLINK")==0) ) {
                parse_qmlink(   keywordstr,
                                reststr,
                                atom_data,
                                global_data,
                                &(my_force->qm),
                                forceout_atoms);
                return 1;
        }

        if (strcmp(keywordstr,"QMSIGMA")==0) {
                double sigma=0.0;

#if PRECISION == DOUBLE
                sscanf(reststr,"%lf",&sigma);
#else
                sscanf(reststr,"%f",&sigma);
#endif
                i = 0;
                for (j=i;j<i+len;j++) {
                        if ((reststr[j]==' ')||(reststr[j]=='\t')) break;
                }
                i = j;
                for (j=i;j<i+len;j++)
                        if ((reststr[j]!=' ')&&(reststr[j]!='\t')) break;

                for (i=0;i<global_data->total_no_atoms;i++) used_atoms[i] = 0;
                if (set_used_atoms(reststr+j,global_data->total_no_atoms,atom_data,used_atoms)<0) {
                        errout("Fatal error: Correct this error in out control file (QMSIGMA)\n");
                        pex(-1);
                }
                for (i=0,j=0;i<global_data->total_no_atoms;i++) {
                        if (used_atoms[i]) {
                                j++;
                                my_force->qm.asize[i] = sigma;
                        }
                }
                txtout("Sigma=%f assigned to %d atoms.\n",sigma,j);
                return 1;
        }

        if (    (strcmp(keywordstr,"QMSTART")==0)       ||
                (strcmp(keywordstr,"QMCONTINUE")==0)    ||
                (strcmp(keywordstr,"QMEXTENDBOX")==0)   ||
                (strcmp(keywordstr,"QMNOEXTELEC")==0) ||
                (strcmp(keywordstr,"QMFIXBOX")==0)    ||
                (strcmp(keywordstr,"QMFIXBOXB")==0)   ||
                (strcmp(keywordstr,"QMATOMFORCE")==0) ||
                (strcmp(keywordstr,"QMCUTOFF")==0)    ||
                (strcmp(keywordstr,"EXTDYN")==0)        ||
                (strcmp(keywordstr,"QMEXCL")==0)        ||
                (strcmp(keywordstr,"QMANGLE")==0)       ||
                (strcmp(keywordstr,"QMBOND")==0)        ||
                (strcmp(keywordstr,"QMENERGYOFFSET")==0) ||
                (strcmp(keywordstr,"QMRESTART")==0)
         ) {
                parse_qmcontrol(keywordstr,reststr,
                                atom_data,
                                global_data);
                return 1;
        }

        return 0;
}

int
set_qmboundbox( int *forceout_atoms,
                STRUCT_ATOMDATA *atom_data,
                LOAD_MSG_ALL_NODE *global_data)
{
        int i;

        if (global_data->nr_qmatoms==0) return 0;

        /**********************************************************/
        /*** Get the bounding box of the QM atoms and store it  ***/
        /*** in global_data->qmbox                              ***/
        /**********************************************************/
        start_bbox(global_data->qmbox);
        for (i=0;i<global_data->total_no_atoms;i++) {
                if (forceout_atoms[i]&FLAG_QMATOM) {
                        double rotko[3];
                        rotforw(atom_data[i].ko,rotko);
                        add_atom2bbox(  rotko[0],
                                        rotko[1],
                                        rotko[2],
                                        global_data->qmbox);
                }
        }

        if (!global_data->qmfixbox) {
                /*** in that case global_data->qmboxext holds an offset ***/
                /*** which should be added to global_data->qmbox in     ***/
                /*** order to get the the QM-simulation volume          ***/

                for (i=0;i<6;i++) {
                        global_data->qmboxext[i] += global_data->qmbox[i];
                }
        }
        /*** otherwise global_data->qmboxext is set through the command ***/
        /*** QMFIXBOX                                                   ***/

        txtout("QM bounding box:  X: %# 7.4g \t %# 7.4g \t dx= %# 7.4g\n",
                global_data->qmbox[0],global_data->qmbox[1],
                global_data->qmbox[1]-global_data->qmbox[0]);

        txtout("                  Y: %# 7.4g \t %# 7.4g \t dy= %# 7.4g\n",
                global_data->qmbox[2],global_data->qmbox[3],
                global_data->qmbox[3]-global_data->qmbox[2]);

        txtout("                  Z: %# 7.4g \t %# 7.4g \t dz= %# 7.4g\n",
                global_data->qmbox[4],global_data->qmbox[5],
                global_data->qmbox[5]-global_data->qmbox[4]);

        txtout("QM simulation box:X: %# 7.4g \t %# 7.4g \t dx= %# 7.4g\n",
                global_data->qmboxext[0],global_data->qmboxext[1],
                global_data->qmboxext[1]-global_data->qmboxext[0]);

        txtout("                  Y: %# 7.4g \t %# 7.4g \t dy= %# 7.4g\n",
                global_data->qmboxext[2],global_data->qmboxext[3],
                global_data->qmboxext[3]-global_data->qmboxext[2]);

        txtout("                  Z: %# 7.4g \t %# 7.4g \t dz= %# 7.4g\n",
                global_data->qmboxext[4],global_data->qmboxext[5],
                global_data->qmboxext[5]-global_data->qmboxext[4]);

        return 0;
}

static int
parse_qmlink(   char *keystr,
                char *selstr,
                STRUCT_ATOMDATA *atom_data,
                LOAD_MSG_ALL_NODE *global_data,
                STRUCT_QM_INFO *finfo,
                int *forceout_atoms)
{
        int     atomid1=0,atomid2=0;
        double  linklen1=0.0,linklen2=0.0,forcec1=0.0,forcec2=0.0;

        if (strcmp(keystr,"QMNOLINK")==0) {
#if PRECISION == DOUBLE
                if (sscanf(selstr,"%d %d",&atomid1,&atomid2)!=2)
#else
                if (sscanf(selstr,"%d %d",&atomid1,&atomid2)!=2)
#endif
                        goto illegal_qmlink_sel;

                if ((atomid1<=0)||(atomid1>global_data->total_no_atoms))
                        goto illegal_qmlink_sel;

                if ((atomid2<=0)||(atomid2>global_data->total_no_atoms))
                        goto illegal_qmlink_sel;

                txtout("No Link-atom between %d %s (%d %s) and atom %d %s (%d %s).\n",
                        atomid2,atom_data[atomid2-1].atomstr,atom_data[atomid2-1].residnr,atom_data[atomid2-1].residstr,
                        atomid1,atom_data[atomid1-1].atomstr,atom_data[atomid1-1].residnr,atom_data[atomid1-1].residstr);
                atomid1--;      /*** EGO-numbering of atoms starts with 0 ***/
                atomid2--;


                finfo->qmlink_p1[ finfo->qmlink_len ]     = atomid1;
                finfo->qmlink_p2[ finfo->qmlink_len ]     = atomid2;
                finfo->qmlink_dist1[ finfo->qmlink_len ]  = -1.0;
                finfo->qmlink_dist2[ finfo->qmlink_len ]  = -1.0;
                finfo->qmlink_fconst1[ finfo->qmlink_len ]= 0.0;
                finfo->qmlink_fconst2[ finfo->qmlink_len ]= 0.0;
                finfo->qmlink_len++;
                forceout_atoms[atomid2] |= FLAG_QMLINK; /*** mark it as a link atom ***/

                return 0;
        }

        /**************************************************/
        /***  Check/Scan the selection string           ***/
        /**************************************************/
#if PRECISION == DOUBLE
        if (sscanf(selstr,"%d %d %lf %lf %lf %lf",&atomid1,&atomid2,
                &linklen1,&forcec1,&linklen2,&forcec2)!=6)
#else
        if (sscanf(selstr,"%d %d %f %f %f %f",&atomid1,&atomid2,
                &linklen1,&forcec1,&linklen2,&forcec2)!=6)
#endif
                goto illegal_qmlink_sel;

        if ((atomid1<=0)||(atomid1>global_data->total_no_atoms))
                goto illegal_qmlink_sel;

        if ((atomid2<=0)||(atomid2>global_data->total_no_atoms))
                goto illegal_qmlink_sel;

        if ((linklen1<0.4)||(linklen1>3.0)) goto illegal_qmlink_sel;

        txtout("Atom %d %s in residue %d %s selected for QM-link atom with atom %d.\n",
                atomid2,atom_data[atomid2-1].atomstr,
                atom_data[atomid2-1].residnr,atom_data[atomid2-1].residstr,
                atomid1);
        atomid1--;      /*** EGO-numbering of atoms starts with 0 ***/
        atomid2--;

        finfo->qmlink_p1[ finfo->qmlink_len ]     = atomid1;
        finfo->qmlink_p2[ finfo->qmlink_len ]     = atomid2;
        finfo->qmlink_dist1[ finfo->qmlink_len ]  = linklen1;
        finfo->qmlink_dist2[ finfo->qmlink_len ]  = linklen2;
        finfo->qmlink_fconst1[ finfo->qmlink_len ]= forcec1;
        finfo->qmlink_fconst2[ finfo->qmlink_len ]= forcec2;
        finfo->qmlink_len++;
        forceout_atoms[atomid2] |= FLAG_QMLINK; /*** mark it as a link atom ***/

        return 0;

illegal_qmlink_sel:
        errout("Fatal error: Illegal selection string <%s> for QM-link atoms!\n",
                selstr);
        pex(-1);
        return -1;
}

/**************************************************/
/*** parses some QM-parameters in forceout.lis  ***/
/**************************************************/
static int
parse_qmcontrol(char *keystr,char *selstr,
                STRUCT_ATOMDATA *atom_data,
                LOAD_MSG_ALL_NODE *global_data)
{
        int     value=0,i=atom_data[0].atomnr;
        int     len=strlen(selstr);
        double  dvalue;

        /*** truncate trailing white spaces ***/
        for (i=len-1;i>=0;i--) {
                if ((selstr[i]==' ')||(selstr[i]=='\t')) {
                        selstr[i]=0;
                } else break;
        }

        if (strcmp(keystr,"QMRESTART")==0) {
                do_qm_restart = 1;
                if (strcmp(selstr,"LATEST")==0) do_qm_restart = 2;
                return 0;
        } else
        if (strcmp(keystr,"QMNOEXTELEC")==0) {
                global_data->qmnoextelec = 1;
                return 0;
        } else
        if (strcmp(keystr,"QMEXCL")==0) {
                int res;
                double  vals[2];
#if PRECISION == DOUBLE
                res = sscanf(selstr,"%le %le",
#else
                res = sscanf(selstr,"%e %e",
#endif
                                &vals[0],
                                &vals[1]);
                if ((vals[0]<0.0) ||(vals[1]<0.0))  goto illegal_qmcontrol;
                global_data->qmexcl[0] = vals[0];
                global_data->qmexcl[1] = vals[1];

                txtout("Setting QM-exclusion function to dist=%f width=%f\n",
                        global_data->qmexcl[0],global_data->qmexcl[1]);

                return 0;
        } else
        if (strcmp(keystr,"QMANGLE")==0) {
                int atid[3],res;
                double  vals[2];
#if PRECISION == DOUBLE
                res = sscanf(selstr,"%d %d %d %le",
#else
                res = sscanf(selstr,"%d %d %d %e",
#endif
                                &atid[0],&atid[1],&atid[2],
                                &vals[0]);

                if (atid[0]<1)    goto illegal_qmcontrol;
                if (atid[1]<1)    goto illegal_qmcontrol;
                if (atid[2]<1)    goto illegal_qmcontrol;

                atid[0]--;
                atid[1]--;
                atid[2]--;
                set_QQL_anglefactor(atid,vals[0]);

                txtout("Setting QM-Angle-Patch %d %d %d with %f\n",
                        atid[0],atid[1],atid[2],vals[0]);

                return 0;
        } else
        if (strcmp(keystr,"QMBOND")==0) {
                int atid[3],res;
                double  vals[2];
#if PRECISION == DOUBLE
                res = sscanf(selstr,"%d %d %le",
#else
                res = sscanf(selstr,"%d %d %e",
#endif
                                &atid[0],&atid[1],
                                &vals[0]);

                if (atid[0]<1)    goto illegal_qmcontrol;
                if (atid[1]<1)    goto illegal_qmcontrol;

                atid[0]--;
                atid[1]--;
                set_bondfactor(atid,vals[0]);

                txtout("Setting QM-Bond-Patch %d %d with %f\n",
                        atid[0],atid[1],vals[0]);

                return 0;
        } else
        if (    (strcmp(keystr,"QMEXTENDBOX")==0)  ||
                (strcmp(keystr,"QMEXTENDBOXB")==0) ||
                (strcmp(keystr,"QMFIXBOX")==0)     ||
                (strcmp(keystr,"QMFIXBOXB")==0) ) {
                int res=0,different=0,firsttime=1;
                double vals[6];

#if PRECISION == DOUBLE
                res = sscanf(selstr,"%le %le %le %le %le %le",
#else
                res = sscanf(selstr,"%e %e %e %e %e %e",
#endif
                                &vals[0],
                                &vals[1],
                                &vals[2],
                                &vals[3],
                                &vals[4],
                                &vals[5]);

                if (res!=6) goto illegal_qmcontrol;

                if ( (strcmp(keystr,"QMFIXBOXB")==0) ||
                     (strcmp(keystr,"QMEXTENDBOXB")==0) ) {

                        for (res=0;res<6;res++)
                                vals[res] /= ANGS_TO_BOHR;

                }

                for (res=0;res<6;res++)
                        if (vals[res]!=global_data->qmboxext[res])
                                different=1;

                for (res=0;res<6;res++)
                        if (global_data->qmboxext[res]!=2.0)
                                firsttime=0;

                if ((global_data->restarting==TRUE)&&(different)&&(firsttime)) {
                        txtout("***************************************************\n");
                        txtout("***************************************************\n");
                        txtout("Info: QMFIXBOX parameters are not taken from the\n");
                        txtout("      restart file but from the control file.\n");
                        txtout("      Control-file: %f %f   %f %f   %f %f\n",
                                vals[0],vals[1],vals[2],vals[3],vals[4],vals[5]);
                        txtout("      Restart-file: %f %f   %f %f   %f %f\n",
                                global_data->qmboxext[0],global_data->qmboxext[1],
                                global_data->qmboxext[2],global_data->qmboxext[3],
                                global_data->qmboxext[4],global_data->qmboxext[5]);
                        txtout("***************************************************\n");
                        txtout("***************************************************\n");
                        sleep((unsigned int) 7);
                }

                memcpy(global_data->qmboxext,vals,6*sizeof(double));

                if (    (strcmp(keystr,"QMFIXBOX" )==0) ||
                        (strcmp(keystr,"QMFIXBOXB")==0) )
                        global_data->qmfixbox=1;
                else    global_data->qmfixbox=0;
                return 0;
        } else
        if (strcmp(keystr,"QMATOMFORCE")==0) {
                int res;
                char dstr[30];
                res = sscanf(selstr,"%s",dstr);
                if (res!=1) goto illegal_qmcontrol;
                if (strcmp(dstr,"OFF")==0) {
                        global_data->qmatomforce=QMATOMFORCE_OFF;
                        txtout("No forces on QM atoms.\n");
                }
                if (strcmp(dstr,"QM")==0) {
                        global_data->qmatomforce=QMATOMFORCE_QM;
                        txtout("Using QM-forces for QM atoms.\n");
                }
                if (strcmp(dstr,"CLASSIC")==0) {
                        global_data->qmatomforce=QMATOMFORCE_CLASSIC;
                        txtout("Using EGO-forces for QM atoms.\n");
                }

                return 0;
        }

        /*** get next key-word ***/
        for (i=0;i<strlen(selstr);i++) {
                if ((selstr[i]==' ')||(selstr[i]=='\t')) {
                        selstr[i] = 0;
                        break;
                }
        }
        for (i++;(i<strlen(selstr))&&
                 ((selstr[i]==' ')||(selstr[i]=='\t'));i++) {
        }
#if PRECISION == DOUBLE
                sscanf(selstr,"%le",&dvalue);
#else
                sscanf(selstr,"%e",&dvalue);
#endif
        if (sscanf(selstr,"%d",&value)!=1) goto illegal_qmcontrol;

        if (strcmp(keystr,"QMENERGYOFFSET")==0) {
                global_data->qmenergyoffset = dvalue;
                txtout("QM-Energy-offset is %f.\n",global_data->qmenergyoffset);
                return 0;
        } else
        if (strcmp(keystr,"QMCUTOFF")==0) {
                if (value<10)                      goto illegal_qmcontrol;
                global_data->qmcutoff = value;
                txtout("WF-Cutoff for QM-Programm is %d.\n",global_data->qmcutoff);
                return 0;
        } else
        if (strcmp(keystr,"QMSTART")==0) {
                if (value<0)                       goto illegal_qmcontrol;
                global_data->qmstart_step = value;
                strcpy(global_data->qmstartstr,selstr+i);
                txtout("Starting QM-program at step %d with command <%s>.\n",
                        global_data->qmstart_step,global_data->qmstartstr);

                /*** delete irritating files for the QM interface ***/
                remove_file(FN_QMSTART,    global_data->outpath);
                remove_file(FN_QMCONTINUE, global_data->outpath);
                remove_file(FN_EGOCONTINUE,global_data->outpath);
                remove_file(FN_QMOUT,      global_data->outpath);
                remove_file(FN_QMEXIT,     global_data->outpath);
                remove_file(FN_EGOEXIT,    global_data->outpath);
                return 0;
        } else
        if (strcmp(keystr,"QMCONTINUE")==0) {
                if (value<0)                       goto illegal_qmcontrol;
                global_data->qmcontin_step = value;
                strcpy(global_data->qmcontinstr,selstr+i);
                txtout("Continuing QM-program every %d-th step with command <%s>.\n",
                        global_data->qmcontin_step,global_data->qmcontinstr);
                return 0;
        } else
        if (strcmp(keystr,"EXTDYN")==0) {
                int res;
                if (value<0)                       goto illegal_qmcontrol;
                global_data->extdyn_step[0] = value;
                res = sscanf(selstr+i,"%d %d %s",
                        &global_data->extdyn_step[1],
                        &global_data->extdyn_step[2],
                        global_data->extdynstr);
                if (res==2) {
                        global_data->extdynstr[0]=0;
                } else  goto illegal_qmcontrol;

                txtout("Starting extern dynamic at step %d with file <%s> for %d steps.\n",
                        global_data->extdyn_step[0],
                        global_data->extdynstr,
                        global_data->extdyn_step[1]);
                txtout("Normal dynamic for %d steps in between.\n",
                        global_data->extdyn_step[2]);
                return 0;
        }
illegal_qmcontrol:
        errout("Fatal error: Illegal line <%s %s> in control-file!\n",
                keystr,selstr);
        pex(-1);
        return -1;
}

int
add_qm_forces(  LOAD_MSG_ONE_NODE *my_data,
                LOAD_MSG_ALL_NODE *global_data,
                ENERGY_INFO *local_energy,
                STRUCT_FORCES *my_force,
                int *skipcounter)
{
        /**********************************************************/
        /*** Here all stuff for the QM-interface is done: collect**/
        /*** QM-atom-positions and dummy-atoms for the QM-program**/
        /**********************************************************/

        int i,j;
        for (i=0;i<my_data->no_atoms;i++) {
                double data[33]; /*** 3*ko, 3*felec, 1*q, 10*local     ***/
                                 /*** 4*bondf,4*anglef,4*dihef,4*imprf ***/
                int atom_id,localindex;
                int force_id = my_data->atom_sele_flag1[i];

                if (!(force_id&FLAG_QMATOM)) continue;

                /*** now we have a QM atom ***/
                atom_id    = my_data->atom_id[i];
                localindex = get_qmnode_index(atom_id,my_data);

                data[0] = my_data->x_new[i];
                data[1] = my_data->y_new[i];
                data[2] = my_data->z_new[i];
                data[3] = my_data->force_elec[4*i+1];
                data[4] = my_data->force_elec[4*i+2];
                data[5] = my_data->force_elec[4*i+3];
                data[6] = my_data->q[i];

                memcpy( data+7,
                        my_data->qmlocal.local+10*localindex,
                        10*sizeof(double));
                /***  my_data->qmlocal.bondf, etc. with EGO sign ***/
                data[17] = my_data->qmlocal.bondf[4*localindex  ];
                data[18] = my_data->qmlocal.bondf[4*localindex+1];
                data[19] = my_data->qmlocal.bondf[4*localindex+2];
                data[20] = my_data->qmlocal.bondf[4*localindex+3];

                data[21] = my_data->qmlocal.anglef[4*localindex  ];
                data[22] = my_data->qmlocal.anglef[4*localindex+1];
                data[23] = my_data->qmlocal.anglef[4*localindex+2];
                data[24] = my_data->qmlocal.anglef[4*localindex+3];

                data[25] = my_data->qmlocal.dihef[4*localindex  ];
                data[26] = my_data->qmlocal.dihef[4*localindex+1];
                data[27] = my_data->qmlocal.dihef[4*localindex+2];
                data[28] = my_data->qmlocal.dihef[4*localindex+3];

                data[29] = my_data->qmlocal.imprf[4*localindex  ];
                data[30] = my_data->qmlocal.imprf[4*localindex+1];
                data[31] = my_data->qmlocal.imprf[4*localindex+2];
                data[32] = my_data->qmlocal.imprf[4*localindex+3];

                add_qmexcl_list(atom_id,my_data);

                sort_to_qminfo( atom_id,i,0,
                                &(my_force->qm),
                                data,
                                my_data->qmlocal.incllen[localindex],
                                my_data->qmlocal.incllist+localindex*MAXDUMMYSPERQMATOM);
        }
#if DEBUG
txtout("QM-Prepare: after loop sort_to_qminfo\n");
#endif
        /***********************************/
        /*** now extract all dummy atoms ***/
        /***********************************/
        my_force->qm.cur_nr_dummys = 0;
        for (i=0,j=0;i<my_data->no_atoms;i++) {

                int qmexcl = test_qmexclusion(my_data->atom_id[i]);

                if  (   (qmexcl!=0) ||
                        (my_data->atom_sele_flag1[i]&FLAG_QMDUMMY) ) {

                        my_data->atom_sele_flag1[i]  |= FLAG_QMDUMMY;

                        my_force->qm.excltype[  j  ] = qmexcl;
                        my_force->qm.dummy_id[  j  ] = my_data->atom_id[i];
                        my_force->qm.dummy_q [  j  ] = my_data->q[i];
                        my_force->qm.dummy_ko[3*j+0] = my_data->x_new[i];
                        my_force->qm.dummy_ko[3*j+1] = my_data->y_new[i];
                        my_force->qm.dummy_ko[3*j+2] = my_data->z_new[i];
                        if (++j>=my_force->qm.len_dummys) {
                                errout("Fatal error: Not enough space allocated\n");
                                errout("for QM-dummy atoms (%d > %d).\n",
                                        j,my_force->qm.len_dummys);
                                pex(-1);
                        }

                        /*** look at O atom ***/
                        if (my_data->q[i]<0.0) {
                                double efield[3];

                                efield[0] = -my_data->force_elec[4*i+1]/my_data->q[i];
                                efield[1] = -my_data->force_elec[4*i+2]/my_data->q[i];
                                efield[2] = -my_data->force_elec[4*i+3]/my_data->q[i];
                                global_data->userdouble[0]=1.0;
                        }

                }
        }
#if DEBUG
txtout("QM-Prepare: after loop extract all dummy atoms\n");
#endif
        my_force->qm.cur_nr_dummys = j;
        my_force->qm.cur_len       = 0;

        /*** here actually the QM-interface is envoced ***/
        manage_qminterface(global_data,my_force,(*skipcounter));

        local_energy->qm_energy[0] = my_force->qm.qm_energy[0];
        local_energy->qm_energy[1] = my_force->qm.qm_energy[1];

        if (global_data->qmcontin_step>1) {
                (*skipcounter)++;
                if ((*skipcounter)>=global_data->qmcontin_step) {
                        (*skipcounter)=0;
                        txtout("Reset skipcounter %d\n",(*skipcounter));
                } else {
                        txtout("Skipping integration %d\n",(*skipcounter));
                }
        }

        return 0;
}

static int
manage_qminterface(LOAD_MSG_ALL_NODE *global_data,
                   STRUCT_FORCES *my_force,
                   int skipcounter)
{
        int difstep=global_data->cycles_passed-global_data->qmstart_step;
        int await_qmresult=0;
#if NOFORK == 1
        char fname[2*MAX_STRING_SIZE];
#endif

#if DEBUG
txtout("Enter manage_qminterface cycles_passed %d  qmstart %d  qm.len %d.\n",
global_data->cycles_passed,global_data->qmstart_step,my_force->qm.len);
#endif
        if (difstep<0)                  goto go_onhere;
        if (my_force->qm.len<=0)        goto go_onhere;

        /**********************************************************/
        /*** Starting of the QM-program first time              ***/
        /**********************************************************/
        if (    (global_data->qmstart_step>=0) &&
                (difstep>=0) && (cpmd_started==0) ) {
                char strprog[MAX_STRING_SIZE],strarg[MAX_STRING_SIZE];
#if DEBUG
                txtout("In STARTQM: Before prepare_qminterface\n");
#endif
                strcpy(qmoutpath,global_data->outpath);
                prepare_qminterface(global_data,&(my_force->qm));
#if DEBUG
                txtout("In STARTQM: Before patch_qm_template\n");
#endif
                patch_qm_template(      global_data,
                                        &(my_force->qm),
                                        WITH_ELEC);
                /*** give the signal for CPMD for reading input file ***/
                touch_file(FN_QMSTART,global_data->outpath);
#if DEBUG
                txtout("In STARTQM: open_qmtraj\n");
#endif
                await_qmresult = 1;
                open_qmtraj(global_data);

                /*** special debug mode for the interface ***/
                if ((global_data->debug_flags[1]&4) ||
                    (global_data->debug_flags[1]&2)) {
                        touch_file(FN_EGOCONTINUE,global_data->outpath);
                        goto skip_cpmdstart;
                }

                /*** give the signal for CPMD to continue ***/
                touch_file(FN_QMCONTINUE,global_data->outpath);

                sscanf(global_data->qmstartstr,"%s %s",strprog,strarg);
                /*** this case is for the situation when cpmd is started not ***/
                /*** by EGO, but seperately                                  ***/
                if (    (strcmp(strprog,"xxx")==0) ||
                        (strcmp(strprog,"XXX")==0) ) {

                        cpmd_started = 1; /*** lets hope that this is or will be done ***/

                } else {
                /*** this case is for the situation when cpmd is started by EGO ***/
#if NOFORK == 1
                        int res_syscall=0;
                        sprintf(fname,"(cd %s ; %s) &", global_data->outpath,
                                                        global_data->qmstartstr);
                        txtout("Starting QM-program with system(\"%s\").\n",
                                fname);
                        res_syscall = system(fname);
                        cpmd_started = 1;
#else
                        char strprog[MAX_STRING_SIZE],strarg[MAX_STRING_SIZE];

                        sscanf(global_data->qmstartstr,"%s %s",strprog,strarg);

                        /******************************************************/
                        /*** here we start the QM-program via fork and exec ***/
                        /******************************************************/
                        switch (cpmd_started = fork()) {
                        case -1 :
                                errout("Fork command for starting QM-program failed!\n");
                                pex(-1);
                        case 0:
                                chdir(qmoutpath);

                                txtout("Starting fork: <%s %s>\n",
                                        strprog,global_data->qmstartstr+strlen(strprog));
                                execlp(strprog,strprog,global_data->qmstartstr+strlen(strprog),0);

                                errout("execlp failed <%s> <%s> !\n",
                                        strprog,global_data->qmstartstr+strlen(strprog));
                                errout("This usually happens if the QM program can not be found!\n");
                                errout("The QM program must be accessable from the current path.\n");

                                touch_file(FN_EGOEXIT,qmoutpath);
                                txtout("EXIT file for EGO created!\n");
                                pex(-1);
                        default:
                                /*** here the parent process continues ***/
                                break;
                        }
                        txtout("QM-program started with PID %d.\n",cpmd_started);
#endif
                }
        } else
        if (    (cpmd_started>0) &&
                (global_data->qmchargemode==QMCHARGEMODE_VARIABLE) &&
                (skipcounter!=0) ) {
                await_qmresult = 0;
                prepare_qminterface(global_data,&(my_force->qm));
                patch_qm_template(      global_data,
                                        &(my_force->qm),
                                        WITH_ELEC);
        } else
        /**********************************************************/
        /*** Restarting the QM-program                          ***/
        /**********************************************************/
        if (cpmd_started>0) {
                /* int res_syscall=0; */
#if DEBUG
                txtout("In CONTINQM: Before prepare_qminterface\n");
#endif
                prepare_qminterface(global_data,&(my_force->qm));
#if DEBUG
                txtout("In CONTINQM: Before patch_qm_template\n");
#endif
                patch_qm_template(      global_data,
                                        &(my_force->qm),
                                        WITH_ELEC);
#if DEBUG
                txtout("In CONTINQM: After patch_qm_template.\n");
#endif
                touch_file(FN_QMCONTINUE,global_data->outpath);

                my_force->qm.cur_len = 0;       /*** very important ***/

                await_qmresult = 1;

                if ((global_data->debug_flags[1]&4) ||
                    (global_data->debug_flags[1]&2)) {
                        touch_file(FN_EGOCONTINUE,global_data->outpath);
                }

                /******************** Later maybe
                *** txtout("Restarting QM-program with system(\"%s\").\n",
                ***     global_data->qmcontinstr);
                *** sprintf(fname,"(cd %s ; %s) &", global_data->outpath,
                                                global_data->qmcontinstr);
                *** res_syscall = system(fname);
                *** txtout("QM-program completed with result %d.\n",res_syscall)
                *********************************/
        } else {
                await_qmresult = 0;
        }

skip_cpmdstart:
        /**********************************************************/
        /*** Reading the results from the QM-program            ***/
        /**********************************************************/
        if (await_qmresult) {
                FILE *fh;
                int locked=1;
                char lockfile[MAX_STRING_SIZE];
                char exitfile[MAX_STRING_SIZE];

                /*** here we wait until the file EGOCONTINUE is ***/
                /*** created by the QM-programm which is the    ***/
                /*** startsignal for EGO to read the QM-forces  ***/

                strcpy(lockfile,global_data->outpath);
                strcat(lockfile,FN_EGOCONTINUE);

                strcpy(exitfile,global_data->outpath);
                strcat(exitfile,FN_EGOEXIT);

                txtout("########## EGO awaits QM-result! ##############\n");
                do {
                        if ((global_data->debug_flags[1]&4) ||
                            (global_data->debug_flags[1]&2)) {
                                /*** no wait needed here ***/
                        } else {
                                sleep((unsigned int) 1);
                        }

                        if (get_terminate_flag()!=DONT_STOP) {
                                errout("STOP-Signal detected!\n");
                                pex(1);
                        }

                        fh=fopen(exitfile,"r");
                        if (fh!=NULL) {
                                errout("EGO-EXIT file detected!\n");
                                fclose(fh);
                                remove(exitfile);
                                pex(-1);
                        }

                        fh=fopen(lockfile,"r");
                        if (fh!=NULL) {
                                fclose(fh);
                                sleep((unsigned int) 1); /*** on a vpp one needs that sleep ***/
                                remove(lockfile);
                                locked=0;
                        }
                } while(locked);
        }
#if DEBUG
        txtout("In CONTINQM: Before readin_qmresult.\n");
#endif

        readin_qmresult(global_data,&(my_force->qm));

#if DEBUG
        txtout("In CONTINQM: After readin_qmresult.\n");
#endif

go_onhere:

#if defined(SEQ)
        set_qmforces(global_data,&(my_force->qm));
#else
        /*** distribute forces to the nodes ***/
        /*** not yet implemented !!         ***/
        PCI_send_qmforces(global_data,&(my_force->qm));
#endif
#if DEBUG
        txtout("In CONTINQM: End of manage_qminterface.\n");
#endif
        return 0;
}

/*** clean up after the simulation    ***/
/*** kill -9 just in case of hang ups ***/
int
finish_qmcalculation()
{
        /* cpmd_started should contain the PID of CPMD.
           On SGI systems this does not work, so we have
           an alternative implementation there. */
        if (cpmd_started>0) {
                char comstr[MAX_STRING_SIZE];

                close_qmtraj();

                touch_file(FN_QMEXIT,qmoutpath);
                txtout("EXIT file for QM program created!\n");

                sprintf(comstr,"sleep 10; kill -9 %d",cpmd_started);
                system(comstr);

                cpmd_started=0;
        }

#if 0
/* for SGIs: */
        if (cpmd_started>0) {
                char comstr[MAX_STRING_SIZE];
                char comstr1[MAX_STRING_SIZE];
                FILE *com;
                char command[MAX_STRING_SIZE];
                char result[MAX_STRING_SIZE];

/*              char strprog[MAX_STRING_SIZE],strarg[MAX_STRING_SIZE];
                sscanf(global_data->qmstartstr,"%s %s",strprog,strarg);
                sprintf(command,"ps | grep %s | tr -s \" \"| cut -f2 -d \" \"\n",strprog); */
                sprintf(command,"ps | grep cpmd.x | tr -s \" \"| cut -f2 -d \" \"\n");
                com = popen(command, "r");
                fscanf(com,"%s",result);
                pclose(com);
#if DEBUG
                txtout(result);
                txtout("\n");
#endif

                close_qmtraj();

                touch_file(FN_QMEXIT,qmoutpath);
                txtout("EXIT file for QM program created!\n");

                sprintf(comstr,"sleep 10; kill -9 %s",result);
#if DEBUG
                sprintf(comstr1,"sleep 10; kill -9 %s\n",result);
                txtout(comstr1);
#endif
                system(comstr);

                cpmd_started=0;
        }
#endif	/* end of SGI part */
        return 0;
}

static int
open_qmtraj(LOAD_MSG_ALL_NODE *global_data)
{
        int newfile;

        strcpy(fnqmtraj,global_data->outpath);
        strcat(fnqmtraj,global_data->idstr);
        strcat(fnqmtraj,"qmtraj.out");

        strcpy(fndipol,global_data->outpath);
        strcat(fndipol,global_data->idstr);
        strcat(fndipol,"dipol.out");

        newfile=0;
        if ((hqmtraj=fopen(fnqmtraj,"r"))==NULL) { /*** test existance of file **/
                newfile = 1;
        } else {
                fclose(hqmtraj);
        }

        if ((hqmtraj=fopen(fnqmtraj,"aw"))==NULL) { /*** open for appending ***/
                errout("Fatal error: Can not open force file <%s>.\n",
                        fnqmtraj);
                pex(-1);
        }

        qmtraj_here=1;

        if (newfile) {
                write_force_header(hqmtraj,global_data,
                                FLAG_QMATOM,global_data->nr_qmatoms,
                                "REMARK CPMD-EGO-Interface version 1\n",
                                "QMTRAJECTORY");
        }
        fclose(hqmtraj);

        /******************************************************************/
        /*** the dipol-file                                             ***/
        /******************************************************************/
        newfile=0;
        if ((hqmtraj=fopen(fndipol,"r"))==NULL) { /*** test existance of file **/
                newfile = 1;
        } else {
                fclose(hqmtraj);
        }

        if ((hqmtraj=fopen(fndipol,"aw"))==NULL) { /*** open for appending ***/
                errout("Fatal error: Can not open force file <%s>.\n",
                        fnqmtraj);
                pex(-1);
        }

        if (newfile) {
                write_force_header(hqmtraj,global_data,
                                FLAG_QMATOM,global_data->nr_qmatoms,
                                "REMARK CPMD-EGO-Interface version 1\n",
                                "QMDIPOL");
                fprintf(hqmtraj,"Step      Ex         Ey         Ez        E          Dx         Dy         Dz           D\n");
        }
        fclose(hqmtraj);

        return 0;
}

int
write_force_header(FILE *hqmtraj,LOAD_MSG_ALL_NODE *global_data,
                int flag,int nr_atoms,
                char *headstr,char *flagstr)
{
        int i;

        fprintf(hqmtraj,headstr);

        if (global_data->nr_qmatoms>0) { /*** if QM-calculation then also include ***/
                                         /*** the input file for the QM-programm  ***/
                char fname[MAX_STRING_SIZE];
                strcpy(fname,global_data->outpath);
                strcat(fname,fncarparrun);

                fprintf(hqmtraj,"REMARK First QM-input-file ====================================================\n");
                writeout_file(hqmtraj,fname);
        }

        fprintf(hqmtraj,"REMARK %s =======================\n",
                        global_data->ctlfile);
        writeout_ctlfile(hqmtraj,global_data);

        fprintf(hqmtraj,"REMARK Data ===================================================================\n");
        fprintf(hqmtraj,"[%s]\n",flagstr);
        fprintf(hqmtraj,"Atoms %d\n",nr_atoms);
        fprintf(hqmtraj,"Atom-Nr  Resid.-Nr  Atom-Str  Resid.-Str\n");

        for (i=0;i<global_data->total_no_atoms;i++) {

                if ((global_data->atom_sele_flag1[i]&flag)==0) continue;

                fprintf(hqmtraj,"%-8d  %-5d     %-4s      %-4s\n",
                        i+1,
                        global_data->atomdata[i].residnr,
                        global_data->atomdata[i].atomstr,
                        global_data->atomdata[i].residstr);
        }

        fflush(hqmtraj);
        return 0;
}

static
int close_qmtraj()
{
        if (qmtraj_here) {
                qmtraj_here = 0;
        }
        return 0;
}

/******************************************************************/
/*** Here we shift the molecule to the center of the cell,      ***/
/*** check the bounding boxes and manage the link-atoms.        ***/
/******************************************************************/

static  int     firsttimebak=1;

static  int
prepare_qminterface(    LOAD_MSG_ALL_NODE *global_data,
                        STRUCT_QM_INFO *finfo)
{
        int     i,h1,h2,j;
        double  dx,dy,dz;
        double  minx,miny,minz,maxx,maxy,maxz;

        qmenecor=0.0;

        /*** extent of the QM-box ***/
        dx = global_data->qmboxext[1]-global_data->qmboxext[0];
        dy = global_data->qmboxext[3]-global_data->qmboxext[2];
        dz = global_data->qmboxext[5]-global_data->qmboxext[4];

        /*** calculate offset-vector to shift the center of QM-atoms into ***/
        /*** the center of the QM-calculation box global_data->qmboxext   ***/
        finfo->mass = 0.0;
        minx = miny = minz =  9999999.0;
        maxx = maxy = maxz = -9999999.0;
        for (i=0;i<finfo->len;i++) { /*** loop over all QM-atoms ***/
                int atom_id  = finfo->atom_id[i];
                int atomtyp  = global_data->atomdata[atom_id].atomtype;
                double ko[3],m     = global_data->AB[atomtyp][atomtyp];

                finfo->mass += m;
                rotforw(finfo->ko+3*i,ko);

                if (minx>ko[0]) minx=ko[0];
                if (miny>ko[1]) miny=ko[1];
                if (minz>ko[2]) minz=ko[2];
                if (maxx<ko[0]) maxx=ko[0];
                if (maxy<ko[1]) maxy=ko[1];
                if (maxz<ko[2]) maxz=ko[2];
        }

        finfo->offs[0] = (minx + maxx)/2.0;
        finfo->offs[1] = (miny + maxy)/2.0;
        finfo->offs[2] = (minz + maxz)/2.0;
#if DEBUG
        txtout("Total mass of QM-atoms: %f\n",finfo->mass);
        txtout("Center of QM-atoms    : %f %f %f\n",
                finfo->offs[0],finfo->offs[1],finfo->offs[2]);
#endif
        /*** shift it to the center of the QM-BOX ***/
        finfo->offs[0] -= dx/2.0;
        finfo->offs[1] -= dy/2.0;
        finfo->offs[2] -= dz/2.0;

        /*** Test positions of QM-atoms ***/
        for (i=0;i<finfo->len;i++) { /*** loop over all QM-atoms ***/
                double ko[3];

                rotforw(finfo->ko+3*i,ko);
                ko[0] = ko[0] - finfo->offs[0];
                ko[1] = ko[1] - finfo->offs[1];
                ko[2] = ko[2] - finfo->offs[2];

                if (ko[0]<QMWARNINGDIST)        goto not_in_QMbox;
                if (ko[0]+QMWARNINGDIST>dx)     goto not_in_QMbox;

                if (ko[1]<QMWARNINGDIST)        goto not_in_QMbox;
                if (ko[1]+QMWARNINGDIST>dy)     goto not_in_QMbox;

                if (ko[2]<QMWARNINGDIST)        goto not_in_QMbox;
                if (ko[2]+QMWARNINGDIST>dz)     goto not_in_QMbox;

                continue;

not_in_QMbox:
                errout("Fatal error: One QM atom is closer than %f to the QM box-edge.\n",QMWARNINGDIST);
                errout("Restart the program with a larger QM-BOX or find a better orientation with QMROT.!\n");
                errout("QM-Atom %d (OrigPos: %f,%f,%f)  (Shifted/Rotated: %f,%f,%f)\n",
                        finfo->atom_id[i]+1,
                        finfo->ko[3*i+0],
                        finfo->ko[3*i+1],
                        finfo->ko[3*i+2],
                        ko[0],
                        ko[1],
                        ko[2]);
                pex(-1);
        }

        /*** Manage the link atoms ***/
        for (i=0;i<finfo->qmlink_len;i++) {     /*** for every link-atom ***/
                double  dx,dy,dz,fac,len,bforce2,newd=0,clforce,ausl;

                if (finfo->qmlink_dist1[i]<=0.0) continue; /*** no link atom here ***/

                /*** get the index in our QM_INFO list ***/
                h1 = h2 = -1;
                for (j=0;j<finfo->len;j++)
                        if (finfo->atom_id[j]==finfo->qmlink_p1[i]) {
                                h1 = j;
                                break;
                        }
                for (j=0;j<finfo->len;j++)
                        if (finfo->atom_id[j]==finfo->qmlink_p2[i]) {
                                h2 = j;
                                break;
                        }
                if ((h1==-1)||(h2==-1)||(h1==h2)) {
                        errout("Fatal error: Illegal QM-link-atom definition!\n");
                        errout("Check your QMLINK %d %d definition and check whether the\n",
                                finfo->qmlink_p1[i]+1,finfo->qmlink_p2[i]+1);
                        errout("link atom is included in the QMATOM selection.\n");
                        pex(-1);
                }

                /*** now h1 is the index of the QM atom   ***/
                /*** and h2 is the index of the Link atom ***/
                dx = finfo->ko[3*h2 + 0] - finfo->ko[3*h1 + 0];
                dy = finfo->ko[3*h2 + 1] - finfo->ko[3*h1 + 1];
                dz = finfo->ko[3*h2 + 2] - finfo->ko[3*h1 + 2];
                len = sqrt(dx*dx + dy*dy + dz*dz);

                /*** now adapt the link atom bond distance ***/
                /*** project QM force along bond axes      ***/
/****
                clforce = (dx*finfo->bondf[4*h1+1] +
                           dy*finfo->bondf[4*h1+2] +
                           dz*finfo->bondf[4*h1+3] )/len;

****/
                /*** set the new link atom position depending on the ***/
                /*** parameters given in the control-file            ***/
                if (global_data->qmlinkmode==QMLINKMODE_ACTIVEPOS) {

                        clforce=2.0*finfo->qmlink_fconst2[i]*(len - finfo->qmlink_dist2[i]);

                        ausl = clforce/(2.0*finfo->qmlink_fconst1[i]);
                        newd = finfo->qmlink_dist1[i] + ausl;

                        /*** subtract link atom bond energy ***/
                        /***
                        qmenecor += finfo->qmlink_fconst1[i]*ausl*ausl;
                        ***/
                        qmenecor = 0.0;
                        /*** in that case we do not need that force, but must set it to zero ***/
                        finfo->bondf[4*h1  ]=0.0;
                        finfo->bondf[4*h1+1]=0.0;
                        finfo->bondf[4*h1+2]=0.0;
                        finfo->bondf[4*h1+3]=0.0;

                        txtout("QMLINKMODE=SPLAM: MM-force %f  equid %f newd %f  enecor %f\n",
                                clforce,finfo->qmlink_dist1[i],newd,qmenecor);

                } else
                /*** the link atom is always positioned in a fixed and given ***/
                /*** distance from the QM atom at the bond axis              ***/
                if (global_data->qmlinkmode==QMLINKMODE_FIXEDPOS) {

                        newd = finfo->qmlink_dist1[i];

                        finfo->bondf[4*h2  ] = -finfo->bondf[4*h1  ];
                        finfo->bondf[4*h2+1] = -finfo->bondf[4*h1+1];
                        finfo->bondf[4*h2+2] = -finfo->bondf[4*h1+2];
                        finfo->bondf[4*h2+3] = -finfo->bondf[4*h1+3];

                        qmenecor = 0.0;
                }
                else
                /*** try to minimize the position of the linkatom ***/
                /*** but it is positioned at the bond axis        ***/
                if (global_data->qmlinkmode==QMLINKMODE_MINIPOS) {
                        bforce2 = (dx*finfo->qm_force[3*h2 + 0] +
                                   dy*finfo->qm_force[3*h2 + 1] +
                                   dz*finfo->qm_force[3*h2 + 2] )/len;

                        newd = finfo->qmlink_dist1[i] + bforce2/(4.0*finfo->qmlink_fconst1[i]);

                        finfo->bondf[4*h2  ] = -finfo->bondf[4*h1  ];
                        finfo->bondf[4*h2+1] = -finfo->bondf[4*h1+1];
                        finfo->bondf[4*h2+2] = -finfo->bondf[4*h1+2];
                        finfo->bondf[4*h2+3] = -finfo->bondf[4*h1+3];

                        qmenecor = 0.0;
                        txtout("QMLINKMODE=CLAMP: force %f  olddist %f  newdist %f\n",
                                        bforce2,finfo->qmlink_dist1[i],newd);
                        finfo->qmlink_dist1[i]=newd;
                } else
                /*** try to minimize the position of the linkatom     ***/
                /*** the first time it is positioned at the bond axis ***/
                if (global_data->qmlinkmode==QMLINKMODE_BAKOWIES) {
                        bforce2 = sqrt(finfo->qm_force[3*h2 + 0]*finfo->qm_force[3*h2 + 0] +
                                       finfo->qm_force[3*h2 + 1]*finfo->qm_force[3*h2 + 1] +
                                       finfo->qm_force[3*h2 + 2]*finfo->qm_force[3*h2 + 2]);

                        finfo->bondf[4*h2  ] = -finfo->bondf[4*h1  ];
                        finfo->bondf[4*h2+1] = -finfo->bondf[4*h1+1];
                        finfo->bondf[4*h2+2] = -finfo->bondf[4*h1+2];
                        finfo->bondf[4*h2+3] = -finfo->bondf[4*h1+3];

                        if (firsttimebak==1) {
                                newd = finfo->qmlink_dist1[i];
                                fac= newd/len;
                                finfo->kolink[3*h2 + 0] = finfo->ko[3*h1 + 0] + dx*fac;
                                finfo->kolink[3*h2 + 1] = finfo->ko[3*h1 + 1] + dy*fac;
                                finfo->kolink[3*h2 + 2] = finfo->ko[3*h1 + 2] + dz*fac;

                                firsttimebak=0;
                        } else {
                                finfo->kolink[3*h2 + 0] += finfo->qm_force[3*h2 + 0]/(4.0*finfo->qmlink_fconst1[i]);
                                finfo->kolink[3*h2 + 1] += finfo->qm_force[3*h2 + 1]/(4.0*finfo->qmlink_fconst1[i]);
                                finfo->kolink[3*h2 + 2] += finfo->qm_force[3*h2 + 2]/(4.0*finfo->qmlink_fconst1[i]);
                        }
                        qmenecor = 0.0;
                        txtout("QMLINKMODE=LAMP: force %f\n",bforce2);
                        finfo->flag[h2]     |= FLAG_ISLINKATOM;
                        continue;
                }

                fac= newd/len;
                finfo->kolink[3*h2 + 0] = finfo->ko[3*h1 + 0] + dx*fac;
                finfo->kolink[3*h2 + 1] = finfo->ko[3*h1 + 1] + dy*fac;
                finfo->kolink[3*h2 + 2] = finfo->ko[3*h1 + 2] + dz*fac;
                finfo->flag[h2]        |= FLAG_ISLINKATOM;
        }

        return 0;
}

static int
touch_file(char *fname,char *fpath)
{
        FILE *fh;
        char fullname[MAX_STRING_SIZE];

        strcpy(fullname,fpath);
        strcat(fullname,fname);

        fh = fopen(fullname,"w");
        if (fh==NULL) {
                errout("Fatal error: Unable to create file <%s>.\n",fname);
                pex(-1);
        }
        fclose(fh);
        return 0;
}

static int
remove_file(char *fname,char *fpath)
{
        char fullname[MAX_STRING_SIZE];

        strcpy(fullname,fpath);
        strcat(fullname,fname);
        remove(fullname);
        return 0;
}

/**************************************************************************/
/*** reads in the output file of the QM-program containing the forces   ***/
/*** acting on each EGO-QM-atom. Writes out coordinates and forces to   ***/
/*** "qmtraj.out".                                                      ***/
/**************************************************************************/

static int      get_vdw_AB(double corec,double *A, double *B);
static double   get_exclusion_fac(double d);

static double
get_exclusion_fac(double d)
{
        double fac;

        if (d<3.0) return (double)0.0;
        if (d>4.5) return (double)1.0;

        fac = (d-3.0)/1.5;
        fac = 1.0-fac*fac;
        return 1.0-fac*fac;
}

static int
get_vdw_AB(double corec,double *A, double *B)
{
        if ((corec>0.5)&&(corec<1.5)) { /*** H with H or H with D ***/
                *A=5270.0;
                *B=30.0;
        } else
        if ((corec>5.5)&&(corec<6.5)) { /*** H with C ***/
                *A=55000.0;
                *B=116.0;
        } else
        if ((corec>6.5)&&(corec<7.5)) { /*** H with N ***/
                *A=61000.0;
                *B=44.0;
        } else
        if ((corec>7.5)&&(corec<8.5)) { /*** H with O ***/
                *A=26000.0;
                *B=29.0;
        } else {                        /*** if unknown then like C ***/
                *A=55000.0;
                *B=116.0;
        }
        return 0;
}

static int
correct_elecvdw_restqm( double *kolink,
                        double *koq,
                        double chargelink,
                        int li,
                        LOAD_MSG_ALL_NODE *global_data,
                        STRUCT_QM_INFO *finfo)
{
        int i;
        double rxl,ryl,rzl,rl_2;
        double rxq,ryq,rzq,rq_2;
        double elecstat_loc_energyl,elecstat_loc_energyq,vdw_loc_energy;
        double r_6,fx,fy,fz,felec,fvdw,facq,facl;
        double A,B,sumene,ko[3],corec=1.0;

        for (i=0,sumene=0.0;i<global_data->nr_qmatoms;i++) {
                if (li==i) continue;

                if (finfo->flag[i]&FLAG_ISLINKATOM) {
                        ko[0]=finfo->kolink[3*i  ];
                        ko[1]=finfo->kolink[3*i+1];
                        ko[2]=finfo->kolink[3*i+2];
                        corec=1.0;
                } else {
                        ko[0]=finfo->ko[3*i  ];
                        ko[1]=finfo->ko[3*i+1];
                        ko[2]=finfo->ko[3*i+2];
                        corec=finfo->qcore[i];
                }

                rxq = koq[0] - ko[0];
                ryq = koq[1] - ko[1];
                rzq = koq[2] - ko[2];
                rq_2 = rxq*rxq + ryq*ryq + rzq*rzq;

                facq=get_exclusion_fac((double)sqrt(rq_2));
                if (facq<=0.00001) continue;

                rxl = kolink[0] - ko[0];
                ryl = kolink[1] - ko[1];
                rzl = kolink[2] - ko[2];
                rl_2 = rxl*rxl + ryl*ryl + rzl*rzl;

                facl=get_exclusion_fac((double)sqrt(rl_2));

                /*** electrostatics of link atom ***/
                rq_2=1.0/rq_2;
                rl_2=1.0/rl_2;
                elecstat_loc_energyl = (facl+facq)*0.25 * global_data->f_elstat * finfo->qm_q[i] * chargelink * sqrt(rl_2);
                felec     = 2 * rl_2 * elecstat_loc_energyl;
                fx = felec * rxl;
                fy = felec * ryl;
                fz = felec * rzl;

                /*** electrostatics of q atom ***/
                elecstat_loc_energyq =-(facl+facq)*0.25 * global_data->f_elstat * finfo->qm_q[i] * chargelink * sqrt(rq_2);
                felec     = 2 * rq_2 * elecstat_loc_energyq;
                fx += felec * rxq;
                fy += felec * ryq;
                fz += felec * rzq;

                /*** vdw of link atom ***/
                get_vdw_AB(corec,&A,&B);

                r_6 = rl_2 * rl_2 * rl_2;
                vdw_loc_energy = facl*r_6 * (r_6*A - B) / 2;
                fvdw           = rl_2 * (24*vdw_loc_energy + 6*r_6*B);

#if DEBUG_LINK
txtout("Corr QM atom %d: elec %f %f %f   vdw %f %f %f\n",
        i,fx,fy,fz,fvdw * rxl,fvdw * ryl,fvdw * rzl);
#endif

                fx += fvdw * rxl;
                fy += fvdw * ryl;
                fz += fvdw * rzl;

                /*** correct only vdw on link atom ***/
                finfo->qm_force[3*li+0]-= fvdw * rxl;
                finfo->qm_force[3*li+1]-= fvdw * ryl;
                finfo->qm_force[3*li+2]-= fvdw * rzl;

                /*** correct elec and vdw on other QM atoms ***/
                finfo->qm_force[3*i+0] += fx;
                finfo->qm_force[3*i+1] += fy;
                finfo->qm_force[3*i+2] += fz;
                sumene -=  elecstat_loc_energyl + elecstat_loc_energyq + vdw_loc_energy;
        }
#if DEBUG_LINK
txtout("Sum QM energy correction %f\n",sumene);
#endif
        return 0;
}

static int
readin_qmresult(LOAD_MSG_ALL_NODE *global_data,STRUCT_QM_INFO *finfo)
{
        double  q,qmenergy[2],force[3],efield[4],sumq=0.0,corq=0.0,qcorperatom;
        int  i,li;
        char fname[MAX_STRING_SIZE],line[MAX_STRING_SIZE];
        FILE *forceh;

        strcpy(fname,global_data->outpath);
        strcat(fname,FN_QMOUT);

        if (global_data->debug_flags[1]&4) {  /*** read from old qmtraj.file ***/
                strcat(fname,".sav");
                if ((forceh=fopen(fname,"r"))==NULL) { /*** open for reading ***/
                        errout("Fatal error: Can not open QM-output file <%s>.\n",
                                fname);
                        pex(-1);
                }
        } else
        if (global_data->debug_flags[1]&2) {
                forceh=NULL;
        } else
        if ((forceh=fopen(fname,"r"))==NULL) { /*** open QM-output for reading ***/
                errout("Fatal error: Can not open QM-output file <%s>.\n",
                        fname);
                pex(-1);
        }

        sumq=0.0;
        txtout("====> EGO reads QM-result from file <%s>.\n",fname);

        /**************************************************/
        /*** readin the energy of the QM system         ***/
        /*** either from an old qmtraj-file for testing ***/
        /*** or just dummy data                         ***/
        /*** or realy from the QM-outputfile            ***/
        /**************************************************/

        /**************************************************/
        /*** case 1: read from old qmtraj.file for tests **/
        /**************************************************/
        if (global_data->debug_flags[1]&4) {
                char tstr[10];
                int fstep,anum;
                double t;

                /*** search step ***/
                do {
                        i = read_string(forceh,line);
                        memcpy(tstr,line,5);
                        tstr[5]=0;
                        if (strcmp(tstr,"Step ")==0) {
#if PRECISION == DOUBLE
                                if (sscanf(line+4,"%d  %lf %lf",
#else
                                if (sscanf(line+4,"%d  %f %f",
#endif
                                        &fstep,qmenergy+0,qmenergy+1)!=3) {
                                        sscanf(line+4,"%d",&fstep);
                                        qmenergy[0] = 0.0;
                                        qmenergy[1] = 0.0;
                                }

                                qmenergy[0] *= KCAL_TO_HART;
                                qmenergy[1] *= KCAL_TO_HART;
                                if (fstep==global_data->cycles_passed) break;
                        }
                } while (i!=-1);

                for (i=0;i<global_data->nr_qmatoms;i++) {
                        if (read_string(forceh,line)==-1) {
                                errout("Fatal error: Not enough lines in QM-output file <%s>!\n",
                                        fname);
                                pex(-1);
                        }
#if PRECISION == DOUBLE
                        sscanf(line,"%d %lf %lf %lf %lf %lf %lf %lf",
#else
                        sscanf(line,"%d %f %f %f %f %f %f %f",
#endif
                                &anum,
                                &finfo->qm_q[i],
                                &t, &t, &t,
                                &finfo->qm_force[3*i+0],
                                &finfo->qm_force[3*i+1],
                                &finfo->qm_force[3*i+2]);
                        sumq+=finfo->qm_q[i];
                }
                efield[0] = efield[1] = efield[2] = efield[3] = 0.0;
                finfo->dipol[0] = 0.0;
                finfo->dipol[1] = 0.0;
                finfo->dipol[2] = 0.0;
                finfo->dipol[3] = 0.0;
        } else
        /***************************************************************/
        /*** case 2: no CPMD started, energies and forces set to 0.0 ***/
        /***************************************************************/
        if (global_data->debug_flags[1]&2) {
                qmenergy[0] = 0.0;
                qmenergy[1] = 0.0;

                for (i=0;i<global_data->nr_qmatoms;i++) {
                        finfo->qm_q[i]         = 0.0;
                        finfo->qm_force[3*i  ] = 0.0;
                        finfo->qm_force[3*i+1] = 0.0;
                        finfo->qm_force[3*i+2] = 0.0;
                }
                sumq+=0.0;
                efield[0] = efield[1] = efield[2] = efield[3] = 0.0;
                finfo->dipol[0] = 0.0;
                finfo->dipol[1] = 0.0;
                finfo->dipol[2] = 0.0;
                finfo->dipol[3] = 0.0;
        } else {
        /***************************************************************/
        /*** case 3: CPMD has calculated forces, read that data      ***/
        /***************************************************************/
                if (read_string(forceh,line)==-1) {
                        errout("Fatal error: Not enough lines in QM-output file <%s>!\n",
                                fname);
                        pex(-1);

                }
#if PRECISION == DOUBLE
                if (sscanf(line,"%lf %lf",qmenergy+0,qmenergy+1)!=2) {
#else
                if (sscanf(line,"%f %f",qmenergy+0,qmenergy+1)!=2) {
#endif
                        errout("Fatal error: Wrong format in QM-output file <%s>!\n",fname);
                        errout("<%s>\n",line);
                        pex(-1);
                }
        }

        finfo->qm_energy[0] = qmenergy[0]/KCAL_TO_HART -  qmenecor;
        finfo->qm_energy[1] = qmenergy[1]/KCAL_TO_HART;

        /******************************************************/
        /*** readin the forces and charges for each QM atom ***/
        /******************************************************/

        if (global_data->debug_flags[1]&4) {

        } else
        if (global_data->debug_flags[1]&2) {

        } else {
        for (i=0;i<global_data->nr_qmatoms;i++) {
                int j,mapi=-1;

                /*** search the right index to put the data in the EGO  ***/
                /*** finfo-structure                                    ***/
                for (j=0;j<global_data->nr_qmatoms;j++)
                        if (finfo->qmmap[j]==i) {
                                mapi=j;
                                break;
                        }
                if (mapi==-1) {
                        errout("Fatal error: Confusion in readin_qmresult.\n");
                        pex(-1);
                }


                if (read_string(forceh,line)==-1) {
                        errout("Fatal error: Not enough lines in QM-output file <%s>!\n",
                                fname);
                        pex(-1);
                }
#if PRECISION == DOUBLE
                if (sscanf(line,"%lf %lf %lf %lf",
#else
                if (sscanf(line,"%f %f %f %f",
#endif
                        force+0,force+1,force+2, /*** forces ***/
                        &q                       /*** charge ***/
                        )!=4) {
                        errout("Fatal error: Wrong format in QM-output file <%s>!\n",
                                fname);
                        errout("<%s>\n",line);
                        pex(-1);
                }

                rotback(force,force);
                /*** convert units from bohr to kCal and store forces.  ***/
                /*** Note that these forces have the right sign in      ***/
                /*** contrast to the EGO-forces                         ***/
                finfo->qm_q[mapi]         = q;
                finfo->qm_force[3*mapi+0] = force[0]*FORCE_AU_TO_KCAL;
                finfo->qm_force[3*mapi+1] = force[1]*FORCE_AU_TO_KCAL;
                finfo->qm_force[3*mapi+2] = force[2]*FORCE_AU_TO_KCAL;
                sumq+=finfo->qm_q[mapi];
        }

        /*** read in mean E-field ***/
        if (read_string(forceh,line)==-1) {
                efield[0] = 0.0;
                efield[1] = 0.0;
                efield[2] = 0.0;
                efield[3] = 0.0;
        } else {
#if PRECISION == DOUBLE
                sscanf(line,"%lf %lf %lf %lf",efield,efield+1,efield+2,efield+3);
#else
                sscanf(line,"%f %f %f %f",efield,efield+1,efield+2,efield+3);
#endif
        }
        /*** read in dipol moment ***/
        if (read_string(forceh,line)==-1) {
                finfo->dipol[0] = 0.0;
                finfo->dipol[1] = 0.0;
                finfo->dipol[2] = 0.0;
                finfo->dipol[3] = 0.0;
        } else {


#if PRECISION == DOUBLE
                sscanf(line,"%lf %lf %lf %lf",
#else
                sscanf(line,"%f %f %f %f",
#endif
                        finfo->dipol  ,finfo->dipol+1,
                        finfo->dipol+2,finfo->dipol+3);
                rotback(finfo->dipol,finfo->dipol);

        }
        }

        if (global_data->debug_flags[1]&2) {

        } else {
                fclose(forceh);
        }


        /******************************************************************/
        /*** now we append all forces to the special qmtraj.out file    ***/
        /******************************************************************/
        if ((hqmtraj=fopen(fnqmtraj,"aw"))==NULL) { /*** open for appending ***/
                errout("Fatal error: Can not open QM-trajectory file <%s>.\n",
                        fnqmtraj);
                pex(-1);
        }
        fprintf(hqmtraj,"Step %d     %# 12.6f %# 12.6f\n",
                global_data->cycles_passed,
                finfo->qm_energy[0],
                finfo->qm_energy[1]);

        finfo->sumlinkforce[0] = 0.0;
        finfo->sumlinkforce[1] = 0.0;
        finfo->sumlinkforce[2] = 0.0;
        for (i=0;i<global_data->nr_qmatoms;i++) {
                double  ko[3];

                if (finfo->flag[i]&FLAG_ISLINKATOM) {
                        ko[0] = finfo->kolink[3*i  ];
                        ko[1] = finfo->kolink[3*i+1];
                        ko[2] = finfo->kolink[3*i+2];
                } else {
                        ko[0] = finfo->ko[3*i  ];
                        ko[1] = finfo->ko[3*i+1];
                        ko[2] = finfo->ko[3*i+2];

                        finfo->sumlinkforce[0] -= (finfo->anglef[4*i+1]+finfo->dihef[4*i+1]+finfo->imprf[4*i+1]);
                        finfo->sumlinkforce[1] -= (finfo->anglef[4*i+2]+finfo->dihef[4*i+2]+finfo->imprf[4*i+2]);
                        finfo->sumlinkforce[2] -= (finfo->anglef[4*i+3]+finfo->dihef[4*i+3]+finfo->imprf[4*i+3]);
                }

                fprintf(hqmtraj,"%5d %# 10.5f   %# 10.5f %# 10.5f %# 10.5f   %# 10.5f %# 10.5f %# 10.5f\n",
                        finfo->atom_id[i]+1,
                        finfo->qm_q[i],
                        ko[0],ko[1],ko[2],
                        finfo->qm_force[3*i+0],
                        finfo->qm_force[3*i+1],
                        finfo->qm_force[3*i+2]);

#if DEBUG_LINK
txtout("Elec : %5d    %# 10.5f  %# 10.5f %# 10.5f   MM charge %# 10.5f\n",
finfo->atom_id[i]+1,
finfo->elecf[4*i+0],finfo->elecf[4*i+1],
finfo->elecf[4*i+2],finfo->elecf[4*i+3]);

txtout("Bond : %5d    %# 10.5f     %# 10.5f %# 10.5f %# 10.5f\n",
finfo->atom_id[i]+1,
finfo->bondf[4*i+0],finfo->bondf[4*i+1],
finfo->bondf[4*i+2],finfo->bondf[4*i+3]);

txtout("Angle: %5d    %# 10.5f     %# 10.5f %# 10.5f %# 10.5f\n",
finfo->atom_id[i]+1,
finfo->anglef[4*i+0],finfo->anglef[4*i+1],
finfo->anglef[4*i+2],finfo->anglef[4*i+3]);

txtout("Dihe : %5d    %# 10.5f     %# 10.5f %# 10.5f %# 10.5f\n",
finfo->atom_id[i]+1,
finfo->dihef[4*i+0],finfo->dihef[4*i+1],
finfo->dihef[4*i+2],finfo->dihef[4*i+3]);

txtout("Impr : %5d    %# 10.5f     %# 10.5f %# 10.5f %# 10.5f\n",
finfo->atom_id[i]+1,
finfo->imprf[4*i+0],finfo->imprf[4*i+1],
finfo->imprf[4*i+2],finfo->imprf[4*i+3]);
#endif

        }
        fflush(hqmtraj);
        fclose(hqmtraj);

#if DEBUG_LINK
txtout("Overlap Force: %f %f %f\n",
finfo->sumlinkforce[0],finfo->sumlinkforce[1],finfo->sumlinkforce[2]);
#endif

        if ((sumq>-0.1) && (sumq< 0.1)) corq= 0.0;
        else
        if ((sumq> 0.9) && (sumq< 1.1)) corq= 1.0;
        else
        if ((sumq>-1.1) && (sumq<-0.9)) corq=-1.0;
        else
        if ((sumq> 1.9) && (sumq< 2.1)) corq= 2.0;
        else
        if ((sumq>-2.1) && (sumq<-1.9)) corq=-2.0;
        else
        if ((sumq> 2.9) && (sumq< 3.1)) corq= 3.0;
        else
        if ((sumq>-3.1) && (sumq<-2.9)) corq=-3.0;
        else
        if ((sumq> 3.9) && (sumq< 4.1)) corq= 4.0;
        else
        if ((sumq>-4.1) && (sumq<-3.9)) corq=-4.0;
        else {
                errout("Fatal error: The total charge in the QM-fragment looks strange (%f)\n",sumq);
                errout("Fatal error: The total charge of the QM fragment should be an integer\n");
                pex(-1);
        }

        qcorperatom = (corq-sumq)/global_data->nr_qmatoms;
#if DEBUG
        txtout("Correction of partial charges by %f to gain total charge %f.\n",qcorperatom,corq);
#endif
        for (i=0;i<global_data->nr_qmatoms;i++) {
                finfo->qm_q[i] += qcorperatom;
        }

        /******************************************************************/
        /****** now all the link atom corrections start *******************/
        /******************************************************************/
        for (li=0;li<global_data->nr_qmatoms;li++) {
                double  flbond[3],flsenk[3],dbond[3],dcc,dcl;
                double  lefield[3],qefield[3],temp;

                int j,h1=-1,qi=-1;
                if (!(finfo->flag[li]&FLAG_ISLINKATOM)) continue;

                for (j=0;j<finfo->qmlink_len;j++)
                        if (finfo->atom_id[li]==finfo->qmlink_p2[j]) {
                                h1 = j;
                                break;
                        }
                if (h1==-1) {errout("Fatal error: #46743 inform programmer !\n");pex(-1);}
                /*** now h1 is the link atom index ***/

                if (finfo->qmlink_dist1[h1]<=0.0) continue; /*** only if real link atom here ***/

                for (j=0;j<global_data->nr_qmatoms;j++)
                        if (finfo->atom_id[j]==finfo->qmlink_p1[h1]) {
                                qi = j;
                                break;
                        }
                if (qi==-1) {errout("Fatal error: #48542 inform programmer !\n");pex(-1);}
                /*** now qi points to the atom to which the link atom is connected. ***/

                /*** get Q-L (dcl) and Q-M (dcc) distance ***/
                dbond[0]=finfo->ko[3*li  ]-finfo->ko[3*qi  ];
                dbond[1]=finfo->ko[3*li+1]-finfo->ko[3*qi+1];
                dbond[2]=finfo->ko[3*li+2]-finfo->ko[3*qi+2];
                dcc=sqrt(dbond[0]*dbond[0]+dbond[1]*dbond[1]+dbond[2]*dbond[2]);

                dbond[0]=finfo->kolink[3*li  ]-finfo->ko[3*qi  ];
                dbond[1]=finfo->kolink[3*li+1]-finfo->ko[3*qi+1];
                dbond[2]=finfo->kolink[3*li+2]-finfo->ko[3*qi+2];
                dcl=sqrt(dbond[0]*dbond[0]+dbond[1]*dbond[1]+dbond[2]*dbond[2]);
                dbond[0]/=dcl;
                dbond[1]/=dcl;
                dbond[2]/=dcl;
#if DEBUG_LINK
txtout("TEST LINK START : dcc=%f  dcl=%f  dir %f %f %f\n",dcc,dcl,dbond[0],dbond[1],dbond[2]);
txtout("TEST LINK ORIG F: q %f %f %f  L %f %f %f\n",
finfo->qm_force[3*qi+0],finfo->qm_force[3*qi+1],finfo->qm_force[3*qi+2],
finfo->qm_force[3*li+0],finfo->qm_force[3*li+1],finfo->qm_force[3*li+2]);
#endif
                /*** now we calculate the elec field at the q and L atom ***/
                if (fabs(finfo->elecf[4*qi+3])>=0.001) {
                        qefield[0] = finfo->elecf[4*qi+0]/finfo->elecf[4*qi+3];
                        qefield[1] = finfo->elecf[4*qi+1]/finfo->elecf[4*qi+3];
                        qefield[2] = finfo->elecf[4*qi+2]/finfo->elecf[4*qi+3];
                } else {
                        qefield[0] = 0.0;
                        qefield[1] = 0.0;
                        qefield[2] = 0.0;
                }

                if (fabs(finfo->elecf[4*li+3])>=0.001) {
                        lefield[0] = finfo->elecf[4*li+0]/finfo->elecf[4*li+3];
                        lefield[1] = finfo->elecf[4*li+1]/finfo->elecf[4*li+3];
                        lefield[2] = finfo->elecf[4*li+2]/finfo->elecf[4*li+3];
                } else {
                        lefield[0] = 0.0;
                        lefield[1] = 0.0;
                        lefield[2] = 0.0;
                }

#if DEBUG_LINK
txtout("TEST EField q: %f %f %f    %f  %f\n",qefield[0],qefield[1],qefield[2],
finfo->qm_q[qi],finfo->elecf[4*qi+3]);
txtout("TEST EField L: %f %f %f    %f  %f\n",lefield[0],lefield[1],lefield[2],
finfo->qm_q[li],finfo->elecf[4*li+3]);
#endif


                /*** substract external force on link atom ***/
/**             finfo->qm_force[3*li+0]-= lefield[0]*finfo->qm_q[li];
                finfo->qm_force[3*li+1]-= lefield[1]*finfo->qm_q[li];
                finfo->qm_force[3*li+2]-= lefield[2]*finfo->qm_q[li];
**/
                /*** add that force to q atom ***/
/**             finfo->qm_force[3*qi+0]+= qefield[0]*finfo->qm_q[qi];
                finfo->qm_force[3*qi+1]+= qefield[1]*finfo->qm_q[qi];
                finfo->qm_force[3*qi+2]+= qefield[2]*finfo->qm_q[qi];
**/
                correct_elecvdw_restqm( finfo->kolink+3*li,
                                        finfo->ko+3*qi,
                                        finfo->qm_q[li],li,
                                        global_data,
                                        finfo);

                /*** now we shift the link atom charge to the ***/
                /*** QM atom to which it is connected         ***/
                finfo->qm_q[qi]+= finfo->qm_q[li];

                /*** now we split up the force on the link atom in a contribution ***/
                /*** parallel to the bond and perpendicular to the bond           ***/
                temp = dbond[0]*finfo->qm_force[3*li+0]+
                       dbond[1]*finfo->qm_force[3*li+1]+
                       dbond[2]*finfo->qm_force[3*li+2];
                flbond[0] = temp*dbond[0];
                flbond[1] = temp*dbond[1];
                flbond[2] = temp*dbond[2];

                flsenk[0] = finfo->qm_force[3*li+0]-flbond[0];
                flsenk[1] = finfo->qm_force[3*li+1]-flbond[1];
                flsenk[2] = finfo->qm_force[3*li+2]-flbond[2];
#if DEBUG_LINK
txtout("TEST LINK       : flbond=%f %f %f  flsenk=%f %f %f\n",
flbond[0],flbond[1],flbond[2],
flsenk[0],flsenk[1],flsenk[2]);
#endif

                /*** now we scale down the perpendicular contribution ***/
                finfo->qm_force[3*li+0] = flbond[0] + flsenk[0]*dcl/dcc;
                finfo->qm_force[3*li+1] = flbond[1] + flsenk[1]*dcl/dcc;
                finfo->qm_force[3*li+2] = flbond[2] + flsenk[2]*dcl/dcc;

                /*** add rest of the perpendicular contribution to qi ***/
                finfo->qm_force[3*qi+0] += flsenk[0]*(1.0-dcl/dcc);
                finfo->qm_force[3*qi+1] += flsenk[1]*(1.0-dcl/dcc);
                finfo->qm_force[3*qi+2] += flsenk[2]*(1.0-dcl/dcc);

                finfo->sumlinkforce[0] += finfo->qm_force[3*li+0];
                finfo->sumlinkforce[1] += finfo->qm_force[3*li+1];
                finfo->sumlinkforce[2] += finfo->qm_force[3*li+2];

                finfo->qm_force[3*li+0]+= finfo->elecf[4*li+0];
                finfo->qm_force[3*li+1]+= finfo->elecf[4*li+1];
                finfo->qm_force[3*li+2]+= finfo->elecf[4*li+2];

#if DEBUG_LINK
txtout("TEST LINK final : q %f %f %f  L %f %f %f\n",
finfo->qm_force[3*qi+0],finfo->qm_force[3*qi+1],finfo->qm_force[3*qi+2],
finfo->qm_force[3*li+0],finfo->qm_force[3*li+1],finfo->qm_force[3*li+2]);
#endif

        }

        for (li=0;li<global_data->nr_qmatoms;li++) {
                if (!(finfo->flag[li]&FLAG_ISLINKATOM)) continue;
                finfo->qm_q[li] = -999.9;/*** signal for the rest of the program   ***/
                                         /*** that we have treated the link charge ***/
        }

        /******************************************************************/
        /****** now all the link atom corrections are finished ************/
        /******************************************************************/

        /******************************************************************/
        /*** now we append dipole-moments to the dipol.out-file         ***/
        /******************************************************************/
        if ((hqmtraj=fopen(fndipol,"aw"))==NULL) { /*** open for appending ***/
                errout("Fatal error: Can not open QM-dipol file <%s>.\n",
                        fndipol);
                pex(-1);
        }

        fprintf(hqmtraj,"%6d %# 10.5f %# 10.5f %# 10.5f %# 10.5f  %# 10.5f %# 10.5f %# 10.5f %# 10.5f\n",
                global_data->cycles_passed,
                efield[0]*FORCE_AU_TO_KCAL,
                efield[1]*FORCE_AU_TO_KCAL,
                efield[2]*FORCE_AU_TO_KCAL,
                efield[3]*FORCE_AU_TO_KCAL,
                finfo->dipol[0],
                finfo->dipol[1],
                finfo->dipol[2],
                finfo->dipol[3]);

        global_data->dipol[0]=finfo->dipol[0];
        global_data->dipol[1]=finfo->dipol[1];
        global_data->dipol[2]=finfo->dipol[2];

        fflush(hqmtraj);
        fclose(hqmtraj);

#if DEBUG
txtout("Reading of QM-output completed.\n");
#endif
        return 0;
}

/**************************************************************************/
/*** parses the QM template file makes some basic checks.               ***/
/*** patches the cell parameters and atomic positions.                  ***/
/*** adds all information about the external electrostatic field.       ***/
/**************************************************************************/
static int
patch_qm_template(      LOAD_MSG_ALL_NODE *global_data,
                        STRUCT_QM_INFO *finfo,
                        int add_elec)
{
        FILE *htempl,*hout;
        char *ptr,fname[MAX_STRING_SIZE],sdum[MAX_STRING_SIZE];
        int i;

        /*** open QM template file for reading. The template file has   ***/
        /*** always the extension .tmpl and is located in the lis-direc.***/
        strcpy(fname,global_data->lispath);

        ptr = strstr(global_data->qmstartstr,".run");
        if (ptr==NULL) {
                errout("Fatal error: Argument for QM-start wrong! <%s>\n",
                        global_data->qmstartstr);
                errout("The start-file must have extension '.run'.\n");
                pex(-1);
        }
        ptr[0] = 0;     /*** replace "." by end of string ***/
        for (i=1;i<strlen(global_data->qmstartstr);i++) {
                if (isspace((int)ptr[-i])) break;
        }
        strcat(fname,ptr-i+1);
        strcat(fname,".tmpl");

        if ((htempl=fopen(fname,"r"))==NULL) {
                errout("Fatal error: Can not open QM-template file <%s>.\n",
                        fname);
                errout("This template file must be located in the directory in which\n");
                errout("the lis-files (e.g., ctl.lis) are located.\n");
                pex(-1);
        }

        interfacemode=-1;
        if (read_string(htempl,sdum)!=-1) {
                sscanf(sdum,"%s",fname);
                if (fname[0]=='&')      interfacemode=CPMD_MODE;
                else                    interfacemode=COLUMBUS_MODE;
        }
        fseek(htempl,0L,SEEK_SET);

        /*** create new input-file for QM-program ***/
        strcpy(fname,global_data->outpath);
        strcat(fname,ptr-i+1);
        strcpy(fncarparrun,ptr-i+1);
        strcat(fncarparrun,".run");
        strcat(fname,".run");
        ptr[0] = '.';

        /*** count on the .run-file for debug ***/
        if (global_data->debug_flags[1]&2) {
                sprintf(sdum,".%d",global_data->cycles_passed);
                strcat(fname,sdum);
        }

        remove_file(fname,global_data->outpath);
        if ((hout=fopen(fname,"w"))==NULL) {
                errout("Fatal error: Can not create QM-input file <%s>.\n",
                        fname);
                pex(-1);
        }

        curqmmap = 0;   /*** important! ***/

        switch (interfacemode) {
                case CPMD_MODE:
                        patch_CPMD(htempl,hout,global_data,finfo,add_elec);
                        break;
                case COLUMBUS_MODE:
                        patch_COLUMBUS(htempl,hout,global_data,finfo,add_elec);
                        break;
                default:
                        errout("Fatal error: Unknown template file for QM-Interface.\n");
                        pex(-1);
        }

        fclose(htempl);
        fclose(hout);

        return 0;
}

static int
patch_COLUMBUS( FILE *htempl,FILE *hout,
                LOAD_MSG_ALL_NODE *global_data,
                STRUCT_QM_INFO *finfo,
                int add_elec)
{
        int i;
        char sdum[MAX_STRING_SIZE];

        i=0;                            /*** number of atoms already patched ***/
        while (read_string(htempl,sdum)!=-1) {

                if (sdum[0]=='#') { /*** modify line ***/
                        double qmx,qmy,qmz;
/***
                        int  atom_id  = finfo->atom_id[i];
                        char atomchar = global_data->atomdata[atom_id].atomstr[0];
***/

                        if (i>=finfo->len) {
                                errout("Fatal error: Number of QM atoms inconsistent with\n");
                                errout("             COLUMBUS template file.\n");
                                pex(-1);
                        }

                        if (finfo->flag[i]&FLAG_ISLINKATOM) {
                                /*** in case of a link atom we write out a corrected ***/
                                /*** atom position stored in finfo->link             ***/

                                qmx = (finfo->kolink[3*i+0] - finfo->offs[0])*ANGS_TO_BOHR;
                                qmy = (finfo->kolink[3*i+1] - finfo->offs[1])*ANGS_TO_BOHR;
                                qmz = (finfo->kolink[3*i+2] - finfo->offs[2])*ANGS_TO_BOHR;
                        } else {
                                qmx = (finfo->ko[3*i+0]   - finfo->offs[0])*ANGS_TO_BOHR;
                                qmy = (finfo->ko[3*i+1]   - finfo->offs[1])*ANGS_TO_BOHR;
                                qmz = (finfo->ko[3*i+2]   - finfo->offs[2])*ANGS_TO_BOHR;
                        }

                        fprintf(hout,"%#13.7f %#13.7f %#13.7f\n",qmx,qmy,qmz);
                        finfo->qmmap[i] = curqmmap++;
                        i++;
                } else { /*** don't modify line ***/
                        fprintf(hout,"%s\n",sdum);
                }
        }
        if (i>=finfo->len) {
                errout("Fatal error: Number of QM atoms inconsistent with\n");
                errout("             COLUMBUS template file.\n");
                pex(-1);
        }
        return 0;
}

/**********************************************************/
/*** parses the CPMD-template-file and makes changes    ***/
/*** like inserting the atom-positions                  ***/
/**********************************************************/
static int
patch_CPMD(     FILE *htempl,FILE *hout,
                LOAD_MSG_ALL_NODE *global_data,
                STRUCT_QM_INFO *finfo,
                int add_elec)
{
        char *sdum,line[MAX_STRING_SIZE],keyword[MAX_STRING_SIZE];
        int cursection=SEC_NO,i;
        int wasthere_IFEGO=0;
        int wasthere_CENTEROFF=0;
        int wasthere_CHARGE=0;
        int firsttime=1;

        /**********************************************************/
        /*** most things from the template file are copied,     ***/
        /*** only atom positions, bounding-boxes, MM atoms      ***/
        /*** every integration step.                            ***/
        /**********************************************************/
        while (read_string(htempl,line)!=-1) {

                /*** skip white spaces at the beginning ***/
                for (i=0,sdum=line;line[i]!=0;i++)
                        if (!isspace((int)line[i])) {sdum=line+i;break;}

                sscanf(sdum,"%s",keyword);

                if (strcmp(keyword,"&END")==0) {
                        if (    (cursection==SEC_SYSTEM) &&
                                (wasthere_CHARGE==0) &&
                                (global_data->qmcharge>-900.0) ) {
                                fprintf(hout,"CHARGE\n");
                                fprintf(hout,"%d.0\n",(int)(global_data->qmcharge));
                        }
                        if (    (cursection==SEC_CPMD) &&
                                (do_qm_restart==1)) {
                                fprintf(hout,"RESTART WAVEFUNCTION\n");
                                do_qm_restart=0;
                        }
                        if (    (cursection==SEC_CPMD) &&
                                (do_qm_restart==2)) {
                                fprintf(hout,"RESTART WAVEFUNCTION LATEST\n");
                                do_qm_restart=0;
                        }
                        cursection=SEC_NO;
                } else
                if (strcmp(keyword,"&CPMD")==0) {
                        cursection=SEC_CPMD;
                } else
                if (strcmp(keyword,"&SYSTEM")==0) {
                        cursection=SEC_SYSTEM;
                } else
                if (strcmp(keyword,"&ATOMS")==0) {
                        cursection=SEC_ATOMS;
                } else
                if (keyword[0]=='&') {
                        cursection=SEC_OTHER;
                }

                /*** yust copy line in the following cases ***/
                if (    (cursection==SEC_OTHER) ||
                        (cursection==SEC_NO)    ) {
                        fprintf(hout,"%s\n",sdum);
                } else
                if (cursection==SEC_CPMD) {
                        fprintf(hout,"%s\n",sdum);

                        if (strstr(sdum,"INTERFACE EGO ESP")!=NULL)
                                wasthere_IFEGO=1;

                        if (strstr(sdum,"INTERFACE EGO")!=NULL)
                                wasthere_IFEGO=1;

                        if (strcmp(sdum,"MOLECULE CENTER OFF")==0)
                                wasthere_CENTEROFF=1;

                } else
                if (cursection==SEC_SYSTEM) {
                        if (strcmp(keyword,"CELL")==0) {
                                double dx,dy,dz;

                                fprintf(hout,"%s\n",sdum);

                                /*** read next line in order to skip it ***/
                                read_string(htempl,sdum);
                                dx = global_data->qmboxext[1]-global_data->qmboxext[0];
                                dy = global_data->qmboxext[3]-global_data->qmboxext[2];
                                dz = global_data->qmboxext[5]-global_data->qmboxext[4];
                                fprintf(hout,"%# 12.7f %# 12.7f %# 12.7f    0.0  0.0  0.0\n",
                                        dx*ANGS_TO_BOHR,
                                        dy/dx,
                                        dz/dx);
                        } else
                        if ( (strcmp(sdum,"CUTOFF")==0) &&
                             (global_data->qmcutoff>0) ) {

                                fprintf(hout,"%s\n",sdum);

                                /*** read next line in order to skip it ***/
                                read_string(htempl,sdum);

                                /*** write cutoff-value to file ***/
                                fprintf(hout,"%d\n",global_data->qmcutoff);
                        } else
                        if ( (strcmp(sdum,"CHARGE")==0) &&
                             (global_data->qmcharge>-900.0) ) {
                                fprintf(hout,"%s\n",sdum);

                                /*** read next line in order to skip it ***/
                                read_string(htempl,sdum);

                                /*** write charge-value to file ***/
                                fprintf(hout,"%d.0\n",(int)(global_data->qmcharge));
                                wasthere_CHARGE=1;
                        } else {
                                fprintf(hout,"%s\n",sdum);
                        }
                } else
                if (cursection==SEC_ATOMS) {
                        if (sdum[0]=='*') {
                                int res=gather_qmatoms( hout,htempl,sdum,
                                                        global_data,
                                                        finfo,firsttime);
                                firsttime=0;
                                /*** skip next line if res==0 as   ***/
                                /*** there is no atom of this type ***/
                                if (res==0) read_string(htempl,sdum);
                        } else {
                                fprintf(hout,"%s\n",sdum);
                        }
                }
        }

        /****** here we add electrostatic information ********/
        if (add_elec) {
                fprintf(hout,"&EXTELEC\n");
                writeout_qmelec(hout,global_data,finfo);
                fprintf(hout,"&END\n");
        }

        if (!wasthere_IFEGO) {
                errout("Fatal error: Missing 'INTERFACE EGO' or 'INTERFACE EGO ESP'\n");
                errout("in the CPMD-template file!\n");
                pex(-1);
        }

        if (!wasthere_CENTEROFF) {
                errout("Fatal error: Missing 'MOLECULE CENTER OFF'\n");
                errout("in the CPMD-template file!\n");
                pex(-1);
        }

        return 0;
}

/***********************************************************************/
/*** this function extracts all atoms of one physical kind from EGO's***/
/*** QM-atom list and writes out its coordinates to the cpmd-input-f ***/
/***********************************************************************/
static int
gather_qmatoms( FILE *hout,FILE *htempl,
                char *sdum,
                LOAD_MSG_ALL_NODE *global_data,
                STRUCT_QM_INFO *finfo,
                int firsttime)
{
        int nr_atoms=0;
        char atomstr[10],resortfn[MAX_STRING_SIZE];
        int i,slen=strlen(sdum);
        FILE *resh;

        /*** create a file usefull for ego2pdb resorting ***/
        strcpy(resortfn,global_data->outpath);
        strcat(resortfn,"resort.txt");
        if (firsttime) {
                resh = fopen(resortfn,"w");
                fprintf(resh,"[FREEFORMAT1]\n");
                fprintf(resh,"%d 1 1\n",finfo->len);
                fprintf(resh,"%%*d %%*c %%lf %%lf %%lf %%*lf %%*lf %%*lf\n");
        } else  resh = fopen(resortfn,"a");

        /*** the first letters up to the '_' of this string determains  ***/
        /*** the physical atom type, e.g. H, C, N, O, P                 ***/
        for (i=0;(i<slen)&&(sdum[i+1]!='_');i++) {
                atomstr[i]  =sdum[i+1];
                atomstr[i+1]=0;
        }

        for (i=0,nr_atoms=0;i<finfo->len;i++) {
                int  atom_id  = finfo->atom_id[i];
                char atomchar = global_data->atomdata[atom_id].atomstr[0];

                /*** special treatment of link-atoms ***/
                if (finfo->flag[i]&FLAG_ISLINKATOM) {
                        atomchar='H';
                }

                if (atomchar==atomstr[0]) nr_atoms++;
        }
#if DEBUG
        txtout("Nr of %c is %2d\n",atomstr[0],nr_atoms);
#endif
        if (nr_atoms==0) {
                fclose(resh);
                return 0;
        }
        fprintf(hout,"%s\n",sdum);
        /*** copy next line ***/
        read_string(htempl,sdum);
        fprintf(hout,"%s\n",sdum);

        fprintf(hout,"%2d\n",nr_atoms);
        for (i=0;i<finfo->len;i++) {
                int  j,atom_id  = finfo->atom_id[i];
                char atomchar = global_data->atomdata[atom_id].atomstr[0];
                double  ko[3];

                for (j=0;j<finfo->qmlink_len;j++)
                        if (    (atom_id==finfo->qmlink_p2[j]) &&
                                (finfo->qmlink_dist1[j]<=0.0) ) {
                                atomchar='X';
                                break;
                        }

                /*** for the CPMD programm all atoms of one kind have to***/
                /*** be written to file consecutive. Therefore I need a ***/
                /*** index table (qmmap) which enables me to sort the   ***/
                /*** output of CPMD back to the EGO finfo structure     ***/

                if (    (finfo->flag[i]&FLAG_ISLINKATOM) &&
                        (atomstr[0]=='H') &&
                        (atomchar!='X') ) {
                        /*** in case of a link atom we write out a corrected **/
                        /*** atom position stored in finfo->kolink ***/

                        rotforw(finfo->kolink+3*i,ko);
                        ko[0] = (ko[0] - finfo->offs[0])*ANGS_TO_BOHR;
                        ko[1] = (ko[1] - finfo->offs[1])*ANGS_TO_BOHR;
                        ko[2] = (ko[2] - finfo->offs[2])*ANGS_TO_BOHR;
                } else
                if (    (atomchar==atomstr[0]) &&
                        (!(finfo->flag[i]&FLAG_ISLINKATOM)) ) {

                        rotforw(finfo->ko+3*i,ko);
                        ko[0] = (ko[0]  - finfo->offs[0])*ANGS_TO_BOHR;
                        ko[1] = (ko[1]  - finfo->offs[1])*ANGS_TO_BOHR;
                        ko[2] = (ko[2]  - finfo->offs[2])*ANGS_TO_BOHR;
                } else {
                        continue;
                }


                fprintf(resh,"%d\n",atom_id+1);         /*** for the resort file ***/
                fprintf(hout,"%# 12.7f %# 12.7f %# 12.7f\n",ko[0],ko[1],ko[2]);
                finfo->qmmap[i] = curqmmap++;
        }
        fclose(resh);
        return 1;
}

/******************************************************************/
/*** this function adds all information about the extern        ***/
/*** electrostatic field to the CPMD input file.                ***/
/******************************************************************/
static  double  fieldfac = -99999.9;

static int
writeout_qmelec(FILE *forceh,
                LOAD_MSG_ALL_NODE *global_data,
                STRUCT_QM_INFO *finfo)
{
        int i,j;

        fprintf(forceh,"QM-Box\n%# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f\n",
                0.0,(global_data->qmboxext[1]-global_data->qmboxext[0])*ANGS_TO_BOHR,
                0.0,(global_data->qmboxext[3]-global_data->qmboxext[2])*ANGS_TO_BOHR,
                0.0,(global_data->qmboxext[5]-global_data->qmboxext[4])*ANGS_TO_BOHR);

        fprintf(forceh,"Bounding-Box\n%# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f\n",
                0.0,(global_data->qmbox[1]-global_data->qmbox[0])*ANGS_TO_BOHR,
                0.0,(global_data->qmbox[3]-global_data->qmbox[2])*ANGS_TO_BOHR,
                0.0,(global_data->qmbox[5]-global_data->qmbox[4])*ANGS_TO_BOHR);

        /**************************************************/
        /*** write out the local expansions             ***/
        /**************************************************/
        if (global_data->qmnoextelec) {
                fprintf(forceh,"Taylorexpansions\n%d\n",0);
                fprintf(forceh,"MM-Atoms\n%d\n",0);
                return 0;
        }

        /**********************************************************/
        /*** this is a special mode for the external field:     ***/
        /*** here we use one local expansion to generate a      ***/
        /*** homogenious E-field.                               ***/
        /**********************************************************/
        if (global_data->debug_flags[1]&8) {
                if (fieldfac<-99999.0) {
                        fieldfac  = global_data->userdouble[1];
                } else {
                        fieldfac += global_data->userdouble[0];
                }

                txtout("Special Mode, External Field: %f %f %f kcal/mol A e\n",
                        fieldfac*global_data->userdouble[2],
                        fieldfac*global_data->userdouble[3],
                        fieldfac*global_data->userdouble[4]);

                fprintf(forceh,"Taylorexpansions\n1\n");
                fprintf(forceh,"%-8d %# 12.7f %# 12.7f %# 12.7f\n",
                        1,
                        0.0,0.0,0.0);

                fprintf(forceh,"%# 12.7f\n%# 12.7f %# 12.7f %# 12.7f\n",
                        0.0,     /*** Phi_0    ***/
                        fieldfac*global_data->userdouble[2]/FORCE_AU_TO_KCAL,   /*** Gradient ***/
                        fieldfac*global_data->userdouble[3]/FORCE_AU_TO_KCAL,
                        fieldfac*global_data->userdouble[4]/FORCE_AU_TO_KCAL);

                fprintf(forceh,"%# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f\n",
                        0.0/HESS_AU_TO_KCAL, /*** Hesse Matrix ***/
                        0.0/HESS_AU_TO_KCAL,
                        0.0/HESS_AU_TO_KCAL,
                        0.0/HESS_AU_TO_KCAL,
                        0.0/HESS_AU_TO_KCAL,
                        0.0/HESS_AU_TO_KCAL);

                fprintf(forceh,"%-2d\n",0);
                fprintf(forceh,"\n");
                fflush(forceh);

        } else {
        fprintf(forceh,"Taylorexpansions\n%d\n",finfo->len-finfo->qmlink_len);
        for (i=0;i<finfo->len;i++) {
                double ko[3],egrad[3],hess[3][3];

                if (finfo->flag[i]&FLAG_ISLINKATOM) continue;

                rotforw(finfo->ko+3*i,ko);
                ko[0]=(ko[0]-finfo->offs[0])*ANGS_TO_BOHR;
                ko[1]=(ko[1]-finfo->offs[1])*ANGS_TO_BOHR;
                ko[2]=(ko[2]-finfo->offs[2])*ANGS_TO_BOHR;

                egrad[0]=finfo->local[10*i+1]/FORCE_AU_TO_KCAL;
                egrad[1]=finfo->local[10*i+2]/FORCE_AU_TO_KCAL;
                egrad[2]=finfo->local[10*i+3]/FORCE_AU_TO_KCAL;
                rotforw(egrad,egrad);

                hess[0][0] = finfo->local[10*i+4]/HESS_AU_TO_KCAL;
                hess[1][1] = finfo->local[10*i+5]/HESS_AU_TO_KCAL;
                hess[2][2] = finfo->local[10*i+6]/HESS_AU_TO_KCAL;
                hess[1][0] = hess[0][1] = finfo->local[10*i+7]/HESS_AU_TO_KCAL;
                hess[2][0] = hess[0][2] = finfo->local[10*i+8]/HESS_AU_TO_KCAL;
                hess[2][1] = hess[1][2] = finfo->local[10*i+9]/HESS_AU_TO_KCAL;
                rothess(hess);

                /*** the positions of the local expansions ***/
                fprintf(forceh,"%-8d %# 12.7f %# 12.7f %# 12.7f\n",
                        finfo->atom_id[i]+1,
                        ko[0],ko[1],ko[2]);

                /*** the local expansion coefficients ***/
                fprintf(forceh,"%# 12.7f\n%# 12.7f %# 12.7f %# 12.7f\n",
                        finfo->local[10*i+0]*KCAL_TO_HART,      /*** Phi_0    ***/
                        egrad[0],                               /*** Gradient ***/
                        egrad[1],
                        egrad[2]);

                fprintf(forceh,"%# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f\n",
                        hess[0][0],                             /*** Hesse Matrix ***/
                        hess[1][1],
                        hess[2][2],
                        hess[0][1],
                        hess[0][2],
                        hess[1][2]);

                /*** the dummy-atoms which have to be included in the   ***/
                /*** calculation of the electrostatic potential         ***/

                if (finfo->flag[i]&FLAG_ISLINKATOM) {
                        int kk;

                        for (kk=0;kk<finfo->cur_nr_dummys;kk++) {
                                if (finfo->dummy_id[kk]==finfo->atom_id[i]) {
                                        finfo->incllist[i*MAXDUMMYSPERQMATOM + finfo->incllen[i]]=finfo->atom_id[i];
                                        finfo->excltype[kk]=3; /*** flag for link-atom ***/
                                        finfo->incllen[i]++;
                                        break;
                                }
                        }
                        if (kk==finfo->cur_nr_dummys) {
                                errout("Fatal error: confusion, exclusion atom %d not in list!\n",
                                        finfo->atom_id[i]+1);
                                pex(-1);
                        }
                }

                fprintf(forceh,"%-2d\n",finfo->incllen[i]);
                for (j=0;j<finfo->incllen[i];j++) {
                        int k,atom_id = finfo->incllist[i*MAXDUMMYSPERQMATOM + j];

                        for (k=0;k<finfo->cur_nr_dummys;k++)
                                if (finfo->dummy_id[k]==atom_id) break;

                        if (k>=finfo->cur_nr_dummys) {
                                errout("Fatal error: that can not be in writeout_qmelec %d %d.\n",
                                        atom_id,k);
                                for (k=0;k<finfo->cur_nr_dummys;k++)
                                        txtout("k=%d   %d\n",k,finfo->dummy_id[k]);
                                pex(-1);
                        }

                        fprintf(forceh,"%-6d ",k);

                        /*** always 10 numbers per line ***/
                        if (((j+1)%10)==0) fprintf(forceh,"\n");
                }

                if (finfo->flag[i]&FLAG_ISLINKATOM) finfo->incllen[i]--;

                fprintf(forceh,"\n");
                fflush(forceh);
        }
        }

        /**************************************************/
        /*** write out the dummy atoms                  ***/
        /**************************************************/
        fprintf(forceh,"MM-Atoms\n%d\n",finfo->cur_nr_dummys);
        for (i=0;i<finfo->cur_nr_dummys;i++) {
                int j = finfo->dummy_id[i];
                double curq = finfo->dummy_q[i];
                double cutoffwidth,cutoffdist;
                double ko[3];

                curq = finfo->dummy_q[i] * global_data->userdouble[0];

                /*** set proper values for exclusion cutoffs ***/
                if (finfo->excltype[i]==0) {
                        cutoffdist  = 0.0;
                        cutoffwidth = 0.0;
                } else {
                        cutoffdist  = global_data->qmexcl[0];
                        cutoffwidth = global_data->qmexcl[1];

                        if (    (finfo->excltype[i]==3) &&
                                (cutoffdist+cutoffwidth<2.0) ) {
                                cutoffdist = 1.0;
                                cutoffwidth= 1.0;
                        }
                }
                rotforw(finfo->dummy_ko+3*i,ko);
                ko[0] = (ko[0]-finfo->offs[0])*ANGS_TO_BOHR;
                ko[1] = (ko[1]-finfo->offs[1])*ANGS_TO_BOHR;
                ko[2] = (ko[2]-finfo->offs[2])*ANGS_TO_BOHR;

                fprintf(forceh,"%-8d %# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f %# 12.7f\n",
                        j+1,                            /*** PDB-atom number    ***/
                        curq,                           /*** partial charge     ***/
                        cutoffwidth*ANGS_TO_BOHR,       /*** special values for ***/
                        cutoffdist*ANGS_TO_BOHR,        /*** excluded atoms     ***/
                        finfo->qcore[i],                /*** charge of the core ***/
                        finfo->asize[i]*ANGS_TO_BOHR,   /*** extension of E-Dens***/
                        ko[0],
                        ko[1],
                        ko[2]);
        }

        return 0;
}

/*** here I sort forces comming from the nodes by a bubblesort  ***/
/*** to increasing numbers. This function is called only on the ***/
/*** master node or by the controller in sequentiel version.    ***/
int
sort_to_qminfo( int atom_id,
                int nodeatomindex,int node_number,
                STRUCT_QM_INFO *finfo,double *data,
                int incllen,int *incllist)
{
        int     i=0;

        if (finfo->cur_len>=finfo->len) {
                errout("Fatal error: Inconsistent number (%d>=%d) of QM-atoms in sort_to_qminfo!\n",
                        finfo->cur_len,finfo->len);
                pex(-1);
        }

        /*** search were to insert the new qm-data:     ***/
        /*** a kind of bubble sort!                     ***/
        if (finfo->cur_len==0) i = 0;
        else
        for (i=0;i<finfo->cur_len;i++)
                if (atom_id<finfo->atom_id[i]) break;

        if (i<finfo->cur_len) { /*** make a gap ***/
                int j;
                for (j=finfo->cur_len;j>=i+1;j--) {
                        finfo->atom_id[j]   = finfo->atom_id[j-1];
                        finfo->nodeatomi[j] = finfo->nodeatomi[j-1];
                        finfo->node[j]      = finfo->node[j-1];
                        finfo->flag[j]      = finfo->flag[j-1];
                        finfo->incllen[j]   = finfo->incllen[j-1];

                        memcpy( finfo->ko+3*j,
                                finfo->ko+3*(j-1),
                                3*sizeof(double));

                        memcpy( finfo->elecf+4*j,
                                finfo->elecf+4*(j-1),
                                4*sizeof(double));

                        memcpy( finfo->local+10*j,
                                finfo->local+10*(j-1),
                                10*sizeof(double));

                        memcpy( finfo->bondf+4*j,
                                finfo->bondf+4*(j-1),
                                4*sizeof(double));

                        memcpy( finfo->anglef+4*j,
                                finfo->anglef+4*(j-1),
                                4*sizeof(double));

                        memcpy( finfo->dihef+4*j,
                                finfo->dihef+4*(j-1),
                                4*sizeof(double));

                        memcpy( finfo->imprf+4*j,
                                finfo->imprf+4*(j-1),
                                4*sizeof(double));

                        memcpy( finfo->incllist+MAXDUMMYSPERQMATOM*j,
                                finfo->incllist+MAXDUMMYSPERQMATOM*(j-1),
                                MAXDUMMYSPERQMATOM*sizeof(int));

                }
        } else {                /*** add it to the end ***/
                i=finfo->cur_len;
        }

        finfo->atom_id[i]   = atom_id;
        finfo->nodeatomi[i] = nodeatomindex;
        finfo->node[i]      = node_number;
        finfo->flag[i]      = 0;
        finfo->incllen[i]   = incllen;
        finfo->ko   [3*i  ] = data[0];  /*** x ***/
        finfo->ko   [3*i+1] = data[1];  /*** y ***/
        finfo->ko   [3*i+2] = data[2];  /*** z ***/

        /*** electrostatic at QM atom, here we switch the hidden EGO-sign ***/
        finfo->elecf[4*i  ] = -data[3]; /*** elec_fx ***/
        finfo->elecf[4*i+1] = -data[4]; /*** elec_fy ***/
        finfo->elecf[4*i+2] = -data[5]; /*** elec_fz ***/
        finfo->elecf[4*i+3] = data[6]; /*** charge  ***/

        memcpy(finfo->local+10*i,data+7,10*sizeof(double));

        /*** bonding forces, here we switch the hidden EGO-sign ***/
        finfo->bondf[4*i  ] = data[17];  /*** bond_E  ***/
        finfo->bondf[4*i+1] = -data[18]; /*** bond_fx ***/
        finfo->bondf[4*i+2] = -data[19]; /*** bond_fy ***/
        finfo->bondf[4*i+3] = -data[20]; /*** bond_fz ***/

        /*** angle forces, here we switch the hidden EGO-sign ***/
        finfo->anglef[4*i  ] = data[21];  /*** angele_E  ***/
        finfo->anglef[4*i+1] = -data[22]; /*** angle_fx ***/
        finfo->anglef[4*i+2] = -data[23]; /*** angle_fy ***/
        finfo->anglef[4*i+3] = -data[24]; /*** angle_fz ***/

        /*** dihe forces, here we switch the hidden EGO-sign ***/
        finfo->dihef[4*i  ] = data[25];  /*** dihe_E  ***/
        finfo->dihef[4*i+1] = -data[26]; /*** dihe_fx ***/
        finfo->dihef[4*i+2] = -data[27]; /*** dihe_fy ***/
        finfo->dihef[4*i+3] = -data[28]; /*** dihe_fz ***/

        /*** impr forces, here we switch the hidden EGO-sign ***/
        finfo->imprf[4*i  ] = data[29];  /*** impr_E  ***/
        finfo->imprf[4*i+1] = -data[30]; /*** impr_fx ***/
        finfo->imprf[4*i+2] = -data[31]; /*** impr_fy ***/
        finfo->imprf[4*i+3] = -data[32]; /*** impr_fz ***/

        memcpy(finfo->incllist+MAXDUMMYSPERQMATOM*i,incllist,incllen*sizeof(int));

        finfo->cur_len++;
        return 0;
}

/******************************************************************/
/*** here we store in the my_data->qmlocal structure            ***/
/******************************************************************/
int
get_qmnode_index(int atom_id,LOAD_MSG_ONE_NODE *my_data)
{
        int i;
/***
errout("in get_qmnode_index %d %d %d\n",
atom_id,my_data->qmlocal.cur_len,my_data->qmlocal.len);
***/
        if (my_data->qmlocal.cur_len==0) {
                my_data->qmlocal.atom_id[0] = atom_id;
                my_data->qmlocal.cur_len++;
                return 0;
        }

        /*** search in list and return index if found ***/
        for (i=0;i<my_data->qmlocal.cur_len;i++) {
                if (my_data->qmlocal.atom_id[i]==atom_id) return i;
        }

        /*** not found in list --> enter it to list ***/
        if (my_data->qmlocal.cur_len>=my_data->qmlocal.len) {
                errout("Fatal error: Not enough space in get_qmnode_index (%d) len=%d.\n",
                        atom_id,my_data->qmlocal.cur_len);
                for (i=0;i<my_data->qmlocal.cur_len;i++) {
                        txtout("TEST %d %d\n",i,my_data->qmlocal.atom_id[i]);
                }
                pex(-1);
        }

        my_data->qmlocal.atom_id[my_data->qmlocal.cur_len] = atom_id;
        my_data->qmlocal.cur_len++;

        return my_data->qmlocal.cur_len-1;
}

int
calculate_qmelec(int my_number,int my_atom,int my_unitnum,
                int other_atom,int other_unitnum,
                int copy_flag,
                LOAD_MSG_ALL_NODE *global_data,
                LOAD_MSG_ONE_NODE *my_data,
                TRAVELLING_PACKETS *packet)
{
        double  dx,dy,dz;
        double  dx2,dy2,dz2,r_2;
        double  QQ=0.0,f,f1;
        double  *local;
        int     index=my_number;
        int     my_qm   = my_data->atom_sele_flag1[my_atom]&FLAG_QMATOM;
        int     other_qm= packet->coordinates.atom_sele_flag1[other_atom]&FLAG_QMATOM;

        /********************************************************/
        /*** Normal calculation if both are classical atoms   ***/
        /********************************************************/
        if ( (!other_qm) && (!my_qm) )  return CALC_NORMAL;
#if DEBUG
txtout("Entering first calculate_qmelec: %d(%d) %d(%d)\n",my_qm,my_atom,other_qm,other_atom);
#endif
        /********************************************************/
        /*** No electrostatics+VDW if both atoms are QM-Atoms ***/
        /*** but not a link atom                              ***/
        /********************************************************/
        if ( (other_qm) && (my_qm) ) {
                if (my_data->atom_sele_flag1[my_atom]&FLAG_QMLINK)               my_qm    = 0;
                if (packet->coordinates.atom_sele_flag1[other_atom]&FLAG_QMLINK) other_qm = 0;
                if ( (other_qm) && (my_qm) ) return SKIP_ELECVDW;

                /*** both are link-atoms! ***/
                if ( (!other_qm) &&(!my_qm) ) my_qm = 1;
        }
#if DEBUG
txtout("Entering second calculate_qmelec: %d(%d) %d(%d)\n",my_qm,my_atom,other_qm,other_atom);
#endif
        /********************************************************/
        /*** One of the atoms is QM atom, the other a MM-atom ***/
        /********************************************************/
        dx = my_data->x_new[my_atom] - packet->coordinates.x[other_atom];
        dx2=dx*dx;

        dy = my_data->y_new[my_atom] - packet->coordinates.y[other_atom];
        dy2=dy*dy;

        dz = my_data->z_new[my_atom] - packet->coordinates.z[other_atom];
        dz2=dz*dz;

        r_2= 1/(dx2 + dy2 + dz2);
        local = NULL;

                        /***************************************************/
        if (my_qm) {    /*** my_atom is QM-atom, other_atom is classical ***/
                        /***************************************************/
                double dxt,dyt,dzt,dist,offset;
                struct_cluster *pa_cl=&(packet->branch.cluster[0][other_unitnum]);
                if (pa_cl->typ==CHARGED) offset=CHARGEDOFFSET;
                else {
                        double maxm;
                        maxm=max(fabs(pa_cl->moment[1]),fabs(pa_cl->moment[2]));
                        maxm=max(maxm,fabs(pa_cl->moment[3]));
/**
                        txtout("Maxm %f  size %f\n",maxm,pa_cl->cen_geom[3]);
**/
                        offset=0.9;
                        if (maxm>100.0) offset=1.5;
                        if (maxm>200.0) offset=2.9;
                }

                dxt = my_data->x_new[my_atom] - pa_cl->cen_geom[0];
                dyt = my_data->y_new[my_atom] - pa_cl->cen_geom[1];
                dzt = my_data->z_new[my_atom] - pa_cl->cen_geom[2];

                dxt*=dxt;
                dyt*=dyt;
                dzt*=dzt;
                dist = sqrt(dxt+dyt+dzt)+pa_cl->cen_geom[3];

                if (dist>QMLOCALDIST) {
                        dist=get_min_unitdist(  0,
                                                my_atom,other_unitnum,
                                                my_data,packet);
/****
                        txtout("Normal dist %f (- boxsize %f) real %f\n",
                                sqrt(dxt+dyt+dzt),sqrt(dxt+dyt+dzt)+pa_cl->cen_geom[3],dist);
****/
                }

                index = get_qmnode_index(my_data->atom_id[my_atom],my_data);

                /*** if other_atom is close to the QM atom, mark it as  ***/
                /*** a dummy atom and add it to list of local expansion ***/
                if (    (dist<QMLOCALDIST+offset) && (copy_flag) &&
                        (fabs(my_data->q[other_atom])>0.00001)
                   ) {
                        int offset = MAXDUMMYSPERQMATOM*index + my_data->qmlocal.incllen[index];
                        my_data->qmlocal.incllen[index]++;
                        if (my_data->qmlocal.incllen[index]>=MAXDUMMYSPERQMATOM) {
                                errout("Fatal error: MAXDUMMYSPERQMATOM (%d) too small. Increase it in ego.h.\n",
                                        MAXDUMMYSPERQMATOM);
                                pex(-1);
                        }

                        my_data->qmlocal.incllist[offset]    = my_data->atom_id[other_atom];
                        my_data->atom_sele_flag1[other_atom] |= FLAG_QMDUMMY;
#if DEBUG
txtout("TEST1 : local-index %4d  added atom-id %4d    len %4d\n",
        index,my_data->atom_id[other_atom],my_data->qmlocal.incllen[index]);
#endif
                        return CALC_NORMAL;
                }

#if DEBUG
txtout("TEST1a: local-index %4d  added atom-id %4d    len %4d  %f\n",
        index,my_data->atom_id[other_atom],my_data->qmlocal.incllen[index],sqrt(dxt+dyt+dzt));
#endif
                /*** else, add electrostatics to local expansion        ***/
                QQ    = packet->coordinates.q[other_atom]*global_data->f_elstat;
                local = my_data->qmlocal.local + 10*index;

                                /***************************************************/
        } else                  /*** my_atom is classical, other_atom is QM-atom ***/
                                /***************************************************/
        if ( (other_qm) && (copy_flag) ) {
                double dxt,dyt,dzt,dist,offset;
                struct_cluster *pa_cl=&(my_data->cluster[0][my_unitnum]);
                if (pa_cl->typ==CHARGED) offset=CHARGEDOFFSET;
                else {
                        double maxm;
                        maxm=max(fabs(pa_cl->moment[1]),fabs(pa_cl->moment[2]));
                        maxm=max(maxm,fabs(pa_cl->moment[3]));
/**
                        txtout("Maxm %f  size %f\n",maxm,pa_cl->cen_geom[3]);
**/
                        offset=0.9;
                        if (maxm>100.0) offset=1.5;
                        if (maxm>200.0) offset=2.9;
                }

                dxt = packet->coordinates.x[other_atom] - pa_cl->cen_geom[0];
                dyt = packet->coordinates.y[other_atom] - pa_cl->cen_geom[1];
                dzt = packet->coordinates.z[other_atom] - pa_cl->cen_geom[2];

                dxt*=dxt;
                dyt*=dyt;
                dzt*=dzt;

                dist = sqrt(dxt+dyt+dzt)+pa_cl->cen_geom[3];
                index = get_qmnode_index(packet->coordinates.atom_id[other_atom],my_data);

                if (dist>QMLOCALDIST) {
                        dist=get_min_unitdist(  1,
                                                other_atom,my_unitnum,
                                                my_data,packet);
/****
                        txtout("Normal dist %f (- boxsize %f) %f\n",
                                sqrt(dxt+dyt+dzt),sqrt(dxt+dyt+dzt)+pa_cl->cen_geom[3],dist);
****/
                }

                /*** if my_atom is close to the QM atom, mark it as     ***/
                /*** a dummy atom and add it to list of local expansion ***/
                if (    (dist<QMLOCALDIST+offset) &&
                        (fabs(my_data->q[my_atom])>0.00001)
                    ) {
                        int offset = MAXDUMMYSPERQMATOM*index + my_data->qmlocal.incllen[index];
                        my_data->qmlocal.incllen[index]++;
                        if (my_data->qmlocal.incllen[index]>=MAXDUMMYSPERQMATOM) {
                                errout("Fatal error: MAXDUMMYSPERQMATOM (%d) too small\n",
                                        MAXDUMMYSPERQMATOM);
                                pex(-1);
                        }
                        my_data->qmlocal.incllist[offset]    = my_data->atom_id[my_atom];
                        my_data->atom_sele_flag1[my_atom]    |= FLAG_QMDUMMY;
#if DEBUG
txtout("TEST2: local-index %4d  added atom-id %4d    len %4d\n",
        index,my_data->atom_id[my_atom],my_data->qmlocal.incllen[index]);
#endif
                        return CALC_NORMAL;
                }

#if DEBUG
txtout("TEST2a: local-index %4d  added atom-id %4d    len %4d  %f\n",
        index,my_data->atom_id[my_atom],my_data->qmlocal.incllen[index],sqrt(dxt+dyt+dzt));
#endif
                /*** else, add electrostatics to local expansion        ***/
                dx = -dx;
                dy = -dy;
                dz = -dz;

                QQ    = my_data->q[my_atom] * global_data->f_elstat;
                local = my_data->qmlocal.local + 10*index;

        } else          /***************************************************/
        if (other_qm) { /*** my_atom is classical, other_atom is QM-atom ***/
                        /***************************************************/
                double dxt,dyt,dzt,dist;
                struct_cluster *pa_cl=&(my_data->cluster[0][my_unitnum]);

txtout("Not yet implemented\n");
pex(-1);
                dxt = packet->coordinates.x[other_atom] - pa_cl->cen_geom[0];
                dyt = packet->coordinates.y[other_atom] - pa_cl->cen_geom[1];
                dzt = packet->coordinates.z[other_atom] - pa_cl->cen_geom[2];

                dxt*=dxt;
                dyt*=dyt;
                dzt*=dzt;

                dist=sqrt(dxt+dyt+dzt);

                if (dist>QMLOCALDIST) {
                        dist=get_min_unitdist(  1,
                                                other_atom,my_unitnum,
                                                my_data,packet);
/****
                        txtout("Normal dist %f (- boxsize %f) %f\n",
                                sqrt(dxt+dyt+dzt),sqrt(dxt+dyt+dzt)-pa_cl->cen_geom[3],dist);
****/
                }

                index = get_qmnode_index(my_data->atom_id[my_atom],my_data);

                /*** if my_atom is close to a QM atom, mark it as       ***/
                /*** a dummy atom and add it to list of local expansion ***/
                if (    (dist<QMLOCALDIST) &&
                        (fabs(my_data->q[my_atom])>0.00001)
                    ) {
                        int offset = MAXDUMMYSPERQMATOM*index + my_data->qmlocal.incllen[index];
                        my_data->qmlocal.incllen[index]++;
                        if (my_data->qmlocal.incllen[index]>=MAXDUMMYSPERQMATOM) {
                                errout("Fatal error: MAXDUMMYSPERQMATOM (%d) too small. Increase it in ego.h\n",
                                        MAXDUMMYSPERQMATOM);
                                pex(-1);
                        }
                        my_data->qmlocal.incllist[offset]    = my_data->atom_id[my_atom];
                        my_data->atom_sele_flag1[my_atom]    |= FLAG_QMDUMMY;
#if DEBUG
txtout("TEST3: local-index %4d  added atom-id %4d    len %4d\n",
        index,my_data->atom_id[my_atom],my_data->qmlocal.incllen[index]);
#endif
                }
                return CALC_NORMAL;
#if DEBUG
txtout("TEST3a: local-index %4d  added atom-id %4d    len %4d  %f\n",
        index,my_data->atom_id[my_atom],my_data->qmlocal.incllen[index],sqrt(dxt + dyt + dzt));
#endif
        }

        f = QQ * sqrt(r_2);     /*** f = Q/|r|      ****/
        local[0] += f;

        f *= -r_2;              /*** f = -Q/(r^3)   ****/
        local[1] += f * dx;
        local[2] += f * dy;
        local[3] += f * dz;

        f1 = 3.0 * f * r_2;     /*** f1= -3 Q/(r^5) ****/
        local[4] += f - dx2*f1;
        local[5] += f - dy2*f1;
        local[6] += f - dz2*f1;

        local[7] -= dx*dy*f1;
        local[8] -= dx*dz*f1;
        local[9] -= dy*dz*f1;

        return CALC_NORMAL;
}


/************************************************************************/

int
handle_qmbond(  int *onthisnode,
                int *forceout_flag,
                int *myatom_id,
                LOAD_MSG_ONE_NODE *my_data,
                double *bforce,
                int ind1,int ind2)
{
        int isqmatom[2],islink[2],sumqm,sumlink;

        /*** isqmatom[...] is 1 only if it is a QM atom ***/
        isqmatom[0] = (forceout_flag[0]&FLAG_QMATOM) ? 1 : 0;
        isqmatom[1] = (forceout_flag[1]&FLAG_QMATOM) ? 1 : 0;
        sumqm = isqmatom[0] + isqmatom[1];

        /*** islink[...] is 1 only if it is a Link atom ***/
        islink[0]   = (forceout_flag[0]&FLAG_QMLINK) ? 1 : 0;
        islink[1]   = (forceout_flag[1]&FLAG_QMLINK) ? 1 : 0;
        sumlink = islink[0] + islink[1];

        if ((sumqm==0)&&(sumlink==0)) return 0; /*** no energies and forces to patch ***/

#if DEBUG_LINK
txtout("in handle_qmbond: atomids %d %d   QM: %d %d   Link: %d %d\n",
myatom_id[0],myatom_id[1],
isqmatom[0],isqmatom[1],
islink[0],islink[1]);
#endif
        /*********************************************/
        /*** subtract energy contributions in case ***/
        /*** if both atoms are QM atoms            ***/
        /*** Case: Q -- Q or Q -- L                ***/
        /*********************************************/
        if (sumqm==2) {

                /*** if no link atom is involved, subtract energy ***/
                /*** Case: Q -- Q                                 ***/
                if (sumlink==0) {
                        double f=0.0;
                        f=get_bondfactor(myatom_id);
                        if (onthisnode[0]) {
                                int offset = get_qmnode_index(myatom_id[0],my_data)*4;
                                int ii = 4*ind1;
                                my_data->force_bond[ii  ]        -= bforce[0];
                                my_data->qmlocal.bondf[offset  ] += f*bforce[0];
                                my_data->qmlocal.bondf[offset+1] += f*bforce[1];
                                my_data->qmlocal.bondf[offset+2] += f*bforce[2];
                                my_data->qmlocal.bondf[offset+3] += f*bforce[3];
                        }

                        if (onthisnode[1]) {
                                int offset = get_qmnode_index(myatom_id[1],my_data)*4;
                                int ii = 4*ind2;
                                my_data->force_bond[ii  ]        -= bforce[0];
                                my_data->qmlocal.bondf[offset  ] += f*bforce[0];
                                my_data->qmlocal.bondf[offset+1] -= f*bforce[1];
                                my_data->qmlocal.bondf[offset+2] -= f*bforce[2];
                                my_data->qmlocal.bondf[offset+3] -= f*bforce[3];
                        }
                } else
                /*** if one link atom is involved, we need that force ***/
                /*** for positioning of the link atom. We store that  ***/
                /*** force in my_data->qmlocal.bondf[...] of the      ***/
                /*** QM-atom and set it later to zero.                ***/
                /*** Case: Q -- L                                     ***/
                if (sumlink==1) {
                        /*** add the classical bond force to the link atom ***/
                        /*** my_data->qmlocal.bondf has the EGO sign       ***/
                        if ((onthisnode[0])&&(islink[0])) {
                                /*** NOTE: offset comes from the index of the QM-atom!!!! ***/
                                int offset = get_qmnode_index(myatom_id[1],my_data)*4;
                                my_data->qmlocal.bondf[offset  ] = bforce[0];
                                my_data->qmlocal.bondf[offset+1] = -bforce[1];
                                my_data->qmlocal.bondf[offset+2] = -bforce[2];
                                my_data->qmlocal.bondf[offset+3] = -bforce[3];
                        }

                        if ((onthisnode[1])&&(islink[1])) {
                                /*** NOTE: offset comes from the index of the QM-atom!!!! ***/
                                int offset = get_qmnode_index(myatom_id[0],my_data)*4;
                                my_data->qmlocal.bondf[offset  ] = bforce[0];
                                my_data->qmlocal.bondf[offset+1] = bforce[1];
                                my_data->qmlocal.bondf[offset+2] = bforce[2];
                                my_data->qmlocal.bondf[offset+3] = bforce[3];
                        }
                } else
                /******************************************************/
                /*** Case L --- L : for example miniretinal HybrB! ****/
                /******************************************************/
                if (sumlink==2) {
                        int offset = get_qmnode_index(myatom_id[0],my_data)*4;
                        int ii = 4*ind1;
                        my_data->force_bond[ii  ]        += bforce[0];
                        my_data->qmlocal.bondf[offset  ] += bforce[0];
                        my_data->qmlocal.bondf[offset+1] += bforce[1];
                        my_data->qmlocal.bondf[offset+2] += bforce[2];
                        my_data->qmlocal.bondf[offset+3] += bforce[3];

                        offset = get_qmnode_index(myatom_id[1],my_data)*4;
                        ii = 4*ind2;
                        my_data->force_bond[ii  ]        += bforce[0];
                        my_data->qmlocal.bondf[offset  ] += bforce[0];
                        my_data->qmlocal.bondf[offset+1] -= bforce[1];
                        my_data->qmlocal.bondf[offset+2] -= bforce[2];
                        my_data->qmlocal.bondf[offset+3] -= bforce[3];
                }
        } else
        if ((sumqm==1)&&(sumlink==1)) {
                /*** add the classical bond force to the link atom ***/
                /*** my_data->qmlocal.bondf has the EGO sign       ***/
                /*** Case: L -- M or M -- L                        ***/

                if ((onthisnode[0])&&(islink[0])) {
                        if (isqmatom[0]) {
                                int offset = get_qmnode_index(myatom_id[0],my_data)*4;
                                int ii = 4*ind1;
                                my_data->force_bond[ii  ]        += bforce[0];
                                my_data->qmlocal.bondf[offset  ] += bforce[0];
                                my_data->qmlocal.bondf[offset+1] += bforce[1];
                                my_data->qmlocal.bondf[offset+2] += bforce[2];
                                my_data->qmlocal.bondf[offset+3] += bforce[3];
#if DEBUG_LINK
txtout("BOND to L: %d -- %d  %f %f %f\n",
myatom_id[0],myatom_id[1],-bforce[1],-bforce[2],-bforce[3]);
#endif
                        }
                        else { /*** only possible if QMNOLINK specified ***/

                                int offset = get_qmnode_index(myatom_id[1],my_data)*4;
                                int ii = 4*ind2;
                                my_data->force_bond[ii  ]        += bforce[0];
                                my_data->qmlocal.bondf[offset  ] += bforce[0];
                                my_data->qmlocal.bondf[offset+1] -= bforce[1];
                                my_data->qmlocal.bondf[offset+2] -= bforce[2];
                                my_data->qmlocal.bondf[offset+3] -= bforce[3];
#if DEBUG_LINK
txtout("Special case QMNOLINK BOND to L: %d -- %d  %f %f %f\n",
myatom_id[0],myatom_id[1],-bforce[1],-bforce[2],-bforce[3]);
#endif
                        }
                }

                if ((onthisnode[1])&&(islink[1])) {
                        if (isqmatom[1]) {
                                int offset = get_qmnode_index(myatom_id[1],my_data)*4;
                                int ii = 4*ind2;
                                my_data->force_bond[ii  ]        += bforce[0];
                                my_data->qmlocal.bondf[offset  ] += bforce[0];
                                my_data->qmlocal.bondf[offset+1] -= bforce[1];
                                my_data->qmlocal.bondf[offset+2] -= bforce[2];
                                my_data->qmlocal.bondf[offset+3] -= bforce[3];
#if DEBUG_LINK
txtout("BOND to L: %d -- %d  %f %f %f\n",
myatom_id[0],myatom_id[1],bforce[1],bforce[2],bforce[3]);
#endif
                        }
                        else {  /*** only possible if QMNOLINK specified ***/
                                int offset = get_qmnode_index(myatom_id[0],my_data)*4;
                                int ii = 4*ind1;
                                my_data->force_bond[ii  ]        += bforce[0];
                                my_data->qmlocal.bondf[offset  ] += bforce[0];
                                my_data->qmlocal.bondf[offset+1] += bforce[1];
                                my_data->qmlocal.bondf[offset+2] += bforce[2];
                                my_data->qmlocal.bondf[offset+3] += bforce[3];
#if DEBUG_LINK
txtout("Special case QMNOLINK BOND to L: %d -- %d  %f %f %f\n",
myatom_id[0],myatom_id[1],bforce[1],bforce[2],bforce[3]);
#endif
                        }
                }
        } else
        if ((sumqm==0)&&(sumlink==1)) { /*** only possible if QMNOLINK specified ***/
                                        /*** nothing to do in that case          ***/
        }

        return 0;
}


static  int     bondnumentries=0;
static  int     bondelements[2*MAX_QMLINK_ATOMS*3];
static  double  bondfactor[3*MAX_QMLINK_ATOMS];

static int
set_bondfactor(int *atomids, double factor)
{
        int     offs=2*bondnumentries;

        bondelements[offs+0]   = atomids[0];
        bondelements[offs+1]   = atomids[1];
        bondfactor[bondnumentries] = factor;
        bondnumentries++;

        if (bondnumentries>=3*MAX_QMLINK_ATOMS) {
                errout("Fatal error: To many Bond-Patches. Increase MAX_QMLINK_ATOMS in ego.h.\n");
                pex(-1);
        }
        return 0;
}

static double
get_bondfactor(int *atomids)
{
        int i,offs;

        if (bondnumentries==0)  return 0.0;
        for (i=0;i<bondnumentries;i++) {
                offs=2*i;

                if (    (bondelements[offs+0] == atomids[0]) &&
                        (bondelements[offs+1] == atomids[1]) ) return bondfactor[i];

                if (    (bondelements[offs+0] == atomids[1]) &&
                        (bondelements[offs+1] == atomids[0]) ) return bondfactor[i];
        }

        return 0.0;     /*** not in list ***/
}


/************************************************************************/
int
handle_qmangles(int *onthisnode,int *forceout_flag,int *atom_id,
                LOAD_MSG_ONE_NODE *my_data,
                double cur_energy,double *f1,double *f3,int *ai)
{
        int isqmatom[3],islink[3],sumqm,sumlink;

        /*** isqmatom[...] is 1 only if it is a QM atom ***/
        isqmatom[0] = (forceout_flag[0]&FLAG_QMATOM) ? 1 : 0;
        isqmatom[1] = (forceout_flag[1]&FLAG_QMATOM) ? 1 : 0;
        isqmatom[2] = (forceout_flag[2]&FLAG_QMATOM) ? 1 : 0;
        sumqm = isqmatom[0] + isqmatom[1] + isqmatom[2];

        if (sumqm==0) return 0; /*** no energies and forces to patch ***/

        /*** islink[...] is 1 only if it is a Link atom ***/
        islink[0]   = (forceout_flag[0]&FLAG_QMLINK) ? 1 : 0;
        islink[1]   = (forceout_flag[1]&FLAG_QMLINK) ? 1 : 0;
        islink[2]   = (forceout_flag[2]&FLAG_QMLINK) ? 1 : 0;
        sumlink = islink[0] + islink[1] + islink[2];

#if DEBUG_LINK
txtout("in handle_qmangles: atomids %d %d %d QM or NOT: %d %d %d\n",
atom_id[0],atom_id[1],atom_id[2],
isqmatom[0],isqmatom[1],isqmatom[2]);
#endif

        if (sumqm==1) {
                /**** Case: L -- M -- M *****/
                if ((onthisnode[0])&&(isqmatom[0])) {
                        int offset=get_qmnode_index(atom_id[0],my_data)*4;
                        int ii = 4*ai[0];
                        my_data->force_angle[ii  ]       +=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=cur_energy/3;
                        my_data->qmlocal.anglef[offset+1]+=f1[0];
                        my_data->qmlocal.anglef[offset+2]+=f1[1];
                        my_data->qmlocal.anglef[offset+3]+=f1[2];
#if DEBUG_LINK
txtout("ANGLE to L: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
f1[0],f1[1],f1[2]);
#endif
                }

                /**** Case: M -- L -- M *****/
                if ((onthisnode[1])&&(isqmatom[1])) {
                        int offset=get_qmnode_index(atom_id[1],my_data)*4;
                        int ii = 4*ai[1];
                        my_data->force_angle[ii  ]       +=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]-=(f1[0]+f3[0]);
                        my_data->qmlocal.anglef[offset+2]-=(f1[1]+f3[1]);
                        my_data->qmlocal.anglef[offset+3]-=(f1[2]+f3[2]);
#if DEBUG_LINK
txtout("ANGLE to L: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
(f1[0]+f3[0]),(f1[1]+f3[1]),(f1[2]+f3[2]));
#endif
                }

                /**** Case: M -- M -- L *****/
                if ((onthisnode[2])&&(isqmatom[2])) {
                        int offset=get_qmnode_index(atom_id[2],my_data)*4;
                        int ii = 4*ai[2];
                        my_data->force_angle[ii  ]       +=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]+=f3[0];
                        my_data->qmlocal.anglef[offset+2]+=f3[1];
                        my_data->qmlocal.anglef[offset+3]+=f3[2];
#if DEBUG_LINK
txtout("ANGLE to L: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
f3[0],f3[1],f3[2]);
#endif
                }
        } else
        /******************************************************************/
        /*** subtract energy contributions in case if all are QM atoms  ***/
        /*** Case: Q -- Q -- Q                                          ***/
        /******************************************************************/
        if ((sumqm==3)&&(sumlink==0)) {

                if (onthisnode[0]) {
                        int ii = 4*ai[0];
                        my_data->force_angle[ii  ] -= cur_energy/3.0;
                }
                if (onthisnode[1]) {
                        int ii = 4*ai[1];
                        my_data->force_angle[ii  ] -= cur_energy/3.0;
                }
                if (onthisnode[2]) {
                        int ii = 4*ai[2];
                        my_data->force_angle[ii  ] -= cur_energy/3.0;
                }
        } else
        /******************************************************************/
        /*** special treatment                                          ***/
        /*** Case: Q -- Q -- L                                          ***/
        /******************************************************************/
        if ((sumqm==3)&&(sumlink==1)) {
                double f=0.0;

                f=get_QQL_anglefactor(atom_id);

                if (onthisnode[0]) {
                        int offset=get_qmnode_index(atom_id[0],my_data)*4;
                        int ii = 4*ai[0];
                        my_data->force_angle[ii  ]       +=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]+=f*f1[0];
                        my_data->qmlocal.anglef[offset+2]+=f*f1[1];
                        my_data->qmlocal.anglef[offset+3]+=f*f1[2];
#if DEBUG_LINK
txtout("ANGLE to Q1: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
-f1[0],-f1[1],-f1[2]);
#endif
                }
                if (onthisnode[1]) {
                        int offset=get_qmnode_index(atom_id[1],my_data)*4;
                        int ii = 4*ai[1];
                        my_data->force_angle[ii  ]       +=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]-=f*(f1[0]+f3[0]);
                        my_data->qmlocal.anglef[offset+2]-=f*(f1[1]+f3[1]);
                        my_data->qmlocal.anglef[offset+3]-=f*(f1[2]+f3[2]);
#if DEBUG_LINK
txtout("ANGLE to Q2: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
(f1[0]+f3[0]),(f1[1]+f3[1]),(f1[2]+f3[2]));
#endif
                }
                if (onthisnode[2]) {
                        int offset=get_qmnode_index(atom_id[2],my_data)*4;
                        int ii = 4*ai[2];
                        my_data->force_angle[ii  ]       +=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]+=f*f3[0];
                        my_data->qmlocal.anglef[offset+2]+=f*f3[1];
                        my_data->qmlocal.anglef[offset+3]+=f*f3[2];
#if DEBUG_LINK
txtout("ANGLE to Q3: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
-f3[0],-f3[1],-f3[2]);
#endif
                }
        } else
        /******************************************************************/
        /*** special treatment                                          ***/
        /*** Case: Q -- L -- L                                          ***/
        /******************************************************************/
        if ((sumqm==3)&&(sumlink==2)) {
                if ((onthisnode[0])&&(isqmatom[0])) {
                        int offset=get_qmnode_index(atom_id[0],my_data)*4;
                        int ii = 4*ai[0];
                        my_data->force_angle[ii  ]       +=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]+=f1[0];
                        my_data->qmlocal.anglef[offset+2]+=f1[1];
                        my_data->qmlocal.anglef[offset+3]+=f1[2];
#if DEBUG_LINK
txtout("ANGLE to Q1: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
-f1[0],-f1[1],-f1[2]);
#endif
                }
                if ((onthisnode[1])&&(isqmatom[1])) {
                        int offset=get_qmnode_index(atom_id[1],my_data)*4;
                        int ii = 4*ai[1];
                        my_data->force_angle[ii  ]       +=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]-=(f1[0]+f3[0]);
                        my_data->qmlocal.anglef[offset+2]-=(f1[1]+f3[1]);
                        my_data->qmlocal.anglef[offset+3]-=(f1[2]+f3[2]);
#if DEBUG_LINK
txtout("ANGLE to Q2: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
(f1[0]+f3[0]),(f1[1]+f3[1]),(f1[2]+f3[2]));
#endif
                }
                if ((onthisnode[2])&&(isqmatom[2])) {
                        int offset=get_qmnode_index(atom_id[2],my_data)*4;
                        int ii = 4*ai[2];
                        my_data->force_angle[ii  ]       +=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]+=f3[0];
                        my_data->qmlocal.anglef[offset+2]+=f3[1];
                        my_data->qmlocal.anglef[offset+3]+=f3[2];
#if DEBUG_LINK
txtout("ANGLE to Q3: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
-f3[0],-f3[1],-f3[2]);
#endif
                }
                if (!(forceout_flag[1]&FLAG_QMLINK)) {
                        errout("Fatal error: No valid QM/MD-Fragmentation for angles %d %d %d.\n",
                        atom_id[0]+1,atom_id[1]+1,atom_id[2]+1);
                        errout("Think about your system!\n");
                        pex(-1);
                }
        } else
        /******************************************************************/
        /*** special treatment only possible if QMNOLINK specified      ***/
        /*** Case: Q -- Q -- L  (L is a MM atom then)                   ***/
        /******************************************************************/
        if (    (sumqm==2)&&
                (sumlink==1)&&
                (       (islink[0]&&(!isqmatom[0])) ||
                        (islink[1]&&(!isqmatom[1])) ||
                        (islink[2]&&(!isqmatom[2]))
                )
            ) {
                double f=1.0;

                if ((onthisnode[0])&&(isqmatom[0])) {
                        int offset=get_qmnode_index(atom_id[0],my_data)*4;
                        int ii = 4*ai[0];
                        my_data->force_angle[ii  ]       +=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]+=f*f1[0];
                        my_data->qmlocal.anglef[offset+2]+=f*f1[1];
                        my_data->qmlocal.anglef[offset+3]+=f*f1[2];
#if DEBUG_LINK
txtout("Special case QMNOLINK ANGLE to Q1: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
-f1[0],-f1[1],-f1[2]);
#endif
                }
                if ((onthisnode[1])&&(isqmatom[1])) {
                        int offset=get_qmnode_index(atom_id[1],my_data)*4;
                        int ii = 4*ai[1];
                        my_data->force_angle[ii  ]       +=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]-=f*(f1[0]+f3[0]);
                        my_data->qmlocal.anglef[offset+2]-=f*(f1[1]+f3[1]);
                        my_data->qmlocal.anglef[offset+3]-=f*(f1[2]+f3[2]);
#if DEBUG_LINK
txtout("Special case QMNOLINK ANGLE to Q2: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
(f1[0]+f3[0]),(f1[1]+f3[1]),(f1[2]+f3[2]));
#endif
                }
                if ((onthisnode[2])&&(isqmatom[2])) {
                        int offset=get_qmnode_index(atom_id[2],my_data)*4;
                        int ii = 4*ai[2];
                        my_data->force_angle[ii  ]       +=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=f*cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]+=f*f3[0];
                        my_data->qmlocal.anglef[offset+2]+=f*f3[1];
                        my_data->qmlocal.anglef[offset+3]+=f*f3[2];
#if DEBUG_LINK
txtout("Special case QMNOLINK ANGLE to Q3: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
-f3[0],-f3[1],-f3[2]);
#endif
                }
        } else
        /******************************************************************/
        /*** add forces to some QM atoms in the overlap region          ***/
        /*** Case: Q -- L -- M                                          ***/
        /******************************************************************/
        if (sumqm==2) {
                if ((onthisnode[0])&&(isqmatom[0])) {
                        int offset=get_qmnode_index(atom_id[0],my_data)*4;
                        int ii = 4*ai[0];
                        my_data->force_angle[ii  ]       +=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]+=f1[0];
                        my_data->qmlocal.anglef[offset+2]+=f1[1];
                        my_data->qmlocal.anglef[offset+3]+=f1[2];
#if DEBUG_LINK
txtout("ANGLE to Q1: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
-f1[0],-f1[1],-f1[2]);
#endif
                }
                if ((onthisnode[1])&&(isqmatom[1])) {
                        int offset=get_qmnode_index(atom_id[1],my_data)*4;
                        int ii = 4*ai[1];
                        my_data->force_angle[ii  ]       +=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]-=(f1[0]+f3[0]);
                        my_data->qmlocal.anglef[offset+2]-=(f1[1]+f3[1]);
                        my_data->qmlocal.anglef[offset+3]-=(f1[2]+f3[2]);
#if DEBUG_LINK
txtout("ANGLE to Q2: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
(f1[0]+f3[0]),(f1[1]+f3[1]),(f1[2]+f3[2]));
#endif
                }
                if ((onthisnode[2])&&(isqmatom[2])) {
                        int offset=get_qmnode_index(atom_id[2],my_data)*4;
                        int ii = 4*ai[2];
                        my_data->force_angle[ii  ]       +=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset  ]+=cur_energy/3.0;
                        my_data->qmlocal.anglef[offset+1]+=f3[0];
                        my_data->qmlocal.anglef[offset+2]+=f3[1];
                        my_data->qmlocal.anglef[offset+3]+=f3[2];
#if DEBUG_LINK
txtout("ANGLE to Q3: %d -- %d -- %d   %f %f %f\n",
atom_id[0],atom_id[1],atom_id[2],
-f3[0],-f3[1],-f3[2]);
#endif
                }

                if (!(forceout_flag[1]&FLAG_QMLINK)) {
                        errout("Fatal error: No valid QM/MD-Fragmentation for angles %d %d %d.\n",
                        atom_id[0]+1,atom_id[1]+1,atom_id[2]+1);
                        errout("Think about your system!\n");
                        pex(-1);
                }
        }
        return 0;
}

static  int     numentries=0;
static  int     elements[3*MAX_QMLINK_ATOMS*3];
static  double  qqlfactor[3*MAX_QMLINK_ATOMS];

static int
set_QQL_anglefactor(int *atomids, double factor)
{
        int     offs=3*numentries;

        elements[offs+0] = atomids[0];
        elements[offs+1] = atomids[1];
        elements[offs+2] = atomids[2];
        qqlfactor[numentries] = factor;
        numentries++;

        if (numentries>=3*MAX_QMLINK_ATOMS) {
                errout("Fatal error: To many QQL-Patches. Increase MAX_QMLINK_ATOMS in ego.h.\n");
                pex(-1);
        }
        return 0;
}

static double
get_QQL_anglefactor(int *atomids)
{
        int i,offs;
        if (numentries==0)      return 0.0;
        for (i=0;i<numentries;i++) {
                offs=3*i;
                if (elements[offs+1] != atomids[1]) continue;

                if (    (elements[offs+0] == atomids[0]) &&
                        (elements[offs+2] == atomids[2]) ) return qqlfactor[i];

                if (    (elements[offs+0] == atomids[2]) &&
                        (elements[offs+2] == atomids[0]) ) return qqlfactor[i];
        }

        return 0.0;             /*** not in list ***/
}

/************************************************************************/

int
handle_qmimpr(  int *onthisnode,int *forceout_flag,int *myatom_id,
                LOAD_MSG_ONE_NODE *my_data,
                double cur_energy,
                double *f1,double *f2,double *f3,double *f4,int *ai)
{
        int isqmatom[4],sumis;

        isqmatom[0] = (forceout_flag[0]&FLAG_QMATOM) ? 1 : 0;
        isqmatom[1] = (forceout_flag[1]&FLAG_QMATOM) ? 1 : 0;
        isqmatom[2] = (forceout_flag[2]&FLAG_QMATOM) ? 1 : 0;
        isqmatom[3] = (forceout_flag[3]&FLAG_QMATOM) ? 1 : 0;
        sumis = isqmatom[0] + isqmatom[1] + isqmatom[2] + isqmatom[3];
        if (sumis==0) return 0;

#if DEBUG_LINK
txtout("IMPRHANDLE: atomids %d %d %d %d QM or NOT: %d %d %d %d\n",
myatom_id[0],myatom_id[1],myatom_id[2],myatom_id[3],
isqmatom[0],isqmatom[1],isqmatom[2],isqmatom[3]);
#endif
        /*** subtract energy contributions in case if all are QM atoms ***/
        if (sumis==4) {
                if (onthisnode[0]) {
                        int ii = 4*ai[0];
                        my_data->force_impr[ii  ] -= cur_energy/4.0;
                }
                if (onthisnode[1]) {
                        int ii = 4*ai[1];
                        my_data->force_impr[ii  ] -= cur_energy/4.0;
                }
                if (onthisnode[2]) {
                        int ii = 4*ai[2];
                        my_data->force_impr[ii  ] -= cur_energy/4.0;
                }
                if (onthisnode[3]) {
                        int ii = 4*ai[3];
                        my_data->force_impr[ii  ] -= cur_energy/4.0;
                }
        } else
        if ((sumis==3)||(sumis==2)||(sumis==1)) {
                if ((onthisnode[0])&&(isqmatom[0])) {
                        int offset=get_qmnode_index(myatom_id[0],my_data)*4;
                        int ii = 4*ai[0];
                        my_data->force_impr[ii  ] += cur_energy/4.0;
                        my_data->qmlocal.imprf[offset  ]+=cur_energy/4.0;
                        my_data->qmlocal.imprf[offset+1]+=f1[0];
                        my_data->qmlocal.imprf[offset+2]+=f1[1];
                        my_data->qmlocal.imprf[offset+3]+=f1[2];
                }

                if ((onthisnode[1])&&(isqmatom[1])) {
                        int offset=get_qmnode_index(myatom_id[1],my_data)*4;
                        int ii = 4*ai[1];
                        my_data->force_impr[ii  ] += cur_energy/4.0;
                        my_data->qmlocal.imprf[offset  ]+=cur_energy/4.0;
                        my_data->qmlocal.imprf[offset+1]+=f2[0];
                        my_data->qmlocal.imprf[offset+2]+=f2[1];
                        my_data->qmlocal.imprf[offset+3]+=f2[2];
                }

                if ((onthisnode[2])&&(isqmatom[2])) {
                        int offset=get_qmnode_index(myatom_id[2],my_data)*4;
                        int ii = 4*ai[2];
                        my_data->force_impr[ii  ] += cur_energy/4.0;
                        my_data->qmlocal.imprf[offset  ]+=cur_energy/4.0;
                        my_data->qmlocal.imprf[offset+1]+=f3[0];
                        my_data->qmlocal.imprf[offset+2]+=f3[1];
                        my_data->qmlocal.imprf[offset+3]+=f3[2];
                }

                if ((onthisnode[3])&&(isqmatom[3])) {
                        int offset=get_qmnode_index(myatom_id[3],my_data)*4;
                        int ii = 4*ai[3];
                        my_data->force_impr[ii  ] += cur_energy/4.0;
                        my_data->qmlocal.imprf[offset  ]+=cur_energy/4.0;
                        my_data->qmlocal.imprf[offset+1]+=f4[0];
                        my_data->qmlocal.imprf[offset+2]+=f4[1];
                        my_data->qmlocal.imprf[offset+3]+=f4[2];
                }
        }
        return 0;
}

/************************************************************************/

int
handle_qmdihe(  int *onthisnode,int *forceout_flag,int *myatom_id,
                LOAD_MSG_ONE_NODE *my_data,
                double cur_energy,
                double *f1,double *f2,double *f3,double *f4,int *ai)
{
        int isqmatom[4],islink[4],sumis,sumlink;

        isqmatom[0] = (forceout_flag[0]&FLAG_QMATOM) ? 1 : 0;
        isqmatom[1] = (forceout_flag[1]&FLAG_QMATOM) ? 1 : 0;
        isqmatom[2] = (forceout_flag[2]&FLAG_QMATOM) ? 1 : 0;
        isqmatom[3] = (forceout_flag[3]&FLAG_QMATOM) ? 1 : 0;
        sumis = isqmatom[0] + isqmatom[1] + isqmatom[2] + isqmatom[3];

        if (sumis==0) return 0;

        /*** islink[...] is 1 only if it is a Link atom ***/
        islink[0]   = (forceout_flag[0]&FLAG_QMLINK) ? 1 : 0;
        islink[1]   = (forceout_flag[1]&FLAG_QMLINK) ? 1 : 0;
        islink[2]   = (forceout_flag[2]&FLAG_QMLINK) ? 1 : 0;
        islink[3]   = (forceout_flag[3]&FLAG_QMLINK) ? 1 : 0;
        sumlink = islink[0] + islink[1] + islink[2] + islink[3];

#if DEBUG_LINK
txtout("DIHEHANDLE: atomids %d %d %d %d QM or NOT: %d %d %d %d\n",
myatom_id[0],myatom_id[1],myatom_id[2],myatom_id[3],
isqmatom[0],isqmatom[1],isqmatom[2],isqmatom[3]);
#endif
        /*** subtract energy contributions in case if all are QM atoms ***/
        if ( (sumis==4) &&
             (  (sumlink==0)||(sumlink==1)||
                ((sumlink==2)&&islink[0]&&islink[3])
             )
            ) {
                if (onthisnode[0]) {
                        int ii = 4*ai[0];
                        my_data->force_dihe[ii  ] -= cur_energy/4.0;
                }
                if (onthisnode[1]) {
                        int ii = 4*ai[1];
                        my_data->force_dihe[ii  ] -= cur_energy/4.0;
                }
                if (onthisnode[2]) {
                        int ii = 4*ai[2];
                        my_data->force_dihe[ii  ] -= cur_energy/4.0;
                }
                if (onthisnode[3]) {
                        int ii = 4*ai[3];
                        my_data->force_dihe[ii  ] -= cur_energy/4.0;
                }
        } else
        if ((sumis==3)||(sumis==2)||(sumis==1)) {
                if ((onthisnode[0])&&(isqmatom[0])) {
                        int offset=get_qmnode_index(myatom_id[0],my_data)*4;
                        int ii = 4*ai[0];
                        my_data->force_dihe[ii  ]       +=cur_energy/4.0;
                        my_data->qmlocal.dihef[offset  ]+=cur_energy/4.0;
                        my_data->qmlocal.dihef[offset+1]+=f1[0];
                        my_data->qmlocal.dihef[offset+2]+=f1[1];
                        my_data->qmlocal.dihef[offset+3]+=f1[2];
                }

                if ((onthisnode[1])&&(isqmatom[1])) {
                        int offset=get_qmnode_index(myatom_id[1],my_data)*4;
                        int ii = 4*ai[1];
                        my_data->force_dihe[ii  ]       +=cur_energy/4.0;
                        my_data->qmlocal.dihef[offset  ]+=cur_energy/4.0;
                        my_data->qmlocal.dihef[offset+1]+=f2[0];
                        my_data->qmlocal.dihef[offset+2]+=f2[1];
                        my_data->qmlocal.dihef[offset+3]+=f2[2];
                }

                if ((onthisnode[2])&&(isqmatom[2])) {
                        int offset=get_qmnode_index(myatom_id[2],my_data)*4;
                        int ii = 4*ai[2];
                        my_data->force_dihe[ii  ]       +=cur_energy/4.0;
                        my_data->qmlocal.dihef[offset  ]+=cur_energy/4.0;
                        my_data->qmlocal.dihef[offset+1]+=f3[0];
                        my_data->qmlocal.dihef[offset+2]+=f3[1];
                        my_data->qmlocal.dihef[offset+3]+=f3[2];
                }

                if ((onthisnode[3])&&(isqmatom[3])) {
                        int offset=get_qmnode_index(myatom_id[3],my_data)*4;
                        int ii = 4*ai[3];
                        my_data->force_dihe[ii  ]       +=cur_energy/4.0;
                        my_data->qmlocal.dihef[offset  ]+=cur_energy/4.0;
                        my_data->qmlocal.dihef[offset+1]+=f4[0];
                        my_data->qmlocal.dihef[offset+2]+=f4[1];
                        my_data->qmlocal.dihef[offset+3]+=f4[2];
                }
        }
        return 0;
}


static  int     list_complete=0;
static  int     curqmexcl=0;
static  int     qmexclbase[MAXQMEXCL][3];

/*** qmatom and excluded_atom are the original EGO numbers ***/
/*** excltype is 1 for total exclusion, 2 for 14-exclusion ***/
int
set_qmexclusionatom(    int qmatom,int excluded_atom,
                        LOAD_MSG_ALL_NODE *global_data,
                        int excltype)
{
        if (qmatom==-1)         list_complete=1;
        if (list_complete)      return 0;

        if (curqmexcl>=MAXQMEXCL) {
                errout("Fatal error: Increase MAXQMEXCL (%d) in egqminter.c!",MAXQMEXCL);
                pex(-1);
        }

        qmexclbase[curqmexcl][0] = qmatom;
        qmexclbase[curqmexcl][1] = excluded_atom;
        qmexclbase[curqmexcl][2] = excltype;

        curqmexcl++;

#if DEBUGEXCL
txtout("Entry %d: QMatom %2d  excludes %2d type %d\n",
        curqmexcl-1,qmatom,excluded_atom,excltype);
#endif
        return 0;
}

/*********************************************/
/*** set dummy-flags due to exclusion list ***/
/*********************************************/
int
test_qmexclusion(int atom_id)
{
        int j;

        for (j=0;j<curqmexcl;j++)
                if (atom_id==qmexclbase[j][1])
                        return qmexclbase[j][2];

        return 0;
}

/*** atom_id is the EGO-atom-number of a QM-atom ***/
int
add_qmexcl_list(int atom_id,LOAD_MSG_ONE_NODE *my_data)
{
        int i,j,entry_there;
        int index,offset,limit;

        for (j=0;j<curqmexcl;j++) {

                if (atom_id!=qmexclbase[j][0]) continue;
                /*** now we have an exclusion-entry for this QM-atom ***/
                /*** now we look in the list of the lokal expansion  ***/
                /*** wheather this atom is included                  ***/

                index = get_qmnode_index(atom_id,my_data);
                offset= MAXDUMMYSPERQMATOM*index;
                limit = offset + my_data->qmlocal.incllen[index];
                entry_there = 0;

                for (i=offset;i<limit;i++) {
                        if (qmexclbase[j][1]==my_data->qmlocal.incllist[i]) {
                                entry_there=1;
                                break;
                        }
                }

                if (!entry_there) {
                        my_data->qmlocal.incllist[limit] = qmexclbase[j][1];
                        my_data->qmlocal.incllen[index]++;
#if DEBUGEXCL
txtout("Add QMEXCL: QM-Atom %d  Excl %d index %d newsize %d\n",
atom_id,qmexclbase[j][1],index,my_data->qmlocal.incllen[index]);
#endif
                }

        }

        return 0;
}

static  double rmatf[3][3]={    {1.0,0.0,0.0},
                                {0.0,1.0,0.0},
                                {0.0,0.0,1.0}
                            };
static  double rmatb[3][3]={    {1.0,0.0,0.0},
                                {0.0,1.0,0.0},
                                {0.0,0.0,1.0}
                            };

/*** rotation matrix first around x-axis, then arround y-axis, then z-axis ***/
static int
set_qmrotmats(double wx, double wy, double wz)
{
        double sa,sb,sg,ca,cb,cg;

        if ((wx==0.0) && (wy==0.0) && (wz==0.0)) {
                rmatf[0][0]=1.0;   rmatf[0][1]=0.0;   rmatf[0][2]=0.0;
                rmatf[1][0]=0.0;   rmatf[1][1]=1.0;   rmatf[1][2]=0.0;
                rmatf[2][0]=0.0;   rmatf[2][1]=0.0;   rmatf[2][2]=1.0;

                rmatb[0][0]=1.0;   rmatb[0][1]=0.0;   rmatb[0][2]=0.0;
                rmatb[1][0]=0.0;   rmatb[1][1]=1.0;   rmatb[1][2]=0.0;
                rmatb[2][0]=0.0;   rmatb[2][1]=0.0;   rmatb[2][2]=1.0;
                return 0;
        }

        sa = sin(wx/RADDEGREEFAC);
        sb = sin(wy/RADDEGREEFAC);
        sg = sin(wz/RADDEGREEFAC);

        ca = cos(wx/RADDEGREEFAC);
        cb = cos(wy/RADDEGREEFAC);
        cg = cos(wz/RADDEGREEFAC);

        rmatf[0][0]=cg*cb;   rmatf[0][1]=-sg*ca-cg*sb*sa;   rmatf[0][2]=sg*sa-cg*sb*ca;
        rmatf[1][0]=sg*cb;   rmatf[1][1]=cg*ca-sg*sb*sa;    rmatf[1][2]=-cg*sa-sg*sb*ca;
        rmatf[2][0]=sb;      rmatf[2][1]=cb*sa;             rmatf[2][2]=cb*ca;

        rmatb[0][0]=cg*cb;              rmatb[0][1]=sg*cb;            rmatb[0][2]=sb;
        rmatb[1][0]=-sg*ca-cg*sb*sa;    rmatb[1][1]=cg*ca-sg*sb*sa;   rmatb[1][2]=cb*sa;
        rmatb[2][0]=sg*sa-cg*sb*ca;     rmatb[2][1]=-cg*sa-sg*sb*ca;  rmatb[2][2]=cb*ca;
        return 0;
}

static int
rotforw(double *ko,double *res)
{
        double vec[3];
        vec[0] = rmatf[0][0]*ko[0] + rmatf[0][1]*ko[1] +rmatf[0][2]*ko[2];
        vec[1] = rmatf[1][0]*ko[0] + rmatf[1][1]*ko[1] +rmatf[1][2]*ko[2];
        vec[2] = rmatf[2][0]*ko[0] + rmatf[2][1]*ko[1] +rmatf[2][2]*ko[2];

        res[0] = vec[0];
        res[1] = vec[1];
        res[2] = vec[2];
        return 0;
}

static int
rotback(double *ko,double *res)
{
        double vec[3];
        vec[0] = rmatb[0][0]*ko[0] + rmatb[0][1]*ko[1] +rmatb[0][2]*ko[2];
        vec[1] = rmatb[1][0]*ko[0] + rmatb[1][1]*ko[1] +rmatb[1][2]*ko[2];
        vec[2] = rmatb[2][0]*ko[0] + rmatb[2][1]*ko[1] +rmatb[2][2]*ko[2];

        res[0] = vec[0];
        res[1] = vec[1];
        res[2] = vec[2];
        return 0;
}

static int
rothess(double hess[3][3])
{
        double MH[3][3];

        MH[0][0] = rmatf[0][0]*hess[0][0] + rmatf[0][1]*hess[1][0] + rmatf[0][2]*hess[2][0];
        MH[0][1] = rmatf[0][0]*hess[0][1] + rmatf[0][1]*hess[1][1] + rmatf[0][2]*hess[2][1];
        MH[0][2] = rmatf[0][0]*hess[0][2] + rmatf[0][1]*hess[1][2] + rmatf[0][2]*hess[2][2];

        MH[1][0] = rmatf[1][0]*hess[0][0] + rmatf[1][1]*hess[1][0] + rmatf[1][2]*hess[2][0];
        MH[1][1] = rmatf[1][0]*hess[0][1] + rmatf[1][1]*hess[1][1] + rmatf[1][2]*hess[2][1];
        MH[1][2] = rmatf[1][0]*hess[0][2] + rmatf[1][1]*hess[1][2] + rmatf[1][2]*hess[2][2];

        MH[2][0] = rmatf[2][0]*hess[0][0] + rmatf[2][1]*hess[1][0] + rmatf[2][2]*hess[2][0];
        MH[2][1] = rmatf[2][0]*hess[0][1] + rmatf[2][1]*hess[1][1] + rmatf[2][2]*hess[2][1];
        MH[2][2] = rmatf[2][0]*hess[0][2] + rmatf[2][1]*hess[1][2] + rmatf[2][2]*hess[2][2];

        hess[0][0] = MH[0][0]*rmatb[0][0] + MH[0][1]*rmatb[1][0] + MH[0][2]*rmatb[2][0];
        hess[0][1] = MH[0][0]*rmatb[0][1] + MH[0][1]*rmatb[1][1] + MH[0][2]*rmatb[2][1];
        hess[0][2] = MH[0][0]*rmatb[0][2] + MH[0][1]*rmatb[1][2] + MH[0][2]*rmatb[2][2];

        hess[1][0] = MH[1][0]*rmatb[0][0] + MH[1][1]*rmatb[1][0] + MH[1][2]*rmatb[2][0];
        hess[1][1] = MH[1][0]*rmatb[0][1] + MH[1][1]*rmatb[1][1] + MH[1][2]*rmatb[2][1];
        hess[1][2] = MH[1][0]*rmatb[0][2] + MH[1][1]*rmatb[1][2] + MH[1][2]*rmatb[2][2];

        hess[2][0] = MH[2][0]*rmatb[0][0] + MH[2][1]*rmatb[1][0] + MH[2][2]*rmatb[2][0];
        hess[2][1] = MH[2][0]*rmatb[0][1] + MH[2][1]*rmatb[1][1] + MH[2][2]*rmatb[2][1];
        hess[2][2] = MH[2][0]*rmatb[0][2] + MH[2][1]*rmatb[1][2] + MH[2][2]*rmatb[2][2];

        return 0;
}

