/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**	Module  : egrotflood.c 	by M.Eichinger and H. Grubmueller	**
**      Belongs2: EGO                                                   **
**	Contents: All calculations for flooding 			**
**									**
**	important variables for control:				**	
**	global_data->target_temp	in K				**
**		->flood_energy_ini	in kT				**
**		->flood_energy_final	in kT				**	
**		->flood_energy_inc	in kT/ps			**
**		->flood_time_const	in s				**
**              ->selimob_start         steps                           **
**		->flooding		1 if flooding is on		**
**		->flag_adaptive_flood	1 if adaptive flooding		**
**		->rot_correction	1 if rot correction		**	
*************************************************************************/

#define DEBUG	0
#define STOP_FLOODING_IF_ENERGY_DROPS_BELOW 0.8
#define STOP_FLOODING_IF_STRUCTURE_DEVIATION_EXCEEDS 300.0    /* A^2 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ego.h"
#include "egalloc.h"
#include "egrotflood.h"
#include "util_lib.h"

static 	void	ini_stoprt(void);
static	void	free_stoprt(void);
static	double	**stoprt(	int nr_of_atoms,
				double *x1,double *y1,double *z1,
				double *x2,double *y2,double *z2);

static	int	my_num_atoms=0;
static	int	my_start=0;

static	double	*xref,*yref,*zref;
static  double	*x2,*y2,*z2;
static	double	*x3,*y3,*z3;
static	double	*x4,*y4,*z4;
static	double	*x2sel,*y2sel,*z2sel;
static	double	*x3sel,*y3sel,*z3sel;
static	double	*x2ord,*y2ord,*z2ord;
static	double	*x3ord,*y3ord,*z3ord;
static	double	*xrefsel,*yrefsel,*zrefsel;
static	int	*map_to_list;
static	int	flooding=0;
static	int	rot_correction=0;
static	int	countptr=0;
static	int	all_filled=0;
static	int	my_number;

static	double	aveval;

int
ini_rot_flood(LOAD_MSG_ALL_NODE *global_data,int no_max_atoms)
{
	int i=no_max_atoms;i++; /*** no warning ***/

	flooding       = global_data->flooding;
	rot_correction = global_data->rot_correction;

	if ((!flooding)&&(!rot_correction))	return 0;

	xref = alloc_doubles(global_data->total_no_atoms+10);
	yref = alloc_doubles(global_data->total_no_atoms+10);
	zref = alloc_doubles(global_data->total_no_atoms+10);

	x2 = alloc_doubles(global_data->total_no_atoms+10);
	y2 = alloc_doubles(global_data->total_no_atoms+10);
	z2 = alloc_doubles(global_data->total_no_atoms+10);

	x3 = alloc_doubles(global_data->total_no_atoms+10);
	y3 = alloc_doubles(global_data->total_no_atoms+10);
	z3 = alloc_doubles(global_data->total_no_atoms+10);

	x4 = alloc_doubles(global_data->total_no_atoms+10);
	y4 = alloc_doubles(global_data->total_no_atoms+10);
	z4 = alloc_doubles(global_data->total_no_atoms+10);

	x2ord = alloc_doubles(global_data->total_no_atoms+10);
	y2ord = alloc_doubles(global_data->total_no_atoms+10);
	z2ord = alloc_doubles(global_data->total_no_atoms+10);

	x3ord = alloc_doubles(global_data->total_no_atoms+10);
	y3ord = alloc_doubles(global_data->total_no_atoms+10);
	z3ord = alloc_doubles(global_data->total_no_atoms+10);

	x2sel = alloc_doubles(global_data->flood_nr_atoms+10);
	y2sel = alloc_doubles(global_data->flood_nr_atoms+10);
	z2sel = alloc_doubles(global_data->flood_nr_atoms+10);

	x3sel = alloc_doubles(global_data->flood_nr_atoms+10);
	y3sel = alloc_doubles(global_data->flood_nr_atoms+10);
	z3sel = alloc_doubles(global_data->flood_nr_atoms+10);

	xrefsel = alloc_doubles(global_data->total_no_atoms+10);
	yrefsel = alloc_doubles(global_data->total_no_atoms+10);
	zrefsel = alloc_doubles(global_data->total_no_atoms+10);
	ini_stoprt();
	map_to_list = alloc_ints(global_data->total_no_atoms+10);
	return 0;
}

int
free_rot_flood()
{
	if ((!flooding)&&(!rot_correction))	return 0;

	myfree(xref);
	myfree(yref);
	myfree(zref);

	myfree(x2);
	myfree(y2);
	myfree(z2);

	myfree(x3);
	myfree(y3);
	myfree(z3);

	myfree(x4);
	myfree(y4);
	myfree(z4);

	myfree(x2ord);
	myfree(y2ord);
	myfree(z2ord);

	myfree(x3ord);
	myfree(y3ord);
	myfree(z3ord);

	myfree(x2sel);
	myfree(y2sel);
	myfree(z2sel);

	myfree(x3sel);
	myfree(y3sel);
	myfree(z3sel);

	myfree(xrefsel);
	myfree(yrefsel);
	myfree(zrefsel);

	free_stoprt();
	myfree(map_to_list);
	flooding = 0;
	rot_correction = 0;
	all_filled = 0;
	aveval     = 0.0;
	return 0;
}

void	reset_merge_buffer()
{
	all_filled = 0;
	aveval     = 0.0;
}

int
merge_packet_to_buffer(	int myn,
			int nr_of_atoms,
			TRAVELLING_COORDINATES *coord)
{
	my_number=myn;
 	 if ((!flooding)&&(!rot_correction))	return 0; 
#if DEBUG
txtout("Node %d: merging packet %d to buffer all %d  act %d\n",
my_number,coord->node,nr_of_atoms,coord->no_atoms);
#endif
	if (coord->node==0) { /*** copy last coords to x3,y3,z3 ***/
		countptr=0;
  	   	memcpy(x3,x2,sizeof(double)*nr_of_atoms);
		memcpy(y3,y2,sizeof(double)*nr_of_atoms);
		memcpy(z3,z2,sizeof(double)*nr_of_atoms); 
	}

	if (my_number==coord->node) {
		my_start     = countptr;
		my_num_atoms = coord->no_atoms;
	}

	/*** merge current coords to x2,y2,z2 ***/
	memcpy(x2+countptr,coord->x,sizeof(double)*coord->no_atoms);
	memcpy(y2+countptr,coord->y,sizeof(double)*coord->no_atoms);
	memcpy(z2+countptr,coord->z,sizeof(double)*coord->no_atoms);
	countptr += coord->no_atoms;

                  if (countptr==nr_of_atoms) {
	              /*** save first coordinate set to xref ***/
	          if (all_filled==0) {
			memcpy(xref,x2,sizeof(double)*nr_of_atoms);
			memcpy(yref,y2,sizeof(double)*nr_of_atoms);
			memcpy(zref,z2,sizeof(double)*nr_of_atoms);
		}



		if (all_filled==1) all_filled=2;
		if (all_filled==0) all_filled=1;
	}


#if DEBUG
txtout("Node %d: exit merge now %d with %d\n",
my_number,coord->node,all_filled);
#endif
	return 0;
}

int
do_rot_correction(	LOAD_MSG_ALL_NODE *global_data,
			double *xnew,double *ynew,double *znew,
			double *xold,double *yold,double *zold)
{
	double	x0,y0,z0,d1;
	double	x0ref,y0ref,z0ref;
	double **W;
	int	*new2old_map;
static	int	i,j; /*** ME: static needed for PARIX ?? ***/

	if ((!flooding)&&(!rot_correction))	return 0;
	if (all_filled!=2)			return 0;
#if DEBUG
txtout("Node ?: doing now transl/rotation correction\n");
#endif
	new2old_map = global_data->new2old_map;
	/**** now we reorder the new atoms x2,y2,z2 *****/
	for (i=0;i<global_data->total_no_atoms;i++) {
		int mapi = new2old_map[i];

		x2ord[mapi] = x2[i];
		y2ord[mapi] = y2[i];
		z2ord[mapi] = z2[i];
	}
	/**** now we extract new atoms *****/
	for (i=0,j=0;i<global_data->total_no_atoms;i++) {				
		x2sel[j] = x2ord[i];
		y2sel[j] = y2ord[i];
		z2sel[j] = z2ord[i];	
		if (global_data->flood_used_atoms[i])	j++;	
	}
	/**** now we reorder the old atoms *****/
	for (i=0;i<global_data->total_no_atoms;i++) {
		int mapi = new2old_map[i];

		x3ord[mapi] = x3[i];
		y3ord[mapi] = y3[i];
		z3ord[mapi] = z3[i];
	}
	/**** now we extract old atoms *****/
	for (i=0,j=0;i<global_data->total_no_atoms;i++) {				
		x3sel[j] = x3ord[i];
		y3sel[j] = y3ord[i];
		z3sel[j] = z3ord[i];	
		if (global_data->flood_used_atoms[i])	j++;	
	}
	/**** now we reorder the ref atoms *****/
	for (i=0;i<global_data->total_no_atoms;i++) {
		int mapi = new2old_map[i];

		xrefsel[mapi] = xref[i];
		yrefsel[mapi] = yref[i];
		zrefsel[mapi] = zref[i];
	}
	/**** now we extract ref atoms *****/
	for (i=0,j=0;i<global_data->total_no_atoms;i++) {				
		xrefsel[j] = xrefsel[i];
		yrefsel[j] = yrefsel[i];
		zrefsel[j] = zrefsel[i];	
		if (global_data->flood_used_atoms[i])	j++;	
	}
	/***** compute center of geometry of ref_coords *******/
	x0ref = y0ref = z0ref = 0.0;
	for (i = 0; i < global_data->flood_nr_atoms; i++) {
		x0ref += xrefsel[i];
		y0ref += yrefsel[i];
		z0ref += zrefsel[i];
	}
	x0ref /= ((double)global_data->flood_nr_atoms);
	y0ref /= ((double)global_data->flood_nr_atoms);
	z0ref /= ((double)global_data->flood_nr_atoms);

	/***** do center of geometry correction on new_coords *******/
	x0 = y0 = z0 = 0.0;
	for (i = 0; i < global_data->flood_nr_atoms; i++) {
		x0 += x2sel[i];
		y0 += y2sel[i];
		z0 += z2sel[i];
	}
	x0 /= ((double)global_data->flood_nr_atoms);
	y0 /= ((double)global_data->flood_nr_atoms);
	z0 /= ((double)global_data->flood_nr_atoms);
	x0 -= x0ref;
	y0 -= y0ref;
	z0 -= z0ref;

	for (i = 0; i < global_data->flood_nr_atoms; i++) {
		x2sel[i] -= x0;
		y2sel[i] -= y0;
		z2sel[i] -= z0;
	}
	/***** do center of geometry correction on old_coords *******/
	x0 = y0 = z0 = 0.0;
	for (i = 0; i < global_data->flood_nr_atoms; i++) {
		x0 += x3sel[i];
		y0 += y3sel[i];
		z0 += z3sel[i];
	}
	x0 /= ((double)global_data->flood_nr_atoms);
	y0 /= ((double)global_data->flood_nr_atoms);
	z0 /= ((double)global_data->flood_nr_atoms);
	x0 -= x0ref;
	y0 -= y0ref;
	z0 -= z0ref;

	for (i = 0; i < global_data->flood_nr_atoms; i++) {
		x3sel[i] -= x0;
		y3sel[i] -= y0;
		z3sel[i] -= z0;
	}
	/***** now do the rotation correction ******/
	W=stoprt(global_data->flood_nr_atoms,xrefsel,yrefsel,zrefsel,x2sel,y2sel,z2sel);
	d1 = 0.0;
	for(i=0;i<global_data->flood_nr_atoms;i++){
		double x0 = W[1][1]*x2sel[i]+W[1][2]*y2sel[i]+W[1][3]*z2sel[i];
		double y0 = W[2][1]*x2sel[i]+W[2][2]*y2sel[i]+W[2][3]*z2sel[i];
		double z0 = W[3][1]*x2sel[i]+W[3][2]*y2sel[i]+W[3][3]*z2sel[i];
		d1 += fabs(x0-x2sel[i]) + fabs(y0-y2sel[i]) + fabs(z0-z2sel[i]);
		x2sel[i] = x0;
		y2sel[i] = y0;
		z2sel[i] = z0;
	}
	
	d1 /= global_data->flood_nr_atoms;
	if (aveval<=0.0) aveval=d1;
	if (d1 > 50.0*aveval) {
		txtout("New initialisation of rot-correction (akt: %g ave:%g)!\n",
			d1,aveval);
		reset_merge_buffer();
		return 0;		
	}
	aveval = aveval*0.9 + 0.1*d1;
	
	W=stoprt(global_data->flood_nr_atoms,xrefsel,yrefsel,zrefsel,x3sel,y3sel,z3sel);
	for(i=0;i<global_data->flood_nr_atoms;i++){
		double x0 = W[1][1]*x3sel[i]+W[1][2]*y3sel[i]+W[1][3]*z3sel[i];
		double y0 = W[2][1]*x3sel[i]+W[2][2]*y3sel[i]+W[2][3]*z3sel[i];
		double z0 = W[3][1]*x3sel[i]+W[3][2]*y3sel[i]+W[3][3]*z3sel[i];
		x3sel[i] = x0;
		y3sel[i] = y0;
		z3sel[i] = z0;
	}
	
	/**** now we merge the new atoms *****/
	for (i=global_data->total_no_atoms-1,j=global_data->flood_nr_atoms-1;i>=0;i--) {
		if (global_data->flood_used_atoms[i])	{
			x2ord[i] = x2sel[j];
			y2ord[i] = y2sel[j];
			z2ord[i] = z2sel[j];	
			j--;	
 		}
	}
	/**** now we bring the new atoms back into old order *****/
	for(i=my_start,j=0;j<my_num_atoms;j++,i++){
		int mapi = new2old_map[i];

		xnew[j] = x2ord[mapi];
		ynew[j] = y2ord[mapi];
		znew[j] = z2ord[mapi];
	}
	/**** now we merge the old atoms *****/
	for (i=global_data->total_no_atoms-1,j=global_data->flood_nr_atoms-1;i>=0;i--) {
		if (global_data->flood_used_atoms[i])	{
			x3ord[i] = x3sel[j];
			y3ord[i] = y3sel[j];
			z3ord[i] = z3sel[j];	
			j--;	
 		}
	}
	/**** now we bring the old atoms back into old order *****/
	for(i=my_start,j=0;j<my_num_atoms;j++,i++){
		int mapi = new2old_map[i];

		xold[j] = x3ord[mapi];
		yold[j] = y3ord[mapi];
		zold[j] = z3ord[mapi];
	}

#if DEBUG
txtout("Node ?: Now did rot-trans-correction\n");					        
#endif	
	return 0;
}

int
add_flood_forces(	LOAD_MSG_ONE_NODE *my_data,
			LOAD_MSG_ALL_NODE *global_data,
			double *fxsum,
			double *fysum,
			double *fzsum,
			ENERGY_INFO *energy)
{
	int	k,i,j,l,cmptr,index=my_data->no_atoms; /*** no warning ***/
	int	nr_of_atoms=global_data->flood_nr_atoms;
	double	ex,extmp,sum,flc,scale,flooding_enr_total;
	double  averaging_factor1,averaging_factor2,desired_flood_energy_strength,timediff;
	double  structure_deviation=0.0;
	double	ke = 2.39e-4;/*   kcal/J */
	double	rg = 8.31441;/*   J/mol/K */
	double	floodc=ke*rg;/** umrechungsfaktor **/
	double	*flooding_cm,*flooding_matrix;
	int	*new2old_map;

	if (!flooding) {
	  	energy->flooding                  = 0.0;
	  	energy->flood_energy_average      = 0.0;
	  	energy->flood_energy_strength     = 0.0;
	  	energy->flooding_distance         = 0.0;
	  	global_data->flood_energy_average = 0.0;
	  	global_data->flood_energy_strength= 0.0;
	  	return 0;
	}
        if (global_data->cycles_passed<global_data->selimob_start) return 0;
/***
txtout("INFLOOD: %f %f %g %g %g\n",
global_data->flood_energy_average,
global_data->flood_energy_strength,
global_data->flood_scale,
global_data->flood_time_distance,
global_data->flood_energy_final
);
***/
	new2old_map = global_data->new2old_map;

	energy->flooding   = 0.0;
	flooding_enr_total = 0.0;

        /*****************************************************************
	   we need to pass the following two parameters from the master
	   to the nodes as well as from to nodes to the master;
	   therefore we include them into structures global_data *and*
	   energy. Ugly, but works. 
	******************************************************************/
	energy->flood_energy_average  = global_data->flood_energy_average;
	energy->flood_energy_strength = global_data->flood_energy_strength;
	energy->flooding_distance     = global_data->flood_distance;

	/**** compute actual flooding energy strength (E_0): *****/
        if (global_data->cycles_passed==global_data->selimob_start) {
          	energy->flood_energy_average = 0.0;
          	energy->flood_energy_strength= 0.0;
	  	energy->flooding_distance    = 0.0;
	}

        timediff = (global_data->cycles_passed - global_data->selimob_start) *
                   ((global_data->time_step * TIMEDIV) / 1.000E-12);         
        averaging_factor1=1.0e-15/global_data->flood_time_const;
        averaging_factor2=1.0e-15/global_data->flood_time_distance;
        
        if (!global_data->flag_adaptive_flood) {
          	energy->flood_energy_strength = global_data->flood_energy_ini + timediff * global_data->flood_energy_inc;
          	if (energy->flood_energy_strength>global_data->flood_energy_final) {
            		energy->flood_energy_strength=global_data->flood_energy_final;
            	}
	}
	if (energy->flood_energy_strength < 0.0001) energy->flood_energy_strength=0.0001;

	flc = floodc*global_data->target_temp*energy->flood_energy_strength;

	/**** now we reorder the atoms *****/
	for (i=0;i<global_data->total_no_atoms;i++) {
		int mapi = new2old_map[i];

		x3[mapi] = x2[i];
		y3[mapi] = y2[i];
		z3[mapi] = z2[i];
	}

	for (i=0,index=0;i<global_data->total_no_atoms;i++) {
		if (global_data->flood_used_atoms[i]) 
			map_to_list[i] = index++;
		else 	map_to_list[i] = -1; /** sure is sure **/
	}

	/**** now we extract atoms *****/
	for (i=0,j=0;i<global_data->total_no_atoms;i++) {				

		x3[j] = x3[i];
		y3[j] = y3[i];
		z3[j] = z3[i];	
		if (global_data->flood_used_atoms[i])	j++;	
	}
	
    	for (k=0; k<global_data->nr_of_flooding_gaussians; k++) { 
	        cmptr = 0;
		flooding_cm = global_data->flood_cm[k];

		for (i=0; i<nr_of_atoms; i++) {
	     		x4[i]=x3[i]-flooding_cm[cmptr++];
        		y4[i]=y3[i]-flooding_cm[cmptr++];
        		z4[i]=z3[i]-flooding_cm[cmptr++];
		}

		structure_deviation = 0.0;
		for (i=0; i<nr_of_atoms; i++) {
			structure_deviation += (x4[i] * x4[i] + y4[i] * y4[i] + z4[i] * z4[i]);
		}


		flooding_matrix = global_data->flood_matrix[k];
		ex=0.0;
		/* nur obere diagonalmatrix wird berechnet, weil matrix symmetrisch */
		/* zuerst die ausserblockdiagonalelemente: */
      		for (i=0; i<nr_of_atoms-1; i++) {
			int offs_x = i*3*3*nr_of_atoms;
			int offs_y = offs_x+3*nr_of_atoms;
			int offs_z = offs_y+3*nr_of_atoms;
			int offs;

			for (j=i+1; j<nr_of_atoms; j++) {
				offs=3*j;
				extmp =  x4[i]*flooding_matrix[offs_x+ offs    ]*x4[j];
				extmp += x4[i]*flooding_matrix[offs_x+ offs + 1]*y4[j];
				extmp += x4[i]*flooding_matrix[offs_x+ offs + 2]*z4[j];

				extmp += y4[i]*flooding_matrix[offs_y+ offs    ]*x4[j];
				extmp += y4[i]*flooding_matrix[offs_y+ offs + 1]*y4[j];
				extmp += y4[i]*flooding_matrix[offs_y+ offs + 2]*z4[j];

				extmp += z4[i]*flooding_matrix[offs_z+ offs    ]*x4[j];
				extmp += z4[i]*flooding_matrix[offs_z+ offs + 1]*y4[j];
				extmp += z4[i]*flooding_matrix[offs_z+ offs + 2]*z4[j];

				ex += (2.0*extmp);
			}
		}

		/* jetzt die blockdiagonalelemente: */
      		for (i=0; i<nr_of_atoms; i++) {
			int offs_x = i*3*3*nr_of_atoms;
			int offs_y = offs_x+3*nr_of_atoms;
			int offs_z = offs_y+3*nr_of_atoms;
			int offs   = 3*i;
			/* die ausserdiagonalelemente in jedem 3er-block: */
			ex += 2.0*x4[i]*flooding_matrix[offs_x+ offs + 1]*y4[i];
			ex += 2.0*x4[i]*flooding_matrix[offs_x+ offs + 2]*z4[i];
			ex += 2.0*y4[i]*flooding_matrix[offs_y+ offs + 2]*z4[i];

			/* die diagonalelemente: */
			ex += x4[i]*flooding_matrix[offs_x+ offs    ]*x4[i];
			ex += y4[i]*flooding_matrix[offs_y+ offs + 1]*y4[i];
			ex += z4[i]*flooding_matrix[offs_z+ offs + 2]*z4[i];
		}

		ex *= (-0.5);

		/***************************************************************/
		/*** global_data->flood_scale is a heuristic scalefactor for ***/
		/*** the width of the flooding-potential		     ***/
		/***************************************************************/
    		scale=sqrt(global_data->flood_scale*energy->flood_energy_strength);
      		ex /= (scale*scale);
		if (ex > 0.0) {
			errout("Fatal error: Flood-exponent > 0 !!\n");
			pex(-1);
		}
      		if (ex < -50.0)	ex = 0.0;
        	else		ex = exp(ex);
		
        	/********************************************************/
        	/*** add up x-forces ************************************/
        	/********************************************************/
        	for (i=my_start,l=0; l<my_num_atoms; i++,l++) {
			int offs_x = map_to_list[new2old_map[i]]*3*3*nr_of_atoms;
			int offs;
        		sum=0.0;
			if (global_data->flood_used_atoms[new2old_map[i]]) {
        		for (j=0,offs=0;j<nr_of_atoms; j++,offs+=3) { 
				sum += (flooding_matrix[offs_x+offs  ]*x4[j]);
				sum += (flooding_matrix[offs_x+offs+1]*y4[j]);
				sum += (flooding_matrix[offs_x+offs+2]*z4[j]);
          		}
        		my_data->force_flood[4*l+1] -= (flc*sum*ex/(scale*scale));
			}
		}
		
        	/********************************************************/
        	/*** add up y-forces ************************************/
        	/********************************************************/
		for (i=my_start,l=0; l<my_num_atoms; i++,l++) {
			int offs_y = map_to_list[new2old_map[i]]*3*3*nr_of_atoms + 3*nr_of_atoms;
			int offs;
        		sum=0.0;
			if (global_data->flood_used_atoms[new2old_map[i]]) {
	       		for (j=0,offs=0;j<nr_of_atoms; j++,offs+=3) { 
          			sum += (flooding_matrix[offs_y+offs  ]*x4[j]);
				sum += (flooding_matrix[offs_y+offs+1]*y4[j]);
				sum += (flooding_matrix[offs_y+offs+2]*z4[j]);
          		}
        		my_data->force_flood[4*l+2] -= (flc*sum*ex/(scale*scale));
			}
		}
		
        	/********************************************************/
        	/*** add up z-forces ************************************/
        	/********************************************************/
		for (i=my_start,l=0; l<my_num_atoms; i++,l++) {
			int offs_z = map_to_list[new2old_map[i]]*3*3*nr_of_atoms + 2*3*nr_of_atoms;
			int offs;
        		sum=0.0;
			if (global_data->flood_used_atoms[new2old_map[i]]) {
        		for (j=0,offs=0;j<nr_of_atoms; j++,offs+=3) { 
          			sum += (flooding_matrix[offs_z+offs  ]*x4[j]);
				sum += (flooding_matrix[offs_z+offs+1]*y4[j]);
				sum += (flooding_matrix[offs_z+offs+2]*z4[j]);
          		}
        		my_data->force_flood[4*l+3] -= (flc*sum*ex/(scale*scale));
			}
		}
		
		if (my_number==0) energy->flooding += (flc*ex); /* nur einmal erwuenscht */
		flooding_enr_total += (flc*ex);    /* for averaging only */
      	}

	/**** compute actual flooding energy strength (E_0): *****/

        desired_flood_energy_strength = global_data->flood_energy_ini + timediff * global_data->flood_energy_inc;
        if (desired_flood_energy_strength>global_data->flood_energy_final) {
		desired_flood_energy_strength=global_data->flood_energy_final;
	}
	
        if (global_data->flag_adaptive_flood) {
		energy->flood_energy_average   = (1.0-averaging_factor1)*energy->flood_energy_average
                         + averaging_factor1*(flooding_enr_total/(floodc*global_data->target_temp));  /* kT */
		energy->flood_energy_strength += (averaging_factor1*
			(desired_flood_energy_strength - energy->flood_energy_average));
	}

	energy->flooding_distance = (1.0-averaging_factor2)*energy->flooding_distance + 
					averaging_factor2*structure_deviation;

	/*****************************************************************/
	/******* one way to detect a conformational transition: **********/
	if (1==0) {   /* not used at the moment */
          if (global_data->flag_adaptive_flood) {
	    if (timediff > 6.0*global_data->flood_time_const/1.0e-12) {  
	    	/*** to skip start-up phase, where flood_energy_average is   ***/
	    	/*** considerably smaller than desired_flood_energy_strength ***/
	      if (energy->flood_energy_average < STOP_FLOODING_IF_ENERGY_DROPS_BELOW*desired_flood_energy_strength) {
	        global_data->flooding=0;
	        flooding=0;
		}
	      }
	    }
          }

	/* another way to detect a conformational transition: */
	if (1==0) {   /* also not used at the moment */
          if (global_data->flag_adaptive_flood) {
	    if (energy->flooding_distance > STOP_FLOODING_IF_STRUCTURE_DEVIATION_EXCEEDS) {
	      global_data->flooding=0;
	      flooding=0;
	      }
	    }
          }

        /* we need to pass the following two parameters from the master
	   to the nodes as well as from to nodes to the master;
	   therefore we include them into structures global_data *and*
	   energy. Ugly, but works. */
	global_data->flood_energy_average  = energy->flood_energy_average;
	global_data->flood_energy_strength = energy->flood_energy_strength;
	global_data->flood_distance        = energy->flooding_distance;

   	return 0;	
}

/******************************************************************/
static	double **B,**V,**D,*w;

static void 	ini_stoprt()
{
	D= dmatrix(1,3,1,3);
	B= dmatrix(1,3,1,3);
	V= dmatrix(1,3,1,3);
	w= dvector(1,3);
}

static void	free_stoprt()
{
	free_dmatrix(D,1,3,1,3);
	free_dmatrix(B,1,3,1,3);
	free_dmatrix(V,1,3,1,3);
	free_dvector(w,1,3);
}

static
double **stoprt( 	int nr_of_atoms,
		double *x1,double *y1,double *z1,
		double *x2,double *y2,double *z2)
{
	int i,j,k;
		
	for(i=1;i<=3;i++) for(j=1;j<=3;j++) B[i][j]=0.0;
	for(i=0;i<nr_of_atoms;i++){
		B[1][1] += x1[i]*x2[i];  
		B[1][2] += x1[i]*y2[i];  
		B[1][3] += x1[i]*z2[i];  
		B[2][1] += y1[i]*x2[i];  
		B[2][2] += y1[i]*y2[i];  
  		B[2][3] += y1[i]*z2[i];  
  		B[3][1] += z1[i]*x2[i];  
   		B[3][2] += z1[i]*y2[i];  
  		B[3][3] += z1[i]*z2[i];  
	}

	svdcmp(B,3,3,w,V);
	for(i=1;i<=3;i++) {
		for(j=1;j<=3;j++){
			D[i][j]=0.0;
 			for(k=1;k<=3;k++) {
				D[i][j] += B[i][k]*V[j][k];
			}
 		}
	}
	return(D);
}

