/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*--------------------------------------------------------------*/
/*								*/
/*			timing.c				*/
/*								*/
/* - Subject : function to manage timing.			*/
/* - Created by : Pierre-Eric BERNARD.                          */
/* - Modified by : Laurent COLOMBET & Christophe CALVIN  	*/
/* - Date : 8/23/1993.					        */
/*								*/
/*--------------------------------------------------------------*/

#include <stdio.h>
#if defined(NEXT)
#include <libc.h>
#endif

#if (NOTIMERS==1)

#else
#include <sys/types.h>
#include <sys/time.h>
#include <sys/times.h>
#if defined(LINUX) || defined(LINUX64)
#include <time.h>
#include <unistd.h>
#endif

#if defined(PARIX) || defined(PARIXPPC) || defined(SEQ_NANO)
#include <gnu/time.h>
#undef HZ
#endif
#endif /*** NOTIMERS ***/

#include "ego.h"
#include "util_lib.h"
#include "timing.h"

#define HZ	(float)60

typedef enum { TIMER_NOT_USED, TIMER_STOP, TIMER_RUN } Timer_Status;

typedef struct 
{
    	float last_user_time;	/* Last CPU time used while executing		*/
				/* instructions in the user space of		*/
				/* the calling process.				*/
    	float last_sys_time;	/* Last CPU time used by the systemm on		*/
				/* behalf of the calling process.		*/
    	long last_real_stime;	/* Last elapsed real time in seconds.	   	*/
    	long last_real_mtime;	/* Last elapsed real time in microseconds. 	*/
    	float user_time;	/* CPU time spend in user space, in second.	*/
    	float sys_time;		/* CPU time spend in system space, in second.	*/
    	long real_stime;	/* elapsed real time in seconds.		*/
    	long real_mtime;	/* elapsed real time in microseconds.		*/
    	Timer_Status status;
} Timer;

static	Timer	timer[NB_TIMER];
static	int	init_status_timer = 0;

/*----------------------------------------------------------------------
- Subject : init a new timer.
 ----------------------------------------------------------------------*/
int init_timer()
{
    	int i;
	
    	if(!init_status_timer) {	/* Timers not initialised. */
		for(i=0;i<NB_TIMER;i++) timer[i].status = TIMER_NOT_USED;
		init_status_timer = 1;
		clear_timer(0);
		return(0);
    	} else {	/** init_status_timer != 0 **/
		for( i=0;(timer[i].status!=TIMER_NOT_USED )&&(i<NB_TIMER);i++);
		if (i<NB_TIMER) {
	    	clear_timer(i);
 	    	return(i);
		} else {	/* No free timer. */
	    	return( -1 );
		}
    	}
}


/*--------------------------------------------------------------------
- Subject : free the timer num_timer.
----------------------------------------------------------------------*/
void free_timer( int num_timer )
{
    if( (0 <= num_timer) && (num_timer<NB_TIMER) ) {
	timer[num_timer].status = TIMER_NOT_USED;
    }
}

/*--------------------------------------------------------------------
- Subject : Reset the timer num_timer.
----------------------------------------------------------------------*/
int clear_timer( int num_timer )
{
	timer[num_timer].status     = TIMER_STOP;
	timer[num_timer].user_time  = (float)0.0;
	timer[num_timer].sys_time   = (float)0.0;
	timer[num_timer].real_stime = 0;
	timer[num_timer].real_mtime = 0;

	return 0;
}

/*--------------------------------------------------------------------
- Subject : Start the timer num_timer.
----------------------------------------------------------------------*/
int start_timer( int num_timer )
{
#if (NOTIMERS==1)
    	timer[num_timer].status          = TIMER_RUN;
	timer[num_timer].last_user_time  = (float)0.0;
	timer[num_timer].last_sys_time   = (float)0.0;
    	timer[num_timer].last_real_stime = 0L; 
    	timer[num_timer].last_real_mtime = 0L;
#else	
#if defined(PARIX) || defined(PARIXHAL) || defined(PARIXPPC) || defined(SEQ_NANO)
    	timer[num_timer].status          = TIMER_RUN;
    	timer[num_timer].last_user_time  = (float)0.0;
    	timer[num_timer].last_sys_time   = (float)0.0;
    	timer[num_timer].last_real_stime = 0L; 
    	timer[num_timer].last_real_mtime = (long)TimeNow();
#elif defined(T3D)
	/*** we have to use a dummy here, because the front-end machine ***/
	/*** uses much! system time if we use timer-functions ****/	
    	timer[num_timer].status          = TIMER_RUN;
	timer[num_timer].last_user_time  = (float)0.0;
	timer[num_timer].last_sys_time   = (float)0.0;
    	timer[num_timer].last_real_stime = 0L; 
    	timer[num_timer].last_real_mtime = 1L;
#else    
#if !defined(NEXT)
    	clock_t times();
#endif
	struct tms buffer;
	struct timeval t;
	struct timezone tz;

	times(&buffer);
	gettimeofday(&t,&tz);
    
	timer[num_timer].status          = TIMER_RUN;
	timer[num_timer].last_user_time  = (float)buffer.tms_utime;
	timer[num_timer].last_sys_time   = (float)buffer.tms_stime;
	timer[num_timer].last_real_stime = (long) t.tv_sec; 
	timer[num_timer].last_real_mtime = (long) t.tv_usec;
#endif
#endif /*** NOTIMERS ***/
	return(0);
}


/*----------------------------------------------------------------------
- Subject : Stop the timer num_timer.
-----------------------------------------------------------------------*/
int stop_timer( int num_timer )
{
#if (NOTIMERS==1)
    	timer[num_timer].status      = TIMER_STOP;
    	timer[num_timer].user_time   = (float)0.0;
    	timer[num_timer].sys_time    = (float)0.0;
    	timer[num_timer].real_stime  = 0;
    	timer[num_timer].real_mtime  = 0;
#else
#if defined(PARIX) || defined(PARIXHAL) || defined(PARIXPPC) || defined(SEQ_NANO)
    	long stop_time=0;
    	timer[num_timer].status     = TIMER_STOP;
    	timer[num_timer].user_time  = (float)0.0;
    	timer[num_timer].sys_time   = (float)0.0;
    	timer[num_timer].real_stime = 0;
    	stop_time=TimeNow();
    	if (stop_time>timer[num_timer].last_real_mtime) {
		timer[num_timer].real_mtime += (long)stop_time - (long)timer[num_timer].last_real_mtime;
    	} else {
		timer[num_timer].real_mtime += (long)stop_time - (long)timer[num_timer].last_real_mtime+
				4294967295L;
    	}
#elif defined(T3D)
	/*** dummy here: see start_timer ****/	
    	timer[num_timer].status      = TIMER_STOP;
    	timer[num_timer].user_time   = (float)0.0;
    	timer[num_timer].sys_time    = (float)0.0;
    	timer[num_timer].real_stime  = 0;
    	timer[num_timer].real_mtime += 1L;
#else
#if !defined(NEXT)
    	clock_t times();
#endif
	struct tms buffer;
	struct timeval t;
	struct timezone tz;
    
	gettimeofday(&t,&tz);
    	times(&buffer);
    
    	timer[num_timer].status      = TIMER_STOP;
#if defined(CRAY)
    	timer[num_timer].user_time  += (float)((float)buffer.tms_utime - timer[num_timer].last_user_time)/(float)CLK_TCK;
#else
    	timer[num_timer].user_time  += (float)((float)buffer.tms_utime - timer[num_timer].last_user_time)/ HZ;
#endif
    	timer[num_timer].sys_time   += (float)((float)buffer.tms_stime - timer[num_timer].last_sys_time) / HZ;
    	timer[num_timer].real_stime += (long) t.tv_sec  - (long) timer[num_timer].last_real_stime;
    	timer[num_timer].real_mtime += (long) t.tv_usec - (long) timer[num_timer].last_real_mtime;
#endif
#endif	/*** NOTIMERS ***/

	return( 0 );
}

/*--------------------------------------------------------------------
- Subject : Print timer num_timer.
----------------------------------------------------------------------*/
int print_timer( int num_timer, char * msg )
{
  	txtout("%s usertime= %.3f systemtime= %.3f realtime= %.3f\n",
  		msg,
		timer[num_timer].user_time,
		timer[num_timer].sys_time,
		get_rt_timer(num_timer));
	
	return 0;
}


/*--------------------------------------------------------------------
- Subject     : Give real time of num_timer.
----------------------------------------------------------------------*/
float get_rt_timer(int num_timer)
{
	float rt;

#if defined(PARIX) || defined(PARIXHAL) || defined(PARIXPPC) || defined(SEQ_NANO)
	rt = (float) timer[num_timer].real_mtime / 1000000.0;
#else
	rt = (float) timer[num_timer].real_stime + 
	     (float) timer[num_timer].real_mtime / 1000000.0;
#endif
	return(rt);

}

float get_user_timer(int num_timer)
{
	float rt;

#if defined(PARIX) || defined(PARIXHAL) || defined(PARIXPPC) || defined(SEQ_NANO)
	rt = (float) timer[num_timer].real_mtime / 1000000.0;
#else
	rt = (float) timer[num_timer].user_time;
#endif
	return(rt);
}

int
get_time_string(char *str)
{
#if defined(PARIX) || defined(PARIXHAL) || defined(PARIXPPC) || defined(SEQ_NANO)
	strcpy(str,"No time");
#else
	time_t aa;
	time(&aa);	
	strcpy(str,ctime(&aa));
	str[strlen(str)-1]=0;
#endif  
	return 0;
}


#if defined(SP2_MPI)
/* the sleep() call does return immediately under MPI on the SP2. So we need
   a workaround. Not nice, but should do. Brrrr .. I hate AIX (HH) 
   Brough this to the attemtion on IBM. They say it is a "feature" which is
   documented! Problem is that signals from MPI wake up sleep(). 
   IBM sent the replacement code below. I hope sleep() works in other MPI
   implementations! (HH) */
#include <errno.h>
#include <sys/time.h>
int SLEEP (unsigned int amount)
{
 struct timestruc_t Requested, Remaining;
 double famount = amount;
 int rc;
 while (famount > 0.0)
 {
  Requested.tv_sec = (int) famount;
  Requested.tv_nsec = (int) ((famount - Requested.tv_sec)*1000000000.);
  rc = nsleep ( &Requested, &Remaining );
  if ((rc == -1) && (errno == EINTR))
  {
   /* Sleep interrupted.  Resume it */
   famount = Remaining.tv_sec + Remaining.tv_nsec / 1000000000.;
   continue;
  }
  else /* Completed sleep.  Set return to zero */
   return (0);
  } /* end of while */
/* famount = 0; exit */
return (0);
}
#else
int SLEEP (unsigned int amount)
{
#if defined(PARIXPPC) /***** PVM-Subcase PPC ****/
  return (0);	/* there is no sleep() in PARIXPPC */
#else
  return(sleep(amount));
#endif
}
#endif /* SP2 MPI bug fix */


