/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: ego2crd.c	by M.Eichinger, Munich			**
**									**
*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <sys/stat.h> 
#include <sys/types.h>
#include <unistd.h>
#if defined(NEXT)
#include <sys/file.h>
#include <libc.h>
#endif
#if defined(RS6K)
#include <sys/mode.h> /* file modes for access command are in here! */
#else
#if !defined(E_ACC)
#define E_ACC 0
#endif
#endif

#include "../hardw.h"
#include "../ego.h"
#include "../util_lib.h"
#include "../wodecls.h"
#include "../woalloc.h"
#include "../version.h"
#include "../eganaly.h"

#define DEBUG 0

#define DEFAULT_OUTFILE "+lastframes"
#define CRD_EXT         ".DCD"
#define ENY_EXT         ".eny"
#define PRS_EXT		".prs"
#define VIR_EXT         ".vir"
#define ENDIAN_LITTLE	0
#define ENDIAN_BIG	1
#define LEN_OF_STRING	80
#define NUM_OF_STRINGS	2
#define intshort	int	/*** 2 or 4 bytes ***/
#define intlong 	long	/*** 4 or 8 bytes ***/

/******************************************************************************/
/**************** functions in this modul *************************************/
static int	write_crd_header(int crd_h,int numcoords,int framenum,int framestep,double timestep);
static int	append_to_crdfile(int crd_h,int numcoords,float *xko,float *yko,float *zko);
static int	search_lastfile(int *fn_startnr,int *fn_endnr,int *framestep);
static int	write_fortran_intl(int crd_h,int val);
static int	write_fortran_double(int crd_h,double val);
static int	write_fortran_float(int crd_h,float val);

static void	print_information(void);

/******************************************************************************/
static	char	cur_egofile[MAX_STRING_SIZE+1];
static	char	crd_outfile[MAX_STRING_SIZE+1];
static	char	coutpath[MAX_STRING_SIZE+1], eoutpath[MAX_STRING_SIZE+1], 
		help[MAX_STRING_SIZE+1];
static	char	ene_outfile[MAX_STRING_SIZE+1], prs_outfile[MAX_STRING_SIZE+1], 
                vir_outfile[MAX_STRING_SIZE+1];
static	char	sdum[MAX_STRING_SIZE+1];
static	char	eneremline[MAX_COMMENT_LINES][MAX_STRING_SIZE];
static	int	numcoords=-1;
static	int	framenum =-1;
static	int	framestep=-1;
static	double	timestep =-1.0; /*** in fs ***/
static	float	*xko,*yko,*zko;
static	int	own_endian=-1;
static	int	out_endian=-1;
static	int	ownsize[4],outsize[4];
static	int	ene_stride=1;
static	int	ene_count=0, prs_count=0, vir_count=0;
static	int	crd_count=0;
static	int	enecolumns=NR_OF_ENERGY_ROWS; /* this is just a default */
static	int	vircolumns=NR_OF_VIRIAL_PRESSURE_ROWS; /* this is just a default */
static	int	prscolumns=0; /* number of pressure values in one output row (can vary!) */
static	int	do_add_file=0;
static	int	new_ene_file=0;
static	int	new_crd_file=0;
static	int	avenumber=1;
static	int	fn_startnr,fn_endnr;
static	int	write_xplor=1;
static	int	nocrd=0,withavecoord=0;

int	get_atomtype_index(char *atomtypstr)
{
        strcpy(atomtypstr,"");
	return 0;
}
/********************************************************************************/
int main(int argc,char *argv[])
{
	int	i,j,startfn=-1,endfn=-1;
	FILE	*ego_h;
	int	crd_h,ene_h,prs_h=-1,vir_h=-1,warnings=0,stride_count=0,avecount;
	int	frames_already_written=0;
	int	curnum=-1,curframe=-1;
	double	curstep=-1.0;
	int	no_comments; /* the number of comment lines is also a switch:
			      * no_comments = 2: no pressure data, old format
			      * no_comments = 3: with pressure data, new format 
			      * no_comments = 4: with virial pressure data */

        txtout("==================================================================\n");
        txtout("ego2crd - Converting EGO-output files to DCD/crd file.\n");
        txtout("(%s) %s\n",VERSION,EGO_MACHINE);
        txtout("==================================================================\n");
	own_endian = get_my_endian();
	out_endian = own_endian;
	outsize[0] = ownsize[0] = sizeof(intshort);
	outsize[1] = ownsize[1] = sizeof(intlong);
	outsize[2] = ownsize[2] = sizeof(float);
	outsize[3] = ownsize[3] = sizeof(double);
	ene_stride = 1;
	ene_count=0;
	prs_count=0;
	crd_count=0;
	strcpy(coutpath,".");
	strcpy(eoutpath,".");
	if ((argc<=2)||(argc>11)) {
		print_information();
		return 0;
	}
	/***************************************************************************/
	/*** Manage INPUT **********************************************************/
	/***************************************************************************/
	strcpy(crd_outfile,DEFAULT_OUTFILE);	/*** default file name ***/
        sscanf(argv[1],"%d",&startfn);
        sscanf(argv[2],"%d",&endfn);
	fn_startnr = startfn;
	fn_endnr   = endfn;

        if ( (argc>=4)&&(argv[3][0]!='-') ) {
		strcpy(crd_outfile,argv[3]);
	}
        if (crd_outfile[0]=='+') { /*** now we have to manage incremental adding ***/
		int step  = -1;    /*** we do not know now ***/
		int res=0;
		fn_startnr = startfn;
		fn_endnr   = endfn;

		res=search_lastfile(&fn_startnr,&fn_endnr,&step);
		if (res<0)	return -1;
		frames_already_written = fn_endnr - fn_startnr + 1;
		if (res==0) {
			txtout("Creating new DCD/eny file !\n");
			do_add_file=1;
			frames_already_written = 0;
		} else
		if (fn_endnr+1==startfn) {
			txtout("Increment of DCD/eny file possible !\n");
			do_add_file=1;			
		} else
		if (fn_endnr+1<startfn) {
                         txtout("There is a gap between last file written to DCD/eny-file\n");
                         txtout("and new files to add ! You may have lost some files.\n");
                         txtout("No incremental adding of files, but using your specified filename!\n");
			 do_add_file=0;
			 strcpy(ene_outfile,crd_outfile+1);
		} else
		if (fn_endnr>=endfn) {
                         txtout("All specified EGO-outputfiles are already in a\n");
                         txtout("DCD/eny file ! This may happen after restarts. Dont worry !\n");
                         txtout("Nothing to do for ego2crd.\n");
			 goto no_error_occured;
		} else {
                         txtout("You specified some EGO-output files, that are\n");
                         txtout("already in the DCD/eny file. I do skip those files !\n");
                         txtout("This may happen after restarts. Dont worry!\n");
			 do_add_file=1;
			 startfn=fn_endnr+1;
                         strcpy(ene_outfile,"add_");	
			 generate_egofname(ene_outfile+4,"",fn_startnr);
                         strcat(ene_outfile,"-");
                         generate_egofname(ene_outfile+strlen(ene_outfile),"",fn_endnr);
		}
		txtout("******************************************************\n\n");
		if (do_add_file) {
			strcpy(ene_outfile,"add_");	
			generate_egofname(ene_outfile+4,"",fn_startnr);
			ene_outfile[strlen(ene_outfile)-4]=0;
			strcat(ene_outfile,"-");
			generate_egofname(ene_outfile+strlen(ene_outfile),"",fn_endnr);
			ene_outfile[strlen(ene_outfile)-4]=0;
		}
		strcpy(crd_outfile,ene_outfile);
	} else {
		do_add_file=0;
	}
	strcpy(ene_outfile,crd_outfile);
	strcpy(prs_outfile,crd_outfile);
        strcpy(vir_outfile,crd_outfile);
	strcat(crd_outfile,CRD_EXT);
	strcat(ene_outfile,ENY_EXT);
	strcat(prs_outfile,PRS_EXT);
        strcat(vir_outfile,VIR_EXT);
	if (frames_already_written==0) {
		remove(ene_outfile); /*** No DCD/crd file ==> erase eny file if there ***/
		remove(prs_outfile); /*** No DCD/crd file ==> erase prs file if there ***/
		remove(vir_outfile); /*** No DCD/crd file ==> erase prs file if there ***/
	}	
	for (i=3;i<argc;i++) {
		if (strcmp(argv[i],"-nocrd")==0)	nocrd = 1;
		if (strcmp(argv[i],"-pc")==0)		strcpy(coutpath,argv[i+1]);
		if (strcmp(argv[i],"-pe")==0)		strcpy(eoutpath,argv[i+1]);
		if (strcmp(argv[i],"-C")==0)		write_xplor= 0;
		if (strcmp(argv[i],"-c")==0)		write_xplor= 0;
                if (strcmp(argv[i],"-l")==0)    	out_endian = ENDIAN_LITTLE;
                if (strcmp(argv[i],"-b")==0)    	out_endian = ENDIAN_BIG;
		if (strcmp(argv[i],"-outave")==0)	withavecoord=1;
		if (strcmp(argv[i],"-s4448")==0) {
			outsize[0] = 4;
			outsize[1] = 4;
			outsize[2] = 4;
			outsize[3] = 8;
		}
                if (strcmp(argv[i],"-s4848")==0) {
			outsize[0] = 4;
			outsize[1] = 8;
			outsize[2] = 4;
			outsize[3] = 8;
		}
                if (    (argv[i][0]=='-')&&
                        (argv[i][1]=='i') ) {
                        int res=sscanf(argv[i]+2,"%d",&ene_stride);
			if ((res!=1)||(ene_stride<1)||(ene_stride>10000)) {
                                txtout("Illegal option <%s>. I ignore this !\n",argv[i]);
				ene_stride = 1;
			}
		}
                if (    (argv[i][0]=='-')&&
                        (argv[i][1]=='a')&&(argv[i][2]=='v')&&(argv[i][3]=='e') ) {
                        int res=sscanf(argv[i]+4,"%d",&avenumber);
			if ((res!=1)||(avenumber<1)) {
                                txtout("Illegal option <%s>. I ignore this !\n",argv[i]);
				avenumber=1;
			}
		}
	}
        txtout("First ego-output file     : %d.\n",startfn);
        txtout("Last  ego-output file     : %d.\n",endfn);

	strcpy(help, crd_outfile);
	strcpy(crd_outfile,coutpath);
	strcat(crd_outfile,"/");
	strcat(crd_outfile,help);

	strcpy(help, ene_outfile);
	strcpy(ene_outfile,eoutpath);
	strcat(ene_outfile,"/");
	strcat(ene_outfile,help);

	strcpy(help, prs_outfile);
	strcpy(prs_outfile,eoutpath);
	strcat(prs_outfile,"/");
	strcat(prs_outfile,help);

	strcpy(help, vir_outfile);
	strcpy(vir_outfile,eoutpath);
	strcat(vir_outfile,"/");
	strcat(vir_outfile,help);

	if (do_add_file) {
                txtout("Add to DCD output file     : >%s<.\n",crd_outfile);
                txtout("Add to energy output file  : >%s<.\n",ene_outfile);
                txtout("Add to pressure output file: >%s<.\n",prs_outfile);
                txtout("Add to virial output file  : >%s<.\n",vir_outfile);
	} else {
                txtout("New DCD output file        : >%s<.\n",crd_outfile);
                txtout("New energy output file     : >%s<.\n",ene_outfile);
                txtout("New pressure output file   : >%s<.\n",prs_outfile);
                txtout("New virial  output file    : >%s<.\n",vir_outfile);
	}
        txtout("Stride energy output       : %d.\n",ene_stride);
	if (withavecoord)
	txtout("Average coordiante set will be written to ave.ego\n");

	if (write_xplor)		txtout("crd/DCD-format            : XPLOR.\n");
	else				txtout("crd/DCD-format            : CHARMM.\n");

        if (out_endian==ENDIAN_LITTLE)  txtout("                          : LITTLE ENDIAN.\n");
        else                            txtout("                          : BIG ENDIAN.\n");

        txtout("Bytes sizes               : int=%1d, long=%1d, float=%1d, double=%1d\n",
		outsize[0],outsize[1],outsize[2],outsize[3]);

	if ((startfn<0) || (endfn<0) || (startfn>endfn)) {
                errout("Illegal parameters for files to convert !\n");
		return -1;
	}
	if (((endfn-startfn+1)%avenumber)!=0) {
                errout("Fatal error: (%d-%d+1) mod %d != 0. Illegal aveing parameter !\n",
			endfn,startfn,avenumber);
		return -1;
	}
	/************************************************************************/
	/****************** readin header of first file for checks **************/
	/************************************************************************/
        generate_egofname(cur_egofile,"",startfn);
        if ((ego_h=fopen(cur_egofile,"r"))==NULL) {
                errout("Can not open/find ego-file <%s>. Fatal error.\n",cur_egofile);
		return -1;
	}
	if (read_trjego_header(ego_h,&numcoords,&framenum,&framestep,&timestep)<0) {
		fclose(ego_h);
		return -1;
	}

	if ((no_comments=read_eneego_header(ego_h,eneremline,&enecolumns))<0) {
                errout("Something is wrong with the ego-output file.\n");
                errout("There should be three remark lines for energies.\n");
                errout(">%s<\n",eneremline[0]);
                errout(">%s<\n",eneremline[1]);
                errout(">%s<\n",eneremline[2]);
		fclose(ego_h);
		return -1;
	}
	fclose(ego_h);
        txtout("There are %d energy columns per output line.\n",enecolumns);
#if DEBUG == 1
	txtout("There are %d comment lines in the ego file.\n", no_comments);
#endif
	/************************************************************************/
	/************************** prepare crd/DCD-file ************************/
	/************************************************************************/
	if (!do_add_file) remove(crd_outfile);
	new_crd_file=(access(crd_outfile,E_ACC) == -1);
	if ((crd_h=open(crd_outfile,O_CREAT|O_RDWR,0400|0200))==0) {
                errout("Fatal error: Can not access crd/DCD output file <%s>.\n",crd_outfile);
		return -1;
	}
	if (new_crd_file)
		if (write_crd_header(crd_h,numcoords,framenum,framestep*avenumber,timestep)<0) {
                        errout("Fatal error: Can not write crd/DCD header.\n");
			close(crd_h);
			return -1;
		}
	lseek(crd_h,0L,SEEK_END);
        txtout("There are %d atoms in one coordinate set.\n",numcoords);
        txtout("======================================================\n\n");

	/************************************************************************/
	/*************************** prepare energy-file ************************/
	/************************************************************************/
	if (!do_add_file) remove(ene_outfile);
	new_ene_file=(access(ene_outfile,E_ACC)==-1);
#if DEBUG == 1
	if (new_ene_file) 
	  txtout("We are creating a new eny file\n");
	else
	  txtout("We are appending to an exisiting eny file\n");
#endif
	if ((ene_h=open(ene_outfile,O_CREAT|O_RDWR,0400|0200))==0) {
                errout("Fatal error: Can not create energy output file <%s>.\n",ene_outfile);
		close(crd_h);
		return -1;
	}
	if (new_ene_file) { /*** write the header ***/
#if DEBUG == 1
		txtout("Writing energy header\n%s\n%s",eneremline[0],eneremline[1]);
#endif
                sprintf(sdum,"2 %2d                 %g \n",enecolumns,timestep);
		write(ene_h,sdum,strlen(sdum));
		write(ene_h,eneremline[0],strlen(eneremline[0]));
		write(ene_h,"\n",1);
		write(ene_h,eneremline[1],strlen(eneremline[1]));
		write(ene_h,"\n",1);
		ene_count = 0;
	} else {
		int a,b;
		lseek(ene_h,0L,SEEK_SET);
                read(ene_h,sdum,20);sdum[19]='\n';
                sscanf(sdum,"%d %d %d\n",&a,&b,&ene_count);
                txtout("Found %d energy lines in file !\n",ene_count);
		if (enecolumns!=b) {
                        errout("Fatal error: Number of energy columns changed (%d,%d) !\n",b,enecolumns);
			close(crd_h);
			close(ene_h);
			return -1;
		}
	}
	
	if ((no_comments == 3) || (no_comments == 4))
	   /* the number of comment lines is also a switch:
	    * no_comments = 2: no pressure data, old format
	    * no_comments = 3: with pressure data, new format 
	    * no_comments = 4: with virial pressure data */
	  {
	  /************************************************************************/
	  /*************************** prepare pressure-file **********************/
	  /************************************************************************/
	  if (!do_add_file) remove(prs_outfile);
	  new_ene_file=(access(prs_outfile,E_ACC)==-1);
	  prs_h=open(prs_outfile,O_CREAT|O_RDWR,0400|0200);
	  if (prs_h==0) {
		  errout("Fatal error: Can not create pressure output file <%s>.\n",prs_outfile);
		  close(crd_h);
		  return -1;
	  }
	  if (new_ene_file) { /*** write the header ***/
	  	  if (no_comments == 3) /* older ego file */
		  {
		    prscolumns = NR_OF_PRESSURE_ROWS;
		  }
		  else if (no_comments == 4)
		  {
		    prscolumns = NR_OF_PRESSURE_ROWS;
		  }
		  sprintf(sdum,"2 %2d                 %g \n",prscolumns,timestep);
		  write(prs_h,sdum,strlen(sdum));
		  write(prs_h,eneremline[0],strlen(eneremline[0]));
    		  write(prs_h,"\n",1);		  
    		  write(prs_h,"Step ",5);
		  strcpy(sdum,eneremline[2]);
		  write(prs_h, sdum,strlen(sdum));
    		  write(prs_h,"\n",1);		  
		  prs_count = 0;
	  } else { /* we append to an existing file */
		  int a,old_prscolumns;
		  lseek(prs_h,0L,SEEK_SET);
		  read(prs_h,sdum,20);sdum[19]='\n';
		  sscanf(sdum,"%d %d %d\n",&a,&old_prscolumns,&prs_count);
		  txtout("Found %d pressure lines in file !\n",prs_count);
		  if (prscolumns!=old_prscolumns) {
			  errout("Fatal error: Number of pressure columns changed (%d,%d) !\n", 
			          old_prscolumns,prscolumns);
			  close(crd_h);
			  close(prs_h);
			  return -1;
		  }
	  }
	}

	/************************************************************************/
	/*************************** prepare virial-file ************************/
	/************************************************************************/
	if (!do_add_file) remove(vir_outfile);
	new_ene_file=(access(vir_outfile,E_ACC)==-1);
	if ((vir_h=open(vir_outfile,O_CREAT|O_RDWR,0400|0200))==0) {
                errout("Fatal error: Can not create energy output file <%s>.\n",vir_outfile);
		close(crd_h);
		return -1;
	}

	if (new_ene_file) { /*** write the header ***/
                sprintf(sdum,"2 %2d                 %g \n",vircolumns,timestep);
		write(vir_h,sdum,strlen(sdum));
		write(vir_h,eneremline[0],strlen(eneremline[0]));
		write(vir_h,"\n",1);
                write(vir_h,"Step ",5);
		write(vir_h,eneremline[3],strlen(eneremline[3]));
		write(vir_h,"\n",1);
		vir_count = 0;
	} else {
		int a,b;
		lseek(vir_h,0L,SEEK_SET);
                read(vir_h,sdum,20);sdum[19]='\n';
                sscanf(sdum,"%d %d %d\n",&a,&b,&vir_count);
                txtout("Found %d energy lines in file !\n",vir_count);
		if (vircolumns != b) {
                        errout("Fatal error: Number of energy columns changed (%d,%d) !\n",b,vircolumns);
			close(crd_h);
			close(ene_h);
			close(prs_h);
			return -1;
		}
	}

	lseek(ene_h,0L,SEEK_END);
	lseek(prs_h,0L,SEEK_END);
	lseek(vir_h,0L,SEEK_END);

	xko = alloc_floats(numcoords+10);
	yko = alloc_floats(numcoords+10);
	zko = alloc_floats(numcoords+10);

	/********************************************************************/
	/****************** here starts the loop over all files *************/
	/********************************************************************/
	warnings = 0;
	stride_count=ene_stride;
	crd_count = 0;
	avecount=avenumber;
	for (i=startfn;i<=endfn;i++) {
		double coords[3];
		curstep=-1.0;
		curnum=-1;
		curframe=-1;
	
                generate_egofname(cur_egofile,"",i);
                if ((ego_h=fopen(cur_egofile,"r"))==NULL) {
                        errout("\nWarning error: Can not open <%s>. Skipping it !\n",
				cur_egofile);
			warnings++;
			continue;
		}
		if (read_trjego_header(ego_h,&curnum,&curframe,&framestep,&curstep)==-1) {
                        errout("\nWarning error: Can not read header of <%s>. Skipping it !\n",
				cur_egofile);
			warnings++;
			continue;
		}
                txtout("<%s>: frame: %-6d timestep: %g fs ...",
			cur_egofile,curframe,curstep);
		if (curnum!=numcoords) {
                        errout("\nFatal error: Number of atoms changed from %d to %d !\n",
				numcoords,curnum);
			goto error_occured;
		}
		/**********************************************************/
		/****** readin coords from ego-file ***********************/
		/**********************************************************/
		if (avecount>=avenumber) {/*** initialize aveing buffer ***/
			for (j=0;j<curnum;j++) {
				xko[j] = (float) 0.0;
				yko[j] = (float) 0.0;
				zko[j] = (float) 0.0;
			}
			avecount=0;
		}
		if (!nocrd) {
		    for (j=0;j<curnum;j++) {
			int res = read_ego_coords(ego_h,coords);
			if (res==-2) goto error_occured;
			if (res==-1) {
                                errout("\nFatal error: End of file after %d read coords !\n",j);
				goto error_occured;
			}
			xko[j] += (float) coords[0];
			yko[j] += (float) coords[1];
			zko[j] += (float) coords[2];
		    }
		    avecount++;
                    txtout("done ! ");
		    /**********************************************************/
		    /******** append coords to crd-file ***********************/
		    /**********************************************************/
		    if (avecount>=avenumber) {
                        txtout("Appending ...");
			for (j=0;j<curnum;j++) {
				xko[j] /= (float) avecount;
				yko[j] /= (float) avecount;
				zko[j] /= (float) avecount;
			}
			if (append_to_crdfile(crd_h,curnum,xko,yko,zko)<0) {
				goto error_occured;
			}
                        txtout("done !\n");
                        if (withavecoord) { /*** output the last average coord set ***/
				FILE *myh;
				char str[30];
				sprintf(str,"ave_%08d.ego",crd_count);
  				if ((myh=fopen(str,"w"))==NULL) {
                       			 errout("\nWarning error: Can not create <ave.ego>.\n");
				}
				fprintf(myh,"[BEGINCOORD]\n");
				fprintf(myh,"%d %d %d 1.0000\n",curnum,i-avenumber+1,avenumber);
				for (j=0;j<curnum;j++) {
					fprintf(myh,"%# 10.7f %# 10.7f %# 10.7f\n",
							xko[j],yko[j],zko[j]);
				}
                        	fclose(myh);
                        	txtout("Average coords written to file <%s>\n",str);
                        }
			crd_count++;
		    } else {
			txtout("Averaging.\n");
		    }
		} else {
			txtout("coords skipped.\n");
		}

		if ((no_comments=read_eneego_header(ego_h,eneremline,&enecolumns))<0) {
                        errout("Fatal error: Something is wrong with <%s>.\n",cur_egofile);
                        errout("There should be three remark lines for energies.\n");
                        errout(">%s<\n",eneremline[0]);
                        errout(">%s<\n",eneremline[1]);;
                        errout(">%s<\n",eneremline[2]);;
			fclose(ego_h);
			goto error_occured;
		}
		if ((no_comments==3)||(no_comments==4)) {
			while (read_string(ego_h,sdum)!=-1) {
				char step[MAX_STRING_SIZE];
                        	strcat(sdum,"\n");
				if (stride_count>=ene_stride) { 
			 		if(sdum[0] == 'E') {
			   		 	write(ene_h,&(sdum[2]),strlen(sdum)-2);
			    			sscanf(&(sdum[2]),"%s",step);
			    			strcat(step," ");
			    			ene_count++;
			 		}
			  		if(sdum[0] == 'P') {
			   			write(prs_h,step,strlen(step));
						if (no_comments==3)
						{
			    			  write(prs_h,&(sdum[2]),strlen(sdum)-2);
			    			  stride_count=0;
			    			  prs_count++;
						}
						else
						{
						  /* With the new vir-file, there is no more difference */
                                                write(prs_h,&(sdum[2]),strlen(sdum)-2);
						}
			  		}
			  		if(sdum[0] == 'V') {
			   			write(vir_h,step,strlen(step));
			    			write(vir_h,&(sdum[2]),strlen(sdum)-2);
			    			stride_count=0;
			    			vir_count++;
			  		}
				}
		                if (no_comments==3) 
				{
				  if(sdum[0] == 'P') stride_count++;
				}
		                else if (no_comments==4) 
				{
				  if(sdum[0] == 'V') stride_count++;
				}
			}
		} else { /*** compatibility-mode to old out-put-style ***/
			 /*** and if only energy-lines in EGO-file    ***/
			while (read_string(ego_h,sdum)!=-1) {
                        	strcat(sdum,"\n");
				if (stride_count>=ene_stride) { 
					if (sdum[0] == 'E') {	/*** only energies ***/
			   		 	write(ene_h,&(sdum[2]),strlen(sdum)-2);
			 		} else {		/*** old output-style ***/
			   			write(ene_h,sdum,strlen(sdum));
			   		}
			    		ene_count++;
			    		stride_count=0;
			 	}
				stride_count++;
			}

		}
		fclose(ego_h);
	}
	/*** now we update the crd/DCD header  ***/
	frames_already_written += crd_count;
	lseek(crd_h,(long)(4L+outsize[0]),SEEK_SET);
	write_fortran_intl(crd_h,frames_already_written);
	lseek(crd_h,(long)(4L+outsize[0]+3*outsize[1]),SEEK_SET);
#if DEBUG == 2
txtout("TEST frames_alrady_written %d   last frame %d\n",frames_already_written,curframe);
#endif
	write_fortran_intl(crd_h,curframe+1);
	close(crd_h);
	
        /*** now we update the energy-file header argh... ***/
	lseek(ene_h,7L,SEEK_SET);
        sprintf(sdum,"%d",ene_count);
	write(ene_h,sdum,strlen(sdum));
	close(ene_h);

        if (no_comments>=3) {
	    /*** now we update the pressure-file header argh... ***/
	    lseek(prs_h,7L,SEEK_SET);
            sprintf(sdum,"%d",prs_count);
	    write(prs_h,sdum,strlen(sdum));
	    close(prs_h);

	    lseek(vir_h,7L,SEEK_SET);
            sprintf(sdum,"%d",vir_count);
	    write(vir_h,sdum,strlen(sdum));
	    close(vir_h);
        }
        
	if (do_add_file) {	/*** now we rename our files 	***/
				/*** add_########-########.crd	***/
				/*** add_########-########.eny	***/
				/*** add_########-########.prs	***/		
				/*** add_########-########.vir	***/		
		char ene_outfile_new[MAX_STRING_SIZE];
		char crd_outfile_new[MAX_STRING_SIZE];
		char prs_outfile_new[MAX_STRING_SIZE];
		char vir_outfile_new[MAX_STRING_SIZE];
		
		strcpy(ene_outfile_new,"add_");	
		generate_egofname(ene_outfile_new+4,"",fn_startnr);
		ene_outfile_new[strlen(ene_outfile_new)-4]=0;
		strcat(ene_outfile_new,"-");
		generate_egofname(ene_outfile_new+strlen(ene_outfile_new),"",endfn);
		ene_outfile_new[strlen(ene_outfile_new)-4]=0;
	
		strcpy(crd_outfile_new,ene_outfile_new);
		strcpy(prs_outfile_new,ene_outfile_new);
		strcpy(vir_outfile_new,ene_outfile_new);
				
		strcat(crd_outfile_new,CRD_EXT);
		strcat(ene_outfile_new,ENY_EXT);
		strcat(prs_outfile_new,PRS_EXT);
		strcat(vir_outfile_new,VIR_EXT);
				
		if (strcmp(crd_outfile_new,crd_outfile)!=0) {
			txtout("\nRename file <%s> to <%s>\n",crd_outfile,crd_outfile_new);
			rename(crd_outfile,crd_outfile_new);

			txtout("Rename file <%s> to <%s>\n",ene_outfile,ene_outfile_new);
			rename(ene_outfile,ene_outfile_new);
			
			txtout("Rename file <%s> to <%s>\n",prs_outfile,prs_outfile_new);
			rename(prs_outfile,prs_outfile_new);

			txtout("Rename file <%s> to <%s>\n",vir_outfile,vir_outfile_new);
			rename(prs_outfile,prs_outfile_new);
		}	
	} 

no_error_occured:	
        txtout("\n======================================================\n");
        txtout("ego2crd finished now. No fatal errors.\n");
        txtout("%d coordinate sets and %d energy steps were added.\n",crd_count,ene_count);
        txtout("%d warning errors occured !\n",warnings);
        txtout("======================================================\n");
	myfree(xko);
	myfree(yko);
	myfree(zko);
	return 0;
error_occured:
	fclose(ego_h);
	close(crd_h);
	close(ene_h);
        txtout("\n======================================================\n");
        errout("ego2crd stopped with a fatal error.\n");
        errout("Ego-output-files may be corrupt.\n");
        errout("%d new coordinate sets and %d energy steps were written.\n",crd_count,ene_count);
        errout("%d warning errors occured !\n",warnings);
        txtout("======================================================\n");
	myfree(xko);
	myfree(yko);
	myfree(zko);
	return -1;
}

static int
search_lastfile(int *fn_startnr,int *fn_endnr,int *framestep)
{
	int res,i;
	char curfile[MAX_STRING_SIZE+1];
	FILE *tmp_h;

	res = system("ls -t add_*-*.DCD > tmp_ego2crd.log");
	if ((tmp_h=fopen("tmp_ego2crd.log","r"))==NULL) {
		errout("ego2crd, fatal error: tmp_ego2crd.log was not created !\n");
		return -1;	
	}
	read_string(tmp_h,curfile);
	fclose(tmp_h);
	remove("tmp_ego2crd.log");
	if (strlen(curfile)<9) { /*** no file found for adding ***/
		return 0;
	}

	i = sscanf(curfile,"add_%d-%d.DCD",fn_startnr,fn_endnr);
	*framestep=-1;
	if (i!=2) {
		errout("ego2crd, fatal error: Found a 'add_*-*.DCD' file, but it has not two filenumbers !\n");
		return -1;
	}
	return 1;	
}

static
int write_fortran_string(int crd_h,char *str)
{
	write(crd_h,str,strlen(str));
	return 0;
}

static
int write_fortran_ints(int crd_h,int val)
{
	intshort h=val;

	if (own_endian==out_endian) {
		write(crd_h,&h,sizeof(intshort));
	} else {
		char *ptr1,swapbuf[64];
		int i;
		ptr1 = (char *) &h;

		for (i=sizeof(intshort)-1;i>=0;i--) swapbuf[i] = *ptr1++;
		write(crd_h,swapbuf,sizeof(intshort));
	}
	return 0;
}

static
int write_fortran_intl(int crd_h,int val)
{
	intlong h=(intlong)val;
	if (outsize[1]!=ownsize[1]) {
		if (outsize[1]==ownsize[0]) {
			write_fortran_ints(crd_h,val);
			return 0;
		} else
		if ((outsize[1]==8)&&(ownsize[0]==4)) {
			if (out_endian==ENDIAN_LITTLE) {
				write_fortran_ints(crd_h,0);
				write_fortran_ints(crd_h,val);
			 } else {
				write_fortran_ints(crd_h,val);
				write_fortran_ints(crd_h,0);
			 }
			return 0;
		}
	}

	if (own_endian==out_endian) {
		write(crd_h,&h,sizeof(intlong));
	} else {
		char *ptr1,swapbuf[64];
		int i;
		ptr1 = (char *) &h;
		for (i=sizeof(intlong)-1;i>=0;i--) swapbuf[i] = *ptr1++;
		write(crd_h,swapbuf,sizeof(intlong));
	}
	return 0;
}

static
int write_fortran_double(int crd_h,double val)
{
	double h=val;

	if (own_endian==out_endian) {
		write(crd_h,&h,sizeof(double));
	} else {
		char *ptr1,swapbuf[64];
		int i;
		ptr1 = (char *) &h;
			for (i=sizeof(double)-1;i>=0;i--) swapbuf[i] = *ptr1++;
		write(crd_h,swapbuf,sizeof(double));
	}
	return 0;
}
static
int write_fortran_float(int crd_h,float val)
{
	float h=val;

	if (own_endian==out_endian) {
		write(crd_h,&h,sizeof(float));
	} else {
		char *ptr1,swapbuf[64];
		int i;
		ptr1 = (char *) &h;

		for (i=sizeof(float)-1;i>=0;i--) swapbuf[i] = *ptr1++;
		write(crd_h,swapbuf,sizeof(float));
	}
	return 0;
}

static int
write_crd_header(int crd_h,int numcoords,int staframe,int framestep,double timestep)
{
	int	i,len1,rlen;
	char	rem1[140],rem2[140];
        sprintf(rem1," REMARKS Trajectory calculated with %s\n",VERSION);
        sprintf(rem2," REMARKS Converted by ego2crd %s\n",EGO_MACHINE);
	len1=min(78,strlen(rem1));
	for (i=len1;i<80;i++) {
                rem1[i]   = ' ';
		rem1[i+1] = 0;
	}
	len1=min(78,strlen(rem2));
	for (i=len1;i<80;i++) {
                rem2[i]   = ' ';
		rem2[i+1] = 0;
	}
	/**********************************/
	rlen = 4 + 18*outsize[1] + outsize[3];
	write_fortran_ints(crd_h,rlen); 	/** start record **/
        write_fortran_string(crd_h,"CORD");
	write_fortran_intl(crd_h,0);		/*** number of frames: Xplor sets it to 0 ***/
	write_fortran_intl(crd_h,staframe+1);	/*** start integr-step			  ***/
	write_fortran_intl(crd_h,framestep);	/*** save integr-step			  ***/
	write_fortran_intl(crd_h,0);		/*** last integr-step: Xplor sets it to 0 ***/
	write_fortran_intl(crd_h,0);
	write_fortran_intl(crd_h,0);
	write_fortran_intl(crd_h,0);
	write_fortran_intl(crd_h,0);		/*** Degrees of freedom: atoms*3 - 6	  ***/
	write_fortran_intl(crd_h,0);		/*** NATOM-NFREAT ***/
	if (write_xplor) {
		write_fortran_double(crd_h,(double)(timestep/(TIMEFAC*1000.0)));
	} else { /*** CHARMM, Quanta - style ***/
		write_fortran_float(crd_h,(double)(timestep/(TIMEFAC*1000.0)));
		write_fortran_ints(crd_h,0);
	}
	write_fortran_intl(crd_h,0);
	write_fortran_intl(crd_h,0);
	write_fortran_intl(crd_h,0);
	write_fortran_intl(crd_h,0);
	write_fortran_intl(crd_h,0);
	write_fortran_intl(crd_h,0);
	write_fortran_intl(crd_h,0);
	write_fortran_intl(crd_h,0);
	write_fortran_intl(crd_h,0);		/*** Charmm Version (22)***/		
	write_fortran_ints(crd_h,rlen); 	/*** terminate record	***/
	/**********************************/
	rlen = LEN_OF_STRING*NUM_OF_STRINGS + outsize[1];
	write_fortran_ints(crd_h,rlen); 	/*** start record	***/
	write_fortran_intl(crd_h,NUM_OF_STRINGS);
	write_fortran_string(crd_h,rem1);	/*** info strings	***/
	write_fortran_string(crd_h,rem2);
	write_fortran_ints(crd_h,rlen); 	/*** terminate record	***/
	/**********************************/
	write_fortran_ints(crd_h,outsize[1]);
	write_fortran_intl(crd_h,numcoords);
	write_fortran_ints(crd_h,outsize[1]);

	return 0;
}

static int
append_to_crdfile(int crd_h,int numcoords,float *xko,float *yko,float *zko)
{
	int b,i;
	b=numcoords*outsize[2];
	write_fortran_ints(crd_h,b);
	if (own_endian==out_endian) {
		write(crd_h,xko,b);
	} else {
		for (i=0;i<numcoords;i++) write_fortran_float(crd_h,(float)xko[i]);
	}
	write_fortran_ints(crd_h,b);
	write_fortran_ints(crd_h,b);
	if (own_endian==out_endian) {
		write(crd_h,yko,b);
	} else {
		for (i=0;i<numcoords;i++) write_fortran_float(crd_h,(float)yko[i]);
	}
	write_fortran_ints(crd_h,b);
	write_fortran_ints(crd_h,b);
	if (own_endian==out_endian) {
		write(crd_h,zko,b);
	} else {
		for (i=0;i<numcoords;i++) write_fortran_float(crd_h,(float)zko[i]);
	}
	write_fortran_ints(crd_h,b);
	return 0;
}

static void
print_information()
{
	txtout("ego2crd <first_egonr> <last_egonr> [<outfilename>] [-C] [-i<number>]\n");
	txtout("        [-ave<number>] [-nocrd] -[b|l] -[s????]\n");
	txtout("        [-pc <crd output path>] [-pe <eny output path>]\n");
	txtout("Description:\n");
	txtout("<outfilename>		Name of DCD-file without extension. If <outfilename>\n");
	txtout("			starts with '+', coordinates are added to the end of\n");
	txtout("			a file 'add_<num_min>-<num_max>.DCD'. If more such\n");
	txtout("			files are found, ego2crd chooses the most recently\n");
	txtout("			written file.\n");
	txtout("			If adding of new files is not possible, the specified\n");
	txtout("			filename after '+' is used. Default is '+lastframes'.\n");	
	txtout("option -pc		Specify a crd output path other than the current directory.\n");				
	txtout("option -pe		Specify an energy output path other than the current directory.\n");				
	txtout("option -C		Use CHARMM DCD-format, default is XPLOR format.\n");				
	txtout("option -i<number>	Stride of writing out energy time steps.\n");
	txtout("option -ave<avenumber>	Number of coordinate sets to ave over.\n");
	txtout("			Default is 1 (no aveing). Important:\n");
	txtout("			(<last_egonr>-<first_egonr>+1)/<avenumber> must be\n");
	txtout("			an integer !\n");
	txtout("option -outave          Output the last averaged coord set to ASCII file ave.ego.\n");
	txtout("option -nocrd		Produce an empty DCD/crd-file.\n");
	txtout("As DCD/crd files are binary files, you have to characterise the target\n");
	txtout("computer on which you want to analyse your DCD/crd files (XPLOR, Quanta, etc.),\n");
	txtout("if you don't envoke 'ego2crd' on this target computer.\n");
	txtout("This is done by the options '-[b|l]' and '-[s????]'\n");
	txtout("option -b              Specifies BIG ENDIAN    (DEC,486)\n");
	txtout("option -l              Specifies LITTLE ENDIAN (SUN,SGI,RS6K,RS8000)\n");
	txtout("option -s????          Spec. number of bytes for int, long, float and double.\n");
	txtout("==>    -s4448          Data format for SGI,SUN,486,RS6K.\n");
	txtout("       -s4848          Data format for DEC-Alpha,R8000.\n");
	txtout("Example: If you want to make a DCD/crd-file for a SGI running ego2crd\n");
	txtout("         on any computer use the options: -l -s4448\n");
	txtout("========================================================================\n");
	txtout("The data format of your current computer is given below:\n");	
	if (own_endian==ENDIAN_LITTLE) {
	txtout("This computer uses LITTLE ENDIAN type     ==> option -l\n");
	} else {
	txtout("This computer uses BIG ENDIAN type        ==> option -b\n");
	}
	txtout("Size of int=%1d long=%1d float=%1d double=%1d     ==> option -s%1d%1d%1d%1d\n",
			(int)sizeof(int),(int)sizeof(long),
			(int)sizeof(float),(int)sizeof(double),
			(int)sizeof(int),(int)sizeof(long),
			(int)sizeof(float),(int)sizeof(double));
	return;
}


