/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: ego2pdb.c	by M.Eichinger, Munich			**
**	A utility to convert koordinates given in one format to another **
**									**
*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "../hardw.h"
#include "../ego.h"
#include "../util_lib.h"
#include "../wodecls.h"
#include "../woalloc.h"
#include "../version.h"

#define DEBUG 0

/******************************************************************************/
static	char	old_pdb_file[MAX_STRING_SIZE+1];
static	char	new_pdb_file[MAX_STRING_SIZE+1];
static	char	ego_file[MAX_STRING_SIZE+1];
static	char	line[MAX_STRING_SIZE+1];
static	char	formatstr[MAX_STRING_SIZE+1];
static  LOAD_MSG_ALL_NODE *global_data;

#define	PDB	0
#define LIS	1
static	int	templatetype=PDB;

#define EGO	2
#define FFORM	3
#define RESTART	4
static	int	inputtype=EGO;

static	int	conversionflag=0,shiftflag=0;

int	get_atomtype_index(char *atomtypstr)
{
        strcpy(atomtypstr,"");
	return 0;
}
/****************************************************************************/
int main(int argc,char *argv[])
{
	int	j,clf=0;
	FILE	*ego_h,*ego_h_new;
	double	*xko,*yko,*zko,*charge,coords[3],timestep;
	int	numcoords,framenum,framestep;
	int *renum;
	
        txtout("==================================================================\n");
        txtout("ego2pdb - Conversion of coordinate formats.\n");
        txtout("(%s) %s\n",VERSION,EGO_MACHINE);
        txtout("==================================================================\n");

	if (argc!=4) {
                txtout("ego2pdb <template-file> <input-file> <output-file>\n");
		txtout("A utility to convert file-formats containing atom positions.\n");
		txtout("1. Example: You have an EGO-output (e.g., 00001200.ego) file and you\n");
		txtout("            want to convert it to a PDB file.\n");	
		txtout("            ego2pdb myoldpdb.pdb 0001200.ego 0001200.pdb\n");
		
		txtout("2. Example: You have an EGO-output (e.g., 00001200.ego) file and you \n");
		txtout("            want to convert it to a new coord-file.\n");
		txtout("            ego2pdb coord.lis 0001200.ego coord_1200.lis\n");
							
       		txtout("3. Example: You have an output of coordinates in an arbitrary format and\n");
		txtout("            you want to convert it to a new coord-file or PDB file.\n");
		txtout("            Assume the following format of your input-file with 321 atoms:\n");		
		txtout("            1   HA    4.4563   7.6577   2.2345\n");
		txtout("            2   CB    5.4769  -0.5787   1.2456\n");
		txtout("            ..........\n");
		txtout("            Modify the input-file and write before the list of coordinates:\n");
		txtout("            [FREEFORMAT]\n");
		txtout("            321 0 -1\n");
		txtout("            %%*d  %%*s  %%lf %%lf %%lf\n\n");
					
		txtout("<template-file>		Either a PDB-file or a coord.lis file.\n");
                txtout("<input-file>		Either a EGO-outputfile or a free-format file.\n");               
                txtout("<output-file>		Depending on the template-file either\n");
                txtout("                        a PDB-file or a coord.lis file.\n");
		txtout("Format of the free-format case:\n");
		txtout("[FREEFORMAT]\n");
		txtout("<number of atoms> <conversion-flag> <shift-flag>\n");
		txtout("<C-style conversion string>\n");
		txtout("<conversion-flag>	==0  -> Perform no unit conversion.\n");	
		txtout("                       ==1  -> Convert from Bohr to Angstroem.\n");
		txtout("<shift-flag>		==-1 -> Perform no origin shifting.\n");	
		txtout("                       ==#  -> Position of atom # is the new origin.\n");	
		txtout("Notes on C-conversion strings:\n");
		txtout("%%s			stands for a string.\n");
		txtout("%%d			stands for an integer.\n");
		txtout("%%lf			stands for a real number.\n");
		txtout("%%*d, %%*s or %%*lf	skips corresponding entry.\n");
		txtout("The entries of the first %%d in the conversion string is used for resorting.\n");		
		txtout("atom order. If no %%d is included (or only %%*d's) no resorting is performed.\n");
		return 0;
	}
	/***************************************************************************/
	/*** Manage INPUT **********************************************************/
	/***************************************************************************/
	strcpy(old_pdb_file,argv[1]);
	strcpy(ego_file,argv[2]);
	strcpy(new_pdb_file,argv[3]);

	if (strstr(old_pdb_file,".pdb")!=NULL) {
		templatetype=PDB;	
        	txtout("Template PDB file          : <%s>.\n",old_pdb_file);
	} else
	if (strstr(old_pdb_file,".lis")!=NULL) {
		templatetype=LIS;	
        	txtout("Template lis-file          : <%s>.\n",old_pdb_file);
	} else {
		errout("Fatal error: Unkown file format for template file.\n");
		return -1;	
	}	


	if (strstr(ego_file,".ego")!=NULL) {
		inputtype=EGO;	
        	txtout("EGO-output file to convert : <%s>.\n",ego_file);
	} else {
		inputtype=FFORM;	
        	txtout("Free file to convert       : <%s>.\n",ego_file);	
	}
	
        if (templatetype==PDB) {
        	txtout("New PDB file               : <%s>.\n",new_pdb_file);
        } else 
        if (templatetype==LIS) {
        	txtout("New lis-file               : <%s>.\n",new_pdb_file);
        }

	if (inputtype==EGO) {
		/***********************************************************/
		/****************** readin ego file  ***********************/
		/***********************************************************/
        	if ((ego_h=fopen(ego_file,"r"))==NULL) {
                	errout("Can not open/find ego-file <%s>. Fatal error.\n",ego_file);
			return -1;
		}
		if (read_trjego_header(ego_h,&numcoords,&framenum,&framestep,&timestep)<0) {
			fclose(ego_h);
			return -1;
		}
		xko = alloc_doubles(numcoords+10);
		yko = alloc_doubles(numcoords+10);
		zko = alloc_doubles(numcoords+10);
		charge = alloc_doubles(numcoords+10);
		for (j=0;j<numcoords;j++) {
			int res = read_ego_coords(ego_h,coords);
			if (res==-2) goto error_occured;
			if (res==-1) {
                        	errout("\nFatal error: End of file after %d read coords !\n",j);
				goto error_occured;
			}
			xko[j] = coords[0];
			yko[j] = coords[1];
			zko[j] = coords[2];
		}
		fclose(ego_h);
	} else
	if (inputtype==FFORM) {
		/***********************************************************/
		/****************** readin free format file  ***************/
		/***********************************************************/		
		int resort=0,table=0;
		double cfactor=1.0,x,y,z,myc;
		
       		if ((ego_h=fopen(ego_file,"r"))==NULL) {
                	errout("Can not open/find input-file <%s>. Fatal error.\n",ego_file);
			return -1;
		}
			
		while(read_string(ego_h,line)!=-1) {
			if (strstr(line,"[FREEFORMAT]")) break;
			if (strstr(line,"[FREEFORMAT1]")) {table=1;break;}
		}
		
		/*** next line contains number of atoms, unit-conversion-flag, shift-flag ***/
		read_string(ego_h,line);
		sscanf(line,"%d %d %d",&numcoords,&conversionflag,&shiftflag);

		txtout("Number of Atoms: %d\n",numcoords);
		if (conversionflag==0)	{
			txtout("No unit conversion is performed.\n");
			cfactor=1.0;
		}
		else
		if (conversionflag==1)	{
			txtout("Unit conversion bohr --> Angstroem.\n");
			cfactor=1.0/ANGS_TO_BOHR;
		}
		if (shiftflag==-1)	txtout("Coordiante system not shifted.\n");
		else			txtout("Setting atom number %d to origin.\n",shiftflag);

		/*** next line contains the format string in C-Style ***/
		read_string(ego_h,formatstr);
		txtout("Conversion string <%s>\n",formatstr);
		
		for (j=0,clf=0;j<strlen(formatstr)-2;j++) {
			if (	(formatstr[j+0]=='%') &&
				(formatstr[j+1]=='l') &&
				(formatstr[j+2]=='f') ) clf++;			
		}
		
		if ((clf<3)||(clf>4)) {
			errout("Fatal error: Illegal free format specification %d.\n",clf);
			fclose(ego_h);
			return -1;	
		} else
		if (strstr(formatstr,"%d")!=NULL) {
			resort=1;
			txtout("Using index numbers for resorting.\n");
		}

		if (clf==4) txtout("Using last %%lfd for charges\n");
		
		renum = alloc_ints(numcoords+10);
		xko = alloc_doubles(numcoords+10);
		yko = alloc_doubles(numcoords+10);
		zko = alloc_doubles(numcoords+10);
		charge = alloc_doubles(numcoords+10);
		
		/*** if we have had [FREEFORMAT1] we first read in ***/
		/*** the renumbering table 			   ***/		
		if (table) {
			for (j=0;j<numcoords;j++) {
				read_string(ego_h,line);	
				sscanf(line,"%d",&renum[j]);
				renum[j]--;
				if ((renum[j]<0)||(renum[j]>=numcoords)) {
					errout("Fatal error: Wrong index renumbering.\n");
					fclose(ego_h);
					return -1;
				}
			}
		}
	
		for (j=0;j<numcoords;j++) {
			read_string(ego_h,line);
			if (resort) {
				if (clf==3) sscanf(line,formatstr,&renum[j],&x,&y,&z);
				else        sscanf(line,formatstr,&renum[j],&x,&y,&z,&myc);
				
				renum[j]--;
				if ((renum[j]<0)||(renum[j]>=numcoords)) {
					errout("Fatal error: Wrong index renumbering.\n");
					fclose(ego_h);
					return -1;
				}
			} else {
				if (clf==3) sscanf(line,formatstr,&x,&y,&z);
				else        sscanf(line,formatstr,&x,&y,&z,&myc);
				if (!table) renum[j]=j;
			}
			xko[renum[j]] = x*cfactor; 
			yko[renum[j]] = y*cfactor; 
			zko[renum[j]] = z*cfactor;
			if (clf==4) charge[renum[j]] = myc;
		}
		
		if (shiftflag>=0) {
			double oriko[3];
			oriko[0] = xko[shiftflag-1];
			oriko[1] = yko[shiftflag-1];
			oriko[2] = zko[shiftflag-1];			
			for (j=0;j<numcoords;j++) {
				xko[j] -= oriko[0];
				yko[j] -= oriko[1];
				zko[j] -= oriko[2];
			}
		}
		
		fclose(ego_h);
	}
	
   	if (templatetype==PDB) {
        	if ((ego_h=fopen(old_pdb_file,"r"))==NULL) {
                	errout("Fatal error: Can not open/find pdb template file <%s>.\n",old_pdb_file);
			goto error_occured;
		}
		if ((ego_h_new=fopen(new_pdb_file,"w"))==NULL) {
                	errout("Fatal error: Can not create new pdb-file <%s>.\n",new_pdb_file);
			goto error_occured;
		}
		for (j=0;j<numcoords;j++) {
			patch_pdb_coords(ego_h,ego_h_new,xko[j],yko[j],zko[j]);
		}
		fprintf(ego_h_new,"END\n");
		fclose(ego_h);
		fclose(ego_h_new);
	} else if (templatetype==LIS) {
		STRUCT_ATOMDATA *atom_data;
		
		atom_data=alloc_atomdata(numcoords+5);
		if (read_coord_file(global_data,numcoords,old_pdb_file,atom_data)==-1) {
			errout("Fatal error in <%s> file!",old_pdb_file);
			return -1;
		}

		for (j=0;j<numcoords;j++) {
			atom_data[j].ko[0] = xko[j];
			atom_data[j].ko[1] = yko[j];
			atom_data[j].ko[2] = zko[j];
			if (clf==4) atom_data[j].charge = charge[j];
		}	

      		if (writeout_coordfile(numcoords,new_pdb_file,atom_data)==-1) {
                	errout("Fatal error while writing new coordinate-file!\n");
                	return -1;
        	}
	}

        txtout("\n======================================================\n");
        txtout("ego2pdb finished now. No fatal errors.\n");
        txtout("======================================================\n");
	return 0;

error_occured:
        txtout("\n======================================================\n");
        txtout("ego2pdb stopped with a fatal error.\n");
        txtout("Ego-output-files or template PDB-file may be corrupt.\n");
        txtout("======================================================\n");
	return -1;
}
