/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: ego2rib.c	by Helmut Heller, Munich		**
**	A utility to convert coordinates given in EGO format to		**
**	RIB format, for rendering scenes with Pixar's Renderman		**
**	which comes for free with all NeXTSTEP systems.			**
**	You may have trouble to compile this file on anything		**
**	that is not NeXTSTEP!						**
**									**
**	prman < ri.rib							**
** 	open ri.pic							**
**									**
**	Tx=19990923.							**
**									**
*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>


/* Special RenderMan interface: */
#include <ri.h>


#include "../hardw.h"
#include "../ego.h"
#include "../egalloc.h"
#include "../util_lib.h"
#include "../wodecls.h"
#include "../woalloc.h"
#include "../version.h"

static  LOAD_MSG_ALL_NODE *global_data;


/* DEBUG ==  1 : reading of ego files
   DEBUG ==  2 : rendering, early stages
*/
#define DEBUG 0

#define ATOM_C 0
#define ATOM_H 1
#define ATOM_O 2
#define ATOM_P 3
#define ATOM_N 4


/******************************************************************************/

int	get_atomtype_index(char *atomtypstr)
{
        strcpy(atomtypstr,"");
	return 0;
}

void   pex (int i)
{
	printf("Exit by pex(%d)!\n",i);
	fflush(stdout);

	exit(i);
	return;
}

int transparent (STRUCT_ATOMDATA *atomdata, int atomtype)
{
  /* decide, if an atom is to be made transparent 
     (means that it is harmonically fixated or,
      if it is a hydrogen, which is never fixated, some atom in its
      residue was fixated) */
  int transp=FALSE;
  static int current_harm_residnr=0;
  
  if(atomdata->harmon != 0)
  {
    transp = TRUE;
    current_harm_residnr = atomdata->residnr;
  }
  else if ((atomtype == ATOM_H)  && (atomdata->residnr == current_harm_residnr))
  {
    transp = TRUE;    
  }
  else if (atomdata->residnr != current_harm_residnr)
  {
    transp = FALSE;
    current_harm_residnr = -1;
  }
  else
  {
    transp = FALSE;
  }
  
  return transp; 
}

void PlaceObject(RtFloat xangle, RtFloat yangle, RtFloat distance)
{
  RtPoint from; 
  from[1]=-distance*sin(xangle/180.0*M_PI);from[2]=distance*cos(xangle/180.0*M_PI);
  from[0]=from[2]*sin(yangle/180.0*M_PI);from[2]=from[2]*cos(yangle/180.0*M_PI);
  RiTranslate(from[0],from[1],from[2]);
  if(xangle != 0) RiRotate(xangle,1,0,0);
  if(yangle != 0) RiRotate(yangle,0,1,0);
}

void PlaceCamera(RtFloat xangle, RtFloat yangle, RtFloat distance)
{ /* since we don't really place the camera, but instead move the world the other
     direction, we have to change a few signs to make it LOOK as if we placed
     the camera!
   */
  RtPoint from; 
  from[1]=distance*sin(xangle/180.0*M_PI);from[2]=-distance*cos(xangle/180.0*M_PI);
  from[0]=from[2]*sin(yangle/180.0*M_PI);from[2]=from[2]*cos(yangle/180.0*M_PI);
  /* since we move the world to simulate a moving camera, we have to
     FIRST translate
     SECOND rotate */
  if(yangle != 0) RiRotate(-yangle,0,1,0);
  if(xangle != 0) RiRotate(-xangle,1,0,0);
  RiTranslate(from[0],from[1],from[2]);
}

int cut_away (STRUCT_ATOMDATA *atom)
{
  /* cut away a part of the membrane to look inside */
  int cut=0;
  float x=atom->ko[0],y=atom->ko[1],z=atom->ko[2];

  if(strcmp(atom->residstr,"TIP3") == 0)
  { /* it is a water */
    cut=((x>0 || z>0) && y>0); 
  }
  else
  { /* no water */
    cut=(x>0 && y>0);
  }
  
  return (cut);
}

void DoObject(int Natom, STRUCT_ATOMDATA *atom_data)
{
  int i;       
  RtColor Transparency = {0.1,0.1,0.1};
  
  /* now we have the array AtomRad of atom vdW radii and the corresponding
      array AtomRadC of 1 letter characters designating the atom types */
#       define MAX_ATOM_TYPES 5
  float	AtomRad[MAX_ATOM_TYPES] ={1.5,0.8,1.4,1.7,1.6};
  /* int	AtomRadC[MAX_ATOM_TYPES]={'C','H','O','P','N'}; */
  RtColor AtomColor[MAX_ATOM_TYPES] ={
	  {133./255.,133./255.,113./255.}, /* C */
	  {255./255.,255./255.,255./255.}, /* H */
	  {230./255.,  0./255.,  0./255.}, /* O */
	  {250./255.,208./255.,  0./255.}, /* P */
	  { 34./255., 34./255.,255./255.}, /* N */
	  };
  int	AtomRadI[256]; /* index from characters to atom radii */
  RtColor CholColor = { 20./255.,160./255., 20./255.};
  AtomRadI['C'-'A'] = ATOM_C;
  AtomRadI['H'-'A'] = ATOM_H;
  AtomRadI['O'-'A'] = ATOM_O;
  AtomRadI['P'-'A'] = ATOM_P;
  AtomRadI['N'-'A'] = ATOM_N;

  RiSolidBegin(RI_UNION);
    for (i=0;i<Natom;i++)
      if(!cut_away(&(atom_data[i]))) /* cut away a part to look inside */
      {
	int atomtype = AtomRadI[(char)atom_data[i].atomstr[0]-'A'];
	RtFloat radius = AtomRad[atomtype];
	RiAttributeBegin();

	  /* We color cholesterols in green, regardless of the atom types */
	  if(strcmp(atom_data[i].residstr,"CHOL") == 0)
	    RiColor(CholColor);
	  else
	    RiColor(AtomColor[atomtype]);

	  /* all harmonically fixated atoms will be very transparent ! */
	  if(transparent(&(atom_data[i]), atomtype)) 
	  {
	    /* somehow the glass didn't work all too well. 
		I will try pure opacity instead: */
	    /* RiSurface("glass",RI_NULL); */
	    RiOpacity(Transparency);
	  }
	  else
	  {
	    RiSurface("plastic",RI_NULL);
	  }

	  RiTranslate(atom_data[i].ko[0],atom_data[i].ko[1],atom_data[i].ko[2]);
	  RiSphere(radius,-radius, radius,360.0,RI_NULL);
	RiAttributeEnd();
      }
  RiSolidEnd();
}

void DoBackground(int noback)
{
  RtColor BackgroundC = {19./255.,17./255.,125./255.};
  RtPoint BackgroundDim[4]={{300,300,50},{300,-300,50},{-300,-300,50},{-300,300,50}};
  /* the background */
  if(!noback)
  {
    RiAttributeBegin();
      RiShadingRate(64); /* high value, since the wall is unicolor */
      RiShadingInterpolation(RI_SMOOTH);
      RiSurface("plastic",RI_NULL);
      RiColor(BackgroundC);
      RiPolygon(4,RI_P,(RtPointer)BackgroundDim,RI_NULL);
    RiAttributeEnd();
  }
}

/****************************************************************************/
int main(int argc,char *argv[])
{
	int	i,j;
	FILE	*ego_handle;
	double	timestep;
	int	Natom,framenum,framestep;
	STRUCT_ATOMDATA *atom_data;

	char	rib_file[MAX_STRING_SIZE];
	char	ego_file[MAX_STRING_SIZE];
	char	template_file[MAX_STRING_SIZE];

	RtInt   xres=400, yres=300;
	int	noback=FALSE, 
	        draft=FALSE,
		verbose = FALSE,
	        shadow = TRUE; /* default is to produce shadows */
	RtFloat ShadingRate = 1;

	/* camera attributes: */
	RtFloat CameraFOV = 25.0,
	/* specification for camera position: */
                CameraXAngle=0, CameraYAngle=0, CameraDistance=-260;

	RtFloat AmInt = 0.1; /* ambient light intensity, keep it LOW */
	RtFloat DiInt = 1.0; /* distant (sun-like) light intensity */
	/* spot light parameters (important for shadow!) */
	RtFloat SpFOV=60, SpConeAngle=SpFOV*M_PI/360.0, SpInt = 20000.0,
		SpXAngle=10,    SpYAngle=40,  SpDistance=-100;

	/* We can rotate just the object alone: */
	RtFloat ObjectXAngle=-15, ObjectYAngle=110, ObjectDistance=0;

	
        txtout("==================================================================\n");
        txtout("ego2rib - Conversion of EGO coordinates to RIB format.\n");
        txtout("ego2rib - prman can then produce nice pictures from RIB files!\n");
        txtout("(%s) %s\n",VERSION,EGO_MACHINE);
        txtout("==================================================================\n");

	if (argc < 4) {
                txtout("ego2rib <template-file> <input-file> <output-file> [options]\n");
                txtout("Conversion of EGO coordinates to Pixar's RenderMan RIB format.\n");
		txtout("Example: You have an EGO-output (e.g., 00001200.ego) file and\n");
		txtout("         want to make a nicely rendered picture of it.\n");	
		txtout("         ego2rib coord.lis 0001200.ego 0001200.rib\n\n");
		txtout("options can be:\n");
		txtout("-draft speedy shading, lower quality\n");
		txtout("-v verbose mode\n");
		txtout("-noshadow do not produce shadows\n");
		txtout("-noback	no background\n");
		txtout("-xres <x-resolution>\n");
		txtout("-yres <y-resolution>\n");
		txtout("-Oxangle <angle in degrees> rotate object around x-axis\n");
		txtout("-Oyangle <angle in degrees> rotate object around y-axis\n");
		txtout("-Odistance <distance> object viewing distance\n");
		txtout("-Lxangle <angle in degrees> rotate spot light around x-axis\n");
		txtout("-Lyangle <angle in degrees> rotate spot light around y-axis\n");
		txtout("-Ldistance <distance> spot light distance\n");
		return 0;
	}
	
	for (i=4;i<argc;i++) 
	{
	  if      (strcmp(argv[i],"-xres") == 0) xres = atoi(argv[++i]);
	  else if (strcmp(argv[i],"-yres") == 0) yres = atoi(argv[++i]);
	  else if (strcmp(argv[i],"-noback") == 0) noback = TRUE;
	  else if (strcmp(argv[i],"-draft") == 0) draft = TRUE;
	  else if (strcmp(argv[i],"-v") == 0) verbose = TRUE;
	  else if (strcmp(argv[i],"-noshadow") == 0) shadow = FALSE;
	  else if (strcmp(argv[i],"-Oxangle") == 0) ObjectXAngle = atof(argv[++i]);
	  else if (strcmp(argv[i],"-Oyangle") == 0) ObjectYAngle = atof(argv[++i]);
	  else if (strcmp(argv[i],"-Odistance") == 0) ObjectDistance = atof(argv[++i]);
	  else if (strcmp(argv[i],"-Lxangle") == 0) SpXAngle = atof(argv[++i]);
	  else if (strcmp(argv[i],"-Lyangle") == 0) SpYAngle = atof(argv[++i]);
	  else if (strcmp(argv[i],"-Ldistance") == 0) SpDistance = atof(argv[++i]);
	  else txtout("\nUnknown option %s, ignored.\n",argv[i]);
	}
	
	/***************************************************************************/
	/*** Manage INPUT **********************************************************/
	/***************************************************************************/
	strcpy(template_file,argv[1]);
	strcpy(ego_file,argv[2]);
	strcpy(rib_file,argv[3]);

	if (strstr(ego_file,".ego") != NULL) {
        	txtout("EGO file to convert: <%s>\n",ego_file);
	} else {
        	txtout("No other input file formats than *.ego files are currently supported.\n");
		return(-1);	
	}

	if (strstr(template_file,".lis")!=NULL) {
        	txtout("Template lis-file  : <%s>\n", template_file);
	}  else {
        	txtout("We need the matching .lis file!\n");
		return(-2);	
	}
	
	if (strstr(rib_file,".rib")!=NULL) {
        	txtout("New RIB-file       : <%s>\n", rib_file);
	}  else {
        	txtout("No other output file formats than *.rib files are currently supported.\n");
		return(-3);	
	}
	

	/***********************************************************/
	/****************** read ego file  *************************/
	/***********************************************************/
	if ((ego_handle=fopen(ego_file,"r"))==NULL) {
		errout("Can not open/find ego-file <%s>. Fatal error.\n",ego_file);
		return -1;
	}
	if (read_trjego_header(ego_handle,&Natom,&framenum,&framestep,&timestep)<0) {
		fclose(ego_handle);
		return -1;
	}
	
#       if DEBUG == 1
          errout("read ego file\n");
#       endif

	/***********************************************************/
	/****************** read template file  ********************/
	/***********************************************************/		
	atom_data=alloc_atomdata(Natom+5);
	if (read_coord_file(global_data, Natom, template_file,atom_data) == -1) {
		errout("Fatal error in <%s> file!", template_file);
		return -1;
	}

#       if DEBUG == 1
          errout("read coord lis file\n");
#       endif

        /* transfer actual coordinates into data structure: */
	for (j=0;j<Natom;j++) {
		int res = read_ego_coords(ego_handle,atom_data[j].ko);
		if (res == -1) {
			errout("\nFatal error: End of file after %d read coords !\n",j);
		}
		if ((res == -1)||(res == -2)) 
		{
		  txtout("\n======================================================\n");
		  txtout("ego2rib stopped with a fatal error.\n");
		  txtout("Ego-output-file or template lis-file may be corrupt.\n");
		  txtout("======================================================\n");
		  return -200;
		}
	}
	fclose(ego_handle);
#       if DEBUG == 1
          errout("transferred data\n");
#       endif


        {
	/***********************************************************/
	/****************** do the actual rendering  ***************/
	/***********************************************************/		
	  char  *archive = "archive", *filename;
	  char	pic_file[MAX_STRING_SIZE];
          char *shadowfile = "shadow.shd";
          RtToken SHADOWFILE, RI_ZFILE, SHADOWSPOT, RI_INTENSITY, 
	          RI_TO, RI_FROM, RI_CONEANGLE, BUCKETSIZE;
          int bucketsize[2] = {20,20}, gridsize;
	  
          /* this is a weird "feature": the storage for filename MUST
	     be allocated on the heap, it seems... */
          filename = alloc_chars(MAX_STRING_SIZE);
	  strcpy(filename,rib_file);
	  
	  /* start the renderer to a file */
	  RiBegin("membrane", RI_RENDERER,&archive, RI_FILEPATH,  &filename, RI_NULL); 
	    SHADOWFILE=RiDeclare("shadowname","uniform string");
	    SHADOWSPOT=RiDeclare("shadowspot","uniform string");
	    RI_ZFILE=RiDeclare("zfile","uniform string");
	    /* now we have a problem with the shadowspot shader: it doesn't recognize the intensity token!
		To circumvent this, I re-declared intensity here, and this works! */   
	    RI_CONEANGLE=RiDeclare("coneangle","uniform float");
	    RI_INTENSITY=RiDeclare("intensity","uniform float");
	    RI_TO =      RiDeclare("to","uniform point");
	    RI_FROM =    RiDeclare("from","uniform point");
            BUCKETSIZE = RiDeclare("bucketsize","uniform int");

	    if(draft)
	    { 
	      RtToken GRIDSIZE;
              GRIDSIZE   = RiDeclare("gridsize","uniform int");

	      txtout("Running in DRAFT mode! Quality is reduced, speed is increased.\n");

	      /* a lower resolution also helps a lot to gain speed! */
	      xres = xres/3.0;
	      yres = yres/3.0;
	      ShadingRate = 10; /* higher values speed up rendering, 
	                           lower values give better resolution */
	      RiShadingInterpolation(RI_SMOOTH); /* this will result in Gouraud shading */
	      RiPixelSamples(1,1); /* no anti aliasing: rough edges, but faster rendering! */
	      RiOption("limits", BUCKETSIZE, (RtPointer)bucketsize, RI_NULL); 
	      gridsize = bucketsize[0]/ShadingRate;
	      RiOption("limits", GRIDSIZE,(RtPointer) &gridsize,RI_NULL);
	      shadow = FALSE; /* no shadows to speed things up! */
	    }
	    else
	    {
	      RiShadingInterpolation(RI_CONSTANT);
	    }
	    RiShadingRate(ShadingRate);

	    if(verbose)
	    {
	      txtout("\nVerbose mode is on and will tell you everything!\n");
	      txtout("xres=%d, yres=%d, FOV=%5.1f\n",xres,yres, CameraFOV);
	      txtout("CameraXAngle=%7.2f, CameraYAngle=%7.2f, CameraDistance=%7.2f\n", 
	              CameraXAngle, CameraYAngle, CameraDistance);
	      txtout("ObjectXAngle=%7.2f, ObjectYAngle=%7.2f, ObjectDistance=%7.2f\n", 
	              ObjectXAngle, ObjectYAngle, ObjectDistance);
	      txtout("LightXAngle= %7.2f, LightYAngle= %7.2f, LightDistance= %7.2f\n", 
	              SpXAngle, SpYAngle, SpDistance);
	      txtout("Spot light intensity=    %4.1f, spot beam angle=%5.1f\n",SpInt, SpFOV);
	      txtout("Ambient light intensity= %4.1f\n",AmInt);
	      txtout("Distant light intensity= %4.1f\n",DiInt);
	      txtout("ShadingRate=%7.2f,  bucketsize=%d,   gridsize=%d\n", 
	              (float)ShadingRate, (int)bucketsize[0], (int)gridsize);
	      if(!shadow)txtout("Not u");else txtout("U");txtout("sing shadows.\n");
	      if(noback) txtout("Not u");else txtout("U");txtout("sing background.\n");
	    }

	    if(shadow)
	    {
	      RtInt shadowxres, shadowyres; /* have to be powers of two! */
	      shadowxres = pow(2,((int) (log(xres)/log(2.0))));
	      shadowyres = pow(2,((int) (log(yres)/log(2.0))));

	      if(verbose)
	      {
		txtout("ShadowXRes=%d, ShadowYRes=%d\n", shadowxres, shadowyres);
	      }
	      /* first we render the shadow frame */
	      RiFrameBegin(1);
	        RiFormat(shadowxres,shadowyres,1.0);
                RiDisplay("shadow.z", RI_ZFILE, RI_Z, RI_NULL); /* this frame will produce a shadow map */
                RiProjection("perspective",RI_FOV, (RtPointer)&SpFOV,RI_NULL);
                PlaceCamera(SpXAngle, SpYAngle, SpDistance) ;
		RiWorldBegin();	
		  RiTransformBegin();
		    /* We can rotate just the membrane: */
		    PlaceObject(ObjectXAngle, ObjectYAngle, ObjectDistance) ;
		    DoObject(Natom, atom_data);
		  RiTransformEnd();
	          DoBackground(noback);
		RiWorldEnd();
              RiFrameEnd();
	      RiMakeShadow("shadow.z",shadowfile,RI_NULL); /* here we convert the shadow map */
	      txtout("Shadow map produced. Now producing RIB file.\n");
	    }

            RiFrameBegin(2);
  	      /* display parameters */
	      strncat(pic_file, rib_file,strlen(rib_file)-4);
	      strcat(pic_file, ".tiff"); /* output file format is TIFF ! */
#             if DEBUG == 2
		errout("picture file: %s, xres %d, yres %d\n", pic_file, xres, yres);
#             endif
	      RiDisplay((RtToken)&pic_file,RI_FILE,RI_RGBA,RI_NULL);
	      RiFormat(xres,yres,1.0);
	      /* RiProjection(RI_ORTHOGRAPHIC,RI_NULL); */
	      RiProjection(RI_PERSPECTIVE,RI_FOV, (RtPointer)&CameraFOV,RI_NULL);
    
	      /* we place the camera as if it was in world space */
	      PlaceCamera(CameraXAngle, CameraYAngle, CameraDistance);
	      RiWorldBegin();
	
		/* Lights: */
                RiTransformBegin(); /* set up the light sources! Detach them from any objects */
		  RiLightSource(RI_LIGHTSOURCE, RI_DISTANTLIGHT, RI_INTENSITY,(RtPointer)&DiInt,RI_NULL);    
		  RiLightSource(RI_LIGHTSOURCE, RI_AMBIENTLIGHT, RI_INTENSITY,(RtPointer)&AmInt,RI_NULL);    
                  PlaceObject(SpXAngle, SpYAngle, SpDistance) ;
                  if(shadow)
		  {
		    RiLightSource(RI_LIGHTSOURCE, SHADOWSPOT, 
		      SHADOWFILE, (RtPointer)&shadowfile,
		      RI_INTENSITY, (RtPointer)&SpInt, 
		      RI_CONEANGLE,(RtPointer)&SpConeAngle,
		      RI_NULL); 
		  }
		  else
		  {
		    RiLightSource(RI_LIGHTSOURCE, RI_SPOTLIGHT, 
		      RI_INTENSITY, (RtPointer)&SpInt, 
		      RI_CONEANGLE,(RtPointer)&SpConeAngle,
		      RI_NULL); 
		  }
                  {
		    RtColor Green={0,1,0};
		    RiSphere(1,-1,1,360.0,RI_CS,Green,RI_NULL); /* to indicate the light position */
		  }
                RiTransformEnd();
		
                RiTransformBegin();
		  /* We can rotate just the membrane: */
		  PlaceObject(ObjectXAngle, ObjectYAngle, ObjectDistance) ;
		  DoObject(Natom, atom_data);
                RiTransformEnd();
		DoBackground(noback); 
	      RiWorldEnd();
	    RiFrameEnd();
	  RiEnd();
	  myfree(filename);
	}
	
        txtout("\n======================================================\n");
        txtout("ego2rib finished now. No fatal errors.\n");
        txtout("======================================================\n");
	txtout("\n\n\n");
	return 0;
}
