/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: force2me.c	by M.Eichinger, Munich			**
**									**
*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#ifdef SUN4
#include <unistd.h>
#endif

#define DEBUG 0

#define MAX_STRING_SIZE	250
#define MAX_ATOMS	100
#define MAX_LISTS	10
#define ATOMSTRLEN	4

static	int	atom_nr[MAX_ATOMS];
static	int	res_nr[MAX_ATOMS];
static	int	atom_str[MAX_ATOMS][ATOMSTRLEN+1];
static	int	res_str[MAX_ATOMS][ATOMSTRLEN+1];
static	double	fx[MAX_ATOMS],fy[MAX_ATOMS],fz[MAX_ATOMS],e[MAX_ATOMS];
static	double	kx[MAX_ATOMS],ky[MAX_ATOMS],kz[MAX_ATOMS],q[MAX_ATOMS];

#define	LIST	0
#define RANGE	1

#define SUM	0
#define AVE	1
#define DIST	2
#define POS	3
#define SUM_NEG	4
#define AVE_NEG	5
#define POS_NEG	6

static	int	numlists=0;
static	int	calcmode[MAX_LISTS];
static	int	listmode[MAX_LISTS];
static	int	listsize[MAX_LISTS];
static	int	listelem[MAX_LISTS][MAX_ATOMS];

static	double	sum_x[MAX_LISTS],sum_y[MAX_LISTS],sum_z[MAX_LISTS];
static	int	sum_count[MAX_LISTS];

static	char	longline[100*MAX_STRING_SIZE];
static	char	sdum[MAX_STRING_SIZE];

static int	print_information(void);
static int	skip_header(void);
static int	scan_header(void);
static int	read_string(FILE *f,char *str);
static int	parse_forces(void);
static int	parse_forces(void);
static int	is_in_list(int atom_nr,int listi);
static int	parse_list(char *newname);

static 	FILE	*oldh,*newh;

#define	EGO_TRAJ	0
#define QM_TRAJ		1
static	int	input_format=EGO_TRAJ;

#define	OUT_DEFAULT	0
#define OUT_SUMALL	1
#define OUT_SUMPART	2
static	int	output_format=OUT_DEFAULT;

static	int	nr_atoms=0;
static	int	step=0;
static	long	seekpos=0L;
static	double	qm_energy[2];
static	int	rows=0;

/*****************************************************************************/
int main(int argc,char *argv[])
{
	int	i,j;
	char	newname[MAX_STRING_SIZE];

	numlists=0;
	
        printf("===========================================================\n");
        printf("force2me - Converting EGO-force files to special format.\n");
        printf("============================================================\n");
	if (argc<2) {
		print_information();
		return 0;
	}

	for (i=1;i<argc;i++) {
		int len=strlen(argv[i]);
		if (len<2)	continue;
		strcpy(newname,argv[i]);
		
		if (newname[0]=='-') { 

			if (strcmp(newname,"-default")==0) {
				printf("Output: s e fx fy fz (all in one line)\n");
				output_format = OUT_DEFAULT;	
			}
			if (strcmp(newname,"-sumall")==0) {
				printf("Output: s sum_e sum_fx sum_fy sum_fz\n");
				output_format |= OUT_SUMALL;	
			}

			if (newname[1]=='s') calcmode[numlists]=SUM;
			if (newname[1]=='a') calcmode[numlists]=AVE;
			if (newname[1]=='p') calcmode[numlists]=POS;
			if (newname[1]=='S') calcmode[numlists]=SUM_NEG;
			if (newname[1]=='A') calcmode[numlists]=AVE_NEG;
			if (newname[1]=='P') calcmode[numlists]=POS_NEG;
			if (newname[1]=='d') {
				int j;

				for (j=2;j<strlen(newname);j++) {
					if (newname[j]=='_') {
						newname[j] = 0;
						calcmode[numlists]=DIST;
						if (parse_list(newname)==-1) return -1;

						newname[j]='_';
						calcmode[numlists]=DIST;
						if (parse_list(newname+j-1)==-1) return -1;
					}
				}
				continue;
			}

			if (parse_list(newname)==-1) return -1;
	
			continue;
		} else {
			strcpy(newname+len-4,".txt");	
			printf("Converting file %s --> %s \n",argv[i],newname);
		}
		
		if ((oldh=fopen(argv[i],"r"))==NULL) {
                        printf("\nFatal error: Can not open <%s>. Skipping it !\n",
				argv[i]);
			continue;
		}
		if ((newh=fopen(newname,"w"))==NULL) {
                        printf("\nFatal error: Can not create <%s>. Skipping it !\n",
				newname);
			fclose(oldh);
			continue;
		}
		
		for (j=0;j<argc;j++) {
			fprintf(newh,"%s ",argv[j]);			
		}
		fprintf(newh,"\n");
		
		if (skip_header()==-1) {
			fclose(oldh);
			fclose(newh);
			return -1;
		}
		
		if (scan_header()==-1) {
			fclose(oldh);
			fclose(newh);
			return -1;	
		}
		
		/*** loop over all steps ***/
		step=0;
		while (read_string(oldh,sdum)!=-1) {
			
			if (input_format==EGO_TRAJ) {
				sscanf(sdum,"%s %d",newname,&step);
			} else
			if (input_format==QM_TRAJ) {
				qm_energy[0] = qm_energy[1] = 0.0;
				sscanf(sdum,"%s %d %lf %lf",
					newname,&step,
					&qm_energy[0],&qm_energy[1]);
			}

			parse_forces();

			if (input_format==EGO_TRAJ) {
				fprintf(newh,"\n");
			} else 
			if (input_format==QM_TRAJ) {
				fprintf(newh,"% 10.5f % 10.5f\n",qm_energy[0],qm_energy[1]);
				rows += 2;
			}
		}
		fseek(newh,seekpos,SEEK_SET);
		fprintf(newh,"%d %d %d",nr_atoms,step+1,rows);
		fclose(oldh);
		fclose(newh);
	}

	return 0;
}

static int
print_information()
{
	printf("Converts the standard EGO force output-formats to\n");
	printf("other formats which are easier to parse.\n");
	printf("-default             all forces of all atoms in one line.\n");
	printf("-p3,4,5              average position of these atoms as axis.\n");
	printf("-d3,4,5_1,2,3        distance of average position as axis.\n");
	printf("-sumall              sum up all forces.\n");
	printf("-s3,5,6              sum up forces of these atoms.\n");
	printf("-s3-5                sum up forces of atom 3 to 5.\n");
	printf("-a3,5,6              average force of these atoms.\n");
	printf("Big key-letters (-P..., -A..., -S...) indicate that\n");
	printf("negative values are written to file.\n\n");
	return 0;	
}

static int	
read_string(FILE *f,char *str)
{
	int	i,ch;
	i=0;	
	str[0]=0;
	do {
		ch=fgetc(f);	
		if ((ch==EOF)||(ch=='\n')) {
			str[i]=0;
			if (ch==EOF)	return -1;
			else		return 0;
		} else {
			str[i++]= (char)ch;
		}
		if (i>=MAX_STRING_SIZE-2) {
			printf("Fatal error: Line too long for string-buffer (read_string).\n");
			printf("Actual stringbuffer-size is %d.\n",MAX_STRING_SIZE);
			str[i]=0;	
			return -1;
		}
	} while (ch!=EOF);
	str[i]=0;	
	return -1;
}

/*** read and write until "[BEGINFORCE]" or [QMTRAJECTORY] ***/
static int
skip_header()
{
	char str[MAX_STRING_SIZE];
	
	while (read_string(oldh,sdum)!=-1) {
		strcpy(str,sdum);
		str[12]=0;
		if (strcmp(str,"[BEGINFORCE]")==0) {
			input_format=EGO_TRAJ;
			return 0;		
		}
		if (strcmp(str,"[QMTRAJECTOR")==0) {
			input_format=QM_TRAJ;
			return 0;		
		}		
		
		fprintf(newh,"%s\n",sdum);
	}
	printf("This is no valid input force format!\n");
	return -1;
}

/*** read/write number of atoms ***/
static int
scan_header()
{
	int j;
	char str[MAX_STRING_SIZE];
		
	read_string(oldh,sdum);
	sscanf(sdum,"%s %d",str,&nr_atoms);
	if (nr_atoms>=MAX_ATOMS) {
		printf("MAX_ATOMS too small! Increase it!");
		return -1;	
	}

	if (input_format==EGO_TRAJ) {
		fprintf(newh,"[BEGINFORCE]  E: kcal/mol    F: kcal/(molA) \n");
	} else
	if (input_format==QM_TRAJ) {
		fprintf(newh,"[QMTRAJECTORY]  E: kcal/mol    F: kcal/(molA) \n");
	}

	seekpos = ftell(newh);
	fprintf(newh,"%d                         \n",nr_atoms);
	
	read_string(oldh,sdum);
	for (j=0;j<nr_atoms;j++) {
		read_string(oldh,sdum);
			
		sscanf(sdum,"%d %d %s %s",	&atom_nr[j],&res_nr[j],
						atom_str[j],res_str[j]);
						
		printf("          %-5d      %-5d      %-4s       %-4s\n",
						atom_nr[j],res_nr[j],
						atom_str[j],res_str[j]);
						
		fprintf(newh,"          %-5d      %-5d      %-4s       %-4s\n",
						atom_nr[j],res_nr[j],
						atom_str[j],res_str[j]);
	}

	if (input_format==EGO_TRAJ) {
		if (calcmode[0]==DIST)
		fprintf(newh,"Step     [Dist       PosX       PosY       PosZ       Fx         Fy         Fz]\n");
		else
		fprintf(newh,"Step     [PosX       PosY       PosZ       Fx         Fy         Fz]\n");
	} else
	if (input_format==QM_TRAJ) {
		if (calcmode[0]==DIST)
		fprintf(newh,"Step     [Dist       PosX       PosY       PosZ       Fx         Fy         Fz]        Etot       Eext\n");
		else
		fprintf(newh,"Step     [PosX       PosY       PosZ       Fx         Fy         Fz]        Etot       Eext\n");
	}
	
	return 0;
}

static int
parse_forces()
{	
	int j,idum,i;
	double	sum_fx,sum_fy,sum_fz;
	double	ave_kx,ave_ky,ave_kz;
	
	rows=0;
	for (j=0;j<nr_atoms;j++) {
		read_string(oldh,sdum);
		
		if (input_format==EGO_TRAJ) {
			sscanf(sdum,"%d %lf %lf %lf %lf %lf %lf %lf",
					&idum,
					&e[j],
					&kx[j],&ky[j],&kz[j],
					&fx[j],&fy[j],&fz[j]);
			q[j] = 0.0;
		} else 
		if (input_format==QM_TRAJ) {
			sscanf(sdum,"%d %lf %lf %lf %lf %lf %lf %lf",
					&idum,
					&q[j],
					&kx[j],&ky[j],&kz[j],
					&fx[j],&fy[j],&fz[j]);
			e[j] = 0.0;
		}
	 
		/*************** Default Output *************************/
		if (output_format==OUT_DEFAULT) {
			if (j==0) {
				fprintf(newh,"%-6d % 10.5f % 10.5f % 10.5f % 10.5f",
						step,e[j],fx[j],fy[j],fz[j]);
				rows += 5;
					
			}
			else if ((j==nr_atoms-1)&&(input_format==QM_TRAJ)) {
				fprintf(newh,"    % 10.5f % 10.5f % 10.5f % 10.5f",
						q[j],fx[j],fy[j],fz[j]);
				
				rows += 4;
			} else {
				fprintf(newh,"    % 10.5f % 10.5f % 10.5f % 10.5f",
						e[j]+q[j],fx[j],fy[j],fz[j]);
				rows += 4;				
			}
			continue;
		}
		
		if (j==0) {
			fprintf(newh,"%-6d ",step);
			rows += 1;
		}
		/*************** Sum all forces *************************/
		/*************** Center of geometry *********************/	
		if (output_format&OUT_SUMALL) {
			/*** set vars to zero at the beginning of a new step ****/
			if (j==0) {
				sum_fx=sum_fy=sum_fz=0.0;
				ave_kx=ave_ky=ave_kz=0.0;
			}
			
			/*** add all up ***/
			sum_fx += fx[j];	/*** forces ***/
			sum_fy += fy[j];
			sum_fz += fz[j];
						
			ave_kx += kx[j];	/*** coords ***/
			ave_ky += ky[j];
			ave_kz += kz[j];
			
			/*** make output to the file ***/
			if (j==nr_atoms-1) {
				ave_kx /= (double) nr_atoms;
				ave_ky /= (double) nr_atoms;
				ave_kz /= (double) nr_atoms;	
			
				fprintf(newh,"% 10.5f % 10.5f % 10.5f % 10.5f % 10.5f % 10.5f ",
						ave_kx,ave_ky,ave_kz,	/*** center of geom ***/
						sum_fx,sum_fy,sum_fz);	/*** total force    ***/
				rows += 6;		
			}		
		}
		
		/*************** Sum parts *************************/
		if (output_format&OUT_SUMPART) {
			int distcount=0;

			/*** loop over all subsets ***/ 
			for (i=0,distcount=0;i<numlists;i++) {		
				if (j==0) {
					sum_x[i]=sum_y[i]=sum_z[i]=0.0;
					sum_count[i]=0;
				}

				if (is_in_list(idum,i)) {
					/*** sum or average of of forces ***/
					if (	(calcmode[i]==SUM)     || (calcmode[i]==AVE)    ||
						(calcmode[i]==SUM_NEG) || (calcmode[i]==AVE_NEG) ) {
						sum_x[i] += fx[j];
						sum_y[i] += fy[j];
						sum_z[i] += fz[j];
						sum_count[i]++;
					} else /*** Position or distance as x-axis ***/
					if (	(calcmode[i]==POS) || 
						(calcmode[i]==POS_NEG) || 
						(calcmode[i]==DIST)) {
						sum_x[i] += kx[j];
						sum_y[i] += ky[j];
						sum_z[i] += kz[j];
						sum_count[i]++;				
					}
				}

				/*** make output to the file ***/
				if (j==nr_atoms-1) {
					if (calcmode[i]==SUM) {
						fprintf(newh,"% 10.5f % 10.5f % 10.5f ",
							sum_x[i],sum_y[i],sum_z[i]);
						rows += 3;
					} else
					if (calcmode[i]==SUM_NEG) {
						fprintf(newh,"% 10.5f % 10.5f % 10.5f ",
							-sum_x[i],-sum_y[i],-sum_z[i]);
						rows += 3;
					} else
					if ((calcmode[i]==AVE)||(calcmode[i]==POS)) {
						fprintf(newh,"% 10.5f % 10.5f % 10.5f ",
							sum_x[i]/sum_count[i],
							sum_y[i]/sum_count[i],
							sum_z[i]/sum_count[i]);
						rows += 3;
					} else
					if ((calcmode[i]==AVE_NEG)||(calcmode[i]==POS_NEG)) {
						fprintf(newh,"% 10.5f % 10.5f % 10.5f ",
							-sum_x[i]/sum_count[i],
							-sum_y[i]/sum_count[i],
							-sum_z[i]/sum_count[i]);
						rows += 3;
					} else
					if (calcmode[i]==DIST) {
						double dx,dy,dz,dist;
						if (distcount==0) {
							distcount=1;
							continue;
						}
						dx=sum_x[i]/sum_count[i]-sum_x[i-1]/sum_count[i-1];
						dy=sum_y[i]/sum_count[i]-sum_y[i-1]/sum_count[i-1];
						dz=sum_z[i]/sum_count[i]-sum_z[i-1]/sum_count[i-1];
						dist = sqrt(dx*dx + dy*dy + dz*dz);

						fprintf(newh,"% 10.5f % 10.5f % 10.5f % 10.5f ",
							dist,dx,dy,dz);
						rows += 4;

						distcount=0;
					}
				}
			} /*** loop over all subsets ***/
		} /*** if sumpart ***/
	} /*** loop over all atoms ***/
	
	return 0;
}

static int
is_in_list(int atom_nr,int listi)
{
	int i;
	
	if (listmode[listi]==RANGE) {
		if (	(atom_nr>=listelem[listi][0]) &&
			(atom_nr<=listelem[listi][1]) ) return 1;
	} else
	if (listmode[listi]==LIST) {
		for (i=0;i<listsize[listi];i++) {
			if (atom_nr==listelem[listi][i]) return 1;
		}
	}	
	return 0;	
}

static int
parse_list(char *newname)
{
	int j;

	listmode[numlists] = LIST;
	listsize[numlists] = 1;

	/*** set all "," and "-" to spaces ***/
	for (j=2;j<strlen(newname);j++) {
		if (newname[j]==',') {
			newname[j] = ' ';
			listmode[numlists] = LIST;
			listsize[numlists]++;	
		}
		if (newname[j]=='-') {
			newname[j] = ' ';
			listmode[numlists] = RANGE;
			listsize[numlists] = 2;		
		}
	}

	if (listmode[numlists]==-1) {
		printf("Illegal paramter format <%s>.\n",newname);
		return -1;	
	}

	if (listmode[numlists]==RANGE) {
		sscanf(newname+2,"%d %d",
			&listelem[numlists][0],&listelem[numlists][1]);
		printf("Listrange: Atom %d - %d\n",
			listelem[numlists][0],listelem[numlists][1]);
		listsize[numlists] = listelem[numlists][1] - listelem[numlists][0];	
	} else 
	if (listmode[numlists]==LIST) {
		int l,k=2;
		printf("List elements: ");
		for (j=0;j<listsize[numlists];j++) {
			sscanf(newname+k,"%d",&listelem[numlists][j]);
			printf("%d,",listelem[numlists][j]);
			for (l=k;l<strlen(newname);l++) {
				if (newname[l]==' ') break;
			}
			k=l+1;
		}
		printf("\n");	
	}
			 
	if (listsize[numlists]<=0) {
		printf("Illegal paramter format <%s>.\n",newname);
		return -1;					
	}
					
	numlists++;
	if (numlists>=MAX_LISTS) {
		printf("MAX_LISTS too small!\n");
		return -1;
	}
	output_format |= OUT_SUMPART;	
	return 0;
}
