/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*********************************************************************************
**	Belongs to: listest - Some basics check of LIS-data for EGO		**
**										**
**	Module	: listest.c	 by M.Eichinger, Munich 			**
**	Contents: Gives some basic data about your lis-files			**
**										**
**      Modif's:                                                                **
**		15.06.94							**
**										**
*********************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#if defined(NEXT) || defined(RS6K) || defined(MAC)
#include <stdlib.h>
#else
#include <malloc.h>
#endif
#include "../ego.h"
#include "../util_lib.h"
#include "../woalloc.h"

#define DEBUG_UNITS	0

/************* defined functions in listest.c *************************/

static	void	print_prog_header();
static	void	print_wrong_call();
static	int	work_on_vdwfile(char *vdw_filen);
static	int	work_on_typefile(char *type_filen);
static	int	work_on_coordfile(LOAD_MSG_ALL_NODE *global_data,char *coord_filen);
static	int	work_on_units(char *unit_filen);
static	int	make_unit_stat(char *cntfn);

/**********************************************************************/

typedef struct {
	int	atoms;
	int	charges;
	int	type;
	int	character;
	int	child_pos;
	double	cen_geom[3];		/* center of geometry */
} unitdata;

/**********************************************************************/

static  char    version_string[]={"Version 1.0, 15.07.1995, by M.Eichinger, Munich\n"};
static	char	curpath[MAX_STRING_SIZE];
static	char	filenames[MAX_FILES][MAX_STRING_SIZE];

static	int	nr_atomtypes=0;
static	int	nr_of_atoms=0;
static	double	mass[MXNO_ATOM_TYPES];
static	char	atom_typestr[MXNO_ATOM_TYPES+1][LEN_ATOM_TYPE+1];

static	char	sdum[MAX_STRING_SIZE];		/* space for dummy-strings */

static	STRUCT_ATOMDATA *atom_data;
static LOAD_MSG_ALL_NODE *global_data;

/*********************************************************************/

int	main(int argc,char *argv[])
{
	FILE	*fopen();
	FILE	*h_ctlfile;
	int	i;
	int	nr_files;
	char	ctlfilen[MAX_STRING_SIZE];

	print_prog_header();

	if (argc==1) {
                strcpy(ctlfilen,"ctl.lis");
	} else
	if (argc==2) {
		strcpy(ctlfilen,argv[1]);
	} else {
		print_wrong_call();
		return -1;
	}

	extract_path(ctlfilen,curpath);
						/* open my ctl-file */
        if ((h_ctlfile=fopen(ctlfilen,"r"))==NULL) {
                errout("\nCan not read control file <%s>.\n",ctlfilen);
		return -1;
	}
	txtout("Using ctontrol-file <%s>\n",ctlfilen);
	
						/* get number of files */
	nr_files=read_int_skip(h_ctlfile);
	if (nr_files!=CUR_LISFILES) {
                errout("Illegal number of files given in control file !\n");
                errout("This may happen if you use different versions of EGO !\n");
		return -1;
	}
						 /* read all filenames */
	for(i=0;i<nr_files;i++) {
		strcpy(filenames[i],curpath);
                fscanf(h_ctlfile,"%s",filenames[i]+strlen(curpath));
	}

						/* number of atoms */
	skip_line(h_ctlfile,2);
	nr_of_atoms=read_int_skip(h_ctlfile);
        txtout("Number of atoms is %d.\n",nr_of_atoms);

	fclose(h_ctlfile);

	/***** buffer for holding atom-data ********************************/
	atom_data=alloc_atomdata(nr_of_atoms+5);

	/**** read vdw-file ===> in diagonal there are the masses *******/
	txtout("Using vdw-file <%s>\n",filenames[VDWPARA_FILE_INDEX]);
	if (work_on_vdwfile(filenames[VDWPARA_FILE_INDEX])==-1) {
                errout("Fatal error in VDW-File !\n");
		free(atom_data);
		return -1;
	}
						/* work on atom-type-file */
	if (work_on_typefile(filenames[TYPE_FILE_INDEX])==-1) {
                errout("Fatal error in Atom-Type-File !\n");
		free(atom_data);
		return -1;
	}
						/* print what we got */
	for (i=0;i<nr_atomtypes;i++) {
                txtout("Mass of atom typ %2d [%-4s] is %.5g\n",
					i,atom_typestr[i],mass[i]);
	}
	
	txtout("Using coord-file <%s>\n",filenames[COORD_FILE_INDEX]);
	if (work_on_coordfile(global_data,filenames[COORD_FILE_INDEX])==-1) {
                errout("Fatal error in coordinate file <%s>.\n",filenames[COORD_FILE_INDEX]);
		free(atom_data);
		return -1;
	}

	if (work_on_units(filenames[UNIT_FILE_INDEX])==-1) {
                errout("Fatal error in unit file <%s>.\n",filenames[UNIT_FILE_INDEX]);
		free(atom_data);
		return -1;
	}

	free(atom_data);
	return 0;
}

static
void	print_prog_header()
{
        txtout("\nPrints some properties of your lis-files\n");
        txtout("========================================\n");
	txtout(version_string);
}

static
void	print_wrong_call()
{
        txtout("Program-Usage:\n");
        txtout("listest <ctlfile.lis>\n");
        txtout("<ctlfile.lis>    Control file (default is <ctl.lis>)\n");
}

/*********************************************************************************/

static
int	work_on_vdwfile(char *vdw_filen)
{
	FILE	*fopen();
	FILE	*h_vdwfile;
	int	i,j;
	double	x;
						/* open my vdw-file */
        if ((h_vdwfile=fopen(vdw_filen,"r"))==NULL) {
                errout("Can not open vdw-file <%s>.\n",vdw_filen);
		return -1;
	}
						/* skip header */
	skip_line(h_vdwfile,1);
        fscanf(h_vdwfile,"%s%s%s%s%s%d",sdum,sdum,sdum,sdum,sdum,&nr_atomtypes);
	if (nr_atomtypes<1)	return -1;

        txtout("Atomtypes in VDW-File : %d\n",nr_atomtypes);
						/* get atommasses in diag */
	for (i = 0; i<=nr_atomtypes; i++) {
		for (j = 0; j<=nr_atomtypes; j++) {
			if (i==j) {
				mass[j]=read_double(h_vdwfile);
			} else {
				x=read_double(h_vdwfile);
			}
		}
	}
	fclose(h_vdwfile);
	return 0;
}

/**************************************************************************************/

static
int	work_on_typefile(char *type_filen)
{
	FILE	*fopen();
	FILE	*h_typefile;
	int	i,index,sl;
	static	char	typestr[MAX_STRING_SIZE];

						/* open my atom-type-file */
        if ((h_typefile=fopen(type_filen,"r"))==NULL) {
                errout("Can not open atom-type-file <%s>.\n",type_filen);
		return -1;
	}

	skip_line(h_typefile,6);

	for (i = 0; i<nr_atomtypes; i++) {
                fscanf(h_typefile,"%d %s",&index,typestr);
		if (i!=index) {
                        errout("Fatal error: Illegal numbering in atom-type-file.\n");
			fclose(h_typefile);
			return -1;
		}
		sl=strlen(typestr);
		if ((sl==0)||(sl>LEN_ATOM_TYPE)) {
                        errout("Fatal error: Illegal string-len in atom-type-file.\n");
			fclose(h_typefile);
			return -1;
		}
		strcpy(atom_typestr[i],typestr);
	}
	fclose(h_typefile);
	return 0;
}

/**************************************************************************************/

static
int	work_on_coordfile(LOAD_MSG_ALL_NODE *global_data,char *coord_filen)
{
	int	i;
	double	sum_charge,sum_mass;
	double	com[3]; /* center of mass */
	double	vel[3]; /* velocity of COM */
	double	bbox[6];

	if (read_coord_file(global_data,nr_of_atoms,coord_filen,atom_data)==-1) {
                errout("Fatal error in coord-file !\n");
		return -1;
	}
						/* sum mass and charges */
	sum_mass=sum_charge=0.0;
	com[0]=com[1]=com[2]=0.0;
	vel[0]=vel[1]=vel[2]=0.0;

	for (i=0;i<nr_of_atoms;i++) {
		if (	(atom_data[i].atomtype<0) ||
			(atom_data[i].atomtype>=nr_atomtypes)) {
                        errout("Fatal error: Illegal line in coord-file (Atom-type-number !).\n");
			return -1;
		}
		sum_mass  += mass[atom_data[i].atomtype];
		sum_charge+= atom_data[i].charge;

		com[0]	  += mass[atom_data[i].atomtype]*atom_data[i].ko[0];
		com[1]	  += mass[atom_data[i].atomtype]*atom_data[i].ko[1];
		com[2]	  += mass[atom_data[i].atomtype]*atom_data[i].ko[2];

		vel[0]	  += mass[atom_data[i].atomtype]*atom_data[i].vel[0];
		vel[1]	  += mass[atom_data[i].atomtype]*atom_data[i].vel[1];
		vel[2]	  += mass[atom_data[i].atomtype]*atom_data[i].vel[2];
		if (i==0) {
		     bbox[0] = bbox[1] = atom_data[i].ko[0];
		     bbox[2] = bbox[3] = atom_data[i].ko[1];
		     bbox[4] = bbox[5] = atom_data[i].ko[2];
		} else {
		     if (bbox[0]>atom_data[i].ko[0]) bbox[0]=atom_data[i].ko[0];
		     if (bbox[1]<atom_data[i].ko[0]) bbox[1]=atom_data[i].ko[0];

		     if (bbox[2]>atom_data[i].ko[1]) bbox[2]=atom_data[i].ko[1];
		     if (bbox[3]<atom_data[i].ko[1]) bbox[3]=atom_data[i].ko[1];

		     if (bbox[4]>atom_data[i].ko[2]) bbox[4]=atom_data[i].ko[2];
		     if (bbox[5]<atom_data[i].ko[2]) bbox[5]=atom_data[i].ko[2];
		}
	}


        txtout("\nTotal charge         : % .6g.\n",sum_charge);
        txtout("Total mass           : % .6g.\n",sum_mass);
        txtout("Center of mass (COM) : (% .6g, % .6g, % .6g).\n",
						com[0]/sum_mass,
						com[1]/sum_mass,
						com[2]/sum_mass);
        txtout("Velocity of COM      : (% .6g, % .6g, % .6g).\n",
						vel[0]/sum_mass,
						vel[1]/sum_mass,
						vel[2]/sum_mass);
        txtout("Bounding Box         : X: [% .6g, % .6g] dx=%.6g.\n",bbox[0],bbox[1],bbox[1]-bbox[0]);
        txtout("                       Y: [% .6g, % .6g] dy=%.6g.\n",bbox[2],bbox[3],bbox[3]-bbox[2]);
        txtout("                       Z: [% .6g, % .6g] dz=%.6g.\n",bbox[4],bbox[5],bbox[5]-bbox[4]);
	return 0;
}

/**********************************************************************************/

static	int	nr_of_units=-1;
static	int	nr_of_neutrals=-1;
static	int	nr_of_polar=-1;
static	int	nr_of_charged=-1;
static	int	max_nr_fields=-1;

static	unitdata	*unit_buf;
static	int		*child_buf;

static
int	work_on_units(char *unit_filen)
{
	FILE	*fopen();
	FILE	*h_unitfile;
	int	i,j,testi;
	int	child_pos;

        if ((h_unitfile=fopen(unit_filen,"r"))==NULL) {
                errout("Can not open unit-file <%s>.\n",unit_filen);
		return -1;
	}
						/* skip header */
	skip_line(h_unitfile,1);

	/*************** get number of units ********************************/
        fscanf(h_unitfile,"%s%s%s%d",sdum,sdum,sdum,&nr_of_units);
	if (nr_of_units<1) {
                errout("Fatal error: Illegal number of units.\n");
		fclose(h_unitfile);
		return -1;
	}
        fscanf(h_unitfile,"%s%s%s%d",sdum,sdum,sdum,&nr_of_neutrals);
	if (nr_of_neutrals<0) {
                errout("Fatal error: Illegal number of neutral units.\n");
		fclose(h_unitfile);
		return -1;
	}
        fscanf(h_unitfile,"%s%s%s%d",sdum,sdum,sdum,&nr_of_polar);
	if (nr_of_polar<0) {
                errout("Fatal error: Illegal number of polar units.\n");
		fclose(h_unitfile);
		return -1;
	}
        fscanf(h_unitfile,"%s%s%s%d",sdum,sdum,sdum,&nr_of_charged);
	if (nr_of_charged<0) {
                errout("Fatal error: Illegal number of charged units.\n");
		fclose(h_unitfile);
		return -1;
	}
        fscanf(h_unitfile,"%s%s%s%s%d",sdum,sdum,sdum,sdum,&max_nr_fields);
	if (max_nr_fields<1) {
                errout("Fatal error: Illegal number of fields.\n");
		fclose(h_unitfile);
		return -1;
	}

	txtout("Number of units (total)    : %d\n",nr_of_units);
	txtout("Number of units (neutral)  : %d\n",nr_of_neutrals);
	txtout("Number of units (polar)    : %d\n",nr_of_polar);
	txtout("Number of units (charged)  : %d\n",nr_of_charged);

	skip_line(h_unitfile,2);
	/************ now we scanned the header **********************************/
	unit_buf=(unitdata*)malloc(sizeof(unitdata)*(nr_of_units+5));
	if (unit_buf==NULL) {
                errout("Fatal error: Can not allocate space for unit-data.\n");
		fclose(h_unitfile);
		return -1;
	}
	child_buf=(int*)malloc(sizeof(int)*(nr_of_atoms+5));
	if (child_buf==NULL) {
                errout("Fatal error: Can not allocate space for unit-data.\n");
		fclose(h_unitfile);
		free(unit_buf);
		return -1;
	}

	/************ lets get data of units *************************************/
	child_pos=0;
	for (i=0;i<nr_of_units;i++) {
                if (fscanf(h_unitfile,"%d %d %d %d %d",
			&testi,
			&unit_buf[i].atoms,
			&unit_buf[i].charges,
			&unit_buf[i].type,
			&unit_buf[i].character)!=5) {
                        errout("Can not read unit-line.\n");
			free(unit_buf);
			free(child_buf);
			fclose(h_unitfile);
			return -1;
		}
#if DEBUG_UNITS
                txtout("%d  %d  %d  %d  %d\n",
			testi,
			unit_buf[i].atoms,
			unit_buf[i].charges,
			unit_buf[i].type,
			unit_buf[i].character);
#endif
		if (i!=testi) {
                        errout("Illegal numbering of units.\n");
                        errout("Occured in line %d\n",i);
			free(unit_buf);
			free(child_buf);
			fclose(h_unitfile);
			return -1;
		}

		/********** center of geometry ****************/
		unit_buf[i].cen_geom[0] = 0.0;
		unit_buf[i].cen_geom[1] = 0.0;
		unit_buf[i].cen_geom[2] = 0.0;

		unit_buf[i].child_pos=child_pos;

		for (j=0;j<unit_buf[i].atoms;j++) {
			int child_nr=read_int(h_unitfile);
			child_buf[child_pos++]=child_nr;
			unit_buf[i].cen_geom[0] += atom_data[child_nr].ko[0];
			unit_buf[i].cen_geom[1] += atom_data[child_nr].ko[1];
			unit_buf[i].cen_geom[2] += atom_data[child_nr].ko[2];
		}
		unit_buf[i].cen_geom[0] /= unit_buf[i].atoms;
		unit_buf[i].cen_geom[1] /= unit_buf[i].atoms;
		unit_buf[i].cen_geom[2] /= unit_buf[i].atoms;
	}

        if (make_unit_stat("unitstat.out")==-1) {
                errout("Fatal error: Error during creating Statistik-file for units.\n");
		free(unit_buf);
		free(child_buf);

		fclose(h_unitfile);
		return 1;
	}
        txtout("Statistic-file <unitstat.out> written.\n");
	free(unit_buf);
	free(child_buf);

	fclose(h_unitfile);
	return 0;
}

/***********************************************************************************/

#define MAX_ATOMS_IN_UNIT	50

#define NEUTRAL 	0
#define POLAR		1
#define CHARGED 	2

static
int	make_unit_stat(char *cntfn)
{
static	int	cnt_total[MAX_ATOMS_IN_UNIT];
static	int	cnt_charges[MAX_ATOMS_IN_UNIT];

static	int	cnt_neutral[MAX_ATOMS_IN_UNIT];
static	int	cnt_polar[MAX_ATOMS_IN_UNIT];
static	int	cnt_charged[MAX_ATOMS_IN_UNIT];

	int	i,j;
	FILE	*h_cnt;

	for (i=0;i<MAX_ATOMS_IN_UNIT;i++) {
		cnt_total[i]	= 0;
		cnt_charges[i]	= 0;

		cnt_neutral[i]	= 0;
		cnt_polar[i]	= 0;
		cnt_charged[i]	= 0;
	}

	for (i=0;i<nr_of_units;i++) {
		cnt_total[ unit_buf[i].atoms ]++;	/* frequency number of atoms */
							/* in unit */

		cnt_charges[ unit_buf[i].charges ]++;	/* frequency number of charged */
							/* atoms in unit */

		if (unit_buf[i].character==NEUTRAL) {
			cnt_neutral[ unit_buf[i].atoms ]++;
		}

		if (unit_buf[i].character==POLAR) {
			cnt_polar[ unit_buf[i].atoms ]++;
		}

		if (unit_buf[i].character==CHARGED) {
			cnt_charged[ unit_buf[i].atoms ]++;
		}
	}

	/************** open statistic-file *******************************************/
        if ((h_cnt=fopen(cntfn,"w"))==NULL) {
                errout("Can not open statistic file for unit-counts.\n");
		return -1;
	}
        fprintf(h_cnt,"Distribution of atoms to units.\n");
        fprintf(h_cnt,"%d\n",max_nr_fields+1);
        fprintf(h_cnt,"Atoms  Count units: total  charges neutral  polar  charged\n");

	for (i=0;i<=max_nr_fields;i++) {
                fprintf(h_cnt,"%3d                 %3d      %3d     %3d    %3d      %3d\n",
			i,
			cnt_total[i],
			cnt_charges[i],
			cnt_neutral[i],
			cnt_polar[i],
			cnt_charged[i]);
	}

        fprintf(h_cnt,"\n");
        fprintf(h_cnt,"%d\n",nr_of_units);
        fprintf(h_cnt,"Charge              Dipol                     RefP                       BBox    R-Gyr\n");

	for (i=0;i<nr_of_units;i++) {

		int	child_pos=unit_buf[i].child_pos;
		int	k;
		double	sum_charge=0.0;
		double	cen[3];
		double	dip[3];
		double	bbox[6];
		double	rgyr=0.0;

				/****************  get me the charges **************/
		for (j=0;j<unit_buf[i].charges;j++) {
			sum_charge += atom_data[child_buf[child_pos++]].charge;
		}
                fprintf(h_cnt,"% 5.2f    ",sum_charge);

		child_pos=unit_buf[i].child_pos;
		cen[0]=cen[1]=cen[2]=0.0;

		if (unit_buf[i].character==CHARGED) {
			/*************** get the center of charge *****************/
			for (j=0;j<unit_buf[i].charges;j++) {
				k=child_buf[child_pos++];
				cen[0]+=atom_data[k].charge*atom_data[k].ko[0];
				cen[1]+=atom_data[k].charge*atom_data[k].ko[1];
				cen[2]+=atom_data[k].charge*atom_data[k].ko[2];
			}
			cen[0]/=sum_charge;
			cen[1]/=sum_charge;
			cen[2]/=sum_charge;
		} else {
			/************* cen of geometry of charged atoms ***********/
			int c= (unit_buf[i].charges==0)?unit_buf[i].atoms:unit_buf[i].charges;
			for (j=0;j<c;j++) {
				k=child_buf[child_pos++];
				cen[0]+=atom_data[k].ko[0];
				cen[1]+=atom_data[k].ko[1];
				cen[2]+=atom_data[k].ko[2];
			}
			cen[0]/=c;
			cen[1]/=c;
			cen[2]/=c;
		}

				/************ get dipolmoment *********************/

		child_pos=unit_buf[i].child_pos;
		dip[0]=dip[1]=dip[2]=0.0;
		for (j=0;j<unit_buf[i].charges;j++) {
			k=child_buf[child_pos++];
			dip[0]+=atom_data[k].charge*(atom_data[k].ko[0]-cen[0]);
			dip[1]+=atom_data[k].charge*(atom_data[k].ko[1]-cen[1]);
			dip[2]+=atom_data[k].charge*(atom_data[k].ko[2]-cen[2]);
		}
                fprintf(h_cnt,"% 6.3f % 6.3f % 6.3f       % 6.3f % 6.3f % 6.3f       ",
			dip[0],dip[1],dip[2],
			cen[0],cen[1],cen[2]);

				/************* boundry box ***************************/
				/************* of all atoms in unit    ***************/
		child_pos=unit_buf[i].child_pos;
		k=child_buf[child_pos++];
		bbox[0]=bbox[1]=atom_data[k].ko[0];
		bbox[2]=bbox[3]=atom_data[k].ko[1];
		bbox[4]=bbox[5]=atom_data[k].ko[2];
		for (j=0;j<unit_buf[i].atoms-1;j++) {
			k=child_buf[child_pos++];
			if (bbox[0]>atom_data[k].ko[0]) bbox[0]=atom_data[k].ko[0];
			if (bbox[1]<atom_data[k].ko[0]) bbox[1]=atom_data[k].ko[0];

			if (bbox[2]>atom_data[k].ko[1]) bbox[2]=atom_data[k].ko[1];
			if (bbox[3]<atom_data[k].ko[1]) bbox[3]=atom_data[k].ko[1];

			if (bbox[4]>atom_data[k].ko[2]) bbox[4]=atom_data[k].ko[2];
			if (bbox[5]<atom_data[k].ko[2]) bbox[5]=atom_data[k].ko[2];
		}

                fprintf(h_cnt,"% 6.3f % 6.3f % 6.3f",
			bbox[1]-bbox[0],
			bbox[3]-bbox[2],
			bbox[5]-bbox[4]);

				/*********** Radius of Gyration ******************/
		rgyr=0.0;
		child_pos=unit_buf[i].child_pos;
		for (j=0;j<unit_buf[i].atoms;j++) {
			double d1,d2,d3;
			k=child_buf[child_pos++];
			d1 = unit_buf[i].cen_geom[0] - atom_data[k].ko[0];
			d2 = unit_buf[i].cen_geom[1] - atom_data[k].ko[1];
			d3 = unit_buf[i].cen_geom[2] - atom_data[k].ko[2];
			rgyr += d1*d1 + d2*d2 + d3*d3;

		}
		rgyr=sqrt(rgyr/unit_buf[i].atoms);

                fprintf(h_cnt," % 6.3f\n",rgyr);
	}

        fprintf(h_cnt,"\n");

	fclose(h_cnt);
	return 0;
}

