/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*********************************************************************************
**										**
**	Module	: mkflood.c	by H. Grubmueller, M.Eichinger , Munich		**
**										**
*********************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#if defined(LINUX64)
#include <stdlib.h>
#endif

#include "../ego.h"
#include "../hardw.h"
#include "../util_lib.h"
#include "../wodecls.h"
#include "../egalloc.h"

#define	MAX_FNLEN	80
#define	MAX_NR_OF_STEPS	1000

/******************************************************************************/
/**************** functions in this module ************************************/
/******************************************************************************/

	int	get_atomtype_index(char	*newstr);
	int	get_atti_by_atomnr(int atom_nr);
	int	goto_psf_atomdata(FILE *h_psf,int *nr_of_atoms);
	double	get_atom_mass(int atom_type_index);
	int	get_atomtype_str(int atom_type_index,char *str);
	int	isit_a_hydrogen(double mass);

	
static	int	readin_psf_atomtypes(	char *psffn,int  *nr_atomtypes,
					char atom_typestr_list[][LEN_ATOM_TYPE+1],
					int  *nr_of_atoms);

static	int	read_basic_data(char *pdbfn,char *psffn);
static	int	free_atom_buffer(void);
static	int	free_all_buffers(int nr_of_atoms);

static int	init_cov_matrix(int nr_of_atoms);
static int	free_cov_matrix(int nr_of_atoms);

static void	sum_covariance(	int nr_of_atoms,
				double *xold1,double *yold1,double *zold1);
static void	calc_cov_matrix(int nr_of_atoms);
static void	diagonalize_cov_matrix(int nr_of_atoms, double dt);
static void	extrapolate_variances(int nr_of_atoms);
static void	calc_flooding_matrix(int nr_of_atoms,int nr_of_dofs);
static void	calc_flooding_matrix_extrapol(int nr_of_atoms,int nr_of_dofs, double dt);

static int	save_flooding_matrix(char *fname,int nr_of_atoms,int nr_flooding_atoms);
static int	create_projected_trajectory(char *fname,int nr_of_atoms,int nr_flooding_atoms, int nr_of_projections);
static int	save_flooding_fake(char *fname,int nr_of_atoms,int nr_flooding_atoms);
static int	save_ews_init(char *fname,int nr_of_atoms,int nr_of_steps);
static int	save_ews_step(char *fname,double *d,int nr_of_atoms,double dt);
static int	save_ews_finish(char *fname,int nr_of_atoms);

static void	tred2(double **a,int n,double *d,double *e);
static void	tqli(double *d,double *e,int n,double **z);
static void	tred2_ew(double **a,int n,double *d,double *e);
static void	tqli_ew(double *d,double *e,int n,double **z);
static void 	sort(int n, double *ra);
static void 	sort2(int n, double *ra, double *rb);

/*************** buffer for holding atom-type-index for each atom ************/
static	int		nr_of_atoms=0;
static	STRUCT_ATOMDATA	*atom_buf;
static	int		atom_buf_here=0;

/******************** buffer for holding atom-type-strings *******************/
static	int	nr_atomtypes=0;
static	char	atom_typestr_list[MXNO_ATOM_TYPES+1][LEN_ATOM_TYPE+1];
static	double	atom_typemass_list[MXNO_ATOM_TYPES+1];
static	int	startfn,endfn,minfn;
static	int	nr_of_steps;
static	int	startfni[MAX_NR_OF_STEPS+1];
static	int	endfni[MAX_NR_OF_STEPS+1];
static  double  ew_sum;
static	double  ew_step_sum;
static	double  ew_step_square_sum;
static	double  *ew_var_sum;
static	double  *ew_step_var_sum;
static  double  *ew_b;
static  double  *ew_a;

static	char	cur_egofile[MAX_STRING_SIZE];

static	double	*xko,*yko,*zko;

static	double	*covsum_i;
static	double	*flooding_cm;
static	double	*cov_eigenvalues;
static	double	**covsum_ij;
static	double	**cov_matrix;
static	double	**largest_cov_eigenvectors;
static	int	*use_this_atom;
static	int	covariance_ctr=0;
static	int	write_eigenvalues_only_flag=0;
static	int	nr_of_projections=0;

/******* this is the range for Xplor to detect Hydrogens *****/
#define H_MIN_LIMIT	0.5
#define H_MAX_LIMIT	3.5

/**************************************************************************************/
void   pex (int i)
{
	errout("Exit by pex(%d) in emachine.c !\n",i);
	exit(i);
	return;
}

int main(int argc,char *argv[])
{
	char	psffilen[MAX_STRING_SIZE];
	char	pdbfilen[MAX_STRING_SIZE];
	int	i,ii,j,k,res;
	int	nr_of_dofs=0,nr_flood_atoms=0;
	FILE	*ego_h;
	double  ex;
	double	coords[3];

	txtout("\n===================================================\n");
	txtout("mkflood - Make Covarianz-Matrix.\n");
	txtout("Version 1.0\n");
	txtout("===================================================\n");

	if (argc<6) {
		txtout("Fatal error: Illegal call of program.\n");
		txtout("mkflood <pdb-file> <psf-file> <start-num> <end-num> <min. #files> <#steps>\n");
		txtout("        <nr_of_dofs> [-ev] [-p <#proj>] ['<Selection-string>']\n\n");	
		txtout("        <pdb-file>: used for atom selection\n");
		txtout("        <psf-file>: used for atom selection\n");
		txtout("       <start-num>: number of first ego-file used\n");
		txtout("         <end-num>: number of last ego-file used\n");
		txtout("     <min. #files>: the first covariance matrix will be computed\n");
		txtout("                    from ego-files <start-num>...<start-num>+<min. #files>-1.\n");
		txtout("          <#steps>: number of covariance matrices to be computed, each\n");
		txtout("                    of the matrices is computed from a different number of\n");
		txtout("                    ego-files, these numbers are distributed logarithmically\n");
		txtout("                    in the range <min. #files>...<end-num>-<start-num>.\n");
		txtout("                    <#steps>=0 means that only one covariance matrix is\n");
		txtout("                    computed, and no extrapolation is performed. In that case\n");
		txtout("                    the value of <min. #files> is meaningless.\n");
		txtout("      <nr_of_dofs>: the number of degrees of freedom used for flooding.\n");
		txtout("               -ev: if set, write only eigenvalues into flooding-file\n");
		txtout("       -p <#proj> : if set, coords are projected on #proj largest eigenvectors\n");
		txtout("<Selection-string>: [+|-]{A|R}{<number|number1-number2|string}\n");
		txtout("                    + : include selected atoms/residues (default)\n");
		txtout("                    - : exclude selected atoms/residues\n");
		txtout("                    A : refer selection to atom numbers/names\n");
		txtout("                    R : refer selection to residue numbers/names\n");
		txtout("               number : select single atom/residue\n");
		txtout("      number1-number2 : select atom/residue range\n");
		txtout("               string : select atom/residue by string-match\n");
		txtout("                          wildcards: *  (string)\n");
		txtout("                                     %%  (single character)\n");
		txtout("                                     #  (string of digits)\n");
		txtout("                                     +  (single digit)\n");
		txtout("                    example: 'AC* AN* -RPRO': include all C and all N atoms\n");
		txtout("                                              except for those in prolines\n");
		txtout("If you want to make an empty flooding-matrix-file, which can only be used for\n");
		txtout("translatation/rotation correction, simply enter:\n");
		txtout("mkflood <pdb-file> <psf-file> 0 0 0 0 0 ['<Selection-string>']\n\n");	
		return -1;
	}
	
	strcpy(pdbfilen,argv[1]);
	strcpy(psffilen,argv[2]);

	startfn = endfn = -1;
	sscanf(argv[3],"%d",&startfn);
	sscanf(argv[4],"%d",&endfn);
	sscanf(argv[5],"%d",&minfn);
	sscanf(argv[6],"%d",&nr_of_steps);
	sscanf(argv[7],"%d",&nr_of_dofs);

	if (	(startfn<0) || (endfn<startfn) || 
		(((minfn >= endfn-startfn) || (minfn < 1))&&(nr_of_steps!=0))) {
		errout("Illegal file-numbers !\n");
		return -1;
	}
	if (nr_of_steps < 0) {
		errout("Number of extrapolation steps must be >= 0.\n");
		return -1;
	}
	if (nr_of_steps >= MAX_NR_OF_STEPS-1) {
		errout("Maximum number of extrapolation steps exceeded.\n");
		return -1;
	}
	txtout("PDB-file       : %s\n",pdbfilen);
	txtout("PSF-file       : %s\n",psffilen);
	txtout("Start-file     : %d.ego\n",startfn);
	txtout("End-file       : %d.ego\n",endfn);
	txtout("Nr of dofs     : %d\n\n",nr_of_dofs);

	txtout("===================================================\n\n");

	/*************************************************************************/
	if (readin_psf_atomtypes(psffilen,&nr_atomtypes,
				 atom_typestr_list,
				 &nr_of_atoms)==-1) 		goto error_occured;
	
	if (read_basic_data(pdbfilen,psffilen)==-1)		goto error_occured;
	use_this_atom =  alloc_ints(nr_of_atoms);

	txtout("Structure-data read.\n");
	for (i=0;i<nr_of_atoms;i++) use_this_atom[i]=0;
	write_eigenvalues_only_flag=0;
	nr_of_projections=0;
       	if (argc>8) {
	  for (i=8; i<argc; i++) {
	    if (strcmp(argv[i],"-ev")==0) {
	      write_eigenvalues_only_flag=1;
              i++;
	    }
	    if (strcmp(argv[i],"-p")==0) {
	      i++;
	      sscanf(argv[i],"%d",&nr_of_projections);
	    }
	    else {
	      if (set_used_atoms(argv[i],nr_of_atoms,atom_buf,use_this_atom)!=0) {
		  errout("The selection string specifying which atoms to write out\n");
		  errout("during analysis is syntactically incorrect.\n");
	      }	
	    }	
	  }
	}
	nr_flood_atoms = 0;
	for (i=0;i<nr_of_atoms;i++) {
	  if (use_this_atom[i]) nr_flood_atoms++;
	}

	if ((startfn==0)&&(endfn==0)&&(minfn==0)&&(nr_of_steps==0)&&(nr_of_dofs==0)){
		txtout("Writing empty flooding-matrix to file <flooding.lis>\n");
		txtout("This file can only be used for translatation/rotation correction !\n");	
		save_flooding_fake("flooding.lis",nr_of_atoms,nr_flood_atoms);
		myfree(use_this_atom);
		free_atom_buffer();
		txtout("\n===================================================\n");
		txtout("mkflood finished now. No fatal errors.\n");
		errout("===================================================\n");
		return 0;	
	}

	txtout("%d atoms will be flooded !\n",nr_flood_atoms);
	if (nr_of_steps <= 0) {
	  nr_of_steps=0;
	  startfni[0]=startfn;
	  endfni[0]=endfn;
	  }
	else {
	  startfni[0]=startfn;
	  for (i=0; i<nr_of_steps; i++) {
	    ex=log(1.0*(endfn-startfn)/(1.0*minfn));
	    startfni[i+1]=startfn+floor(0.5+(endfn-startfn)*exp(-(nr_of_steps-i*1.0)/(1.0*nr_of_steps)*ex));
	    endfni[i]=startfni[i+1]-1;
	    }
          endfni[nr_of_steps]=endfn;
	  }

	txtout("\nExtrapolation intervals:\n");

	for (i=0; i<=nr_of_steps; i++) {  /*xxx*/
	  txtout("%d %d %d\n",i,startfni[i],endfni[i]);  /*xxx*/  
	  }  /*xxx*/
	xko =  alloc_doubles(nr_flood_atoms);
	yko =  alloc_doubles(nr_flood_atoms);
	zko =  alloc_doubles(nr_flood_atoms);
	
	init_cov_matrix(nr_flood_atoms);

	errout("All allocated.\n");
	save_ews_init("ew.lis",nr_flood_atoms,nr_of_steps);

	/****************** here starts the loop over all files *****************/

	for (ii=0;ii<=nr_of_steps;ii++) {
		for (i=startfni[ii];i<=endfni[ii];i++) {
			generate_egofname(cur_egofile,"",i);
			if ((ego_h=fopen(cur_egofile,"r"))==NULL) {
				if (i==startfni[ii]) 
					txtout("Found no new ego-output name. Trying old output-name !\n");
				sprintf(cur_egofile,"%d.ego",i);
				if ((ego_h=fopen(cur_egofile,"r"))==NULL) {
					errout("Can not open ego-file <%s>. Skipping it !\n",cur_egofile);
					continue;
				}
			}	
			txtout("Reading data from file <%s> ...",cur_egofile);
			skip_ego_header(ego_h);
			for (j=0,k=0;j<nr_of_atoms;j++) {
				res = read_ego_coords(ego_h,coords);
				if (res==-2) goto error_occured;
				if (res==-1) {
					errout("Fatal error: End of file after %d read coords !",j);
					goto error_occured;
				}
				if (use_this_atom[j]) {
					xko[k] = coords[0];
					yko[k] = coords[1];
					zko[k++] = coords[2];
				}
			}
			fclose(ego_h);
			txtout("Done ! Summing now ...");
			sum_covariance(nr_flood_atoms,xko,yko,zko);
			errout("Done !\n");
		}

		txtout("Calculating the covariance-matrix...\n");
		calc_cov_matrix(nr_flood_atoms);
		if (ii!=nr_of_steps) {
			txtout("Diagonalizing the covariance-matrix...\n");
			diagonalize_cov_matrix(nr_flood_atoms, 1.0/((double)(endfni[ii]+1-startfn)));
		}
	}

	if (nr_of_steps >0) {
	  calc_flooding_matrix_extrapol(nr_flood_atoms,nr_of_dofs,1.0/((double)(endfn+1-startfn)));
	  }
	else {   /* no extrapolation needed */
	  txtout("Calculating the flooding-matrix....\n");
	  calc_flooding_matrix(nr_flood_atoms,nr_of_dofs);
	  }

	txtout("Writing flooding-matrix to file <flooding.lis>.\n");
	save_flooding_matrix("flooding.lis",nr_of_atoms,nr_flood_atoms);

	if (nr_of_projections != 0) {
	  if (nr_of_steps >0) {
	    errout("\n=======================================================================\n");
	    errout("ERROR: Projecting is only allowed if no extrapolation is used (#steps 0).\n");
	    errout("=========================================================================\n");
	    return -1;
	    }
	  create_projected_trajectory("proj.lis",nr_of_atoms,nr_flood_atoms,nr_of_projections);
	  }
	free_all_buffers(nr_flood_atoms);

	txtout("\n===================================================\n");
	txtout("mkflood finished now. No fatal errors.\n");
	txtout("===================================================\n");

	if (nr_flood_atoms*3 >= (endfni[0]-startfni[0])) {
	  txtout("*** WARNING: Matrix dimension >= number of data. ***\n");
	  txtout("***          Some eigenvalues will be zero.      ***\n\n");
	  }
	return 0;

error_occured:
	free_all_buffers(nr_flood_atoms);

	errout("\n=================================================================\n");
	errout("mkflood stopped with a fatal error.\n");
	errout("Check your Xplor-input-files or the ego-data-files for this error.\n");
	errout("==================================================================\n");
	return -1;
}

/**************************************************************************************/

int	get_atomtype_index(char	*atomtypstr)
{
	int i;

	if (nr_atomtypes<=0)	return -1;		/* list is empty */	

	for (i=0;i<LEN_ATOM_TYPE;i++) {			/* add up spaces */
		if (atomtypstr[i]==0) {
			atomtypstr[i]=' ';
			atomtypstr[i+1]=0;
		}
	}

	for (i=0;i<nr_atomtypes;i++) {			/* search for matching string */
		if (strcmp(atomtypstr,atom_typestr_list[i])==0)	return i;
	}
	return -1;					/* str not in list */
}

/************ use by other moduls to get info about atom-type-index ********************/
/*****        /-------- read: ATom-Type-Index ! ****************************************/
int	get_atti_by_atomnr(int atom_nr)
{
	return atom_buf[atom_nr].atomtype;
}

/*************** get the mass of atom by atom-type-index *******************************/
double	get_atom_mass(int att_index)
{
	if ((att_index<0)||(att_index>=nr_atomtypes)) {
		errout("Fatal error: Illegal atomtype-index in get_atom_mass.\n");
		return 0.0;
	}
	return (double) atom_typemass_list[att_index];	
}

/*************** get the atom-type-string by atom-type-index ***************************/
int	get_atomtype_str(int att_index,char *str)
{

	if ((att_index<0)||(att_index>=nr_atomtypes)) {
		errout("Fatal error: Illegal atomtype-index in get_atomtype_str.\n");
		return 0.0;
	}

	strcpy(str,atom_typestr_list[att_index]);
	return 0;
}

int	isit_a_hydrogen(double mass)
{
	if (	(mass>=H_MIN_LIMIT) &&
		(mass<=H_MAX_LIMIT) ) {
		return 1;		/* YES */
	}
	return 0;			/* NO */
}


/***************************************************************************************/

static	
int	free_all_buffers(int nr_of_atoms)
{
	free(xko);
	free(yko);
	free(zko);
	free_cov_matrix(nr_of_atoms);
	free_atom_buffer();

	return 0;
}

/***************************************************************************************/
/**************************************************************************************/

static
int	readin_psf_atomtypes(	char *psffn,
				int  *nr_atomtypes,
				char atom_typestr_list[][LEN_ATOM_TYPE+1],
				int  *nr_of_atoms)
{
	FILE	*h_psf;
	int	res=0;
	int	readin_atoms=0;
	int	atom_index=0,ati;
	double	atom_charge=0.0,atom_mass=0.0;	/* not analyzed here */
	static	char	tmpstr[MAX_STRING_SIZE];
/**** next variables are not used *****/
	int	psf_residnr;
static	char	psf_residstr[LEN_ATOM_TYPE+1];
static	char	psf_segidstr[LEN_ATOM_TYPE+1];
static	char	psf_atomstr [LEN_ATOM_TYPE+1];

	*nr_of_atoms=0;
	*nr_atomtypes=0;

	if ((h_psf=fopen(psffn,"r"))==NULL) {
		errout("Can not open specified psf-file.\n");
		return -1;
	}

	if (goto_psf_atomdata(h_psf,nr_of_atoms)==-1) return -1;	
	
	txtout("Number of atoms in PSF-file: %d\n",*nr_of_atoms);
	
	do { /**** for each atom in psf-file *******/
		res=read_next_psfatom(h_psf,	&atom_index,
						tmpstr,
						&atom_charge,
						&atom_mass,
						psf_atomstr,
						psf_residstr,
						&psf_residnr,
						psf_segidstr);

		if (res==0) {
					/* try to get the atomtype index */
			ati=get_atomtype_index(tmpstr);
					
			if (ati==-1) {	/* -1 means,... not in list, so add it  */

				strcpy(atom_typestr_list[*nr_atomtypes],tmpstr);
				atom_typemass_list[*nr_atomtypes]=atom_mass;

				if (++(*nr_atomtypes)>MXNO_ATOM_TYPES) {
					errout("Too many atomtypes in PSF-file !\n");
					errout("Actual limit is %d (MXNO_ATOM_TYPES)\n",
						MXNO_ATOM_TYPES);
					fclose(h_psf);
					return -1;
				}
			}
			readin_atoms++;	
		} else {
			errout("Error detected after %d read atoms in PSF-file.\n",
				readin_atoms);
			fclose(h_psf);
			return -1;
		}
	} while(readin_atoms<*nr_of_atoms);

	txtout("Number of atom-types       : %d\n",*nr_atomtypes);
	fclose(h_psf);
	return 0;
}

/***************************************************************************************/

int	goto_psf_atomdata(FILE *h_psf,int *nr_of_atoms)
{
	static	char	lstr[MAX_STRING_SIZE];
	static	char	tmpstr[MAX_STRING_SIZE];
	int 	res;

	while (read_string(h_psf,lstr)!=-1) {
		strup(lstr);
		if (sscanf(lstr,"%d %s",&res,tmpstr)==2) {
			if (strcmp(tmpstr,"!NATOM")==0) {
				*nr_of_atoms=res;
				return 0;
			}
		}	
	}
	errout("Fatal error: No atom-data-section (!NATOM) in PSF-file.\n");
	return -1;
}

static
int	read_basic_data(char *pdbfn,char *psffn)
{
	FILE	*h_pdb,*h_psf;
	int	pdbres=0,psfres=0;
	int	nr_of_atoms=0,nr_read_atoms=0;
	int	pdb_atom_index=0,psf_atom_index,atom_type_index=0;
	int	psf_residnr,pdb_residnr;
	double	coords[3],vel[3];
	double	beta=0.0,harmon=0.0,atom_charge=0.0;
	double	atom_mass=0.0;
static	char	atomtypestr[MAX_STRING_SIZE];
static	char	psf_segidstr[LEN_ATOM_TYPE+1],pdb_segidstr[LEN_ATOM_TYPE+1];
static	char	psf_residstr[LEN_ATOM_TYPE+1],pdb_residstr[LEN_ATOM_TYPE+1];
static	char	psf_atomstr[LEN_ATOM_TYPE+1],pdb_atomstr[LEN_ATOM_TYPE+1];


	if ((h_pdb=fopen(pdbfn,"r"))==NULL) {
		errout("Can not open specified pdb-file.\n");
		return -1;
	}
	
	if ((h_psf=fopen(psffn,"r"))==NULL) {
		errout("Can not open specified psf-file.\n");
		fclose(h_pdb);
		return -1;
	}
	if (goto_psf_atomdata(h_psf,&nr_of_atoms)==-1) {
		fclose(h_psf);
		fclose(h_pdb);
		return -1;	
	}

	atom_buf=alloc_atomdata(nr_of_atoms+5);
	if (atom_buf==NULL) {
		errout("Fatal error: Can not allocate space for atom-data.\n");
		return -1;
	}	
	atom_buf_here=1;

	vel[0]=vel[1]=vel[2]=0.0;

	do {	/***** for each atom ***********************************/

		pdbres=read_next_pdbdata(h_pdb,	&pdb_atom_index,
						coords,
						&beta,
						&harmon,
						pdb_atomstr,
						pdb_residstr,
						&pdb_residnr,
						pdb_segidstr);
		
		psfres=read_next_psfatom(h_psf,	&psf_atom_index,
						atomtypestr,
						&atom_charge,
						&atom_mass,
						psf_atomstr,
						psf_residstr,
						&psf_residnr,
						psf_segidstr);

		/****** many, many checks ***************************************/
		if ((pdbres==-2)||(psfres!=0)) {
			errout("Error detected after %d read atoms.\n",
				nr_read_atoms);
			goto error_in_wcoord;
		}
		
		if (psf_atom_index!=pdb_atom_index) {
			errout("PSF and PDB-data out of sync after %d read atoms.\n",
				nr_read_atoms);
			goto error_in_wcoord;
		}

		if (	(strcmp(pdb_atomstr, psf_atomstr )!=0) ||
			(strcmp(pdb_residstr,psf_residstr)!=0) ||
			(strcmp(pdb_segidstr,psf_segidstr)!=0) ||
			(pdb_residnr != psf_residnr) ) {

			errout("(<%s> <%s>) (<%s> <%s>) (<%s> <%s>) (%d %d)\n",
				pdb_atomstr,psf_atomstr,
				pdb_residstr,psf_residstr,
				pdb_segidstr,psf_segidstr,
				pdb_residnr,psf_residnr);
			errout("Fatal error: PDB and PSF-data are inconsistent after %d read atoms.\n",
				nr_read_atoms);
			goto error_in_wcoord;
		}

		if (pdbres!=0) continue;

		atom_type_index=get_atomtype_index(atomtypestr);
		if (atom_type_index==-1) {	
			errout("Unknown atom-type <%s> after %d read atoms in PDB-file.\n",
				atomtypestr,nr_read_atoms);
			goto error_in_wcoord;
		}

		if ((pdb_atom_index<1)||(pdb_atom_index>nr_of_atoms)) {
			errout("Illegal atom-number %d in PDB-file.\n",
				pdb_atom_index);
			goto error_in_wcoord;
		}

		/********** keep data in mem for use in other modules ****************/

		atom_buf[pdb_atom_index-1].atomnr=pdb_atom_index-1;

		atom_buf[pdb_atom_index-1].atomtype=atom_type_index; 
		strcpy(atom_buf[pdb_atom_index-1].atomtypestr,atomtypestr);

		atom_buf[pdb_atom_index-1].residnr=psf_residnr;
		strcpy(atom_buf[pdb_atom_index-1].residstr,psf_residstr);
		
		strcpy(atom_buf[pdb_atom_index-1].atomstr,psf_atomstr);
		strcpy(atom_buf[pdb_atom_index-1].segidstr,psf_segidstr);
		atom_buf[pdb_atom_index-1].charge=atom_charge;
		
		nr_read_atoms++;

	} while (nr_read_atoms<nr_of_atoms); /***** end: for each atom ****/

	fclose(h_psf);
	fclose(h_pdb);

	txtout("Read atoms in PDB-file     : %d\n",nr_read_atoms);
	
	return 0;

error_in_wcoord:
	fclose(h_psf);
	fclose(h_pdb);

	return -1;
}

static
int	free_atom_buffer(void)
{
	if (atom_buf_here) {
		free(atom_buf);
		atom_buf_here=0;
	}
	return 0;
}


static int	init_cov_matrix(int nr_of_atoms)
{
	int i,k;
	
	covsum_i 		=  alloc_doubles(3*nr_of_atoms);
	flooding_cm 		=  alloc_doubles(3*nr_of_atoms);
	cov_eigenvalues 	=  alloc_doubles(3*nr_of_atoms);
	covsum_ij  		=(double**) alloc_ptrs(3*nr_of_atoms);
	cov_matrix  		=(double**) alloc_ptrs(3*nr_of_atoms);
	largest_cov_eigenvectors  	=(double**) alloc_ptrs(3*nr_of_atoms);

	for (i = 0; i < 3*nr_of_atoms; i++) {
		covsum_i[i]		= 0.0;
		flooding_cm[i]= 0.0;		/*** not really needed ***/
		covsum_ij[i]  	=  alloc_doubles(3*nr_of_atoms);
		if (covsum_ij[i]==NULL) {
			errout("Can not allocate covarince-matrix\n");
			return -1;
		}
		for (k = 0; k < 3*nr_of_atoms; k++) {
			covsum_ij[i][k] =0.0;
		}
		cov_matrix[i]  	=  alloc_doubles(3*nr_of_atoms);
		if (cov_matrix[i]==NULL) {
			errout("Can not allocate covarince-matrix\n");
			return -1;
		}
		for (k = 0; k < 3*nr_of_atoms; k++) {
			cov_matrix[i][k] =0.0;
		}
		largest_cov_eigenvectors[i]  	=  alloc_doubles(3*nr_of_atoms);
		if (largest_cov_eigenvectors[i]==NULL) {
			errout("Can not allocate covarince-eigenvectors\n");
			return -1;
		}
		for (k = 0; k < 3*nr_of_atoms; k++) {
			largest_cov_eigenvectors[i][k] =0.0;
		}
	}
	covariance_ctr = 0;

	ew_sum = 0.0;
	ew_step_sum = 0.0;
	ew_step_square_sum = 0.0;
	ew_var_sum 		=  alloc_doubles(3*nr_of_atoms);
	ew_step_var_sum		=  alloc_doubles(3*nr_of_atoms);
	ew_a			=  alloc_doubles(3*nr_of_atoms);
	ew_b			=  alloc_doubles(3*nr_of_atoms);
	for (i = 0; i < 3*nr_of_atoms; i++) {
	  ew_var_sum[i] = 0.0;
	  ew_step_var_sum[i] = 0.0;
	  }
	return 0;
}

static int	free_cov_matrix(int nr_of_atoms)
{
	int	i;

	free(covsum_i);
	free(flooding_cm);
	free(cov_eigenvalues);
	for (i = 0; i < 3*nr_of_atoms; i++) {
		free(covsum_ij[i]);
		free(cov_matrix[i]);
		free(largest_cov_eigenvectors[i]);
	}
	free(covsum_ij);
	free(largest_cov_eigenvectors);
	free(ew_var_sum);
	free(ew_step_var_sum);
	free(ew_b);
	free(ew_a);
	return 0;
}

/***********************************************************************/

static void	sum_covariance(	int nr_of_atoms,
				double *xold1,double *yold1,double *zold1)
{
	int	i,j;
	int	offsx,offsy;

	for (i=0,offsx=0;i<nr_of_atoms;i++,offsx+=3) {
		covsum_i[offsx  ] += xold1[i];
		covsum_i[offsx+1] += yold1[i];
		covsum_i[offsx+2] += zold1[i];
	}

	for (i=0,offsx=0;i<nr_of_atoms;i++,offsx+=3) {
		for (j=0,offsy=0;j<nr_of_atoms;j++,offsy+=3) {
			covsum_ij[offsx  ][offsy  ] += xold1[i]*xold1[j];
			covsum_ij[offsx  ][offsy+1] += xold1[i]*yold1[j];
			covsum_ij[offsx  ][offsy+2] += xold1[i]*zold1[j];

			covsum_ij[offsx+1][offsy  ] += yold1[i]*xold1[j];
			covsum_ij[offsx+1][offsy+1] += yold1[i]*yold1[j];
			covsum_ij[offsx+1][offsy+2] += yold1[i]*zold1[j];

			covsum_ij[offsx+2][offsy  ] += zold1[i]*xold1[j];
			covsum_ij[offsx+2][offsy+1] += zold1[i]*yold1[j];
			covsum_ij[offsx+2][offsy+2] += zold1[i]*zold1[j];
		}
	}
	covariance_ctr++;
	return ;
}

static void	calc_cov_matrix(int nr_of_atoms)
{
	int i,j;
	double covctr=(double)covariance_ctr;

	for (i = 0; i < 3*nr_of_atoms; i++) {
		for (j = 0; j < 3*nr_of_atoms; j++) {
			cov_matrix[i][j]=(covsum_ij[i][j]-covsum_i[i]*covsum_i[j]/covctr)/
						covctr;
		}
	}
	for (i = 0; i < 3*nr_of_atoms; i++) {
		flooding_cm[i] = covsum_i[i]/covctr;
	}
	return;
}

/*********************************************************/

static void	diagonalize_cov_matrix(int nr_of_atoms, double dt)
{
	int i,j;
	double sigma2;
	double **c,*d,*e;

	c = dmatrix(1,3*nr_of_atoms,1,3*nr_of_atoms);
	d = dvector(1,3*nr_of_atoms);
	e = dvector(1,3*nr_of_atoms);

	for (i = 1; i <= 3*nr_of_atoms; i++) {
		for (j = 1; j <= 3*nr_of_atoms; j++) {
			c[i][j]=cov_matrix[i-1][j-1];
		}
	}

	txtout("tred2...\n");
	tred2(c,3*nr_of_atoms,d,e);

	txtout("tqli...\n");
	tqli(d,e,3*nr_of_atoms,c);

	txtout("\n");
        sort(3*nr_of_atoms,d);

	/* prepare cumulative quantities for linear fit: */
	sigma2=dt;  /* sigma decreases with the sqrt of simulation time */
        ew_sum+=1.0/sigma2;
	ew_step_sum += dt/sigma2;   /* here reciprocal time! */
	ew_step_square_sum += dt*dt/sigma2;
	for (i = 1; i <= 3*nr_of_atoms; i++) {
	  ew_var_sum[i-1] += d[i]/sigma2;
	  ew_step_var_sum[i-1] += dt*d[i]/sigma2;
	  }

	save_ews_step("ew.lis",d,nr_of_atoms,dt);
	free_dvector(d,1,3*nr_of_atoms);
	free_dvector(e,1,3*nr_of_atoms);
	free_dmatrix(c,1,3*nr_of_atoms,1,3*nr_of_atoms);   /*xxx*/

	return;
}


static void	extrapolate_variances(int nr_of_atoms)
{
	int i;

	double s,sx,sy,sxx,sxy,stt;

	for (i = 0; i < 3*nr_of_atoms; i++) {
	  s=ew_sum;
	  sx = ew_step_sum;
	  sy = ew_var_sum[i];
	  sxx = ew_step_square_sum;
	  sxy = ew_step_var_sum[i];
	  stt = sxx - sx*sx/s;
          ew_b[i] = (sxy - sx*sy/s) / stt;
          ew_a[i] = (sy - ew_b[i]*sx)/s;
	  if (ew_a[i]<0.0) {
	    errout("Error: A negative eigenvalue was computed during extrapolation.\n");
	    errout("       You should try to increase <min. #files>\n");
	    errout("       or dispense with extrapolation, i.e., <#steps> := 0\n");
	    exit(-1);
	    }
	  }
	save_ews_finish("ew.lis",nr_of_atoms);

	return;
}


static int	save_ews_init(char *fname,int nr_of_atoms,int nr_of_steps)
{
	FILE	*fp;

	if ((fp = fopen(fname,"w")) == NULL) {
		errout("Can not open eigenvalue-file <%s> !",fname);
		return -1;
	}
	fprintf(fp,"%d\n",nr_of_atoms);		/*** nr of atoms ***/
	fprintf(fp,"%d\n",nr_of_steps);		/*** nr of extrapolaion steps ***/
	fclose(fp);
	return 0;
}

static int	save_ews_step(char *fname,double *d,int nr_of_atoms,double dt)
{
	int	i;
	FILE	*fp;

	if ((fp = fopen(fname,"a")) == NULL) {
		errout("Can not append to eigenvalue-file <%s> !",fname);
		return -1;
	}
	fprintf(fp,"%f\n",dt);
	for (i = 1; i <= 3*nr_of_atoms; i++) {
	  fprintf(fp,"%f\n",d[i]);
	  }
	fclose(fp);
	return 0;
}

static int	save_ews_finish(char *fname,int nr_of_atoms)
{
	int	i;
	FILE	*fp;

	if ((fp = fopen(fname,"a")) == NULL) {
		errout("Can not append to eigenvalue-file <%s> !",fname);
		return -1;
	}
	for (i = 0; i < 3*nr_of_atoms; i++) {
	  fprintf(fp,"%f\n",ew_a[i]);
	}
	for (i = 0; i < 3*nr_of_atoms; i++) {
	  fprintf(fp,"%f\n",ew_b[i]);
	}
	fclose(fp);
	return 0;
}

/*****************************************/
/*********************************************************/

static void	calc_flooding_matrix(int nr_of_atoms,int nr_of_flooding_dofs)
{
	int i,itmp,ip,j,k,jmin=0;
	double dmin,sum,tmp;
	double **c,*d,*dtmp,*e;

	c = dmatrix(1,3*nr_of_atoms,1,3*nr_of_atoms);
	d = dvector(1,3*nr_of_atoms);
	dtmp = dvector(1,3*nr_of_atoms);
	e = dvector(1,3*nr_of_atoms);

	for (i = 1; i <= 3*nr_of_atoms; i++) {
		for (j = 1; j <= 3*nr_of_atoms; j++) {
			c[i][j]=cov_matrix[i-1][j-1];
		}
	}

	txtout("tred2...\n");
	if ((write_eigenvalues_only_flag==0) || (nr_of_projections>0)) {
	  tred2(c,3*nr_of_atoms,d,e);
	  }
        else {
	  tred2_ew(c,3*nr_of_atoms,d,e);
	  }

	txtout("tqli...\n");
	if ((write_eigenvalues_only_flag==0) || (nr_of_projections>0)) {
	  tqli(d,e,3*nr_of_atoms,c);
	  }
        else {
	  tqli_ew(d,e,3*nr_of_atoms,c);
	  }

	txtout("Scaling eigenvalues...\n");
	for (i = 1; i <= 3*nr_of_atoms; i++) {
	  cov_eigenvalues[i-1]=d[i];
          dtmp[i]=d[i];
	  }
	if (nr_of_projections>0) {
		itmp=0;
                
		for (ip = 0; ip <= nr_of_projections; ip++) {
			tmp=-1.0;
			for (i = 1; i <= 3*nr_of_atoms; i++) {
				if (dtmp[i]>tmp) {
			  		tmp=dtmp[i];
			  		itmp=i;
				}
			}
                        dtmp[itmp]=-1.0e20;
		  	for (j = 1; j <= 3*nr_of_atoms; j++) {
			  	largest_cov_eigenvectors[ip][j-1] = c[j][itmp]; /* last index is ev-nr */
		  	}
		}
  	}
	if ((write_eigenvalues_only_flag==0) || (nr_of_projections>0)) {
	  for (i = 1; i <= 3*nr_of_atoms - nr_of_flooding_dofs; i++) {
		  dmin=1.0e19;
		  for (j = 1; j <= 3*nr_of_atoms; j++) {
			  if (d[j] <= dmin) {
				  dmin=d[j];
				  jmin=j;
			  }
	  	  }
		  d[jmin]=1.0e20;
	  }
	
	  txtout("Matrix multiply part 1...\n");
	  /* C^T * F * C:  (f ist symmetrisch): */
	  for (i = 1; i <= 3*nr_of_atoms; i++) {
		  for (j = 1; j <= 3*nr_of_atoms; j++) {
			  cov_matrix[i-1][j-1] = c[i][j]/d[j];
		  }
  	  }

	  txtout("Matrix multiply part 2...\n");
	  for (i = 1; i <= 3*nr_of_atoms; i++) {
		  txtout("*");
		  for (j = 1; j <= 3*nr_of_atoms; j++) {
			  sum=0.0;
			  for (k = 1; k <= 3*nr_of_atoms; k++) {
				  sum += cov_matrix[i-1][k-1]* c[j][k];   /*new_flooding_matrix:=G*C^T*/
			  }
			  d[j]=sum;
		  }
		  memcpy(cov_matrix[i-1],d+1,3*nr_of_atoms*sizeof(double));
	  }
	}
	txtout("\n");
	free_dvector(d,1,3*nr_of_atoms);
	free_dvector(dtmp,1,3*nr_of_atoms);
	free_dvector(e,1,3*nr_of_atoms);
	free_dmatrix(c,1,3*nr_of_atoms,1,3*nr_of_atoms);
	return;
}


static void	calc_flooding_matrix_extrapol(int nr_of_atoms,int nr_of_flooding_dofs, double dt)
{
	int i,j,k,jmin=0;
	double dmin,sum,sigma2;
	double **c,*d,*e,*ix;

	c = dmatrix(1,3*nr_of_atoms,1,3*nr_of_atoms);
	d = dvector(1,3*nr_of_atoms);
	e = dvector(1,3*nr_of_atoms);
	ix = dvector(1,3*nr_of_atoms);

	for (i = 1; i <= 3*nr_of_atoms; i++) {
		for (j = 1; j <= 3*nr_of_atoms; j++) {
			c[i][j]=cov_matrix[i-1][j-1];
		}
	}

	txtout("Diagonalizing the covariance-matrix...\n");

	txtout("tred2...\n");
	tred2(c,3*nr_of_atoms,d,e);

	txtout("tqli...\n");
	tqli(d,e,3*nr_of_atoms,c);

	txtout("\n");
	/* this time we also prepare an index table so that we can sort back the d's: */
	for (i = 1; i <= 3*nr_of_atoms; i++) ix[i]=(double)i;
        sort2(3*nr_of_atoms,d,ix);

	/* prepare cumulative quantities for linear fit: */
	sigma2=dt;  /* sigma decreases with the sqrt of simulation time */
        ew_sum+=1.0/sigma2;
	ew_step_sum += dt/sigma2;   /* here reciprocal time! */
	ew_step_square_sum += dt*dt/sigma2;
	for (i = 1; i <= 3*nr_of_atoms; i++) {
	  ew_var_sum[i-1] += d[i]/sigma2;
	  ew_step_var_sum[i-1] += dt*d[i]/sigma2;
	  }

	save_ews_step("ew.lis",d,nr_of_atoms,dt);

	extrapolate_variances(nr_of_atoms);

	txtout("Scaling eigenvalues...\n");
	for (i = 1; i <= 3*nr_of_atoms; i++) {
	  cov_eigenvalues[i-1]=ew_a[(int)ix[i]];
	  d[i]=ew_a[(int)ix[i]];
	  }
	for (i = 1; i <= 3*nr_of_atoms - nr_of_flooding_dofs; i++) {
		dmin=1.0e19;
		for (j = 1; j <= 3*nr_of_atoms; j++) {
			if (d[j] <= dmin) {
				dmin=d[j];
				jmin=j;
			}
		}
		d[jmin]=1.0e20;
	}
	
	txtout("Matrix multiply part 1...\n");
	/* C^T * F * C:  (f ist symmetrisch): */
	for (i = 1; i <= 3*nr_of_atoms; i++) {
		for (j = 1; j <= 3*nr_of_atoms; j++) {
			cov_matrix[i-1][j-1] = c[i][j]/d[j];
		}
	}

	txtout("Matrix multiply part 2...\n");
	for (i = 1; i <= 3*nr_of_atoms; i++) {
		txtout("*");
		for (j = 1; j <= 3*nr_of_atoms; j++) {
			sum=0.0;
			for (k = 1; k <= 3*nr_of_atoms; k++) {
				sum += cov_matrix[i-1][k-1]* c[j][k];   /*new_flooding_matrix:=G*C^T*/
			}
			d[j]=sum;
		}
		memcpy(cov_matrix[i-1],d+1,3*nr_of_atoms*sizeof(double));
	}
	txtout("\n");
	free_dvector(d,1,3*nr_of_atoms);
	free_dvector(e,1,3*nr_of_atoms);
	free_dvector(ix,1,3*nr_of_atoms);
	free_dmatrix(c,1,3*nr_of_atoms,1,3*nr_of_atoms);
	return;
}

static int	save_flooding_matrix(char *fname,int nr_of_atoms,int nr_flooding_atoms)
{
	int	i,j;
	FILE	*fp;

	if ((fp = fopen(fname,"w")) == NULL) {
		errout("Can not open flood-file <%s> !",fname);
		return -1;
	}
	if (write_eigenvalues_only_flag==0) {
	  fprintf(fp,"[HEADER]\n");
	  fprintf(fp,"%d\n",1);			/*** nr of gaussians ***/
	  fprintf(fp,"%d\n",nr_of_atoms);		/*** nr of atoms ***/
	  fprintf(fp,"%d\n",nr_flooding_atoms);
  /*
	  fprintf(fp,"%d\n",nr_of_flooding_dofs);
  */ 
	  fprintf(fp,"[USINGARRAY]\n");
	  for (i=0;i<nr_of_atoms;i++) {
		  fprintf(fp,"%d\n",use_this_atom[i]);
	  }
  
	  fprintf(fp,"[CENTER_OF_MASS]\n");

	  for (i = 0; i < 3*nr_flooding_atoms; i++) {
		  fprintf(fp,"%f\n",flooding_cm[i]);
	  }
  	
	  fprintf(fp,"[FLOODING_MATRIX]\n");
	  for (i = 0; i < 3*nr_flooding_atoms; i++) {
		  for (j = 0; j < 3*nr_flooding_atoms; j++) {
			  fprintf(fp,"%f\n",cov_matrix[i][j]);
		  }
	  }
        }

	fprintf(fp,"[COVARIANCE EIGENVALUES]\n");
	for (i = 0; i < 3*nr_flooding_atoms; i++) {
		fprintf(fp,"%f\n",cov_eigenvalues[i]);
	}
	
	fclose(fp);
	return 0;
}


static int	create_projected_trajectory(char *fname,int nr_of_atoms,int nr_flood_atoms, int nr_of_projections)
{
	int	i,jj,j,k,res;
	double  *crd_proj;
	double	coords[3];
	FILE	*fp, *ego_h;


	if ((fp = fopen(fname,"w")) == NULL) {
		errout("Can not open projections-file <%s> !",fname);
		return -1;
	}

	/****************** here starts the loop over all files *****************/

	crd_proj =  alloc_doubles(nr_of_projections);

	for (i=startfn;i<=endfn;i++) {
		for (jj=0; jj<nr_of_projections; jj++) crd_proj[jj]=0.0;
		generate_egofname(cur_egofile,"",i);
		if ((ego_h=fopen(cur_egofile,"r"))==NULL) {
			errout("Can not open ego-file <%s>. Skipping it !\n",cur_egofile);
			continue;
			}	
		txtout("Reading data from file <%s> ...",cur_egofile);
		skip_ego_header(ego_h);
		for (j=0,k=0;j<nr_of_atoms;j++) {
			res = read_ego_coords(ego_h,coords);
			if (res==-2) goto error_occured;
			if (res==-1) {
				errout("Fatal error: End of file after %d read coords !",j);
				goto error_occured;
			}
			if (use_this_atom[j]) {
				for (jj=0; jj<nr_of_projections; jj++) {
					crd_proj[jj] += ((coords[0]-flooding_cm[k])*largest_cov_eigenvectors[jj][k])
					     	       +((coords[1]-flooding_cm[k+1])*largest_cov_eigenvectors[jj][k+1])
					     	       +((coords[2]-flooding_cm[k+2])*largest_cov_eigenvectors[jj][k+2]);
				}
				k += 3;
			}
		}
		fclose(ego_h);
	  	for (jj = 0; jj < nr_of_projections; jj++) {
		  	fprintf(fp,"%f ",crd_proj[jj]);
	  	}
	  	fprintf(fp,"\n");
		errout("Projection done !\n");
	}
	fprintf(fp,"[CENTER_OF_MASS_SELECTED]\n");
	for (j=0,k=0;j<nr_of_atoms;j++) {
		if (use_this_atom[j]) {
                        fprintf(fp,"%f %f %f\n",flooding_cm[k],flooding_cm[k+1],flooding_cm[k+2]);
			k += 3;
		}
	}
	fprintf(fp,"[EIGENVECTORS]\n");
	for (jj=0; jj<nr_of_projections; jj++) {
        	for (j=0,k=0;j<nr_flood_atoms;j++) {
			fprintf(fp,"%f %f %f\n",largest_cov_eigenvectors[jj][k],
                                                largest_cov_eigenvectors[jj][k+1],
                                                largest_cov_eigenvectors[jj][k+2]);
                        k += 3;
			}
		}
	fclose(fp);
	free(crd_proj);
	return 0;

error_occured:
	free_all_buffers(nr_flood_atoms);

	errout("\n=================================================================\n");
	errout("mkflood stopped with a fatal error.\n");
	errout("Check your Xplor-input-files or the ego-data-files for this error.\n");
	errout("==================================================================\n");
	return -1;
}




static int	save_flooding_fake(char *fname,int nr_of_atoms,int nr_flooding_atoms)
{
	int	i;
	FILE	*fp;

	if ((fp = fopen(fname,"w")) == NULL) {
		errout("Can not open flood-file <%s> !",fname);
		return -1;
	}
	fprintf(fp,"[HEADER]\n");
	fprintf(fp,"%d\n",0);			/*** nr of gaussians ***/
	fprintf(fp,"%d\n",nr_of_atoms);		/*** nr of atoms ***/
	fprintf(fp,"%d\n",nr_flooding_atoms);

	fprintf(fp,"[USINGARRAY]\n");
	for (i=0;i<nr_of_atoms;i++) {
		fprintf(fp,"%d\n",use_this_atom[i]);
	}
	fprintf(fp,"[CENTER_OF_MASS]\n");
	fprintf(fp,"[FLOODING_MATRIX]\n");
	fprintf(fp,"[COVARIANCE EIGENVALUES]\n");	
	fclose(fp);
	return 0;
}


/*****************************************/
/*** came from numerical recipies ********/
/*****************************************/

static void	tred2(double **a,int n,double *d,double *e)
{
	int l,k,j,i;
	double scale,hh,h,g,f;

	for (i=n;i>=2;i--) {
		txtout(".");
		l=i-1;
		h=scale=0.0;
		if (l > 1) {
			for (k=1;k<=l;k++)
				scale += fabs(a[i][k]);
			if (scale == 0.0)
				e[i]=a[i][l];
			else {
				for (k=1;k<=l;k++) {
					a[i][k] /= scale;
					h += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = f>0 ? -sqrt(h) : sqrt(h);
				e[i]=scale*g;
				h -= f*g;
				a[i][l]=f-g;
				f=0.0;
				for (j=1;j<=l;j++) {
			/* Next statement can be omitted if eigenvectors not wanted */
					a[j][i]=a[i][j]/h;
					g=0.0;
					for (k=1;k<=j;k++)
						g += a[j][k]*a[i][k];
					for (k=j+1;k<=l;k++)
						g += a[k][j]*a[i][k];
					e[j]=g/h;
					f += e[j]*a[i][j];
				}
				hh=f/(h+h);
				for (j=1;j<=l;j++) {
					f=a[i][j];
					e[j]=g=e[j]-hh*f;
					for (k=1;k<=j;k++)
						a[j][k] -= (f*e[k]+g*a[i][k]);
				}
			}
		} else
			e[i]=a[i][l];
		d[i]=h;
	}
	/* Next statement can be omitted if eigenvectors not wanted */
	d[1]=0.0;
	e[1]=0.0;
	/* Contents of this loop can be omitted if eigenvectors not
			wanted except for statement d[i]=a[i][i]; */
	txtout("\n");
	for (i=1;i<=n;i++) {
		txtout("+");
		l=i-1;
		if (d[i]) {
			for (j=1;j<=l;j++) {
				g=0.0;
				for (k=1;k<=l;k++)
					g += a[i][k]*a[k][j];
				for (k=1;k<=l;k++)
					a[k][j] -= g*a[k][i];
			}
		}
		d[i]=a[i][i];
		a[i][i]=1.0;
		for (j=1;j<=l;j++) a[j][i]=a[i][j]=0.0;
	}
	txtout("\n");
}

/****************************************************************/

static void	tqli(double *d,double *e,int n,double **z)
{
	int m,l,iter,i,k;
	double s,r,p,g,f,dd,c,b;

	for (i=2;i<=n;i++) e[i-1]=e[i];
	e[n]=0.0;
	for (l=1;l<=n;l++) {
		txtout("x");
		iter=0;
		do {
			for (m=l;m<=n-1;m++) {
				dd=fabs(d[m])+fabs(d[m+1]);
				if (fabs(e[m])+dd == dd) break;
			}
			if (m != l) {
				if (iter++ == 30) errout("Too many iterations in TQLI");
				g=(d[l+1]-d[l])/(2.0*e[l]);
				r=sqrt((g*g)+1.0);
				g=d[m]-d[l]+e[l]/(g+SIGN(r,g));
				s=c=1.0;
				p=0.0;
				for (i=m-1;i>=l;i--) {
					f=s*e[i];
					b=c*e[i];
					if (fabs(f) >= fabs(g)) {
						c=g/f;
						r=sqrt((c*c)+1.0);
						e[i+1]=f*r;
						c *= (s=1.0/r);
					} else {
						s=f/g;
						r=sqrt((s*s)+1.0);
						e[i+1]=g*r;
						s *= (c=1.0/r);
					}
					g=d[i+1]-p;
					r=(d[i]-g)*s+2.0*c*b;
					p=s*r;
					d[i+1]=g+p;
					g=c*r-b;
				/* Next loop can be omitted if eigenvectors not wanted */
					for (k=1;k<=n;k++) {
						f=z[k][i+1];
						z[k][i+1]=s*z[k][i]+c*f;
						z[k][i]=c*z[k][i]-s*f;
					}
				}
				d[l]=d[l]-p;
				e[l]=g;
				e[m]=0.0;
			}
		} while (m != l);
	}
        txtout("\n");
}

/***************************************************************************************/
/*********************************************/

static void	tred2_ew(double **a,int n,double *d,double *e)
{
	int l,k,j,i;
	double scale,hh,h,g,f;

	for (i=n;i>=2;i--) {
		txtout(".");
		l=i-1;
		h=scale=0.0;
		if (l > 1) {
			for (k=1;k<=l;k++)
				scale += fabs(a[i][k]);
			if (scale == 0.0)
				e[i]=a[i][l];
			else {
				for (k=1;k<=l;k++) {
					a[i][k] /= scale;
					h += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = f>0 ? -sqrt(h) : sqrt(h);
				e[i]=scale*g;
				h -= f*g;
				a[i][l]=f-g;
				f=0.0;
				for (j=1;j<=l;j++) {
					g=0.0;
					for (k=1;k<=j;k++)
						g += a[j][k]*a[i][k];
					for (k=j+1;k<=l;k++)
						g += a[k][j]*a[i][k];
					e[j]=g/h;
					f += e[j]*a[i][j];
				}
				hh=f/(h+h);
				for (j=1;j<=l;j++) {
					f=a[i][j];
					e[j]=g=e[j]-hh*f;
					for (k=1;k<=j;k++)
						a[j][k] -= (f*e[k]+g*a[i][k]);
				}
			}
		} else
			e[i]=a[i][l];
		d[i]=h;
	}
	e[1]=0.0;
	txtout("\n");
	for (i=1;i<=n;i++) {
		txtout("+");
		d[i]=a[i][i];
	}
	txtout("\n");
}

/****************************************************************/

static void	tqli_ew(double *d,double *e,int n,double **z)
{
	int m,l,iter,i;
	double s,r,p,g,f,dd,c,b;

	for (i=2;i<=n;i++) e[i-1]=e[i];
	e[n]=0.0;
	for (l=1;l<=n;l++) {
		txtout("x");
		iter=0;
		do {
			for (m=l;m<=n-1;m++) {
				dd=fabs(d[m])+fabs(d[m+1]);
				if (fabs(e[m])+dd == dd) break;
			}
			if (m != l) {
				if (iter++ == 30) errout("Too many iterations in TQLI");
				g=(d[l+1]-d[l])/(2.0*e[l]);
				r=sqrt((g*g)+1.0);
				g=d[m]-d[l]+e[l]/(g+SIGN(r,g));
				s=c=1.0;
				p=0.0;
				for (i=m-1;i>=l;i--) {
					f=s*e[i];
					b=c*e[i];
					if (fabs(f) >= fabs(g)) {
						c=g/f;
						r=sqrt((c*c)+1.0);
						e[i+1]=f*r;
						c *= (s=1.0/r);
					} else {
						s=f/g;
						r=sqrt((s*s)+1.0);
						e[i+1]=g*r;
						s *= (c=1.0/r);
					}
					g=d[i+1]-p;
					r=(d[i]-g)*s+2.0*c*b;
					p=s*r;
					d[i+1]=g+p;
					g=c*r-b;
				}
				d[l]=d[l]-p;
				e[l]=g;
				e[m]=0.0;
			}
		} while (m != l);
	}
        txtout("\n");
}

/***************************************************************************************/

void sort(int n, double *ra)
{
        int l,j,ir,i;
        double rra;

        l=(n >> 1)+1;
        ir=n;
        for (;;) {
                if (l > 1) {
                        rra=ra[--l];
                } else {
                        rra=ra[ir];
                        ra[ir]=ra[1];
                        if (--ir == 1) {
                                ra[1]=rra;
                                return;
                        }
                }
                i=l;
                j=l << 1;
                while (j <= ir) {
                        if (j < ir && ra[j] < ra[j+1]) ++j;
                        if (rra < ra[j]) {
                                ra[i]=ra[j];
                                j += (i=j);
                        }
                        else j=ir+1;
                }
                ra[i]=rra;
        }
}


void sort2(int n, double *ra, double *rb)
{
        int l,j,ir,i;
        double rra;
        int rrb;

        l=(n >> 1)+1;
        ir=n;
        for (;;) {
                if (l > 1) {
                        rra=ra[--l];
                        rrb=rb[l];
                } else {
                        rra=ra[ir];
                        rrb=rb[ir];
                        ra[ir]=ra[1];
                        rb[ir]=rb[1];
                        if (--ir == 1) {
                                ra[1]=rra;
                                rb[1]=rrb;
                                return;
                        }
                }
                i=l;
                j=l << 1;
                while (j <= ir) {
                        if (j < ir && ra[j] < ra[j+1]) ++j;
                        if (rra < ra[j]) {
                                ra[i]=ra[j];
                                rb[i]=rb[j];
                                j += (i=j);
                        }
                        else j=ir+1;
                }
                ra[i]=rra;
                rb[i]=rrb;
        }
}
