/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/****************************************************************************
**	Belongs to: mkmaxw.c - Assign a Maxwell-distribution to coord.lis  **									**									   **
**	Module	: mkmaxw.c	by M.Eichinger				   **
**									   **
**      Modif's:                                                           **
**      19950907.  HH corrected several output messages (typos...)         **
**	19980611.  HH changed default behaviour: reads tmp from ctl.lis	   **
**									   **
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "../hardw.h"
#include "../ego.h"
#include "../egcolib.h"
#include "../util_lib.h"
#include "../woalloc.h"

/*********************************************************************/
static	void	print_prog_header();
static	void	print_wrong_call();
static	int	readin_vdwfile(char *vdw_filen);
static	double	gauss_distri(double width);
static	double	maxwell(double temperature, double mass, double *vel);

/**********************************************************************/
/**********************************************************************/

static	double	temperature = -1.0; /* This is a flag since negative temps. don't make sense here */

static  char    version_string[]=
                  {"Version 1.1, 11.06.1998,\nby M. Eichinger, H. Grubmueller, H. Heller, Munich, Germany\n"};
static	char	curpath[MAX_STRING_SIZE];

static	int	nr_atomtypes=0;
static	int	nr_of_atoms=0;
static	double	mass[MXNO_ATOM_TYPES];
static	char	sdum[MAX_STRING_SIZE];		/* space for dummy-strings */

static	double	harmon=0.0;
static	int	harmonmode=0;
static 	char	selstr[MAX_STRING_SIZE];

static	STRUCT_ATOMDATA *atom_data;

/**********************************************************************/
/* Here some dummy definitions to satisfy the loader.... */

int set_qmboundbox( int *atom_sele_flag1,
		STRUCT_ATOMDATA *atom_data,
		LOAD_MSG_ALL_NODE *global_data)
{
return -1;
}

int parse_qminterface(	char *keywordstr,char *reststr,
			STRUCT_ATOMDATA *atom_data,
			int *atom_sele_flag1,
			LOAD_MSG_ALL_NODE *global_data,
			STRUCT_FORCES *my_force,
			int *used_atoms)
{
return -1;
}

void zero_pressure (PRESSURE *pressure)
{
return ;
}

/*** pex is the abbreviation for "Program EXit" ***/
void   pex(int i)
{	
	txtout("Exit by pex(%d) in mkmaxw.c !\n",i);
	exit(i);
	return;
}



/**********************************************************************/
int	main(int argc,char *argv[])
{
	FILE	*fopen(), *fp;
	int	i,numopt, noupdate=FALSE;
	char	ctlfilenam[MAX_STRING_SIZE];
	char	outfilen[MAX_STRING_SIZE];
	double	vx,vy,vz,mges;
	long	rseed=-123652L;
	LOAD_MSG_ALL_NODE global_data;

	print_prog_header();

        strcpy(ctlfilenam,"ctl.lis");
        strcpy(curpath,"");

	numopt=0;
	/*** parse the argument list ***/
	for (i=1;i<argc;i++) {
		if (strcmp(argv[i],"-h") == 0) {
			print_wrong_call();
			return 0;
		}
		else if ((argv[i][0]=='-') && (argv[i][1]=='s')) {
			sscanf(argv[i]+2,"%ld",&rseed);
			if (rseed>0L) rseed=-rseed;
			txtout("Random generator seed set to %ld\n",-rseed);
			set_rand_seed(rseed);
			numopt++;
		}
		else if (	(strcmp(argv[i],"-harmon")==0) ||
			(strcmp(argv[i],"-Harmon")==0) ) {
			
			if (argv[i][1]=='H') harmonmode=1;
			
			if (sscanf(argv[i+1],"%lf",&harmon)!=1) {
				print_wrong_call();
				return -1;
			}
			strcpy(selstr,argv[i+2]);
			txtout("Setting harmonic constraints %f for <%s>\n",harmon,selstr);
			if (harmonmode==1) txtout("Deleting all other harmonic constraints.\n");
			numopt += 3;
		}
		else if (strcmp(argv[i],"-noupdate")==0)
		{
		  noupdate = TRUE;
		  numopt++;
		}
		else if (sscanf(argv[1],"%lf",&temperature)!=1) {
			print_wrong_call();
			return -1;
		}
	}
	
	if (argc == (4+numopt)) {
		strcpy(ctlfilenam,argv[3]);
		extract_path(ctlfilenam,curpath);
	}
  	read_ctl_file (ctlfilenam, &global_data);
        if (temperature < 0.0) temperature = global_data.target_temp;
	txtout("Assigned temperature is %f Kelvin.\n",temperature);

        sprintf(outfilen,"%scoord_%dK.lis",curpath,(int)(temperature));
	if (argc >= (3+numopt)) {
                sprintf(outfilen,"%s%s",curpath,argv[2]);
	}

	nr_of_atoms=global_data.total_no_atoms;
        txtout("Number of atoms is %d.\n",nr_of_atoms);

	/***** buffer for holding atom-data ********************************/
	atom_data=alloc_atomdata(nr_of_atoms+5);
						/* work on vdw-file */
        txtout("Reading atom masses from file <%s>.\n",global_data.filenames[VDWPARA_FILE_INDEX]);
	if (readin_vdwfile(global_data.filenames[VDWPARA_FILE_INDEX])==-1) {
                errout("Fatal error in VDW-file!\n");
		return -1;
	}

        txtout("Reading atom coordinates from file <%s>.\n",global_data.filenames[COORD_FILE_INDEX]);
	if (read_coord_file(&global_data,nr_of_atoms,global_data.filenames[COORD_FILE_INDEX],atom_data)==-1) {
                errout("Fatal error in coordinate-file!\n");
		return -1;
	}
 
        txtout("Setting new velocities ... ");
        mges = 0.0;
        vx = 0.0;
        vy = 0.0;
        vz = 0.0;
	for (i=0;i<nr_of_atoms;i++) {
		mges += mass[atom_data[i].atomtype];
		maxwell(temperature,
			mass[atom_data[i].atomtype],
			atom_data[i].vel);
		vx += mass[atom_data[i].atomtype]*atom_data[i].vel[0];
		vy += mass[atom_data[i].atomtype]*atom_data[i].vel[1];
		vz += mass[atom_data[i].atomtype]*atom_data[i].vel[2];		
	}
        txtout("Done !\n");

	vx /= mges*nr_of_atoms;
	vy /= mges*nr_of_atoms;
	vz /= mges*nr_of_atoms;	
	txtout("Test total velocity %f %f %f    Mass: %f\n",vx,vy,vz,mges);
	for (i=0;i<nr_of_atoms;i++) {
		atom_data[i].vel[0] -= mges*vx/mass[atom_data[i].atomtype];
		atom_data[i].vel[1] -= mges*vy/mass[atom_data[i].atomtype];
		atom_data[i].vel[2] -= mges*vz/mass[atom_data[i].atomtype];		
	}

	/***** setting harmonic constants for constraints ***/
	if (selstr[0]!=0) {
		int *used_atoms;
		used_atoms = alloc_ints(nr_of_atoms);
		for (i=0;i<nr_of_atoms;i++) used_atoms[i] = 0;
		if (set_used_atoms(selstr,nr_of_atoms,atom_data,used_atoms)<0) {
			errout("Fatal error: Correct this error in your parameter <%s>\n",selstr);
			return -1;
		}	
		for (i=0;i<nr_of_atoms;i++) {
			if (used_atoms[i]!=0)   atom_data[i].harmon=harmon;
			else if (harmonmode==1) atom_data[i].harmon=0.0;
		}
		myfree(used_atoms);
	}
		
	if (writeout_coordfile(nr_of_atoms,outfilen,atom_data)==-1) {
                errout("Fatal error while writing new coordinate-file!\n");
		return -1;
	}
	free(atom_data);

	if(noupdate)
	{
	  txtout("Change the %d. filename (default <coord.lis>) in your control file\n",COORD_FILE_INDEX+1);
	  txtout("to <%s>, if you want to use the new velocities for your future EGO runs.\n",outfilen);
	}
	else
	{
	  sprintf(sdum,"mv %s %s.bak\n", ctlfilenam, ctlfilenam);
	  system(sdum);
	  txtout("The %d. filename (previously %s) in your control file\n",
	          COORD_FILE_INDEX+1,global_data.filenames[COORD_FILE_INDEX]);
	  txtout("has been changed to <%s>.\n",outfilen);
	  txtout("Your old ctl-file (<%s>) has been moved to <%s.bak>.\n", ctlfilenam, ctlfilenam);
	  strcpy(global_data.filenames[COORD_FILE_INDEX], outfilen);
	  if ((fp=fopen(ctlfilenam,"w"))==NULL) {
		  errout("Can not open ctl-file <%s> for writing.\n", ctlfilenam);
		  return -1;
	  }
	  writeout_ctlfile(fp,&global_data);
	  fclose(fp);
	}
	return 0;
}


static
void	print_prog_header()
{
	txtout("===================================================================\n");
  	txtout("\nmkmaxw - Assigns a Maxwell-distribution to an EGO coordinate file.\n");
	txtout(version_string);
        txtout("===================================================================\n\n");
}

static
void	print_wrong_call()
{
        txtout("This program reads the EGO control file and creates a coordinate\n");
        txtout("lis-file with the specified temperature.\n\n");

        txtout("Program-Usage:\n");
        txtout("mkmaxw [<temperature>] [<newcoord.lis>] [<controlfile>] [other options]\n");
        txtout("<temperature>	       Temperature in Kelvin (otherwise taken from ctl.lis).\n");
        txtout("<newcoord.lis>	       Name for new coordinate file (default: coord_<temp>.lis).\n");
        txtout("<controlfile>	       Controlfile to use (default: ctl.lis).\n");
	txtout("-s<longnumber>	       Set seed of random generator.\n");	
        txtout("-harmon <k> selstr     Set harmonic constants only for selected atoms.\n");
        txtout("-Harmon <k> selstr     Set harmonic constants for selected atoms and \n");
        txtout("                       delete all other harmonic constants.\n");
	txtout("-noupdate              Do NOT update control file\n");
	txtout("                       (default: DO update the coordinate file name).\n");
	txtout("-h                     Help. Give this message.\n");
}

static
int	readin_vdwfile(char *vdw_filen)
{
	FILE	*fopen();
	FILE	*h_vdwfile;
	int	i,j;
	double	read_double(),x;
						/* open my vdw-file */
        if ((h_vdwfile=fopen(vdw_filen,"r"))==NULL) {
                errout("Can not open vdw-file <%s>.\n",vdw_filen);
		return -1;
	}
						/* skip header */
	skip_line(h_vdwfile,1);
        fscanf(h_vdwfile,"%s%s%s%s%s%d",sdum,sdum,sdum,sdum,sdum,&nr_atomtypes);
	if (nr_atomtypes<1)	return -1;

        txtout("Atom types in VDW-File : %d\n",nr_atomtypes);
						/* get atommasses in diag */
	for (i = 0; i<=nr_atomtypes; i++) {
		for (j = 0; j<=nr_atomtypes; j++) {
			if (i==j) {
				mass[j]=read_double(h_vdwfile);
				if (	(mass[j]<0.0) ||
					(mass[j]>100.0) ) {
                                        errout("Error: strange mass = %f !\n",mass[j]);
				}
			} else {
				x=read_double(h_vdwfile);
			}
		}
	}

	fclose(h_vdwfile);
	return 0;
}

/*****************************************************************************
produces a Gauss distribution of random numbers
width is the width (standard deviation) of the Gauss distribution
the mean is always zero
P(x) = exp (-(x**2)/2), variance = 1 (that is why it is called NORMAL distribution)
P(x) * width: Gauss distribution with new variance width*width and
	      standardard deviation width

The underlaying principle is that the mean and variance of uniformly distributed
random numbers is 1/2 and 1/12 resp.

******************************************************************************/

static
double gauss_distri(double width)
{
	int i;
	double	p=0.0;

	for (i=0;i<12;i++) p+=ego_rand();
	return (p-6.0)*width;
}

/**********
in ego.h KBOLTZ is defined in AKMA units,
temperature is in Kelvin and masses are in amu.
Velocities are given in Angstroem/sec
***********/

static
double	maxwell( double temperature, double mass, double *vel)
{
	double width;
	double gauss_distri(double width);

	if (temperature<SMALL) {
		vel[0]=vel[1]=vel[2]=0.0;
		return 0;
	}
	if (mass<SMALL) {
                errout("Fatal error: illegal mass for Maxwell-Distribution.\n");
		vel[0]=vel[1]=vel[2]=0.0;
		return -1;
	}

	width = sqrt((KBOLTZ*temperature)/mass);

	vel[0]=gauss_distri(width);
	vel[1]=gauss_distri(width);
	vel[2]=gauss_distri(width);

	return 0;
}




