/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: woangle.c	by M.Eichinger, Munich			**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis)**
**	Contents: scan in and write out angle-list and parameters.	**
**									**
**  	Modif's:							**
**		05.06.94						**
**		08.12.94 Impruved error-output in case of errors	**
**		16.01.94 now numbers of angles is written to header of file	**
**									**
*************************************************************************/
#include <math.h>

#include "hardw.h"
#include "ego.h"
#include "woalloc.h"
#include "util_lib.h"
#include "wodecls.h"


#define TEST_ANGLEPARA	0

typedef struct {
	double	kt;
	double	theta0;
	double	ubk;
	double	ubd;
	int	att1;
	int	att2;
	int	att3;
} anglepar_type;

/******************************************************************************/
/**************** functions in this modul *************************************/
/******************************************************************************/

	int	writeout_anglelist(char *psffn,char *anglelistfn,int nr_of_atoms);
	int	writeout_anglepara(char *angleparafn);
	int	free_anglebuffers(void);

static	int	get_anglepara_usedindex(int *angletrip);
static	int	get_anglepara_allindex(int *angletrip);
static	int	add_anglepara_usedindex(int all_index);
static	int	fill_angle_paralist(FILE *h_para,anglepar_type *all_buf, char atom_typestr_list[][LEN_ATOM_TYPE+1]);
static	int	allready_in_list(anglepar_type *angparabuf, char atom_typestr_list[][LEN_ATOM_TYPE+1]);
	
static	int	add_to_anglelist(int *anglelist,int *angletrip,int nr_read_angles);
static	int	get_next_angletrips(FILE *h_psf,int *angletrip);
static	int	goto_psf_angledata(FILE *h_psf,int *nr_of_angles);

/******************************************************************************/
/********************* some vars **********************************************/
/******************************************************************************/

static	int		nr_angle_allpara=0;
static	anglepar_type	*all_buf;
static	int		nr_angle_usedpara=0;
static	anglepar_type	*used_buf;
static	int		angle_buffers_here=0;

/*****************************************************************************/

int	readin_all_anglepara(int argc,char *argv[], char atom_typestr_list[][LEN_ATOM_TYPE+1])
{
	int	file_ind;		/* 3 is the first parameter file */
	FILE 	*h_para;
	
	/****************************** allocate *****************************/
	all_buf=(anglepar_type*) malloc(MXNO_ANGLE_PARAM*sizeof(anglepar_type));
	if (all_buf==NULL) {
		errout("Fatal Error: Can not allocate space for angle-list.\n");
		return -1;
	}	
	nr_angle_allpara=0;	

	used_buf=(anglepar_type*) malloc(MXNO_ANGLE_PARAM*sizeof(anglepar_type));
	if (used_buf==NULL) {
		myfree(all_buf);
		errout("Fatal error: Can not allocate space for angle-list.\n");
		return -1;
	}	
	nr_angle_usedpara=0;

	/******************* scan each parmater-file *************************/
	for (file_ind=3;file_ind<argc;file_ind++) {
		if ((h_para=fopen(argv[file_ind],"r"))==NULL) {
			errout("Can not read parameter-file: %s\n",argv[file_ind]);
			myfree(all_buf);
			myfree(used_buf);	
			return -1;
		}

		if (fill_angle_paralist(h_para,all_buf, atom_typestr_list)==-1) {
			errout("Error occured in parameter-file: %s\n",argv[file_ind]);
			fclose(h_para);
			myfree(all_buf);
			myfree(used_buf);
			return -1;
		}
		
		fclose(h_para);
	}
	angle_buffers_here=1;
	return 0;
}

/*****************************************************************************/

int	free_anglebuffers(void)
{
	if (angle_buffers_here) {
		myfree(all_buf);
		myfree(used_buf);
		angle_buffers_here=0;
	}
	return 0;
}

/*****************************************************************************/
static	int	*anglelist;

int	writeout_anglelist(char *psffn,char *anglelistfn,int nr_of_atoms)
{
	int	i,res1,off=0;
	int	atom1,atom2,atom3;
	int	nr_of_angles=0,nr_read_angles=0;
	int	angletrips[4*3],nr_converted=0;
	int	all_index=0,anglelistbuf_here=0;
	FILE	*h_psf,*h_angle;

	
	if ((h_psf=fopen(psffn,"r"))==NULL) {
		errout("Can not open specified psf-file.\n");
		return -1;
	}

	if (goto_psf_angledata(h_psf,&nr_of_angles)==-1) {
		goto write_angle_list;		/* no anglelist-section ? */
	}

	anglelist=alloc_ints((nr_of_angles+10)*4);
	if (anglelist==NULL) {
		errout("Can not allocate space for anglelist.\n");
		fclose(h_psf);
		return -1;
	}
	anglelistbuf_here=1;

	/******************** build up sorted anglelist **********************/

	while ((nr_converted=get_next_angletrips(h_psf,angletrips))>0) {
		for (i=0;i<nr_converted;i+=4) {
	
			atom1=angletrips[i  ];		/* just for better reading */
			atom2=angletrips[i+1];
			atom3=angletrips[i+2];
							/* some tests 	*/		
			if (	(atom1<0) || (atom1>=nr_of_atoms) ||
				(atom2<0) || (atom2>=nr_of_atoms) ||
				(atom3<0) || (atom3>=nr_of_atoms) ) {
				errout("Fatal error: Illegal atomnumber in PSF-file.\n");
				goto error_in_anglelist;	
			}		

			/******* build up the angle-PARAMTER-list and get the index ************/		
			if ((angletrips[i+3]=get_anglepara_usedindex(angletrips+i))==-1) {
				all_index=get_anglepara_allindex(angletrips+i);
				if (all_index==-1) {
					errout("Fatal error: No angle parameters for atom# %d %d %d in psf-file.\n",
						atom1+1,atom2+1,atom3+1);
					{ char s1[20],s2[20],s3[20];
					get_atomtype_str(get_atti_by_atomnr(atom1),s1);
					get_atomtype_str(get_atti_by_atomnr(atom2),s2);
					get_atomtype_str(get_atti_by_atomnr(atom3),s3);
					errout("Atom-type-strings <%s> <%s> <%s>\n",s1,s2,s3);
					}
/**
					fclose(h_psf);
					return -1;				
**/

				}
				if ((angletrips[i+3]=add_anglepara_usedindex(all_index))==-1) {
					errout("Fatal error: Too many used-angle-paramaters.\n");
					fclose(h_psf);
					return -1;	
				}	
			}
			/**** now angletrips[i+3] holds the index to the angle-PARAMETER-list */
			/**** save the angle-trips & the index too the angle-PARMETER-list ****/		
			res1=add_to_anglelist(anglelist,angletrips+i,nr_read_angles);
						
			if (res1==-1) goto error_in_anglelist;
				
			if (++nr_read_angles>nr_of_angles) {
				errout("Fatal error: Incompatible number (to much !) of angles in PSF-file.\n");
		goto error_in_anglelist;
			}
		} /******* for next: for each triplet in line */

	} /******* while: for each line in angle-pair-section */

	if (nr_converted<0)	goto error_in_anglelist;

	if (nr_read_angles!=nr_of_angles) {
		errout("Fatal error: Incompatible number (to less !) of angles in PSF-file.\n");
		goto error_in_anglelist;		
	}
		
	/******************** write out anglelist to file ********************/
write_angle_list:	
	if ((h_angle=fopen(anglelistfn,"w"))==NULL) {
		errout("Can not open angle-list-file.\n");
		goto error_in_anglelist;
	}
	fprintf(h_angle,"%d  Number of angle.\n",nr_of_angles);
	fprintf(h_angle,"Format of the ANGLE list:\n");
	fprintf(h_angle,"Number   1. atom  2. atom  3. atom    index\n");

	for (i=0,off=0;i<nr_of_angles;i++,off+=4) {		/***** for each angle *****/
		fprintf(h_angle,"%4d      %4d     %4d     %4d      %4d\n",
			i,
			anglelist[off+0],anglelist[off+1],anglelist[off+2],
			anglelist[off+3]);
	}

	txtout("Read angles in PSF-file    : %d\n",nr_read_angles);

	if (anglelistbuf_here)	myfree(anglelist);
	fclose(h_psf);
	fclose(h_angle);

	return 0;

error_in_anglelist:
	errout("Error after %d read angles.\n",nr_read_angles);
	if (anglelistbuf_here) myfree(anglelist);
	fclose(h_psf);
	return -1;
}

/*****************************************************************************/

int	writeout_anglepara(char *angleparafn)
{
	int	i;
	FILE 	*h_anglep;

	if ((h_anglep=fopen(angleparafn,"w"))==NULL) {
		errout("Can not open angle-parameter-file.\n");
		return -1;
	}

	fprintf(h_anglep,"This is the ANGLE parameter list.\n");
	fprintf(h_anglep,"Number of parameter sets:   %4d .\n",nr_angle_usedpara);
	fprintf(h_anglep,"Format:\n");
	fprintf(h_anglep," Index        Kt        Theta0        UB-k         UB-d\n");

	for (i=0;i<nr_angle_usedpara;i++) {
		fprintf(h_anglep,"  %4d     %f     %f     %f     %f\n",
				i,
				used_buf[i].kt,
				used_buf[i].theta0,
				used_buf[i].ubk,
				used_buf[i].ubd);
	}
	fclose(h_anglep);

	txtout("Used angle-parameters      : %d\n",nr_angle_usedpara);
	return 0;
}

/****************************************************************************/

static
int	get_anglepara_usedindex(int *angletrip)
{
	int	i;
	int	att1,att2,att3;

	if (nr_angle_usedpara==0)	return -1;	/* nothing at all in list */
							/* no error here */	
	att1=get_atti_by_atomnr(angletrip[0]);
	att2=get_atti_by_atomnr(angletrip[1]);
	att3=get_atti_by_atomnr(angletrip[2]);

	for (i=0;i<nr_angle_usedpara;i++) {
		if (	(att1==used_buf[i].att1) &&
			(att2==used_buf[i].att2) &&
			(att3==used_buf[i].att3) ) {
				return i;		/* found match in used-list */
		}
		/****** swap atoms **********/	
		if (	(att3==used_buf[i].att1) &&
			(att2==used_buf[i].att2) &&
			(att1==used_buf[i].att3) ) {
				return i;		/* found match in used-list */
		}
	}
	return -1;
}

/*****************************************************************************/

static
int	get_anglepara_allindex(int *angletrip)
{
	int	i;
	int	att1,att2,att3;

	if (nr_angle_allpara==0) {
		errout("Fatal error: no angle parameters in list.\n");
		return -1;				/* nothing at all in list */
	}						/* that should not be !!! */
	att1=get_atti_by_atomnr(angletrip[0]);
	att2=get_atti_by_atomnr(angletrip[1]);
	att3=get_atti_by_atomnr(angletrip[2]);

	for (i=0;i<nr_angle_allpara;i++) {

		if (	(att1==all_buf[i].att1) &&
			(att2==all_buf[i].att2) &&
			(att3==all_buf[i].att3) ) {
				return i;		/* found match in all-list */
		}
		/******* swap atoms **************/
		if (	(att3==all_buf[i].att1) &&
			(att2==all_buf[i].att2) &&
			(att1==all_buf[i].att3) ) {
				return i;		/* found match in all-list */
		}
	}
	errout("Fatal error: Unknown angle-Tripplet !\n");
	return -1; 				/* unknown angle-tripplet, thats an error ! */			
}

/****************************************************************************/

static
int	add_anglepara_usedindex(int all_index)
{
	memcpy(&used_buf[nr_angle_usedpara++],&all_buf[all_index],sizeof(anglepar_type));
	
	return nr_angle_usedpara-1;
}

/*****************************************************************************/

static	
int	fill_angle_paralist(FILE *h_para,anglepar_type *all_buf, char atom_typestr_list[][LEN_ATOM_TYPE+1])
{
	double 	angle_kt,angle_theta0,ubk,ubd;
	int	att1,att2,att3;	/* atom-type-index ! */
	int 	res,res1;

	while ((res=read_next_anglepara(h_para,	&att1,&att2,&att3,
						&angle_kt,&angle_theta0,
						&ubk,&ubd))>=0) {

		if (res==0)	continue;	/* found atom-type which don't occur in */
						/* PDB/PSF-file, no error, skip it  */

		all_buf[nr_angle_allpara].att1	=att1;
		all_buf[nr_angle_allpara].att2	=att2;
		all_buf[nr_angle_allpara].att3	=att3;
		all_buf[nr_angle_allpara].kt	=angle_kt;
		all_buf[nr_angle_allpara].theta0=angle_theta0/180.0*3.1415926;
		all_buf[nr_angle_allpara].ubk	=ubk;
		all_buf[nr_angle_allpara].ubd	=ubd;
				
#if TEST_ANGLEPARA
	printf("TEST: Angle-params: %3d  %3d  %3d   %f  %f %f %f\n",
	all_buf[nr_angle_allpara].att1,all_buf[nr_angle_allpara].att2,all_buf[nr_angle_allpara].att3,
	all_buf[nr_angle_allpara].kt,
	all_buf[nr_angle_allpara].theta0);
#endif

		res1=allready_in_list(all_buf,atom_typestr_list); /* last added element is checked against */
								  /* all others */
		if (res1==1)	continue;			  /* allready in list */


		if (++nr_angle_allpara>=MXNO_ANGLE_PARAM) {
			errout("Fatal error: Too many angle-parameters (MXNO_ANGLE_PARAM is %d).\n",MXNO_ANGLE_PARAM);
			errout("Increase MXNO_ANGLE_PARAM in ego.h !\n");
			return -1;
		}
	}
	if (res==-2) {
		errout("Fatal error: Illegal parameter-line while reading angle parameters.\n");
		return -1;		/* thats an error */
	}
	return 0;
}
/******************************************************************************/
/***** last added element is checked against all others *****/

static
int	allready_in_list(anglepar_type *angparabuf, char atom_typestr_list[][LEN_ATOM_TYPE+1])
{
	int	i;

	for (i=0;i<nr_angle_allpara;i++) {
	    if ( ((angparabuf[i].att1==angparabuf[nr_angle_allpara].att1) &&
		  (angparabuf[i].att2==angparabuf[nr_angle_allpara].att2) &&
		  (angparabuf[i].att3==angparabuf[nr_angle_allpara].att3)) || 
										/* swap it */
		 ((angparabuf[i].att1==angparabuf[nr_angle_allpara].att3) &&
		  (angparabuf[i].att2==angparabuf[nr_angle_allpara].att2) &&
		  (angparabuf[i].att3==angparabuf[nr_angle_allpara].att1)) 
	       ) {

		    if ((fabs(angparabuf[i].kt     -angparabuf[nr_angle_allpara].kt    )<SMALL) &&
		 	(fabs(angparabuf[i].theta0 -angparabuf[nr_angle_allpara].theta0)<SMALL) 
		       ) {

			return 1;	/* thats ok , totally identic, so skip it */
		    } else {
			
			errout("Warning error: Inconsistent angle-parameters for atomtype <%s> <%s> <%s>\n",
				atom_typestr_list[angparabuf[i].att1],
				atom_typestr_list[angparabuf[i].att2],
				atom_typestr_list[angparabuf[i].att3]);

			errout("old vals (kt,theta0) (%f, %f) skipped vals: (%f, %f)\n",
				angparabuf[i].kt,angparabuf[i].theta0,
				angparabuf[nr_angle_allpara].kt,angparabuf[nr_angle_allpara].theta0);
			errout("Skipping new angle-parameters !\n");

			return 1;	/* strings match, but not the values ==> inconsitency */
					/* skip this too */
		    }
	    }
	}
	return 0;			/* not in list */
}


/*****************************************************************************/

static
int	add_to_anglelist(int *anglelist,int *angletrip,int nr_read_angles)
{
	int	i;
	int 	off;
	int	swapbuf[4];

	off=4*nr_read_angles;	
							/* add to end of (sorted) list */
	anglelist[off+0]=angletrip[0];
	anglelist[off+1]=angletrip[1];
	anglelist[off+2]=angletrip[2];
	anglelist[off+3]=angletrip[3];

	if (nr_read_angles==0)	return 1;		/* first triplet, no sorting */

							/* sort new entry in ascending order */
							/* of the middle atom (angletrip[1]) */
	for (i=nr_read_angles,off=4*nr_read_angles;i>0;i--,off-=4) {
		if (anglelist[off+1]<anglelist[off-4+1]) {
			memcpy(swapbuf,		anglelist+off-4,4*sizeof(int));
			memcpy(anglelist+off-4,	anglelist+off,	4*sizeof(int));
			memcpy(anglelist+off,	swapbuf,	4*sizeof(int));
		} else	break;				/* now sorted to right place */
	}	
	return 1;
}

/*****************************************************************************/

static
int	get_next_angletrips(FILE *h_psf,int *angletrip)		/* returns number of integers */
{								/* in bondpair ==> even       */
	static	char	linestr[MAX_STRING_SIZE];
	int 	i,j;
	int	nr_converted;

	if (read_string(h_psf,linestr)==-1)	return -1;	/* error : can not read at all */

	nr_converted=sscanf(linestr,"%d %d %d %d %d %d %d %d %d",
				angletrip+0,angletrip+1,angletrip+2,
				angletrip+4,angletrip+5,angletrip+6,
				angletrip+8,angletrip+9,angletrip+10
				);
	
	if (nr_converted<=0) {		/* nothing converted, may be the end */
		return 0;
	} else
	if (nr_converted%3!=0) {
		errout("Fatal Error: even converted numbers in PSF-angle-list\n");
		return -1;
	} else {
		for (i=0,j=0;i<nr_converted/3;i++,j+=4) {
			angletrip[j  ]--;			/* atomnumbers in ego */
			angletrip[j+1]--;			/* start with 0 */
			angletrip[j+2]--;
		}
		angletrip[3 ]=0;
		angletrip[7 ]=0;
		angletrip[11]=0;	
		return (nr_converted/3)*4;			
	}
}

/*****************************************************************************/

static
int	goto_psf_angledata(FILE *h_psf,int *nr_of_angles)
{
	static	char	lstr[MAX_STRING_SIZE];
	static	char	tmpstr[MAX_STRING_SIZE];
	int 	res;

	while (read_string(h_psf,lstr)!=-1) {			/* read a psf-line */
		strup(lstr);
		if (sscanf(lstr,"%d %s",&res,tmpstr)==2) {	/* check it */
			tmpstr[7]=0;
			if (strcmp(tmpstr,"!NTHETA")==0) {
				*nr_of_angles=res;
				return 0;
			}
		}	
	}

	*nr_of_angles=0;
	errout("Warning error: No anglelist-data-section (!NTHETA) in PSF-file.\n");
	return -1;
}

