/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: wobond.c	by M.Eichinger, Munich			**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis)**
**	Contents: scan in and write out the bond-list of the atoms.	**
**		  provides a function for other moduls to get the 	**
**                bondlist for an atom.					**
**									**
**  	Modif's:							**
**		05.06.94						**
**									**
*************************************************************************/

#include <math.h>

#include "hardw.h"
#include "ego.h"
#include "woalloc.h"
#include "util_lib.h"
#include "wodecls.h"


/******************************************************************************/
/**************** functions in this modul *************************************/
/******************************************************************************/

	int	readin_bondlist(char *psffn,int nr_of_atoms);
	int	writeout_bondlist(char *bondlistfn);
	int	free_bondlist(void);
	int	get_atom_bondlist(int at_nr,int *bondl);

static	int	add_to_bondlist(int *curbondlist,int atom_nr);
static	int	get_next_bondpairs(FILE *h_psf,int *bondpair);
static	int	goto_psf_bonddata(FILE *h_psf,int *nr_of_bonds);

/******************************************************************************/
/**************** static vars for this modul **********************************/
/******************************************************************************/

static	int	buffer_here=0;
static	int	*bondlist;		/* pointer to buffer for the bondlist */
static	int	nr_read_bonds=0;
static	int	nr_of_atoms=0;

/*******************************************************************************/
/******************* global functions ******************************************/
/*******************************************************************************/

int	readin_bondlist(char *psffn,int nrofatoms)
{
	int	i,res1,res2;
	int	atom1,atom2;
	int	nr_of_bonds=0;
	int	bondpairs[2*5],nr_converted;
	FILE	*h_psf;
	
	nr_of_atoms=nrofatoms;

	if ((h_psf=fopen(psffn,"r"))==NULL) {
		errout("Can not open specified psf-file.\n");
		return -1;
	}

	bondlist=alloc_ints((nr_of_atoms+10)*(MXNO_FIELDS_BOND+1));
	buffer_here=1;

	if (goto_psf_bonddata(h_psf,&nr_of_bonds)==-1) {
		goto error_in_bondlist;		/* no bondlist-section ? */
	}

	/******************** build up sorted bondlist **************************/

	for (i=0;i<nr_of_atoms;i++) bondlist[i*MXNO_FIELDS_BOND+0]=-1; /*ini list*/

	while ((nr_converted=get_next_bondpairs(h_psf,bondpairs))>0) {

		for (i=0;i<nr_converted;i+=2) {

			atom1=bondpairs[i];
			atom2=bondpairs[i+1];

			if (	(atom1<0) || (atom1>=nr_of_atoms) ||
				(atom2<0) || (atom2>=nr_of_atoms) ) {
				errout("Fatal error: Illegal atomnumber in PSF-file.\n");
				goto error_in_bondlist;	
			}

			res1=add_to_bondlist(bondlist+atom1*MXNO_FIELDS_BOND,atom2);
			res2=add_to_bondlist(bondlist+atom2*MXNO_FIELDS_BOND,atom1);
			
			if ( (res1==-1)||(res2==-1) ) goto error_in_bondlist;
				
			if (++nr_read_bonds>nr_of_bonds) {
				errout("Fatal error: Incompatible number (to much !) of bonds in PSF-file.\n");
				goto error_in_bondlist;
			}
		} /******* for next: for each pair in line */

	} /******* while: for each line in bond-pair-section */

	if (nr_converted<0)	goto error_in_bondlist;

	if (nr_read_bonds!=nr_of_bonds) {
		errout("Fatal error: Incompatible number (to less !) of bonds in PSF-file.\n");
		goto error_in_bondlist;		
	}
	
	txtout("Read bonds in PSF-file     : %d\n",nr_read_bonds);
	
	fclose(h_psf);
	return 0;
	
error_in_bondlist:
	errout("Error after %d read bonds in PSF-file.\n",nr_read_bonds);
	fclose(h_psf);
	return -1;
}

/********************************************************************************/

int	free_bondlist(void)
{
	if (buffer_here) {
		myfree(bondlist);
		buffer_here=0;
	}
	return 0;
}

/*********************************************************************************/

static	int	rexinpcounter=0;

int	writeout_bondlist(char *bondlistfn)
{
	int	i,j,tmp;
	FILE	*h_bondl;

	if ((h_bondl=fopen(bondlistfn,"w"))==NULL) {
		errout("Fatal error: Can not open bond-list-file.\n");
		return -1;
	}

	for (i=0;i<nr_of_atoms;i++) {			/***** for each atom *****/
		fprintf(h_bondl,"Atom  %4d  bonds to  ",i);

		for (j=0;j<MXNO_FIELDS_BOND;j++) {	/***** for each bond *****/
			tmp=bondlist[i*MXNO_FIELDS_BOND+j];

			if (tmp!=-1)	fprintf(h_bondl," %4d ",tmp);
			else		fprintf(h_bondl," %4d\n",tmp);

			if (tmp==-1) break;	/* exit inner loop: for each bond */
		}
	}
	fclose(h_bondl);

	if (nr_of_atoms>100) return 0;

	/******** for the rexinp file ***************/
	if ((h_bondl=fopen("rexinp.1","w"))==NULL) {
		errout("Fatal error: Can not open rexinp-file.\n");
		return -1;
	}
	fprintf(h_bondl,"# molecule name\n");
	fprintf(h_bondl,"# definitions of internal coordinates\n");	
	fprintf(h_bondl,"#   >>>>>>>>  DIMENSION OF INTERNAL COORDINATE SPACE IS       !  <<<<<<<<\n");
	fprintf(h_bondl,"#   type  index list        coflc          cnorm          c5ring     no.  fdiag\n");	
	fprintf(h_bondl,"# stre\n");
	rexinpcounter=0;	
	
	/********** write intern stretch *******************/
	for (i=0;i<nr_of_atoms;i++) {
		for (j=0;j<MXNO_FIELDS_BOND;j++) {	/***** for each bond *****/
			char str1[10],str2[10];
			tmp=bondlist[i*MXNO_FIELDS_BOND+j];
			if (tmp==-1) 	break;		/* exit inner loop: for each bond */
			if (tmp<i) 	continue;

			rexinpcounter++;

			get_atomtype_str(get_atti_by_atomnr(i),str1);
			get_atomtype_str(get_atti_by_atomnr(tmp),str2);
			fprintf(h_bondl,"# Stre %c%d - %c%d\n",str1[0],i+1,str2[0],tmp+1);
			fprintf(h_bondl,"k   stre %2d %2d  0  0     0.00000000     0.00000000     0.00000000    %2d         \n",
			        i+1,tmp+1,rexinpcounter);
		}
	}
	
	/***** write intern bend **********/
	for (i=0;i<nr_of_atoms;i++) {
		int    atti2,atti1 = get_atti_by_atomnr(i);
		double mass2,mass1 = get_atom_mass(atti1);
		char   str1[10];
		int numbonds=0,numh=0,numc=0,numo=0;
		int noth[10];
		int ish[10],iso[10];
				
		if (mass1<3.0)	continue;
		get_atomtype_str(atti1,str1);
		
		for (j=0,numbonds=0,numh=0,numc=0;j<MXNO_FIELDS_BOND;j++) {	/***** for each bond *****/
			tmp=bondlist[i*MXNO_FIELDS_BOND+j];
			if (tmp==-1) 	break;				/* exit inner loop: for each bond */
			numbonds++;
			atti2 = get_atti_by_atomnr(tmp);
			mass2 = get_atom_mass(atti2);		
			if (mass2<3.0)	{ish[numh] =tmp+1; numh++;}
			else		{noth[numc]=tmp+1; numc++;}

			if ((mass2>15.5)&&(mass2<16.5)) {iso[numo] = tmp+1; numo++;}
		}
		
		/**** methyl ******/		
		if ( (str1[0]=='C') && (numbonds==4) && (numh>=3) ) {
			rexinpcounter++;
			fprintf(h_bondl,"# methyl %c%d sym. def.\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],ish[1],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        ish[0],ish[2],i+1);			
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        ish[1],ish[2],i+1);	

			if (numh==4) noth[0]=ish[3];	/*** special case methan ***/
			
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[0],noth[0],i+1);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[1],noth[0],i+1);			
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[2],noth[0],i+1);

			rexinpcounter++;			    
			fprintf(h_bondl,"# methyl %c%d asym. def.\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     2.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[1],ish[2],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[1],ish[0],i+1);    
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[2],ish[0],i+1);

			rexinpcounter++;		    
			fprintf(h_bondl,"# methyl %c%d asym. def.*\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[1],ish[0],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[2],ish[0],i+1);    

			rexinpcounter++;		    
			fprintf(h_bondl,"# methyl %c%d rocking\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     2.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],noth[0],i+1,rexinpcounter);		    
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[1],noth[0],i+1);
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[2],noth[0],i+1);						    

			rexinpcounter++;		    
			fprintf(h_bondl,"# methyl %c%d rocking*\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[1],noth[0],i+1,rexinpcounter);		    
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[2],noth[0],i+1);			    
		} else
		/**** methylene sp3 (>CH2) ******/		
		if ( (str1[0]=='C') && (numbonds==4) && (numh==2) ) {
			rexinpcounter++;
			fprintf(h_bondl,"# methylene %c%d CH2 scissoring\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     5.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],ish[1],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        noth[0],noth[1],i+1);			
	
			rexinpcounter++;
			fprintf(h_bondl,"# methylene %c%d CXY scissoring* (To delete if part of ring)\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],ish[1],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0     5.00000000     0.00000000     0.00000000               \n",
			        noth[0],noth[1],i+1);	
			    
			rexinpcounter++;
			fprintf(h_bondl,"# methylene %c%d rocking\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],noth[0],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[1],noth[0],i+1);		    
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        ish[0],noth[1],i+1);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[1],noth[1],i+1);	
			    
			rexinpcounter++;
			fprintf(h_bondl,"# methylene %c%d wagging\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],noth[0],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        ish[1],noth[0],i+1);		    
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[0],noth[1],i+1);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[1],noth[1],i+1);	
			    
			rexinpcounter++;
			fprintf(h_bondl,"# methylene %c%d twisting\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],noth[0],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[1],noth[0],i+1);		    
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[0],noth[1],i+1);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        ish[1],noth[1],i+1);			    		
		} else
		/**** methylene sp2 ******/		
		if ( ((str1[0]=='C')||(str1[0]=='N')) && (numbonds==3) && (numh==2) ) {
			rexinpcounter++;
			if (str1[0]=='C') fprintf(h_bondl,"# methylene %c%d sym. def. (To delete if part of ring)\n",str1[0],i+1); else
			if (str1[0]=='N') fprintf(h_bondl,"# amino %c%d scissoring (To delete if part of ring)\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     2.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],ish[1],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[0],noth[0],i+1);			
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[1],noth[0],i+1);		

			rexinpcounter++;
			if (str1[0]=='C') fprintf(h_bondl,"# methylene %c%d rocking\n",str1[0],i+1); else
			if (str1[0]=='N') fprintf(h_bondl,"# amino %c%d rocking\n",str1[0],i+1);		
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],noth[0],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[1],noth[0],i+1);

			rexinpcounter++;			    
			if (str1[0]=='C') fprintf(h_bondl,"# methylene %c%d X out CH2\n",str1[0],i+1); else
			if (str1[0]=='N') fprintf(h_bondl,"# amino %c%d X out NH2\n",str1[0],i+1);		
			fprintf(h_bondl,"k   outp %2d %2d %2d %2d     1.00000000     0.00000000     0.00000000    %2d         \n",
			        noth[0],ish[0],ish[1],i+1,rexinpcounter);				
			    			    
		} else
		/**** methine sp2 ******/		
		if ( ((str1[0]=='C')||(str1[0]=='N')) && (numbonds==3) && (numh==1) ) {
			rexinpcounter++;
			if (str1[0]=='C') fprintf(h_bondl,"# methine %c%d sp2 rocking\n",str1[0],i+1); else
			if (str1[0]=='N') fprintf(h_bondl,"# imino %c%d sp2 rocking\n",str1[0],i+1);				
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],noth[0],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[0],noth[1],i+1);

			rexinpcounter++;
			if (str1[0]=='C') fprintf(h_bondl,"# methine %c%d XCY def. (To delete if part of ring)\n",str1[0],i+1); else
			if (str1[0]=='N') fprintf(h_bondl,"# imino %c%d XNY def. (To delete if part of ring)\n",str1[0],i+1);				
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     2.00000000     0.00000000     0.00000000    %2d         \n",
			        noth[0],noth[1],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[0],noth[0],i+1);			
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[0],noth[1],i+1);		

			rexinpcounter++;			    
			if (str1[0]=='C') fprintf(h_bondl,"# methine %c%d H out of XCY plane\n",str1[0],i+1); else
			if (str1[0]=='N') fprintf(h_bondl,"# amino %c%d H out of XNY plane\n",str1[0],i+1);		
			fprintf(h_bondl,"k   outp %2d %2d %2d %2d     1.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],noth[0],noth[1],i+1,rexinpcounter);    
		} else
		if (	(str1[0]=='O') && (numbonds==1) ) {
			/*** is ok ***/
		} else
		/**** a >C=O structure ******/		
		if ( (str1[0]=='C') && (numbonds==3) && (numh==0) && (numo==1) ) {
			rexinpcounter++;

			if (noth[0]==iso[0]) noth[0] = noth[2];
			if (noth[1]==iso[0]) noth[1] = noth[2];	
					 
			fprintf(h_bondl,"# C=O %c%d rocking\n",str1[0],i+1);				
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000    %2d         \n",
			        iso[0],noth[0],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        iso[0],noth[1],i+1);

			rexinpcounter++;
			fprintf(h_bondl,"# C=O %c%d XCY def.\n",str1[0],i+1);				
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     2.00000000     0.00000000     0.00000000    %2d         \n",
			        noth[0],noth[1],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        iso[0],noth[0],i+1);			
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        iso[0],noth[1],i+1);		

			rexinpcounter++;			    
			fprintf(h_bondl,"# C=O %c%d O out of XCY plane\n",str1[0],i+1);	
			fprintf(h_bondl,"k   outp %2d %2d %2d %2d     1.00000000     0.00000000     0.00000000    %2d         \n",
			        iso[0],noth[0],noth[1],i+1,rexinpcounter);    
		} else
		/**** methine sp3 ******/		
		if ( (str1[0]=='C') && (numbonds==4) && (numh==1) ) {
			rexinpcounter++;
			fprintf(h_bondl,"# methine %c%d sp3 rocking\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     2.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],noth[0],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[0],noth[1],i+1);
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[0],noth[2],i+1);

			rexinpcounter++;
			fprintf(h_bondl,"# methine %c%d sp3 rocking*\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000    %2d         \n",
			        ish[0],noth[1],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0    -1.00000000     0.00000000     0.00000000               \n",
			        ish[0],noth[2],i+1);

			rexinpcounter++;
			fprintf(h_bondl,"# methine %c%d XCY def.\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     4.00000000     0.00000000     0.00000000    %2d         \n",
			        noth[0],noth[1],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        noth[0],noth[2],i+1);			
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        noth[1],noth[2],i+1);		

			rexinpcounter++;
			fprintf(h_bondl,"# methine %c%d XCZ def.\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     4.00000000     0.00000000     0.00000000    %2d         \n",
			        noth[0],noth[2],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        noth[0],noth[1],i+1);			
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        noth[1],noth[2],i+1);
			    
			    rexinpcounter++;
			fprintf(h_bondl,"# methine %c%d YCZ def.\n",str1[0],i+1);
			fprintf(h_bondl,"k   bend %2d %2d %2d  0     4.00000000     0.00000000     0.00000000    %2d         \n",
			        noth[1],noth[2],i+1,rexinpcounter);				
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        noth[0],noth[1],i+1);			
			fprintf(h_bondl,"    bend %2d %2d %2d  0     1.00000000     0.00000000     0.00000000               \n",
			        noth[0],noth[2],i+1);
		}  else
		if (str1[0]=='I') { /* We have an ion, therefore no bonds at all */}

		else {
			txtout("Found an unknown structure. Don't know how to make internal\n");	
			txtout("coordinate parameters for that! Make it by hand.\n");
			txtout("This happened for atom %c%d\n",str1[0],i+1);	
		}						
	}
	

	/***** write intern tors **********/
	for (i=0;i<nr_of_atoms;i++) {
		int    atti2,atti1 = get_atti_by_atomnr(i);
		double mass2,mass1 = get_atom_mass(atti1);
		char   str1[10],str2[10];
		int numbonds=0,numh=0,numc=0,numtors=0;
		int noth[10],nextt[10];
		int ish[10];
				
		if (mass1<3.0)	continue;
		get_atomtype_str(atti1,str1);
		
		for (j=0,numbonds=0,numh=0,numc=0,numtors=0;j<MXNO_FIELDS_BOND;j++) {	/*** for each bond ***/
			tmp=bondlist[i*MXNO_FIELDS_BOND+j];
			if (tmp==-1) 	break;		/* exit inner loop: for each bond */
			numbonds++;
			atti2 = get_atti_by_atomnr(tmp);
			mass2 = get_atom_mass(atti2);		
			if (mass2<3.0)	{ish[numh] =tmp+1; numh++;}
			else		{
				noth[numc]=tmp+1; numc++;
				if (tmp>i) {nextt[numtors] = tmp; numtors++;}
			}
		}

		if (numc==0) continue;
		if (numtors==0) continue;		
		
		do {
			int kk,jj;
			int p1[10],p2[10];
			int c1,c2;
			int atti2;
			numtors--;
			
			atti2 = get_atti_by_atomnr(nextt[numtors]);
			get_atomtype_str(atti2,str2);		
	
			for (j=0,c1=0;j<MXNO_FIELDS_BOND;j++) {	/***** for each bond *****/
				tmp=bondlist[i*MXNO_FIELDS_BOND+j];
				if (tmp==-1) 	break;		/* exit inner loop: for each bond */
				if (tmp==nextt[numtors]) continue;
				p1[c1++] = tmp+1;
			}
			if (c1==0) continue;
			
			for (j=0,c2=0;j<MXNO_FIELDS_BOND;j++) {	/***** for each bond *****/
				tmp=bondlist[nextt[numtors]*MXNO_FIELDS_BOND+j];
				if (tmp==-1) 	break;		/* exit inner loop: for each bond */
				if (tmp==i) 	continue;
				p2[c2++] = tmp+1;
			}
			if (c2==0) continue;
			
			for (kk=0;kk<c2;kk++) {
				for (jj=0;jj<c1;jj++) {
					if ((jj==0)&&(kk==0)) {
						rexinpcounter++;
						fprintf(h_bondl,"# Torsion  X - %c%d - %c%d - Y\n",str1[0],i+1,str2[0],nextt[numtors]+1);
						fprintf(h_bondl,"k   tors %2d %2d %2d %2d     1.00000000     0.00000000     0.00000000    %2d         \n",
			        			p1[jj],i+1,nextt[numtors]+1,p2[kk],rexinpcounter);
			        	} else {
			        		fprintf(h_bondl,"    tors %2d %2d %2d %2d     1.00000000     0.00000000     0.00000000               \n",
			        			p1[jj],i+1,nextt[numtors]+1,p2[kk]);	
			        	}
				}	
			}
		} while (numtors>0);
	}

	/***** write template for 5-ring **********/
	fprintf(h_bondl,"# 5-ring deformation (template)\n");
	fprintf(h_bondl,"#k   bend  2  5  1  0     1.00000000     0.00000000     0.00000000    ??         \n");
	fprintf(h_bondl,"#    bend  3  1  2  0    -0.80901699     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  1  4  5  0    -0.80901699     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  4  2  3  0     0.30901699     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  5  3  4  0     0.30901699     0.00000000     0.00000000               \n");

	fprintf(h_bondl,"# 5-ring deformation*\n");
	fprintf(h_bondl,"#k   bend  3  1  2  0    -1.11803399     0.00000000     0.00000000    ??         \n");
	fprintf(h_bondl,"#    bend  1  4  5  0     1.11803399     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  4  2  3  0     1.80901699     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  5  3  4  0    -1.80901699     0.00000000     0.00000000               \n");

	fprintf(h_bondl,"# 5-ring torsion\n");
	fprintf(h_bondl,"#k   tors  2  3  4  5     1.00000000     0.00000000     0.00000000    ??         \n");
	fprintf(h_bondl,"#    tors  1  2  3  4    -0.80901699     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  3  4  5  1    -0.80901699     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  5  1  2  3     0.30901699     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  4  5  1  2     0.30901699     0.00000000     0.00000000               \n");

	fprintf(h_bondl,"# 5-ring torsion*\n");
	fprintf(h_bondl,"#k   tors  1  2  3  4     1.11803399     0.00000000     0.00000000    ??         \n");
	fprintf(h_bondl,"#    tors  3  4  5  1    -1.11803399     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  4  5  1  2     1.80901699     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  5  1  2  3    -1.80901699     0.00000000     0.00000000               \n");

	/***** write template for 6-ring **********/
	fprintf(h_bondl,"# 6-ring trigonal def (template)\n");
	fprintf(h_bondl,"#k   bend  6  2  1  0     1.00000000     0.00000000     0.00000000    ??         \n");
	fprintf(h_bondl,"#    bend  1  3  2  0    -1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  2  4  3  0     1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  3  5  4  0    -1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  4  6  5  0     1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  5  1  6  0    -1.00000000     0.00000000     0.00000000               \n");

	fprintf(h_bondl,"# 6-ring asym def\n");
	fprintf(h_bondl,"#k   bend  6  2  1  0     2.00000000     0.00000000     0.00000000    ??         \n");
	fprintf(h_bondl,"#    bend  1  3  2  0    -1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  2  4  3  0    -1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  3  5  4  0     2.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  4  6  5  0    -1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  5  1  6  0    -1.00000000     0.00000000     0.00000000               \n");

	fprintf(h_bondl,"# 6-ring asym def*\n");
	fprintf(h_bondl,"#k   bend  1  3  2  0     1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  2  4  3  0    -1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  4  6  5  0     1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    bend  5  1  6  0    -1.00000000     0.00000000     0.00000000               \n");

	fprintf(h_bondl,"# 6-ring puckering\n");
	fprintf(h_bondl,"#k   tors  6  1  2  3     1.00000000     0.00000000     0.00000000    ??         \n");
	fprintf(h_bondl,"#    tors  1  2  3  4    -1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  2  3  4  5     1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  3  4  5  6    -1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  4  5  6  1     1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  5  6  1  2    -1.00000000     0.00000000     0.00000000               \n");

	fprintf(h_bondl,"# asym torsion\n");
	fprintf(h_bondl,"#k   tors  6  1  2  3     1.00000000     0.00000000     0.00000000    ??         \n");
	fprintf(h_bondl,"#    tors  2  3  4  5    -1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  3  4  5  6     1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  5  6  1  2    -1.00000000     0.00000000     0.00000000               \n");

	fprintf(h_bondl,"# asym torsion*\n");
	fprintf(h_bondl,"#k   tors  6  1  2  3    -1.00000000     0.00000000     0.00000000    ??         \n");
	fprintf(h_bondl,"#    tors  1  2  3  4     2.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  2  3  4  5    -1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  3  4  5  6    -1.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  4  5  6  1     2.00000000     0.00000000     0.00000000               \n");
	fprintf(h_bondl,"#    tors  5  6  1  2    -1.00000000     0.00000000     0.00000000               \n");

	fclose(h_bondl);

	return 0;
}

/**********************************************************************************/
/**** returns number of bonds for at_nr and fills int_array bondl with the indices*/
/**** of the bonded atoms *********************************************************/

int	get_atom_bondlist(int at_nr,int *bondl)
{
	int j;

	for (j=0;j<MXNO_FIELDS_BOND;j++) {
		if ((bondl[j]=bondlist[at_nr*MXNO_FIELDS_BOND+j])==-1) {
			return j;	/* number of bonds */
		}
	}
	return 0;
}


/*********************************************************************************/
/********* some static functions *************************************************/
/*********************************************************************************/

static
int	add_to_bondlist(int *curbondlist,int atom_nr)
{
	int	i,tmp;
	int	nr_in_list=0;

	for (i=0;i<MXNO_FIELDS_BOND;i++) {
		if (curbondlist[i]==-1) {	/* add atom to end of list */
			curbondlist[i]=atom_nr;
			curbondlist[i+1]=-1;
			nr_in_list=i+1;	
			break;
		} else
		if (curbondlist[i]==atom_nr) {	/* exit if already in list */
			return 0;
		}
	}

	if (nr_in_list==1)	return 1;	/* first atom, no sorting needed */
	
	if (nr_in_list>MXNO_FIELDS_BOND) {	/* check overflow */
		errout("Too many bonds for one atom (MXNO_FIELDS_BOND is %d). Change it in ego.h\n",
			MXNO_FIELDS_BOND);
		return -1;
	}

	for (i=nr_in_list-1;i>0;i--) {		/* sort atom-nr in ascending order */
		if (curbondlist[i]<curbondlist[i-1]) {	/* swap it */
			tmp=curbondlist[i-1];
			curbondlist[i-1]=curbondlist[i];
			curbondlist[i]=tmp;
		} else 	break;			/* now sorted to right position */
	}
	return 1;
}

/********************************************************************************************/

static
int	get_next_bondpairs(FILE *h_psf,int *bondpair)	/* returns number of integers */
{							/* in bondpair ==> even       */
	static	char	linestr[MAX_STRING_SIZE];
	int 	i;
	int	nr_converted;

	if (read_string(h_psf,linestr)==-1) {
		errout("Fatal error: Can not read bonds in psf-file !\n");
		return -1; /* error : can not read at all*/
	}

	nr_converted=sscanf(linestr,"%d %d %d %d %d %d %d %d",
				bondpair+0,bondpair+1,
				bondpair+2,bondpair+3,
				bondpair+4,bondpair+5,
				bondpair+6,bondpair+7);
	if (nr_converted<=0) {		/* nothing converted, may be the end */
		return 0;
	} else
	if (nr_converted%2!=0) {
		errout("Fatal Error: odd converted numbers in PSF-bond-list\n");
		return -1;
	} else {
		for (i=0;i<nr_converted;i++) bondpair[i]--;	/* atomnumbers in ego */
		return nr_converted;				/* start with 0 */
	}
}

/***************************************************************************************/

static
int	goto_psf_bonddata(FILE *h_psf,int *nr_of_bonds)
{
	static	char	lstr[MAX_STRING_SIZE];
	static	char	tmpstr[MAX_STRING_SIZE];
	int 	res;

	while (read_string(h_psf,lstr)!=-1) {			/* read a psf-line */
		strup(lstr);
		if (sscanf(lstr,"%d %s",&res,tmpstr)==2) {	/* check it */
			tmpstr[6]=0;
			if (strcmp(tmpstr,"!NBOND")==0) {
				*nr_of_bonds=res;
				return 0;
			}
		}	
	}

	errout("Fatal Error: No bondlist-data-section (!NBOND) in PSF-file.\n");
	return -1;
}
