/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: wodihe.c	by M.Eichinger, Munich			**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis)**
**	Contents: scan in and write out dihedral-list and parameters.	**
**		  recursive implementation of wild-card-string compare.	**
**									**
**  	Modif's:							**
**		05.06.94						**
**		05.07.94 swapping of dihedral if mirrored order fits	**
**		11.07.94 bug in xplor_strcmp repaired			**
**		16.01.94 now numbers of diehdrals is written to file    **
**		06.04.95 Dihedral-multiplicity dihe_n == 1 fixed	**				
**									**				
*************************************************************************/

#include <math.h>

#include "hardw.h"
#include "ego.h"
#include "woalloc.h"
#include "util_lib.h"
#include "wodecls.h"

#define TEST_DIHEPARA	0

typedef struct {
	int	att[4];			    /*holds the atom-type-indices,-1 if wild-card*/
	char	attstr[4][LEN_ATOM_TYPE+1]; /*holds the atom-type-strings or wild-cards */
					
	double	kd;
	double	d;
	int	n;
	int	multi;
} dihepar_type;

/******************************************************************************/
/**************** functions in this modul *************************************/
/******************************************************************************/

	int	readin_dihepara(int argc,char *argv[]);
	int	writeout_dihelist(char *psffn,char *dihelistfn,int nr_of_atoms);
	int	writeout_dihepara(char *diheparafn);
	int	free_dihebuffers(void);


static	int	get_dihepara_index(int *angletrip);
static	int	allready_in_list(dihepar_type *diheparabuf,int charmm);

static	int	fill_dihe_paralist(FILE *h_para,dihepar_type *diheparabuf);
static	int	match_dihedral(int *att,dihepar_type *par,int wild_cards_allowed);

static	int	add_to_dihelist(int *dihelist,int *dihequads,int nr_read_dihe);

static	int	get_next_dihequads(FILE *h_psf,int *dihequads);
static	int	goto_psf_dihedata(FILE *h_psf,int *nr_of_dihe);

/******************************************************************************/
/********************* some vars **********************************************/
/******************************************************************************/

static	int		nr_dihe_para=0;
static	dihepar_type	*diheparabuf;
static	int		dihe_buffers_here=0;

/**********************************************************************************/

int	readin_dihepara(int argc,char *argv[])
{
	int	file_ind;		/* 3 is the first parameter file */
	FILE 	*h_para;
	
	/****************************** allocate ***********************************/
	diheparabuf=(dihepar_type*)malloc((size_t)MXNO_DIHED_PARAM*sizeof(dihepar_type));
	if (diheparabuf==NULL) {
		errout("Fatal error: Can not allocate space for dihedral-list.\n");
		return -1;
	}	
	nr_dihe_para=0;	

	/******************* scan each parmater-file *********************************/
	for (file_ind=3;file_ind<argc;file_ind++) {
		if ((h_para=fopen(argv[file_ind],"r"))==NULL) {
			errout("Fatal error: Can not read parameter-file: %s\n",argv[file_ind]);
			myfree(diheparabuf);
			return -1;
		}

		if (fill_dihe_paralist(h_para,diheparabuf)==-1) {
			errout("Fatal error: Error occured in parameter-file: %s\n",argv[file_ind]);
			fclose(h_para);
			myfree(diheparabuf);
			return -1;
		}
		
		fclose(h_para);
	}

	dihe_buffers_here=1;
	return 0;
}

/***************************************************************************************/

int	free_dihebuffers(void)
{
	if (dihe_buffers_here) {
		myfree(diheparabuf);
		dihe_buffers_here=0;
	}
	return 0;
}

/***************************************************************************************/

int	writeout_dihelist(char *psffn,char *dihelistfn,int nr_of_atoms)
{
	int	i,res1,off=0;
	int	atom1=-1,atom2=-1,atom3=-1,atom4=-1;
	int	nr_of_dihe=0,nr_read_dihe=0;
	int	dihequad[2*5],nr_converted=0;
	int	dihelistbuf_here=0,m,t,fatalerror=0;
	FILE	*h_psf,*h_dihe;
static	int	*dihelist;

	/*************** open psf-file and allocate space *******************************/	
	if ((h_psf=fopen(psffn,"r"))==NULL) {
		errout("Fatal error: Can not read psf-file <%s>\n",psffn);
		return -1;
	}
	if (goto_psf_dihedata(h_psf,&nr_of_dihe)==-1) {
		fclose(h_psf);
		goto write_dihe_list;			/* no dihelist-section ? */
	}

	dihelist=alloc_ints(3*(nr_of_dihe+10)*5);
	dihelistbuf_here=1;

	/******************** build up sorted dihedrallist ******************************/

	while ((nr_converted=get_next_dihequads(h_psf,dihequad))>0) {
		for (i=0;i<nr_converted;i+=5) {
			if (	(atom1==dihequad[i  ]) &&	/*** Multi dihedrals ***/
				(atom2==dihequad[i+1]) &&	/*** in xplor        ***/
				(atom3==dihequad[i+2]) &&
				(atom4==dihequad[i+3]) ) {
				continue;
			}	
			atom1=dihequad[i  ];		/* just for better reading */
			atom2=dihequad[i+1];
			atom3=dihequad[i+2];
			atom4=dihequad[i+3];
							/* some tests 	*/		
			if (	(atom1<0) || (atom1>=nr_of_atoms) ||
				(atom2<0) || (atom2>=nr_of_atoms) ||
				(atom3<0) || (atom3>=nr_of_atoms) ||
				(atom4<0) || (atom4>=nr_of_atoms) ) {

				errout("Fatal error: Illegal atomnumber in PSF-file.\n");
				goto error_in_dihelist;	
			}		

			/******* build up the dihe-PARAMTER-list and get the index ************/
			
			if ((dihequad[i+4]=get_dihepara_index(dihequad+i))==-1) {
				char s1[20],s2[20],s3[20],s4[20];
				errout("Fatal error: Unknown dihedral-quadripple in psf-file.\n");
				errout("PDB atom-numbers:%4d %4d %4d %4d\n",
					atom1+1,atom2+1,atom3+1,atom4+1);
				
				get_atomtype_str(get_atti_by_atomnr(atom1),s1);
				get_atomtype_str(get_atti_by_atomnr(atom2),s2);
				get_atomtype_str(get_atti_by_atomnr(atom3),s3);
				get_atomtype_str(get_atti_by_atomnr(atom4),s4);
				errout("Atom-type-strings <%s> <%s> <%s> <%s>\n",s1,s2,s3,s4);
				fatalerror=1;			
			}
				
			/**** now dihequad[i+4] holds the index to the dihe-PARAMETER-list */
			/**** save the dihequad & the index too the dihe-PARMETER-list ****/		
			t=diheparabuf[dihequad[i+4]].multi;	
			for (m=0;m<t;m++) {
				res1=add_to_dihelist(dihelist,dihequad+i,nr_read_dihe);
				if (res1==-1) goto error_in_dihelist;
				if (++nr_read_dihe>3*nr_of_dihe) {
					errout("Fatal error: Too much dihedrals in PSF-file.\n");
					goto error_in_dihelist;		
				}
				dihequad[i+4]++;
			}
		} /******* for next: for each dihe-quad in line */

	} /******* while: for each line in dihe-section */
	fclose(h_psf);
	
	if (nr_converted<0)	goto error_in_dihelist;
	if (fatalerror)		return -1;

	if (nr_read_dihe!=nr_of_dihe) {
		errout("*****************************************************************************\n");
		errout("Warning error: Incompatible number of dihedrals in PSF-file.\n");
		errout("Nr_read_dihedrals %d, expected dihedrals %d\n",nr_read_dihe,nr_of_dihe);
		errout("This may be caused by a wrong setup of the multi-dihedral list in the PSF-file.\n");		
		errout("Don't worry! It is very likely that EGO can handle this situation.\n");
		errout("But for sure crosscheck energy-printouts of EGO with XPLOR or CHARMM.\n");		
		errout("*****************************************************************************\n");
	}
		
	/******************** write out dihedral-list to file ****************************/
write_dihe_list:	
	if ((h_dihe=fopen(dihelistfn,"w"))==NULL) {
		errout("Fatal error: Can not open dihedral-list-file <%s>.\n",dihelistfn);
		goto error_in_dihelist;
	}
	fprintf(h_dihe,"%d  Number of dihedrals.\n",nr_read_dihe);
	fprintf(h_dihe,"Format of the DIHEDRAL list:\n");
	fprintf(h_dihe,"Number   1. atom  2. atom  3. atom   4. atom    index\n");

	for (i=0,off=0;i<nr_read_dihe;i++,off+=5) {		/***** for each dihe *****/
		fprintf(h_dihe,"%4d      %4d     %4d     %4d     %4d      %4d\n",
			i,
			dihelist[off+0],dihelist[off+1],dihelist[off+2],dihelist[off+3],
			dihelist[off+4]);
	}

	txtout("Read dihedrals in PSF-file   : %d\n",nr_read_dihe);

	if (dihelistbuf_here)	myfree(dihelist);
	fclose(h_dihe);

	return 0;

error_in_dihelist:
	errout("Error after %d read dihedrals.\n",nr_read_dihe);
	if (dihelistbuf_here)	myfree(dihelist);
	fclose(h_psf);
	return -1;
}

/**********************************************************************************/

int	writeout_dihepara(char *diheparafn)
{
	int	i;
	FILE 	*h_dihep;

	if ((h_dihep=fopen(diheparafn,"w"))==NULL) {
		errout("Can not open dihedral-parameter-file.\n");
		return -1;
	}

	fprintf(h_dihep,"This is the DIHEDRAL parameter list.\n");
	fprintf(h_dihep,"Number of parameter sets:   %4d .\n",nr_dihe_para);
	fprintf(h_dihep,"Format:\n");
	fprintf(h_dihep,"    Index    Kd              d           n\n");

	for (i=0;i<nr_dihe_para;i++) {
		fprintf(h_dihep," %4d    %10.8g     %10.8g      %4d\n",
				i,
				diheparabuf[i].kd,
				diheparabuf[i].d,
				diheparabuf[i].n);
	}
	fclose(h_dihep);

	txtout("Written dihedral-parameters  : %d\n",nr_dihe_para);

	return 0;
}

/**************************************************************************************/
/*** Xplor try's to match first DIHE-Params not containing wild-cards, and after that */
/*** Xplor try's to match with DIHE-Params containing wild_cards **********************/

static
int	get_dihepara_index(int *dihequad)
{
	int	i,res,swap;
	int	att[4];	
	int	wild_cards_allowed;
	
	if (nr_dihe_para==0)	return -1;		/* nothing at all in list */
							/* that should not be !!! */
	att[0]=get_atti_by_atomnr(dihequad[0]);
	att[1]=get_atti_by_atomnr(dihequad[1]);
	att[2]=get_atti_by_atomnr(dihequad[2]);
	att[3]=get_atti_by_atomnr(dihequad[3]);

	/**** first try match with not wild-cards allowed ********/
	wild_cards_allowed=0;
	for (i=0;i<nr_dihe_para;i++) {
		res=match_dihedral(att,&(diheparabuf[i]),wild_cards_allowed);
		if (res==1)	return i;
		if (res==2) 	goto ok_but_swap_it;

		if (res==-1)	return -1;	/* thats an error */
	}

	
	/**** then try match with wild-cards allowed ********/
	/** a la Xplor wildcards are allowed in this way: * a b * but that don't matter here ***/
	wild_cards_allowed=1;
	for (i=0;i<nr_dihe_para;i++) {
		res=match_dihedral(att,&(diheparabuf[i]),wild_cards_allowed);
		if (res==1)	return i;
		if (res==2) 	goto ok_but_swap_it;

		if (res==-1)	return -1;	/* thats an error */
	}

	return -1; 				/* unknown dihe-quad, thats an error ! */

ok_but_swap_it:

	swap=dihequad[3];dihequad[3]=dihequad[0];dihequad[0]=swap;
	swap=dihequad[2];dihequad[2]=dihequad[1];dihequad[1]=swap;
	return i;
}

/***********************************************************************************/

static
int	match_dihedral(int *att,dihepar_type *par,int wild_cards_allowed)
{
	int 	i,res;
	int	matches;
	static	char	tmpstr[MAX_STRING_SIZE];
	static	int	swapatt[4];
				/*********** first try normal order ****************/
	matches=0;	
	for (i=0;i<4;i++) {
		if (par->att[i]!=-1) {
			if (par->att[i]==att[i])	matches++;
		} else {
			get_atomtype_str(att[i],tmpstr);
			res=xplor_strcmp(tmpstr,par->attstr[i],wild_cards_allowed);
			if (res==1)	matches++;
			if (res==-1) 	return -1;
		}
	}
	if (matches==4)	return 1;	/* all ok, no swapping needed */

				/********* if not matching, try mirror-order ********/
	swapatt[0]=att[3];	swapatt[3]=att[0];
	swapatt[1]=att[2];	swapatt[2]=att[1];

	matches=0;
	for (i=0;i<4;i++) {
		if (par->att[i]!=-1) {
			if (par->att[i]==swapatt[i])	matches++;
		} else {
			get_atomtype_str(swapatt[i],tmpstr);
			res=xplor_strcmp(tmpstr,par->attstr[i],wild_cards_allowed);
			if (res==1)	matches++;
			if (res==-1)	return -1;
		}
	}
	if (matches==4)	return 2;	/* all ok, swapping needed */

	return 0;
}

/***********************************************************************************/

static	
int	fill_dihe_paralist(FILE *h_para,dihepar_type *diheparabuf)
{
	double 	dihe_kd[3],dihe_d[3];
	int	dihe_n[3],i;
	static	char	at1str[LEN_ATOM_TYPE+1];
	static	char	at2str[LEN_ATOM_TYPE+1];
	static	char	at3str[LEN_ATOM_TYPE+1];
	static	char	at4str[LEN_ATOM_TYPE+1];
	int 	res,res1,charmm=0;

	while ((res=read_next_dihepara(h_para,
			at1str,at2str,at3str,at4str,
			dihe_kd,dihe_d,dihe_n))>=0) {

		strcpy(diheparabuf[nr_dihe_para].attstr[0],at1str);
		strcpy(diheparabuf[nr_dihe_para].attstr[1],at2str);
		strcpy(diheparabuf[nr_dihe_para].attstr[2],at3str);
		strcpy(diheparabuf[nr_dihe_para].attstr[3],at4str);

		/**** replaces 'X' by '*' because 'X' acts here as a wild-card *********/
		for (i=0;i<4;i++) {
			if (	(strlen(diheparabuf[nr_dihe_para].attstr[i])==1) &&
				(diheparabuf[nr_dihe_para].attstr[i][0]=='X') ) {
				diheparabuf[nr_dihe_para].attstr[i][0]='*';
			}
		}

		if (res>=100) {
			res -= 100; 
			charmm=1;
		} else	charmm=0;
		
		for (i=0;i<res;i++) {
			if (i>0) {
				strcpy(diheparabuf[nr_dihe_para].attstr[0],diheparabuf[nr_dihe_para-1].attstr[0]);
				strcpy(diheparabuf[nr_dihe_para].attstr[1],diheparabuf[nr_dihe_para-1].attstr[1]);
				strcpy(diheparabuf[nr_dihe_para].attstr[2],diheparabuf[nr_dihe_para-1].attstr[2]);
				strcpy(diheparabuf[nr_dihe_para].attstr[3],diheparabuf[nr_dihe_para-1].attstr[3]);
			}
			diheparabuf[nr_dihe_para].att[0]	=get_atomtype_index(at1str);
			diheparabuf[nr_dihe_para].att[1]	=get_atomtype_index(at2str);
			diheparabuf[nr_dihe_para].att[2]	=get_atomtype_index(at3str);
			diheparabuf[nr_dihe_para].att[3]	=get_atomtype_index(at4str);
		
			diheparabuf[nr_dihe_para].kd	=dihe_kd[i];
			diheparabuf[nr_dihe_para].d	=dihe_d[i]/180.0*3.1415926;
			diheparabuf[nr_dihe_para].multi =res;
			if (!((dihe_n[i]==2) || (dihe_n[i]==3) || (dihe_n[i]==4) || (dihe_n[i]==6))){
                  		dihe_n[i]=dihe_n[i] + 50;	/* This is a flag to go into the
								general dihedral routine */
			}

			if (dihe_d[i]==180.0) {		/* Most Sig. Byte indicates 180 degree */
				diheparabuf[nr_dihe_para].n	=dihe_n[i]|0x80;
			} else {
				diheparabuf[nr_dihe_para].n	=dihe_n[i];
			}
		
#if TEST_DIHEPARA
		printf("TEST: Dihe-Params %d: <%s>  <%s>  <%s>  <%s>  %lf  %lf  %d multi %d\n",
			nr_dihe_para,
			diheparabuf[nr_dihe_para].attstr[0],diheparabuf[nr_dihe_para].attstr[1],
			diheparabuf[nr_dihe_para].attstr[2],
			diheparabuf[nr_dihe_para].attstr[3],
			diheparabuf[nr_dihe_para].kd,
			diheparabuf[nr_dihe_para].d,
			diheparabuf[nr_dihe_para].n,
			diheparabuf[nr_dihe_para].multi);
#endif
			if (i==0) {
				res1=allready_in_list(diheparabuf,charmm);/* last added element is checked */
									  /* against all others entries.   */
				if (res1==1)	continue;		  /* allready in list              */
			}
							/* ok, keep last parameters */
			if (++nr_dihe_para>=MXNO_DIHED_PARAM) {
				errout(
				"Fatal error: Too many dihedral-parameters (MXNO_DIHED_PARAM is %d)\n",MXNO_DIHED_PARAM);
				return -1;
			}
		}
	}
	if (res==-2)	return -1;		/* thats an error */

	return 0;
}

/**************************************************************************************/
/***** last added element is checked against all others *****/

static
int	allready_in_list(dihepar_type *diheparabuf,int charmm)
{
	int	i;

	for (i=0;i<nr_dihe_para;i++) {
	     if (((strcmp(diheparabuf[i].attstr[0],diheparabuf[nr_dihe_para].attstr[0])==0) &&
		  (strcmp(diheparabuf[i].attstr[1],diheparabuf[nr_dihe_para].attstr[1])==0) &&
		  (strcmp(diheparabuf[i].attstr[2],diheparabuf[nr_dihe_para].attstr[2])==0) &&
		  (strcmp(diheparabuf[i].attstr[3],diheparabuf[nr_dihe_para].attstr[3])==0))||
		 						/* swap */
		 ((strcmp(diheparabuf[i].attstr[0],diheparabuf[nr_dihe_para].attstr[3])==0) &&
		  (strcmp(diheparabuf[i].attstr[1],diheparabuf[nr_dihe_para].attstr[2])==0) &&
		  (strcmp(diheparabuf[i].attstr[2],diheparabuf[nr_dihe_para].attstr[1])==0) &&
		  (strcmp(diheparabuf[i].attstr[3],diheparabuf[nr_dihe_para].attstr[0])==0))
		)  {

		    if ((fabs(diheparabuf[i].kd-diheparabuf[nr_dihe_para].kd)<SMALL) &&
		 	(fabs(diheparabuf[i].d -diheparabuf[nr_dihe_para].d )<SMALL) &&
			(diheparabuf[i].n==diheparabuf[nr_dihe_para].n)             ) {

			return 1;	/* thats ok , totally identic, so skip it */
		    } else {
		    	
		    	if (charmm) {
		    		
		    		if (diheparabuf[i].multi==1)
		    		txtout("\nINFO: CHARMM-Multidihedral for <%s> <%s> <%s> <%s>:  (%f, %f, %d)\n",
		    		diheparabuf[i].attstr[0],
				diheparabuf[i].attstr[1],
				diheparabuf[i].attstr[2],
				diheparabuf[i].attstr[3],
				diheparabuf[i].kd,diheparabuf[i].d,diheparabuf[i].n);
			
				txtout("INFO: CHARMM-Multidihedral for <%s> <%s> <%s> <%s>:  (%f, %f, %d)\n",
		    		diheparabuf[nr_dihe_para].attstr[0],
				diheparabuf[nr_dihe_para].attstr[1],
				diheparabuf[nr_dihe_para].attstr[2],
				diheparabuf[nr_dihe_para].attstr[3],
				diheparabuf[nr_dihe_para].kd,diheparabuf[nr_dihe_para].d,diheparabuf[nr_dihe_para].n);
			
		    		diheparabuf[i].multi++;
		    		diheparabuf[nr_dihe_para].multi=diheparabuf[i].multi;
		    		
				return 0;
		    	} else {
				errout(
				"Warning error: Inconsistent dihedral-params for atomtypes <%s> <%s> <%s> <%s>\n",
				diheparabuf[i].attstr[0],
				diheparabuf[i].attstr[1],
				diheparabuf[i].attstr[2],
				diheparabuf[i].attstr[3]);

				errout("old vals (kd,d,n) (%f, %f, %d) skipped vals: (%f, %f, %d)\n",
				diheparabuf[i].kd,diheparabuf[i].d,diheparabuf[i].n,
				diheparabuf[nr_dihe_para].kd,diheparabuf[nr_dihe_para].d,
				diheparabuf[nr_dihe_para].n);
				errout("Skipping new dihedral-parameters !\n");

				return 1;	/* strings match, but not the values ==> inconsitency */
						/* skip this too */
			}
		    }
	    }
	}
	return 0;			/* not in list */
}	

/*************************************************************************************/

static
int	add_to_dihelist(int *dihelist,int *dihequad,int nr_read_dihe)
{
	int	i;
	int 	off;
	int	swapbuf[5];

	off=5*nr_read_dihe;	
						/* add to end of (sorted) list */
	dihelist[off+0]=dihequad[0];
	dihelist[off+1]=dihequad[1];
	dihelist[off+2]=dihequad[2];
	dihelist[off+3]=dihequad[3];
	dihelist[off+4]=dihequad[4];

	if (nr_read_dihe==0)	return 1;	/* first quad, no sorting */

						/* sort new entry in ascending order */
						/* of the middle atom (dihequad[1]) */
	for (i=nr_read_dihe,off=5*nr_read_dihe;i>0;i--,off-=5) {
		if (dihelist[off+1]<dihelist[off-5+1]) {
			memcpy(swapbuf,		dihelist+off-5,	5*sizeof(int));
			memcpy(dihelist+off-5,	dihelist+off,	5*sizeof(int));
			memcpy(dihelist+off,	swapbuf,	5*sizeof(int));
		} else	break;			/* now sorted to right place */
	}	
	return 1;
}

/**************************************************************************************/

static
int	get_next_dihequads(FILE *h_psf,int *dihequads)		/* returns number of integers */
{								/* in dihe ==> even           */
	static	char	linestr[MAX_STRING_SIZE];
	int 	i,j;
	int	nr_converted;

	if (read_string(h_psf,linestr)==-1)	return -1;	/* error : can not read at all */

	nr_converted=sscanf(linestr,"%d %d %d %d %d %d %d %d",
				dihequads+0,dihequads+1,dihequads+2,dihequads+3,
				dihequads+5,dihequads+6,dihequads+7,dihequads+8
				);
	
	if (nr_converted<=0) {		/* nothing converted, may be the end */
		return 0;
	} else
	if (nr_converted%4!=0) {
		errout("Fatal error: odd converted numbers in PSF-dihedral-list.\n");
		return -1;
	} else {
		for (i=0,j=0;i<nr_converted/4;i++,j+=5) {
			dihequads[j  ]--;			/* atomnumbers in ego */
			dihequads[j+1]--;			/* start with 0 */
			dihequads[j+2]--;
			dihequads[j+3]--;
		}
		dihequads[4 ]=0;
		dihequads[9 ]=0;	
		return (nr_converted/4)*5;			
	}
}

/**************************************************************************************/

static
int	goto_psf_dihedata(FILE *h_psf,int *nr_of_dihe)
{
	static	char	lstr[MAX_STRING_SIZE];
	static	char	tmpstr[MAX_STRING_SIZE];
	int 	res;

	while (read_string(h_psf,lstr)!=-1) {			/* read a psf-line */
		strup(lstr);
		if (sscanf(lstr,"%d %s",&res,tmpstr)==2) {	/* check it */
			tmpstr[5]=0;
			if (strcmp(tmpstr,"!NPHI")==0) {
				*nr_of_dihe=res;
				return 0;
			}
		}	
	}
	
	*nr_of_dihe=0;
	errout("Warning error: No dihedral-list-data-section (!NPHI) in PSF-file.\n");
	return -1;
}


