/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: woimpr.c	by M.Eichinger, Munich			**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis)**
**	Contents: scan in and write out improper-list and parameters.	**
**									**
**  	Modif's:							**
**		05.06.94						**
**		05.07.94 swapping of impropers if mirrored order fits	**
*************************************************************************/

#include <math.h>

#include "hardw.h"
#include "ego.h"
#include "woalloc.h"
#include "util_lib.h"
#include "wodecls.h"


/***** still to do: MULTI impropers,now: ==> warning, if occured but not used  
       or fatal error, if used *******/  

#define TEST_IMPRPARA	0

typedef struct {
	int	att[4];				/*holds the atom-type-indices,-1 if wild-card*/
	char	attstr[4][LEN_ATOM_TYPE+1];	/*holds the atom-type-strings or wild-cards*/
					
	double	ki;
	double	omega0;
} imprpar_type;

/******************************************************************************/
/**************** functions in this modul *************************************/
/******************************************************************************/

	int	readin_imprpara(int argc,char *argv[]);
	int	writeout_imprlist(char *psffn,char *imprlistfn,int nr_of_atoms);
	int	writeout_imprpara(char *imprparafn);
	int	free_imprbuffers(void);

static	int	get_imprpara_index(int *imprquad);
static	int	allready_in_list(imprpar_type *imprparabuf);

static	int	fill_impr_paralist(FILE *h_para,imprpar_type *imprparabuf);
static	int	match_imprdral(int *att,imprpar_type *par,int *wild_cards_allowed);

static	int	add_to_imprlist(int *imprlist,int *imprquads,int nr_read_impr);

static	int	get_next_imprquads(FILE *h_psf,int *imprquads);
static	int	goto_psf_imprdata(FILE *h_psf,int *nr_of_impr);

/******************************************************************************/
/********************* some vars **********************************************/
/******************************************************************************/

static	int		nr_impr_para=0;
static	imprpar_type	*imprparabuf;

static	int		impr_buffers_here=0;

/**********************************************************************************/

int	readin_imprpara(int argc,char *argv[])
{
	int	file_ind;		/* 3 is the first parameter file */
	FILE 	*h_para;
	
	/****************************** allocate ***********************************/
	imprparabuf=(imprpar_type*) malloc((size_t)MXNO_IMPROP_PARAM*sizeof(imprpar_type));
	if (imprparabuf==NULL) {
		errout("Fatal error: Can not allocate space for improper-list.\n");
		return -1;
	}	
	nr_impr_para=0;	

	/******************* scan each parmater-file *********************************/
	for (file_ind=3;file_ind<argc;file_ind++) {
		if ((h_para=fopen(argv[file_ind],"r"))==NULL) {
			errout("Fatal error: Can not read parameter-file: <%s>\n",argv[file_ind]);
			myfree(imprparabuf);
			return -1;
		}

		if (fill_impr_paralist(h_para,imprparabuf)==-1) {
			errout("Fatal error:  Error occured in parameter-file: <%s>\n",argv[file_ind]);
			fclose(h_para);
			myfree(imprparabuf);
			return -1;
		}
		fclose(h_para);
	}

	impr_buffers_here=1;
	return 0;
}

/****************************************************************************/

int	free_imprbuffers(void)
{
	if (impr_buffers_here) {
		myfree(imprparabuf);
		impr_buffers_here=0;
	}
	return 0;
}

/*****************************************************************************/

int	writeout_imprlist(char *psffn,char *imprlistfn,int nr_of_atoms)
{
	int	i,res1,off=0;
static	int	*imprlist;
	int	atom1,atom2,atom3,atom4;
	int	nr_of_impr=0,nr_read_impr=0;
	int	imprquad[2*5],nr_converted=0;
	int	imprlistbuf_here=0,fatalerror=0;
	FILE	*h_psf,*h_impr;

	/*************** open psf-file and allocate space ********************/	
	if ((h_psf=fopen(psffn,"r"))==NULL) {
		errout("Fatal error: Can not read psf-file <%s>.\n",psffn);
		return -1;
	}

	if (goto_psf_imprdata(h_psf,&nr_of_impr)==-1) {
                fclose(h_psf);
		goto write_impr_listfile;	/* no imprlist-section */
	}
	
	imprlist=alloc_ints((nr_of_impr+10)*5);
	imprlistbuf_here=1;
	
	/******************** build up sorted imprdrallist *******************/

	while ((nr_converted=get_next_imprquads(h_psf,imprquad))>0) {
		for (i=0;i<nr_converted;i+=5) {
	
			atom1=imprquad[i  ];		/* just for better reading */
			atom2=imprquad[i+1];
			atom3=imprquad[i+2];
			atom4=imprquad[i+3];
							/* some tests 	*/		
			if (	(atom1<0) || (atom1>=nr_of_atoms) ||
				(atom2<0) || (atom2>=nr_of_atoms) ||
				(atom3<0) || (atom3>=nr_of_atoms) ||
				(atom4<0) || (atom4>=nr_of_atoms) ) {

				errout("Fatal error: Illegal atomnumber (%d %d %d %d) in PSF-file.\n",
					atom1,atom2,atom3,atom4);
				goto error_in_imprlist;	
			}		

			if ((imprquad[i+4]=get_imprpara_index(imprquad+i))<0) {
				errout("Fatal error: Unknown improper-quadripple in psf-file.\n");
				errout("atom-numbers:%4d %4d %4d %4d\n",
					atom1+1,atom2+1,atom3+1,atom4+1);
				{ char s1[20],s2[20],s3[20],s4[20];
				get_atomtype_str(get_atti_by_atomnr(atom1),s1);
				get_atomtype_str(get_atti_by_atomnr(atom2),s2);
				get_atomtype_str(get_atti_by_atomnr(atom3),s3);
				get_atomtype_str(get_atti_by_atomnr(atom4),s4);
				errout("Atom-type-strings <%s> <%s> <%s> <%s>\n",s1,s2,s3,s4);
				}
				fatalerror=1;
			}
				
			/**** now imprquad[i+4] holds the index to the impr-PARAMETER-list */
			/**** save the imprquad & the index too the impr-PARMETER-list ****/		

			res1=add_to_imprlist(imprlist,imprquad+i,nr_read_impr);
						
			if (res1==-1) goto error_in_imprlist;
				
			if (++nr_read_impr>nr_of_impr) {
				errout("Fatal error: Incompatible number (to much !) of impropers in PSF-file.\n");
				goto error_in_imprlist;
			}
		} /******* for next: for each impr-quadripple in line */

	} /******* while: for each line in impr-section */

	fclose(h_psf);

	if (nr_converted<0)	goto error_in_imprlist;
	if (fatalerror) 	return -1;
	
	if (nr_read_impr!=nr_of_impr) {
		errout("Fatal error: Incompatible number (to less !) of impropers in PSF-file.\n");
		goto error_in_imprlist;		
	}
	
	/********************************************************************************/
	/******************** write out improperlist to file ****************************/
write_impr_listfile:	
	if ((h_impr=fopen(imprlistfn,"w"))==NULL) {
		errout("Fatal error: Can not open improper-list-file <%s>.\n",imprlistfn);
		goto error_in_imprlist;
	}

	fprintf(h_impr,"%d  Number of impropers.\n",nr_of_impr);
	fprintf(h_impr,"Format of the IMPROPER list:\n");
	fprintf(h_impr,"Number   1. atom  2. atom  3. atom   4. atom    index\n");

	for (i=0,off=0;i<nr_of_impr;i++,off+=5) {		/***** for each impr *****/
		fprintf(h_impr,"%4d      %4d     %4d     %4d     %4d      %4d\n",
			i,
			imprlist[off+0],imprlist[off+1],imprlist[off+2],imprlist[off+3],
			imprlist[off+4]);
	}

	txtout("Read impropers in PSF-file   : %d\n",nr_read_impr);

	if (imprlistbuf_here)	myfree(imprlist);
	fclose(h_impr);

	return 0;

error_in_imprlist:
	errout("Fatal error: Error after %d read impropers.\n",nr_read_impr);

	if (imprlistbuf_here)	myfree(imprlist);
	fclose(h_psf);

	return -1;
}

/**********************************************************************************************/

int	writeout_imprpara(char *imprparafn)
{
	int	i;
	FILE 	*h_imprp;

	if ((h_imprp=fopen(imprparafn,"w"))==NULL) {
		errout("Fatal error: Can not open improper-parameter-file <%s>.\n",imprparafn);
		return -1;
	}

	fprintf(h_imprp,"This is the IMPROPER parameter list.\n");
	fprintf(h_imprp,"Number of parameter sets:   %4d .\n",nr_impr_para);
	fprintf(h_imprp,"Format:\n");
	fprintf(h_imprp,"    Index    Ki           omega0\n");

	for (i=0;i<nr_impr_para;i++) {
		fprintf(h_imprp," %4d    %10.8g     %10.8g\n",
				i,
				imprparabuf[i].ki,
				imprparabuf[i].omega0);
	}
	fclose(h_imprp);

	txtout("Written improper-parameters  : %d\n",nr_impr_para);

	return 0;
}

/************************************************************************************/
/**************************************************************************************/
/*** Xplor try's to match first IMPR-Params not containing wild-cards, and after that */
/*** Xplor try's to match with IMPR-Params containing wild_cards in a certain order.***/
/*** see Xplor manual IMPRoper_parameter section for it (Parameter Statement 3.2)   ***/

static
int	get_imprpara_index(int *imprquad)
{
	int	i,res,swap;
	int	wild_cards_allowed[4];	/* array for signals of allowed wild-cards */
					/* 1=wc allowed, 0=wc not allowed */
	int	att[4];	
	
	if (nr_impr_para==0)	return -1;		/* nothing at all in list */
							/* that should not be !!! */
	att[0]=get_atti_by_atomnr(imprquad[0]);
	att[1]=get_atti_by_atomnr(imprquad[1]);
	att[2]=get_atti_by_atomnr(imprquad[2]);
	att[3]=get_atti_by_atomnr(imprquad[3]);

		/**** first try match with not wild-cards allowed ********/
	wild_cards_allowed[0]=0;
	wild_cards_allowed[1]=0;
	wild_cards_allowed[2]=0;
	wild_cards_allowed[3]=0;
	for (i=0;i<nr_impr_para;i++) {
		res=match_imprdral(att,&(imprparabuf[i]),wild_cards_allowed);
		if (res==1)	return i;
		if (res==2) 	goto ok_but_swap_it;
					
		if (res==-1)	return -1;	/* thats an error */
	}
		/**** then try match with wild-cards allowed ********/

	wild_cards_allowed[0]=0;
	wild_cards_allowed[1]=1;
	wild_cards_allowed[2]=1;
	wild_cards_allowed[3]=0;
	for (i=0;i<nr_impr_para;i++) {
		res=match_imprdral(att,&(imprparabuf[i]),wild_cards_allowed);
		if (res==1)	return i;
		if (res==2) 	goto ok_but_swap_it;

		if (res==-1)	return -1;	/* thats an error */
	}

	wild_cards_allowed[0]=1;
	wild_cards_allowed[1]=0;
	wild_cards_allowed[2]=0;
	wild_cards_allowed[3]=0;
	for (i=0;i<nr_impr_para;i++) {
		res=match_imprdral(att,&(imprparabuf[i]),wild_cards_allowed);
		if (res==1)	return i;
		if (res==2) 	goto ok_but_swap_it;

		if (res==-1)	return -1;	/* thats an error */
	}

	wild_cards_allowed[0]=1;
	wild_cards_allowed[1]=0;
	wild_cards_allowed[2]=0;
	wild_cards_allowed[3]=1;
	for (i=0;i<nr_impr_para;i++) {
		res=match_imprdral(att,&(imprparabuf[i]),wild_cards_allowed);
		if (res==1)	return i;
		if (res==2) 	goto ok_but_swap_it;

		if (res==-1)	return -1;	/* thats an error */
	}

	wild_cards_allowed[0]=1;
	wild_cards_allowed[1]=1;
	wild_cards_allowed[2]=0;
	wild_cards_allowed[3]=0;
	for (i=0;i<nr_impr_para;i++) {
		res=match_imprdral(att,&(imprparabuf[i]),wild_cards_allowed);
		if (res==1)	return i;
		if (res==2) 	goto ok_but_swap_it;

		if (res==-1)	return -1;	/* thats an error */
	}

	return -1; 				/* unknown impr-quad, thats an error ! */

ok_but_swap_it:

	swap=imprquad[3];imprquad[3]=imprquad[0];imprquad[0]=swap;
	swap=imprquad[2];imprquad[2]=imprquad[1];imprquad[1]=swap;
	return i;
}

static
int	match_imprdral(int *att,imprpar_type *par,int *wc_allowed)
{
	int	i,res;
	int	matches=0;
	int	swap_wc_allowed[4];
	static	char	tmpstr[MAX_STRING_SIZE];
	static	int	swapatt[4];

				/*********** first try normal order ****************/
	for (i=0;i<4;i++) {
		if (par->att[i]!=-1) {
			if (par->att[i]==att[i])	matches++;
		} else {
			get_atomtype_str(att[i],tmpstr);
			res=xplor_strcmp(tmpstr,par->attstr[i],wc_allowed[i]);
			if (res==1)	matches++;
			if (res==-1)	return -1;
		}
	}
	if (matches==4)	return 1;	/* all ok, no swap nedded */

				/********* if not matching, try mirror-order ********/
	swapatt[0]=att[3];				swapatt[3]=att[0];
	swapatt[1]=att[2];				swapatt[2]=att[1];
	swap_wc_allowed[0]=wc_allowed[3];		swap_wc_allowed[3]=wc_allowed[0];
	swap_wc_allowed[1]=wc_allowed[2];		swap_wc_allowed[2]=wc_allowed[1];

	matches=0;
	for (i=0;i<4;i++) {
		if (par->att[i]!=-1) {
			if (par->att[i]==swapatt[i])	matches++;
		} else {
			get_atomtype_str(swapatt[i],tmpstr);
			res=xplor_strcmp(tmpstr,par->attstr[i],swap_wc_allowed[i]);
			if (res==1)	matches++;
			if (res==-1)	return -1;
		}
	}
	if (matches==4)	return 2;	/* all ok, but swap needed */

	return 0;
}

/***********************************************************************************/

static	
int	fill_impr_paralist(FILE *h_para,imprpar_type *imprparabuf)
{
	double 	impr_ki,impr_omega0;
	static	char	at1str[LEN_ATOM_TYPE+1];
	static	char	at2str[LEN_ATOM_TYPE+1];
	static	char	at3str[LEN_ATOM_TYPE+1];
	static	char	at4str[LEN_ATOM_TYPE+1];
	int 	res,i,res1;

	while ((res=read_next_imprpara(h_para,
			at1str,at2str,at3str,at4str,
			&impr_ki,&impr_omega0))>=0) {

		strcpy(imprparabuf[nr_impr_para].attstr[0],at1str);
		strcpy(imprparabuf[nr_impr_para].attstr[1],at2str);
		strcpy(imprparabuf[nr_impr_para].attstr[2],at3str);
		strcpy(imprparabuf[nr_impr_para].attstr[3],at4str);

		/**** replaces 'X' by '*' because 'X' acts here as a wild-card *********/
		for (i=0;i<4;i++) {
			if (	(strlen(imprparabuf[nr_impr_para].attstr[i])==1) &&
				(imprparabuf[nr_impr_para].attstr[i][0]=='X') ) {
				imprparabuf[nr_impr_para].attstr[i][0]='*';
			}
		}

		imprparabuf[nr_impr_para].att[0]	=get_atomtype_index(at1str);
		imprparabuf[nr_impr_para].att[1]	=get_atomtype_index(at2str);
		imprparabuf[nr_impr_para].att[2]	=get_atomtype_index(at3str);
		imprparabuf[nr_impr_para].att[3]	=get_atomtype_index(at4str);
		
		imprparabuf[nr_impr_para].ki		=impr_ki;
		imprparabuf[nr_impr_para].omega0	=impr_omega0/180.0*3.1415926;
		
#if TEST_IMPRPARA
		txtout("%4s  %4s  %4s  %4s  %lf  %lf\n",
			imprparabuf[nr_impr_para].attstr[0],
			imprparabuf[nr_impr_para].attstr[1],
			imprparabuf[nr_impr_para].attstr[2],
			imprparabuf[nr_impr_para].attstr[3],
			imprparabuf[nr_impr_para].ki,
			impr_omega0);
#endif
		res1=allready_in_list(imprparabuf);	/* last added element is checked against */
							/* all others */
		if (res1==1)	continue;		/* allready in list */

							/* ok, keep last parameters */
		if (++nr_impr_para>=MXNO_IMPROP_PARAM) {
			errout(
			"Fatal error: Too many improper-parameters (MXNO_IMPROP_PARAM is %d)\n",MXNO_IMPROP_PARAM);
			return -1;
		}
	}
	if (res==-2)	return -1;		/* thats an error */

	return 0;
}

/**************************************************************************************/
/***** last added element is checked against all others *****/

static
int	allready_in_list(imprpar_type *imprparabuf)
{
	int	i;

	for (i=0;i<nr_impr_para;i++) {
	    if (((strcmp(imprparabuf[i].attstr[0],imprparabuf[nr_impr_para].attstr[0])==0) &&
		 (strcmp(imprparabuf[i].attstr[1],imprparabuf[nr_impr_para].attstr[1])==0) &&
		 (strcmp(imprparabuf[i].attstr[2],imprparabuf[nr_impr_para].attstr[2])==0) &&
		 (strcmp(imprparabuf[i].attstr[3],imprparabuf[nr_impr_para].attstr[3])==0))||
									/* swap */
		((strcmp(imprparabuf[i].attstr[0],imprparabuf[nr_impr_para].attstr[3])==0) &&
		 (strcmp(imprparabuf[i].attstr[1],imprparabuf[nr_impr_para].attstr[2])==0) &&
		 (strcmp(imprparabuf[i].attstr[2],imprparabuf[nr_impr_para].attstr[1])==0) &&
		 (strcmp(imprparabuf[i].attstr[3],imprparabuf[nr_impr_para].attstr[0])==0))
               )  {

		    if ((fabs(imprparabuf[i].ki    -imprparabuf[nr_impr_para].ki)    <SMALL) &&
			(fabs(imprparabuf[i].omega0-imprparabuf[nr_impr_para].omega0)<SMALL) ) {

			return 1;	/* thats ok , totally identic,... so skip it */

		    } else {
			errout(
			"Warning error: Inconsistent improper-params for atomtypes <%s> <%s> <%s> <%s>\n",
			imprparabuf[i].attstr[0],
			imprparabuf[i].attstr[1],
			imprparabuf[i].attstr[2],
			imprparabuf[i].attstr[3]);

			errout("old vals (ki,omega0): (%f, %f) skipped vals:(%f, %f)\n",
			imprparabuf[i].ki,		imprparabuf[i].omega0,
			imprparabuf[nr_impr_para].ki,	imprparabuf[nr_impr_para].omega0);
			errout("Skipping new improper-parameters !\n");

			return 1;	/* strings match, but not the values ==> inconsitency */
					/* skip this too */
		    }
	    }
	}
	return 0;
}

/*************************************************************************************/

static
int	add_to_imprlist(int *imprlist,int *imprquad,int nr_read_impr)
{
	int	i;
	int 	off;
	int	swapbuf[5];

	off=5*nr_read_impr;	
							/* add to end of (sorted) list */
	imprlist[off+0]=imprquad[0];
	imprlist[off+1]=imprquad[1];
	imprlist[off+2]=imprquad[2];
	imprlist[off+3]=imprquad[3];
	imprlist[off+4]=imprquad[4];			/* index-pointer to parameter-list */

	if (nr_read_impr==0)	return 1;		/* first quad, no sorting */

							/* sort new entry in ascending order */
							/* of the middle atom (imprquad[1]) */
	for (i=nr_read_impr,off=5*nr_read_impr;i>0;i--,off-=5) {
		if (imprlist[off+1]<imprlist[off-5+1]) {
			memcpy(swapbuf,		imprlist+off-5,	5*sizeof(int));
			memcpy(imprlist+off-5,	imprlist+off,	5*sizeof(int));
			memcpy(imprlist+off,	swapbuf,	5*sizeof(int));
		} else	break;				/* now sorted to right place */
	}	
	return 1;
}

/**************************************************************************************/

static
int	get_next_imprquads(FILE *h_psf,int *imprquads)		/* returns number of integers */
{								/* in impr ==> even           */
	static	char	linestr[MAX_STRING_SIZE];
	int 	i,j;
	int	nr_converted;

	if (read_string(h_psf,linestr)==-1)	return -1;	/* error : can not read at all */

	nr_converted=sscanf(linestr,"%d %d %d %d %d %d %d %d",
				imprquads+0,imprquads+1,imprquads+2,imprquads+3,
				imprquads+5,imprquads+6,imprquads+7,imprquads+8
				);
	
	if (nr_converted<=0) {		/* nothing converted, may be the end */
		return 0;
	} else
	if (nr_converted%4!=0) {
		errout("Fatal error: odd converted numbers in PSF-improper-list\n");
		return -1;
	} else {
		for (i=0,j=0;i<nr_converted/4;i++,j+=5) {
			imprquads[j  ]--;			/* atomnumbers in ego */
			imprquads[j+1]--;			/* start with 0 */
			imprquads[j+2]--;
			imprquads[j+3]--;
		}
		imprquads[4 ]=0;
		imprquads[9 ]=0;	
		return (nr_converted/4)*5;			
	}
}

/**************************************************************************************/


static
int	goto_psf_imprdata(FILE *h_psf,int *nr_of_impr)
{
	static	char	lstr[MAX_STRING_SIZE];
	static	char	tmpstr[MAX_STRING_SIZE];
	int 	res;

	while (read_string(h_psf,lstr)!=-1) {			/* read a psf-line */
		strup(lstr);
		if (sscanf(lstr,"%d %s",&res,tmpstr)==2) {	/* check it */
			tmpstr[7]=0;
			if (strcmp(tmpstr,"!NIMPHI")==0) {
				*nr_of_impr=res;
				return 0;
			}
		}	
	}

	errout("Fatal error: No improper-list-data-section (!NIMPHI) in PSF-file.\n");
	return -1;
}

