/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: womain.c	by M.Eichinger, Munich			**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis	**
**	Contents: main-body, some small utils				**
**		  scan and write out atom-types and atom-data.		**
**		  scan and write bond-params, write ctl-file with	**
**		  default values					**
**		  writeout default (uniform) loadfile (load.lis)	**
**									**
**  	Modif's:							**
**		05.06.94						**
**		11.07.94 minor changes (Tabs in Ctl-file,		**
**			 Default-Values for Distance-Classes)		**
**		25.07.94 Clustersequence for Clustering added		**
**		15.02.94 SBound added					**
**		19.10.95 mcb unit identification via units.def added	**
**									**
*************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "hardw.h"
#include "ego.h"
#include "util_lib.h"
#include "wodecls.h"
#include "woalloc.h"
#include "wounitid.h"
#include "wounits.h"
#include "version.h"

/*************************************************************************/
/**************** functions in this modul ********************************/
/*************************************************************************/

	int	get_atomtype_index(char	*newstr);
	int	get_restype_index(char	*newstr);
	int	get_atti_by_atomnr(int atom_nr);
	int	goto_psf_atomdata(FILE *h_psf,int *nr_of_atoms);
	double	get_atom_mass(int atom_type_index);
	int	get_atomtype_str(int atom_type_index,char *str);
	int	get_pdb_atomdata(int atom_nr,STRUCT_ATOMDATA *data);
	int	isit_a_hydrogen(double mass);
	
static	int	readin_psf_atomtypes(	char *psffn,
				int  *nr_atomtypes,
				char atom_typestr_list[][LEN_ATOM_TYPE+1],
				int  *nr_restypes,
				char res_typestr_list[][LEN_ATOM_TYPE+1],
				int  *nr_of_atoms);

static	int	writeout_atomlisb(	int	nr_atomtypes,
					char	atom_typestr_list[][LEN_ATOM_TYPE+1],
					int     nr_restypes,
					char    res_typestr_list[][LEN_RES_TYPE+1],
					char	*atomlisbfn);

static	int	readin_atomdata(char *pdbfn,char *psffn);
static	int	free_atom_buffer(void);

static	int	writeout_bondpara(int argc,char *argv[],char *bondparafn);
static	int	fill_bondmatrix(FILE *h_para,double *bondmat);

static	int	free_all_buffers(void);

static	int	writeout_loadfile(char *loadfn);
static	int	writeout_forcefile(char *forcefn);

/*************** buffer for holding atom-type-index for each atom ************/
static	int		nr_of_atoms=0;
static	STRUCT_ATOMDATA	*atom_buf;
static	int		atom_buf_here=0;

/******************** buffer for holding atom-type-strings *******************/
static	int	nr_atomtypes=0;
static	char	atom_typestr_list[MXNO_ATOM_TYPES+1][LEN_ATOM_TYPE+1];
static	double	atom_typemass_list[MXNO_ATOM_TYPES+1];

/******************** buffer for holding res-type-strings ********************/
static	int	nr_restypes=0;
static	char	res_typestr_list[MXNO_RES_TYPES+1][LEN_RES_TYPE+1];

/******************** buffer for holding data from units.def ******************/
static	res_description_t **descriptions;

/*****************************************************************************/
static	LOAD_MSG_ALL_NODE global_data;

/******* this is the range for Xplor to detect Hydrogens *****/
#define H_MIN_LIMIT	0.0
#define H_MAX_LIMIT	3.5

/*****************************************************************************/

int main(int argc,char *argv[])
{
	char	unitfilen[MAX_STRING_SIZE];
	char	psffilen[MAX_STRING_SIZE];
	char	pdbfilen[MAX_STRING_SIZE];
	int	i,res=0,num_units=0,nr_of_neutrals=0;
	
	txtout("\n===================================================\n");
	txtout("xpl2lis - Converting Xplor-data to EGO-data-format.\n");
	txtout("Version for %s\n",VERSION);
	txtout("===================================================\n");

	if (argc<5) {
		errout("Fatal error: Illegal call of program.\n");
		errout("You have to specify a units definition file (units.def),\n");
		errout("a PDB, a PSF and at least one Xplor-Parameter-file !\n");
		errout("xpl2lis <units.def> <pdb-file> <psf-file> <param-file1> [<param-file2> ...]\n\n");
		return -1;
	}
	
	fill_filename_list(&global_data);
	strcpy(unitfilen,argv[1]);
	strcpy(pdbfilen,argv[2]);
	strcpy(psffilen,argv[3]);

	txtout("units.def file : %s\n",unitfilen);
	txtout("PDB-file       : %s\n",pdbfilen);
	txtout("PSF-file       : %s\n",psffilen);
	txtout("Parameter-files: %d\n",argc-4);
	for (i=4;i<argc;i++) {
		txtout("                 %s\n",argv[i]);
	}
	txtout("===================================================\n\n");

	/****************************************************************/
	if (readin_psf_atomtypes(psffilen,&nr_atomtypes,
				 atom_typestr_list,
				 &nr_restypes,
				 res_typestr_list,
				 &nr_of_atoms)==-1) 		goto error_occured;
	
	if (writeout_atomlisb(	nr_atomtypes,
				atom_typestr_list,
				nr_restypes,
				res_typestr_list,
				global_data.filenames[TYPE_FILE_INDEX])==-1) 	goto error_occured;
	
	if (readin_atomdata(	pdbfilen,psffilen)==-1)		goto error_occured;

	if (writeout_coordfile( nr_of_atoms,global_data.filenames[COORD_FILE_INDEX],
				atom_buf)==-1)			goto error_occured;

	if (readin_bondlist(	psffilen,
				nr_of_atoms)==-1)		goto error_occured;

	if (writeout_bondlist(	global_data.filenames[BONDLIST_FILE_INDEX])==-1) goto error_occured;

		
	if (writeout_bondpara(	argc,argv,
				global_data.filenames[BONDPARA_FILE_INDEX])==-1) goto error_occured;

	if (readin_all_anglepara(argc,argv, atom_typestr_list)==-1)		goto error_occured;
	
	if (writeout_anglelist(	psffilen,
				global_data.filenames[ANGLELIST_FILE_INDEX],
				nr_of_atoms)==-1) 		goto error_occured;

	if (writeout_anglepara(	global_data.filenames[ANGLEPARA_FILE_INDEX])==-1) goto error_occured;

	/*****************************************************************/
	/***** do not change the calling-order (see woexcl.c) *****/
	if (buildup_1excl_list(	psffilen,
				nr_of_atoms)==-1)		goto error_occured;

	if (writeout_1excl (	global_data.filenames[EXCL_FILE_INDEX+0],
				nr_of_atoms)==-1)		goto error_occured;

	if (writeout_m1excl(	global_data.filenames[EXCL_FILE_INDEX+5],
				nr_of_atoms)==-1)		goto error_occured;

	if (writeout_2excl (	global_data.filenames[EXCL_FILE_INDEX+1],
				global_data.filenames[EXCL_FILE_INDEX+6],
				nr_of_atoms)==-1)		goto error_occured;

	if (writeout_3excl (	global_data.filenames[EXCL_FILE_INDEX+2],
				global_data.filenames[EXCL_FILE_INDEX+7],
				nr_of_atoms)==-1)		goto error_occured;

	if (writeout_4excl (	global_data.filenames[EXCL_FILE_INDEX+3],
				global_data.filenames[EXCL_FILE_INDEX+8],
				nr_of_atoms)==-1)		goto error_occured;

	if (writeout_14excl(	global_data.filenames[EXCL_FILE_INDEX+4],
				nr_of_atoms)==-1)		goto error_occured;

	free_1excl_buffer();
	/****************************************************************/
	
	if (writeout_shakelist(	nr_of_atoms,
				global_data.filenames[SHAKE_FILE_INDEX])==-1)	goto error_occured;

	if (readin_vdw_params(	argc,argv,
				nr_atomtypes)==-1)		goto error_occured;

	if (writeout_vdwparams(	global_data.filenames[VDWPARA_FILE_INDEX],
				global_data.filenames[VDW14PARA_FILE_INDEX])==-1) goto error_occured;

	/*****************************************************************/
	if (readin_dihepara(argc,argv)==-1)			goto error_occured;

	if (writeout_dihelist(	psffilen,
				global_data.filenames[DIHELIST_FILE_INDEX],
				nr_of_atoms)==-1)		goto error_occured;

	if (writeout_dihepara(global_data.filenames[DIHEPARA_FILE_INDEX])==-1)	goto error_occured;

	/*****************************************************************/
	if (readin_imprpara(argc,argv)==-1)			goto error_occured;

	if (writeout_imprlist(	psffilen,
				global_data.filenames[IMPRLIST_FILE_INDEX],
				nr_of_atoms)==-1)		goto error_occured;

	if (writeout_imprpara(	global_data.filenames[IMPRPARA_FILE_INDEX])==-1) goto error_occured;

	/******************************************************************/	
	/** do not change the calling order (see in wohbond.c) ***/
	if (writeout_hbacclis(	psffilen,
				global_data.filenames[HBACC_FILE_INDEX],
				nr_of_atoms,nr_atomtypes)==-1)	goto error_occured;

	if (writeout_hbdonlis(	psffilen,
				global_data.filenames[HBDON_FILE_INDEX],
				nr_of_atoms)==-1)		goto error_occured;

	if (buildup_hbmats(argc,argv)==-1)			goto error_occured;

	if (writeout_hbhtype(	global_data.filenames[HBHT_FILE_INDEX],
				global_data.filenames[HBTRAN_FILE_INDEX])==-1)	goto error_occured;

	if (writeout_hbmats(	global_data.filenames[HBAH_FILE_INDEX],
				global_data.filenames[HBBH_FILE_INDEX])==-1)	goto error_occured;

	/****************************************************************/
	
	if ((descriptions=read_res_descriptions(unitfilen))==NULL) goto error_occured;

#if DEBUG
	if (write_res_descriptions(descriptions,"hallo.txt")!=0) goto error_occured;
#endif	
	num_units = identify_units(descriptions,nr_of_atoms,&nr_of_neutrals);

	if (num_units==-1)			goto error_occured;
	if (writeout_units(global_data.filenames[UNIT_FILE_INDEX])==-1)	goto error_occured;
	res=print_restinfo();

	/*****************************************************************/
	if (writeout_loadfile(global_data.filenames[PGA_FILE_INDEX])==-1)	goto error_occured;
	if (writeout_forcefile(global_data.filenames[FORCE_FILE_INDEX])==-1)	goto error_occured;

	fill_dummy_ctl_data(&global_data);	
	global_data.total_no_atoms	= nr_of_atoms;
	/*** negative values for global_data->cluster_levels and 
	     global_data->cluster_branches ***/
	/*** indicate, that they are not realy used by EGO. Only if a positive ***/
	/*** value is entered, these values are used by EGO.		       ***/
	global_data.treeinfo.levels = -set_cluster_params(nr_of_atoms,
						global_data.outer_classrad,
						global_data.treeinfo.clusterseq);
	global_data.treeinfo.branches = -num_units;
	for (i=1;i<-global_data.treeinfo.levels;i++) {
		global_data.treeinfo.branches /= global_data.treeinfo.clusterseq[i];
	}

	
	{FILE *h_ctl;
	if ((h_ctl=fopen("ctl.lis","w"))==NULL) {
		errout("Can not write control-file <%s>.\n","ctl.lis");
		goto error_occured;
	}
	if (writeout_ctlfile(h_ctl,&global_data)==-1)	goto error_occured;	
	fclose(h_ctl);
	txtout("Control-file ctl.lis written.\n");
	}

	/*******************************************************************/
	if (res==-1) goto error_occured;

	free_all_buffers();
	
	txtout("\n===================================================\n");
	txtout("xpl2lis finished now. No fatal errors.\n");
	txtout("===================================================\n\n");
	return 0;

error_occured:
	free_all_buffers();

	errout("\n===================================================\n");
	errout("xpl2lis stopped with a fatal error.\n");
	errout("Check your Xplor-input-files for this error.\n");
	errout("===================================================\n");
	return -1;
}

/****************************************************************************/

int	get_atomtype_index(char	*atomtypstr)
{
	int i;

	if (nr_atomtypes<=0)	return -1;	/* list is empty */	

	for (i=0;i<LEN_ATOM_TYPE;i++) {		/* add up spaces */
		if (atomtypstr[i]==0) {
			atomtypstr[i]=' ';
			atomtypstr[i+1]=0;
		}
	}

	for (i=0;i<nr_atomtypes;i++) {		/* search for matching string */
		if (strcmp(atomtypstr,atom_typestr_list[i])==0)	return i;
	}
	return -1;				/* str not in list */
}

/****************************************************************************/

int	get_restype_index(char	*restypstr)
{
	int i;

	if (nr_restypes<=0)	return -1;	/* list is empty */	

	for (i=0;i<LEN_RES_TYPE;i++) {		/* add up spaces */
		if (restypstr[i]==0) {
			restypstr[i]=' ';
			restypstr[i+1]=0;
		}
	}

	for (i=0;i<nr_restypes;i++) {		/* search for matching string */
		if (strcmp(restypstr,res_typestr_list[i])==0)	return i;
	}
	return -1;				/* str not in list */
}

/************ use by other moduls to get info about atom-type-index *******/
/*****        /-------- read: ATom-Type-Index ! ***************************/
int	get_atti_by_atomnr(int atom_nr)
{
	return atom_buf[atom_nr].atomtype;
}

/*************** get the mass of atom by atom-type-index ******************/
double	get_atom_mass(int att_index)
{
	if ((att_index<0)||(att_index>=nr_atomtypes)) {
		errout("Fatal error: Illegal atomtype-index in get_atom_mass.\n");
		return 0.0;
	}
	return (double) atom_typemass_list[att_index];	
}

/*************** get the atom-type-string by atom-type-index **************/
int	get_atomtype_str(int att_index,char *str)
{

	if ((att_index<0)||(att_index>=nr_atomtypes)) {
		errout("Fatal error: Illegal atomtype-index in get_atomtype_str.\n");
		return 0.0;
	}

	strcpy(str,atom_typestr_list[att_index]);
	return 0;
}

int	get_pdb_atomdata(int atom_nr,STRUCT_ATOMDATA *data)
{
	if ((atom_nr<0)||(atom_nr>=nr_of_atoms)) {
		errout("Fatal error: Illegal Atom-Number. Check code.\n");
		return -1;
	}
	memcpy(data,&(atom_buf[atom_nr]),sizeof(STRUCT_ATOMDATA));

	return 0;
}

int	isit_a_hydrogen(double mass)
{
	if (	(mass>=H_MIN_LIMIT) &&
		(mass<=H_MAX_LIMIT) ) {
		return 1;		/* YES */
	}
	return 0;			/* NO */
}


/**************************************************************************/

static	
int	free_all_buffers(void)
{
	free_bondlist();
	free_atom_buffer();
	free_anglebuffers();
	free_dihebuffers();
	free_imprbuffers();

	return 0;
}

/*************************************************************************/

static
int	writeout_bondpara(int argc,char *argv[],char *bondparafn)
{
	int	file_ind,i,j,errors=0;		/* 3 is the first parameter file */
	FILE 	*h_para,*h_bond;
	double	*bondmat;
	
	/************* allocate and ini bond-matrix with nonsense *********/
	bondmat=alloc_doubles((nr_atomtypes+1)*(nr_atomtypes+1));
	for (i=0;i<=nr_atomtypes;i++) {
		for (j=0;j<=nr_atomtypes;j++) {
			bondmat[i*(nr_atomtypes+1)+j]=-999.0;
		}
	}
	
	/******************* scan each parmater-file *********************/
	for (file_ind=3;file_ind<argc;file_ind++) {
		if ((h_para=fopen(argv[file_ind],"r"))==NULL) {
			errout("Can not read parameter-file: <%s>\n",argv[file_ind]);
			myfree(bondmat);
			return -1;
		}

		if (fill_bondmatrix(h_para,bondmat)==-1) {
			fclose(h_para);
			myfree(bondmat);
			errout("Error occured in parameter-file: %s\n",argv[file_ind]);
			return -1;
		}
		
		fclose(h_para);
	}


	for (i=0;i<nr_of_atoms;i++) {
		int mybondl[MXNO_FIELDS_BOND+1],numbonds,k,ind1,ind2;
		
		ind1 = get_atti_by_atomnr(i);
		numbonds=get_atom_bondlist(i,mybondl);
		
		for (k=0;k<numbonds;k++) {
			 ind2 = get_atti_by_atomnr(mybondl[k]);
			 if (ind1==ind2) ind2=nr_atomtypes;

			 if ((bondmat[ind1*(nr_atomtypes+1)+ind2])<-990.0) {
			 	char str1[10],str2[10];
			 	get_atomtype_str(ind1,str1);
				ind2 = get_atti_by_atomnr(mybondl[k]);
			 	get_atomtype_str(ind2,str2);			 	
			 	txtout("Fatal error: No bond parameters between atometype <%s> and <%s>\n",
			 		str1,str2);
			 	errors=1;
			 }
		}
	}

	if (errors!=0) return -1;	

	/***************** write bond-params to file **********************/
	if ((h_bond=fopen(bondparafn,"w"))==NULL) {
		errout("Can not open bond-parameter-file.\n");
		myfree(bondmat);
		return -1;
	}
	
	fprintf(h_bond,"This is the BONDed parameter matrix.\n");
	fprintf(h_bond,"Number of atom typs encountered:    %4d\n",nr_atomtypes);

	for (i=0;i<=nr_atomtypes;i++) {
		for (j=0;j<=nr_atomtypes;j++) {
			if (	((j+1)%MXNO_FIELDS_BOND==0) ||
				(j==nr_atomtypes)	) {
				fprintf(h_bond,"%8.4f\n",bondmat[i*(nr_atomtypes+1)+j]);
			} else {
				fprintf(h_bond,"%8.4f  ",bondmat[i*(nr_atomtypes+1)+j]);
			}
		}
		fprintf(h_bond,"\n");
	}
	fclose(h_bond);
	myfree(bondmat);
	
	txtout("Bond parameters written.\n");
	return 0;
}

/***************************************************************************/

static
int	fill_bondmatrix(FILE *h_para,double *bondmat)
{
	double 	bond_kb,bond_r0;
	int	att1,att2;	/* atom-type-index ! */
	int 	mini,maxi,res;
	int	kb_index=0,r0_index=0;

	while ((res=read_next_bondpara(h_para,&att1,&att2,&bond_kb,&bond_r0))>=0) {
		if (res==0)	continue;	/* found atom-type which don't occur in */
						/* PDB/PSF-file, no error, skip it  */
		if (att1==att2) {		
			kb_index=att1*(nr_atomtypes+1)+nr_atomtypes;	/* last row */
			r0_index=nr_atomtypes*(nr_atomtypes+1)+att1;	/* last line */
		} else {
			mini=min(att1,att2);
			maxi=max(att1,att2);
							/* upper-right part of matrix */
			kb_index=mini*(nr_atomtypes+1)+maxi;
							/* lower-left part of matrix */
			r0_index=maxi*(nr_atomtypes+1)+mini;
		}

		if (bondmat[kb_index]>-990.0) {		/* we got that value before */
			if (	(fabs(bondmat[kb_index]-bond_kb)>SMALL) ||
				(fabs(bondmat[r0_index]-bond_r0)>SMALL) ) {
				txtout("Warning error: Inconsistent bond-params for atomtypes <%s> <%s>\n",
					atom_typestr_list[att1],atom_typestr_list[att2]);

				txtout("old vals (kb,r0): (%f, %f)  skipped vals: (%f, %f)\n",
				bondmat[kb_index],bondmat[r0_index],bond_kb,bond_r0);
				txtout("Skipping new bond-parameters !\n");
			}
		} else {
			bondmat[kb_index]=bond_kb;
			bondmat[r0_index]=bond_r0;
		}
	}
	if (res==-2) {
		errout("Fatal error: while reading bond-parameters !\n");
		return -1;		/* thats an error */
	}

	return 0;
}

/**************************************************************************/

static
int	readin_psf_atomtypes(	char *psffn,
				int  *nr_atomtypes,
				char atom_typestr_list[][LEN_ATOM_TYPE+1],
				int  *nr_restypes,
				char res_typestr_list[][LEN_ATOM_TYPE+1],
				int  *nr_of_atoms)
{
	FILE	*h_psf;
	int	res=0;
	int	readin_atoms=0;
	int	atom_index=0,ati,rti;
	double	atom_charge=0.0,atom_mass=0.0;	/* not analyzed here */
	static	char	tmpstr[MAX_STRING_SIZE];
/**** next variables are not used *****/
	int	psf_residnr;
static	char	psf_residstr[LEN_ATOM_TYPE+1];
static	char	psf_segidstr[LEN_ATOM_TYPE+1];
static	char	psf_atomstr [LEN_ATOM_TYPE+1];

	*nr_of_atoms=0;
	*nr_atomtypes=0;
	*nr_restypes=0;

	if ((h_psf=fopen(psffn,"r"))==NULL) {
		errout("Can not open specified psf-file.\n");
		return -1;
	}

	if (goto_psf_atomdata(h_psf,nr_of_atoms)==-1) return -1;	
	
	txtout("Number of atoms in PSF-file: %d\n",*nr_of_atoms);
	
	do { /**** for each atom in psf-file *******/
		res=read_next_psfatom(h_psf,	&atom_index,
						tmpstr,
						&atom_charge,
						&atom_mass,
						psf_atomstr,
						psf_residstr,
						&psf_residnr,
						psf_segidstr);

		if (res==0) {
					/* try to get the atomtype index */
			ati=get_atomtype_index(tmpstr);
					
			if (ati==-1) {	/* -1 means,... not in list, so add it  */

				strcpy(atom_typestr_list[*nr_atomtypes],tmpstr);
				atom_typemass_list[*nr_atomtypes]=atom_mass;

				if (++(*nr_atomtypes)>MXNO_ATOM_TYPES) {
					errout("Too many atomtypes in PSF-file !\n");
					errout("Actual limit is %d (MXNO_ATOM_TYPES)\n",
						MXNO_ATOM_TYPES);
					fclose(h_psf);
					return -1;
				}
			}
					/* try to get the restype index */
			rti=get_restype_index(psf_residstr);
					
			if (rti==-1) {	/* -1 means,... not in list, so add it  */

				strcpy(res_typestr_list[*nr_restypes],psf_residstr);

				if (++(*nr_restypes)>MXNO_RES_TYPES) {
					errout("Too many residue types in PSF-file !\n");
					errout("Actual limit is %d (MXNO_RES_TYPES)\n",
						MXNO_RES_TYPES);
					fclose(h_psf);
					return -1;
				}
			}
			readin_atoms++;	
		} else {
			errout("Error detected after %d read atoms in PSF-file.\n",
				readin_atoms);
			fclose(h_psf);
			return -1;
		}
	} while(readin_atoms<*nr_of_atoms);

	txtout("Number of atom-types       : %d\n",*nr_atomtypes);

	fclose(h_psf);
	return 0;
}

/***************************************************************************/

static
int	writeout_atomlisb(	int	nr_atomtypes,
				char	atom_typestr_list[][LEN_ATOM_TYPE+1],
				int     nr_restypes,
				char    res_typestr_list[][LEN_RES_TYPE+1],
				char	*atomlisbfn)
{
	FILE	*h_file;
	int	i;
		
	if ((h_file=fopen(atomlisbfn,"w"))==NULL) {
		errout("Can not open atomtype-list-file.\n");
		return -1;
	}

	fprintf(h_file,"The format is:\n");
	fprintf(h_file,"number of atom entries\n");
	fprintf(h_file,"atom.typ         user readable string\n");
	fprintf(h_file,"number of residue entries\n");
	fprintf(h_file,"residue.typ      user readable string\n");
	
	fprintf(h_file,"%d\n",nr_atomtypes);
	for (i= 0;i<nr_atomtypes;i++) {
		fprintf(h_file,"       %2d        %s\n",i,atom_typestr_list[i]);
	}
	fprintf(h_file,"%d\n",nr_restypes);
	for (i= 0;i<nr_restypes;i++) {
		fprintf(h_file,"       %2d        %s\n",i,res_typestr_list[i]);
	}
	
	fclose(h_file);
	return 0;
}

/***************************************************************************/

int	goto_psf_atomdata(FILE *h_psf,int *nr_of_atoms)
{
	static	char	lstr[MAX_STRING_SIZE];
	static	char	tmpstr[MAX_STRING_SIZE];
	int 	res;

	while (read_string(h_psf,lstr)!=-1) {
		strup(lstr);
		if (sscanf(lstr,"%d %s",&res,tmpstr)==2) {
			if (strcmp(tmpstr,"!NATOM")==0) {
				*nr_of_atoms=res;
				return 0;
			}
		}	
	}
	errout("Fatal error: No atom-data-section (!NATOM) in PSF-file.\n");
	return -1;
}


/*****************************************************************************/
/********** this functions write outs atom-index, coords, atom-type-index,.***/
/*********** and fills a buffer with atom-data for use by other moduls *******/
/*****************************************************************************/

static
int	readin_atomdata(char *pdbfn,char *psffn)
{
	FILE	*h_pdb,*h_psf;
	int	pdbres=0,psfres=0;
	int	nr_of_atoms=0,nr_read_atoms=0;
	int	pdb_atom_index=0,psf_atom_index;
	int	atom_type_index=0,res_type_index=0;
	int	psf_residnr,pdb_residnr;
	double	coords[3],vel[3];
	double	beta=0.0,harmon=0.0,atom_charge=0.0;
	double	atom_mass=0.0;
static	char	atomtypestr[MAX_STRING_SIZE];
static	char	psf_segidstr[LEN_ATOM_TYPE+1],pdb_segidstr[LEN_ATOM_TYPE+1];
static	char	psf_residstr[LEN_ATOM_TYPE+1],pdb_residstr[LEN_ATOM_TYPE+1];
static	char	psf_atomstr[LEN_ATOM_TYPE+1],pdb_atomstr[LEN_ATOM_TYPE+1];


	if ((h_pdb=fopen(pdbfn,"r"))==NULL) {
		errout("Can not open pdb-file <%s>.\n",pdbfn);
		return -1;
	}
	
	if ((h_psf=fopen(psffn,"r"))==NULL) {
		errout("Can not open psf-file <%s>.\n",psffn);
		fclose(h_pdb);
		return -1;
	}
	if (goto_psf_atomdata(h_psf,&nr_of_atoms)==-1) {
		fclose(h_psf);
		fclose(h_pdb);
		return -1;	
	}

	atom_buf=alloc_atomdata(nr_of_atoms+5);
	atom_buf_here=1;
	vel[0]=vel[1]=vel[2]=0.0;

	do {	/***** read data of each atom ***************************/

		pdbres=read_next_pdbdata(h_pdb,	&pdb_atom_index,
						coords,
						&beta,
						&harmon,
						pdb_atomstr,
						pdb_residstr,
						&pdb_residnr,
						pdb_segidstr);
		
		psfres=read_next_psfatom(h_psf,	&psf_atom_index,
						atomtypestr,
						&atom_charge,
						&atom_mass,
						psf_atomstr,
						psf_residstr,
						&psf_residnr,
						psf_segidstr);

		/****** many, many checks ********************************/
		if ((pdbres==-2)||(psfres!=0)) {
			errout("Error detected after %d read atoms.\n",
				nr_read_atoms);
			goto error_in_wcoord;
		}
		
		if (psf_atom_index!=pdb_atom_index) {
			errout("PSF and PDB-data out of sync after %d read atoms.\n",
				nr_read_atoms);
			errout("PSF-Atom-Number %d  !+ PDB-Atom-Number %d\n",
				psf_atom_index,
				pdb_atom_index);
			goto error_in_wcoord;
		}

		if (	(strcmp(pdb_atomstr, psf_atomstr )!=0) ||
			(strcmp(pdb_residstr,psf_residstr)!=0) ||
			(strcmp(pdb_segidstr,psf_segidstr)!=0) ||
			(pdb_residnr != psf_residnr) ) {

			errout("(<%s> <%s>) (<%s> <%s>) (<%s> <%s>) (%d %d)\n",
				pdb_atomstr,psf_atomstr,
				pdb_residstr,psf_residstr,
				pdb_segidstr,psf_segidstr,
				pdb_residnr,psf_residnr);
			errout("Fatal error: PDB and PSF-data are inconsistent after %d read atoms.\n",
				nr_read_atoms);
			goto error_in_wcoord;
		}

		if (pdbres!=0) continue;

		atom_type_index=get_atomtype_index(atomtypestr);
		if (atom_type_index==-1) {	
			errout("Unknown atom-type <%s> after %d read atoms in PDB-file.\n",
				atomtypestr,nr_read_atoms);
			goto error_in_wcoord;
		}

		res_type_index=get_restype_index(psf_residstr);
		if (res_type_index==-1) {	
			errout("Unknown res-type <%s> after %d read atoms in PDB-file.\n",
				psf_residstr,nr_read_atoms);
			goto error_in_wcoord;
		}

		if ((pdb_atom_index<1)||(pdb_atom_index>nr_of_atoms)) {
			errout("Illegal atom-number %d in PDB-file.\n",
				pdb_atom_index);
			goto error_in_wcoord;
		}

		/********** keep data in mem for use in other modules ******/
		atom_buf[pdb_atom_index-1].atomnr=pdb_atom_index-1;

		atom_buf[pdb_atom_index-1].atomtype=atom_type_index; 
		strcpy(atom_buf[pdb_atom_index-1].atomtypestr,atomtypestr);

		atom_buf[pdb_atom_index-1].residnr=psf_residnr;
		strcpy(atom_buf[pdb_atom_index-1].residstr,psf_residstr);
		
		strcpy(atom_buf[pdb_atom_index-1].atomstr,psf_atomstr);
		strcpy(atom_buf[pdb_atom_index-1].segidstr,psf_segidstr);

		if (	(coords[0]<-990.0) || (coords[0]>9990.0) ||
			(coords[1]<-990.0) || (coords[1]>9990.0) ||
			(coords[2]<-990.0) || (coords[2]>9990.0) ) {
			errout("There seem to be unkown atom positions in the PDB file!\n");
			errout("The valid range for atom positions is [-990.0, 9990.0].\n");		
			errout("Atom-Nr %d: Position %f %f %f\n",pdb_atom_index,
								 coords[0],coords[1],coords[2]);
			goto error_in_wcoord;	
		}

		atom_buf[pdb_atom_index-1].ko[0]  = coords[0];
		atom_buf[pdb_atom_index-1].ko[1]  = coords[1];
		atom_buf[pdb_atom_index-1].ko[2]  = coords[2];

		atom_buf[pdb_atom_index-1].vel[0] = vel[0];
		atom_buf[pdb_atom_index-1].vel[1] = vel[1];
		atom_buf[pdb_atom_index-1].vel[2] = vel[2];

		atom_buf[pdb_atom_index-1].charge = atom_charge;
		atom_buf[pdb_atom_index-1].beta   = beta;
		atom_buf[pdb_atom_index-1].harmon = harmon;

		nr_read_atoms++;

	} while (nr_read_atoms<nr_of_atoms); /***** end: for each atom ****/

	fclose(h_psf);
	fclose(h_pdb);

	txtout("Read atoms in PDB-file     : %d\n",nr_read_atoms);
	return 0;

error_in_wcoord:
	fclose(h_psf);
	fclose(h_pdb);
	return -1;
}

static
int	free_atom_buffer(void)
{
	if (atom_buf_here) {
		myfree(atom_buf);
		atom_buf_here=0;
	}
	return 0;
}

/**************************************************************************/

/*************************************************************************/

static
int	writeout_loadfile(char *loadfn)
{
	FILE	*h_load;
	int	nr_of_nodes	=4;
	int	divres=(int)(nr_of_atoms/nr_of_nodes);
	int	i;
	int	rest=nr_of_atoms;

	if ((h_load=fopen(loadfn,"w"))==NULL) {
		errout("Can not write load-file.\n");
		return -1;
	}

	fprintf(h_load,"%4d \t\t %s\n",
		nr_of_nodes,"last number of nodes used; next the atom distribution:");

	for (i=0;i<nr_of_nodes-1;i++) {
		fprintf(h_load,"%5d \t\t\n",divres);
		rest-=divres;
	}
	fprintf(h_load,"%5d\t\t\n",rest);
	fclose(h_load);
	txtout("Load-file written.\n");
	return 0;
}

/*************************************************************************/

static
int	writeout_forcefile(char *forcefn)
{
	FILE	*h_force;

	if ((h_force=fopen(forcefn,"w"))==NULL) {
		errout("Can not write force-file <%s>.\n",forcefn);
		return -1;
	}

	fprintf(h_force,"! Control file for force output.\n");
	fprintf(h_force,"! =============================================================\n");
	fprintf(h_force,"! Comment lines are introduced by '!'.\n");
	fprintf(h_force,"! Comment lines which are printed to screen and added to\n");
	fprintf(h_force,"! EGO-output files are introduced by '#'.\n\n");
	fprintf(h_force,"! Format of atom force-selection: <FORCETYPE> <atomselection string>\n");
	fprintf(h_force,"! <FORCETYPE> can be one of the following keywords:\n");
	fprintf(h_force,"! TOTAL        Total force (ELECVDW+INTERN).\n");
	fprintf(h_force,"! ELECVDW      Sum of electrostatic and VDW force.\n");
	fprintf(h_force,"! INTERN       Sum of BOND, ANGLE, DIHEDRAL and IMPROPER forces.\n");
	fprintf(h_force,"! ELEC         Electrostatic force.\n");
	fprintf(h_force,"! VDW          VDW force.\n");
	fprintf(h_force,"! BOND         Bond force.\n");
	fprintf(h_force,"! ANGLE        Angle force.\n");
	fprintf(h_force,"! DIHE         Dihedral force.\n");
	fprintf(h_force,"! IMPR         Improper force.\n");
	fprintf(h_force,"! RESTR        Harmonic restraint force.\n");
	fprintf(h_force,"! SBOUND       SBound force (not yet implemented).\n");
	fprintf(h_force,"! FLOOD        Flooding force (not yet implemented).\n");
	fprintf(h_force,"! KIN          Velocity (not yet implemented).\n");
	fprintf(h_force,"! =============================================================\n");
	fprintf(h_force,"! Through PICK-statements specific forces can be selected:\n");
	fprintf(h_force,"! Format of pick force-selection (Atom :\n");
	fprintf(h_force,"! PICKBOND	<AtomNr1>    <AtomNr2>\n");
	fprintf(h_force,"! PICKELEC	<AtomNr1>    <AtomNr2>\n");
	fprintf(h_force,"! PICKVDW	<AtomNr1>    <AtomNr2>\n");
	fprintf(h_force,"! PICKANGLE	<AtomNr1>    <AtomNr2>    <AtomNr3>\n");
	fprintf(h_force,"! PICKDIHE	<AtomNr1>    <AtomNr2>    <AtomNr3>    <AtomNr4>\n");
	fprintf(h_force,"! PICKIMPR	<AtomNr1>    <AtomNr2>    <AtomNr3>    <AtomNr4>\n");
	fprintf(h_force,"! =============================================================\n");
	fprintf(h_force,"! Other control parameters are:\n");
	fprintf(h_force,"! START    <IntegrStepNum>               Sets the start integration step.\n");
	fprintf(h_force,"! STRIDE   <StrideStep>                  Sets the write-out stride of forces.\n");
	fprintf(h_force,"! IDSTRING <string>                      Modifies the filenames for force-output.\n");
	fprintf(h_force,"! FORCEOFF <atomselection string>        Switch off force for selected atoms.\n");
	fprintf(h_force,"! SPEEDOFF <atomselection string>        Switch off velocity for selected atoms.\n");
	fprintf(h_force,"! SHAKEOFF <atomselection string>        Switch off SHAKE for selected atoms.\n");
	fprintf(h_force,"! SBOUNDMODE    <0, 1, 2, 3>             Switch SBOUND MODE (ranged SBOUND)feature on (1) or\n");
	fprintf(h_force,"!                                        off (0) or (2) use water parameters only,\n");
	fprintf(h_force,"!                                        or (3) use ranged SBOUND in numerical_sbound.\n");
	fprintf(h_force,"! PART1    <atomselection string>        undocumented.\n");
	fprintf(h_force,"! PART2    <atomselection string>        undocumented.\n");		
	fprintf(h_force,"! =============================================================\n");
	fprintf(h_force,"! Parameters for the QM interface:\n");
	fprintf(h_force,"! QMATOM <atomselection string>          Note that also the linkatoms have\n");
	fprintf(h_force,"!                                        to be included in the selection.\n");
	fprintf(h_force,"! QMLINK <AtomNr1> <AtomNr2> <distance>  Selects AtomNr2 as a linkatom with\n");
	fprintf(h_force,"!                                        the given distance.\n");
	fprintf(h_force,"! QMSIGMA <value> <atomselection string> Width of electronic density.\n");
	fprintf(h_force,"! QMSTART <IntegrStepNum> <script>       Starts the QM-program at integration\n");
	fprintf(h_force,"!                                        step with the given script-file.\n");
	fprintf(h_force,"! QMCUTOFF <value>                       Sets WF-cutoff value for QM calculation.\n");	
	fprintf(h_force,"! QMCONTINUE <stride> <script>           Recall/reinvoce the QM-program after\n");
	fprintf(h_force,"!                                        stride steps.\n");
	fprintf(h_force,"! QMEXTENDBOX -x +x -y +y -z +z          Extends the QM-box around QM-atoms.\n");
	fprintf(h_force,"! QMFIXBOX    -x +x -y +y -z +z          Sets QM-box to a fixed size.\n");
	fprintf(h_force,"! QMNOEXTELEC                            Set this if you want no external\n");
	fprintf(h_force,"!                                        field passed to the QM-program.\n");
	fprintf(h_force,"! QMATOMFORCE OFF                        Switch off forces on QM atoms.\n");
	fprintf(h_force,"! QMATOMFORCE QM               (default) Use the forces from the QM program.\n");
	fprintf(h_force,"! QMATOMFORCE CLASSIC                    Use the classical forces from EGO.\n");
	fprintf(h_force,"! =============================================================\n");
	fclose(h_force);
	txtout("Force-output control file written.\n");
	return 0;
}

