/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*********************************************************************************
**										**
**	Module	: wounits.c	by M.Eichinger, Munich				**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis)	**
**	Contents:  								**
**		Scans the molecule structure and defines the structural units,  **
**		which represents the first clustering-level needed for FMM !.	** 
**  	Modif's:								**
*********************************************************************************/

#include <math.h>
#include <stdlib.h>

#include "hardw.h"
#include "ego.h"
#include "woalloc.h"
#include "util_lib.h"
#include "wodecls.h"
#include "wounitid.h"
#include "wounits.h"


#define DEBUG_UNITS		0
#define DEBUG_LIPID		0

#define	UNIT_FACTOR		1
#define MAX_ATOMS_IN_UNIT	50
#define	UNCHECKED		-1

/***********************************************************************/

	int	scan4_units(int nrofatoms,int *nr_of_neutrals);
	int	writeout_units(char *unitfn);
	int	print_restinfo(void);

	int	free_scan4_units(void);

static	int	reset_atombuffer();
static	int	sort_to_buffer(int atom_nr);
static	int	save_unit(int tag);

void 	sort_H_to_buffer(abuf_entry_t **buffer, int atom_id);
int	find_H_partner(int	atom_id);
int 	identify_units(res_description_t ** descriptions, int nrofatoms, int *nr_of_neutrals);
void 	init_globals();
int 	build_units(res_description_t ** descriptions, abuf_entry_t **buffer);

/*************************************************************************/

static	int	nr_of_atoms=0;
			
static	int	*atom_2_unit;		/* -1 if atom not yet in a unit, otherwise  */
static	int	atom_2_unit_here=0;	/* index number of the unit	*/

static	int	*atom_charged;		/* 1 if atom is charged, otherwise 0 */

static	int	max_nr_units=0;		/* limit of possible units */

static	int	cur_nr_units=0;
static	int	cur_nr_neutral=0;
static	int	cur_nr_polar=0;
static	int	cur_maxsize_units=0;	/* current maximal number of atoms in a unit */

static	int	*atoms_in_unit;		/* number of atoms in unit */ 
static	int	*charges_in_unit;	/* number of charged atoms in unit */
static	int	*type_of_unit;		/* Tag to indetify chemical character of unit */
static	int	*character_of_unit;	/* NEUTRAL, POLAR or CHARGED */

/*********************************************************************************/


/*********************************************************************************/

int	writeout_units(char *unitfn)
{
	int i,j,check;
	FILE	*h_unit;

	/************** open lis-file *******************************************/
	if ((h_unit=fopen(unitfn,"w"))==NULL) {
		errout("Can not create unit-lis-file <%s>.\n",unitfn);
		return -1;
	}

	fprintf(h_unit,"This is the list of units.\n");
	fprintf(h_unit,"Number of units:    %4d\n",cur_nr_units);
	fprintf(h_unit,"Number of neutrals: %4d\n",cur_nr_neutral);
	fprintf(h_unit,"Number of polar:    %4d\n",cur_nr_polar);
	fprintf(h_unit,"Number of charged:  %4d\n",
				cur_nr_units-cur_nr_neutral-cur_nr_polar);
	fprintf(h_unit,"Max atoms per unit: %4d\n",cur_maxsize_units);
	fprintf(h_unit,"Unit-Nr  Nr-Atoms   Nr-Charged  Type  Character\n");

	for (i=0;i<cur_nr_units;i++) {	/* for every unit */

		fprintf(h_unit,"%4d      %4d      %4d      %4d      %4d\n      ",
			i,
			atoms_in_unit[i],
			charges_in_unit[i],
			type_of_unit[i],
			character_of_unit[i]);
	
		check=0;
		/******** first writeout the charged atoms ****************/
		for (j=0;j<nr_of_atoms;j++) {

			if (	(atom_2_unit[j]==i) &&
				(atom_charged[j]==1) ) {
				fprintf(h_unit,"%4d ",j);
				check++;
			}
		}
		/********* then writeout the uncharged atoms **************/
		for (j=0;j<nr_of_atoms;j++) {

			if (	(atom_2_unit[j]==i) &&
				(atom_charged[j]==0)  ) {
				fprintf(h_unit,"%4d ",j);
				check++;
			}
		}
		fprintf(h_unit,"\n");

		if (check!=atoms_in_unit[i]) {
			errout("Fatal error: Inconsistent unit found (Unit %d). Check code.\n",i);
			goto error_writeout;
		}
	}

	fclose(h_unit);
	txtout("Units-file written           : %d\n",cur_nr_units);
	return 0;

error_writeout:
	errout("Fatal error: Can't write unit-lis-file.\n");
	fclose(h_unit);
	return -1;
}

/*************************************************************************************/
/***** here we are warning the user of atoms, that do not belong to a unit. **********/
/***** This my occure 1. if wounits.c is not yet informed about the structures *******/
/***** 			 that appear in the given data. 			******/
/*****			 ===> work for you in wounits.c or path units.lis by hand ****/
/*****		      2. if the strings for identifying known structures are    ******/
/****			 different from what we assume.	(see definitions above)  *****/
/****			 ===> change definitions or PDB-PSF-files		******/
/****		      3. if there is a bug in wounits.c or other modules	 *****/
/****			 ===> work for you (or me)				******/
/*************************************************************************************/

int	print_restinfo()
{
	int i,count=0;
	STRUCT_ATOMDATA	atom;

	for (i=0;i<nr_of_atoms;i++) {
		if (atom_2_unit[i]==UNCHECKED) count++;	
	}
	if (count==0) return 0;

	errout("SEVERE WARNING ERROR: %d atom(s) do(es) not belong to a unit !\n",count);
	for (i=0;i<nr_of_atoms;i++) {
		if (atom_2_unit[i]==UNCHECKED) {
			if (get_pdb_atomdata(i,&atom)==-1)	return -1;	/* error */
			errout("EGO-Nr=%4d  Segi=<%s> Astr=<%s>  Resid-Nr=%4d <%s>  C=%f\n",
				atom.atomnr,
				atom.segidstr,
				atom.atomstr,
				atom.residnr,
				atom.residstr,
				atom.charge );
		}
	}
	errout("Every atom must belong to a unit! The reason for this\n");
	errout("error may be:\n");
	errout("1. The units definition file units.def does not know about\n");
	errout("   your molecule, because nobody did told it to the file.\n");
	errout("2. The strings for identifying known structures are different\n");
	errout("   from what is assumed in the units.def file.\n");
	errout("Check out for this problems and have a look at the manual (Section 4.2)\n");

	return -1;
}


/***************************************************************************/

static	int	charged_in_buffer=0;
static	int	charged_buffer[MAX_ATOMS_IN_UNIT/2+1];

static	int	neutral_in_buffer=0;
static	int	neutral_buffer[MAX_ATOMS_IN_UNIT/2+1];

static	double	total_charge=0.0;

static
int sort_to_buffer(int atom_nr)
{
	STRUCT_ATOMDATA	atom;

	if ((atom_nr<0)||(atom_nr>=nr_of_atoms)) {
		errout("Fatal error: Uncomplete or unknown structure.\n");
		return -1;
	}

	if (get_pdb_atomdata(atom_nr,&atom)==-1)	return -1;	/* error */

	if (	(atom.charge>SMALL) ||
		(atom.charge<-SMALL) ) {	/* save to charged buffer */

		charged_buffer[charged_in_buffer]=atom_nr;
		total_charge+=atom.charge;

		if (++charged_in_buffer>=MAX_ATOMS_IN_UNIT/2) {
			errout("Fatal error: Too many atoms in unit. Limit is 	%d\n",
				MAX_ATOMS_IN_UNIT);
			return -1;
		}

	} else {				/* save to neutral buffer */
		neutral_buffer[neutral_in_buffer]=atom_nr;

		if (++neutral_in_buffer>=MAX_ATOMS_IN_UNIT/2) {
			errout("Fatal error: Too many atoms in unit. Limit is %d\n",
				MAX_ATOMS_IN_UNIT);
			return -1;
		}
	}
	return 0;
}

static
int	reset_atombuffer()
{
	charged_in_buffer=0;
	neutral_in_buffer=0;
	total_charge=0.0;
	return 0;
}


/****************************************************************************/

static				/* here we write out the atoms for each list. */
int	save_unit(int tag)	/* first all charged atoms and then the neutral */
{				/* atoms for each unit */
	int	i;

	if (cur_nr_units>=max_nr_units) {
		errout("Fatal error: Too many units.\n");
		return -1;
	}

	if (charged_in_buffer+neutral_in_buffer==0) {
		errout("Fatal error: No atoms in unit. Check code.\n");
		return -1;
	}

	atoms_in_unit[cur_nr_units]	= charged_in_buffer+neutral_in_buffer;
	charges_in_unit[cur_nr_units]	= charged_in_buffer;
	type_of_unit[cur_nr_units]	= tag;

	if (charged_in_buffer==0) {
		character_of_unit[cur_nr_units]=NEUTRAL;
		cur_nr_neutral++;
	} else
	if ( 	(total_charge>SMALL) ||
		(total_charge<-SMALL) ) {

		character_of_unit[cur_nr_units]=CHARGED;
	} else {
		character_of_unit[cur_nr_units]=POLAR;
		cur_nr_polar++;
	}

	for (i=0;i<charged_in_buffer;i++) {
		atom_2_unit [charged_buffer[i]] = cur_nr_units;
		atom_charged[charged_buffer[i]] = 1;
	}		
	for (i=0;i<neutral_in_buffer;i++) {
		atom_2_unit[neutral_buffer[i]] = cur_nr_units; 		
	}		
	/* now each atom knows to which unit it belongs */	

	if (cur_maxsize_units<atoms_in_unit[cur_nr_units]) {
		cur_maxsize_units = atoms_in_unit[cur_nr_units];
	}
	cur_nr_units++;
	
	reset_atombuffer();	/* shure, is shure */
	return 0;
}


/********************************************************************
18.10.95 mcb: unit identification based on unit definition file units.def

*********************************************************************/


/******************************************************************************************
**	init_globals()
**	initialises all needed globals for building units
**	
*******************************************************************************************/

void init_globals()
{
	int	i;
	
	
	/******** the following code is inherited (by copy&paste) from scan4units() */
	

	/**************** allocate and ini arrays ********************/
	atom_2_unit=alloc_ints(nr_of_atoms+5);
	atom_charged=alloc_ints(nr_of_atoms+5);
	atoms_in_unit=alloc_ints(max_nr_units+5);
	type_of_unit=alloc_ints(max_nr_units+5);
	charges_in_unit=alloc_ints(max_nr_units+5);
	character_of_unit=alloc_ints(max_nr_units+5);

	for (i=0;i<nr_of_atoms;i++) {
		atom_2_unit[i] = UNCHECKED;
		atom_charged[i]= 0;
	}

	cur_nr_units=0;
	cur_nr_neutral=0;
	cur_nr_polar=0;

	for (i=0;i<max_nr_units;i++) {
		atoms_in_unit[i]	=-1;
		type_of_unit[i] 	=-1;
		charges_in_unit[i]	=-1;
		character_of_unit[i]	=-1;
	}

	atom_2_unit_here=1;
	
	/******** end of inheritance section */
}


/******************************************************************************************
**	find_H_partner()
**	returns the atom bondend to atom_id
**	
*******************************************************************************************/

int	find_H_partner(int	atom_id)
{
	int	bondl[MXNO_FIELDS_BOND+1];
	int	nr_bonds;

	nr_bonds=get_atom_bondlist(atom_id,bondl);
	if (nr_bonds>1) {
		return bondl[0];	 /* more than one bond to an H???? */
	} else if (nr_bonds==0)
		return -1;
        else
		return bondl[0];
}


/******************************************************************************************
**	sort_H_to_buffer()
**	sorts all H-atoms bonded to atom_id to the unit-buffer
**	
*******************************************************************************************/

void sort_H_to_buffer(abuf_entry_t **buffer, int atom_id)
{
	int	i;
	
	for (i = 0; buffer[i] != NULL; i++) {
		if (is_hydrogen(buffer[i])) {
			if (find_H_partner(buffer[i]->atom_id) == atom_id) {
				buffer[i]->assigned = 1;
				sort_to_buffer(buffer[i]->atom_id);
			}
		}
	}
}


/******************************************************************************************
**	build_units()
**	builds the units from the atoms in buffer, based on the informtion in descriptions
**	returns 0 if all atoms were assigned to a unit
*******************************************************************************************/

int build_units(res_description_t ** descriptions, abuf_entry_t **buffer)
{
	int	res_idx, alt_idx, unit_idx, atom_idx;
	int	atom_id;
	int	i;
	
	if (buffer[0] == NULL){
		printf("Fatal error in build_units(): empty buffer passed.\n");
		exit(-1);
	}
	
	res_idx = identify_residue(descriptions, buffer);
	
	alt_idx = identify_alternative(descriptions[res_idx], buffer);
	
	for (unit_idx = 0; descriptions[res_idx]->alternatives[alt_idx]->units[unit_idx] != NULL; unit_idx++) {
		for (atom_idx = 0; descriptions[res_idx]->alternatives[alt_idx]->units[unit_idx]->atoms[atom_idx] != NULL; atom_idx++) {
			atom_id = get_and_mark(buffer,descriptions[res_idx]->alternatives[alt_idx]->units[unit_idx]->atoms[atom_idx]);
			if (atom_id <0) {
				printf("Fatal error in build_units(): didn't find atom <%s> in buffer for residue <%s> (%d).\n",
					descriptions[res_idx]->alternatives[alt_idx]->units[unit_idx]->atoms[atom_idx],
					descriptions[res_idx]->name,res_idx);
				return -1;
			}

			sort_to_buffer(atom_id);
			
			/* now we handle the hydrogens */
			sort_H_to_buffer(buffer, atom_id);
			
		}
		save_unit(0);	/* no tags */
	}
	
	/* sanity check */
	for (i = 0; buffer[i] != NULL; i++)
		if (buffer[i]->assigned == 0) {
			printf("Fatal error in build_units(): not all atoms of resid <%s>\n",
				buffer[i]->res_name);
			printf("have been assigned to a unit\n");
                        printf("Namley atom <%s> with number %d\n",
				buffer[i]->atom_name,buffer[i]->atom_id+1);
			return -1;
		}
	
	return 0;
}
	

/******************************************************************************************
**	identify_units()
**	creates units based on unit.def 
**	and returns the number of units created
**	ASUMPTION: atoms belonging to one residue are in sequence in pdb_atomdata
*******************************************************************************************/
					
int identify_units(	res_description_t ** descriptions, 
			int nrofatoms, 
			int *nr_of_neutrals)
{
	int				i;
	int				residue_nr;
	STRUCT_ATOMDATA	atom;
	abuf_entry_t	**buffer;		/* NULL terminated buffer of atoms */
	
	/******** the following code is inherited (by copy&paste) from scan4units() */
	
	nr_of_atoms=nrofatoms;
	max_nr_units=max(nr_of_atoms/UNIT_FACTOR,10);
	init_globals();
	
	/* end of inheritance */
	
	buffer = (abuf_entry_t **) malloc(sizeof(abuf_entry_t *));
	buffer[0] = NULL;
	
	for (i = 0, get_pdb_atomdata(i,&atom), residue_nr = atom.residnr; i < nr_of_atoms; i++) {
		
		if (get_pdb_atomdata(i,&atom) == -1) return -1;

		if (residue_nr != atom.residnr){
			if (build_units(descriptions, buffer) != 0) {
                                printf("This error occured in resid-nr %d\n",
					residue_nr);
				return -1;
			}
			buffer = free_buffer(buffer);
			residue_nr = atom.residnr;
		}
		
		buffer = add_to_buffer(buffer, &atom);
	}
	
	/* one more time with the remaining atoms */
	if (build_units(descriptions, buffer) != 0) 
		return -1;
	buffer = free_buffer(buffer);


	/* now we assign the remaining protons to units */
	
	*nr_of_neutrals = cur_nr_neutral;	/* once again copy&paste inheritance */
	return cur_nr_units;
}



