/*
#    Molecular Dynamics program EGO
#    Copyright (C) 1989  Markus Eichinger, Helmut Grubmueller, and Helmut Heller
#
#    official EGO web site: http://www.lrz.de/~heller/ego/
#
#    Contact the authors here:
#    Dr. Markus Eichinger:   entwicklung@eichinger-net.de
#    Dr. Helmut Grubmueller: hgrubmu@gwdg.de
#    Dr. Helmut Heller:      heller@lrz.de
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#    For the full text of the license, see file LICENSE
#
#########################################################################
*/

/*************************************************************************
**									**
**	Module	: wovdw.c	by M.Eichinger, Munich			**
**	Belongs2: xpl2lis	Converter Xplor-data to EGO-data (*.lis)**
**	Contents: scan in and write out the vdw-param's and the 	**
**		  vdw-14-param's given in Parameter-files by 		**	
** 		  NONB-Param's and NBFIX-param's			**								**										**
**	Remarks: 							**	
**		  not yet checked, whether a matrix-element is changed  **
**		  more then one time	by a NBFIX-Statement.		**										**										**
**  	Modif's:							**
**		05.06.94						**
**									**
*************************************************************************/

#include <math.h>

#include "hardw.h"
#include "ego.h"
#include "woalloc.h"
#include "util_lib.h"
#include "wodecls.h"

#define TESTNBFIX	0
#define TESTVDW		0

/******************************************************************************/
/**************** functions in this modul *************************************/
/******************************************************************************/

	int	readin_vdw_params(int argc,char *argv[],int nr_atomtypes);
	int	writeout_vdwparams(char *vdwparafn,char *vdw14parafn);

static	int	fill_vdwparams(FILE *h_para);
static	int	respect_nbfix(FILE *h_para);
static	int	free_all_buffers(void);
static	int	alloc_all_buffers(int nr_atomtypes);
static	double	sigmaeps2a(double sigma,double eps);
static	double	sigmaeps2b(double sigma,double eps);

/*****************************************************************************/

static	double	*sigma,*eps;
static	double	*sigma14,*eps14;
static	int	*value_present;
static	double	*vdwmat,*vdw14mat;

static	int	nr_atomtypes=0;
static	int	nr_valid_nbfix=0;

/*****************************************************************************/

int	readin_vdw_params(int argc,char *argv[],int nratt)
{
	int	file_ind;
	int	i,j;
	FILE	*h_para;

	nr_valid_nbfix=0;
	nr_atomtypes=nratt;
	if (alloc_all_buffers(nr_atomtypes)==-1)	return -1;
		
	/******************* scan each parmater-file for nonbond-params *****/
	for (file_ind=3;file_ind<argc;file_ind++) {
		if ((h_para=fopen(argv[file_ind],"r"))==NULL) {
			errout("Can not read parameter-file: <%s>\n",argv[file_ind]);
			free_all_buffers();
			return -1;
		}

		if (fill_vdwparams(h_para)==-1) {
			errout("Error occured in parameter-file: <%s>\n",argv[file_ind]);
			free_all_buffers();
			fclose(h_para);
			return -1;
		}
		
		fclose(h_para);
	}

	/******************** check, whether each data is here ***************/
	for (i=0;i<nr_atomtypes;i++) {
		if (!value_present[i]) {
			errout("Missing VDW-Data for atomtype %d.\n",i);
			free_all_buffers();
			return -1;
		}
	}
	txtout("All NONBonded params read (VDW).\n");

	/************************ fill the vdw-matrix ************************/
	for (i=0;i<nr_atomtypes;i++) {

		for (j=0;j<=i;j++) {
			if (i==j) {	/* masses in diag */
				vdwmat  [i*(nr_atomtypes+1)+j]=get_atom_mass(i);
				vdw14mat[i*(nr_atomtypes+1)+j]=(double) 0.0;
			} else {	/*** convert sigma,eps ==> a,b *****/
					/*** and fill the matrix ***********/
				double	sig_ij	=(sigma[i]+sigma[j])/2.0;
				double	eps_ij	=sqrt(eps[i]*eps[j]);			
			
				vdwmat[i*(nr_atomtypes+1)+j]=sigmaeps2b(sig_ij,eps_ij);
				vdwmat[j*(nr_atomtypes+1)+i]=sigmaeps2a(sig_ij,eps_ij);
				
				sig_ij	=(sigma14[i]+sigma14[j])/2.0;
				eps_ij	=sqrt(eps14[i]*eps14[j]);			
		
				vdw14mat[i*(nr_atomtypes+1)+j]=sigmaeps2b(sig_ij,eps_ij);
				vdw14mat[j*(nr_atomtypes+1)+i]=sigmaeps2a(sig_ij,eps_ij);	
			}
		}
		/******* WW between same atomtypes in last row and line *********/ 	
		vdwmat[nr_atomtypes*(nr_atomtypes+1)+i]=sigmaeps2b(sigma[i],eps[i]);
		vdwmat[i*(nr_atomtypes+1)+nr_atomtypes]=sigmaeps2a(sigma[i],eps[i]);

		vdw14mat[nr_atomtypes*(nr_atomtypes+1)+i]=sigmaeps2b(sigma14[i],eps14[i]);
		vdw14mat[i*(nr_atomtypes+1)+nr_atomtypes]=sigmaeps2a(sigma14[i],eps14[i]);
	}

	/***** bottom-right cell not used ***********************/	
	vdwmat  [nr_atomtypes*(nr_atomtypes+1)+nr_atomtypes]=0.0;
	vdw14mat[nr_atomtypes*(nr_atomtypes+1)+nr_atomtypes]=0.0;

	/******************* scan each parmater-file for nbfix-params ********/
	for (file_ind=3;file_ind<argc;file_ind++) {
		if ((h_para=fopen(argv[file_ind],"r"))==NULL) {
			errout("Can not read parameter-file: <%s>\n",argv[file_ind]);
			free_all_buffers();
			return -1;
		}

		if (respect_nbfix(h_para)==-1) {
			errout("Error occured in parameter-file: <%s>\n",argv[file_ind]);
			free_all_buffers();
			fclose(h_para);
		}
		
		fclose(h_para);
	}

	txtout("NBFIX-Statments respected    : %d\n",nr_valid_nbfix);

	return 0;
}

/*****************************************************************************/

int	writeout_vdwparams(char *vdwparafn,char *vdw14parafn)
{
	FILE	*h_vdw,*h_vdw14;
	int	i,j;

		/***************** write vdw-params to file ******************/
	if ((h_vdw=fopen(vdwparafn,"w"))==NULL) {
		errout("Fatal error: Can not create VDW-parameter-file <%s>.\n",vdwparafn);
		free_all_buffers();
		return -1;
	}
	
	fprintf(h_vdw,"This is the van der Waals matrix with the masses in the diagonal.\n");
	fprintf(h_vdw,"Number of atom typs encountered:   %4d\n",nr_atomtypes);

	for (i=0;i<=nr_atomtypes;i++) {
		for (j=0;j<=nr_atomtypes;j++) {
			if (	((j+1)%MXNO_FIELDS_VDW==0) ||
				(j==nr_atomtypes)	) {
				fprintf(h_vdw,"%11.9g\n",vdwmat[i*(nr_atomtypes+1)+j]);
			} else {
				fprintf(h_vdw,"%11.9g  ",vdwmat[i*(nr_atomtypes+1)+j]);
			}
		}
		fprintf(h_vdw,"\n");
	}
	fclose(h_vdw);
	txtout("VDW-parameter-file written.\n");

			/***************** write vdw14-params to file ********/
	if ((h_vdw14=fopen(vdw14parafn,"w"))==NULL) {
		errout("Fatal error: Can not create 14-VDW-parameter-file <%s>.",vdw14parafn);
		free_all_buffers();
		return -1;
	}
	
	fprintf(h_vdw14,"This is the 1-4 van der Waals matrix.\n");
	fprintf(h_vdw14,"Number of atom typs encountered:   %4d\n",nr_atomtypes);

	for (i=0;i<=nr_atomtypes;i++) {
		for (j=0;j<=nr_atomtypes;j++) {
			if (	((j+1)%MXNO_FIELDS_VDW==0) ||
				(j==nr_atomtypes)	) {
				fprintf(h_vdw14,"%11.9g\n",vdw14mat[i*(nr_atomtypes+1)+j]);
			} else {
				fprintf(h_vdw14,"%11.9g  ",vdw14mat[i*(nr_atomtypes+1)+j]);
			}
		}
		fprintf(h_vdw14,"\n");
	}
	fclose(h_vdw14);
	txtout("VDW-14-parameter-file written.\n");
	return 0;
}

/*****************************************************************************/

static	
double	sigmaeps2a(double sig,double eps)
{
	double	sig_6	=sig*sig*sig*sig*sig*sig;
	double	sig_12	=sig_6*sig_6;

	double	a	=4.0*sig_12*eps;

	return a;
}

static
double	sigmaeps2b(double sig, double eps)
{
	double	sig_6	=sig*sig*sig*sig*sig*sig;
	double	b	=4.0*sig_6*eps;

	return b;
}

/*****************************************************************************/

static
int	fill_vdwparams(FILE *h_para)
{
	double 	Sigma,Sigma14,Eps,Eps14;
	int	att;	/* atom-type-index ! */
	int	res;

	while ((res=read_next_vdwpara(h_para,	&att,
						&Sigma,&Eps,
						&Sigma14,&Eps14))>=0) {

		if (res==0)	continue;	/* found atom-type which don't occur in */
						/* PDB/PSF-file, no error, just skip it */

		if (value_present[att]) {
			if (	(fabs(sigma[att]-Sigma)>SMALL)		||
				(fabs(eps[att]-Eps)>SMALL)		||
				(fabs(sigma14[att]-Sigma14)>SMALL) 	||
				(fabs(eps14[att]-Eps14)>SMALL)	 ) {
				errout("Warning error: Inconsistent NONBond-parameters for atomtype %d.\n",att);
				errout("old vals     (sigma,eps, sigma14,eps14): (%f, %f, %f, %f)\n",
					sigma[att],eps[att],sigma14[att],eps14[att]);
				errout("skipped vals (sigma,eps, sigma14,eps14): (%f, %f, %f, %f)\n",
					Sigma,Eps,Sigma14,Eps14);
				errout("Skipping new NONBond-parameters !\n");
			}
		} else {
			sigma[att]	=	Sigma;
			eps[att]	=	Eps;
			sigma14[att]	=	Sigma14;
			eps14[att]	=	Eps14;			
			value_present[att]=1;
#if TESTVDW
			txtout("TEST: VDW-Params:%d   %lf  %lf      %lf   %lf\n",
				att,Sigma,Eps,Sigma14,Eps14); 
#endif
		}

	}
	if (res==-2) {
		errout("Fatal error: illegal parameter-line while reading vdw-parameters.\n");
		return -1;		/* thats an error */
	}
	return 0;
}

/*********************************************************************************/

static
int	respect_nbfix(FILE *h_para)
{
	double 	a,a14,b,b14;
	int	att1,att2;	/* atom-type-index ! */
	int	res;
	int	i,j;

	while ((res=read_next_nbfixpara(h_para,	&att1,&att2,
						&a,&b,
						&a14,&b14))>=0) {
#if TESTNBFIX
		printf("TEST: NBFIX:%d %d  %lf  %lf      %lf   %lf\n",
				att1,att2,a,b,a14,b14); 
#endif
		if (res==0)	continue; /*found atom-type which don't occur in*/
					  /*PDB/PSF-file, no error, just skip it*/

/****** not yet checked, whether a matrix-element is changed more then one time *****/	

		if (att1==att2)	{
			i=nr_atomtypes;
			j=att1;
		} else {
			j=min(att1,att2);
			i=max(att1,att2);
		}

		/******** change now the vdw matrix with this new values ***********/
		vdwmat[i*(nr_atomtypes+1)+j]	=b;
		vdwmat[j*(nr_atomtypes+1)+i]	=a;

		/******** change now the vdw14 matrix with this new values *********/
		vdw14mat[i*(nr_atomtypes+1)+j]	=b14;
		vdw14mat[j*(nr_atomtypes+1)+i]	=a14;
		
		nr_valid_nbfix++;
	}
	if (res==-2) {
		errout("Fatal error: illegal parameter-line while reading nbfix-statements.\n");
		return -1;		/* thats an error */
	}
	return 0;
}

/**********************************************************************************/

static
int	alloc_all_buffers(int nr_atomtypes)
{
	int	i;

	sigma	= alloc_doubles(nr_atomtypes+1);
	sigma14	= alloc_doubles(nr_atomtypes+1);
	eps	= alloc_doubles(nr_atomtypes+1);
	eps14	= alloc_doubles(nr_atomtypes+1);
	value_present=alloc_ints(nr_atomtypes+1);
	vdwmat  = alloc_doubles((nr_atomtypes+1)*(nr_atomtypes+1));
	vdw14mat= alloc_doubles((nr_atomtypes+1)*(nr_atomtypes+1));
	for (i=0;i<nr_atomtypes;i++) {
		sigma[i]=(double)0.0;
		sigma14[i]=(double)0.0;
		
		eps[i]=(double)0.0;
		eps14[i]=(double)0.0;

		value_present[i]=0;	/* 0 = no value present, 1 = value present*/
	}

	return 0;
}

/**********************************************************************************/

static
int	free_all_buffers(void)
{
	myfree(sigma);
	myfree(sigma14);
	myfree(eps);
	myfree(eps14);
	myfree(value_present);
	myfree(vdwmat);
	myfree(vdw14mat);
	return 0;
}
